#ifndef LANG_CORE_TYPES_HPP
#define LANG_CORE_TYPES_HPP

#include <memory>
#include <boost/exception/all.hpp>
#include <boost/optional.hpp>
#include <typeinfo>
#include <type_traits>

namespace leaf {

typedef char src_char_t;
typedef int src_symbol_t;

} //eon leaf

using std::unique_ptr;
using std::shared_ptr;
using std::make_shared;

namespace error {
	//generic reason tag for the error
	typedef boost::error_info<struct error_reason,std::string> tag_error_reason;
	//condition which failed, if based on a condition
	typedef boost::error_info<struct error_condition,std::string> tag_error_condition;
	//the class based on a typeinfo structure related to the error
	typedef boost::error_info<struct error_class,std::string> tag_error_class;
	//primary symbol is related to a symbol
	typedef boost::error_info<struct error_symbol,std::string> tag_error_symbol;
	
	template<typename T>
	struct class_tag { };

	/*
		The variadic template allows a very simple reusable macro that can be used with
		just a condition, with a message, and optional additional exception parameters.
	*/
	
	template<typename X>
	void extend_exception( X & /*ex*/ )
	{ }

	//generic fallback
	//refer: http://mortoray.com/2013/06/03/overriding-the-broken-universal-reference-t/
	template<typename X, typename T, typename Tag>
	void extend_exception_value( X & ex, T&& value, Tag ) {
		ex << value;
	}
	
	template<typename X, typename TF>
	void extend_exception_value( X & ex, TF && value, class_tag<std::string> ) { 
		ex << tag_error_reason( value ); 
	}
	
	template<typename X, typename TF>
	void extend_exception_value( X & ex, TF && ti, class_tag<std::type_info> ) {
		ex << tag_error_class( ti.name() );
	}

	template<typename X, typename TF>
	void extend_exception_value( X & ex, TF && value, class_tag<const char*> ) {
		ex << tag_error_reason( value );
	}

	template<typename X, typename TF, typename ... TR>
	void extend_exception( X & ex, TF&& f, TR&& ... rest ) {
		class_tag<typename std::decay<TF>::type> tag;
		extend_exception_value( ex, std::forward<TF>(f), tag );
		extend_exception( ex, std::forward<TR>(rest)... );
	}

	template<typename X, typename ... TA>
	X create_exception( TA&& ... all ) {
		X ex;
		extend_exception( ex, std::forward<TA>(all)... );
		return ex;
	}
} //eon error
	
/**
	Recoverable errors are thrown when the internal state has been corrupted
	and the throwing object chain is no longer valid.
	
	Refer to:
	http://mortoray.com/2013/01/03/how-to-handle-an-error/
*/
struct recoverable_error : virtual boost::exception, virtual std::exception
{
	recoverable_error() { }
	recoverable_error( char const * msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	recoverable_error( std::string const & msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	recoverable_error( recoverable_error const & o )
		: boost::exception( o ), std::exception( o )
	{ }
};

/**
	Unwindable errors are thrown when the internal state has not been changed
	and the objects can continue to be used.
	
	It is derived from a recoverable_error since a recovery catch can deal with these
	whereas an unwindable catch cannot deal with recoverable.
*/
struct unwindable_error : public recoverable_error {
	unwindable_error() { }
	unwindable_error( char const * msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	unwindable_error( std::string const & msg )
	{
		(*this) << error::tag_error_reason( msg );
	}
	unwindable_error( unwindable_error const & o )
		: boost::exception( o ), std::exception( o ), recoverable_error()
	{ }
};

#define STATE_CHECK( cond_, ... )	\
	if( !(cond_) ) { BOOST_THROW_EXCEPTION( \
		error::create_exception<recoverable_error>( __VA_ARGS__ ) \
		<< error::tag_error_condition( #cond_ ) ); }
#define STATE_FAIL( ... )	\
	BOOST_THROW_EXCEPTION( \
		error::create_exception<recoverable_error>( __VA_ARGS__ )  )
	
#define PRE_CHECK( cond_, ... )	\
	if( !(cond_) ) { BOOST_THROW_EXCEPTION( \
		error::create_exception<unwindable_error>( __VA_ARGS__ ) \
		<< error::tag_error_condition( #cond_ ) ); }
#define PRE_FAIL( ... )	\
	BOOST_THROW_EXCEPTION( \
		error::create_exception<unwindable_error>( __VA_ARGS__ )  )
	
#endif