#ifndef LEAF_IR_BUILDER_HPP
#define LEAF_IR_BUILDER_HPP

#include "language.hpp"
#include "platform/platform.hpp"
#include "type_manager.hpp"

namespace ir
{

struct function_builder;
struct switch_builder;
struct builder;

struct module_builder
{
	compiler_context * cctx;

	explicit module_builder( compiler_context & cctx )
		: cctx( &cctx )
	{ }
	
	value_defn define_value( value::form_t form, shared_ptr<type const> direct_type, 
		std::string const & name ) {
		auto use_value = make_shared<value>( form );
		use_value->direct_type = direct_type;
		use_value->name = name;
		return use_value;
	}
	
	shared_ptr<value> create_local( shared_ptr<type const> direct_type, std::string const & name ) {
		auto l = make_shared<value>( value::f_local );
		l->name = name;
		l->direct_type = direct_type;
		return l;
	}
	
	function_builder create_function( std::string const & name, bool can_fail );
	function_builder create_function( std::string const & name, shared_ptr<type const> func_type,
		shared_ptr<value> use_value = shared_ptr<value>() );
		
protected:
	type_manager & get_types() {
		return cctx->get_types();
	}
};

struct function_builder : public module_builder
{
	function * owner_function;
	shared_ptr<type> mutable_type; //may be null
	
	explicit function_builder( function & owner_function, shared_ptr<type> mutable_type ) : 
		module_builder( owner_function.cctx )	, 
		owner_function( &owner_function ), 
		mutable_type( mutable_type ) {
	}
	
	shared_ptr<function> get_function() { 
		return owner_function->shared_from_this(); 
	}
	
	value_defn get_value_defn() {
		return owner_function->defn_value; 
	}
	
	value_defn define_arg( shared_ptr<type const> dt, std::string const & name ) {
		auto vd = define_value( value::f_arg_ref, dt, name );
		owner_function->args.push_back( vd );
		if( mutable_type )
			mutable_type->args.push_back( dt );
		//TODO: else check type match
		return vd;
	}
	
	value_defn define_return( shared_ptr<type const> dt, std::string const & name ) {
		auto vd = define_value( value::f_return_ref, dt, name );
		vd.origin->anon_ref = owner_function->returns.size();
		owner_function->returns.push_back( vd );
		if( mutable_type )
			mutable_type->sub.push_back( dt );
		//TODO: else check type match
		return vd;
	}
	
	value_defn define_context_arg( shared_ptr<type const> direct_type, std::string const & name ) {
		auto vd = define_value( value::f_capture_ref, direct_type, name );
		owner_function->context_args.push_back( vd );
		return vd;
	}
};

struct funccall_builder
{
	shared_ptr<funccall_inst> inst;
	block * where;
	
	funccall_builder( shared_ptr<funccall_inst> inst, block & where )
		: inst( inst ), where( &where )
	{ }
	
	void add_arg( value_ref p ) {
		inst->args.push_back( p );
	}
	
	void set_context( value_ref ctx ) {
		inst->context = ctx;
	}
	
	value_defn define_return( shared_ptr<type const> t ) {
		auto ref = where->make_temp( t );
		inst->returns.push_back( value_defn(ref) );
		return ref;
	}
	
	//TODO: a generic "inst" base to handle this
	void add() { 
		where->push_back( inst, block_flow::sequential ); 
	}
	void defer() { 
		where->push_back( inst, block_flow::defer ); 
	}
	void defer_fail() { 
		where->push_back( inst, block_flow::defer_fail ); 
	}
	void push_back( block_flow flow ) { 
		where->push_back( inst, flow ); 
	}
};

/**
	The builder simplifies the creation of the IR elements. This helps reduce the amount of
	code in the compiler thus not hiding the important details.
	
	NOTE: This class does not maintain any IR state of its own and is thus safe to instantiate
	and discard without any loss.
*/
struct builder : public module_builder
{
	//which primary block is being built
	block * where;
	//special block where defer's should put placed
	block * split_defer;
	
	function & get_owner_function() {
		return where->get_owner_function();
	}
	
	explicit builder( block & where ) : 
		module_builder( where.get_owner_function().cctx ), 
		where( &where ), 
		split_defer( nullptr ) { 
	}
	explicit builder( function & func ) : 
		module_builder( func.cctx ), 
		where( func.entry_block.get() ), 
		split_defer( nullptr ) { 
	}
	explicit builder( function_builder & fb ) : 
		module_builder( *fb.cctx ), 
		where( fb.owner_function->entry_block.get() ), 
		split_defer( nullptr ) { 
	}

	template<typename T>
	void push_back( T const & i, block_flow flow ) { 
		if( flow == block_flow::defer && split_defer ) {
			split_defer->push_back( i, flow );
		} else {
			where->push_back( i, flow ); 
		}
	}
	template<typename T>
	void add( T const & i ) { 
		push_back(i, block_flow::sequential ); 
	}
	template<typename T>
	void defer( T const & i ) { 
		push_back(i, block_flow::defer );
	}
	template<typename T>
	void defer_fail( T const & i ) { 
		push_back(i, block_flow::defer_fail ); 
	}

	shared_ptr<store_inst> add_store_inst( value_ref left, value_ref right )
	{
		auto si = make_shared<store_inst>();
		si->left = left;
		si->right = right;
		add(si);
		return si;
	}
	
	block & get_block() { 
		return *where; 
	}
	
	shared_ptr<value> make_temp( shared_ptr<type const> dt ) { 
		return where->make_temp( dt ); 
	}

	//TODO: should return block builders...
	shared_ptr<block> create_block() { 
		return where->create_child_block();
	}
	
	shared_ptr<block> insert_block()
	{ return where->insert_block(); }
	
	value_defn define_local( shared_ptr<type const> direct_type, std::string const & name )
	{
		auto l = make_shared<value>( value::f_local );
		l->name = name;
		l->direct_type = direct_type;
		get_owner_function().locals.push_back( l );
		return value_defn(l);
	}
	
	shared_ptr<bind_raw_func_inst> add_bind_raw_func_inst( value_ref raw_func )
	{
		auto wf = make_shared<bind_raw_func_inst>();
		wf->raw_func = raw_func;
		auto rt = raw_func.result_type()->clone();
		rt->fundamental = type::f_function_ptr;
		rt->access = type::a_bound;
		wf->result = where->make_temp(rt);
		add(wf);
		return wf;
	}
	
	shared_ptr<alloc_array_inst> add_alloc_array_inst( shared_ptr<type const> array_type,
		shared_ptr<type const> item_type, value_ref length )
	{
		auto aa = make_shared<alloc_array_inst>();
		aa->result = where->make_temp( array_type );
		aa->item_type = item_type;
		aa->length = length;
		add( aa );
		return aa;
	}
	
	shared_ptr<free_array_inst> create_free_array_inst( value_ref array )
	{
		auto fa = make_shared<free_array_inst>();
		fa->array = array;
		return fa;
	}
	
	shared_ptr<pack_tuple_inst> create_pack_tuple_inst( shared_ptr<type const> result_type ) {
		auto pt = make_shared<pack_tuple_inst>();
		pt->result = where->make_temp(result_type);
		//this is a "create" function so don't add it to the block
		return pt;
	}
	
	switch_builder create_switch( shared_ptr<type const> result_type );
	switch_builder create_switch();

	funccall_builder create_funccall_inst( value_ref call )
	{
		auto fc = make_shared<funccall_inst>();
		fc->call = call;
		return funccall_builder(fc, *where);
	}
	funccall_builder add_funccall_inst( value_ref call )
	{
		auto fc = make_shared<funccall_inst>();
		fc->call = call;
		add( fc );
		return funccall_builder(fc, *where);
	}
	
	shared_ptr<bind_func_inst> create_bind_func_inst( value_ref ctx_func,
		shared_ptr<type const> data_type )
	{
		auto ci = make_shared<bind_func_inst>();
		ci->ctx_func = ctx_func;
		ci->ctx_data.direct_type = data_type;
		ci->ctx_data.result = where->make_temp( get_types().get_raw_pointer( ci->ctx_data.direct_type ) );
		
		auto bound_func = ctx_func.result_type()->clone();
		bound_func->access = type::a_bound;
		ci->result = where->make_temp( bound_func );
		return ci;
	}
	
	shared_ptr<bind_func_explicit_inst> add_bind_func_explicit_inst( 
		value_ref ctx_func, value_ref ctx_data )
	{
		auto ci = make_shared<bind_func_explicit_inst>();
		ci->ctx_func = ctx_func;
		ci->ctx_data = ctx_data;
		
		auto bound_func = ctx_func.result_type()->clone();
		bound_func->access = type::a_bound;
		ci->result = where->make_temp( bound_func );
		add(ci);
		return ci;
	}
	
	shared_ptr<void_context_inst> create_void_context_inst( shared_ptr<type const> t )
	{
		auto vc = make_shared<void_context_inst>();
		vc->result = where->make_temp( t );
		return vc;
	}
	
	shared_ptr<release_shared_inst> create_release_shared_inst( value_ref object )
	{
		auto rs = make_shared<release_shared_inst>();
		rs->object = object;
		return rs;
	}

	shared_ptr<acquire_shared_inst> create_acquire_shared_inst( value_ref object )
	{
		auto rs = make_shared<acquire_shared_inst>();
		rs->object = object;
		return rs;
	}

	value_defn const_integer( shared_ptr<type const> t, intmax_t val ) {
		auto v = make_shared<value>(value::f_const_integer);
		v->int_value = val;
		v->direct_type = t;
		return v;
	}
	
	value_defn const_binary( shared_ptr<type const> t, intmax_t val ) {
		auto v = make_shared<value>(value::f_const_binary);
		v->uint_value = val;
		v->direct_type = t;
		return v;
	}
	
	value_defn const_integer_normal( int val ) //of native integer type
	{
		auto t = make_shared<type>( get_types().shared_from_this(), type::f_integer );
		t->trivial_bitsize = platform::natural_integer_size;
		return const_integer( t, val );
	}
	
	value_defn const_float( shared_ptr<type const> t, double val )
	{
		auto v = make_shared<value>(value::f_const_float);
		v->float_value = val;
		v->direct_type = t;
		return v;
	}

	value_defn const_boolean( bool val )
	{
		auto v = make_shared<value>( value::f_const_boolean );
		v->boolean_value = val;
		v->direct_type = get_types().get_boolean();
		return v;
	}
	
	value_defn add_special_field( special_field_inst::field_t field, shared_ptr<type const> field_type, 
		value_ref src )
	{
		auto sf = make_shared<special_field_inst>();
		sf->field = field;
		sf->object = src;
		sf->result = make_temp( field_type );
		add( sf );
		return sf->result;
	}
	
	value_defn add_binary( binary_inst::op_t op, value_ref a, value_ref b, shared_ptr<type const> ret ) {
		auto bin = make_shared<binary_inst>();
		bin->op = op;
		bin->a = a;
		bin->b = b;
		bin->result = make_temp( ret );
		add( bin );
		return bin->result;
	}

	value_defn add_unary( unary_inst::op_t op, value_ref arg, shared_ptr<type const> ret ) {
		auto un = make_shared<unary_inst>();
		un->op = op;
		un->arg = arg;
		un->result = make_temp( ret );
		add( un );
		return un->result;
	}
	
	value_ref builtin_function_unary( std::string const & name,
		shared_ptr<type const> a, shared_ptr<type const> ret= shared_ptr<type const>() ) {
		auto f = make_shared<value>( value::f_import_code );
		f->name = name;
		auto dt = make_shared<type>( get_types().shared_from_this(), type::a_raw );
		f->direct_type = dt;
		if( ret )
			dt->sub.push_back( ret );
		dt->args.push_back( a );
		return value_ref::as_value(f);
	}
	
	value_ref call_builtin_unary( std::string const & name,
		value_ref a, shared_ptr<type const> ret = shared_ptr<type const>() ) {
		auto f = builtin_function_unary( name, a.result_type(), ret );
		auto c = create_funccall_inst( f );
		c.add_arg( a );
		c.add();
		if( ret )
			return c.define_return( ret );		
		return value_ref();
	}
	
	value_defn add_array_offset( value_ref array, value_ref offset ) {
		auto ao = make_shared<array_offset_inst>();
		ao->array = array;
		ao->offset = offset;
		
		auto arr_type = array.result_type();
		if( arr_type->is_raw_pointer()) {
			//assume pointer to array
			ao->result = where->make_temp( get_types().get_raw_pointer( arr_type->item_type()->item_type() ) );
		} else {
			ao->result = where->make_temp( arr_type->item_type() );
		}
			
		add( ao );
		return ao->result;
	}

	value_defn add_array_ptr_offset( value_ref array, value_ref offset, bool lvalue ) {
		auto ao = make_shared<array_ptr_offset_inst>();
		ao->array = array;
		ao->offset = offset;
		ao->lvalue = lvalue;
// 		if( lvalue ) {
// 			ao->result = where->make_temp( get_types().get_raw_pointer( array.result_type()->item_type() ) );
// 		} else {
			ao->result = where->make_temp( array.result_type()->item_type() );
// 		}
		add( ao );
		return ao->result;
	}

	shared_ptr<loop_inst> add_loop()
	{
		auto loop = make_shared<loop_inst>( get_block() );
		add( loop );
		return loop;
	}
	
	value_defn add_promote_loss( bool loss, value_ref src, shared_ptr<type const> to_type, 
		promote_loss_inst::what_t from_what, promote_loss_inst::what_t to_what )
	{
		shared_ptr<promote_loss_inst> pi;
		if( loss )
			pi = make_shared<loss_inst>();
		else
			pi = make_shared<promote_inst>();
		pi->source = src;
		pi->result = where->make_temp( to_type );
		pi->to_what = to_what;
		pi->from_what = from_what;
		pi->to_bitsize = to_type->trivial_bitsize;
		add( pi );
		return pi->result;
	}

	value_defn add_clone( value_ref source )
	{
		auto ci = make_shared<clone_inst>();
		ci->source = source;
		ci->result = where->make_temp( ci->source.result_type() );
		add( ci );
		return ci->result;
	}
	
	value_defn add_cast( value_ref source, shared_ptr<type const> to_type )
	{
		auto ci = make_shared<cast_inst>();
		ci->source = source;
		ci->result = where->make_temp( to_type );
		add(ci);
		return ci->result;
	}
	
	value_defn add_clone_value( value_ref source )
	{
		auto ci = make_shared<clone_inst>();
		ci->source_value = true;
		ci->source = source;
		ci->result = where->make_temp( get_types().get_raw_pointer( ci->source.result_type() ) );
		add( ci );
		return ci->result;
	}
	
	value_defn resolve( value_ref cur ) 
	{
		//if already a noref_value then do nothing
		if( cur.is_noref_value() )
			return value_defn( cur.origin );
			
		auto cx = make_shared<resolve_inst>();
		cx->source = cur;
		cx->result = value_defn( make_temp( cur.result_type() ) );
		add( cx );
		
		return cx->result;
	}
	
	value_ref access_pre_field( value_ref cur, size_t index )
	{
		if( !cur.pre_field ) {
			cur.pre_field = index;
			return cur;
		}

		value_ref vr = resolve( cur );
		vr.pre_field = index;
		return vr;
	}

	value_ref access_post_field( value_ref cur, size_t index )
	{
		if( !cur.pre_field && cur.how == value_ref::h_value ) {
			cur.pre_field = index;
			return cur;
		}
		
		value_ref vr = resolve( cur );
		vr.pre_field = index;
		return vr;
	}

	value_defn create_optional( value_ref value, shared_ptr<type const> result_type )
	{
		auto pt = create_pack_tuple_inst( result_type );
		pt->sub.push_back( const_boolean( true ) );
		pt->sub.push_back( value );
		add( pt );
		return pt->result;
	}
	
	value_ref access_optional_set( value_ref cur )
	{
		//TODO: does this belong at this level, perhaps better in compiler_leaf.cpp
		//this supports that an optional lvalue can be derefed...
		if( cur.result_type()->is_raw_pointer() ) 
			return access_post_field( access_deref( cur ), 0 );
		return access_post_field( cur, 0 );
	}

	value_ref access_optional_value( value_ref cur )
	{
		//...for an lvalue this also results in an lvalue
		return access_post_field( cur, 1 );
	}

	value_ref access_deref( value_ref cur )
	{
		if( cur.how == value_ref::h_value ) {
			cur.how = value_ref::h_deref;
			return cur;
		}
		
		if( cur.how == value_ref::h_raw_pointer ) {
			cur.how = value_ref::h_value;
			return cur;
		}
		
		return value_ref::as_deref( resolve( cur ) );
	}

	value_ref access_deref_shared( value_ref cur )
	{
		if( cur.how == value_ref::h_value ) {
			cur.how = value_ref::h_deref_shared;
			return cur;
		}
		
		return value_ref::as_deref_shared( resolve( cur ) );
	}

	value_ref access_raw_pointer( value_ref cur )
	{
		if( cur.how == value_ref::h_deref_shared ) {
			cur.how = value_ref::h_raw_shared;
			return cur;
		}
		if( cur.how == value_ref::h_deref ) {	
			cur.how = value_ref::h_value;
			return cur;
		}
		
		if( cur.is_direct() ) {
			cur.origin->need_pointer = true;
			cur.how = value_ref::h_raw_pointer;
			return cur;
		}
		
		return value_ref::as_raw_pointer( resolve( cur ) );
	}

	value_ref access_raw_shared( value_ref cur )
	{
		if( cur.is_direct() ) {
			cur.how = value_ref::h_raw_shared;
			return cur;
		}
		
		return value_ref::as_raw_shared( resolve( cur ) );
	}
	
	value_ref access_raw_array( value_ref cur )
	{
		if( cur.is_direct() ) {
			cur.how = value_ref::h_raw_array;
			return cur;
		}
			
		return value_ref::as_raw_array( resolve(cur) );
	}
	
	//commenting allows a nice way to track what is happening in the dumper IR code
	void add_comment( std::string const & comment ) {
		auto ci = make_shared<comment_inst>();
		ci->comment = comment;
		add( ci );
	}
	template<typename ... T>
	void add_comment( const T & ... all ) {
		std::ostringstream out;
		add_comment_impl( out, all... );
		add_comment( out.str() );
	}
	template<typename T, typename ... TR>
	void add_comment_impl( std::ostream & out, const T & value, const TR & ... rest ) 	{
		out << value << " ";
		add_comment_impl( out, rest... );
	}
	template<typename T>
	void add_comment_impl( std::ostream & out, const T & value ) {
		out << value;
	}
};

inline function_builder module_builder::create_function( std::string const & name, bool can_fail )
{
	auto f = cctx->get_cur_module().create_function( name );
	auto mt = make_shared<type>( get_types().shared_from_this(), type::f_function_ptr );
	mt->access = type::a_raw;
	mt->can_fail = can_fail;
	f->defn_value->direct_type = mt;
	return function_builder( *f, mt );
}

inline function_builder module_builder::create_function( std::string const & name,
	shared_ptr<type const> func_type, shared_ptr<value> use_value )
{
	auto f = cctx->get_cur_module().create_function( name, use_value );
	f->defn_value->direct_type = func_type;
	return function_builder( *f, shared_ptr<type>() );
}

/**
	Convenience builder of switches.
*/
struct switch_builder 
{
	shared_ptr<switch_inst> inst;
	builder parent;
	
	shared_ptr<switch_inst::conditional_block> cur_block;
	shared_ptr<type const> result_type;
	boost::optional<builder> cond_block, expr_block;
	bool on_else, phi_mode;
	value_defn local, cond_result;
	
	switch_builder( builder & parent, shared_ptr<type const> result_type, bool phi_mode = false ) 
		: parent( parent )
		, result_type( result_type )
		, phi_mode( phi_mode ) { 
		if( result_type ) {
			if( phi_mode ) {
				inst = make_shared<switch_inst>(true);
				inst->result = parent.make_temp( result_type );
			} else {
				inst = make_shared<switch_inst>(false);
				static int count = 0;
				char buf[128];
				sprintf( buf, "switch_eval_%d", ++count );
				local = parent.define_local( result_type, buf );
			}
		} else {
			inst = make_shared<switch_inst>(false);
		}
		on_else = false;
		static int count = 0;
		char buf[128];
		sprintf( buf, "cond_result_%d", ++count );
		cond_result = parent.define_local( parent.cctx->get_types().get_boolean(), buf );
	}
	
	void push_back( block_flow flow ) { 
		STATE_CHECK( !result_type );
		parent.push_back( inst, flow ); 
	}
	void defer() {
		STATE_CHECK( !result_type );
		parent.defer( inst );
	}
	void add() {
		parent.add( inst );
		if( result_type ) {
			//TODO: this won't account for an error post-assign, we need defers with markers to say
			//when they are triggered (consider unit_tests/block/optional/release.simple.lfb
			parent.cctx->release_pointer( value_ref::as_raw_pointer(local),
				parent, block_flow::defer );
		}
	}
	void add_case() {
		STATE_CHECK( !on_else );
		cur_block = make_shared<switch_inst::conditional_block>( parent.get_block() );
		inst->blocks.push_back( cur_block );
		cond_block = builder( *cur_block->cond_block );
		expr_block = builder( *cur_block->expr_block );
	}
	void add_else() {
		add_case();
		on_else = true;
	}
	/**
		NOTE: call as a last step in creating the condition as this may add code to the
		condition block.
	*/
	void set_cond_value( value_ref val ) {
		STATE_CHECK( !on_else );
		//cur_block->cond_value = val;
		cond_block->add_store_inst( value_ref::as_raw_pointer(cond_result), val );
		cur_block->cond_value = cond_result;
	}
	void set_expr_value( value_ref val ) {
		if( phi_mode ) {
			cur_block->expr_value = val;
		} else {
			parent.cctx->assign( value_ref::as_raw_pointer(local), val, *expr_block, *expr_block, 
				compiler_context::af_no_release );
		}
	}
	value_defn get_result() {
		if( phi_mode ) {
			return inst->result;
		} else {
			return local;
		}
	}
	builder & get_cond_block() {
		return *cond_block;
	}
	builder & get_expr_block() {
		return *expr_block;
	}
};

inline switch_builder builder::create_switch( shared_ptr<type const> result_type ) {
	return switch_builder( *this, result_type );
}

inline switch_builder builder::create_switch() {
	return switch_builder( *this, shared_ptr<type const>() );
}

} //eon ir
#endif