#ifndef LEAF_IR_INSTRUCTION_HPP
#define LEAF_IR_INSTRUCTION_HPP

#include <vector>

#include "fwd.hpp"

namespace ir {

#define IR_INSTRUCTIONS \
	IR_INST(acquire_shared) \
	IR_INST(alloc_array) \
	IR_INST(alloc_shared) \
	IR_INST(alloc_uninit) \
	IR_INST(array_offset) \
	IR_INST(array_ptr_offset) \
	IR_INST(binary) \
	IR_INST(bind_func) \
	IR_INST(bind_func_explicit) \
	IR_INST(bind_raw_func) \
	IR_INST(cast) \
	IR_INST(clone) \
	IR_INST(comment) \
	IR_INST(fail) \
	IR_INST(free_array) \
	IR_INST(funccall) \
	IR_INST(loop) \
	IR_INST(loop_flow) \
	IR_INST(malloc) \
	IR_INST(pack_tuple) \
	IR_INST(promote) \
	IR_INST(loss) \
	IR_INST(release_shared) \
	IR_INST(resolve) \
	IR_INST(resume) \
	IR_INST(return) \
	IR_INST(special_field) \
	IR_INST(store) \
	IR_INST(struct_pointer) \
	IR_INST(switch) \
	IR_INST(unary) \
	IR_INST(void_context)


struct instruction {
	enum type_t
	{
		#define IR_INST(type) t_##type,
		IR_INSTRUCTIONS
		#undef IR_INST
	} type;
	
	virtual ~instruction() {};
	
	/**
		Checks that the state of the isntruction is valid. This will throw a recoverable_error
		if something is wrong (corruption).
	*/
	virtual void validate( util::validation_context & ) const { }
protected:
	instruction( type_t type )
		: type( type )
	{ }
};

/**
	Stores data at the given location. The left types supported are:
		raw_pointer<T> : in which case the "right" value T is stored at that location
		shared_pointer<T> : in which case the "right" value T is stored in the shared value location
*/
struct store_inst : public instruction
{
	value_ref left, right;
	
	store_inst()
		: instruction( t_store )
	{}
	
	void validate(util::validation_context & vctx) const;
};

/**	
	Exposes the results of the funciton in the "returns" values.
*/
struct funccall_inst : public instruction
{
	std::vector<value_ref> returns, args;
	//which function to call
	value_ref call;
	//provide an explicit context for a context function
	boost::optional<value_ref> context;
	
	funccall_inst()
		: instruction( t_funccall )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

struct pack_tuple_inst : public instruction
{
	std::vector<value_ref> sub;
	value_defn result;
	
	pack_tuple_inst()
		: instruction( t_pack_tuple )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

/**
	Allows both a value returning and non-value returning conditional series.
*/
struct switch_inst : public instruction
{
	struct conditional_block
	{
		shared_ptr<block> cond_block, expr_block; //init'd in ctor
		value_ref expr_value;
		boost::optional<value_ref> cond_value; //if not set then assumed true
		
		conditional_block( block & parent );
	};
	std::vector<shared_ptr<conditional_block>> blocks;
	value_defn result; //a phi of the individual expr_value's

	//true if expr_value and result are defined, otherwise no result values
	bool with_result;
	
	switch_inst( bool with_result )
		: instruction( t_switch )
		, with_result( with_result )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

/**
	Does iteration over a block so long as a condition remains true.
*/
struct loop_inst : public instruction
{
	shared_ptr<block> cond_block, body_block; //init'd in ctor
	value_ref cond_value; //must be produced by cond_block
	
	loop_inst( block & parent);
	
	void validate(util::validation_context & vctx) const;
};

/**
	Reduces an addressing value_ref to a value_defn.
*/
struct resolve_inst : public instruction
{
	value_ref source;
	value_defn result;
	
	resolve_inst( )
		: instruction( t_resolve )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

/**
	Obtains a pointer to a structure field given the current pointer.
*/
struct struct_pointer_inst : public instruction
{
	value_ref struct_ptr;
	size_t index;
	
	value_defn result;
	
	struct_pointer_inst()
		: instruction( t_struct_pointer )
	{ }
	
	void validate(util::validation_context & ) const { /*TODO:*/ };
};

/**
	Allocate memory on the heap.
*/
struct malloc_inst : public instruction
{
	//allocate an area to hold this type
	shared_ptr<ir::type const> direct_type;
	//the resulting memory, this type should be a pointer to direct_type
	value_defn result;
	
	malloc_inst()
		: instruction( t_malloc )
	{ }
	
	void validate(util::validation_context & vctx) const;
};


/**
	Creates an uninitalized value of the type.  This creates the space for the type but does not initialize yet -- the individual bits do exist, but are undefined.
*/
struct alloc_uninit_inst : public instruction {

	alloc_uninit_inst() :
		instruction( t_alloc_uninit ) {
	}
	
	shared_ptr<ir::type const> direct_type;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

/**
	Contructs a bound function from a raw function.
*/
struct bind_raw_func_inst : public instruction
{
	value_ref raw_func;
	value_defn result;
	
	bind_raw_func_inst()
		: instruction( t_bind_raw_func )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

struct shared_object_defn
{
	//allocate an are to hold a shared object of this type
	shared_ptr<ir::type const> direct_type;
	//the resulting memory, this type should be a pointer to direct_type (see users for ptr type)
	value_defn result;
	//which function accepts a raw_pointer to direct_type and destroys it
	value_ref destructor;
	//destructor is not specified, will be provided at time of release
	bool static_destructor;
	
	shared_object_defn()
		: static_destructor( false )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

struct alloc_shared_inst : public instruction
{
	//result is a shared_ptr
	shared_object_defn sod;
	
	alloc_shared_inst()
		: instruction( t_alloc_shared )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

/**
	Builds a context to call a function. The context is a shared object.
*/
struct bind_func_inst : public instruction
{
	//the ctx_func to call
	value_ref ctx_func;
	//the context which is bound
	shared_object_defn ctx_data;
	
	//the resulting bound function
	value_defn result;
	
	bind_func_inst()
		: instruction( t_bind_func )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

/**
	Binds the provided context to the unbound function to produce a
	bound function.
	
	TODO: the naming of this and bind_func_inst should be clarified, and if possible
	merged. This one is more logically binding, the other is creating the context.
*/
struct bind_func_explicit_inst : public instruction 
{
	value_ref ctx_func;
	value_ref ctx_data;
	
	//the data is a shared object, will be released with closure
	bool data_shared;
	//the context pointer is a shared pointer, otherwise a raw pointer
	bool ctx_shared;

	value_defn result;
	
	bind_func_explicit_inst()
		: instruction( t_bind_func_explicit ) { 
		data_shared = false;
	}
	
	void validate(util::validation_context & vctx) const;
};

/**
	Builds a ctx_function which will result in an exception at runtime
	if ever called. There is no need for the resulting object to be unique.
*/
struct void_context_inst : public instruction
{
	value_defn result;
	
	void_context_inst()
		: instruction( t_void_context )
	{ }
	
	void validate(util::validation_context & ) const { /*TODO:*/ }
};

/**
	Release a shared object
*/
struct release_shared_inst : public instruction
{
	enum destroy_how_t
	{
		//no destructor is invoked
		dh_trivial,
		//destructor was registered at creation time
		dh_dynamic,
	} destroy_how;
	
	value_ref object;
	
	release_shared_inst()
		: instruction( t_release_shared )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

struct acquire_shared_inst : public instruction
{
	value_ref object;
	
	acquire_shared_inst()
		: instruction( t_acquire_shared )
	{ }
	
	void validate(util::validation_context & vctx) const;
};

struct return_inst : public instruction
{
	return_inst()
		: instruction( t_return )
	{ }
	
	void validate(util::validation_context &) const { };
};

struct alloc_array_inst : public instruction
{
	alloc_array_inst()
		: instruction( t_alloc_array )
	{ }
	
	value_defn result;
	shared_ptr<ir::type const> item_type;
	value_ref length;
	
	void validate(util::validation_context & vctx) const;
};

struct free_array_inst : public instruction
{
	free_array_inst()
		: instruction( t_free_array )
	{ }
	
	value_ref array;
	
	void validate(util::validation_context & vctx) const;
};

struct array_offset_inst : public instruction
{
	array_offset_inst()
		: instruction( t_array_offset )
	{ }
	
	value_ref array, offset;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

struct array_ptr_offset_inst : public instruction
{
	array_ptr_offset_inst()
		: instruction( t_array_ptr_offset )
	{ }
	
	bool lvalue;
	value_ref array, offset;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};


struct special_field_inst : public instruction
{
	enum field_t
	{
		//size_t result, object must be array
		f_array_len,
	} field;
	
	special_field_inst()
		: instruction( t_special_field )
	{}
	
	value_ref object;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

struct promote_loss_inst : public instruction
{
protected:
	promote_loss_inst( type_t t ) 
		: instruction( t ) {
	}

public:
	enum what_t {
		w_integer,
		w_float,
		w_binary,
	} to_what, from_what;
	int to_bitsize;
	
	value_ref source;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

struct promote_inst : public promote_loss_inst
{
	promote_inst()
		: promote_loss_inst( t_promote )
	{}
};
struct loss_inst : public promote_loss_inst
{
	loss_inst()
		: promote_loss_inst( t_loss )
	{}
};

struct fail_inst : public instruction
{
	fail_inst()
		: instruction( t_fail )
	{ }
	
	//at this point we just use an integer level (all handled the same)
	int level;
	//the data to associate with the error
	boost::optional<value_ref> data;
};

struct resume_inst : public instruction
{
	resume_inst()
		: instruction( t_resume )
	{ }
	
	//the data that was captured in a fail_inst
	boost::optional<value_ref> data;
};

struct loop_flow_inst : public instruction
{
	loop_flow_inst()
		: instruction( t_loop_flow )
	{ }
	
	enum flow_t {
		f_break,
		f_next,
	} flow;
	void validate(util::validation_context & vctx) const;
};

/**
	A fundamental binary operation. These will be limited to generally those
	that would benefit from a close-to-the-host compilation.
*/
struct binary_inst : public instruction
{
	binary_inst()
		: instruction( t_binary )
	{ }
	
	enum op_t {
		o_add, 
		o_sub,
		o_mul,
		o_div,
		
		o_mod,
		
		o_bit_shl,
		o_bit_shr,
		
		o_or,
		o_and,
		o_xor,
		o_and_not,
		o_nor,
		o_nand,
		
		o_eq,
		o_neq,
		o_lt,
		o_lte,
		o_gt,
		o_gte,
	} op;
	static const op_t min_op_ = o_add;
	static const op_t max_op_ = o_gte;
	bool is_boolean_result() const {
		return op >= o_eq && op <= o_gte;
	}
	
	value_ref a, b;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

struct unary_inst : public instruction
{
	unary_inst()
		: instruction( t_unary )
	{ }

	enum op_t {
		o_neg, 
		o_abs,
		
		o_not,
	} op;
	static const op_t min_op_ = o_neg;
	static const op_t max_op_ = o_not;

	value_ref arg;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

/**
	Allows adding comments to the IR. Has no affect on the execution.
*/
struct comment_inst : public instruction
{
	comment_inst()
		: instruction( t_comment )
	{ }

	std::string comment;
};

/**
	Creates a temporary copy of the pointer-to-value and returns a
	pointer to that temporary.
*/
struct clone_inst : public instruction
{
	clone_inst()
		: instruction( t_clone )
	{ 
		source_value = false;
	}
	
	//the source is not a pointer, the result should be a pointer to a copy of the source value
	bool source_value;
	value_ref source;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

/**
	Treat one type as another: the size must be the same.
*/
struct cast_inst : public instruction
{
	cast_inst()
		: instruction( t_cast )
	{ }
	
	value_ref source;
	value_defn result;
	
	void validate(util::validation_context & vctx) const;
};

} //eon ir

#endif