#ifndef LEAF_IR_LLVM_BUILDER_HPP
#define LEAF_IR_LLVM_BUILDER_HPP

#include "gen.hpp"
#include "type_manager.hpp"

//LLVM: they don't make it easy to forward declare this
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
//NOTE: I'm not entirely clear on these defines
	#define __STDC_LIMIT_MACROS
	#define __STDC_CONSTANT_MACROS 
	
	#include "llvm/IR/IRBuilder.h"
	#include "llvm/IR/LLVMContext.h"
	#include "llvm/IR/Instructions.h"
#pragma GCC diagnostic pop

namespace llvm
{
	class IntegerType;
}

/**
	NOTE: The naming convention here follows the LLVM one to provide a consistent 
	style for using the LLVM interface.
*/
namespace ir_llvm
{

enum block_flow : int8_t {
	//normal sequential flow
	bf_seq = 0,
	//breaking out of inner-most loop
	bf_loop_break,
	//next iteration of inner-most loop
	bf_loop_next,
	//returning from function
	bf_return,
	//invoke resulting in error 
	bf_error,
};
static int8_t const block_flow_size = bf_error + 1;

//used to create a set of flows in an unsigned
inline unsigned block_flow_flag( block_flow bf ) {
	return 1<<bf;
}

struct block_data;
struct block_context;
typedef shared_ptr<block_data> block_ptr;

struct block_data {
	//an in flow is converted to a different out_flow
	static int const max_cvt = 2;
	int cvt_size;
	block_flow cvt_in[2];
	block_flow cvt_out[2];

	//all flow that lands here (CALCULATED)
	unsigned flow_in;
	//the flows that leave the block (CALCULATED)
	unsigned flow_out;
	
	//this block has already been completed
	bool complete;
	void set_complete() {
		complete = true;
	}

	shared_ptr<block_context> context;
	llvm::BasicBlock * ll_block;

	//where the seq flow should go (if null then to context flow)
	block_ptr block_seq;
	//where the sequence flows if condition is true
	block_ptr block_cond_true;
	llvm::Value * cond_value;
	//the undiverted flow ends up here (returns with original code to block_seq)
	block_ptr block_undivert;
	//block ends with undiversion to block_seq
	bool undiverts;
	
	block_data() = delete;
	block_data( block_data const & o ) = delete;
	
	void add_cvt( block_flow in, block_flow out ) {
		STATE_CHECK( cvt_size < max_cvt );
		cvt_in[cvt_size] = in;
		cvt_out[cvt_size] = out;
		cvt_size++;
	}
	
	block_ptr get_flow( block_flow bf ) const;

	static shared_ptr<block_data> create(shared_ptr<block_context> context, llvm::BasicBlock * ll_block ) {
		STATE_CHECK( context );
		STATE_CHECK( ll_block );
		return shared_ptr<block_data>( new block_data( context, ll_block ) );
	}
private:
	block_data( shared_ptr<block_context> context, llvm::BasicBlock * ll_block )
		: context( context )
		, ll_block( ll_block ) {
		flow_in = 0;
		flow_out = 0;
		complete = false;
		cond_value = nullptr;
		cvt_size = 0;
		undiverts = false;
	}
};

struct block_context {
	block_ptr flow[block_flow_size];
	//error handling level of this block
	int error_level;
	
	shared_ptr<block_context> parent;
	
	block_context() {
		error_level = 0;
	}
	block_context( shared_ptr<block_context> parent ) 
		: parent( parent ) {
		error_level = 0;
	}
	
	block_ptr get_flow( block_flow bf ) const;
	int get_max_error_level() const;
};

/**
	Extends the IRBuilder with more convenience functions to clarify the gen code. This is intended
	to be used as the scope of a function and as such retains state for the function.
*/
class function_builder
{
	llvm::IRBuilder<> lbuilder;

	llvm::Module & module;
	llvm::LLVMContext & llvm_context;
	llvm::Function & function;
	type_manager & types;
	
	unsigned label_count = 0;
	unsigned name_count = 0;
	block_ptr cur_block;
	gen::options const & gopts;
	
	block_ptr entry_block;
 	block_ptr resume_block, rethrow_block;
 	block_ptr get_rethrow_block();
	
 	void complete_block( block_ptr block );
	std::vector<block_ptr> all_blocks;
	void propagate_flow( block_ptr block, unsigned flow_flags );
public:
	function_builder( llvm::Module & module, llvm::Function & function, 
		type_manager & types, gen::options const & gopts );

 	block_ptr get_resume_block();
 	
 	void set_rethrow_block(block_ptr b) {
		rethrow_block = b;
	}
	
	llvm::LLVMContext & get_llvm_context() {
		return llvm_context;
	}
	
	/**
		Call once the function has all its blocks and instructions created. This will finalize
		the incomplete blocks by adding appropriate flow branching instructions.
	*/
	void complete_blocks();
	
	//intended to override those in IRBuilder
	block_ptr get_insert_block() { 
		return cur_block; 
	}
	void set_insert_point( block_ptr block ) {
		STATE_CHECK( block );
		cur_block = block;
		lbuilder.SetInsertPoint( block->ll_block );
	}
	
	//a function local set to (alloca function variable)
	enum return_path_value
	{
		rpv_sequential = 0,
		rpv_return = 1,
		rpv_resume = 2,
		rpv_cleanup = 3,
		rpv_loop_break = 4,
		rpv_loop_next = 5,
	};
	llvm::Value * return_path;
	//the source context pointer provided to the function
	llvm::Value * context_arg;
	//the current error value
	llvm::Value * error_value;
	llvm::Value * error_info;
	
	llvm::ConstantInt * ConstInt( llvm::IntegerType * type, int64_t val )  {
		return llvm::ConstantInt::get( type, val );
	}
	llvm::ConstantInt * const_natural_int( int64_t val ) {
		return llvm::ConstantInt::get( types.f_natural_int(), val );
	}
	llvm::ConstantInt * const_abi_int( int64_t val ) {
		return llvm::ConstantInt::get( types.f_abi_int(), val );
	}
	llvm::ConstantInt * ConstInt8( int8_t val ) {
		return llvm::ConstantInt::get( llvm::Type::getInt8Ty( llvm_context ), val );
	}
	llvm::ConstantInt * const_int( llvm::IntegerType * type, int64_t val ) {
		return llvm::ConstantInt::get( type, val );
	}
	llvm::Constant * const_zero( llvm::Type * ty ) {
		if( ty->isPointerTy() ) {
			return llvm::ConstantPointerNull::get( llvm::cast<llvm::PointerType>(ty) );
		} else if( ty->isFloatingPointTy() ) {
			return llvm::ConstantFP::get( ty, 0 );
		} else {
			return llvm::ConstantInt::get( ty, 0 );
		}
	}
	
	char name_buffer[128];
	/**
		Create a unique name for this function.
		@param base to use for the name, if "" then "" is returned
		@return the name, this is only valid until another call to local_name
	*/
	char const * local_name( char const * base ) {
		if( *base == 0 )
			return base;
		sprintf( name_buffer, "%s_%d", base, name_count++ );
		return name_buffer;
	}
	
	shared_ptr<block_context> get_context() {
		return cur_block->context;
	}
	
	block_ptr create_child_block( block_ptr parent, char const * label_part = "label" ) {
		char buf[32];
		sprintf( buf, "_%d", label_count++ );
		std::string label( label_part );
		label += buf;
		auto mb = block_data::create( parent->context, 
			llvm::BasicBlock::Create( llvm_context, label.c_str(), &function ) );
		all_blocks.push_back( mb );
		return mb;
	}
	
	block_ptr create_child_block( char const * label_part = "label" ) {
		return create_child_block( cur_block, label_part );
	}
	
	block_ptr create_child_context( char const * label_part ) {
		char buf[32];
		sprintf( buf, "_%d", label_count++ );
		std::string label( label_part );
		label += buf;
		auto mb = block_data::create( make_shared<block_context>(get_context()), 
			llvm::BasicBlock::Create( llvm_context, label.c_str(), &function ) );
		all_blocks.push_back( mb );
		return mb;
	}
	
	llvm::Value * CreateStructLoad( llvm::Value * st, size_t idx, char const * name = "" )
	{
		return lbuilder.CreateLoad( CreateStructGEP( st, unsigned(idx) ), local_name(name) );
	}
	
	void CreateStructStore( llvm::Value * Agg, llvm::Value * Val, size_t idx, char const * name = "" )
	{
		lbuilder.CreateStore( Val, CreateStructGEP( Agg, unsigned(idx) ), local_name(name) );
	}
	
	llvm::Value * CreateInsertValue( llvm::Value *Agg, llvm::Value *Val, size_t idx )
	{
		unsigned i = unsigned(idx);
		return lbuilder.CreateInsertValue( Agg, Val, llvm::ArrayRef<unsigned>( &i, &i+1 ) );
	}
	
	llvm::Value * CreateExtractValue( llvm::Value * Agg, size_t idx, char const * name = "" )
	{
		unsigned i = unsigned(idx);
		return lbuilder.CreateExtractValue( Agg, llvm::ArrayRef<unsigned>( &i, &i + 1 ),
			local_name(name) );
	}
	
	llvm::Value * direct_call( llvm::Value *Callee )
	{ return lbuilder.CreateCall( Callee ); }
	llvm::Value * direct_call( llvm::Value *Callee, llvm::Value * arg )
	{ return lbuilder.CreateCall( Callee, arg ); }
	llvm::Value * direct_call( llvm::Value *Callee, llvm::ArrayRef< llvm::Value * > Args )
	{ return lbuilder.CreateCall( Callee, Args ); }
	
	llvm::Value * Call( llvm::Value * callee, llvm::ArrayRef<llvm::Value*> args ) {
		return lbuilder.CreateCall( callee, args );
	}
	
	llvm::Value * call( type_manager::func_type_entry const & fte, llvm::Value * callee,
		std::vector<llvm::Value*> const args, llvm::Value** out_ret_ptr = nullptr );

	llvm::Value * Call( llvm::Value * Callee ) {
		llvm::Value* args[0];
		return Call( Callee, llvm::ArrayRef<llvm::Value*>( args + 0, args + 0 ) );
	}
	
	llvm::Value * Call( llvm::Value * Callee, llvm::Value* p1 ) {
		llvm::Value* args[1] = { p1 };
		return Call( Callee, llvm::ArrayRef<llvm::Value*>( args  + 0, args + 1 ) );
	}

 	llvm::Value * CreateBr( block_ptr dest ) {
		return lbuilder.CreateBr( dest->ll_block );
	}
	
	llvm::LandingPadInst * CreateLandingPad( llvm::Type *Ty, llvm::Value */*PersFn*/ )
	{ //TODO: Remove since unused, and PersFn no longer exists
			return lbuilder.CreateLandingPad( Ty, 0 ); }
	
	llvm::AllocaInst* CreateAlloca( llvm::Type *Ty, llvm::Value *ArraySize=0, const llvm::Twine &Name="")
	{ return lbuilder.CreateAlloca( Ty, ArraySize, Name ); }
	
	llvm::StoreInst* CreateStore( llvm::Value *Val, llvm::Value *Ptr )
	{ return lbuilder.CreateStore( Val, Ptr ); }
	
	llvm::Value* CreateBitCast( llvm::Value *V, llvm::Type *DestTy, char const * name = "" ) { 
		return lbuilder.CreateBitCast( V, DestTy, local_name(name) ); 
	}
	llvm::Value * create_int_to_ptr( llvm::Value * v, llvm::Type * ty ) { 
		return lbuilder.CreateIntToPtr( v, ty ); 
	}
	llvm::Value * create_ptr_to_int( llvm::Value * v, llvm::Type * ty ) {
		return lbuilder.CreatePtrToInt( v, ty );
	}
	
	llvm::Value* CreateStructGEP( llvm::Value *Ptr, unsigned Idx, char const * name = "" ) {
		auto p = Ptr->getType()->getPointerElementType();
		return lbuilder.CreateStructGEP( p, Ptr, Idx, local_name(name) );
	}
	
	llvm::LoadInst * CreateLoad( llvm::Value *Ptr, char const * name = "" )
	{ return lbuilder.CreateLoad( Ptr, local_name(name) ); }
	
	llvm::AtomicRMWInst * CreateAtomicRMW( llvm::AtomicRMWInst::BinOp Op, 
		llvm::Value *Ptr, llvm::Value *Val, llvm::AtomicOrdering Ordering, 
		llvm::SynchronizationScope SynchScope=llvm::CrossThread) 
	{ return lbuilder.CreateAtomicRMW( Op, Ptr, Val, Ordering, SynchScope ); }
	
	llvm::Value* CreateICmpSLT( llvm::Value * lhs, llvm::Value * rhs )	
	{ return lbuilder.CreateICmpSLT( lhs, rhs ); }
	llvm::Value* CreateICmpEQ(llvm::Value * a, llvm::Value * b )
	{ return lbuilder.CreateICmpEQ( a, b ); }
	
	llvm::PHINode * create_phi( llvm::Type * ty, unsigned num)
	{ return lbuilder.CreatePHI( ty, num ); }
	
	llvm::Value * create_ptr_to_int( llvm::Value * val ) {
		return lbuilder.CreatePtrToInt( val, types.f_ptr_int() );
	}
	
	llvm::Value * create_sign_ext( llvm::Value * val, llvm::Type * dest)	
	{ return lbuilder.CreateSExt( val, dest ); }
	llvm::Value * create_zero_ext( llvm::Value * val, llvm::Type * dest)	
	{ return lbuilder.CreateZExt( val, dest ); }
	llvm::Value * create_si_to_fp( llvm::Value * val, llvm::Type * dest )
	{ return lbuilder.CreateSIToFP( val, dest ); }
	llvm::Value * create_ui_to_fp( llvm::Value * val, llvm::Type * dest )
	{ return lbuilder.CreateUIToFP( val, dest ); }
	llvm::Value * create_fp_ext( llvm::Value * val, llvm::Type * dest )
	{ return lbuilder.CreateFPExt( val, dest ); }
	llvm::Value * create_trunc( llvm::Value * val, llvm::Type * dest )
	{ return lbuilder.CreateTrunc( val, dest ); }
	llvm::Value * create_fp_to_si( llvm::Value * val, llvm::Type * dest )
	{ return lbuilder.CreateFPToSI( val, dest ); }
	llvm::Value * create_fp_to_ui( llvm::Value * val, llvm::Type * dest )
	{ return lbuilder.CreateFPToUI( val, dest ); }
	llvm::Value * create_sext_or_trunc( llvm::Value * val, llvm::IntegerType * dest )
	{ return lbuilder.CreateSExtOrTrunc( val, dest ); }
	llvm::Value * create_fp_trunc( llvm::Value *val, llvm::Type * dest ) 
	{ return lbuilder.CreateFPTrunc( val, dest ); }
	
	#define BINOP( name, fop, iop, uop )  \
		llvm::Value * create_##name( llvm::Value * lhs, llvm::Value * rhs, bool is_signed = true ) \
		{  \
			if( lhs->getType()->isFloatingPointTy() ) \
				return lbuilder.Create##fop( lhs, rhs ); \
			if( is_signed ) \
				return lbuilder.Create##iop( lhs, rhs );  \
			return lbuilder.Create##uop( lhs, rhs ); \
		}
	BINOP(add,FAdd,Add,Add)
	BINOP(sub,FSub,Sub,Sub)
	BINOP(mul,FMul,Mul,Mul)
	BINOP(div,FDiv,SDiv,UDiv)
	BINOP(mod,FRem,SRem,URem)
	
	BINOP(eq,FCmpOEQ,ICmpEQ,ICmpEQ)
	BINOP(neq,FCmpONE,ICmpNE,ICmpNE)
	BINOP(lt,FCmpOLT,ICmpSLT,ICmpULT) 
	BINOP(lte,FCmpOLE,ICmpSLE,ICmpULE)
	BINOP(gt,FCmpOGT,ICmpSGT,ICmpUGT)
	BINOP(gte,FCmpOGE,ICmpSGE,ICmpUGE)
	#undef BINOP
	
	//integer only support
	#define BINOP( name, iop )  \
		llvm::Value * create_##name( llvm::Value * lhs, llvm::Value * rhs, bool is_signed = false ) \
		{  \
			STATE_CHECK(!is_signed); \
			return lbuilder.Create##iop( lhs, rhs );  \
		}
	BINOP(bit_shl,Shl)
	BINOP(bit_shr,LShr)
	#undef BINOP

	//boolean/binary only support
	#define BINOP( name, iop )  \
		llvm::Value * create_##name( llvm::Value * lhs, llvm::Value * rhs, bool is_signed = false ) \
		{  \
			STATE_CHECK(!is_signed); \
			return lbuilder.Create##iop( lhs, rhs );  \
		}
	BINOP(or,Or)
	BINOP(and,And)
	BINOP(xor,Xor)
	#undef BINOP

	llvm::Value * create_neg( llvm::Value * arg ) 
	{
		if( arg->getType()->isFloatingPointTy() )
			return lbuilder.CreateFNeg( arg );
		return lbuilder.CreateNeg( arg );
	}
	llvm::Value * create_not( llvm::Value * arg )
	{
		//-1 is all bits set in 2's complement
		return lbuilder.CreateXor( arg, ConstInt( 
			llvm::cast<llvm::IntegerType>(arg->getType()), -1 ) ); 
	}
	llvm::Value * create_and_not( llvm::Value * lhs, llvm::Value * rhs, bool is_signed = false )
	{
		STATE_CHECK(!is_signed);
		return create_and( lhs, create_not(rhs) );
	}
	llvm::Value * create_nor( llvm::Value * lhs, llvm::Value * rhs, bool is_signed = false )
	{
		STATE_CHECK(!is_signed);
		return create_not( create_or(lhs, rhs) );
	}
	llvm::Value * create_nand( llvm::Value * lhs, llvm::Value * rhs, bool is_signed = false )
	{
		STATE_CHECK(!is_signed);
		return create_not( create_and(lhs, rhs) );
	}
	
	llvm::Value * create_abs( llvm::Value * arg );
	
	llvm::Value* create_inbounds_gep( llvm::Value * ptr, llvm::Value * idx )
	{ return lbuilder.CreateInBoundsGEP( ptr, idx ); }
	llvm::Value* create_gep( llvm::Value * ptr, unsigned idx )	
	{ return lbuilder.CreateConstGEP1_32( ptr, idx ); }

	void create_function_entry();
	
	/////////////////////////////////////////////
	// terminators (also "Call" sometimes)
	void branch_sequence( block_ptr block ) {
		cur_block->block_seq = block;
	}
	void branch_conditional( llvm::Value * val, block_ptr on_true, block_ptr on_false ) {
		cur_block->block_cond_true = on_true;
		cur_block->block_seq = on_false;
		cur_block->cond_value = val;
	}
	
	void flow_loop_break() {
 		CreateStore( ConstInt8(bf_loop_break), return_path );
		cur_block->add_cvt( bf_seq, bf_loop_break );
		cur_block->block_seq.reset();
	}
	
	void flow_loop_next() {
 		CreateStore( ConstInt8(bf_loop_next), return_path );
		cur_block->add_cvt( bf_seq, bf_loop_next );
		cur_block->block_seq.reset();
	}
	
	void flow_return() {
		CreateStore( ConstInt8(bf_return), return_path );
		cur_block->add_cvt( bf_seq, bf_return );
		cur_block->block_seq.reset();
	}
	
 	void flow_resume() {
 		CreateStore( ConstInt8(bf_error), return_path );
 		cur_block->add_cvt( bf_seq, bf_error );
 		cur_block->block_seq.reset();
 	}
 	
	void unreachable() { 
		lbuilder.CreateUnreachable(); 
		cur_block->set_complete();
	}
	
	void func_return( llvm::Value *V) { 
		lbuilder.CreateRet( V ); 
		cur_block->set_complete();
	}
	
	void func_return_void() { 
		lbuilder.CreateRetVoid(); 
		cur_block->set_complete();
	}
	
	///////////////////////////////////////////////////////
	// Other flow control
	struct diversion {
		llvm::Value * flow;
		llvm::Value * error;
		llvm::Value * error_info;
	};
	
	diversion flow_divert( block_ptr return_flow ) {
		cur_block->block_undivert = return_flow;
		
		diversion d;
		d.flow = CreateLoad( return_path, "return_path" );
		d.error = CreateLoad( error_value, "error_value" );
		d.error_info = CreateLoad( error_info, "error_info" );
		CreateStore( const_zero( types.f_error_code() ), error_value );
		CreateStore( const_zero( types.generic_shared_ptr() ), error_info );
		CreateStore( ConstInt8(bf_seq), return_path );
		return d;
	}
	
	void flow_undivert( block_ptr return_flow, diversion const & old ) {
		cur_block->block_seq = return_flow;
		cur_block->undiverts = true;
		CreateStore( old.flow, return_path );
		CreateStore( old.error, error_value );
		CreateStore( old.error_info, error_info );
	}
};

} //eon ir_llvm

#endif