#ifndef LEAF_IR_LLVM_GEN_HPP
#define LEAF_IR_LLVM_GEN_HPP

#include "ir/fwd.hpp"
#include "ir/instruction.hpp"

#include <deque>
#include <map>

//these are all forward declared so a user of this class doesn't need to include LLVM headers
namespace llvm
{
	class LLVMContext;
	class Module;
	class Value;
	class Function;
	class BasicBlock;
	class GlobalValue;
	class Linker;
	class ExecutionEngine;
	class DataLayout;
	class Type;
	class PointerType;
	class FunctionType;
}

namespace ir_llvm
{

class function_builder;
class type_manager;

struct block_data;
typedef shared_ptr<block_data> block_ptr;

class gen
{
public:
	struct options
	{
		//enable backend(LLVM) optimizations
		bool llvm_opt;
		//try to pre-reduce branching constructs to only those needed
		bool reduce_branch;
		//try to reduce blocks used in flow
		bool reduce_flow;
		//generate PIC objects
		bool pic;
		
		options()
			: llvm_opt( false )
			, reduce_branch( true )
			, reduce_flow( true )
			, pic( false )
		{ }
	};
	gen( options gopts );
	~gen();
	
	int execute( ir::module const & m, std::string const & show_code );
	void emit_object( ir::module const & m, std::string const & base_path,
		std::string const & show_code );
private:
	void show_code( std::string const & sc );
	
	options gopts;
	
	unique_ptr<llvm::LLVMContext> llvm_context;
	//LLVM 3.9 changed owernship rules, so we can't have unique_ptr's here anymore
	llvm::Module *module;
	bool del_module;
	unique_ptr<llvm::DataLayout> data_layout;
	unique_ptr<llvm::Linker> linker;
	unique_ptr<type_manager> types;
	std::multimap<std::string,llvm::Function*> func_map;

	//used to do initiliazation of global values
	llvm::Function * startup_func;
	unique_ptr<function_builder> startup_func_builder;
	
	llvm::Function * get_release_deref_shared();
	llvm::Function * get_context_dtor();
	
	void link();
	llvm::ExecutionEngine * create_engine();
	
	void create_std_functions();
	llvm::Value * gen_malloc( llvm::Value * size, function_builder & bb );
	llvm::Value * gen_malloc( size_t size, function_builder & bb );
	llvm::Value * heap_alloc_type( llvm::Type * t, function_builder & bb );
	void gen_release_optional_shared( llvm::Value * ptr, function_builder & bb );
	void gen_acquire_optional_shared( llvm::Value * ptr, function_builder & bb );
	
	void gen_free( llvm::Value * ptr, function_builder & bb );
	
	void build_module( ir::module const & m );
	void build_wrapped_entry( ir::module const & m );
	void build_block( ir::block const & bk, function_builder & bb );
	void build_block_part( ir::block_seq_item const & bk, function_builder & bb );
	void setup_function( ir::function const & func );
	void build_function( ir::function const & func );
	
	void build_instruction( ir::instruction const & i, function_builder & bb );
	void build_global_variable( ir::value const & gv );

	#define IR_INST(type) \
		void build_##type##_inst( ir::type##_inst const & i, function_builder & bb);
		IR_INSTRUCTIONS
	#undef IR_INST
	
	llvm::Value* obtain_value( ir::value_ref const & vr, function_builder & bb );
	void build_release_shared( llvm::Value * object, bool call_dtor, function_builder & bb );
	void build_acquire_shared( llvm::Value * object, function_builder & bb );
	
	/**
		Initialize an allocated shared object.
		
		@param dtor destructor to register, if null then _void_destructor is  used
		@return ptr to the sf_data area
	*/
	llvm::Value* init_shared( llvm::Value * sptr, llvm::Value * dtor, function_builder & bb );
	
	enum landing_type_t : int {
		lt_cleanup = -1,
		lt_all = 0,
		//positive values are level to catch
	};
 	block_ptr create_catch_block( block_ptr next, function_builder & bb, int landing_type );
	std::map<int,llvm::Value*> clause_level_globals;
	
	//internal tracking of symbols
	struct symbol_cref
	{
		llvm::Value * value;
		
		enum ref_type_t
		{
			rt_value,
			rt_pointer,
			rt_global
		} ref_type;
	};
	typedef std::deque<symbol_cref> symbol_deque_t;
	symbol_deque_t symbol_deque;
	void create_symbol_cref( intptr_t & track, llvm::Value * v, symbol_cref::ref_type_t rt );
};

} //eon ir

#endif