#ifndef LEAF_IR_LLVM_TYPE_MANAGER_HPP
#define LEAF_IR_LLVM_TYPE_MANAGER_HPP

#include "ir/fwd.hpp"

#include <map>

namespace llvm
{
	class Type;
	class FunctionType;
	class PointerType;
	class LLVMContext;
	class DataLayout;
	class IntegerType;
	class StructType;
	class ArrayType;
}

namespace ir_llvm
{

/**
	The fields of a shared object. The last "sf_data" is specific to each shared object, with the generic
	structure having an empty array. The "sf_align" is used to force sf_data to be maximally aligned.
*/
enum shared_field : unsigned {
	sf_ref_count = 0,
	sf_destructor,
	//ensures sf_data is always maximally aligned
	sf_align,
	//a pointer to this data is provided to the destructor
	sf_data,
	
	sf_size
};

/**
	The fields of a bound context object (note this is usually wrapped within the "sf_data" or a
	shared object). The generic version for a particular function will have cf_data as an empty
	array. The fully generic version will have generic_ptr() for the func_ptr.
*/
enum context_field : unsigned {
	//a pointer to the function to call (unbound function)
	cf_func_ptr,
	//the data to pass as the first argument to func_ptr
	cf_ctx_ptr,
	//pointer to the data destructor
	cf_data_destructor,
	//ensure cff_ctx_offset is maximally aligned
	cf_align,
	//this *must* be last, and it is actaully the start of the data, not a pointer itself (not always used)
	//a pointer to this data is provided to the data_destructor
	cf_data,
	cf_size
};

/**
	The fields of the error_return that carries error information out of a function.
*/
enum error_field : unsigned {
	// <f_error_code>
	ef_error_code = 0,
	// <generic_shared_ptr> if non-0 this is a shared_field
	ef_error_data
};

/**
	Separates logic dealing with types from the gen class to make it clearer.
*/
class type_manager
{
	llvm::LLVMContext & context;
	llvm::DataLayout & data_layout;
	
	llvm::PointerType * generic_ptr_;
	llvm::Type * void_;
	llvm::FunctionType * destructor_;
	llvm::StructType * extype_type_;
	llvm::StructType * generic_shared_;
	llvm::StructType * generic_ctx_frag_;
	llvm::StructType * error_return_;
public:
	type_manager( llvm::LLVMContext & context, llvm::DataLayout & data_layout );
	
	llvm::PointerType * generic_ptr() { 
		return generic_ptr_; 
	}
	llvm::StructType * generic_shared() {
		return generic_shared_;
	}
	llvm::PointerType * generic_shared_ptr();
	llvm::StructType * generic_ctx_frag() {
		return generic_ctx_frag_;
	}
	llvm::Type * type_void() { 
			return void_; 
	}
	llvm::IntegerType * f_intx( unsigned bits );
	llvm::IntegerType * f_natural_int();
	llvm::IntegerType * f_abi_char();
	llvm::IntegerType * f_abi_int();
	llvm::IntegerType * f_byte();
	llvm::IntegerType * f_error_code() {
		return f_byte();
	}
	llvm::PointerType * ptr_to( llvm::Type * t );
	llvm::FunctionType * destructor() { 
			return destructor_; 
	}
	llvm::Type * f_float( unsigned bits );
	llvm::StructType * extype_type() { 
		return extype_type_; 
	}
	llvm::StructType * error_return() {
		return error_return_;
	}
	
	llvm::IntegerType* f_ptr_int();

	llvm::IntegerType* main_argc();
	llvm::PointerType* main_argv();
	
	size_t alloc_size( llvm::Type * t );
	size_t max_align();
	
	//types
	struct type_ref {
		//all items are compared to idenfity uniqueness
		ir::type const * ir_type;
		bool ctx, shared, direct;
		ir::type const * ctx_a;
		llvm::Type *ctx_b, *func_a;
		
		type_ref() : 
			ir_type( nullptr ), 
			ctx( false ), 
			shared( false ), 
			direct( false ), 
			ctx_a( nullptr ),
			ctx_b( nullptr ),
			func_a( nullptr ) {
		}
		
		bool operator<( type_ref const & o ) const;
	};
	struct type_entry {
		llvm::Type * ll_type;
		//can it be used directly as a return value
		bool direct_return;
		
		type_entry()
			: ll_type( nullptr )
			, direct_return( false )
		{ }
	};
	typedef std::map<type_ref, type_entry> type_map_t;
	type_map_t type_map;
	
	//The type when used as a typical value
	type_entry value( ir::type const & type ); //assume safe to hold pointer
	
	struct func_type_entry {
		//the full function type
		llvm::FunctionType * ll_func_type;
		//the return type (logical, whether returned or as argument)
		llvm::Type * ll_logical_return_type;
		//true if returned directly from funciton, false if as argument
		bool is_direct_return;
		//true if this function uses a context
		bool is_context;
		//true if the context is shared
		bool is_context_shared;
		//true if the function has an error return
		bool is_error_return;
		
		func_type_entry() :
			ll_func_type( nullptr ),
			ll_logical_return_type( nullptr ),
			is_direct_return( false ),
			is_context( false ),
			is_context_shared( false ),
			is_error_return( false ) {
		}
	};
	typedef std::map<type_ref, func_type_entry> func_type_map_t;
	func_type_map_t func_type_map;
	func_type_entry function( ir::type const & type );
	
	//create shared_ptr to this type
	type_entry shared_ptr( ir::type const & t );
	
	//ll_type is a StructType
	type_entry shared_struct( ir::type const & t, bool direct = false );
	type_entry shared_context_struct( ir::type const & t, ir::type const * ctx = nullptr,
		llvm::Type * ll_ctx = nullptr );
	
	/**
		Generalizes lookup and caching code.
		
		LLVM's asserts check types by pointer equality thus we must ensure identical
		types get the exact same type object back. I have no idea how this will work
		accross modules! 
	*/
	template<typename Map, typename TypeEntry>
	struct type_map_helper_ {
		type_map_helper_( Map & tmap, ir::type const * gtype )
			: tmap( tmap ) { 
			tr.ir_type = gtype;
		}
		
		Map & tmap;
		type_ref tr;
		
		bool has() {
			return tmap.find( tr ) != tmap.end();
		}
		TypeEntry get() {
			return tmap.find( tr )->second;
		}
		void set( TypeEntry te ) {
			tmap[tr] = te;
		}
	};

	type_map_helper_<type_map_t,type_entry> type_map_helper( ir::type const * gtype ) {
		return type_map_helper_<type_map_t,type_entry>( type_map, gtype );
	}
	type_map_helper_<func_type_map_t,func_type_entry> func_type_map_helper( ir::type const * gtype ) {
		return type_map_helper_<func_type_map_t,func_type_entry>( func_type_map, gtype );
	}
	
	/**
		Create a zero-length array to align the next field in the structure.
		
		@param len the alignment size (in bytes) of the next field, 0 indicates maximally aligned.
	*/
	llvm::ArrayType * get_align_field( size_t len = 0 );
};

} //eon ir_llvm

#endif