#ifndef LEAF_IR_VALUE_HPP
#define LEAF_IR_VALUE_HPP

#include "fwd.hpp"

#include <boost/optional.hpp>

namespace ir
{

struct value
{
	enum form_t {
		//the noref types
		f_import_code = 1,
		f_import_data,
		f_const_integer,
		f_const_binary,
		f_const_float,
		f_const_boolean,
		f_nullptr,
		f_temp,
		f_undefined,
		f_raw_binary,
		
		//the ref types
		f_return_ref,
		f_arg_ref,
		f_capture_ref,
		f_global,
		f_local,
	} form;
	
	value() : 
		form( form_t(0) ), 
		is_export(false), 
		need_pointer(false), 
		cref(0) { 
	}
	value( form_t form ) : 
		form( form ), 
		is_export( false ), 
		need_pointer( false ), 
		cref( 0 ) { 
	}
	
	virtual ~value() {}
	virtual void validate( util::validation_context & vctx) const;
	
	//is this directly usable as storage
	bool is_direct_storage() const { 
		return form == f_return_ref || form == f_global || form == f_local || form == f_capture_ref; 
	}
	bool is_any_storage() const;
	
	bool is_noref() const { 
		return form == f_import_code || form == f_import_data || form == f_const_integer 
		|| form == f_const_float || form == f_const_boolean || form == f_temp 
		|| form == f_const_binary; 
	}
		
	std::string name;
	size_t anon_ref; //only relevant if name is empty
	shared_ptr<type const> direct_type; //if null then not yet initialized
	bool is_export;
	
	//TODO: a variant that includes the raw_binary
	union {
		//TODO: a larger data type is required for larger integers
		intmax_t int_value;
		uintmax_t uint_value;
		long double float_value;
		
		bool boolean_value;
		struct {
			//index into capture structure for contexts
			size_t index;
			size_t parent;
		} capture;
	};
	std::vector<uint8_t> raw_binary;
	
	//a pointer will be required of this object
	bool need_pointer;
	
	//used by compilation stage
	mutable intptr_t cref;
};

struct value_defn {
	shared_ptr<value> origin;
	
	template<typename T>
	value_defn( shared_ptr<T> origin ) : 
		origin( origin ) { 
	}
	
	shared_ptr<type const> result_type() const { 
		return origin->direct_type; 
	}
	
	value_defn() { 
	}
	
	void validate( util::validation_context & vctx) const;
};

/**
	Stores a reference to a value allowing a variety of accessing modes. This cuts
	down on the instructions needed to deal with pointers and tuples -- which are
	trivial to lower later.
*/
struct value_ref
{
	shared_ptr<value> origin;
	
	enum how_t
	{
		//the direct value (of type origin->direct_type)
		h_value = 1,
		//a pointer to the value (of type raw_pointer<origin->direct_type>)
		h_raw_pointer,
		//derefence the ptr (direct_type must be raw_pointer)
		h_deref,
		//derefence the shared object
		h_deref_shared,
		//raw_pointer to shared object data
		h_raw_shared,
		//raw pointer to the array data
		h_raw_array,
	} how;
	
	//access this field of origin (prior to applying how)
	//if the underlying type is a raw_pointer then the result will be a raw_ptr to the sub-field of that pointer
	boost::optional<size_t> pre_field;
	
	static value_ref as_value( value_defn vd )
	{ return value_ref( vd.origin, h_value ); }
	
	static value_ref as_raw_pointer( value_defn vd )
	{ return value_ref( vd.origin, h_raw_pointer ); }
	
	static value_ref as_deref( value_defn vd )
	{ return value_ref( vd.origin, h_deref ); }
	
	static value_ref as_deref_shared( value_defn vd )
	{ return value_ref( vd.origin, h_deref_shared ); }
	
	static value_ref as_raw_shared( value_defn vd )
	{ return value_ref( vd.origin, h_raw_shared ); }
	
	static value_ref as_raw_array( value_defn vd )
	{ return value_ref( vd.origin, h_raw_array ); }
	
	/**
		@return true if only a plain "value" mode with no modifiers is specified.
	*/
	bool is_direct() const
	{ return how == h_value && !pre_field; }
	
	/**
		Return true if this is_direct and the value is not a variable reference type (not
		bound to a symbol).
	*/
	bool is_noref_value() const
	{ return is_direct() && origin->is_noref(); }
	

	/**
		The resulting type after applying access modifiers. This assumes the value_ref
		is valid, so if it isn't the result may be broken, or you get an error.
	*/
	shared_ptr<type const> result_type() const;
	
	value_ref() : 
		how( how_t(0) ) { 
	}
	value_ref( value_defn const & vd ) { 
		*this = as_value(vd); 
	}
	
	//simplify conversion from value_defn (used commonly and unambiguous)
	value_ref & operator=(value_defn const & vd ) { 
		*this = as_value( vd ); 
		return *this;
	}
	
	void validate( util::validation_context & vctx) const;
	
	bool is_valid() const {
		return how != how_t(0);
	}
private:
	value_ref( shared_ptr<value> origin, how_t how ) : 
		origin( origin ), 
		how( how ) { 
	}
};

} //eon ir

#endif