#ifndef LEAF_LANG_COMMON_HPP
#define LEAF_LANG_COMMON_HPP

#include <vector>

namespace leaf {

enum class scope_style_t
{
	//expresisons may forward reference variables so long as a proper tree is formed
	declarative,
	//expressions may only reference prevously declared variables
	imperative,
	//a special mode for block test files, funcdefn is pending but evaluation is imperative
	//decl's are required only at the point they might actually be used (TODO: not done yet, quite unsafe now)
	dynamic,
};

/**
	Failure severity levels.
*/
enum class fail_level {
	//not a real level, used to continue an existing failure
	resume = 0,
	pre, //TODO: will probably disappear
	normal,
	critical,
};

/**
	The logical "completeness" of a definition. This refers either to both the completeness of the
	types involved and the overall completeness of the object.
	
	Most processing should continue as long as the requisites are at the "resolved" level. This allows
	parametrics to be part of the language. Operations  that can't deal with parametrics can check
	the completeness of the types involved. Operations that can deal with parametrics either
	forward the parametric or instantiate it.
	
	Resolved/Unresolved are the level groups, not concrete ones. Everything >= resolved is considered
	resolved, everything less is unresolved.
*/
enum class extant_level {
	// not enough is known to make progress
	incomplete = 0,
	// enough is known to determine inputs to the expression (in particular arguments to a function call)
	input,
	
	// completely known, but it is parametric
	parametric,
	// complete well-known and static
	complete,
	
	// anything at this level and above is resolved (it need not be, and will not be further evaluated)
	resolved = parametric
};

typedef std::vector<shared_ptr<declaration>> var_vector_t;

//macros for doing checks on statements/expressions
#define STMT_CHECK( stmt_, cond_, ... )	\
	if( !(cond_) ) { auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
		x << error::tag_error_condition( #cond_ ); \
		if( (stmt_).src_loc ) x << util::source_location::error( *(stmt_).src_loc ); \
		BOOST_THROW_EXCEPTION( x ); }
#define STMT_FAIL( stmt_, ... )	{ \
	auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
	if( (stmt_).src_loc ) x << util::source_location::error( *(stmt_).src_loc ); \
	BOOST_THROW_EXCEPTION( x ); }

} //eon leaf

#endif