#ifndef LEAF_CONTEXT_HPP
#define LEAF_CONTEXT_HPP

#include "fwd.hpp"

#include "type_converter.hpp"
#include "type_identifier.hpp"
#include "conversion_applicator.hpp"

namespace leaf {

class typer;

struct job {
	virtual ~job() { }
	virtual bool process( typer * ty ) = 0;
};

/**
	Tracks the context in which typing is being done. This also serves as a builder class
	which provides convenience entry points and common funtions. 
	
	No state is maintained here, it merely wraps the scope and owner. It is important that
	this context can be recreated at any time, copied, etc.
*/
struct context {
	//the scope currently being processed by this context
	shared_ptr<scope> sc;
	
	typer * get_owner() const { 
		return owner; 
	}
	type_converter converter() const { 
		return type_converter( *this );
	}
	conversion_applicator applicator() const {
		return conversion_applicator( *this );
	}
	type_manager & types() const;
	bool no_incomplete() const;
	
	shared_ptr<expression> proc_expression( shared_ptr<expression> expr ) const;
	void add_pending_job( shared_ptr<job> job ) const;
	bool proc_statement_block( shared_ptr<statement_block> block, context ctx ) const;
	bool proc_statement_block( shared_ptr<statement_block> block ) const;
	void scan_symbols( shared_ptr<statement_block> block ) const;
	void proc_statement( shared_ptr<statement> s, shared_ptr<scope> sc ) const;
	
	shared_ptr<statement_block> get_static_init_block();

	shared_ptr<expression> convert_expression( type_ref const & target,
		shared_ptr<expression> source ) const;
	shared_ptr<expression> if_convert_expression( type_ref const & target,
		shared_ptr<expression> source ) const;
	shared_ptr<expression> convert_expression( type_converter::unary_convert target,
		shared_ptr<expression> source ) const;
	shared_ptr<expression> convert_expression( type_converter::binary_convert how,
		shared_ptr<expression> target, shared_ptr<expression> source ) const;
	
	void rewrite( shared_ptr<expression> orig, shared_ptr<expression> replace ) const;
	
	shared_ptr<expr_funcdefn> instance_funcdefn( expr_funcdefn & func, type_ref target ) const;
	
	template<typename T>
	type_identifier::resolve_t determine_type( T const & source, 
		shared_ptr<type_spec> const & spec_type ) const {
		auto rt = type_identifier::determine( *this, spec_type );
		STMT_CHECK( *source, rt.okay(), rt.explain_error() );
		return rt;
	}
	
	//the scope in which a declaration occurred, will be null if not indicated (then use sc)
	shared_ptr<scope> get_decl_scope() const {
		if( decl_sc ) {
			return decl_sc;
		} 
		return sc;
	}
	
	void add_export( shared_ptr<declaration> vd );
	void add_export( shared_ptr<intr_type> type );
	context( typer * owner, shared_ptr<scope> sc, 
		shared_ptr<scope> decl_sc = shared_ptr<scope>() );
	context() = delete;
	context( context const & ) = default;
	
	context derive( shared_ptr<scope> symbols ) const;
private:	
	typer * owner;
	
	shared_ptr<scope> decl_sc;
};

} //eon leaf

#endif //LEAF_CONTEXT_HPP