#ifndef LEAF_declaration_HPP
#define LEAF_declaration_HPP

#include "core_types.hpp"
#include "type.hpp"

namespace leaf
{

struct intr_type;
struct type_spec;

struct declaration : public std::enable_shared_from_this<declaration> {
	enum form_t : int8_t {
		//declaring a variable with a single mapping
		f_single,
		//overloading a variable (implicit declaration)
		f_multi,
		//implicit var_statement for a function argument
		f_arg,
		//implict var_statement for a function return
		f_return,
		//capture variable from previous scope
		f_capture,
	} form;

	//import annotation/style
	std::string import_name;
	enum import_type_t : int8_t {
		//not an import
		it_none,
		//an ABI code import
		it_abi_code,
		//an ABI data import,
		it_abi_data
	} import_type;
	//name of module for the import (blank if undefined -- linker resolved)
	std::string import_module;
	//name mangling for the import/export
	enum name_form_t : int8_t {
		//import_name is the exact name
		nf_bare,
		//module/name are Leaf style (type included for mutli)
		nf_leaf,
		//automatic resolution (scope rules)
		nf_auto,
	} name_form;
	
	//this symbol must be exported
	bool is_export;
	
	//this variable can be resolved automatically when referenced if needed
	bool auto_resolve;
	
	//the lifetime of this declaration
	enum lifetime_t : int8_t {
		//the default, lifetime is tied to containing scope/type
		l_local,
		//lifetime is tied to the module
		l_global,
	} lifetime;
	
	bool is_lifetime_local() const {
		return lifetime == l_local;
	}
	bool is_lifetime_global() const {
		return lifetime == l_global;
	}
	
	//the variable symbol
	std::string symbol;
	//the type of this variable (as provided by the statement)
	type_ref type;
	shared_ptr<type_spec> spec_type;

	//used within the typer to get at original code, may be null
	shared_ptr<statement> source_statement;
	//the scope in which this decl appears (may be zero prior to adding)
	shared_ptr<scope> var_scope;
	
	//if type is literal then this is the expression of that literal
	shared_ptr<expression> literal_value;
	
	//level away from root (just calls var_scope->get_level() )
	unsigned get_root_depth() const;
	unsigned get_capture_level() const;
	
	//usable by compiler/interpreter to create a map (initially 0)
	mutable intptr_t cref;
	mutable intptr_t xref;

	//for f_capture, which the immediate next declaration captured is (an f_capture decl)
	shared_ptr<declaration> proximal_capture;
	//for f_capture the final captured declartion (not an f_capture)
	shared_ptr<declaration> ultimate_capture;
	
	
	bool is_first_class() const {
		return form == f_single || form == f_multi;
	}
	
	void validate(util::validation_context & vctx) const;
	
	shared_ptr<expr_symbol> create_expr_symbol();

	static shared_ptr<declaration> create() {
		return shared_ptr<declaration>(new declaration());
	}
	static shared_ptr<declaration> create( declaration const & o ) {
		return shared_ptr<declaration>(new declaration(o));
	}
private:
	declaration() {
		form = f_single;
		cref = 0;
		xref = 0;
		is_export = false;
		import_type = it_none;
		name_form = nf_bare;
		var_scope = nullptr;
		lifetime = l_local;
		auto_resolve = false;
	}
	declaration( declaration const & o ) :
		std::enable_shared_from_this<declaration>() {
		cref = 0;
		xref = 0;
		import_type = o.import_type;
		form = o.form;
		import_name = o.import_name;
		import_type = o.import_type;
		import_module = o.import_module;
		name_form = o.name_form;
		is_export = o.is_export;
		symbol = o.symbol;
		spec_type = o.spec_type;
		type = o.type;
		var_scope = o.var_scope;
		literal_value = o.literal_value;
		lifetime = o.lifetime;
		proximal_capture = o.proximal_capture;
		ultimate_capture = o.ultimate_capture;
		auto_resolve = o.auto_resolve;
	}
};

} //eon leaf

#endif