#ifndef DUMP_HPP
#define DUMP_HPP

#include "fwd.hpp"
#include "../util/dump_base.hpp"

#include <string>
#include <set>

namespace leaf
{

struct type_ref;

/**
	Converst this into a standard format suitable for checking that
	the parser actually did the right thing. This doesn't generate parsable code
	since we wish to dump additional information. It should however look close
	to real code.
*/
class dump : public leaf_util::dump_base
{
public:
	dump( std::ostream & out, unsigned flags = 0 );
	dump( unsigned flags = 0 );

	void set_show_all_types( bool on ) { 
		show_all_types = on; 
	}
	void set_validate( bool on ) { 
		do_validate = on; 
	}
	void set_show_full_type( bool on ) { 
		show_full_type = on; 
	}
	void set_show_spec_type( bool on ) { 
		show_spec_type = on; 
	}
	void set_hide_var_spec_no_fundamental( bool on ) { 
		hide_var_spec_no_fundamental = on; 
	}
	void set_keep_variant( bool on ) { 
		keep_variant = on; 
	}
	void set_annotate( bool on ) {
		annotate = on;
	}
	void set_expand_defined( bool on ) {
		expand_defined = on;
	}
	void set_omit_value( bool on ) {
		omit_value = on;
	}

	//allow showing implicit settings like default/unset, default is false
	void set_show_spec_implicit( bool on )
	{ show_spec_implicit = on; }
	
	enum type_flags_t {
		tf_expand_defined = 1<<0,
		tf_no_validate = 1 <<1,
		tf_annotate = 1 << 2,
	};
	void write_type_spec( type_spec const & vt )
	{ dump_type_spec( &vt ); }
	void write_type_spec( shared_ptr<type_spec const> vt )
	{ dump_type_spec( vt.get() ); }
	void write_var_statement( var_statement const & vs )
	{ dump_var_statement( &vs ); }
	void write_statement_block( statement_block const & block ) { 
		dump_statement_block( &block ); 
		dump_outline_types();
	}
	void write_statement( statement const & st ) { 
		dump_statement( &st ); 
	}
	void write_declaration( declaration const & vd ) {
		dump_declaration( &vd );
	}
	void write_cvt_expression( cvt_expression const * ce );
	void write_module( module const & mod );
	
	template<typename T>
	static std::string get_impl( T const * obj, unsigned flags = 0 ) {
		dump d( flags );
		d.do_validate = false;
		d.write_impl( obj );
		return d.str();
	}
	
	template<typename T>
	static std::string get( T const * obj, unsigned flags = 0 ) { 
		return get_impl( obj, flags ); 
	}
	
	template<typename T>
	static std::string get( T const & obj, unsigned flags = 0 ) { 
		return get_impl( &obj, flags ); 
	}
	
	template<typename T>
	static std::string get( shared_ptr<T> const & obj, unsigned flags = 0 ) {
		return get_impl( obj.get(), flags ); 
	}
	
	template<typename T>
	void write( T const * obj ) { 
		write_impl( obj ); 
	}
	
	template<typename T>
	void write( T const & obj ) { 
		write_impl( &obj ); 
	}
	
	template<typename T>
	void write( shared_ptr<T> const & obj ) {
		write_impl( obj.get() ); 
	}
	 
private:
	void write_impl( intr_type const * obj );
	void write_impl( type_ref const * obj ) {
		dump_type_ref( *obj );
	}
	void write_impl( extr_type const * obj ) {
		dump_extr_type( *obj ); 
	}
	void write_impl( type_spec const * obj ) {
		dump_type_spec( obj );
	}
	void write_impl( expression const * obj ) {
		dump_expression( obj );
	}
	void write_impl( statement const * obj ) {
		dump_statement( obj );
	}
	void write_impl( statement_block const * obj ) {
		dump_statement_block( obj );
	}
	void write_impl( cvt_expression const * obj ) {
		write_cvt_expression( obj );
	}
	
	void dump_outline_types();
	
	void dump_assign_statement( assign_statement const * as );
	void dump_defer_statement( defer_statement const * defer );
	void dump_do_statement( do_statement const * ds );
	void dump_expression_statement( expression_statement const * es );
	void dump_fail_statement( fail_statement const * fs );
	void dump_for_statement( for_statement const * fs );
	void dump_import_statement( import_statement const * is );
	void dump_init_statement( init_statement const * is );
	
	void dump_type_ref( type_ref const & tr, unsigned flags = 0 );
	void dump_extr_type( extr_type const & tr, unsigned flags = 0 );
	void dump_intr_type( intr_type const & tr, unsigned flags = 0 );
	void dump_intr_type( intr_type const & tr, std::vector<std::string> & markers, unsigned flags = 0 );
	void dump_loop_flow_statement( loop_flow_statement const * ls );
	void dump_noop_statement( noop_statement const * ns );
	void dump_return_statement( return_statement const * rs );
	void dump_statement_block( statement_block const * block );
	void dump_statement( statement const * st );
	void dump_statement_inner( statement const * st );
	void dump_type_spec( type_spec const * vt );
	void dump_typedef_statement( typedef_statement const * ts );
	enum decl_flags_t {
		//don't emit the literal value if there is one
		df_omit_value = 1<<0,
	};
	void dump_declaration( declaration const * vd, unsigned flags = 0 );
	void dump_var_statement( var_statement const * vs );
	void dump_error_statement( error_statement const * es );
	
	enum expr_flags_t {
		//expression is being written on it's own (not a sub-expression)
		ef_isolated = 1 << 0,
		//don't just write the named literal, but the whole expression
		ef_expand_name = 1 <<1,
	};
	void dump_expression( expression const * ex, unsigned flags = 0 );
	void dump_funcdefn( expr_funcdefn const & ef );
	void dump_tuple( expr_with_sub const * ex, char open, char close, char sep );

	bool show_all_types, do_validate, show_full_type, show_spec_type, hide_var_spec_no_fundamental,
		keep_variant, show_spec_implicit, annotate, expand_defined, omit_value;
	//track all items for which the type has been shown
	std::set<void *> shown_type;
	//items which have already been shown
	std::set<void const *> shown;
	
	std::set<intr_type const*> outline_types;

	//show either type or spec_type depending on mode
	void show_type( type_ref type, type_spec const * spec );
	
	template<typename T>
	void dump_intr_type_tuple_inner( const char* o, const char* c, T const * vt, bool nameless = false );
	void _dvtti_sub( shared_ptr<type_spec const> const & vt )
	{ dump_type_spec( vt.get() ); }
	void _dvtti_sub( type_spec const * vt )
	{ dump_type_spec( vt ); }
	void _dvtti_sub( type_ref const & vt )
	{ dump_type_ref( vt ); }
	
	template<typename T>
	void validate( T const & o );
};

} //eon leaf

#endif