#ifndef EXPRESSION_HPP
#define EXPRESSION_HPP

#include "statement.hpp"
#include "type.hpp"

#include "util/number.hpp"
#include "util/unicode.hpp"

#include <boost/optional.hpp>

#include <vector>

namespace leaf
{

struct statement_block;
struct var_statement;

#define EXPRESSION_FORMS \
	EXPRESSION_FORM(binary) \
	EXPRESSION_FORM(bind) \
	EXPRESSION_FORM(block) \
	EXPRESSION_FORM(cast) \
	EXPRESSION_FORM(compat) \
	EXPRESSION_FORM(compound) \
	EXPRESSION_FORM(conditional) \
	EXPRESSION_FORM(current) \
	EXPRESSION_FORM(literal_number) \
	EXPRESSION_FORM(field) \
	EXPRESSION_FORM(funccall) \
	EXPRESSION_FORM(funcdefn) \
	EXPRESSION_FORM(new_func) \
	EXPRESSION_FORM(special_field) \
	EXPRESSION_FORM(subscript) \
	EXPRESSION_FORM(switch) \
	EXPRESSION_FORM(symbol) \
	EXPRESSION_FORM(tag) \
	EXPRESSION_FORM(temp) \
	EXPRESSION_FORM(text) \
	EXPRESSION_FORM(transform) \
	EXPRESSION_FORM(type) \
	EXPRESSION_FORM(unary) \
	EXPRESSION_FORM(value_list)

#define EXPRESSION_PRE_CLONE(t) \
	shared_ptr<t> pre_clone_derived() const { \
		auto c = make_shared<t>(); \
		c->pre_copy( *this ); \
		return c; \
	} \
	virtual shared_ptr<expression> pre_clone() const { \
		return pre_clone_derived(); \
	}

#define EXPRESSION_SERIAL() \
	void write( serial_out & so ) const { \
		write_impl(so); \
	} \
	void read( serial_in & si ) { \
		read_impl(si); \
	}
	
#define EXPRESSION_SERIAL_LITERAL() \
	EXPRESSION_SERIAL() \
	bool export_literal() const { \
		return true; \
	}
	
/**
	Expressions are a series of variables, operators and functions and evaluate to a particular
	value.
	
	SUBSCRIPT: This is not treated like a function call since it has special typing rules and
	special considerations for constants and tuples.
*/
struct expression : public util::source_provider {
	enum form_t
	{
		#define EXPRESSION_FORM(form) f_##form,
		EXPRESSION_FORMS
		#undef EXPRESSION_FORM
	} form;

	static char const * name_of( form_t f );

	//the resulting type of this expression
	type_ref type; //POST-TYPE
	
	//this expression is the initializer for this declaration (may be null)
	shared_ptr<declaration> bound_decl; //POST-TYPE
	
	virtual ~expression() { }
	
	virtual void validate(util::validation_context & vctx) const;
	
	/**
		If the expression has an output and its type is constructable from the inputs
		then do that now. By default this throws since it shouldn't be called unless it
		is known safe to do so.
	*/
	virtual void construct_result_type() { 
		STMT_FAIL(*this); 
	}

	/**
		If the expression is merely a transform of an inner expression this returns
		the number of those inner expressions. If the inner expression is not strictly
		transformed then this should return 0
	*/
	virtual size_t num_inner_trans() const { 
		return 0; 
	}
	virtual shared_ptr<expression> get_inner_trans( size_t /*ndx*/ ) { 
		STMT_FAIL(*this); 
	}
	virtual void set_inner_trans( size_t /*ndx*/, shared_ptr<expression> /*expr*/ ) { 
		STMT_FAIL(*this); 
	}
	
	/**
		@return true if the expression could have side-effects and/or evaluate differently
			each time it is evaluated.
	*/
	virtual bool may_have_side_effects() const { 
		return true; /*safe response*/ 
	}
	
	bool is_resolved() const;
	
	extant_level get_extant() const;
	
	void mark_resolved();
	
	virtual shared_ptr<expression> pre_clone() const {
		STATE_FAIL( "unsupported-pre-clone", typeid(*this) );
	}
	
	virtual void write( serial_out & /*so*/ ) const {
		STATE_FAIL( "unsupported-write", typeid(*this) );
	}
	virtual void read( serial_in & /*so*/ ) {
		STATE_FAIL( "unsupported-read", typeid(*this) );
	}
	virtual bool export_literal() const {
		STATE_FAIL( "unsupported-export-literal", typeid(*this) );
	}
	
	//for the compiler to use
	mutable intptr_t cref;
protected:
	/**
		Pre-copy does pre-typing field copying (to produce a clone from pre_clone). It may
		assume this instance has been default initialized and no other changes.
	*/
	void pre_copy( expression const & o );
	
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & so );
	
	expression( form_t form ) : 
		form( form ) { 
		cref = 0;
		local_resolved = false;
	}
	
private:
	bool local_resolved; //POST-TYPE
};

struct expr_literal_number : public expression
{
	typedef expression super;
	
	expr_literal_number()
		: expression( f_literal_number )
	{ }
	
	 bool may_have_side_effects() const
	 { return false; }
	
	enum src_form_t 
	{
		//pre-typing
		sf_rational,
		sf_boolean,
		//post-typing types
		sf_integer,
		sf_binary,
		sf_float,
	} src_form;
	
	long double float_value() const
	{ return value.as_double(); } //TODO: need long_double
	int64_t integer_value() const
	{ return value.as_int(); }
	bool boolean_value() const
	{ return val_boolean; }
		
	void set_integer_value( int64_t val )
	{ value = util::number::integer(val); }
	
	util::number value;
	bool val_boolean;
	
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_literal_number)
	EXPRESSION_SERIAL_LITERAL()
protected:
	void pre_copy( expr_literal_number const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	Refers to a declaration by symbol name.
	
	This is also introduced during typing to create a reference to a known declaration without a current symbol.
*/
struct expr_symbol : public expression {
	typedef expression super;
	
	expr_symbol()
		: expression( f_symbol )
	{ }
	
	std::string symbol;
	
	//a parameter type specified on the symbol. This is a type_spec since it only affects selection
	//of symbols and resolution of the expression, it's not part of the final symbol
	shared_ptr<type_spec const> param;
	
	bool may_have_side_effects() const
	{ return false; }

	//for symbol (POST-TYPING), to which var_statement this symbol refers
	shared_ptr<declaration> sref;
	 
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_symbol)
	EXPRESSION_SERIAL()
protected:
	void pre_copy( expr_symbol const & o );
	
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	A constant textual expression.
*/
struct expr_text : public expression {
	typedef expression super;
	expr_text()
		: expression( f_text )
	{ }
	
	enum what_t {
		w_string,
		w_char,
	} what;
	
	unicode::codestr text;
	
	EXPRESSION_PRE_CLONE(expr_text)
	EXPRESSION_SERIAL_LITERAL()
protected:
	void pre_copy( expr_text const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

struct expr_with_block : public expression {
	typedef expression super;
	
	shared_ptr<statement_block> block;

	void validate(util::validation_context & vctx) const;
protected:
	expr_with_block( form_t form  )
		: expression( form )
	{ }
	
	void pre_copy( expr_with_block const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	A sequence of code that optionally produces a result.
*/
struct expr_block : public expr_with_block
{
	typedef expr_with_block super;
	
	expr_block()
		: expr_with_block( f_block )
	{ }
	
	//result of the expression, null if it has none (void type)
	shared_ptr<expression> result;
	
	EXPRESSION_PRE_CLONE(expr_block)
protected:
	void pre_copy( expr_block const & o );
};

/**
	The definition of a function.
*/
struct expr_funcdefn : public expr_with_block
{
	typedef expr_with_block super;
	
	expr_funcdefn()
		: expr_with_block( f_funcdefn ) { 
		pend_block = false;
		added_args = false;
		added_returns = false;
		clone = false;
	}

	shared_ptr<type_spec> func_type;
	
	//this a cloned closure, as opposed to a reference closure
	bool clone;
	
	//an entry in arg_defaults is set for each argument which has a default value
	std::map<size_t,shared_ptr<expression>> arg_defaults;
	
	//used for progress in typing
	bool added_args, added_returns, pend_block; //POST-TYPE
	//a list of returns which still need to have conversions done for inferred returns
	struct pend_return {
		shared_ptr<return_statement> rs;
		shared_ptr<scope> sc;
	};
	std::vector<pend_return> pend_returns; //POST-TYPE
	
	bool may_have_side_effects() const {
		return false;
	}
	
	//for parametric functions (similar to setup in intr_type, perhaps this should be a type to share that?)
	shared_ptr<expr_funcdefn> instance_of;
	type_ref applied_type;
	mutable std::vector<shared_ptr<expr_funcdefn>> instances;
	shared_ptr<expr_funcdefn> if_find_instance( type_ref result ) const;
	
	
	EXPRESSION_PRE_CLONE(expr_funcdefn)
	//TODO: direct access to an exported function would be better, it bypasses the
	//variable lookup (slow)
	EXPRESSION_SERIAL()
	bool export_literal() const;
protected:
	void pre_copy( expr_funcdefn const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	A common base for expressions that work on a "base" and access a "field".
*/
struct expr_field_like : public expression {
	typedef expression super;
	
	shared_ptr<expression> base;
	bool optional;
	
	void validate(util::validation_context & vctx) const;
protected:
	expr_field_like( form_t form ) :
		super(form) {
		optional = false;
	}

	void pre_copy( expr_field_like const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	Array indexing/offsetting. 
	
	On input this may also contain field access, since the parser cannot know if the variable
	is an array. Typing however should convert such accesses to "expr_field" and expr_subscript
	is strictly for indexing.
*/
struct expr_subscript : public expr_field_like {
	typedef expr_field_like super;
	
	expr_subscript()
		: super( f_subscript ) { 
		optional = false;
	}
	
	shared_ptr<expression> offset;
	
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_subscript)
	EXPRESSION_SERIAL()
protected:
	void pre_copy( expr_subscript const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	Access to a field of a tuple or custom type.
	
	If `base.type.is_addressable` then the resulting type is an `lvalue`.
*/
struct expr_field : public expr_field_like {
	typedef expr_field_like super;
	
	expr_field() : 
		super( f_field ) {
	}
	
	//the name of the field
	std::string field;
	//a field_index overrides name saying exactly which item
	boost::optional<size_t> field_index;

	//a scope_ref is a dependent field in a custom type
	shared_ptr<declaration> sref; //POST-TYPE
	
	//a parameter type specified on the symbol. This is a type_spec since it only affects selection
	//of symbols and resolution of the expression, it's not part of the final symbol
	//TODO: Make this an operator and remove it from expr_symbol as well
	shared_ptr<type_spec const> param;
	
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_field)
	EXPRESSION_SERIAL()
protected:
	void pre_copy( expr_field const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	Accesses a special field/attribute of a given object.
	
	NOTE: Though the form is basically the same as "transform" or "unary" these classes
	are kept distinct since they convey a different meaning.
	
	POST-TYPE: converted from other forms
*/
struct expr_special_field : public expression {
	expr_special_field() : 
		expression( f_special_field ) { 
	}
	
	enum field_t 	{
		f_array_size,
		f_optional_set,
		f_contextual_context,
		f_contextual_unbound,
	} field;
	shared_ptr<expression> base;
	
	void validate(util::validation_context & vctx) const;
};

struct expr_switch : public expression
{
	typedef expression super;
	
	expr_switch()
		: expression( f_switch )
	{ }
	
	struct expr_list_item
	{
		shared_ptr<expression> cond; //null for an else clause
		shared_ptr<expression> value;
		
		void validate(util::validation_context & vctx) const;
	};
	typedef std::vector<expr_list_item> expr_list_t;
	expr_list_t expr_list;
	
	size_t num_inner_trans() const
	{ return expr_list.size(); }
	shared_ptr<expression> get_inner_trans( size_t ndx )
	{ return expr_list[ndx].value; }
	void set_inner_trans( size_t ndx, shared_ptr<expression> expr )
	{ expr_list[ndx].value = expr; }
	
	bool may_have_side_effects() const;
	void construct_result_type();
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_switch)
protected:
	void pre_copy( expr_switch const & o );
};

struct expr_with_sub : public expression
{
	typedef expression super;

	struct sub_ref {
		std::string name;
		shared_ptr<expression> expr;
	};
	std::vector<sub_ref> sub;

	void push_sub( std::string name, shared_ptr<expression> expr ) {
		sub.push_back( sub_ref{ name, expr } );
	}
	void push_sub( shared_ptr<expression> expr ) {
		sub.push_back( sub_ref{ "", expr } );
	}

	 bool may_have_side_effects() const;

	void validate(util::validation_context & vctx) const;
protected:
	expr_with_sub( form_t form )
		: expression( form )
	{ }
	
	void pre_copy( expr_with_sub const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

struct expr_value_list : public expr_with_sub
{	
	typedef expr_with_sub super;
	
	expr_value_list()
		: expr_with_sub( f_value_list )
	{ }
	
	EXPRESSION_PRE_CLONE(expr_value_list)
	EXPRESSION_SERIAL_LITERAL()
};

struct expr_compound : public expr_with_sub
{
	typedef expr_with_sub super;
	
	expr_compound()
		: expr_with_sub( f_compound )
	{ }
	
	EXPRESSION_PRE_CLONE(expr_compound)
	EXPRESSION_SERIAL_LITERAL()
};

struct expr_funccall : public expression
{
	typedef expression super;
	
	expr_funccall()
		: expression( f_funccall )
	{ }
	
	shared_ptr<expression> func, args;
	//for explicit context/member functions
	shared_ptr<expression> context;
	
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_funccall)
	EXPRESSION_SERIAL()
protected:
	void pre_copy( expr_funccall const & o );
	
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

/**
	Creates a bound result (function) from a context and unbound function.
	
	POST-TYPE: created from field access calls (will eventually be allowed as explicit)
*/
struct expr_bind : public expression {
	typedef expression super;
	
	expr_bind() :
		expression( f_bind ) {
		clone = false;
	}
	
	//create a clone binding
	bool clone;
	
	//if null then binding a raw function
	shared_ptr<expression> context;
	shared_ptr<expression> unbound;
	
	void validate(util::validation_context & vctx) const;
	void construct_result_type();
};

/**
	Transforms change the type of the variable without changing the underlying semantic
	value.
	
	POST-TYPE: create from conversions and builtin functions
*/
struct expr_transform : public expression
{
	expr_transform()
		: expression( f_transform )
	{ }
	
	enum transform_t
	{
		t_deref,
		t_drop_lvalue,
		//indicates that "arg" exists multiple times in this statement but should be evaluated only once
		t_shared_expr,
		//indicates a basic promotions/conversion of a fundmental from arg->type to this->type
		t_promote,
		//indicates a basic loss conversion of a fundamental from arg->type to this->type
		t_loss,
		//create an optional from "arg"
		t_make_optional,
		//deref an optional
		t_deref_optional,
		//get the value_ptr to value, or array_ptr of array
		t_addr_of,
		//integer of char code point
		t_code_val,
		//char code point of integer
		t_char_val,
	} transform;
	
	shared_ptr<expression> arg;

	size_t num_inner_trans() const
	{ return transform == t_promote ? 0 : 1; }
	shared_ptr<expression> get_inner_trans( size_t ndx ) { 
		STMT_CHECK( *this, ndx == 0 ); 
		return arg; 
	}
	void set_inner_trans( size_t ndx, shared_ptr<expression> expr ) {
		STMT_CHECK( *this, ndx == 0 );
		 arg = expr; 
	}
	
	void construct_result_type();
	void validate(util::validation_context & vctx) const;
};

/**
	Constructs a temporary object of a given type, optionally initializing it.
	
	POST-TYPE: this exists only post-typing, usually as a result of a `expr_new`
*/
struct expr_temp : public expression {
	typedef expression super;
	
	expr_temp() :
		expression( f_temp ) {
	}
	
	enum init_t {
		i_none,
		i_default,
		i_intrinsic,
	} init;
	
	static const init_t init_min_ = i_none;
	static const init_t init_max_ = i_intrinsic;
	
	shared_ptr<expression> intrinsic_value;
	
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_temp)
protected:
	void pre_copy( expr_temp const & o );
};
	
/**
	Provides access to a new function for a given type.
*/
struct expr_new_func : public expression {
	typedef expression super;
	
	expr_new_func()
		: expression( f_new_func ) {
	}
	
	type_ref new_type;
	shared_ptr<type_spec> new_spec_type;
	
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_new_func)
protected:
	void pre_copy( expr_new_func const & o );
};

/**
	A family of conditional expressions. These are guaranteed lazy-evaluated of
	unneeded components.
*/
struct expr_conditional : public expression
{
	typedef expression super;
	
	expr_conditional()
		: expression( f_conditional )
	{ }
	
	enum op_t {
		//if 'cond' is set then `unopt(cond)` (set's this to `a` on output), otherwise 'b'
		o_optional_or,
		//if 'cond' is true then optional set to a, otherwise unset optional
		o_optional_cond,
	} op;
	static const op_t min_op_ = o_optional_or;
	static const op_t max_op_ = o_optional_cond;
	
	shared_ptr<expression> cond, a, b;
	
	size_t num_inner_trans() const;
	shared_ptr<expression> get_inner_trans( size_t ndx );
	void set_inner_trans( size_t ndx, shared_ptr<expression> expr );
	
	void construct_result_type();
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_conditional)
protected:
	void pre_copy( expr_conditional const & o );
};

/**
	A fundamental scalar binary operation.
	
	Operations are marked to work on (i)integers, (b)binary, (f)floats, (x)boolean, (c)char,  on all
	sizes. The type of both arguments and the result will be exactlythe same.
	(t)tags support all equality for the purpose of algorithms, not because they have any natural ordering.
	
	POST-TYPE: result from builtin function calls
*/
struct expr_binary : public expression
{
	expr_binary()
		: expression( f_binary )
	{ }
	
	enum op_t {
		o_add, //ifb
		o_sub, //ifb
		o_mul, //ifb
		o_div, //ifb
		
		o_mod, //i
		
		o_bit_shl, //b
		o_bit_shr, //b
		
		o_or, //xb
		o_and, //xb
		o_xor, //xb
		o_and_not, //xb
		o_nor, //xb
		o_nand, //xb
		
		//boolean results
		o_eq, //ifbxct
		o_neq, //ifbxct
		o_lt, //ifbxct
		o_lte, //ifbxct
		o_gt, //ifbxct
		o_gte, //ifbxct
		
	} op;
	static const op_t min_op_ = o_add;
	static const op_t max_op_ = o_gte;
	bool is_boolean_result() const {
		return op >= o_eq && op <= o_gte;
	}
	
	shared_ptr<expression> a, b;
	
	void validate(util::validation_context & vctx) const;
	
	//doesn't exist pretyping, thus shouldn't need a PRE_CLONE
};

/**
	A fundamental scalar unary operation.
	
	POST-TYPE: result of builtin function calls
*/
struct expr_unary : public expression {
	expr_unary() : 
		expression( f_unary ) { 
	}
	
	enum op_t {
		o_neg, //if
		o_abs, //if

		o_not, //xb
	} op;
	static const op_t min_op_ = o_neg;
	static const op_t max_op_ = o_not;
	
	shared_ptr<expression> arg;
	
	void validate(util::validation_context & vctx) const;
	
	//doesn't exist pretyping, thus shouldn't need a PRE_CLONE
};

/**
	A value based on the current scope/context.
*/
struct expr_current : public expression {
	typedef expression super;
	
	expr_current() :
		expression( f_current ) {
	}
	
	enum what_t {
		//current class object PRE-TYPE only
		w_this,
		//context at given depth
		w_context,
	} what;
	static const what_t min_what_ = w_this;
	static const what_t max_what_ = w_context;
	
	shared_ptr<scope> which_scope;
	int depth;
	
	void validate(util::validation_context & vctx) const;
	
	EXPRESSION_PRE_CLONE(expr_current)
protected:
	void pre_copy( expr_current const & o );
};

struct expr_cast : public expression {
	typedef expression super;
	
	expr_cast() :
		expression( f_cast ) {
	}
	
	//the source expression
	shared_ptr<expression> source;
	//the desired type ths expression
	shared_ptr<type_spec> spec_type;

	void validate( util::validation_context & vctx ) const;
	
	EXPRESSION_PRE_CLONE(expr_cast)
protected:
	void pre_copy( expr_cast const & o );
};

struct expr_compat : public expression {
	typedef expression super;
	
	expr_compat() :
		expression( f_compat ) {
	}
	
	//the source expression
	shared_ptr<expression> source;
	
	void validate( util::validation_context & vctx ) const;
	
	EXPRESSION_PRE_CLONE(expr_compat)
protected:
	void pre_copy( expr_compat const & o );
};

struct expr_type : public expression {
	typedef expression super;
	
	expr_type() :
		expression( f_type ) {
	}
	
	//the inner_type of the type instance being created: type｢inner｣
	shared_ptr<type_spec const> inner_type_spec;
	type_ref inner_type;
	
	void validate( util::validation_context & vctx ) const;
	
	EXPRESSION_PRE_CLONE(expr_type)
	EXPRESSION_SERIAL()
protected:
	void pre_copy( expr_type const & o );
	
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};	

/**
	Creates a tagged value. This is a literal which is stored in the code segment on generation
	and can be uniquely identified by the pointer to this data.
	Only an array｢char｣ is supported for now.
*/
struct expr_tag : public expression {
	typedef expression super;
	
	expr_tag() :
		expression( f_tag ) {
	}
	
	shared_ptr<expression> value;
	
	void validate( util::validation_context & vctx ) const;
	
	EXPRESSION_PRE_CLONE(expr_tag)
	EXPRESSION_SERIAL()
protected:
	void pre_copy( expr_tag const & o );
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & si );
};

} //eon leaf

#endif