#ifndef LANG_TYPE_HPP
#error include <lang/type.hpp>
#endif

#ifndef LANG_EXTR_TYPE_HPP
#define LANG_EXTR_TYPE_HPP

namespace leaf {

struct intr_type;

/**
	An extrinsic type adds modified to an intrinsic type. These are modifiers that change
	the access and conversion rules but do not introduce new intrinsic types.
*/
struct extr_type : public type_node {
	//lvalue wraps optional
	bool lvalue;
	//optional wraps reference
	bool optional;
	
	enum reference_t : int8_t {
		//stores directly the fun type
		r_value = 0,
		//a shared reference to the value
		r_shared,
		//an unsafe pointer to a value
		r_value_ptr,
	} reference;

	/**
		Lossy indicates the type is allowed to convert to another with a series of defined
		loss-of-precision/range rules.
	*/
	bool lossy;
	
	//the optional allows underlying conversion
	bool cvt_optional;
	
	//the entire value/reference is literal (known at compile-time)
	bool literal;
	
	shared_ptr<intr_type const> intr; //may not be null

	explicit extr_type( serial_in & si );
	explicit extr_type( shared_ptr<const intr_type> o ) :
		intr( o ) 	{ 
		lvalue = false;
		optional = false;
		cvt_optional = false;
		lossy = false;
		reference = r_value;
		literal = false;
	}
	extr_type() = delete;
	extr_type( extr_type const & ) = default;
	
	bool is_plain() const { 
		return !is_refer() && !optional && !lossy && !lvalue; 
	}

	bool is_lvalue() const {
		return lvalue;
	}
	void set_lvalue( bool s ) {
		lvalue = s;
	}
	
	bool is_optional() const {
		return optional;
	}
	void set_optional( bool s ) {
		optional = s;
	}
	
	bool is_literal() const {
		return literal;
	}
	void set_literal( bool s ) {
		literal = s;
	}
	
	bool is_cvt_optional() const {
		return cvt_optional;
	}
	void set_cvt_optional( bool s ) {
		cvt_optional = s;
	}
	
	bool is_lossy() const {
		return lossy;
	}
	void set_lossy( bool s) {
		lossy = s;
	}
	
	void set_ref_shared() { 
		reference = r_shared; 
	}
	void set_ref_value_ptr() {
		reference = r_value_ptr;
	}
	void set_ref_value() {
		reference = r_value;
	}
	void set_reference( reference_t r ) {
		reference = r;
	}
	reference_t get_reference() const {
		return reference;
	}

	bool is_fixated_impl( util::validation_context & vctx ) const;
	extant_level get_extant_impl( util::validation_context & vctx ) const;
	bool contains_parametric_impl( util::validation_context & ctx ) const;
	
	bool is_value() const {
		return reference == r_value;
	}
	bool is_value_ptr() const {
		return reference == r_value_ptr;
	}
	bool is_shared() const { 
		return reference == r_shared; 
	}
	bool is_refer() const { 
		return is_shared() || is_value_ptr(); 
	}
	bool is_addressable() const {
		return lvalue || is_refer();
	}

	bool is_decorated() const {
		return lvalue ||  optional || lossy || cvt_optional || literal;
	}
	
	bool allow_copy() const;
	
	///////////////////////////////////////////////////////
	// Simpilfied casting support
	template<typename T>
	T const & cast() const;
	template<typename T>
	T const * if_cast() const;
	template<typename T>
	bool is() const;
	
	void validate(util::validation_context & vctx) const;
	
	int compare( extr_type const& o, bool logical ) const;
	int compare_impl( extr_type const & o, bool logical, util::validation_context & vctx ) const;
	bool operator<( extr_type const & o ) const { 
		return compare( o, false ) < 0;
	}
	bool operator==( extr_type const& o ) const { 	
		return compare( o, false ) == 0;
	}
	bool operator!=( extr_type const & o ) const {
		return compare( o, false ) != 0;
	}
	bool same_as( extr_type const & o ) const {
		return compare( o, true ) == 0;
	}
	
	void write( serial_out & so ) const;
	void read( serial_in & si );
	
	std::string get_qualified_id() const;	
};

} //eon leaf

#endif