#ifndef LANG_TYPE_HPP
#error include <lang/type.hpp>
#endif

#ifndef intr_type_HPP
#define intr_type_HPP

#include "platform/platform.hpp"
#include "fwd.hpp"

#include <boost/optional.hpp>

#include <vector>

namespace util {
	struct validation_context;
}

/*
	CPP: This hierarchy is a very good example of the problem of creating a virtual type hierarchy
	in C++. It's a real pain to provide clone functions, comparisons, calling super, etc. Most of the
	code here is boilerplate.
*/
namespace leaf
{

class serial_out;
class serial_in;
class scope;

#define INTR_TYPE_CLONE(t) \
	virtual shared_ptr<intr_type> clone() const override \
	{ return make_shared<t>( *this ); } \
	shared_ptr<t> clone_derived() const \
	{ return std::static_pointer_cast<t>( clone() ); }

#define INTR_TYPE_CLONE_DERIVED(t) \
	shared_ptr<t> clone_derived() const \
	{ return std::static_pointer_cast<t>( clone() ); }

/**
	Defines an intrinsic type. This represents the data which is actually being stored, the value of
	a variable.
	
	IMMUTABLE: Post-init intr_type is assumed to be immutable by all other users. This allows
	"clone" to do a shallow copy and avoids several needs of "clone" itself.
*/
struct intr_type : public type_node, public std::enable_shared_from_this<intr_type> {
	friend struct type_ref;
	
	//used to repeat boilerplatr functions
	#define INTR_TYPE_FUN_CLASSES \
		FUN_CLASS_TYPE(array) \
		FUN_CLASS_TYPE(binary) \
		FUN_CLASS_TYPE(boolean) \
		FUN_CLASS_TYPE(char) \
		FUN_CLASS_TYPE(class) \
		FUN_CLASS_TYPE(closure) \
		FUN_CLASS_TYPE(compat) \
		FUN_CLASS_TYPE(contextual) \
		FUN_CLASS_TYPE(float) \
		FUN_CLASS_TYPE(function) \
		FUN_CLASS_TYPE(integer) \
		FUN_CLASS_TYPE(module) \
		FUN_CLASS_TYPE(rational) \
		FUN_CLASS_TYPE(service) \
		FUN_CLASS_TYPE(tuple) \
		FUN_CLASS_TYPE(type) \
		FUN_CLASS_TYPE(value_list)
		
	enum fun_class_t : int8_t {
		#define FUN_CLASS_TYPE(s) f_##s,
		INTR_TYPE_FUN_CLASSES
		#undef FUN_CLASS_TYPE
	};
	static fun_class_t const f_invalid = (fun_class_t)-1;
	
	fun_class_t get_fun_class() const { 
		return fun_class; 
	}

	static char const* name_of( fun_class_t f );
	
	enum style_t : int8_t {
		s_plain,
		/* A core Leaf type that is always available. These are built using the core intrinsics but can
			be consider a builtin Leaf type. These should be quite limited, only things fundamentally
			required by the core typer. */
		s_inherent,
		/* A distinct type, it is not compatible with another distinct type with an otherwise identical
			instrinsic definition. */
		s_distinct,
	};
	style_t style;
	
	/**
		The type should be treated as a whole and not broken down into parts for the purpose of display
		or comparison.
		
		Most comparisions of style should use this function instead of checking `style` directly. Other 
		than for exporting there shouldn't be a significant distinction between inherent and distinct.
	*/
	bool is_whole() const {
		return style == s_distinct || style == s_inherent;
	}
	
	std::string get_qualified_id() const;
	virtual std::string get_qualified_id_part() const = 0;
	
	bool has_distinct_name() const {
		return !!distinct_name;
	}
	std::string get_distinct_name() const {
		STATE_CHECK( !!distinct_name );
		return *distinct_name;
	}
	void set_distinct_name( std::string const & name ) {
		distinct_name = name;
	}
	void clear_distinct_name() {
		distinct_name = boost::optional<std::string>();
	}
	
	/**
		The type is marked as exported. Note that this is only relevant for distinct types, since
		non-distinct ones can always be exported.
	*/
	bool is_export;
	
	//in which scope was this named type declared (may be null prior to scope addition)
	mutable shared_ptr<scope> name_scope;
	
	/**
		A type may exist prior to being completed. This allows forward references and recursive
		types to be supported. Incomplete types can only be used in special circumstances.
	*/
	bool local_complete;

	//type_node
	extant_level get_extant_impl( util::validation_context & ) const {
		return !local_complete ? extant_level::incomplete :
			params.size()>0 ? extant_level::parametric : extant_level::complete;
	}
	
	bool contains_parametric_impl( util::validation_context & ) const {
		return params.size() > 0;
	}
	void validate( util::validation_context & vctx ) const;
	//end of type_node
	
	~intr_type();

	virtual shared_ptr<intr_type> clone() const = 0;
	
	/**
		Compares the types. This does a logical type comparison and will be equal only if the types 
		are exactly the same (this includes names).
		
		@param logical if true then extrinsic non-value markes are ignored. These don't affect the
			value or storage but merely track knowledge of the type.
	*/
	int compare ( intr_type const & o, bool logical ) const;
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const;
	
	bool operator<( intr_type const & o ) const
	{ return compare( o, false ) < 0; }
	bool operator==(intr_type const& o ) const
	{ return compare( o, false ) == 0; }
	bool operator!=(intr_type const& o ) const
	{ return compare( o, false ) != 0; }
	bool same_as(intr_type const& o) const
	{ return compare( o, true ) == 0; }
	

	struct param_spec {
		std::string name;
	};
	//the unspecified parametric parts of the type
	std::vector<param_spec> params;
	//a list of all instances of this type
	mutable std::vector<shared_ptr<intr_type const>> instances;
	shared_ptr<intr_type const> if_find_instance( std::vector<extr_type> const & params ) const;
	
	//an instance of this parametric type
	shared_ptr<intr_type const> instance_of;
	//with these parameters applies
	std::vector<extr_type> applied_params;
	
	bool is_instance() const {
		return !!instance_of;
	}
	
	void clear_param();

	virtual bool allow_copy() const {
		return true;
	}
	
	///////////////////////////////////////////////////////
	// Simpilfied casting support
	template<typename T>
	bool is() const {
		return T::is_fun_class( fun_class );
	}
	template<typename T>
	T & cast() {
		PRE_CHECK( T::is_fun_class( fun_class ) );
		return static_cast<T&>(*this); 
	}
	template<typename T>
	T * if_cast() {
		if( T::is_fun_class( fun_class ) )
			return static_cast<T*>(this); 
		return nullptr;
	}
	template<typename T>
	T const & cast() const {
		PRE_CHECK( T::is_fun_class( fun_class ) );
		return static_cast<T const&>(*this); 
	}
	template<typename T>
	T const * if_cast() const {
		if( T::is_fun_class( fun_class ) )
			return static_cast<T const*>(this); 
		return nullptr;
	}

	virtual void write( serial_out & so ) const;
	virtual void read( serial_in & si );
	
	//intrusive values usable by the compiler
	mutable intptr_t cref;
protected:
	intr_type( fun_class_t fun_class );
	
	void complete_with( intr_type const & o );
private:
	intr_type() = delete;
	fun_class_t fun_class;
	boost::optional<std::string> distinct_name;
};

struct intr_type_scalar : public intr_type {
	static bool is_fun_class( fun_class_t val ) {
		return val == f_rational || val == f_integer || val == f_binary || val == f_float || val == f_boolean
			|| val == f_char;
	}
	
protected:
	intr_type_scalar( fun_class_t f ) :
		intr_type( f ) {
	}
};

struct intr_type_char final : public intr_type_scalar {
	typedef intr_type_scalar super;

	intr_type_char()
		: super( f_char ) {
	}
	INTR_TYPE_CLONE(intr_type_char)
	
	static bool is_fun_class( fun_class_t val ) {
		return val == f_char;
	}
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
	std::string get_qualified_id_part() const override {
		return "$C";
	}
};


struct intr_type_boolean final : public intr_type_scalar {
	typedef intr_type_scalar super;
	
	intr_type_boolean() : 
		super( f_boolean ) {
	}
	INTR_TYPE_CLONE(intr_type_boolean)
	
	static bool is_fun_class( fun_class_t val ) {
		return val == f_boolean;
	}
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
	std::string get_qualified_id_part() const override {
		return "$B";
	}
};

struct intr_type_number : public intr_type_scalar
{
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_rational || val == f_integer || val == f_binary || val == f_float; 
	}
	
protected:
	intr_type_number( fun_class_t f )
		: intr_type_scalar( f )
	{ }
};

struct intr_type_rational final : public intr_type_number
{
	typedef intr_type_number super;
	static bool is_fun_class( fun_class_t val ) { return val == f_rational; }
	
	int compare_impl( intr_type const& o, bool logical, util::validation_context & vctx ) const override;
	intr_type_rational()
		: intr_type_number( f_rational )
		, integer_signed( false )
	{}
	INTR_TYPE_CLONE(intr_type_rational)
	
	//if it is an integer it has a bitsize which can store that integer
	boost::optional<int> integer_bitsize;
	bool integer_signed;

	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	
	bool is_fixated_impl(util::validation_context &) const {
		return false;
	}
	
	std::string get_qualified_id_part() const override;		
};

struct intr_type_fixed_number : public intr_type_number
{
	typedef intr_type_number super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_integer || val == f_float || val == f_binary; 
	}
	
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const override;
	virtual void validate(util::validation_context & vctx) const;
	
	int data_bitsize;
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
	
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	
	INTR_TYPE_CLONE_DERIVED(intr_type_fixed_number)
protected:
	intr_type_fixed_number( fun_class_t f )	
		: intr_type_number( f ) { 
	}
	
	std::string get_qualified_id_end() const;
};

struct intr_type_integral : public intr_type_fixed_number {
	typedef intr_type_fixed_number super;
	
	static bool is_fun_class( fun_class_t val ) {
		return val == f_integer || val == f_binary;
	}
	
	INTR_TYPE_CLONE_DERIVED(intr_type_integral)
protected:
	intr_type_integral( fun_class_t f ) :
		intr_type_fixed_number( f ) {
	}
};

struct intr_type_integer final : public intr_type_integral {
	typedef intr_type_integral super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_integer; 
	}

	intr_type_integer( int data_bitsize = platform::natural_integer_size ) : 
		super( f_integer ) { 
		this->data_bitsize = data_bitsize;
	}
	INTR_TYPE_CLONE(intr_type_integer)
	
	std::string get_qualified_id_part() const override {
		return std::string("$I") + super::get_qualified_id_end();
	}
};

struct intr_type_binary final : public intr_type_integral {
	typedef intr_type_integral super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_binary; }
	intr_type_binary( ) : 
		super( f_binary ) { 
		data_bitsize = platform::natural_integer_size;
	}
	INTR_TYPE_CLONE(intr_type_binary)
	
	std::string get_qualified_id_part() const override {
		return std::string( "$X" ) + super::get_qualified_id_end();
	}
};

struct intr_type_float final : public intr_type_fixed_number
{
	typedef intr_type_fixed_number super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_float; }
	intr_type_float( )
		: intr_type_fixed_number( f_float )
	{ 
		data_bitsize = platform::natural_float_size;
	}
	
	virtual void validate(util::validation_context & vctx) const;
	INTR_TYPE_CLONE(intr_type_float)
	
	std::string get_qualified_id_part() const override {
		return std::string( "$F" ) + super::get_qualified_id_end();
	}
};

struct intr_type_compound : public intr_type
{
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_tuple || val == f_value_list || val == f_contextual;
	}
	virtual int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const override;
	virtual void validate(util::validation_context & vctx) const;

	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;

	struct sub_ref {
		std::string name;
		type_ref type;
	};
	std::vector<sub_ref> sub;
	
	void push_sub( std::string name, type_ref type ) {
		sub.push_back( sub_ref{ name, type } );
	}
	void push_sub( type_ref type ) {
		sub.push_back( sub_ref{ "", type } );
	}
	
	boost::optional<size_t> find_name( std::string const & name ) const;
	
	bool is_fixated_impl(util::validation_context & vctx) const;
	extant_level get_extant_impl(util::validation_context & vctx) const;
	bool contains_parametric_impl(util::validation_context & vctx) const;
	
	INTR_TYPE_CLONE_DERIVED(intr_type_compound)
protected:
	intr_type_compound( fun_class_t f )
		: intr_type( f )
	{ }
	
	void complete_with( intr_type_compound const & o );
	std::string get_qualified_id_end() const;
};

struct intr_type_tuple final : public intr_type_compound
{
	typedef intr_type_compound super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_tuple; }

	intr_type_tuple()
		: intr_type_compound( f_tuple )
	{ 
		pack = p_normal;
		selectable = false;
	}
	
	enum pack_t : int8_t {
		p_normal,
		p_abi,
		p_byte,
	} pack;
	INTR_TYPE_CLONE(intr_type_tuple)
	
	//indicates that in type conversion this tuple can convert to any of types
	//of the elements of the tuple
	bool selectable;
	
	//complets an incomplete tuple with another definition
	void complete_with( intr_type_tuple const & o );
	
	void validate(util::validation_context & vctx) const;
	int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const override;
	//TODO: needing this implies intr_type_contextual is *not* a intr_type_compound!
	extant_level get_extant_impl( util::validation_context & ) const;

	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	std::string get_qualified_id_part() const override;	
};

struct intr_type_value_list final : public intr_type_compound {
	typedef intr_type_compound super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_value_list; }
	intr_type_value_list()
		: intr_type_compound( f_value_list )
	{ }
	
	void validate(util::validation_context & vctx) const;
	INTR_TYPE_CLONE(intr_type_value_list)
	std::string get_qualified_id_part() const override {
		return std::string( "$V" ) + super::get_qualified_id_end();
	}
};

struct intr_type_scope : public intr_type {
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) {
		return val == f_class || val == f_service || val == f_closure || val == f_module;
	}
	
	//the conext for this scope type
	shared_ptr<scope const> context;
	
	bool is_fixated_impl(util::validation_context &) const { 
		return true; //TODO: assumed for now (would need to handle recursion)
	}
	
	int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const override;
	virtual void validate(util::validation_context & vctx) const;
	virtual extant_level get_extant_impl( util::validation_context & vctx ) const {
		auto b = super::get_extant_impl( vctx );
		if( context ) {
			b = std::max( extant_level::input, b );
		}
		return b;
	}
	
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	std::string get_qualified_id_part() const override;
	
	INTR_TYPE_CLONE_DERIVED(intr_type_scope)
protected:
	intr_type_scope() = delete;
	intr_type_scope( fun_class_t f ) 
		: intr_type( f )
	{ }
};

/**
	Custom types are backed by a scope and may have a parent and destructor.
	This common base to class,closure exists because those two behave identical
	in most cases. There are just a few restrictions about accessing members
	and exporting that are different. They are kept as distinct classes to reinforce
	that the user should not know how closures are implemented.
*/
struct intr_type_custom : public intr_type_scope {
	typedef intr_type_scope super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_class || val == f_closure || val == f_service; 
	}
	
	//custom destroy code (called prior to builtin cleanup code)
	shared_ptr<declaration> destroy; //may be null
	//parent context, may be undefined
	type_ref parent;
	
	virtual void validate(util::validation_context & vctx) const;
	
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	
	//return the declarations that form the content of the type
	virtual var_vector_t get_type_decls() const = 0;
protected:
	intr_type_custom() = delete;
	intr_type_custom( fun_class_t f ) : 
		intr_type_scope( f ) { 
	}
};

struct intr_type_classlike : public intr_type_custom {
	typedef intr_type_custom super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_class || val == f_service; 
	}
	
	//block of code associated with class (may be null for non-parametrics)
	shared_ptr<statement_block> block;
	var_vector_t get_type_decls() const;
	
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	
	void validate(util::validation_context & vctx) const;
	
protected:
	intr_type_classlike() = delete;
	intr_type_classlike( fun_class_t f ) : 
		super( f ) { 
	}
};

struct intr_type_class final : public intr_type_classlike {
	typedef intr_type_classlike super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_class; 
	}
	intr_type_class() :
		super( f_class ) {
	}
	
	INTR_TYPE_CLONE(intr_type_class);
};

struct intr_type_service final : public intr_type_classlike {
	typedef intr_type_classlike super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_service; 
	}
	intr_type_service() :
		super( f_service ) {
	}
	
	bool allow_copy() const override {
		return false;
	}
	
	INTR_TYPE_CLONE(intr_type_service);
};

struct intr_type_closure final : public intr_type_custom {
	typedef intr_type_custom super;
	
	static bool is_fun_class( fun_class_t val ) { 
		return val == f_closure;
	}
	intr_type_closure() :
		intr_type_custom( f_closure ) {
	}
	//TODO: rename this somehow to mention "clone"
	enum what_t {
		w_share_context,
		w_captures,
	} what;
	
	INTR_TYPE_CLONE(intr_type_closure);
	
	var_vector_t get_type_decls() const;
};

struct intr_type_module final : public intr_type_scope {
	static bool is_fun_class( fun_class_t val ) { return val == f_module; }
	intr_type_module() : 
		intr_type_scope( f_module ) { 
	}
	
	INTR_TYPE_CLONE(intr_type_module)
};

/**
	This should ideally be instanced with an `instance_of` some primordial array type so normal
	parametric typing appears on it. It must retain `sub` though, since `applied_params` is for the
	whole type and does not denote location (note that a parametric tuple still has concrete types in it
	in addition to `applied_params`).
*/
struct intr_type_wrap : public intr_type {
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_array; }
	int compare_impl( intr_type const & o, bool logical,util::validation_context & vctx ) const override;
	virtual void validate(util::validation_context & vctx) const;
	void write( serial_out & so ) const override;
	virtual void read( serial_in & si ) override;
	type_ref sub;

	bool is_fixated_impl(util::validation_context & vctx) const { 
		return sub.is_fixated_impl(vctx); 
	}
	
protected:
	intr_type_wrap( fun_class_t f ) : 
		intr_type( f )
	{ }
	
	std::string get_qualified_id_end() const;
};

struct intr_type_array final : public intr_type_wrap {
	typedef intr_type_wrap super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_array; }
	intr_type_array() : 
		intr_type_wrap( f_array ) { 
		raw = false;
	}
	
	//this is a raw array pointer as opposed to a managed/counted array
	bool raw;
	
	INTR_TYPE_CLONE(intr_type_array)
	int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const override;
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	std::string get_qualified_id_part() const override;
};

/**
	A runtime type.
*/
struct intr_type_type final : public intr_type_wrap {
	typedef intr_type_wrap super;
	static bool is_fun_class( fun_class_t val ) {
		return val == f_type;
	}
	
	intr_type_type() : 
		super( f_type ) {
	}
	INTR_TYPE_CLONE(intr_type_type)
	int compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const override;
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	std::string get_qualified_id_part() const override;
};

/**
	Note that since functions are immutable there is no distinction between the
	type of the function and a reference to a function (at this level at least, in
	the IR that changes).
*/
struct intr_type_function final : public intr_type
{
	typedef intr_type super;
	
	static bool is_fun_class( fun_class_t val ) { return val == f_function; }
	intr_type_function()
		: intr_type( f_function )
		, can_fail(true)
		, access( a_bound ) { 
	}
	
	shared_ptr<intr_type_value_list const> args;
	//may be null if implied (an incomplete type)
	shared_ptr<intr_type_value_list const> returns;
	//if defined must be a extr_type/intr_type_custom defining the context
	type_ref context;
	//whether this function can fail (raise an error) or not
	bool can_fail;
	//for a_construct indicates the resulting type of construction, as it may have different extrinsic modifiers than `context`
	type_ref construct_result;
	
	enum access_t {
		//the type has a bound context
		a_bound,
		//the function has no context
		a_raw,
		//an explicit context must be passed to the function
		a_unbound,
		//a constructor for the context (type of unbound)
		a_construct,
	} access;
	
	bool has_context() const {
		return access != a_raw;
	}
	INTR_TYPE_CLONE(intr_type_function)
	
	bool is_bound() const { 
		return access == a_bound; 
	}
	bool is_raw() const { 
		return access == a_raw; 
	}
	bool is_unbound() const { 
		return access == a_unbound || access == a_construct; 
	}
	
	bool is_fixated_impl(util::validation_context & vctx) const;
	int compare_impl( intr_type const & o, bool logical,util::validation_context & vctx ) const override;
	void validate(util::validation_context & vctx) const;
	extant_level get_extant_impl( util::validation_context & vctx ) const;
	bool contains_parametric_impl( util::validation_context & vctx ) const;
	
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	std::string get_qualified_id_part() const override;
};

/**
	The contextual type wraps a context and an unbound type together.
*/
struct intr_type_contextual final : public intr_type_compound {
	typedef intr_type_compound super;
	
	static bool is_fun_class( fun_class_t val ) {
		return val == f_contextual;
	}
	intr_type_contextual() :
		super( f_contextual ) {
		sub.resize(2);
	}
	
	type_ref & get_context() { 
		return sub.at(0).type;
	}
	type_ref & get_unbound() {
		return sub.at(1).type;
	}
	type_ref const & get_context() const { 
		return sub.at(0).type;
	}
	type_ref const & get_unbound() const {
		return sub.at(1).type;
	}
	
	INTR_TYPE_CLONE(intr_type_contextual)
	void validate( util::validation_context & vctx ) const;
	//TODO: needing this implies intr_type_contextual is *not* a intr_type_compound!
	extant_level get_extant_impl( util::validation_context & ) const;
	
	std::string get_qualified_id_part() const override {
		return std::string( "$M" ) + super::get_qualified_id_end();
	}
};

/**
	Types that have very limited handling and limited introspection. They serve very specific 
	roles rather than general ones.
	
	TODO: find a better name
*/
struct intr_type_compat final : public intr_type {
	typedef intr_type super;
	static bool is_fun_class( fun_class_t val ) {
		return val == f_compat;
	}
	
	intr_type_compat()
		: super( f_compat ) {
	}
	INTR_TYPE_CLONE(intr_type_compat)
	
	enum type_t : int8_t {
		abi_var_args,
		string_tag,
	} type;
	
	bool is_fixated_impl(util::validation_context &) const {
		return true;
	}
	
	int compare_impl( intr_type const & o, bool logical,util::validation_context & vctx ) const override;
	void write( serial_out & so ) const override;
	void read( serial_in & si ) override;
	std::string get_qualified_id_part() const override {
		return std::string("$Q") + (char)('0'+ (int)type);
	}
};

#undef INTR_TYPE_CLONE
} //eon leaf

#endif