#ifndef SCOPE_HPP
#define SCOPE_HPP

#include "type.hpp"
#include "fwd.hpp"
#include "cerr.hpp"

#include <map>
#include <vector>
#include <set>

namespace leaf
{

struct expression;

struct scope_symbol {
	enum what_t {
		w_single_var,
		w_multi_var,
		w_fun_type,
		w_type_alias,
		w_capture_var,
	} what;

	bool is_single_var() const {
		return what == w_single_var;
	}
	bool is_multi_var() const {
		return what == w_multi_var;
	}
	bool is_fun_type() const {
		return what == w_fun_type;
	}
	bool is_type_alias() const {
		return what == w_type_alias;
	}
	bool is_capture_var() const {
		return what == w_capture_var;
	}
	bool is_var() const {
		return is_single_var() || is_multi_var() || is_capture_var();
	}
	
	//only the appropriate one will be set
	shared_ptr<intr_type const> fun_type;
	shared_ptr<type_spec const> type_alias;
	var_vector_t vars; //single, multi, capture
};

/**
	Encapsulates a logical scope in the code. This could be a function block, a
	structure, or an anonymous block.
*/
class scope : public std::enable_shared_from_this<scope> {
	scope( scope_style_t style );
	scope() = delete;
	scope( scope const & o ) = delete;
public:
	static shared_ptr<scope> create( scope_style_t style );
	static shared_ptr<scope> create_pseudo( scope_style_t style );
	
	typedef boost::error_info<struct tag_error_symbol,std::string> error_symbol;
	
	enum base_type_t {
		bt_none = 0,
		//is a proper child of base (like a method in a class, or a closure)
		bt_parent, 
		//a cloned child of base (a cloned closure) 
		bt_clone,
		//a non-method/context-free function
		bt_free,
		//is a local extension of base (standard sub-scopes inside functions)
		bt_local,
		//is a namespace of base
		bt_namespace,
	} base_type;
	void set_base( shared_ptr<scope> base, base_type_t type );
	
	bool is_base_reachable( shared_ptr<scope const> base ) const;
	
	/**
		Assign a name to this scope (meant for namespaces)
	*/
	void set_name( std::string name );
	
	/**
		Detaches this scope from it's base.
	*/
	void detach_base();
	
	bool is_imperative() const { 
		return scope_style == scope_style_t::imperative; 
	}
	bool is_declarative() const { 
		return scope_style == scope_style_t::declarative; 
	}
	bool is_dynamic() const { 
		return scope_style == scope_style_t::dynamic; 
	}
	scope_style_t get_scope_style() const {
		return scope_style;
	}

	bool no_init_release; //TODO: get rid of this, it's a patch to get tests running
	bool has_init_block() const {
		return !!init_block;
	}
	shared_ptr<statement_block> get_init_block() {
		PRE_CHECK( has_init_block() );
		return init_block;
	}
	
	void set_init_block( shared_ptr<statement_block> init_block ) {
		this->init_block = init_block;
	}
	
	/**
		Indicates this is a class scope and the context is explicitly represented by
		the given type.
	*/
	void mark_class_root( type_ref context );
	bool is_class_root() const {
		return class_root;
	}
	type_ref get_class_scope_context() {
		if( class_root ) {
			PRE_CHECK( context.is_defined() );
			return context;
		}
		PRE_CHECK( base );
		return base->get_class_scope_context();
	}
	
	shared_ptr<scope> get_base() {
		PRE_CHECK( base_type != bt_none );
		return base;
	}
	bool has_base() const {
		return base_type != bt_none;
	}
	
	shared_ptr<scope> find_context( shared_ptr<intr_type const> context_type, int & depth );
	shared_ptr<scope> find_class_root( int & depth );
	shared_ptr<scope> resolve_depth( int depth );

	/**
		Return a string that briefly describes the structure of the sceop from here.
		Meant only for debugging.
	*/
	std::string desc_structure() const;
	
	/**
		The depth away from the root scope.
	*/
	unsigned get_level() const;
	
	/**
		Check if a decl is local within this scope.
	*/
	bool is_local( declaration const & vd ) const;
	bool is_strict_local( declaration const & vd ) const;
	
	enum add_mode {
		add_in_order = 0x1,
	};
	/**
		Add/bind a declaration to this scope.
	*/
	shared_ptr<declaration> add( shared_ptr<declaration> vs, unsigned flags = 0);

	/**
		Brings a variable from a parent scope into the lookup of the current
		scope. This creates a scope-local binding for the variable: all references to this
		variable in this scope have this local binding. The captures table then contains
		a map between the local and source bindings.
		
		@param shared the capture should be a shared capture
		@return returns the new reference to the captured variable (it
			may be the same, for example if already local or global)
	*/
	shared_ptr<declaration> capture_var( shared_ptr<declaration> vs, 
		bool shared = false );
	
	bool can_capture_var( shared_ptr<declaration> vs ) const;
	
	/**
		Imports a variable captured in another scope to this one... NOTE: see
		statement_block::import and note a lot of assumptions/limitations.
	*/
	shared_ptr<declaration> import( shared_ptr<declaration> sr );
	
	/**
		Obtain all variables bound to this name in the immediate scope only (no parent/extend)
	*/
	var_vector_t get_local_vars( std::string const & name ) const;

	boost::optional<scope_symbol> if_get_symbol( std::string const & name ) const;
	boost::optional<scope_symbol> if_get_local_symbol( std::string const & name ) const;
	
	typedef bool (*decl_match_func)( const declaration & );
	/**
		Obtain all local variables matching the given criteria */
	var_vector_t get_local_vars( decl_match_func criteria ) const;
	
	bool has_local( declaration const * sr ) const;

	/**
		Get all locally declared variables: those which first exist in this scope and are not arguments,
		returns, captures, or otherwise.
		
		@param local_life_only only those with an actual local lifetime, otherwise those marked
			global will still be included
	*/
	var_vector_t get_local_decls( bool local_life_only = false ) const;
	
	var_vector_t const & get_in_order() const { 
		return in_order; 
	}
	
	var_vector_t get_captures() const;
	
	var_vector_t get_context_vars() const {
		return context_vars;
	}
	
	/**
		Get or create the capture refers type. If none has been previously forced this
		returns undefined.
	*/
	type_ref get_context_type( bool force ) const;
	
	/**
		Introduces a defined type.
	*/
	void add_fun_type( shared_ptr<intr_type const> type );
	void add_fun_type( std::string const & name, shared_ptr<intr_type const> type );
	
	enum get_flags_t : unsigned {
		//only return local items, no parent/extends/continues
		gf_strict_local = 1<<0,
	};	
	
	/**
		Introduce a new type alias.
	*/
	void add_type_alias( std::string const & name, shared_ptr<type_spec const> spec );
	
	/**
		Allows dot notation in the name, used for type resolution. Non-type resolution should not use
		this function, the typer handles the dots.
	*/
	cerr_impl<scope_symbol> resolve_symbol( std::string const & name ) const;
	
	/**
		Mark this scope as the root of the scope for a function (it still has a parent,
		but this is where args/returns for the function are defined).
	*/
	void mark_function_root(shared_ptr<expr_funcdefn> expr);
	
	/**
		Obtain the most recent function root for this scope.
		@throw error if not root found
	*/
	scope const & get_function_root() const;
	scope const * if_get_function_root() const;
	shared_ptr<expr_funcdefn> get_function_root_expr() const;
	
	bool is_function_root() const { 
		return !!function_root; 
	}
	
	//intrusives usable by compiler
	mutable intptr_t fref;
	
	std::string qualify_name( std::string const & local, std::string const & join ) const;

	//the minimal serialization does not include contents, only scope structure
	void read_minimal( serial_in & in );
	void write_minimal( serial_out & out ) const;
	
	void validate(util::validation_context & vctx) const;
	
private:
	std::map<std::string, scope_symbol> symbols;
	
	//list of variables which belong in the local shared context
	var_vector_t context_vars;
	mutable type_ref context_type;

	bool requires_shared_parent() const;
	
	//all statements added with the order flag (used for parameter compilation)
	var_vector_t in_order;

	shared_ptr<scope> base;
	std::string scope_name;
	
	scope_style_t scope_style;
	
	shared_ptr<expr_funcdefn> function_root;
	bool class_root;
	type_ref context;
	
	//where instance initializations go in this declarative block
	shared_ptr<statement_block> init_block;
	
	//this is not a real scope and should not alter it's parent (used for type inferencing)
	bool pseudo;
	
	scope & get_local_base();
	void capture_var_add_to_context(shared_ptr<declaration> sr);
	shared_ptr<declaration> capture_var_impl(shared_ptr<declaration> sr, bool shared);
};

} //eon leaf
#endif