#ifndef LANG_SERIAL_HPP
#define LANG_SERIAL_HPP

#include "../core_types.hpp"
#include "type.hpp"
#include "expression.hpp"
#include "type_spec.hpp"

#include <vector>
#include <set>
#include <map>
#include <type_traits>
#include <boost/optional.hpp>

namespace leaf {

/**
	Items in the stream are represented by serial_item's. This is mainly for loading.
*/
struct serial_item {
	enum type_t : int8_t {
		t_invalid = 0,
		t_end_marker,
		t_extr_type,
		t_intr_type,
		t_type_ref,
		t_declaration,
		t_expression,
		t_scope,
		t_type_spec,
		t_statement,
		//not a real type, just a marker
		t_key_marker,
	} type;
	
	enum key_ref_t : int8_t {
		t_none = 0,
		t_root,
	} key_ref;
	
	serial_item() : 
		type( t_invalid ),
		key_ref( t_none ) { 
	}

	boost::optional<type_ref> v_type_ref;
	boost::optional<extr_type> v_extr_type;
	boost::optional<type_spec> v_type_spec;
	shared_ptr<intr_type> v_intr_type;
	shared_ptr<declaration> v_declaration;
	shared_ptr<expression> v_expression;
	shared_ptr<scope> v_scope;
	shared_ptr<statement> v_statement;
	
	static type_t type_of( extr_type const & ) {
		return t_extr_type;
	}
	static type_t type_of( intr_type const & ) {
		return t_intr_type;
	}
	static type_t type_of( declaration const & ) {
		return t_declaration;
	}
	static type_t type_of( expression const & ) {
		return t_expression;
	}
	static type_t type_of( scope const & ) {
		return t_scope;
	}
};

struct serial_resolver {
	/**
		Obtain a reference name for this type to be used at load time. If nothing is returned 
		then the type is stored. This allows some types to not be serialized since they are the
		same in all contexts.
	*/
	virtual boost::optional<std::string> get_ref_name( intr_type const & t ) = 0;
	
	/**
		Resolve a name returned by get_ref_name into a type.
	*/
	virtual std::shared_ptr<intr_type const> get_intr_type( std::string const & ref_name ) = 0;
};

/**
	A binary serialization processor for leaf code.
*/
class serial_out {
	std::ostream & out;
	serial_resolver & resolver;
	bool is_exporting_;
public:
	serial_out( std::ostream & out, serial_resolver & resolver );

	//mark this output as performing an export (some transformations are done)
	void set_is_exporting() {
		is_exporting_ = true;
	}
	bool is_exporting() const {
		return is_exporting_;
	}
	
	std::ostream & get_stream() {
		return out;
	}
	
	///////////////////////////////////////////////////////
	// The high-level used to add objects to the output
	void add_type_ref( type_ref t );
	void add_declaration( shared_ptr<declaration const> vd ) {
		queue_declaration.add(vd);
	}
	void add_type_spec( type_spec const & ts );
	void add_intr_type( shared_ptr<intr_type> type ) {
		queue_intr_type.add(type);
	}
	
	/**
		Should be called after all items have been written. This closes off the data stream
		and complete any internal processing.
	*/
	void done();

	///////////////////////////////////////////////////////
	// The low level interface used by the serializers
	template<typename T>
	void write( T const & t ) {
		static_assert( std::is_trivial<T>::value, "" );
		out.write( (char const*)&t, sizeof(t) );
	}
	template<typename T>
	void write( boost::optional<T> const & t ) {
		bool has = !!t;
		write(has);
		if( has )
			write(*t);
	}
	template<typename T>
	void write( std::vector<T> const & v ) {
		size_t len = v.size();
		write(len);
		for( auto & t : v ) {
			write( t );
		}
	}
	template<typename T, typename W>
	void write( std::vector<T> const & v, W writer ) {
		size_t len = v.size();
		write(len);
		for( auto & t : v ) {
			writer(t);
		}
	}
	void write( std::string const & t );
	
	/**
		Writes a reference to the intr_type object. The object will also be queued for serialization
		if not already done.
	*/
	void write_ref( shared_ptr<intr_type const> t );
	void write_ref( shared_ptr<declaration const> vd );
	void write_ref( shared_ptr<expression const> expr );
	void write_ref( shared_ptr<scope const> sc );
	void write_ref( shared_ptr<statement const> st );
	
	template<typename T> // : shared_ptr
	void write_key_ref( serial_item::key_ref_t key, T const & item ) {
		write( serial_item::t_key_marker );
		write( key );
		write( serial_item::type_of( *item ) ); 
		write_ref( item );
	}
private:
	void write_pending( shared_ptr<intr_type const> t );
	void write_pending( shared_ptr<declaration const> t );
	void write_pending( shared_ptr<expression const> t );
	void write_pending( shared_ptr<scope const> t );
	void write_pending( shared_ptr<statement const> t );
	
	template<typename T>
	void process_pending( T & q );
	
	template<typename T>
	struct queue {
		typedef std::vector<shared_ptr<T const>> pending_t;
		pending_t pending;
		typedef std::set<intptr_t> known_t;
		known_t known;
		
		intptr_t id_of( shared_ptr<T const> t ) {
			return reinterpret_cast<intptr_t>(t.get());
		}
		void add( shared_ptr<T const> t ) {
			auto id = id_of(t);
			if( known.find(id) != known.end() ) {
				return;
			}
			pending.push_back(t);
			known.insert(id);
		}
	};
	
	queue<intr_type> queue_intr_type;
	queue<declaration> queue_declaration;
	queue<expression> queue_expression;
	queue<scope> queue_scope;
	queue<statement> queue_statement;
};

/**
	Reads the serialization written by serial_out.
*/
class serial_in {
	std::istream & in;
	serial_resolver & resolver;
public:

	/**
		After construction the stream is no longer used (all items will have been loaded into
		memory).
	*/
	serial_in( std::istream & in, serial_resolver & resolver );

	std::istream & get_stream() {
		return in;
	}
	
	std::vector<serial_item> const & get_all_items() {
		return loaded_items;
	}
	std::vector<serial_item> const & get_key_items() {
		return key_items;
	}
	
	/**
		Reads the primary assuming it is a intr_type.
	*/
	type_ref read_primary_type_ref();
	type_spec read_primary_type_spec();
	
	template<typename T>
	void read( T & t ) {
		static_assert( std::is_trivial<T>::value, "" );
		in.read( (char*)&t, sizeof(t) );
	}
	template<typename T>
	void read( boost::optional<T> & t ) {
		bool has = !!t;
		read(has);
		if( !has ) {
			t = boost::optional<T>();
			return;
		}
			
		T tmp;
		read(tmp);
		t = tmp;
	}
	template<typename T>
	void read( std::vector<T> & v ) {
		size_t len;
		read(len);
		v.resize(len);
		for( size_t i=0; i < len; ++i ) {
			read( v[i] );
		}
	}
	template<typename T, typename R>
	void read( std::vector<T> & v, R reader ) {
		size_t len;
		read(len);
		v.resize(len);
		for( size_t i=0; i < len; ++i ) {
			reader( v[i] );
		}
	}
	void read( std::string & t );
	
	shared_ptr<intr_type> read_intr_type_ref();
	shared_ptr<declaration> read_declaration_ref();
	shared_ptr<expression> read_expression_ref();
	shared_ptr<scope> read_scope_ref();
	shared_ptr<statement> read_statement_ref();

	shared_ptr<scope> get_key_root_scope();
	
	template<typename T>
	shared_ptr<T> read_intr_type_ref_known() {
		auto x = read_intr_type_ref();
		STATE_CHECK( !x || x->is<T>() )
		return std::static_pointer_cast<T>(x);
	}
	
	/**
		Defines a callback that will be used after all data has been loaded. This is used
		for loading things that can't be completely until after the reference has been
		resolved (like a scope).
	*/
	struct post_loader {
		virtual void operator()() = 0;
	};
	
	void add_post_loader( shared_ptr<post_loader> pl ) {
		post_loaders.push_back( pl );
	}
private:
	template<typename T>
	struct queue {
		typedef std::map<intptr_t,shared_ptr<T>> map_t;
		map_t known, pending;
		
		void add_pending( intptr_t id, shared_ptr<T> t ) {
			STATE_CHECK( !!t );
			pending[id] = t;
			known[id] = t;
		}
		
		shared_ptr<T> find( intptr_t id ) {
			auto k = known.find(id);
			if( k != known.end() ) {
				return k->second;
			}
			
			return shared_ptr<T>();
		}
		
		shared_ptr<T> pop_pending( intptr_t id) {
			auto p = pending.find(id);
			if( p == pending.end() ) {
				STATE_CHECK( known.find(id) == known.end() );
				return shared_ptr<T>();
			}
			
			auto ret = p->second;
			pending.erase( p );
			return ret;
		}
	};
	
	queue<intr_type> queue_intr_type;
	queue<declaration> queue_declaration;
	queue<expression> queue_expression;
	queue<scope> queue_scope;
	queue<statement> queue_statement;
		
	std::vector<serial_item> loaded_items, key_items;
	void load_all_items();
	
	shared_ptr<intr_type> read_intr_type();
	shared_ptr<intr_type> create_intr_type(intr_type::fun_class_t cl );
	shared_ptr<declaration> read_declaration();
	shared_ptr<expression> read_expression();
	shared_ptr<expression> create_expression(expression::form_t form );
	shared_ptr<scope> read_scope();
	shared_ptr<statement> read_statement();
	shared_ptr<statement> create_statement(statement_form_t form);
	
	void complete();
	
	std::vector<shared_ptr<post_loader>> post_loaders;
};

} //eon leaf
#endif