#ifndef STATEMENT_HPP
#define STATEMENT_HPP

#include "../core_types.hpp"
#include "boost/optional.hpp"
#include "util/source_location.hpp"
#include "fwd.hpp"
#include "util/validation_context.hpp"

#define STATEMENT_PRE_CLONE(t) \
	shared_ptr<t> pre_clone_derived() const { \
		auto c = make_shared<t>(); \
		c->pre_copy( *this ); \
		return c; \
	} \
	shared_ptr<statement> pre_clone() const override { \
		return pre_clone_derived(); \
	}

#define STATEMENT_SERIAL() \
	void write( serial_out & so ) const override { \
		write_impl(so); \
	} \
	void read( serial_in & si ) override { \
		read_impl(si); \
	}
	
namespace leaf {

struct statement_block;

#define STATEMENT_FORMS \
	STATEMENT_FORM(assign) \
	STATEMENT_FORM(defer) \
	STATEMENT_FORM(do) \
	STATEMENT_FORM(error) \
	STATEMENT_FORM(expression) \
	STATEMENT_FORM(fail) \
	STATEMENT_FORM(for) \
	STATEMENT_FORM(init) \
	STATEMENT_FORM(import) \
	STATEMENT_FORM(loop_flow) \
	STATEMENT_FORM(noop) \
	STATEMENT_FORM(return) \
	STATEMENT_FORM(typedef) \
	STATEMENT_FORM(var)
	
/**
	NOt a part of statement itself to avoid conflicts/confustion in naming. Things
	like expression have their own form, but derive from statement, thus an
	inner-name would conflict.
*/
enum statement_form_t {
	#define STATEMENT_FORM(form) sf_##form##_statement,
	STATEMENT_FORMS
	#undef STATEMENT_FORM
};

enum flow_type_t {
	flow_sequence = 1<<1, //the non-terminator type
	flow_return = 1<<2,
	flow_fail = 1<<3,
	flow_loop_break = 1<<4,
	flow_loop_next = 1<<5,
	flow_resume = 1<<6,
	flow_skippable = 1<<7,
};

char const * name_of( statement_form_t f );

/**
	A base class for all types of statements.
*/
struct statement : public util::source_provider
{
	statement();
	virtual ~statement() { }
	
	virtual statement_form_t statement_form() const = 0;
	
	virtual bool is_declarative() const { 
		return false; 
	}
	
	/**
		A terminator ends a scope block. No code can exist past a terminator.
	*/
	bool is_terminator() const { 
		return (get_flow() & ~flow_sequence); 
	}
	bool is_skippable() const {
		return (get_flow() & flow_skippable);
	}
	virtual int get_flow() const { 
		return flow_sequence; 
	}
	
	/**
		Checks the aspects of the expression which must be valid post-typing. This
		includes verifying all the types are correct/consistent.
	*/
	virtual void validate(util::validation_context &) const { }
	
	/**
		Used by the typer to track statements which are not yet complete.
	*/
	bool local_complete;
	
	virtual extant_level get_extant() const {
		return local_complete ? extant_level::complete : extant_level::incomplete;
	}
	bool is_resolved() const {
		return get_extant() >= extant_level::resolved;
	}
	
	/**
		Creates a deep-clone of the object. All mutable objects will have clones
		created. Only members which are set prior to typing are cloned. Post-typing
		members are left in their default state.
	*/
	virtual shared_ptr<statement> pre_clone() const {
		STATE_FAIL( "unsupported-pre-clone", typeid(*this) );
	}
	
	virtual void write( serial_out & /*so*/ ) const {
		STATE_FAIL( "Unsupported-write", typeid(*this) );
	}
	virtual void read( serial_in & /*si*/ ) { 
		STATE_FAIL( "unsupported-read", typeid(*this) );
	}
protected:
	
	/**
		Copies to this object from the provided one. Meant to be used as part of
		a derived clone function.
	*/
	void pre_copy( statement const & o );
	
	void write_impl( serial_out & so ) const;
	void read_impl( serial_in & so );
};

} //eon leaf

#endif