#ifndef LEAF_TYPE_CONVERTER_HPP
#define LEAF_TYPE_CONVERTER_HPP

#include "fwd.hpp"
#include "platform/platform.hpp"
#include "lang/cerr.hpp"
#include "lang/type.hpp"

#include <boost/optional.hpp>

namespace leaf
{

//these conversions are listed in the order they are allowed to happen. The costing algorithm
//will make use of this, and other places are allowed to assume this
#define CVT_EXPRESSION_FORMS \
		/*placeholder marker to indicate a parametric (not an actual conversion)*/ \
		CVT_EXPRESSION_FORM(parametric) \
		/*nothing to convert at this level*/ \
		CVT_EXPRESSION_FORM(direct) \
		CVT_EXPRESSION_FORM(field) \
		/*similar to c_field[0] but implies the input size as 1, also only for value lists*/ \
		CVT_EXPRESSION_FORM(collapse) \
		/*get the address of a variable*/ \
		CVT_EXPRESSION_FORM(addr_of) \
		CVT_EXPRESSION_FORM(drop_lvalue) \
		/*deref an optional value (may trigger an error)*/ \
		CVT_EXPRESSION_FORM(deref_optional) \
		/*applies to shared or value_ptr*/ \
		CVT_EXPRESSION_FORM(deref) \
		CVT_EXPRESSION_FORM(bind_raw_func) \
		CVT_EXPRESSION_FORM(bind_func) /*TODO: must be renamed, it creates and binds*/ \
		CVT_EXPRESSION_FORM(bind_context) \
		/*convert a rational to a sufficiently sized integer*/ \
		CVT_EXPRESSION_FORM(fixate) \
		CVT_EXPRESSION_FORM(promote) \
		CVT_EXPRESSION_FORM(loss) \
		/*wrap the sub-fields into a tuple*/ \
		CVT_EXPRESSION_FORM(pack) \
		/*wrap the sub-fields into a value-list*/ \
		CVT_EXPRESSION_FORM(pack_value) \
		/*wrap the sub-fileds into a contextual*/ \
		CVT_EXPRESSION_FORM(pack_contextual) \
		/*create an optional value (set)*/ \
		CVT_EXPRESSION_FORM(optional) \
		/*creates a new unset optional value of target type*/ \
		CVT_EXPRESSION_FORM(unset_optional) \
		/*switches type to the defined "target_type" (current type is exact non-distinct match)*/ \
		CVT_EXPRESSION_FORM(set_distinct) \
		/*finally the literal can be removed if needed*/ \
		CVT_EXPRESSION_FORM(drop_literal) \
		/*convert to a compatibility type "target_type"*/ \
		CVT_EXPRESSION_FORM(make_compat) \
		/*instantiate the parametric to "target_type"*/ \
		CVT_EXPRESSION_FORM(param_inst)
		
struct cvt_expression
{
	enum form_t
	{
		#define CVT_EXPRESSION_FORM(n) c_##n,
		CVT_EXPRESSION_FORMS
		#undef CVT_EXPRESSION_FORM
	} form;

	cvt_expression( form_t form )
		: form( form ) {

		pack = intr_type_tuple::p_normal;
	}

	//if form == c_field
	size_t field;
	
	//if form == c_pack/c_pack_value
	struct sub_part {
		shared_ptr<cvt_expression> cvt;
		std::string name;
	};
	std::vector<sub_part> sub;
	
	//if form == c_pack
	intr_type_tuple::pack_t pack;
	//if form == c_promote, c_fixate (no from), c_loss
	enum fundamental_t {
		//in cost order
		p_integer,
		p_binary,
		p_float,
	} from_fun, to_fun;
	int to_bitsize;

	//NOTE: will only be set for c_unset_optional/c_set_distinct (TODO: perhaps it should be set for all?)
	//TODO: hmm, should this not just be a fundamental...?
	type_ref target_type;
	
	inline bool no_cross_promote() const {
		return from_fun == to_fun;
	}
	
	//all forms
	shared_ptr<cvt_expression> then;
	
	bool chain( shared_ptr<cvt_expression> then ) {
		if( !then || then->is_none() )
			return false;
		this->then = then;
		return true;
	}
	shared_ptr<cvt_expression> extend( shared_ptr<cvt_expression> next ) {
		then = next;
		return next;
	}
	
	//true if there is no conversion operation to perform
	bool is_none() const {
		return form == c_direct && sub.empty();
	}
};

/**
	Tracks conversions between types and calculates the costs of conversions.
*/
class type_converter
{
	context const & ctx;
public:
	type_converter( context const & ctx );
	~type_converter();
	
	typedef cerr_impl<type_ref> resolve_t;
	
	struct convert_t : public cerr_base
	{
		//the tree of conversions to perform, null if none/error
		shared_ptr<cvt_expression> conversions;
		
		int compare_cost( convert_t const & o ) const;
		
		convert_t() 
		{ }
		//providing a conversion assumes it is okay
		convert_t( shared_ptr<cvt_expression> cvt )
			: conversions( cvt )
		{ }
		//assumes there is no conversion
		convert_t( cerr result )
			: cerr_base( result )
		{ }
		//explicit to prevent accidentla cross-type conversion
		explicit convert_t( cerr_base base )
			: cerr_base( base )
		{ }
		
	};
	
	
	//Refers to any of the binary conversion functions
	typedef convert_t (type_converter::*binary_convert)( type_ref const & /*to*/, 
		type_ref const & /*from*/ );
	
	/**
		Calculates the cost of transforming one type to another.
	*/
	convert_t convert( type_ref const & to, type_ref const & from )
	{ return convert_impl( to, from, cf_allow_cvt ); }
	
	//to must be an lvalue, and then convers to non-lvalue type
	convert_t convert_to_lvalue( type_ref const & to, type_ref const & from ) { 
		return convert_impl( to, from, cf_allow_cvt | cf_to_lvalue ); 
	}
	//to must be a reference, and then convertes to non-shared type
	convert_t convert_to_refer( type_ref const & to, type_ref const & from ) { 
		return convert_impl( to, from, cf_allow_cvt | cf_to_refer ); 
	}
	//retains a literal on from, even if not specifed in to
	convert_t convert_retain_literal( type_ref const & to, type_ref const & from ) {
		return convert_impl( to, from, cf_allow_cvt | cf_retain_literal );
	}

	//Refers to any of the unary conversion functions
	typedef convert_t (type_converter::*unary_convert)( type_ref const & );
	
	/**
		Produces a "callable" type: a function or a selectable tuple. In the latter case it may
		not actually contain any functions, but the type itself is "callable".
	*/
	convert_t convert_any_callable( type_ref const & from );
	convert_t convert_function_call( intr_type_value_list const & to_args,
		intr_type_value_list const & from_args, bool allow_parametric = false );

	convert_t convert_any_refer( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_shared,
			acf_retain_refer | acf_require_refer | acf_retain_optional );
	}
	convert_t convert_any_lvalue( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_lvalue,
			acf_retain_lvalue | acf_require_lvalue | acf_retain_refer | acf_retain_optional );
	}
	convert_t convert_nonlvalue_optional( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_optional ,
			acf_require_optional | acf_retain_shared );
	}
	convert_t convert_any_nonlvalue( type_ref const & from ) {
		return convert_any_common( from, cerr::mismatch_lvalue,
			acf_retain_shared | acf_retain_optional );
	}
	convert_t convert_any_scalar( type_ref const & from ) {
		return convert_any_common( from,cerr::requires_scalar, acf_require_scalar);
	}
	convert_t convert_unqualified( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_unqualified, 0 );
	}
	convert_t convert_retain_lvalue( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_unqualified, acf_retain_lvalue );
	}
	convert_t convert_shared_or_unqualified( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_unqualified, acf_retain_shared );
	}
	convert_t remove_optional( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_unqualified,
			acf_retain_all | acf_deref_optional );
	}
	convert_t convert_require_value( type_ref const & from ) {
		return convert_any_common( from, cerr::requires_value,
			acf_retain_all | acf_require_value );
	}
		
	/**
		Fixates all fixables types to their real forms (like rationals to integers/floats).
	*/
	resolve_t fixate( type_ref const & from );
	
	/**
		Calculate the best type to which both given types can be converted. The best
		type is the one with the lowest cost of conversion.
	*/
	resolve_t unify( type_ref const & a, type_ref const & b );
	resolve_t unify( type_ref const * begin, size_t len );
	
	/**
		Does matching of arguments against a selectable set of signatures.
	*/
	std::vector<size_t> match_function( intr_type_tuple const & select, intr_type const & args );
	
	
private:
	enum ci_flags {
		cf_allow_cvt = 1<<0,
		cf_to_lvalue = 1<<1,
		cf_to_refer = 1<<2,
		//a literal on from will be retained even if not specified on to
		cf_retain_literal = 1<<3,
	};
	convert_t convert_impl( type_ref const & to, type_ref const & from, unsigned flags );
	convert_t convert_impl( extr_type const & to, extr_type const & from, unsigned flags );
		
	convert_t convert_list( intr_type_compound const& to,
		intr_type_compound const & from, unsigned sub_flags, cvt_expression::form_t form );
		
	enum any_common_flags {
		acf_require_lvalue = 1<<0,
		acf_retain_lvalue = 1<<1,
		acf_require_shared = 1<<2,
		acf_retain_shared = 1<<3,
		acf_require_optional = 1<<4,
		acf_deref_optional = 1<<5,
		acf_retain_optional = 1<<6,
		acf_retain_value_ptr = 1 << 7,
		acf_require_refer = 1 << 8,
		acf_retain_contextual = 1 << 9,
		acf_require_scalar = 1 << 10,
		acf_require_value = 1 << 11,
		
		acf_retain_refer = acf_retain_shared | acf_retain_value_ptr,
		acf_retain_all = acf_retain_lvalue | acf_retain_shared | acf_retain_refer | acf_retain_contextual,
	};
	convert_t convert_any_common( type_ref const & from, cerr list_fail, unsigned flags );
	convert_t convert_any_common( extr_type const & from, cerr list_fail, unsigned flags );
		
	type_ref unify_impl( extr_type const & a, extr_type const & b );
	enum unify_flags {
		uf_allow_collapse = 1 << 0,
	};
	type_ref unify_value_list( extr_type const & a, extr_type const & b, unsigned flags = 0 );
		
	type_ref fixate_impl( extr_type const & from );
	
	convert_t select_convert( extr_type const & to, extr_type const & from, unsigned flags );
};

} //eon leaf

#endif