#ifndef LANG_TYPE_IDENTIFIER_HPP
#define LANG_TYPE_IDENTIFIER_HPP

#include "fwd.hpp"
#include "cerr.hpp"
#include "type.hpp"

#include <map>

namespace leaf {

namespace type_identifier {

	typedef cerr_impl<type_ref> resolve_t;
	typedef cerr_impl<shared_ptr<type_spec>> resolve_spec_t;
	typedef cerr_impl<shared_ptr<intr_type_value_list>> resolve_value_list_t;
	typedef cerr_impl<shared_ptr<intr_type const>> resolve_intr_type_t;
	
	/**
		Expands the type_spec to remove any unknown symbols and complete
		the other elements.
		
		@param from the source spec
	*/
	resolve_spec_t expand( context const & ctx, type_spec const & from );
	
	/**
		Convert a concrete type into a specification. If passed to determine the
		same type_ref should result.
	*/
	resolve_spec_t get_spec( context const & ctx, type_ref t );
	resolve_spec_t get_spec( context const & ctx, extr_type t );
	resolve_spec_t get_spec( context const & ctx, intr_type const & t );
	
	enum determine_flags {
		//the type is sub-part of a higher type (does not apply to value-lists)
		df_sub_type = 1<<0,
		//the context of a function can be implied from the scope
		df_imply_context = 1<<1,
		//allows a parametric type to be determined
		df_allow_param = 1<<2,
		//registers value-list names in the scope
		df_register_names = 1<<3,
		
		//a type is being inferred, not just constrained
		df_infer = 1 << 4,
		//a reference type is being inferred
		df_refer = 1 << 5,
		//a type is being instantiated
		df_instantiate = 1 << 6,
		//this is not a constrain or infer, but an expansion to a single type
		df_single = 1 << 7,
	};
	
	/**
		Translates an input source type into a fully expanded and completed type.
		
		The input source type is something provided in the source language, an abbreviated type_spec
		as defined by the syntax of the language. Determination completes the semantics of that
		type and converts to a concrete represenation (where appropriate).
	*/
	resolve_t determine( context const & ctx, type_spec const & vt, unsigned flags = 0 );
	resolve_t determine( context const & ctx, shared_ptr<type_spec const> vt, unsigned flags = 0 );
	
	/**
		Determines what the inferred type of the given type should be. 
		
		There are only a few key locations where types are inferred:
			- in variable assigned, the value being assigned has it's type inferred
			- inferred return values
			- in an expression program (.lfx) the high-level type
			
		@param constraint the type specified (may be null)
		@param source which type is being assigned to the "spec" var
		@param init_refer indicates infer is being done for reference initialization
	*/
	resolve_t infer( context const & ctx, type_spec const & constraint, type_ref const & source, bool init_refer );
	resolve_t infer( context const & ctx, shared_ptr<type_spec const> constraint, 
		type_ref const & source, bool init_refer );
	resolve_spec_t infer_spec( context const & ctx, type_spec const & constraint, 
		type_ref source, bool init_refer );
	resolve_t infer( context const & ctx, type_ref const & source, bool init_refer );
	
	/**
		Constrains the given type to the indicated spec. This is a lesser reduction
		than "infer" since it may retain literals and virtual types like rationals.
		
		@param assign The type which is being constrained. this is assumed to be determined type.
		@param spec The constraint being applied
	*/
	resolve_t constrain( context const & ctx, type_spec const & constraint, type_ref const & source );
	resolve_t constrain( context const & ctx, shared_ptr<type_spec const> constraint, type_ref const & source );
	
	/**
		Infers a value-list (such as for inferred return types).
	*/
	resolve_value_list_t infer_value_list( context const & ctx, type_ref assign, bool init_refer );

	/**
		Applies parameters to a parametric type. This does not convert the types, but
		merely produces the type which should result.
	*/
	resolve_t parameterize_type( context const & ctx, type_ref const & param_type, type_ref const & apply,
		unsigned flags = 0);
	resolve_t parameterize_type( context const & ctx, intr_type const & param_type, intr_type const & apply,
		unsigned flags = 0);
	
} //eon type_identifier


} //eon leaf

#endif