#ifndef LANG_TYPE_HPP
#error include <lang/type.hpp>
#endif

#ifndef LANG_TYPE_REF_HPP
#define LANG_TYPE_REF_HPP

#include <boost/optional.hpp>

namespace leaf {

/**
	Encapsulates a reference to a type. In most places this could be undefined, a type_spec,
	or an extrinsic type.
	
	This is a value-type, and extr_type is also considered a value type. Thus the type_ref
	should be considered equivalent to the extr_type itself.
*/
struct type_ref : public type_node {
	enum type_t : int8_t {
		t_undefined,
		t_extr,
		t_spec,
	};
private:
	type_t type_;
	boost::optional<extr_type> extr_;
	shared_ptr<type_spec const> spec_;
public:	
	type_t type() const {
		return type_;
	} 
	bool is_defined() const {
		return type_ != t_undefined;
	}

	bool is_extr() const {
		return type_ == t_extr;
	}
	extr_type const & extr() const {
		PRE_CHECK( extr_ );
		return *extr_;
	}
	extr_type & extr() {
		PRE_CHECK( extr_ );
		return *extr_;
	}
	
	bool is_spec() const {
		return type_ == t_spec;
	}
	type_spec const & spec() const {
		PRE_CHECK( spec_ );
		return *spec_;
	}

	type_ref( serial_in & si );
	type_ref( extr_type const & o ) :
		type_( t_extr ),
		extr_( o ) {
	}
	type_ref() {
		type_ = t_undefined;
	}
	type_ref( type_ref const & ) = default;

	type_ref & operator=( extr_type const & o ) {
		type_ = t_extr;
		extr_ = o;
		return *this;
	}

	static type_ref create_spec( type_spec const & ts );
	
	//type_node
	bool is_fixated_impl( util::validation_context & vctx ) const;
	extant_level get_extant_impl( util::validation_context & vctx ) const;
	bool contains_parametric_impl( util::validation_context & ctx ) const;
	void validate(util::validation_context & vctx) const;
	//end type_node
	
	int compare( type_ref const& o, bool logical ) const;
	int compare_impl( type_ref const & o, bool logical, util::validation_context & vctx ) const;
	bool operator<( type_ref const & o ) const { 
		return compare( o, false ) < 0;
	}
	bool operator==( type_ref const& o ) const { 	
		return compare( o, false ) == 0;
	}
	bool operator!=( type_ref const & o ) const {
		return compare( o, false ) != 0;
	}
	bool same_as( type_ref const & o ) const {
		return compare( o, true ) == 0;
	}
	
	void write( serial_out & so ) const;
	void read( serial_in & si );
	
	std::string get_qualified_id() const;	
};

} //eon leaf

#endif