#ifndef LANG_TYPE_SPEC_HPP
#define LANG_TYPE_SPEC_HPP

#include "fwd.hpp"
#include "util/spec_optional.hpp"
#include "lang/type.hpp"

#include <boost/variant.hpp>

#include <vector>

namespace leaf {

struct node;

#define TYPE_SPEC_PARTS \
	/* \
		singular specs*/ \
	TYPE_SPEC_PART(signed) \
	TYPE_SPEC_PART(data_bitsize) \
	TYPE_SPEC_PART(precision_variant) \
	TYPE_SPEC_PART(selectable) \
	TYPE_SPEC_PART(member) \
	TYPE_SPEC_PART(fun_class) \
	TYPE_SPEC_PART(function_access) \
	TYPE_SPEC_PART(function_throw) \
	TYPE_SPEC_PART(pack) \
	TYPE_SPEC_PART(fun_prototype) \
	TYPE_SPEC_PART(context) \
	TYPE_SPEC_PART(context_name) \
	TYPE_SPEC_PART(compat_type) \
	/* \
	extrinsic modifiers */ \
	TYPE_SPEC_PART(lossy) \
	TYPE_SPEC_PART(reference) \
	TYPE_SPEC_PART(lvalue) \
	TYPE_SPEC_PART(optional) \
	TYPE_SPEC_PART(cvt_optional) \
	TYPE_SPEC_PART(literal) \
	/* \
	qualifiers/constraints*/ \
	TYPE_SPEC_PART(tuple) \
	/* \
	multiple specs*/ \
	TYPE_SPEC_PART(symbol) \
	TYPE_SPEC_PART(param) /*reference to a parameter*/
	
enum part_type_t {
	#define TYPE_SPEC_PART(s) pt_##s,
	TYPE_SPEC_PARTS
	#undef TYPE_SPEC_PART
};
	
constexpr inline bool is_singular_part_type( part_type_t pt ) {
	return pt != pt_symbol && pt != pt_param;
}

std::string name_of( part_type_t pt );

template<part_type_t PT>
struct part_type_descriptor {
};

struct type_spec_symbol {
	std::vector<std::string> parts; //will have >= 1 part to be valid
	
	std::string const & primary() const {
		STATE_CHECK(parts.size() > 0);
		return parts[0];
	}
	
	static type_spec_symbol create( std::string const & part ) {
		type_spec_symbol tss;
		tss.parts.push_back(part);
		return tss;
	}
	
	bool operator==(type_spec_symbol const &o) const;
};

#define DESC_PART(pt_,type_) \
	template<> struct part_type_descriptor<pt_> { \
		typedef type_ type; \
	};
DESC_PART(pt_signed, bool)
DESC_PART(pt_data_bitsize, int)
DESC_PART(pt_precision_variant, int)
DESC_PART(pt_selectable, bool)
DESC_PART(pt_member, bool)
DESC_PART(pt_fun_class, intr_type::fun_class_t)
DESC_PART(pt_function_access, intr_type_function::access_t)
DESC_PART(pt_function_throw, bool)
DESC_PART(pt_pack, intr_type_tuple::pack_t)
DESC_PART(pt_compat_type, intr_type_compat::type_t)

DESC_PART(pt_fun_prototype, shared_ptr<intr_type const>)
DESC_PART(pt_context, extr_type)
DESC_PART(pt_context_name, std::string) //same as context, but just a name

DESC_PART(pt_lossy, bool)
DESC_PART(pt_reference,extr_type::reference_t)
DESC_PART(pt_lvalue,bool)
DESC_PART(pt_optional,bool)
DESC_PART(pt_cvt_optional,bool)
DESC_PART(pt_literal,bool)

DESC_PART(pt_tuple,bool)

DESC_PART(pt_symbol,type_spec_symbol)
DESC_PART(pt_param,std::string)
#undef DESC_PART

/**
	A provisional intr_type specification which results from parsing. This contains details of what was parsed.
	It is meant to be converted to a intr_type during typing (where the additional information will be
	available to complete the type).
	
	If a spec includes a prototype then it becomes the dominant form. All other settings must be
	compliant otherwise the spec cannot determine a valid type.
*/
struct type_spec
{
	struct part {
		part_type_t type;
		boost::variant<
			bool,
			extr_type,
			extr_type::reference_t,
			int,
			intr_type_compat::type_t,
			intr_type::fun_class_t,
			intr_type_function::access_t,
			intr_type_tuple::pack_t,
			shared_ptr<intr_type const>, //MUST not be an instance!
			std::string,
			type_spec_symbol
		> value;
		//sub-parametrics
		std::vector<shared_ptr<type_spec const>> sub;
		//attached expression
		shared_ptr<node const> node_expr;
		shared_ptr<expression const> expr;
	};

	//meant only for multiple types
	template<part_type_t PT>
	part & push_part( typename part_type_descriptor<PT>::type const & value ) {
		static_assert( !is_singular_part_type( PT ), "" );
		part p;
		p.type = PT;
		p.value = value;
		parts.push_back( p );
		return parts.back();
	}
	
	template<part_type_t PT>
	part & set( typename part_type_descriptor<PT>::type value ) {
		return set_part( PT, value );
	}
	
	template<typename T>
	part & set_part( part_type_t pt, T const & value ) {
		for( auto & p : parts ) {
			if( p.type == pt ) {
				p.value = value;
				return p;
			}
		}
		
		part np;
		np.type = pt;
		np.value = value;
		parts.push_back( np );
		return parts.back();
	}
	
	template<part_type_t PT>
	bool has() const {
		return has_part(PT);
	}
	
	bool has_part(part_type_t pt) const {
		for( auto & p : parts ) {
			if( p.type == pt ) {
				return true;
			}
		}
		return false;
	}

	template<part_type_t PT>
	void clear() {
		clear_part(PT);
	}
	
	void clear_part(part_type_t pt) {
		for( int i=int(parts.size())-1; i >= 0; --i ) {
			if( parts[(unsigned)i].type == pt ) {
				parts.erase( parts.begin() + i );
			}
		}
	}
	
	template<part_type_t PT>
	boost::optional<typename part_type_descriptor<PT>::type> if_get() const {
		for( auto & p : parts ) {
			if( p.type == PT ) {
				return boost::get<typename part_type_descriptor<PT>::type>( p.value );
			}
		}
		return boost::optional<typename part_type_descriptor<PT>::type>();
	}

	template<part_type_t PT>
	typename part_type_descriptor<PT>::type get() const {
		for( auto & p : parts ) {
			if( p.type == PT ) {
				return boost::get<typename part_type_descriptor<PT>::type>( p.value );
			}
		}
		STATE_FAIL();
	}

	template<typename T>
	T get_part_value(part_type_t pt) const {
		for( auto & p : parts ) {
			if( p.type == pt ) {
				return boost::get<T>( p.value );
			}
		}
		STATE_FAIL();
	}

	part const & get_part(part_type_t pt) const {
		for( auto & p : parts ) {
			if( p.type == pt ) {
				return p;
			}
		}
		STATE_FAIL();
	}
	part & get_part(part_type_t pt) {
		for( auto & p : parts ) {
			if( p.type == pt ) {
				return p;
			}
		}
		STATE_FAIL();
	}
	
	template<part_type_t PT>
	typename part_type_descriptor<PT>::type get(
		typename part_type_descriptor<PT>::type const & def ) const {
		for( auto & p : parts ) {
			if( p.type == PT ) {
				return boost::get<typename part_type_descriptor<PT>::type>( p.value );
			}
		}
		return def;
	}
	
	template<part_type_t PT>
	part & default_set( typename part_type_descriptor<PT>::type value ) {	
		if( has<PT>() ) {
			return get_part(PT);
		}
		return set<PT>(value);
	}
	
	struct sub_ref {
		std::string name;
		shared_ptr<type_spec const> type;
	};
	std::vector<sub_ref> sub;	
	void push_sub( std::string name, shared_ptr<type_spec const> type ) {
		sub_ref sr;
		sr.name = name;
		sr.type = type;
		sub.push_back( sr );
	}
	void push_sub( shared_ptr<type_spec const> type ) {
		push_sub( "", type );
	}
	
	//the parametric aspects of this type
	//the < T , Q > part of the spec (pt_param parts refer to these items)
	std::vector<sub_ref> params;
	void push_param( std::string name, shared_ptr<type_spec const> type ) {
		sub_ref sr;
		sr.name = name;
		sr.type = type;
		params.push_back( sr );
	}

	/** Is the spec free of any extrinsic modifiers (none are "set", default is okay) */
	bool extrinsic_free() const;
	
	/** Does this, or any of the `sub` items contain a pt_param value */
	bool has_pt_param() const;
	
	type_spec( intr_type::fun_class_t fun_class ) {
		set<pt_fun_class>( fun_class );
	}
	
	type_spec( type_spec const & ) = default;
	type_spec() {} //apprently not the same as =default
	type_spec( serial_in & si ) {
		read(si);
	}
	
	void write( serial_out & so ) const;
	void read( serial_in & si );

	//I'd like this to be a const, but some things are too hard then
	std::vector<part> & get_parts() {
		return parts;
	}
	std::vector<part> const & get_parts() const {
		return parts;
	}
	
	bool has_known_intrinsic() const {
		return has<pt_fun_class>() ||
			has<pt_fun_prototype>();
	}
	
	intr_type::fun_class_t get_any_fun_class() const {
		auto fc = get<pt_fun_class>( intr_type::f_invalid );
		if( fc == intr_type::f_invalid && has<pt_fun_prototype>() ) {
			fc = get<pt_fun_prototype>()->get_fun_class();
		}
		return fc;
	}
		
	boost::optional<part_type_t> has_alternate_part( part_type_t p ) const;
private:
	std::vector<part> parts;
};

} //eon leaf

#endif