#ifndef LEAF_TYPER_HPP
#define LEAF_TYPER_HPP

#include "fwd.hpp"

#include "context.hpp"

namespace leaf
{

struct typer_serial_resolver;

/**
	The typer assembles the parts of a complete module and applies the typing to the
	statements.
*/
class typer {
	friend struct typer_serial_resolver;
	friend struct context;
	
	shared_ptr<module> mod;
	shared_ptr<module_loader> mloader;
	shared_ptr<type_manager> types;
	
	void init();
public:
	typer( boost::optional<std::string> module_name, 
		shared_ptr<module_loader> mloader );
	//meant for testing
	typer( shared_ptr<scope> root_scope );
	typer( ) = delete;
	typer( typer const & o ) = delete;
	
	shared_ptr<module> get_module() {
		return mod;
	}
	
	enum type_expression_flags
	{
		//for full typing force the result to be converted to its inferred type
		tef_convert_inferred = 1<<0,
	};
	
	/**
		Completes the typing of this modules. This can be run multiple times to allow adding
		of new blocks.
	*/
	void full_complete_type();
	
	typedef std::vector<shared_ptr<statement>> statement_vector_t;
	
	/** Add statements as a root block. */
	shared_ptr<statement_block> add_root_block( statement_vector_t block );

	/** Adds statements as a special dynamic block. */
	shared_ptr<statement_block> add_dynamic_block( statement_vector_t block );
	
	/** Adds statement as a global block. */
	shared_ptr<statement_block> add_global_block( statement_vector_t block );
	
	shared_ptr<expression> full_type( shared_ptr<expression> expr, unsigned flags = 0 );
	
	//bool value unused, just a placeholder for this tag
	typedef boost::error_info<struct tag_typer_error,bool> error_typer;
	//tracks additional information about errors
	typedef boost::error_info<struct cerr_source_type,std::string> tag_cerr_source_type;
	typedef boost::error_info<struct cerr_target_type,std::string> tag_cerr_target_type;
	
	friend class job_statement_block; //TODO: replace friend system with distinct open typer and closed high-level interface
	struct job_statement_block : public job {
		shared_ptr<statement_block> sb;
		bool process( typer * ty );
		job_statement_block( shared_ptr<statement_block> sb )
			: sb( sb) {
		}
	};
	
	serial_resolver & get_serial_resolver();
private:
	typedef std::vector<shared_ptr<job>> job_vector_t;
	job_vector_t pending_jobs;
	void complete_type();
	size_t completed_statements, progress_count;
	//no longer allow incomplete items and report errors instead
	bool no_incomplete;
	
	bool proc_statement_block( shared_ptr<statement_block> block, context ctx );
	bool proc_statement_block( shared_ptr<statement_block> block );
	bool proc_error_block( shared_ptr<statement> st, error_block & block, context ctx );
	
	shared_ptr<statement> proc_statement( shared_ptr<statement> s, context ctx );

	shared_ptr<statement> proc_assign_statement( shared_ptr<assign_statement> as, context ctx );
	shared_ptr<statement> proc_defer_statement( shared_ptr<defer_statement> ds, context ctx );
	shared_ptr<statement> proc_do_statement( shared_ptr<do_statement> s, context ctx );
	shared_ptr<statement> proc_expression_statement( shared_ptr<expression_statement> es, context ctx );
	shared_ptr<statement> proc_fail_statement( shared_ptr<fail_statement> fs, context ctx );
	shared_ptr<statement> proc_for_statement( shared_ptr<for_statement> fs, context ctx );
	shared_ptr<statement> proc_import_statement( shared_ptr<import_statement> is, context ctx );
	shared_ptr<statement> proc_init_statement( shared_ptr<init_statement> is, context ctx );
	shared_ptr<statement> proc_loop_flow_statement( shared_ptr<loop_flow_statement> ls, context ctx );
	shared_ptr<statement> proc_return_statement( shared_ptr<return_statement> rs, context ctx );
	shared_ptr<statement> proc_typedef_statement( shared_ptr<typedef_statement> ts, context ctx );
	shared_ptr<statement> proc_var_statement( shared_ptr<var_statement> vs, context ctx );
	shared_ptr<statement> proc_noop_statement( shared_ptr<noop_statement> ns, context ctx );
	shared_ptr<statement> proc_error_statement( shared_ptr<error_statement> es, context ctx );
	
	void scan_symbols( shared_ptr<statement_block> block );
	
	std::shared_ptr<typer_serial_resolver> tsr;
};

namespace expression_typer {
	//entry point to type expressions
	shared_ptr<expression> process_expression( shared_ptr<expression> expr, 	context ctx );
}


} //eon leaf

#endif