#ifndef LEAF_PARSER_NODE_HPP
#define LEAF_PARSER_NODE_HPP

#include "util/source_location.hpp"
#include "util/number.hpp"
#include "fwd.hpp"

#include <vector>
#include <map>

namespace leaf
{

/**
	A Node is the generic syntax structure of the language. It works basically likely an advanced
	lexer converting the input into a structure tree which is further processed into the abstract
	syntax tree (and checked for correctness).
	
	NOTE: value-list's may not contain value-list's! These are not real types but rather
	syntactic constructs. They do have a intr_type but variables cannot be assigned this type.
	
	unary-operator and binary-operator share the same lexing, but only allow a fixed number
	of operators. The parser should not allow unknown operators to parse, but nonetheless
	parses according to a lenient lexer (greedy series of any operator symbols)
*/
struct node
{
	enum form_t
	{
		f_symbol,
		
		f_literal_number,
		f_literal_boolean,
		f_literal_string,
		f_literal_char,

		f_expr_list, //sub contains sequenced nodes
		f_value_list, //sub contains value-list parts
		f_param_list, //sub contains parametric parts
		f_tuple, //sub contains tuple parts
		f_block, //sub contains parts (likely statements)
		
		f_funccall, //sub contains two parts [0] = function symbol, [1] = args tuple
		f_funcdefn, //function definition [0] = args, [1] = returns, [2] = block
		f_type, //node is exclusively a type
		f_named, //sub[0] = name, sub[1] = expression
		f_on, //sub[0] = statement, sub[1] = error class, sub[2] = error block
		f_implied, //there is an unspecified implied type component here
		f_decl_name,  //sub contains the parametric name parts
	} form;

	std::string text;

	std::vector<shared_ptr<node> > sub;

	//the explicit type (may be null)
	node_type_spec type;
	
	//the parameter type (may be null) -- this will always be a value-list
	node_type_spec param;

	//any annotations at this node (node will always be a value list)
	typedef std::map<std::string,shared_ptr<node>> annotate_map_t;
	annotate_map_t annotations;

	node( form_t form, util::source_location src_loc )
		: form( form ) 
		, src_loc( src_loc ) {
	}

	//the logical location of this node in the source
	util::source_location src_loc;
};

struct node_literal_number : public node
{
	node_literal_number( util::source_location src_loc )
		: node( f_literal_number, src_loc )
	{ }
	
	bool was_float;
	util::number value;
};

} //eon leaf

#endif