#ifndef LEAF_PARSER_SOURCE_HPP
#define LEAF_PARSER_SOURCE_HPP

#include "util/source_location.hpp"
#include "symbols.hpp"

#include <boost/regex.hpp>
#include <memory>

namespace leaf {

/**
	The source code which is to be processed.
	
	This is always attached to a shared pointer so that the source_location aspect works.
*/
class source : public std::enable_shared_from_this<source>, public util::source_base
{
	typedef boost::error_info<struct tag_text_at_location,std::string> text_at_location;
	typedef boost::error_info<struct tag_name,std::string> except_name;

	struct private_for_create { };
public:
	source( private_for_create, char const * code, size_t len )
		: data( code, len )
	{
		begin = at = &data[0];
		end = at + len;
	}

	source( private_for_create, std::string const & code )
		: data( code )
	{
		begin = at = &data[0];
		end = at + data.size();
	}
	
	static shared_ptr<source> create( char const * code, size_t len )
	{ return make_shared<source>( private_for_create(), code, len ); }
	static shared_ptr<source> create( std::string const & code )
	{ return make_shared<source>( private_for_create(), code ); }

	bool has_next( unsigned flags = 0)
	{
		skip_space(flags);
		if( at >= end )
			return false;
		return true;
	}

	bool has_raw_next() { 
		return at < end; 
	}
	
	src_char_t raw_next() {
		return *at++;
	}
	
	struct raw_bit {
		src_symbol_t sym;
		int count;
	};
	
	src_symbol_t peek_next() const { 
		auto a = next_bit();
		return a.sym;
	}

	src_symbol_t next() { 
		auto a = next_bit();
		at += a.count;
		return a.sym;
	}

	src_char_t safe_bit(int off) const {
		return at + off < end ? *(at+off) : (char)0;
	}
	
	raw_bit next_bit() const;
	
	enum scan_flags_t {
		skip_line_ends = 1<<1,
	};
	/**
		Skips all non-token characters.
	*/
	void skip_space( unsigned flags = 0 );

	void skip_all_space() {
		skip_space( skip_line_ends );
	}
	bool has_all_next() {
		return has_next( skip_line_ends );
	}
	
	/**
		Perform a match at the current location against the given regex. If there
		is a match then the results are stored in "what" and the current location
		advanced to the end of the match.
	*/
	bool match( boost::regex const & re )
	{
		boost::match_flag_type flags = boost::match_continuous;
		if( !boost::regex_search( at, end, what, re, flags ) )
			return false;

		at = what[0].second;
		return true;
	}
	bool peek_match( boost::regex const & re )
	{
		boost::match_flag_type flags = boost::match_continuous;
		if( !boost::regex_search( at, end, what, re, flags ) )
			return false;
		return true;
	}

	std::string cur_match( int group = 0 )
	{ return std::string( what[group].first, what[group].second ); }
	
	std::string cur_text() const;
	void throw_expecting( std::string const & msg );
	void throw_invalid( std::string const & msg );
	void throw_unexpected( std::string const & msg );

	void set_name( std::string const & name )
	{ this->name = name; }
	std::string get_name() const
	{ return name; }
	
	/**
		Obtain the cursor position in the source file. These are calculated on demand and
		not tracked, thus don't call this unless really needed.
		
		Note these values are 1-based cardinals as normal for text file references.
	*/
	size_t cur_line() const
	{ return line_at( (size_t)(at - begin) ); }
	size_t cur_column() const
	{ return column_at( (size_t)(at - begin) ); }
	size_t cur_loc() const
	{ return (size_t)(at - begin); }
	
	size_t line_at( size_t loc ) const;
	size_t column_at( size_t loc, size_t tab_size = 4 ) const;
	
	util::source_location cur_location() const {
		util::source_location l;
		l.src = shared_from_this();
		l.loc = cur_loc();
		return l;
	}
private:
	void throw_extend( recoverable_error & ex );
	
	std::string data;
	src_char_t const * begin;
	src_char_t const * at;
	src_char_t const * end;

	src_char_t const * convert_loc( size_t loc ) const;
	
	//name of this source (used in exceptions)
	std::string name;
	
	boost::match_results<src_char_t const*> what;
};

} //eon leaf


#endif