#ifndef LEAF_RUNNER_RUNNER_HPP
#define LEAF_RUNNER_RUNNER_HPP

#include "runner/fwd.hpp"

#include "lang/language.hpp"
#include "lang/typer.hpp"
#include "lang/dump.hpp"
#include "lang/serial.hpp"
#include "lang/module_loader.hpp"

#include "parser/node_converter.hpp"
#include "parser/node_parser.hpp"
#include "parser/dump.hpp"
#include "parser/source.hpp"

#include "ir/compiler_leaf.hpp"
#include "ir/dump.hpp"
#include "ir/module.hpp"
#include "ir/trace_functions.hpp"

#include "ir/llvm/gen.hpp"
#include "ir/llvm/runtime_test.hpp"

#include "util/file.hpp"

/**
	Encapsulates the basic logic of compilation and execution (as required for tracing).
	
	This is not a settings class, but a stateful processor.
*/
struct runner {

	//debugging options
	bool dump_parse, dump_parse_flat, dump_base, dump_pre_type, 
		dump_all_types, dump_ir, dump_ir_types, dump_post_type,
		list_functions;
	std::string dump_llvm;

	//all input files (reduced modes expect only 1 file)
	std::vector<std::string> input_files;
	//the name of the module being generated (if any)
	boost::optional<std::string> module_name;
	
	//dynamic libraries to open before execution
	std::vector<std::string> libraries;
	
	//optimizations
	ir_llvm::gen::options gen_opts;
	
	//for module resolution
	shared_ptr<leaf::module_loader> mloader;
	
	runner() {
		capture_stdout = false;
		
		dump_ir_types = false;
		dump_pre_type = false;
		dump_base = false;
		dump_post_type = false;
		dump_all_types = false;
		dump_ir = false;
		dump_parse = false;
		dump_parse_flat = false;
		list_functions = false;
	}
	
	shared_ptr<leaf::statement_block> global;
	shared_ptr<leaf::typer> module_typer;
	shared_ptr<leaf::module> module;
	
	typedef std::vector<shared_ptr<leaf::statement>> statement_vector_t;
	
	void x_dump_parse( leaf::node const & n );
	void x_dump_post_type( leaf::module const & mod );
	void x_dump_pre_type( statement_vector_t const & sb );
	void do_typing( bool base = false );
	
	shared_ptr<leaf::expression> parse_expr( std::string const & expr );
	void init_module();
	void add_base_block( shared_ptr<leaf::source> source );
		
	statement_vector_t load_parse_block( std::string const & filename, 
		leaf::scope_style_t scope_tyle, bool dump = true );
	statement_vector_t load_parse_block( shared_ptr<leaf::source> src, 
		leaf::scope_style_t scope_tyle, bool dump = true );
	
	shared_ptr<leaf::source> create_string_source( std::string const & name, std::string const & content );

	/**
		Execute the main function of the module.
		
		@return For a normal execution that cleanly completes this is the return value from the
			main entry point. For an expectation test that succeeds 0 is returned, an error is
			thrown on failure. Other errors will be thrown.
	*/
	int execute( shared_ptr<ir::module> module );
	
	enum gen_what_t {
		//static object file
		gw_object,
		//shared library
		gw_shared,
		//main executable
		gw_exe,
	};
	/**
		Generate an object file of the module.
		
		@param path_base the base path/name of the output (extensions will be added)
		@param gw what type of object file to generate
	*/
	void emit_object( shared_ptr<ir::module> module, std::string const & path_base, gen_what_t gw );
	
	/** 
		The expected provides a way to intercept output and error conditions of executed code.
	*/
	shared_ptr<expecter> expect;
	
	/**
		The output of the execution should be captured.
	*/
	bool capture_stdout;
	

	/**
		Build the source files down to the IR module.
	*/
	shared_ptr<ir::module> build_module();
private:
	shared_ptr<ir::module> get_expr();
	shared_ptr<ir::module> get_block();
	shared_ptr<ir::module> get_full();
};

#endif