#ifndef LEAF_UTIL_NUMBER_HPP
#define LEAF_UTIL_NUMBER_HPP

#include <gmpxx.h>
#include <string>
#include <cstdint>
#include <iosfwd>

namespace util
{

/**
	Stores an arbitrary precision number. This may be exact, or approximate if converted to a float.
	
	CLN: Should we ever need more operations than what GMP supports CLN may be an option. It is
	not being used now since for basic decimal handling it doesn't offer anything more.
*/
class number
{
	mpq_class value;
	
	number(mpq_class value);
public:
	number();
	number( number const & o );
	
	/**
		Convert a decimal string into a number representation. The decimal will be converted
		under these rules:
		- exactly to a rational number up to an exponent limit
		- otherwise an exception
	*/
	#define PARSE_BASE(name,base) \
		static number parse_##name( std::string const & str ) \
		{ unsigned t; return parse_base( str, t, base ); } \
		static number parse_##name( std::string const & str, unsigned & t ) \
		{ return parse_base( str, t, base ); }
	PARSE_BASE(decimal,10)
	PARSE_BASE(binary,2)
	PARSE_BASE(octal,8)
	PARSE_BASE(hexadecimal,16)
	
	static number parse_base( std::string const & str, int base )
	{ unsigned t; return parse_base( str, t, base ); }

	enum pd_flags_t {
		//the input number contained a '.' separator
		pdf_had_decimal = 0x1,
		//the input contained an 'e' exponent indicator
		pdf_had_exponent = 0x2,
		//the output is exactly equal to the input
		pdf_is_exact = 0x4,
	};
	static number parse_base( std::string const & str, unsigned & meta, int base );

	#if defined(__i386__) or defined(IS_DARWIN)
	//TODO: libgmp can't handel 64-bit values in this case
	static number integer( long long int val ) {
		return number( (long int)val );
	}
	#endif

	template<typename T>
	static number integer( T val )
	{ return number( mpz_class(val) ); }
	
	std::string to_string(bool force_denom=false) const;
	
	number & operator=(number const & o);
	bool operator==(number const & o ) const;
	
	bool is_negative() const;
	bool is_integral() const;
	
	/*
		Does the value fit in the given integer of this bit-size (2's complement)? These
		make the assumption that the type is already an integer and non-negative
		for unsigned.
	*/
	bool fits_signed( unsigned bits ) const;
	bool fits_unsigned( unsigned bits ) const;
	unsigned signed_bits() const;
	unsigned unsigned_bits() const;
	
	/*
		Conversion rules:
		- out-of-range values will result in undefined results (check first!)
		- precision will be truncated to target type
	*/
	intmax_t as_int() const;
	uintmax_t as_uint() const;
	double as_double() const;
	
	/*
		Basic operations.
	*/
	number neg() const;
	number abs() const;
	number operator+( number const & o ) const;
	number operator-( number const & o ) const;
	number operator*( number const & o ) const;
	number operator/( number const & o ) const;
	
	/*
		Serialization
	*/
	void write( std::ostream & out ) const;
	void read( std::istream & in );
};

} //eon
#endif