#ifndef UTIL_SPEC_OPTIONAL_HPP
#define UTIL_SPEC_OPTIONAL_HPP

#include "../core_types.hpp"

namespace util
{

/**
	This base is just to make it easier to work with lambdas which cannot be templated.
*/
struct spec_optional_base
{
protected:
	enum state_t {
		//all items start in their default state (but "value" is not defined)
		s_default,
		//unset items are those which have been explicitly cleared ("value" is not defined)
		s_unset,
		//set items have a defined "value"
		s_set
	} state;
	
	spec_optional_base( state_t state )
		: state( state )
	{ }
public:
	bool is_default() const {
		return state == s_default;
	}
	bool is_set() const {
		return state == s_set;
	}
	bool is_unset() const {
		return state == s_unset;
	}
};

/**
	Like boost optional but has more than one not-set state. This is used by the type_spec
	to properly track what the user specified and what can be merged for inferrence and
	determiniation.
	
	NOTE: this was only needed for trivial data types and thus may not be as robust as optional
	in it's type handling.
*/
template<typename T, bool StartUnset = true>
struct spec_optional : public spec_optional_base
{
private:
	T value;
	
public:
	spec_optional()
		: spec_optional_base( StartUnset ? s_unset : s_default )
	{ }
	spec_optional( spec_optional const & o ) 
		: spec_optional_base( o.state )
	{
		if( state == s_set )
			value = o.value;
	}
	spec_optional( T const & v )
		: spec_optional_base( s_set )
		, value( v )
	{ }
	
	void revert_default() {
		state = s_default;
	}
	void revert_default_if_unset() {
		if( state == s_unset )
			state = s_default;
	}
	void mark_unset() {
		state = s_unset;
	}
	
	spec_optional & operator=( spec_optional const & o ) {
		state = o.state;
		if( state == s_set )
			value = o.value;
		return *this;
	}
	spec_optional & operator=( T const & v ) {
		state = s_set;
		value = v;
		return *this;
	}
	
	T & operator*() {
		PRE_CHECK( state == s_set );
		return value;
	}
	T const & operator*() const {
		PRE_CHECK( state == s_set );
		return value;
	}
};

} //eon util

#endif