#ifndef UTIL_VALIDATION_CONTEXT_HPP
#define UTIL_VALIDATION_CONTEXT_HPP

#include <unordered_set>

namespace util
{

/**
	The validation_context  allows validation routines to operator on potentially
	recrusive structures.
*/
struct validation_context
{
	std::unordered_set<void const*> validating;
	
	bool in( void const * what ) {
		return validating.find( what ) != validating.end();
	}

	/**
		Leaves 'what' in the structure. This is useful for 'validate' where there is
		no need to recheck an object multiple times.
		
		Only use on structures that are typically used in multiple places, like types.
		Many items, like statements/instructions, can only belong to one block
		so it's just wasteful to "cover" them.
	*/
	bool cover( void const * what ) {
		auto result = validating.insert( what );
		return !result.second;
	}
};

/**
	Provides a generic validate entry point that creates the context. This removes
	the need to have an overloaded validate on all objects.
*/
template<class T>
void validate( T const & what ) {
	validation_context vctx;
	what.validate( vctx );
}


/**
	A potentially recursive validation should use a scope to mark itself. This ensures
	it is removed once the scope is exited.
*/
struct validation_scope
{
	validation_context & ctx;
	void const * what;
	
	validation_scope( validation_context & ctx, void const * what ) 
		: ctx( ctx )
		, what( what ) {
		ctx.validating.insert( what );
	}
	~validation_scope() {
		ctx.validating.erase( what );
	}
};

} //eon util

#endif