# init for pyleaf
# -*- coding: utf-8 -*-
import os, yaml, platform
from shelljob import job, fs

# use job.FileMonitor for rough logging (I'm avoiding a better solution until I really need one)
class JobMonitor(job.Monitor):
	def __init__(self, shadow_monitor = None):
		super(JobMonitor,self).__init__()
		self.passed = 0
		self.ignored = 0
		self.omitted = 0
		self.failed = 0
		self.okay = True
		self.msg_failed = []
		self.msg_ignored = []
		self.msg_omitted = []
		self.shadow_monitor = shadow_monitor
		
	def job_started(self,jobj):
		super(JobMonitor,self).job_started(jobj)
		if self.shadow_monitor != None:
			self.shadow_monitor.job_started(jobj)
			
	def job_finished(self,jobj):
		super(JobMonitor,self).job_finished(jobj)
		if self.shadow_monitor != None:
			self.shadow_monitor.job_finished(jobj)
			
		if jobj.exit_code == 0:
			self.passed += 1
			msg = "\x1b[0;32mPASS 😊\x1b[m"
		elif jobj.exit_code == 2:
			self.omitted += 1
			msg = "\x1b[0;36mOMIT 😒\x1b[m"
		elif jobj.exit_code == 3:
			self.ignored += 1
			msg = "\x1b[0;33mIGNORE 😱\x1b[m"
		else:
			self.failed += 1
			msg = "\x1b[0;31mFAIL 😠\x1b[m"
			self.okay = False
			
		f = "{} {}".format( jobj.name, msg )
		print( f )
		
		if jobj.exit_code == 0:
			pass
		elif jobj.exit_code == 2:
			self.msg_omitted.append( f )
		elif jobj.exit_code == 3:
			self.msg_ignored.append( f )
		else:
			self.msg_failed.append( f )
		
	def job_output(self, jobj, line):
		super(JobMonitor,self).job_output(jobj,line)
		if self.shadow_monitor != None:
			self.shadow_monitor.job_output(jobj,line)

	def gen_feedback(self):
		pass
		

class LeafJob:
	def __init__(self):
		self.leaf_opts = ''
	
	def common_job( self, test, line ):
		jobj = job.Job('build/leaf {} --skip-ignore --skip-omit {}'.format( self.leaf_opts, line ))
		jobj.name = test
		return jobj
		
	def single_job( self, test ):
		return self.common_job( test, '--unit-test ' + test )

	def has_files( self, test ):
		files = fs.find( test, include_dirs = False, name_regex = '.*\\.leaf$', limit_depth = 0 )
		return len("".join(files)) > 0
		
	def dir_job( self, test ):
		if not self.has_files(test):
			return None
			
		files = fs.find( test, include_dirs = False, name_regex = '.*\\.leaf$', limit_depth = 0 )
		files_str = " ".join(files)
		if len(files_str) == 0:
			return None
		return self.common_job( test, '--unit-test ' + files_str )

	def exe_jit_job( self, test ):
		if not self.has_files(test):
			return None
			
		files = fs.find( test, include_dirs = False, name_regex = '.*\\.leaf$' )
		return self.common_job( test, '--exe-test ' + " ".join(files) )

	def lib_job( self, test ):
		files = fs.find( test, include_dirs = False, name_regex = '.*\\.leaf$' )
		return self.common_job( test, '--shared build/libs/{} {}'.format( 
			os.path.basename(test), " ".join(files)  ) )

	def exe_job( self, test ):
		if not self.has_files(test):
			return None
			
		jobj = job.Job('src/bin/exe_test.py {}'.format( test ) )
		jobj.name = test
		return jobj
	
	def build_libs( self, path, monitor ):
		def build( paths ):
			cmds = [ self.lib_job(test) for test in sorted(paths) ]
			# allows buidler to return None on some inputs
			cmds = filter(None,cmds)
			print( "Building libraries ({})".format( self.leaf_opts ) )
			monitor.run( cmds )
		
		if os.path.exists( path + '/order.yaml' ):
			# quick way to handle library dependencies
			decl = yaml.load( open(path + '/order.yaml').read() )
			for libs in decl['order']:
				build( [ path + '/' + lib for lib in libs ] )
				if not monitor.okay:
					break
		else:
			build( fs.find( path, include_files = False, exclude_root = True ) )

# This fixes up a defect in `env` on OSX where system protection causes it to strip vital environment variables
# The workaround involves exporting them also as APPLE_ and patching the environment here
def patch_environ():
	if platform.system() == "Darwin":
		patch = {}
		for key, value in os.environ.items():
			if key.startswith( "APPLE_"):
				patch[key[6:]] = value
				
		for key, value in patch.items():
			os.environ[key] = value
				