#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
	Execute a group of tests.
"""
import os, sys, argparse
import pyleaf
from shelljob import job, fs
import subprocess

pyleaf.patch_environ()

parser = argparse.ArgumentParser( description = 'Run Leaf tests' )
parser.add_argument( '--all', action='store_const', const=True )
parser.add_argument( '--unit', action='store_const', const=True )
parser.add_argument( '--mod', action='store_const', const=True )
parser.add_argument( '--exe', action='store_const', const=True )
parser.add_argument( '--exe-jit', action='store_const', const=True )

parser.add_argument( '--list-ignored', action='store_const', const=True )
parser.add_argument( '--list-omitted', action='store_const', const=True )
parser.add_argument( '--leaf-opts', action='store_const', const=True )
parser.add_argument( '--no-opts', action='store_const', const=True )
parser.add_argument( '--libs', action='store_const', const=True )

args = parser.parse_args()

# Collect build options
leaf_opts = ''
if args.leaf_opts:
	leaf_opts = ''
if args.no_opts:
	# TODO: reduce-flow not support, reduce-branch no longer relevant
	leaf_opts = '--reduce-branch=0 --reduce-flow=0'

leaf = pyleaf.LeafJob()
leaf.leaf_opts = leaf_opts

class TestMonitor(pyleaf.JobMonitor):
	def report(self):
		if args.list_omitted and len( self.msg_omitted ) > 0:
			print( " - - - - " )
			print( "\n".join( self.msg_omitted ) )
		if args.list_ignored and len( self.msg_ignored ) > 0:
			print( " - - - - " )
			print( "\n".join( self.msg_ignored ) )
		if len( self.msg_failed ) > 0:
			print( " - - - - " )
			print( "\n".join( self.msg_failed ) )
			
		print( " - - - - " )
		sys.stdout.write( '{} of {} Passed'.format( self.passed, self.passed + self.failed ) )
		if self.ignored != 0:
			sys.stdout.write( " (\x1b[0;33m{} Ignored\x1b[m)".format( self.ignored ) )
		if self.failed != 0:
			sys.stdout.write( " (\x1b[0;31m{} Failed\x1b[m)".format( self.failed ) )
		print( "")


def build_libs( path ):
	libs_run = TestMonitor()
	leaf.build_libs(path, libs_run)
	libs_run.report()
	if not libs_run.okay:
		sys.exit(1)
	
def run( name, tests, builder ):
	# run tests
	if not args.libs:
		cmds = [ builder(test) for test in sorted(tests) ]
		# allows buidler to return None on some inputs
		cmds = filter(None,cmds)

		print( "Running {} tests ({})".format( name, leaf_opts ) )
		tests_run = TestMonitor()
		tests_run.run( cmds )
		tests_run.report()
		if not tests_run.okay:
			sys.exit(1)

# Run selected tests
has_suite = False
if args.unit or args.all:
	has_suite = True
	tests = fs.find( 'share/unit_tests', include_dirs = False, name_regex = '.*\\.(leaf|lfx|lfb)$' )
	run( "Unit", tests, leaf.single_job )
	subprocess.call( ['build/leaf', '--unit-test', 'share/unit_tests/block/defer/basic.lfb' ])

if args.mod or args.all:
	has_suite = True
	build_libs( 'share/mod_tests/_libs' )
	dirs = fs.find( 'share/mod_tests', include_dirs = False, not_whole_name_regex = '.*/_.*', 
		name_regex = '.*\\.(lfb)$' )
	run( "Modules", dirs, leaf.single_job )

if args.exe or args.exe_jit or args.all:
	has_suite = True
	build_libs( 'share/exe_tests/_libs' )
	dirs = list(fs.find( 'share/exe_tests', include_files = False, 
		not_whole_name_regex = '.*/_.*', exclude_root = True) )
	if args.exe or args.all:
		run( "Standalone", dirs, leaf.exe_job )
	if args.exe_jit or args.all:
		# don't do those with tests.yaml since they can't be tested in JIT now
		run( "JIT", [ d for d in dirs if not os.path.exists( d+ '/tests.yaml' ) ], leaf.exe_jit_job )

if not has_suite:
	print( "Missing test suite" )
	sys.exit(1)
