#include "ir/compiler.hpp"
#include "ir/builder.hpp"
#include "ir/type_manager.hpp"
#include "ir/std_functions.hpp"
#include "ir/dump.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

/*
	STATE_CHECK: Note that we use STATE_CHECK here in many places where you might think
	PRE_CHECK should be used. This is because all functions tend to modify directly the compiled
	IR module/function. We aren't concerned with internal unwinding at all, thus we simply 
	cause a recovery error.
*/
namespace ir
{

compiler::compiler( ) {
}

void compiler::init_module( std::string const & name ) {
	cur_module = make_shared<ir::module>( name, *this );
}

value_ref compiler::initialize( value_ref ptr, shared_ptr<type const> init_type, 
	unsigned flags, builder & where )
{
	if( init_type->is_shared_pointer() ) {
		auto mi = gen_shared( init_type, where );
		assign( ptr, mi, where, where, af_no_acquire | af_no_release );
		ptr = where.access_raw_shared( mi );

		init_type = init_type->sub[0];
	}
		
	if( flags & if_undefined_value )
		return ptr;
		
	assign( ptr, gen_default( init_type, where ), 
		where, where, af_no_release | af_no_acquire );
	return ptr;
}

value_ref compiler::gen_shared( shared_ptr<type const> t, builder & where )
{
	STATE_CHECK( t->is_shared_pointer() );
	
	auto mi = make_shared<alloc_shared_inst>();
	mi->sod.direct_type = t->sub[0];
	mi->sod.result = where.make_temp( t );
	configure_destructor( mi->sod, where );
	where.add( mi );
	
	return mi->sod.result;
}

value_ref compiler::gen_default( shared_ptr<type const> gt, builder & where )
{
	switch( gt->fundamental )
	{
		case type::f_optional: {
			auto pt = where.create_pack_tuple_inst( gt );
			pt->sub.push_back( where.const_boolean( false ) );
			
			auto undef = make_shared<value>(value::f_undefined);
			undef->direct_type = pt->result.result_type()->sub[1];
			pt->sub.push_back( value_ref::as_value(undef) );
			where.add( pt );
			return pt->result;
		}
		
		case type::f_shared_pointer: {
			auto nv = gen_shared( gt, where );
			auto d = gen_default( gt->sub[0], where );
			assign( nv, d, where, where, af_no_acquire | af_no_release | af_init_refer );
			return nv;
		}

		case type::f_integer: {
			auto v = make_shared<value>(value::f_const_integer);
			v->direct_type = gt;
			v->int_value = 0;
			return value_ref::as_value(v);
		}
			
		case type::f_binary: {
			auto v = make_shared<value>(value::f_const_binary);
			v->direct_type = gt;
			v->uint_value = 0;
			return value_ref::as_value(v);
		}
		
		case type::f_float: {
			auto v = make_shared<value>(value::f_const_float);
			v->direct_type = gt;
			v->float_value = 0;
			
			return value_ref::as_value(v);
		}
		
		case type::f_boolean: {
			auto v = make_shared<value>(value::f_const_boolean);
			v->direct_type = gt;
			v->boolean_value = false;
			return value_ref::as_value(v);
		}
		
		case type::f_function_ptr: {
			if( gt->access == type::a_bound ) {
				auto vc = where.create_void_context_inst( gt );
				where.add( vc );
				return vc->result;
			} else {
				//TODO: this should also be a throw function, but do this for now
				auto undef = make_shared<value>(value::f_undefined);
				undef->direct_type = gt;
				return value_ref::as_value( undef );
			}
		}
		
		case type::f_tuple: {
			auto pt = where.create_pack_tuple_inst( gt );
			for( size_t i=0; i < gt->sub.size(); ++i )
				pt->sub.push_back( gen_default( gt->sub[i], where ) );
			where.add(pt);
			return pt->result;
		}
		
		case type::f_array: {
			auto aa = where.add_alloc_array_inst( gt, gt->sub[0], 
				where.const_integer_normal( 0 ) );
			return aa->result;
		}
		
		case type::f_raw_pointer: {
			auto v = make_shared<value>(value::f_nullptr);
			v->direct_type = gt;
			return value_ref::as_value(v);
		}
		
		case type::f_abi_var_args: {
			STATE_FAIL( "unsupported-default-abi-var-args" );
		}
	}
	STATE_FAIL();
}

value_ref compiler::gen_static_binary_data( uint8_t const * data, size_t len, builder & /*where*/ ) {
	
	auto v = make_shared<value>(value::f_raw_binary);
	v->raw_binary.insert( v->raw_binary.begin(), data, data + len );

	auto st = get_types().get_octet();
	auto t = get_types().get_raw_pointer(st);
	v->direct_type = t;
	
	//TODO: mark as static and must persist
	return value_ref::as_value(v);
}


void compiler::assign( value_ref_vector_t & left, value_ref_vector_t & right, builder & where,
	unsigned flags )
{
	auto sb = where.create_block();
	builder store_block( *sb );

	DEBUG( left.size(), right.size() );
	STATE_CHECK( left.size() == right.size() );
	if( left.size() > 1 )
		flags |= af_force_resolve;
		
 	for( size_t i=0; i < left.size(); ++i )
 		assign( left[i], right[i], where, store_block, flags );
		
	where.add( sb );
}

/**
	@param left this must be a pointer and it will be derefenced for the store. It needs to be
		like this since the lvalue's get passed around as a values for a while before they are
		actually stored to.
*/
void compiler::assign( value_ref left, value_ref right, builder & where, builder & store_block, unsigned flags )
{
	where.add_comment( "assign" );
	if( flags & af_force_resolve )
	{
		//the store block may not immediately follow the access block, this allows proper
		//ordering where variables are used on the left and right (we also must forced
		//the evaluation now). This is only needed for value-list.size()>1 assigns since
		//the store_inst itself guarantees the ordering in single-value cases
		left = where.resolve( left );
		right = where.resolve( right );
	}
	
	assign_copy_value( left, right, store_block, flags );
	where.add_comment( "assign-end" );
}

value_ref compiler::assign_arg( value_ref in, builder & where )
{
	auto t = in.result_type();
	if( t->fundamental == type::f_array ) {
		auto call = where.add_funccall_inst( get_array_copy_func( *this, t->item_type() ) );
		call.add_arg( in );
		auto dup = call.define_return( t );
		release( dup, where, block_flow::defer );
		return dup;
	} else {
		return in;
	}
}

void compiler::assign_copy_value( value_ref dst, value_ref src, builder & where, unsigned flags )
{
	auto dst_t = dst.result_type();
	
	if( flags & af_init_refer ) {
		if( dst_t->is_shared_pointer() )
			dst = where.access_raw_shared(dst);
		STATE_CHECK( dst.result_type()->is_raw_pointer() );
		assign_copy_value( dst, src, where, flags & ~(af_init_refer) );
		return;
	}
	
	auto t = src.result_type();
	STATE_CHECK( t->copy == type::c_default, dump::get_type(dst_t.get()) );
	if( t->fundamental == type::f_tuple ) {
		for( size_t i=0; i < t->sub.size(); ++i ) {
			assign_copy_value( 
				where.access_post_field( dst, i ),
				where.access_post_field( src, i ),
				where, flags );
		}
		return;
	}

	boost::optional<value_ref> prev;
	if( !(flags & af_no_release) ) {
		prev = get_release_pointer( dst, where, rf_pointer );
	}
		
	if( t->fundamental == type::f_array ) {
		//in this case we have a temporray we can just take
		if( flags & af_no_acquire ) {
			where.add_store_inst( dst, src );
		} else {
			auto call = where.add_funccall_inst( get_array_copy_func( *this, t->item_type() ) );
			call.add_arg( src );
			auto dup = call.define_return( t );
			where.add_store_inst( dst, dup );
		}
	} else {
		if( !(flags & af_no_acquire) )
			acquire( src, where );
		where.add_store_inst( dst, src );
	}
	
	if( prev ) {
		//must be deferred to end of assign block since the same object could be used mutiple times
		//on both sides of the assignment (shared swap is canonical example)
		//NOTE: this may seem wrong, since the right-side must be evaluated/acquired prior to this point, but
		//assignment is piecemeal, not all acquires are done first
		release_pointer( *prev, where, block_flow::defer ); 
	}
}

/**
	Acquire the object if it is a shared object type.
	
	@return true if something must be done, false if nothing
*/
bool compiler::acquire( value_ref value, builder & where )
{
	auto rt = value.result_type();
	switch( rt->fundamental )
	{
		case type::f_function_ptr:
			if( rt->access == type::a_bound ) {
				where.add( where.create_acquire_shared_inst( value ) );
				return true;
			}
			return false;
		case type::f_shared_pointer:
			where.add( where.create_acquire_shared_inst( value ) );
			return true;
			
		case type::f_tuple: {
			//FEATURE: if the acqurie of a tuple is large enough it makes sense to push it into
			//a function instead (or perhaps just always a function and let inlining optimized if appropriate)
			bool d = false;
			for( size_t i=0; i < rt->sub.size(); ++ i ) {
				//OPT: if return is false then the access_pre_field code need not be generated...
				bool e = acquire( where.access_pre_field( value, i ), where );
				d = d || e;
			}
			return d;
		}
			
		case type::f_optional: {
			auto sb = where.create_switch();
			sb.add_case();
			sb.set_cond_value( sb.get_cond_block().access_optional_set( value ) );
			bool d = acquire( sb.get_expr_block().access_optional_value( value ), sb.get_expr_block() );
			if( d )
				sb.add();
			return d;
		}
			
		default:
			return false;
	}
}

/**
	Obtains a value_ref which is suitable for later releasing via release_pointer. 
	This ensure the correct version of the the value is being released.
	
	@return null if nothing needs to be released, otherwise the value to release
*/
boost::optional<value_ref> compiler::get_release_pointer( value_ref value, builder & where, unsigned flags ) {
	auto rt = value.result_type();
	if( flags & rf_pointer ) {
		STATE_CHECK( rt->is_raw_pointer() );
		rt = rt->sub[0];
	}
	//short-circuit if we know nothing is to be done
	if( rt->is_trivial_destruction() )
		return boost::optional<value_ref>();

	if( flags & rf_cur_value ) {
		return where.access_raw_pointer( value );
	} else {
		if( flags & rf_pointer ) {
			return value_ref(where.add_clone( value ));
		} else {
			return value_ref(where.add_clone_value( value ));
		}
	}
}

void compiler::release( value_ref value, builder & where, block_flow flow, unsigned flags )
{
	auto rv = get_release_pointer( value, where, flags );
	if( rv ) {	
		release_pointer( *rv, where, flow, flags );
	}
}

/**
	Release the object if it is a shared object type.
*/
void compiler::release_pointer( value_ref value, builder & where, block_flow flow, unsigned flags )
{
	auto rt = value.result_type();
	STATE_CHECK( rt->fundamental == type::f_raw_pointer );
	rt = rt->sub[0];
	//short-circuit if we know nothing is to be done
	if( rt->is_trivial_destruction() )
		return;

	if( !(flags & rf_inline) ) {
		auto dtor = get_release_function( *this, rt );
		auto fc = where.create_funccall_inst( dtor );
		fc.add_arg( value );
		where.push_back( fc.inst, flow );
		return;
	}
		
	//custom destructor (happend prior to standard destruction)
	if( rt->destroy ) {
		auto fc = where.create_funccall_inst( *rt->destroy );
		fc.set_context( value );
		where.push_back( fc.inst, flow );
	}

	auto referred = [&](){
		auto ad = where.access_deref( value );
		return ad;
	};
	
	switch( rt->fundamental )
	{
		case type::f_function_ptr: {
			if( rt->access == type::a_bound ) {
				auto rs = where.create_release_shared_inst( referred() );
				rs->destroy_how = release_shared_inst::dh_dynamic;
				where.push_back( rs, flow );
				break;
			}
		}
		
		case type::f_shared_pointer: {
			auto rs = where.create_release_shared_inst( referred() );
			if( rt->sub[0]->is_trivial_destruction() )
				rs->destroy_how = release_shared_inst::dh_trivial;
			else
				rs->destroy_how = release_shared_inst::dh_dynamic;
			where.push_back( rs, flow );
			break;
		}

		case type::f_optional: {
			auto sb = where.create_switch();
			
			sb.add_case();
			sb.set_cond_value( sb.get_cond_block().access_optional_set( referred() ) );
			release_pointer( sb.get_expr_block().access_optional_value( value ), 
				sb.get_expr_block(), block_flow::sequential, flags );
			sb.push_back(flow);
			break;
		}
		
		case type::f_tuple: {
			for( size_t i=0; i < rt->sub.size(); ++i ) {
				auto field = where.access_pre_field( value, i );
				release_pointer( field, where, flow, flags );
			}
			break;
		}
		
		case type::f_array: {
			auto fc = where.create_funccall_inst( get_delete_array_func( *this, rt->item_type() ) );
			fc.add_arg( referred() );
			fc.push_back(flow);
			break;
		}
			
		default: 	
			break;
	}
}

void compiler::configure_destructor( shared_object_defn & sod, builder & /*where*/ )
{
	if( sod.direct_type->is_trivial_destruction() )
		sod.static_destructor = true;
	else
		sod.destructor = get_release_function( *this, sod.direct_type );
}

/**
	NOTE: Refer to leaf::expr_transform::t_shared_expr which also creates these
*/
compiler::symbol_cref & compiler::get_symbol_cref( intptr_t & cref )
{
	if( cref )
		return *reinterpret_cast<symbol_cref*>(cref);
		
	symbol_deque.resize( symbol_deque.size() + 1 );
	symbol_cref & sc = symbol_deque.back();
	sc.origin = make_shared<value>();
	cref = reinterpret_cast<intptr_t>(&sc);
	sc.ref = value_ref( sc.origin );
	return sc;
}

void compiler::define_symbol_cref( intptr_t & cref, value_ref vr ) {
	STATE_CHECK( !cref );
	
	symbol_deque.resize( symbol_deque.size() + 1 );
	symbol_cref & sc = symbol_deque.back();
	sc.origin = vr.origin;
	sc.ref = vr;
	cref = reinterpret_cast<intptr_t>(&sc);
}

} //eon ir