#include "ir/compiler_leaf.hpp"
#include "ir/builder.hpp"
#include "ir/type_manager_leaf.hpp"
#include "ir/std_functions.hpp"

#include "lang/language.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"
#include "ir/dump.hpp"

namespace ir {

namespace {
	std::string export_name_of( leaf::declaration const & decl, type const & t ) {
		std::string use_name = decl.var_scope->qualify_name( decl.symbol, platform::export_sep);
		//multi's must have a distinct name, but predictable as export symbol
		if( decl.form == leaf::declaration::f_multi ) {
			use_name += platform::export_sep;
			//each override must have a unique type, so this works
			use_name += t.get_export_name();
		}
		
		return use_name;
	}
	
	std::string import_name_of( leaf::declaration const & decl, type const & t ) {
		std::string use_name;
		switch( decl.name_form ) {
			case leaf::declaration::nf_bare:
				use_name = decl.import_name;
				break;
				
			//TODO: it's questionable whether this should even be supported. From Leaf=>Leaf must always
			//have the .lo available
			case leaf::declaration::nf_leaf:
				if( decl.import_module.size() ) {
					use_name = decl.import_module;
					use_name += platform::export_sep;
				}
				use_name += decl.import_name;
				
				if( decl.form == leaf::declaration::f_multi ) {
					use_name += platform::export_sep;
					use_name += t.get_export_name();
				}
				break;
			
			case leaf::declaration::nf_auto:
				use_name = decl.var_scope->qualify_name( decl.symbol, platform::export_sep );
				if( decl.form == leaf::declaration::f_multi ) {
					use_name += platform::export_sep;
					use_name += t.get_export_name();
				}
				break;
		}
		
		return use_name;
	}
}

#include "compiler_leaf/statement.inc.cpp"
#include "compiler_leaf/expression.inc.cpp"

compiler_leaf::compiler_leaf( shared_ptr<leaf::scope> globals )
	: globals( globals ) 
	, types( type_manager_leaf::create(*this) ) {
}

type_manager & compiler_leaf::get_types() {
	return *types;
}

shared_ptr<ir::module> compiler_leaf::compile_expr( leaf::expression const & expr ) {
	util::validate( expr );
	init_module( "single_expr" );
	
	module_builder mb( *this );
	auto fb = mb.create_function( "eval_expr", true );
	//no args
	auto rv = fb.define_return( types->get( expr.type ), "value" );
	
	builder bb( fb );
	bb.add_store_inst( bb.access_raw_pointer(rv), eval_single_expr( expr, bb ) );
	return cur_module;
}

shared_ptr<ir::module> compiler_leaf::compile_block( leaf::statement_block const & bk,
	leaf::statement_block const & init ) {
	util::validate( bk );
	init_module( "single_block" );

	module_builder mb( *this );
	auto fb = mb.create_function( "_entry", true );
	//no args
	//no returns

	builder bb(fb);
	build_imperative_block( init, bb );
	build_imperative_block( bk, bb );
	return cur_module;
}

shared_ptr<ir::module> compiler_leaf::compile_module( leaf::module const & leaf_module ) {
	util::validate( leaf_module );
	init_module( leaf_module.name ? *leaf_module.name : "main" );

	//split the initialization into init and final for the module. This decomposes all
	//"init" statements into the init part, and the deferred deletion (put in another function).
	//deferred statements are also put in the _final function.
	module_builder mb( *this );
	auto initf = mb.create_function( globals->qualify_name("_init_module",platform::export_sep), false);
	initf.owner_function->is_export = true;
	auto finalf = mb.create_function( globals->qualify_name("_final_module",platform::export_sep), false );
	finalf.owner_function->is_export = true;
	
	builder bb(initf);
	bb.split_defer = finalf.owner_function->entry_block.get();
	build_imperative_block( *leaf_module.static_init_block, bb );
	bool has_imperative = false;
	for( auto & block : leaf_module.blocks ) {
		if( block->get_scope_style() == leaf::scope_style_t::declarative ) {
			build_declarative_block( *block, bb );
		} else {
			has_imperative = true;
		}
	}
	
	STATE_CHECK( !has_imperative );
	
	//create an entry point if "main" exists
	auto main = cur_module->if_get_global_variable( "main" );
	if( main ) {
		auto entryf = mb.create_function( "_entry", true );
		builder eb( entryf );
		
		eb.add_funccall_inst( initf.get_value_defn() );
		
		auto & ftype = *main->direct_type;
		STATE_CHECK( ftype.is_callable() );
		auto fc = eb.add_funccall_inst( value_defn(main) );
		
		//is there a return value
		if( ftype.sub.size() ) {
			auto func_ret = entryf.define_return( ftype.sub_type(0), "rv" );
			auto rv = fc.define_return( ftype.sub_type(0) );
			assign( eb.access_raw_pointer(func_ret), rv, eb, eb );
		}
		
		eb.add_funccall_inst( finalf.get_value_defn() );
	}
	
	return cur_module;
}

	
/**
	At this level it isn't easy (and thus not done) to determine if a variable may be accessed
	prior to the "init" for the variable. The LLVM lowering will however catch it.
*/
value_ref compiler_leaf::obtain_symbol_value_ref( shared_ptr<leaf::declaration> sr ) {
	auto & sref = get_symbol_cref( sr->cref );
	auto sval = sref.origin;
	if( sval->direct_type ) {
		//already done
		return sref.ref;
	}
	sval->direct_type = types->get( sr->type );
	
	switch( sr->import_type ) {
		case leaf::declaration::it_none: {
			if( sr->form == leaf::declaration::f_multi || sr->form == leaf::declaration::f_single ) {
				if( sr->get_root_depth() == 0 || sr->lifetime == leaf::declaration::l_global ) {
					sval->form = value::f_global;
				} else {
					sval->form = value::f_local;
				}
			} else if( sr->form == leaf::declaration::f_capture ) {
				sval->form = value::f_capture_ref;
			} else {
				STATE_FAIL( sr->symbol );
			}
			
			sval->name = export_name_of( *sr, *sval->direct_type );
			break;
		}
		case leaf::declaration::it_abi_code: {
			sval->form = value::f_import_code;
			sval->name = import_name_of( *sr, *sval->direct_type );
			break;
		}
		case leaf::declaration::it_abi_data: {
			sval->form = value::f_import_data;
			sval->name = import_name_of( *sr, *sval->direct_type );
			break;
		}
	}
	return sref.ref;
}

void compiler_leaf::initialize_leaf( value_ref ptr, leaf::extr_type const & init_type_, 
	shared_ptr<leaf::expression> value, unsigned flags, builder & where ) {
	auto init_type = types->get( init_type_ );
	if( value ) {
		if( flags & if_init_refer_value )
			ptr = initialize( ptr, init_type, flags | if_undefined_value, where );
		
		auto right = eval_single_expr( *value, where );
		assign( ptr, right, where, where, af_no_release );
	} else {
		initialize( ptr, init_type, flags, where );
	}
}

} //eon ir