/**
	This evaluates an expression where a value-list is allowed. This is done to avoid creating
	tuples where they are not actually in the source-code -- which involves potentially 
	needless copying	and a loss of clarity in the output IR.
	
	Compound value-lists are flattened here. As it is assumed the input leaf structure is valid
	this shouldn't introduce any problems.
*/
compiler_leaf::value_ref_vector_t compiler_leaf::eval_list_expr( leaf::expression const & expr, builder & where, 
	unsigned flags ) {
	value_ref_vector_t out;
	eval_list_expr( out, expr, where, flags );
	return out;
}

void compiler_leaf::eval_list_expr( value_ref_vector_t & out, leaf::expression const & expr, builder & where,
	unsigned flags ) {
	switch( expr.form )
	{
		case leaf::expression::f_value_list:
		{
			auto & evl = static_cast<leaf::expr_value_list const&>(expr);
			for( size_t i=0; i < evl.sub.size(); ++i )
				eval_list_expr( out, *evl.sub[i].expr, where, flags );
			break;
		}
			
		case leaf::expression::f_funccall:
			eval_funccall_expr( out, static_cast<leaf::expr_funccall const&>(expr), where, flags );
			break;

		default: 
			out.push_back( eval_single_expr( expr, where, flags ) );
			break;
	}
}

void compiler_leaf::eval_funccall_expr( value_ref_vector_t & out, leaf::expr_funccall const & expr, builder & where,
	unsigned /*flags*/ ) {
	auto fi = where.create_funccall_inst( eval_single_expr( *expr.func, where, ef_calling ) );
	
	auto & ft = expr.func->type.extr().cast<leaf::intr_type_function>();
	auto ptypes = ft.args;
	
	STATE_CHECK( expr.args->form == leaf::expression::f_value_list );
	auto & ep = static_cast<leaf::expr_value_list&>(*expr.args);
	STATE_CHECK( ep.sub.size() == ptypes->sub.size() );
	for( size_t i=0; i < ep.sub.size(); ++i )
	{
		unsigned flags = 0;
			
		value_ref ex = eval_single_expr( *ep.sub[i].expr, where, flags );
		fi.add_arg( assign_arg( ex, where ) );
	}
		
	if( expr.context ) {
		value_ref ex = eval_single_expr( *expr.context, where );
		//TODO: this access_raw_pointer seems highly suspect!
		if( ex.result_type()->is_shared_pointer() ) {
			fi.set_context( where.access_raw_shared(ex) );
		} else {
			fi.set_context( where.access_raw_pointer(ex) );
		}
	}
	fi.add();
	
	leaf::intr_type const & vt = *ft.returns;
	auto & vl = vt.cast<leaf::intr_type_value_list>();
	for( size_t i=0; i < vl.sub.size(); ++i )
	{
		auto ref = fi.define_return( types->get( vl.sub[i].type ) );
		out.push_back( ref );
		//returns will be acquired in target function, so release them now
		//TODO: should release at end of current statement (not end of block)
		release( ref, where, block_flow::defer );
	}
	
}

value_ref compiler_leaf::eval_single_expr( leaf::expression const & expr, builder & where, unsigned /*flags*/ ) {
	value_ref vr;
	
	switch( expr.form )
	{
		case leaf::expression::f_block: {
			auto b = static_cast<leaf::expr_block const &>(expr);
			build_imperative_block( *b.block, where );
			if( b.result ) {
				return eval_single_expr( *b.result, where );
			}
			
			//hmm, this works, but should it (nothing uses the value)
			return vr; 
		}

		case leaf::expression::f_symbol: {
			auto es = static_cast<leaf::expr_symbol const&>(expr);
			
			STATE_CHECK( es.sref );
			value_ref vr = obtain_symbol_value_ref( es.sref );
			if( es.type.extr().is_lvalue() )
				vr = where.access_raw_pointer(vr);
			return vr;
		}
			
		case leaf::expression::f_funccall:
		{
			value_ref_vector_t rets;
			eval_funccall_expr( rets, static_cast<leaf::expr_funccall const&>(expr), where );
			if( rets.size() != 1 ) {
				auto bt = where.create_pack_tuple_inst( types->get( expr.type ) );
				for( size_t i=0; i < rets.size(); ++i )
					bt->sub.push_back( rets[i] );
				where.add( bt );
				vr = bt->result;
			} else {
				vr = rets[0];
			}
			return vr;
		}

		//NOTE: I'm trying to prevent needing this, I'd prefer the value_list's don't exist at this
		//point as actual values. Certain conversions may require this to be reintroduced (just
		//remove the check and allow it to work as a tuple)
		case leaf::expression::f_value_list:
			STATE_CHECK(false);
			
		case leaf::expression::f_compound: {
			auto & et = static_cast<leaf::expr_with_sub const&>(expr);
			auto bt = where.create_pack_tuple_inst( types->get( expr.type ) );
			for( size_t i=0; i < et.sub.size(); ++i )
				bt->sub.push_back( eval_single_expr( *et.sub[i].expr, where ) );
			where.add( bt );
			vr = bt->result;
			return vr;
		}
		
		case leaf::expression::f_type: {
			//empty tuple
			auto bt = where.create_pack_tuple_inst( types->get( expr.type ) );
			where.add( bt );
			vr = bt->result;
			return vr;
		}
		
		case leaf::expression::f_literal_number: {
			auto ecn = static_cast<leaf::expr_literal_number const&>(expr);
			switch( ecn.src_form ) {
				case leaf::expr_literal_number::sf_rational:
					STATE_FAIL( "rational no longer valid" );
				case leaf::expr_literal_number::sf_integer:
					vr = where.const_integer( types->get( expr.type ), ecn.integer_value() );
					break;
				case leaf::expr_literal_number::sf_binary:
					vr = where.const_binary( types->get( expr.type ), ecn.integer_value() ); //TODO: requires unsigned conversion
					break;
				case leaf::expr_literal_number::sf_float:
					vr = where.const_float( types->get( expr.type ), double(ecn.float_value()) );
					break;
				case leaf::expr_literal_number::sf_boolean:
					vr = where.const_boolean( ecn.boolean_value() );
					break;
			}
			return vr;
		}
			
		case leaf::expression::f_switch:
			return eval_switch( static_cast<leaf::expr_switch const&>(expr), where );
		
		case leaf::expression::f_funcdefn:
			return build_function( static_cast<leaf::expr_funcdefn const&>(expr), where );
			
		case leaf::expression::f_subscript:
			return eval_subscript( static_cast<leaf::expr_subscript const&>(expr), where );
		
		case leaf::expression::f_bind:
			return eval_bind( static_cast<leaf::expr_bind const&>(expr), where );
			
		case leaf::expression::f_field:
			return eval_field( static_cast<leaf::expr_field const&>(expr), where );
			
		case leaf::expression::f_special_field:
			return eval_special_field( static_cast<leaf::expr_special_field const&>(expr), where );

		case leaf::expression::f_transform:
			return eval_transform( static_cast<leaf::expr_transform const&>(expr), where );
			
		case leaf::expression::f_binary:
			return eval_binary( static_cast<leaf::expr_binary const&>(expr), where );
			
		case leaf::expression::f_unary:
			return eval_unary( static_cast<leaf::expr_unary const&>(expr), where );
		
		case leaf::expression::f_new_func:
		{
			auto & nf = static_cast<leaf::expr_new_func const&>(expr);
			vr = get_new_func( *this, types->get( nf.new_type ) );
			return vr;
		}
		
		case leaf::expression::f_temp:
			return eval_temp( static_cast<leaf::expr_temp const&>(expr), where );
		
		case leaf::expression::f_conditional:
			return eval_conditional( static_cast<leaf::expr_conditional const&>(expr), where );
			
		case leaf::expression::f_text:
			return eval_text( static_cast<leaf::expr_text const&>(expr), where );
			
		case leaf::expression::f_current:
			return eval_current( static_cast<leaf::expr_current const&>(expr), where );
			
		case leaf::expression::f_cast:
			//should disappear during typing
			STATE_FAIL( "unexpected-expression" );
			
		case leaf::expression::f_compat:
			return eval_single_expr( *static_cast<leaf::expr_compat const&>(expr).source, where );
			
		case leaf::expression::f_tag:
			return eval_tag( static_cast<leaf::expr_tag const &>(expr), where );
	}

	STATE_FAIL("unsupported-expression" );
}

/**
	To use a lazy-switch in evaluation requires a local variable of the result. This is required
	since the actual result will be elevated out of the block in which is created, which is
	not allowed. So instead we store in a local variable. This also allows the deferred destruction
	to happen in the correct scope.
*/
value_ref compiler_leaf::eval_switch( leaf::expr_switch const & es, builder & where ) {
	auto sb = where.create_switch( types->get(es.type) );
	for(size_t i=0; i < es.expr_list.size(); ++i ) {
		sb.add_case();
		if( es.expr_list[i].cond ) {
			sb.set_cond_value( eval_single_expr( *es.expr_list[i].cond, sb.get_cond_block() ));
		}
			
		sb.set_expr_value( eval_single_expr( *es.expr_list[i].value, sb.get_expr_block() ) );
	}
	sb.add();
	return sb.get_result();
#if 0
	//this form could be used if we can prove there is no defer in any of the contained blocks
	auto sb = where.create_switch( types->get(es.type) );
	for(size_t i=0; i < es.expr_list.size(); ++i ) {
		sb.add_case();
		if( es.expr_list[i].cond )
			sb.set_cond_value( eval_single_expr( *es.expr_list[i].cond, sb.get_cond_block() ));
		sb.set_expr_value( eval_single_expr( *es.expr_list[i].value, sb.get_expr_block() ));
	}
	vr = sb.get_result();
	sb.add();
	return vr;
#endif
}

value_ref compiler_leaf::eval_transform( leaf::expr_transform const & et, builder & where, 
	unsigned /*eval_flags*/ ) {
	auto what_pl = []( leaf::intr_type const & vt ) {
		if( vt.is<leaf::intr_type_integer>() )
			return promote_loss_inst::w_integer;
		if( vt.is<leaf::intr_type_binary>() )
			return promote_loss_inst::w_binary;
		if( vt.is<leaf::intr_type_float>() )
			return promote_loss_inst::w_float;
		STATE_FAIL( "unsupported-promotion" );
	};
	
	switch( et.transform )
	{
		case leaf::expr_transform::t_deref: {
			bool retain_lvalue = et.arg->type.extr().is_lvalue();
			value_ref vr = eval_single_expr( *et.arg, where );
			if( et.arg->type.extr().is_shared() ) {
				if( retain_lvalue ) {
					return where.access_raw_shared( where.access_deref( vr ) );
				}
				return where.access_deref_shared( vr );
			} else {
				STATE_CHECK( et.arg->type.extr().is_value_ptr() );
//  				if( retain_lvalue ) {
//  					return vr;
//  				}
				return where.access_deref( vr );
			}
		}
	
		case leaf::expr_transform::t_drop_lvalue: {
			value_ref vr = eval_single_expr( *et.arg, where );
			return where.access_deref( vr );
		}
		
		case leaf::expr_transform::t_addr_of: {
			value_ref vr = eval_single_expr( *et.arg, where );
			if( et.arg->type.extr().is_lvalue() ) {
				//already a pointer
			} else if( et.arg->type.extr().is<leaf::intr_type_array>() ) {
				vr = where.access_raw_array( vr );
			} else {
				STATE_FAIL( "expecting-address" );
			}
			return vr;
		}
			
		case leaf::expr_transform::t_shared_expr: {
			if( et.cref ) {
				return reinterpret_cast<symbol_cref*>(et.cref)->ref;
			} else {
				value_ref vr = eval_single_expr( *et.arg, where );
				
				symbol_deque.resize( symbol_deque.size() + 1 );
				symbol_cref & sc = symbol_deque.back();
				sc.origin = vr.origin;
				sc.ref = vr;
				et.cref = reinterpret_cast<intptr_t>(&sc);
				return vr;
			}
		}
			
		case leaf::expr_transform::t_promote: {
			return where.add_promote_loss( false, 
				eval_single_expr( *et.arg, where ), 
				types->get( et.type ), what_pl( *et.arg->type.extr().intr ), 
					what_pl( *et.type.extr().intr ) );
		}
		case leaf::expr_transform::t_loss: {
			return where.add_promote_loss( true,
				eval_single_expr( *et.arg, where ), 
				types->get( et.type ), what_pl( *et.arg->type.extr().intr ), 
					what_pl( *et.type.extr().intr ) );
		}
		
		case leaf::expr_transform::t_code_val: {
			return where.add_cast( eval_single_expr( *et.arg, where ), types->get( et.type ) );
		}
		case leaf::expr_transform::t_char_val: {
			return where.add_cast( eval_single_expr( *et.arg, where ), types->get( et.type ) );
		}
		
		case leaf::expr_transform::t_make_optional: {
			return where.create_optional( 
				eval_single_expr( *et.arg, where ),
				types->get( et.type )
			);
		}
		
		case leaf::expr_transform::t_deref_optional: {
			auto sb = where.create_switch();
			
			sb.add_case();
			auto base = eval_single_expr( *et.arg, sb.get_cond_block() );
			sb.set_cond_value( sb.get_cond_block().access_optional_set( base ) );
			
			//TODO: rather than else a "not" on the cond
			sb.add_else();
			auto fail = make_shared<type>( get_types().shared_from_this(), type::a_raw );
			auto v = where.define_value( value::f_import_code, fail, "_access_unset_optional" );
			sb.get_expr_block().add_funccall_inst( v );
			
			sb.add();
			return where.access_optional_value( base );
		}
	}
	
	STATE_FAIL("invalid transform");
}

value_ref compiler_leaf::eval_conditional( leaf::expr_conditional const & ec, builder & where, 
	unsigned /*eval_flags*/ ) {
	switch( ec.op ) {
		case leaf::expr_conditional::o_optional_or: {
			auto sb = where.create_switch( types->get(ec.type) );
			
			sb.add_case();
			auto base = eval_single_expr( *ec.cond, sb.get_cond_block() );
			sb.set_cond_value( sb.get_cond_block().access_optional_set( base ) );
			auto value = eval_single_expr( *ec.a, sb.get_expr_block() );
			sb.set_expr_value( value );
						
			sb.add_else();
			value = eval_single_expr( *ec.b, sb.get_expr_block() );
			sb.set_expr_value(value);
			
			sb.add();
			return sb.get_result();
		}
		
		case leaf::expr_conditional::o_optional_cond: {
			auto sb = where.create_switch( types->get(ec.type) );
			
			sb.add_case();
			sb.set_cond_value( eval_single_expr( *ec.cond, sb.get_cond_block() ) );
			auto value = eval_single_expr( *ec.a, sb.get_expr_block() );
			value = sb.get_expr_block().create_optional( value, types->get(ec.type) );
			sb.set_expr_value( value);
			
			sb.add_else();
			//type is optional, so default is an unset optional
			value = gen_default( types->get(ec.type), sb.get_expr_block() );
			sb.set_expr_value( value );
			
			sb.add();
			return sb.get_result();
		}
	}
	
	STATE_FAIL("invalid conditional");
}

value_ref compiler_leaf::eval_text( leaf::expr_text const & expr, builder & where ) {
	switch( expr.what) {
		case leaf::expr_text::w_char: {
			return where.const_binary( types->get(expr.type), uintmax_t(expr.text[0]) );
		}
		
		case leaf::expr_text::w_string: {
			auto at = types->get(expr.type);
			auto ctype = at->sub[0];
			auto ar = where.add_alloc_array_inst( at, ctype, 
				where.const_integer_normal( int(expr.text.size())) );
				
			auto local = where.define_local( at, "const_string" );
			auto lptr = where.access_raw_pointer( local );
			where.add_store_inst( lptr, ar->result );
			auto aptr = lptr;
			for( size_t i=0; i < expr.text.size(); ++i ) {
				auto off = where.add_array_offset( aptr, 
					where.const_integer_normal(int(i)) );
				auto val = where.const_binary( ctype, uintmax_t(expr.text[i]) );
				assign( off, val, where, where );
			}
			
			release( ar->result, where, block_flow::defer );
			return ar->result;
		}
	}
	
	STATE_FAIL("invalid text" );
}

value_ref compiler_leaf::eval_clone_func( leaf::expr_bind const & et, builder & where ) {
	STATE_CHECK( et.unbound->form == leaf::expression::f_funcdefn );
	auto & func_defn = static_cast<leaf::expr_funcdefn const&>(*et.unbound);
	
	auto & func_type = et.unbound->type.extr().cast<leaf::intr_type_function>();
	
	builder assign_block( *where.insert_block() );
	
	value_ref func = eval_single_expr( *et.unbound, where );
	
	STATE_CHECK(func_type.context.is_defined());
	auto ci = assign_block.create_bind_func_inst( func, 
		types->get( func_type.context.extr() ) );
	assign_block.add( ci );

	auto const & q = func_defn.block->symbols->get_captures();
	unsigned index = 0;
	for( auto cit = q.begin(), end = q.end(); cit != end; ++cit, ++index ) {
		auto & local_decl = **cit;
		auto local_type = types->get( local_decl.type );
		
		auto sp = make_shared<struct_pointer_inst>();
		sp->struct_ptr = ci->ctx_data.result;
		STATE_CHECK( local_decl.xref );
		sp->index = local_decl.xref - 1;
		sp->result = where.make_temp( types->get_raw_pointer( local_type ) );
		assign_block.add( sp );
		
		auto & src_decl = *(*cit)->proximal_capture;
		STATE_CHECK(src_decl.cref);
		symbol_cref  & sc = *reinterpret_cast<symbol_cref*>(src_decl.cref);
		
		assign( sp->result, sc.ref, assign_block, assign_block, af_no_release ); //don't need distinct store_block
	}
	
	configure_destructor( ci->ctx_data, where );
	
	//release at end of construction block
	release( ci->result, where, block_flow::defer );
	return ci->result;	
}

value_ref compiler_leaf::eval_binary( leaf::expr_binary const & et, builder & where, unsigned /*flags*/ ) {
	auto a = eval_single_expr( *et.a, where );
	auto b = eval_single_expr( *et.b, where );
	
	binary_inst::op_t op = binary_inst::o_add; //keep GCC happy
	switch( et.op ) {
		#define MAP(t) case leaf::expr_binary::o_##t: op = binary_inst::o_##t; break;
		MAP(add)
		MAP(sub)
		MAP(mul)
		MAP(div)
		
		MAP(mod)
		
		MAP(bit_shl)
		MAP(bit_shr)
		
		MAP(or)
		MAP(and)
		MAP(xor)
		MAP(and_not)
		MAP(nor)
		MAP(nand)
		
		MAP(eq)
		MAP(neq)
		MAP(lt)
		MAP(lte)
		MAP(gt)
		MAP(gte)
		#undef MAP
	}
	
	return where.add_binary( op, a, b, types->get( et.type ) );
}

value_ref compiler_leaf::eval_unary( leaf::expr_unary const & et, builder & where, unsigned /*flags*/ ) {
	auto arg = eval_single_expr( *et.arg, where );
	
	unary_inst::op_t op = unary_inst::o_neg; //keep GCC happy;
	switch( et.op ) {
		#define MAP(t) case leaf::expr_unary::o_##t: op = unary_inst::o_##t; break;
		MAP(neg)
		MAP(abs)
		
		MAP(not)
		#undef MAP
	}
	
	return where.add_unary( op, arg, types->get( et.type ) );
}

value_ref compiler_leaf::eval_subscript( leaf::expr_subscript const & ex, builder & where, unsigned /*flags*/ ) {
	auto & arr = ex.base->type.extr().cast<leaf::intr_type_array>();
	
	 if( !arr.raw ) {
		return where.add_array_offset(
			eval_single_expr( *ex.base, where ),
			eval_single_expr( *ex.offset, where ) );
	} else {
		return where.add_array_ptr_offset(
			eval_single_expr( *ex.base, where ),
			eval_single_expr( *ex.offset, where ),
			ex.base->type.extr().is_lvalue() );
	}
}

value_ref compiler_leaf::eval_bind( leaf::expr_bind const & expr, builder & where, unsigned /*eval_flags*/ ) {

	if( expr.clone ) {
		return eval_clone_func( expr, where );
	}
	
	if( !expr.context ) {
		value_ref vr = eval_single_expr( *expr.unbound, where );
		auto wf = where.add_bind_raw_func_inst( vr );
		release( wf->result, where, block_flow::defer );
		return wf->result;
	}
	
	bool is_shared = false;
	value_ref base = eval_single_expr( *expr.context, where );
	if( expr.context->type.extr().is_shared() ) {
		is_shared = true;
		if( expr.context->type.extr().is_lvalue() ) {
			base = where.access_deref( base ); //get to shared
			acquire( base, where );
		} else {
			acquire( base, where );
		}
	} else if( expr.context->type.extr().is_value_ptr() ) {
		base = where.access_deref( base );
	} else if( !expr.context->type.extr().is_lvalue() ) {
		base = where.access_raw_pointer( base );
	}
	
	value_ref unbound = eval_single_expr( *expr.unbound, where );
	auto & func_type = expr.unbound->type.extr().cast<leaf::intr_type_function>();
	auto ex = where.add_bind_func_explicit_inst( unbound, base );
	ex->data_shared = is_shared;
	ex->ctx_shared = func_type.context.extr().is_shared();
	release( ex->result, where, block_flow::defer );
	return ex->result;
}

value_ref compiler_leaf::eval_field( leaf::expr_field const & expr, builder & where, unsigned /*eval_flags*/ ) {
	STATE_CHECK( expr.base->type.extr().is<leaf::intr_type_compound>() 
		|| expr.base->type.extr().is<leaf::intr_type_classlike>() );

	value_ref base = eval_single_expr( *expr.base, where );
	
	size_t index;
	if( expr.sref ) {
		//resolving the base and its type above should have fixated all fields
		STATE_CHECK( expr.sref->xref );
		//scope_ref's must be resolved to an index
		index = expr.sref->xref - 1;
	} else {
		index = *expr.field_index;
	}
	
	bool ptr = false;
	if( expr.base->type.extr().is_lvalue() ) {
		if( expr.base->type.extr().is_shared() ) {
			base = where.access_deref( base ); //drop lvalue on raw_ptr to get shared
			base = where.access_raw_shared( base ); //get pointer to shared value
			ptr = true;
		} else if( expr.base->type.extr().is_value_ptr() ) {
			base = where.access_deref( base ); 
		}
	} else {
		if( expr.base->type.extr().is_shared() ) {
			base = where.access_raw_shared( base );
			//base = where.access_deref_shared( base );
			ptr = true;
		} else if( expr.base->type.extr().is_value_ptr() ) {
		}
	}

	auto vr = where.access_post_field( base, index );
	auto q = where.resolve( vr );
	
	//Dereference if we end up with an unintentional pointer. I'm not clear if this shouldn't be handled
	//at the typing stage.
	//TESTCASE(share/unit_tests/block/type/class_shared_shared_2.lfb)
	if( !expr.type.extr().is_lvalue() && ptr ) {
		vr = where.access_deref(vr);
	}
	return vr;
}

value_ref compiler_leaf::eval_special_field( leaf::expr_special_field const & expr, builder & where, 
	unsigned eval_flags ) {
	value_ref base = eval_single_expr( *expr.base, where, eval_flags );
	
	switch( expr.field )
	{
		case leaf::expr_special_field::f_array_size: {
			auto sf = make_shared<special_field_inst>();
			sf->object = base;
			sf->field = special_field_inst::f_array_len;
			sf->result = where.make_temp( types->get( expr.type ) );
			where.add( sf );
			return sf->result;
		}
		case leaf::expr_special_field::f_optional_set: {
			return where.access_optional_set( base );
		}
		case leaf::expr_special_field::f_contextual_context: {
			return where.access_post_field( base, 0 );
		}
		case leaf::expr_special_field::f_contextual_unbound: {
			return where.access_post_field( base, 1 );
		}
	}
	
	STATE_FAIL( "invalid field" );
}

value_ref compiler_leaf::eval_current( leaf::expr_current const & expr, builder & where ) {

	switch( expr.what ) {
		case leaf::expr_current::w_context: {
			if (expr.depth == 0) {
				return *where.get_owner_function().local_context;
			} else if (expr.depth >= 1) {
				auto pcap = where.get_owner_function().context_arg;
				value_ref q = pcap;
				auto d = expr.depth;
				while (--d > 0){
					//TODO: this is still suspect here, why don't we need to deference the other types (which
					//are also structres/pointers?
					if (q.result_type()->is_shared_pointer() )
						q = where.access_deref_shared( q );
					q = where.access_post_field( q, 0 );
				}

				return q;
			}
			STATE_FAIL( "uh-oh" );
		}
		
		case leaf::expr_current::w_this: 
			STATE_FAIL( "unexpected-at-stage" );
	}
	
	STATE_FAIL( "unknown-what" );
}

value_ref compiler_leaf::eval_temp( leaf::expr_temp const & expr, builder & where ) {
	auto t = types->get(expr.type);
	
	if( expr.init == leaf::expr_temp::i_default ) {
		auto nv = gen_default( t, where );
		release( nv, where, block_flow::defer );
		return nv;
	}
	
	if( expr.init == leaf::expr_temp::i_intrinsic ) {
		auto intr = eval_single_expr( *expr.intrinsic_value, where );
		if( t->fundamental != type::f_shared_pointer) {
			return intr;
		}
	
		switch( t->fundamental ) {
			case type::f_shared_pointer: {
				auto nv = gen_shared( t, where );
				release( nv, where, block_flow::defer );
				assign( nv, intr, where, where, af_no_release | af_init_refer );
				return nv;
			}
			
			default:
				STATE_FAIL( "unsupported" );
		}
		
	}
	
	switch( t->fundamental ) {
		case type::f_shared_pointer: {
			auto nv = gen_shared( t, where );
			release( nv, where, block_flow::defer );
			return nv;
		}
		
		default: {
			auto au = make_shared<alloc_uninit_inst>();
			au->direct_type = types->get(expr.type);
			au->result = where.make_temp(au->direct_type);
			where.add(au);
			release( au->result, where, block_flow::defer, rf_cur_value );
			return au->result;
		}
	}
}

value_ref compiler_leaf::eval_tag( leaf::expr_tag const & expr, builder & where ) {
	STATE_CHECK( expr.value->form == leaf::expression::f_text );
	auto & text = static_cast<leaf::expr_text const &>(*expr.value).text;
	
	std::ostringstream out;
	STATE_CHECK( text.size() <= std::numeric_limits<uint16_t>::max() );
	uint16_t len = (uint16_t)text.size();
	out.write( (char const *)&len, sizeof(len) );
	out.write( (char const*)text.data(), len * sizeof(unicode::char_code) );
	auto res = out.str();
	
	auto data = gen_static_binary_data( (uint8_t const*)res.data(), res.size(), where );
	
	return where.add_cast( where.access_raw_pointer(data), types->get_binary(platform::pointer_size) );
}

value_ref compiler_leaf::build_function( leaf::expr_funcdefn const & expr, builder & where ) {
	//generate a suitable name
	static unsigned count = 0;
	char buf[128];
	std::string use_name = where.get_owner_function().defn_value->name;
	if( use_name == "_entry" ) //kind of cheap, but good enough for debugging
		use_name = "";
		
	if( expr.bound_decl ) {
		sprintf( buf, "_%d", ++count );
		use_name += '_';
		use_name += expr.bound_decl->symbol;
		use_name += buf;
	} else {
		sprintf( buf, "_anon_%d", ++count );
		use_name += buf;
	}
	auto & func_type = expr.type.extr().cast<leaf::intr_type_function>();
	
	//DEBUG( "BUILD_FUNCTION", use_name, leaf::dump::get_intr_type( func_type ) );
	
	//the symbol_cref may already be defined, thus we must use its value here
	auto & result = get_symbol_cref( expr.cref );
	//it may already have been built (literal references)
	if( result.built ) {
		return result.ref;
	}
	//set built now in case of recursion
	result.built = true;
	
	auto fbb = where.create_function( use_name, types->get(expr.type), result.origin );
	builder fbody( fbb );
	
	//track current scope function for later reference
	//STATE_CHECK( expr.block->symbols->is_function_root(), use_name );
	STATE_CHECK( !expr.block->symbols->fref );
	expr.block->symbols->fref = reinterpret_cast<intptr_t>( fbb.owner_function );
	
	fbody.add_comment("define args and returns" );
	auto const & srs = expr.block->symbols->get_in_order();
	for( size_t i=0; i < srs.size(); ++i )
	{
		leaf::declaration const & decl = *srs[i];
		if( decl.form == leaf::declaration::f_arg ) {
			auto p = fbb.define_arg( types->get( decl.type ), decl.symbol );
			define_symbol_cref( decl.cref, p );
			
		} else if( decl.form == leaf::declaration::f_return ) {
			auto r = fbb.define_return( types->get( decl.type ), decl.symbol );
			define_symbol_cref( decl.cref, r );
			//an init statement is expected for returns (it is not done automatically here)
		}
	}

	fbody.add_comment( "bind captures" );
	if( func_type.access != leaf::intr_type_function::a_raw ) {
		STATE_CHECK( func_type.is_unbound() );
		STATE_CHECK( func_type.context.is_defined() );
		
		fbb.owner_function->context_arg = fbb.define_value( value::f_arg_ref /*TODO:???*/, 
			types->get(func_type.context), "context" );
		
		auto const & q = expr.block->symbols->get_captures();
		for( auto cit = q.begin(), end = q.end(); cit != end; ++cit ) {
			auto & local_decl = **cit;
			auto src_decl = (*cit)->proximal_capture;
			auto local_type = types->get( local_decl.type );
			//DEBUG( "CAPTURE", &local_decl.cref, local_decl.symbol, local_decl.get_capture_level(), local_decl.xref, 
			//	src_decl->symbol, src_decl->get_capture_level(), src_decl->xref );
			auto v = fbb.define_context_arg( local_type, local_decl.symbol );
			define_symbol_cref( local_decl.cref, v );			

			//typing of the function will resolve the "xref" of the type. If there local decl has an xref
			//we use that (happens for distinct closure context) otherwise the source must have
			//an xref (happens for class members)
			if( local_decl.xref ) {
				v.origin->capture.index = local_decl.xref - 1;
				v.origin->capture.parent = 0;
			} else {
				unsigned level = src_decl->get_capture_level();
				v.origin->capture.parent = level;
				
				//NOTE: This loop is just for debugging
				/*while( level > 0 ) {
					src_decl = src_decl->proximal_capture;
					STATE_CHECK( src_decl );
					//DEBUG( "UNCAPTURE", src_decl->symbol, src_decl->get_capture_level(), src_decl->xref );
					level--;
				}*/
				if( level > 0 ) {
					src_decl = src_decl->ultimate_capture;
				}
				STATE_CHECK( src_decl->xref );
				v.origin->capture.index = src_decl->xref -1;
			}
		}
	}
		
	build_capture_shared( expr, fbody );
	
	build_imperative_block( *expr.block, fbody );

	return result.ref;
}

/**
	Creates the binding between all those variables referenced in this scope that are
	also captured. This binds the local reference to the correct location.
*/
void compiler_leaf::build_capture_shared( leaf::expr_funcdefn const & expr, builder & where ) {
	//if there isn't a type already it means we don't need to do this
	auto shared_ty = expr.block->symbols->get_context_type(false);
	if( !shared_ty.is_defined() ) {
		return;
	}
	where.add_comment( "capture shared context" );
	auto & scope_ty = shared_ty.extr().intr->cast<leaf::intr_type_custom>();
	
	std::string use_name = where.get_owner_function().defn_value->name + "_shares";
	auto local_ty = types->get( shared_ty );
	auto local_value = where.define_local( local_ty, use_name );
	where.get_owner_function().local_context = local_value;
	auto local_ptr = initialize( value_ref::as_raw_pointer(local_value), local_ty, 0, where ); 
	release_pointer( value_ref::as_raw_pointer(local_value), where, block_flow::defer );
	
	where.add_comment( "prepare" );
	
	//link to parent context (shared pointer placed at index 0)
	if( scope_ty.parent.is_defined() ) {
		auto dst = where.access_post_field(local_ptr, 0);
		assign(dst, where.get_owner_function().context_arg, where, where);
	}
	DEBUG("scopy_ty", expr.block->symbols.get(), dump::get_type( local_ty.get(), true ) );
	
	for( auto & share : scope_ty.get_type_decls() ) {
 		auto & sref = get_symbol_cref( share->cref );
 		STATE_CHECK( share->xref );
 		size_t index = share->xref-1;
 		
 		auto vr = obtain_symbol_value_ref( share );
 		STATE_CHECK( vr.is_direct() );

 		//the context exists beyond
 		sref.no_release = true;
 		
 		auto direct_refer = [&](){
			//refer directly to shared version
			sref.origin = local_value.origin;
			sref.ref = where.access_deref( where.access_post_field( local_ptr, index ) );
 		};
 		
 		DEBUG( "CAPTURE_VAR", share->symbol, sref.origin->form );
 		if( sref.origin->form == value::f_local ) {
			direct_refer();
		} else if( sref.origin->form == value::f_arg_ref ) {
			//copy argument into the shared data
			assign( where.access_post_field( local_ptr, index ), sref.ref, where, where );
			direct_refer();
		} else if( sref.origin->form == value::f_return_ref ) {
			//no need to copy since an init is expected
			//direct_refer();
			//FEATURE: it's undecided if this should be supported, the above line is not enough, the function
			//actually needs to return this new value (not easy to do)
			STATE_FAIL( "return-not-supported-in-capture" );
 		} else {
			STATE_FAIL( "unsupported-capture-shared" );
 		}
	}
	
	where.add_comment( "end capture shared context" );
}