void compiler_leaf::build_statement( leaf::statement const & st, builder & where ) {
	switch( st.statement_form() )
	{
		#define BUILDS(type) \
			case leaf::sf_##type: \
				/*DEBUG("Statement: ", #type);*/ \
				build_##type( static_cast<leaf::type const&>(st), where ); \
				break;
		BUILDS(assign_statement)
		BUILDS(defer_statement)
		BUILDS(do_statement)
		BUILDS(expression_statement)
		BUILDS(fail_statement)
		BUILDS(init_statement)
		BUILDS(loop_flow_statement)
		BUILDS(return_statement)
		BUILDS(error_statement)
		#undef BUILDS
		
		case leaf::sf_var_statement:
			//TODO: hmm, adding symbol scanner leaves some around (likely pushed as incomplete statements)
			//STATE_FAIL("unexpected"); //shouldn't be any left
			break;
		case leaf::sf_noop_statement:
		case leaf::sf_import_statement:
		case leaf::sf_for_statement:
		case leaf::sf_typedef_statement:
			break;
	}
}

void compiler_leaf::build_do_statement( leaf::do_statement const & ds, builder & where ) {
	if( !ds.expr_list.size() )
		return;

	//a local must be defined otherwise the condition value would elevate out of the containing
	//block. This is problematic where there is a defere, since two blocks merge and only one
	//has set the actual value. Refer to the test <unit_tests/block/do/then_defer.lfb>

	if( ds.loop ) {
		STATE_CHECK( ds.expr_list.size() == 1, "unsupported-looping-construct" );
		auto loop_inst = where.add_loop();
		
		auto & item = ds.expr_list[0];
		if( item.cond ) {
			builder cond_block( *loop_inst->cond_block );
			auto result = where.define_local( types->get_boolean(), "cond_result" );
			auto lres = eval_single_expr( *item.cond, cond_block );
			assign( value_ref::as_raw_pointer( result ), lres, cond_block, cond_block );
			loop_inst->cond_value = result;
		} else {
			loop_inst->cond_value = where.const_boolean( true );
		}
		
		builder body_block( *loop_inst->body_block );
		build_imperative_block( *item.block, body_block );
	} else {

		auto result = where.define_local( types->get_boolean(), "cond_result" );
		auto sb = where.create_switch();
		for( size_t i=0; i < ds.expr_list.size(); ++i ) {
			auto & item = ds.expr_list[i];
			sb.add_case();
			if( item.cond ) {
				auto lres = eval_single_expr( *item.cond, sb.get_cond_block() );
				assign( value_ref::as_raw_pointer( result ), lres, sb.get_cond_block(), sb.get_cond_block() );
				sb.set_cond_value( result );
			}
				
			build_imperative_block( *item.block, sb.get_expr_block() );
		}
		sb.add();
	}
}

void compiler_leaf::build_error_statement( leaf::error_statement const & es, builder & where ) {

	auto db = where.insert_block();
	db->error_block = db->create_child_block();
	db->error_block->error_level = int(es.error.on_match);

	builder eb( *db->error_block );
	db->error_block->error_info = eb.define_value(  
		value::f_temp, types->get(es.error.error_info->type), es.error.error_info->symbol );
	auto leaf_error_info = eb.add_cast( *db->error_block->error_info, types->get(es.error.error_info->type) );
	define_symbol_cref( es.error.error_info->cref, leaf_error_info );
	
	build_imperative_block( *es.error.on_error, eb );

	auto body_where = builder( *db );

	build_statement( *es.inner, body_where );
}

void compiler_leaf::build_defer_statement( leaf::defer_statement const & ds, builder & where ) {
	auto db = where.create_block();
	builder bb( *db );
	if( ds.error_only ) {
		//TODO: copiedf rom build_error_statement
		db->error_info = bb.define_value(  
			value::f_temp, types->get(ds.error.error_info->type), ds.error.error_info->symbol );
		auto leaf_error_info = bb.add_cast( *db->error_info, types->get(ds.error.error_info->type) );
		define_symbol_cref( ds.error.error_info->cref, leaf_error_info );
	}
	
	build_imperative_block( *ds.error.on_error, bb );
	if( ds.error_only ) {
		where.defer_fail( db );
	} else {
		where.defer( db );
	}
}

void compiler_leaf::build_return_statement( leaf::return_statement const & rs, builder & where ) {
	builder bb(where);
	if( rs.value ) {
		builder assign_block( *where.insert_block() );
		
		value_ref_vector_t out;
		auto & returns = where.get_owner_function().returns;
		for( size_t i=0; i < returns.size(); ++i )
			out.push_back( value_ref::as_raw_pointer( returns[i] ) );
			
		auto right = eval_list_expr( *rs.value, assign_block );
		assign( out, right, assign_block );
	}
	
	auto ri = make_shared<return_inst>();
	bb.add( ri );
}

void compiler_leaf::build_fail_statement( leaf::fail_statement const & fs, builder & where ) {
	if( fs.level == leaf::fail_level::resume ) {
		auto err_info = where.where->find_error_info();
		STATE_CHECK( !!err_info && err_info->is_valid() );
		DEBUG("resume", err_info->origin.get() );
		auto ri = make_shared<resume_inst>();
		ri->data = err_info;
		where.add( ri );
		
	} else {
		auto fi = make_shared<fail_inst>();
		fi->level = int(fs.level);
		
		if( fs.tag ) {
			auto tag = eval_single_expr( *fs.tag, where );
			
			//TODO: this lookup isn't good, we don't have a reference to the high-level "type_manager" from 
			//the leaf namespace though
			auto err_type = globals->if_get_symbol( "⁑error_info" );
			STATE_CHECK( !!err_type && err_type->is_fun_type() );
			auto serr = leaf::extr_type(err_type->fun_type);
			serr.set_ref_shared();
			
			//TODO: this is kind of suspicious that this is done here. Perhaps the error_info should be
			//created at the higher level instead...
			auto shared_err = gen_default( types->get(serr), where );
			auto err = where.access_raw_shared( shared_err );
			auto dst_tag = where.access_post_field( err, 1); //...If not then this needs to be looked up
			assign( dst_tag, tag, where, where );
			fi->data = shared_err;
		}
		where.add( fi );
	}
}

void compiler_leaf::build_loop_flow_statement( leaf::loop_flow_statement const & ls, builder & where ) {
	auto lfi = make_shared<loop_flow_inst>();
	switch( ls.flow ) {
		case leaf::loop_flow_statement::f_break: 
			lfi->flow = loop_flow_inst::f_break; 
			break;
		case leaf::loop_flow_statement::f_next: 
			lfi->flow = loop_flow_inst::f_next; 
			break;
	}
	where.add(lfi);
}

void compiler_leaf::build_expression_statement( leaf::expression_statement const & es, builder & where ) { 
	eval_list_expr( *es.expr, where ); 
}

void compiler_leaf::build_block_impl( leaf::statement_block const & bk, builder & where ) {
	for( size_t i=0; i < bk.list.size(); ++i )
		build_statement( *bk.list[i], where );
}

void compiler_leaf::build_declarative_block( leaf::statement_block const & bk, builder & where ) {
	STATE_CHECK( bk.symbols->is_declarative() );
	build_block_impl( bk, where );
}

void compiler_leaf::build_imperative_block( leaf::statement_block const & bk, builder & where ) {
	STATE_CHECK( !bk.symbols->is_declarative() );
	build_block_impl( bk, where );
}

void compiler_leaf::build_assign_statement( leaf::assign_statement const& as, builder & where )
{
	//context of implicit refer/relese is this statement itself.
	builder assign_block( *where.insert_block() );
	
	bool left_refer = as.left->type.extr().is_refer();
	bool right_refer = as.right->type.extr().is_refer();
	
	unsigned flags = 0;
	
	auto left = eval_list_expr( *as.left, assign_block ); 
	
	switch( as.how ) {
		case leaf::assign_statement::h_value:
			STATE_CHECK( !left_refer || right_refer );
			break;
		case leaf::assign_statement::h_init_refer:
			//STATE_CHECK( left_shared ); //could also be a value-list
			flags |= af_init_refer;
			break;
		case leaf::assign_statement::h_default: {
			STATE_CHECK( left.size() == 1 ); //TODO: shoudl be safe to just default each part
			
			auto sb = where.create_switch();
			//TODO: a "not" expression is really needed now...
			sb.add_case();
			sb.set_cond_value( sb.get_cond_block().access_optional_set( 
				sb.get_cond_block().access_deref( left[0] ) ) );
			
			sb.add_else();
			//further evaluation happens inside this conditional block
			assign_block = sb.get_expr_block();
			
			sb.add();
			break;
		}
	}
	
	auto right = eval_list_expr( *as.right, assign_block );
	assign( left, right, assign_block, flags );
}

void compiler_leaf::build_init_statement( leaf::init_statement const & vs, builder & outer_where )
{
	auto & decl = *vs.sref;
	if( decl.import_type != leaf::declaration::it_none )
		return;

	builder where( *outer_where.insert_block() );
	
	value_ref vd = obtain_symbol_value_ref( vs.sref );
	DEBUG( vs.sref->symbol, vs.sref->cref, vs.sref.get(), vd.origin.get(), vd.origin->cref );
	if( vd.is_direct() ) {
		if( vd.origin->form == value::f_global ) {
			if( decl.is_export ) {
				vd.origin->is_export = true;
			}
			cur_module->register_global( vd.origin );
		} else if( vd.origin->form == value::f_local ) {
			where.get_owner_function().locals.push_back( vd.origin );
		}
	}
	
	//if already in the shared context it means it's been default initialized already
	//TODO: name no_release is not very good
	bool in_shared_context = get_symbol_cref( vs.sref->cref ).no_release;
	if( in_shared_context ) {
		release( vd, where, block_flow::sequential, 0 );
	}
	
	value_ref nv = where.access_raw_pointer(vd);

	unsigned flags = 0;
	if( vs.init_refer_value )
		flags |= if_init_refer_value;
	if( vs.undefined )
		flags |= if_undefined_value;
	where.add_comment( "init" );
	initialize_leaf( nv, vs.sref->type.extr(),  vs.value, flags, where );
	where.add_comment( "init-2" );

	//initialization comes before release, this is why there is a "code" block
	if( vs.code ) {
		build_statement( *vs.code, where );
	}
	
	//shared context variables are not released
	if( !in_shared_context ) {
		//by default release at end of block, if "no_release" specified it means only normal flow doesn't
		//need release but error flow still does
		//TODO: this probably needs to happen prior to the value assignment,
		//such that for shared pointers the item can be released if there is an error in the assignment
		release( vd, outer_where, 
			vs.no_release ? block_flow::defer_fail : block_flow::defer, 
			rf_cur_value );
	}
}

