#include "ir/dump.hpp"
#include "ir/language.hpp"
#include "util/validation_context.hpp"

#include <boost/algorithm/string.hpp>

#include <iostream>

namespace ir
{

dump::dump( std::ostream & out )
	: dump_base( out )
{
}

#define CHECK_NULL( v, msg ) if( !v ) { error( "null "  msg ); return; }
#define CHECK_SIZE( v, msg ) if( !v.size() ) { error( "empty " msg ); return; }
#define CHECK_SIZE_MIN( v, i, msg ) if( v.size() < i ) { error( "invalid-size " msg ); return; }
/**
	Calls the object validation and emits the error. This doesn't throw since the dump should
	then continue as best it can to write out the code. Doing so greatly aids in debugging (to
	see exactly where the problem is).
*/
template<typename T>
void dump::validate( T const & o ) {
	try {
		util::validate(o);
	} catch( recoverable_error & re ) {
		std::string err = boost::current_exception_diagnostic_information();
		boost::trim( err );
		boost::replace_all( err, "\n", std::string("\n") + indent_str() + "// " );
		out << "\e[31m";
		out << indent_str() << "// " << err << std::endl;
		out << "\e[0m";
	}
}

void dump::write( module const * m )
{
	CHECK_NULL( m, "module" )
	
	out << "/* ** ** ** **" << std::endl
		<< "\tMODULE: " << m->name << std::endl
		<< "*/" << std::endl
		<< std::endl;

	for( auto it = m->global_variable_map.begin(), end = m->global_variable_map.end(); it != end; ++it )
		write_global_variable( it->second.get() );
		
	for( auto it = m->function_map.begin(), end = m->function_map.end(); it != end; ++it )
		write( it->second.get() );
		
	for( auto & it : named_types ) {
		write( it, true );
		out << endl();
	}
}

void dump::write_block_seq_item( block_seq_item const & si )
{
	boost::optional<std::string> flow;
	switch( si.flow ) {
		case block_flow::sequential: break;
		case block_flow::defer: flow = "defer"; break;
		case block_flow::defer_fail: flow = "defer_fail"; break;
	}
	if( flow ) {
		out << indent_str() << *flow << " {" << endl();
		indent();
	}
	
	if( si.an_instruction )
		write_instruction( si.an_instruction.get() );
		
	if( si.a_block ) {
		if( !flow )
			indent( "    ", "{   " );
		write_block( si.a_block.get() );
		if( !flow )
			unindent( "}" );
	}
	
	if( flow ) {
		unindent("}");
	}
}

void dump::write_block( block const * b )
{
	CHECK_NULL( b, "block" )
	
	if( b->error_block )
		indent( "    ", "try " );
		
	for( size_t i = 0; i < b->items.size(); ++i )
		write_block_seq_item( b->items[i] );
		
	if( b->error_block ){
		unindent();
		indent( "    ", "on  " );
		out << indent_str() << "<= " << b->error_level << " " << endl();
		write_block( b->error_block.get() );
		unindent();
	}
}

void dump::write( function const * f )
{
	CHECK_NULL( f, "function" )
	validate( *f );
	
	//we show signature so don't show type later again
	shown_type.insert( f->defn_value.get() );
	for( size_t i=0; i < f->args.size(); ++i )
		shown_type.insert( f->args[i].origin.get() );
	for( size_t i=0; i < f->returns.size(); ++i )
		shown_type.insert( f->returns[i].origin.get() );
		
	out << indent_str() << "function " << f->defn_value->name;
	
	//write out args/returns to include names
	//FEATURE: ideally the names/types would be shown together (harder since the types now have their names stripped)
	out << "(";
	for( size_t i=0; i < f->args.size(); ++i ) {
		if( i )
			out << ",";
		out << f->args[i].origin->name;
	}
	out << ")->(";
	for( size_t i=0; i < f->returns.size(); ++i ) {
		if( i )
			out << ",";
		out << f->returns[i].origin->name;
	}
	out << ") : ";
	//also type to ensure match/show access
	write( f->defn_value->direct_type.get() );
	out << std::endl;
	out << indent_str() << "{" << std::endl;

	indent();
	for( size_t i=0; i < f->locals.size(); ++i )
		write_local_variable( 	f->locals[i].get() );
	write_block( f->entry_block.get() );
	unindent();
	
	out << indent_str() << "}" << std::endl
		<< std::endl;
}

template<typename T>
void dump::write_value_list( std::vector<T> const & values)
{
	out << "(";
	for( size_t i=0; i < values.size(); ++i )
	{
		if( i )
			out << ",";
		write( values[i] );					
	}
	out << ")";
}

template<typename T>
void dump::write_tuple( std::vector<T> const & values )
{
	out << "[";
	for( size_t i=0; i < values.size(); ++i )
	{
		if( i )
			out << ",";
		write( values[i] );					
	}
	out << "]";
}

std::string dump::get_type( type const * t, bool expand ) {
	dump d;
	d.write( t, expand );
	return d.str();
}

std::string dump::get_instruction( instruction const * i ) {
	dump d;
	d.write_instruction( i );
	return d.str();
}

std::string dump::get_value_ref( value_ref vr ) {
	dump d;
	d.write_value_ref( vr );
	return d.str();
}

void dump::write( type const * t, bool force_expand )
{
	CHECK_NULL( t, "type" )
	
	if( t->name.size() ) {
		named_types.insert( t );
		out << t->name;
		if( !force_expand ) {
			return;
		}
		out << " : ";
	}
		
	switch( t->fundamental )
	{
		case type::f_function_ptr: {
			switch( t->access ) {
				case type::a_bound: out << "bound"; break;
				case type::a_raw: out << "raw"; break;
				case type::a_unbound: out << "unbound"; break;
				case type::a_unbound_shared: out << "unbound_shared"; break;
			}
			out << "_func<";
			if( t->context ) {
				write( t->context.get(), force_expand );
				out << ", ";
			}
			write_value_list( t->args );
			out << " -> ";
			write_value_list( t->sub );
			
			if( !t->can_fail ) {
				out << " no_throw";
			}
			out << ">";
			break;
		}
		case type::f_integer: {
			out << "integer_" << t->trivial_bitsize;
			break;
		}
		case type::f_binary: {
			out << "binary_" << t->trivial_bitsize;
			break;
		}		
		case type::f_float: {
			out << "float_" << t->trivial_bitsize;
			break;
		}
		case type::f_boolean: {
			out << "boolean";
			break;
		}
		case type::f_raw_pointer: {
			CHECK_SIZE_MIN( t->sub, 1, "raw_ptr" );
			out << "raw_ptr<";
			write( t->sub[0].get(), force_expand );
			out << ">";
			break;
		}
		case type::f_shared_pointer: {
			CHECK_SIZE_MIN( t->sub, 1, "shared_ptr" );
			out << "shared_ptr<";
			write( t->sub[0].get(), force_expand );
			out << ">";
			break;
		}
		case type::f_optional: {
			CHECK_SIZE_MIN( t->sub, 2, "optional" );
			out << "{optional,";
			write( t->sub[1].get(), force_expand );
			out << "}";
			break;
		}			
		case type::f_tuple: {
			write_tuple( t->sub );
			switch( t->pack ) {
				case type::p_normal: break;
				case type::p_abi: out << " abi_pack"; break;
				case type::p_byte: out << " byte_pack"; break;
			}
			break;
		}
		case type::f_array: {
			CHECK_SIZE_MIN( t->sub, 1, "array" );
			out << "array<";
			write( t->sub[0].get(), force_expand );
			out << ">";
			break;
		}
		case type::f_abi_var_args: {
			out << "abi_var_args";
			break;
		}
	}
}

void dump::write_instruction( instruction const * i )
{
	CHECK_NULL( i, "instruction" )
	validate( *i );
	
	switch( i->type )
	{
		#define IR_INST(type) \
			case instruction::t_##type: \
				write_##type##_inst( static_cast<type##_inst const&>(*i) ); \
				return;
			IR_INSTRUCTIONS
		#undef IR_INST
	}
	
	STATE_CHECK( false );
}

void dump::write_store_inst( store_inst const & ai )
{
	out << indent_str();
	write_value_ref( ai.left );
	out << " => ";
	write_value_ref( ai.right );
	out << std::endl;
}

void dump::write_value_defn( value_defn vd, bool suppress_type )
{
	write_value( vd.origin.get(), suppress_type );
}

void dump::write_value_ref( value_ref vr, bool suppress_type )
{
	bool need_end = true;
	if( vr.how == value_ref::h_raw_pointer )
		out << "raw_ptr(";
	else if( vr.how == value_ref::h_deref )
		out << "deref(";
	else if( vr.how == value_ref::h_deref_shared )
		out << "deref_share(";
	else if( vr.how == value_ref::h_raw_shared )
		out << "raw_share(";
	else if( vr.how == value_ref::h_raw_array )
		out << "raw_array(";
	else
		need_end = false;
		
	write_value_name_part( vr.origin.get(), suppress_type );
	
	if( vr.pre_field )
		out << "#" << *vr.pre_field;
	
	write_value_type_part( vr.origin.get(), suppress_type );
	
	if ( need_end )
		out << ")";
}

void dump::write_value_name_part( value const * v, bool /*suppress_type*/ )
{
	CHECK_NULL( v, "value" )
		
	if( v->need_pointer)
		out << "(&";
		
	switch( v->form )
	{
		case value::f_const_integer:
			out << v->int_value;
			break;
		case value::f_const_binary:
			out << v->uint_value;
			break;
		case value::f_const_float:
			out << v->float_value;
			break;
		case value::f_const_boolean:
			out << (v->boolean_value ? "true" : "false");
			break;
		case value::f_nullptr:
			out << "nullptr";
			break;
			
		case value::f_return_ref:
			out << "returns";
			if( v->name.size() ) {
				out << "." << v->name;
			} else {
				out << "#" << v->anon_ref;
			}
			break;
			
		case value::f_arg_ref:
			CHECK_SIZE( v->name, "args." )
			out << "args." << v->name;
			break;
			
		case value::f_capture_ref:
			out << "captures";
			if( v->capture.parent ) {
				out << "^" << v->capture.parent;
			}
			if( v->name.size() ) {
				out << "." << v->name;
			} else {
				out << "#" << v->capture.index;
			}
			break;
			
		case value::f_import_code:
			CHECK_SIZE( v->name, "import_code." )
			out << "import_code." << v->name;
			break;
			
		case value::f_import_data:
			CHECK_SIZE( v->name, "import_data." )
			out << "import_data." << v->name;
			break;
			
		case value::f_temp:
			CHECK_SIZE( v->name, "temp " )
			out << v->name;
			break;
			
		case value::f_global:
			CHECK_SIZE( v->name, "global " )
			out << v->name;
			break;
			
		case value::f_local:
			CHECK_SIZE( v->name, "local " )
			out << v->name;
			break;
			
		case value::f_undefined:
			out << "undefined";
			break;
			
		case value::f_raw_binary:
			out << "[";
			for( size_t i=0; i < v->raw_binary.size(); ++i) {
				if( i ) {
					out << ",";
				}
				char buf[3];
				sprintf(buf, "%x", v->raw_binary[i] );
				out << buf;
			}
			out << "]";
			break;
	}
	
	if( v->need_pointer)
		out << ")";
}

void dump::write_value_type_part( value const * v, bool suppress_type )
{
	CHECK_NULL( v, "value" );
	//show the type of each value only once
	if( show_all_types || (!suppress_type && shown_type.find( v ) == shown_type.end()) )
	{
		out << " : ";
		write( v->direct_type.get() );
		shown_type.insert( v );
	}
}

void dump::write_funccall_inst( funccall_inst const & fc )
{
	out << indent_str();
	write_value_list( fc.returns );
	out << " = call ";
	if( fc.context ) {
		out << "[";
		write_value_ref( *fc.context, true );
		out << "] ";
	}
	write_value_ref( fc.call, true );
	write_value_list( fc.args );
	out << std::endl;
}

void dump::write_pack_tuple_inst( pack_tuple_inst const & pt )
{
	out << indent_str();
	write_value_defn( pt.result );
	out << " = pack ";
	write_tuple( pt.sub );
	out << std::endl;
}

void dump::write_switch_inst( switch_inst const & si )
{
	out << indent_str();
	if( si.with_result )
	{
		write_value_defn( si.result );
		out << " = switch" << std::endl;
	}
	else
	{
		out << "do" << std::endl;
	}
	indent();
	
	for( size_t i=0; i < si.blocks.size(); ++i ) 
	{
		auto & cb = *si.blocks[i];
		
		if( cb.cond_value ) {
			out << indent_str() << "when ";
			write_value_ref( *cb.cond_value );
			out << std::endl;
		
			indent();
			write_block( cb.cond_block.get() );
			unindent();
		
			out << indent_str() << "then ";
		} else {
			out << indent_str() << "else ";
		}
		
		if( si.with_result )
			write_value_ref( cb.expr_value );
		out << std::endl;
		
		indent();
		write_block( cb.expr_block.get() );
		unindent();
	}
	
	unindent();
}

void dump::write_resolve_inst( resolve_inst const & ei )
{
	out << indent_str();
	write_value_defn( ei.result );
	out << " = resolve ";
	write_value_ref( ei.source );
	out << std::endl;
}

void dump::write_struct_pointer_inst( struct_pointer_inst const & sp )
{
	out << indent_str();
	write_value_defn( sp.result );
	out << " = ";
	write_value_ref( sp.struct_ptr );
	out << " + " << sp.index;
	out << std::endl;
}

void dump::write_global_variable( value const * gv )
{
	CHECK_NULL( gv, "global_variable" );
	out << indent_str();
	out << "global ";
	write_value( gv );
	out << std::endl;
}

void dump::write_local_variable( value const * lv )
{
	CHECK_NULL( lv, "local_variable" );
	out << indent_str();
	out << "local ";
	write_value( lv );
	out << std::endl;
}

void dump::write_malloc_inst( malloc_inst const & mi ) {
	out << indent_str();
	write_value_defn( mi.result );
	out << " = malloc ";
	write( mi.direct_type );
	out << std::endl;
}

void dump::write_alloc_uninit_inst( alloc_uninit_inst const & au ) {
	out << indent_str();
	write_value_defn( au.result );
	out << " = uninit ";
	write( au.direct_type );
	out << std::endl;
}

void dump::write_alloc_shared_inst( alloc_shared_inst const & as ) {
	out << indent_str();
	write_shared_object_defn( as.sod, "alloc_shared" );
	out << std::endl;
}

void dump::write_shared_object_defn( shared_object_defn const & sod, char const * op ) {
	write_value_defn( sod.result );
	out << " = " << op << " ";
	write( sod.direct_type );
	if( sod.static_destructor ) {
		out << " nodtor";
	} else {
		out << " dtor:";
		write_value_ref( sod.destructor );
	}
}

void dump::write_bind_raw_func_inst( bind_raw_func_inst const & wi )
{
	out << indent_str();
	write_value_defn( wi.result );
	out << " = bind_raw_func ";
	write( wi.raw_func );
	out << std::endl;
}

void dump::write_bind_func_explicit_inst( bind_func_explicit_inst const & ci )
{
	out << indent_str();
	write_value_defn( ci.result );
	out << " = bind_func_explicit ";
	write_value_ref( ci.ctx_func );
	out << ", ";
	write_value_ref( ci.ctx_data );
	if( ci.data_shared ) {
		out << " data_shared";
	}
	if( ci.ctx_shared ) {
		out << " ctx_shared";
	}
	out << std::endl;
}

void dump::write_bind_func_inst( bind_func_inst const & ci )
{
	out << indent_str();
	write_value_defn( ci.result );
	out << " = bind_func ";
	write_value_ref( ci.ctx_func );
	out << std::endl;
	
	out << indent_str() << "\xe2\x86\xAA ";
	write_shared_object_defn( ci.ctx_data, "ctx_data" );
	out << std::endl;
}

void dump::write_void_context_inst( void_context_inst const & ci )
{
	out << indent_str();
	write_value_defn( ci.result );
	out << " = void_context ";
	out << std::endl;
}

void dump::write_release_shared_inst( release_shared_inst const & rs )
{
	out << indent_str();
	out << "release ";
	write( rs.object );
	out << " ";
	switch( rs.destroy_how )
	{
		case release_shared_inst::dh_trivial:
			out << "trivial";
			break;
			
		case release_shared_inst::dh_dynamic:
			out << "dynamic";
			break;
	}
	out << std::endl;
}

void dump::write_acquire_shared_inst( acquire_shared_inst const & as )
{
	out << indent_str();
	out << "acquire ";
	write( as.object );
	out << std::endl;
}

void dump::write_return_inst( return_inst const & /*ri*/ )
{
	out << indent_str();
	out << "return";
	out << endl();
}

void dump::write_alloc_array_inst( alloc_array_inst const & aa )
{
	out << indent_str();
	write_value_defn( aa.result );
	out << " = alloc_array ";
	write( aa.item_type.get() );
	out << " ";
	write_value_ref( aa.length );
	out << endl();
}

void dump::write_free_array_inst( free_array_inst const & fa )
{
	out << indent_str();
	out << "free_array ";
	write_value_ref( fa.array );
	out << endl();
}

void dump::write_special_field_inst( special_field_inst const & sf )
{
	out << indent_str();
	write_value_defn( sf.result );
	out << " = special.";
	switch( sf.field )
	{
		case special_field_inst::f_array_len:
			out << "array_len";
			break;
	}
	out << " ";
	write_value_ref( sf.object );
	out << endl();
}

void dump::write_array_offset_inst( array_offset_inst const & ao )
{
	out << indent_str();
	write_value_defn( ao.result );
	out << " = array_offset ";
	write_value_ref( ao.array );
	out << " ";
	write_value_ref( ao.offset );
	out << endl();
}

void dump::write_array_ptr_offset_inst( array_ptr_offset_inst const & ao )
{
	out << indent_str();
	write_value_defn( ao.result );
	out << " = array_ptr_offset ";
	write_value_ref( ao.array );
	out << " ";
	write_value_ref( ao.offset );
	out << endl();
}

void dump::write_loop_inst( loop_inst const & lo )
{
	out << indent_str() << "loop" << endl();
	indent("    ", "?    " );
	write_block( lo.cond_block.get() );
	unindent();
	out << indent_str() << "while ";
	write_value_ref( lo.cond_value );
	out << endl();
	indent();
	write_block( lo.body_block.get() );
	unindent();
}

void dump::write_promote_inst( promote_inst const & pi )
{
	write_promote_loss_inst( pi, "promote" );
}

void dump::write_loss_inst( loss_inst const & pi )
{
	write_promote_loss_inst( pi, "loss" );
}

void dump::write_promote_loss_inst( promote_loss_inst const & inst, char const * label )
{
	out << indent_str();
	write_value_defn( inst.result );
	out << " = " << label << " ";
	write_value_ref( inst.source );
	out << " ";
	auto what = [&]( promote_loss_inst::what_t w ) {
		switch( w ) {
			case promote_loss_inst::w_integer: out << "integer"; break;
			case promote_loss_inst::w_float: out << "float"; break;
			case promote_loss_inst::w_binary: out << "binary"; break;
		}
	};
	what( inst.from_what );
	out << "->";
	what( inst.to_what );
	out << " " << inst.to_bitsize << "bits";
	out << endl();
}

void dump::write_resume_inst( resume_inst const & /*ri*/ )
{
	out << indent_str();
	out << "resume_fail";
	out << endl();
}

void dump::write_fail_inst( fail_inst const & fi )
{
	out << indent_str();
	out << "fail(" << fi.level;
	if( fi.data ) {
		out << ", ";
		write_value_ref( *fi.data );
	}
	out << ")";
	out << endl();
}

void dump::write_loop_flow_inst( loop_flow_inst const & lfi )
{
	out << indent_str();
	switch( lfi.flow ) {
		case loop_flow_inst::f_break: out << "break"; break;
		case loop_flow_inst::f_next: out << "next"; break;
	}
	out << endl();
}

void dump::write_binary_inst( binary_inst const & bi )
{
	out << indent_str();
	write_value_defn( bi.result );
	out << " = ";
	switch( bi.op ) {
		#define OP(t) case binary_inst::o_##t: out << #t; break;
		OP(add)
		OP(sub)
		OP(mul)
		OP(div)
		
		OP(mod)
		
		OP(bit_shl)
		OP(bit_shr)
		
		OP(or)
		OP(and)
		OP(xor)
		OP(and_not)
		OP(nor)
		OP(nand)
		
		OP(eq)
		OP(neq)
		OP(lt)
		OP(lte)
		OP(gt)
		OP(gte)
		
		#undef OP
	}
	out << " ";
	write_value_ref( bi.a );
	out << " ";
	write_value_ref( bi.b );
	out << endl();
}

void dump::write_unary_inst( unary_inst const & ui )
{
	out << indent_str();
	write_value_defn( ui.result );
	out << " = ";
	switch( ui.op ) {
		#define OP(t) case unary_inst::o_##t: out << #t; break;
		OP(neg)
		OP(abs)
		
		OP(not)
		#undef OP
	}
	out << " ";
	write_value_ref( ui.arg );
	out << endl();
}

void dump::write_comment_inst( comment_inst const & ci )
{
	out << indent_str();
	out << "// " << ci.comment;
	out << endl();
}

void dump::write_clone_inst( clone_inst const & ci )
{
	out << indent_str();
	write_value_defn( ci.result );
	out << " = " << (ci.source_value ? "clone_value " : "clone ");
	write_value_ref( ci.source );
	out << endl();
}

void dump::write_cast_inst( cast_inst const & ci )
{
	out << indent_str();
	write_value_defn( ci.result );
	out << " = cast ";
	write_value_ref( ci.source );
	out << endl();
}

} //eon ir
