#include "ir/instruction.hpp"
#include "ir/block.hpp"
#include "ir/type.hpp"
#include "ir/dump.hpp"
#include "util/validation_context.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace ir
{

switch_inst::conditional_block::conditional_block( block & parent )
{
	cond_block = parent.create_child_block();
	expr_block = parent.create_child_block();
}

void store_inst::validate(util::validation_context & vctx) const
{
	left.validate(vctx);
	right.validate(vctx);

	auto t = left.result_type();
	if( t->is_raw_pointer() ) {
		DEBUG( "store_inst", dump::get_type(left.result_type().get()), dump::get_type(right.result_type().get()));
		STATE_CHECK( *left.result_type()->sub.at(0) == *right.result_type() );
	} else if( t->is_shared_pointer() ) {
		STATE_CHECK( *left.result_type()->sub.at(0) == *right.result_type() );
	} else {
		STATE_FAIL( "requires raw/shared pointer for storage" );
	}
}

void funccall_inst::validate(util::validation_context & vctx) const
{
	call.validate(vctx);
	auto ct = call.result_type();
	STATE_CHECK( ct->is_callable() );
	
	STATE_CHECK( returns.size() == ct->sub.size() );
	STATE_CHECK( args.size() == ct->args.size() );
	
	for( size_t i=0; i < returns.size(); ++i )
		STATE_CHECK( *returns[i].result_type() == *ct->sub[i] );
	for( size_t i=0; i < args.size(); ++i ) {
		DEBUG( "ARG", dump::get_type( args[i].result_type().get() ), dump::get_type( ct->args[i].get() ) );
		if( ct->args[i]->fundamental == type::f_abi_var_args ) {
			STATE_CHECK( args[i].result_type()->fundamental == type::f_tuple );
		} else {
			STATE_CHECK( *args[i].result_type() == *ct->args[i] );
		}
	}
}

void pack_tuple_inst::validate(util::validation_context & vctx) const
{
	result.validate(vctx);
	auto pt = result.result_type();
	DEBUG( "VALIDATE", dump::get_type( pt.get() ) );
	STATE_CHECK( pt->sub.size() == sub.size() );
	for( size_t i=0; i < sub.size(); ++i ) {
		DEBUG( "#", i, dump::get_type( sub[i].result_type().get() ), dump::get_type(pt->sub[i].get()) );
		STATE_CHECK( *sub[i].result_type() == *pt->sub[i] );
	}
}

void switch_inst::validate(util::validation_context & vctx) const
{
	if( with_result )
		result.validate(vctx);

	for( auto & cb : blocks )
	{
		cb->cond_block->validate(vctx);
		cb->expr_block->validate(vctx);
		if( cb->cond_value ) {
			cb->cond_value->validate(vctx);
			STATE_CHECK( cb->cond_value->result_type()->is_boolean() );
		}
		
		if( with_result )
		{
			cb->expr_value.validate(vctx);
			STATE_CHECK( *result.result_type() == *cb->expr_value.result_type() );
		}
	}
}

void resolve_inst::validate(util::validation_context & vctx) const
{
	source.validate(vctx);
	result.validate(vctx);
	STATE_CHECK( *result.result_type() == *source.result_type() );
}

void malloc_inst::validate(util::validation_context & vctx) const
{
	STATE_CHECK( direct_type );
	direct_type->validate(vctx);
	result.validate(vctx);
	auto rtype = result.result_type();
	STATE_CHECK( rtype->is_raw_pointer() && *rtype->sub.at(0) == *direct_type );
}

void alloc_shared_inst::validate(util::validation_context & vctx) const
{
	sod.validate(vctx);
	auto rtype = sod.result.result_type();
	STATE_CHECK( rtype->is_shared_pointer() && *rtype->sub.at(0) == *sod.direct_type );
}

void bind_raw_func_inst::validate(util::validation_context & vctx) const
{
	raw_func.validate(vctx);
	STATE_CHECK( raw_func.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( raw_func.result_type()->access == type::a_raw );
	result.validate(vctx);
	STATE_CHECK( result.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( result.result_type()->access == type::a_bound );
	STATE_CHECK( result.result_type()->sub_compare( *raw_func.result_type() ) == 0 );
}

void bind_func_inst::validate(util::validation_context & vctx) const
{
	ctx_data.validate(vctx);
	auto rtype = ctx_data.result.result_type();
	STATE_CHECK( rtype->is_raw_pointer() && *rtype->sub.at(0) == *ctx_data.direct_type );
	
	ctx_func.validate(vctx);
	STATE_CHECK( ctx_func.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( ctx_func.result_type()->is_any_unbound() );
	
	result.validate(vctx);
	STATE_CHECK( result.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( result.result_type()->access == type::a_bound );
}

void bind_func_explicit_inst::validate(util::validation_context & vctx) const
{
	ctx_data.validate(vctx);
	
	ctx_func.validate(vctx);
	STATE_CHECK( ctx_func.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( ctx_func.result_type()->is_any_unbound() );
	
	result.validate(vctx);
	STATE_CHECK( result.result_type()->fundamental == type::f_function_ptr );
	STATE_CHECK( result.result_type()->access == type::a_bound );
	
	STATE_CHECK( !ctx_shared || data_shared );
}

void shared_object_defn::validate(util::validation_context & vctx) const
{
	STATE_CHECK( direct_type );
	direct_type->validate(vctx);
	result.validate(vctx);
	if( !static_destructor )
		destructor.validate(vctx);
}

void release_shared_inst::validate(util::validation_context & vctx) const
{
	object.validate(vctx);
	STATE_CHECK( object.result_type()->is_shared_object_pointer() );
	
	switch( destroy_how )
	{
		case dh_trivial:
		case dh_dynamic:
			break;
	}
}

void acquire_shared_inst::validate(util::validation_context & vctx) const
{
	object.validate(vctx);
	STATE_CHECK( object.result_type()->is_shared_object_pointer() );
}

void alloc_array_inst::validate(util::validation_context & vctx) const
{
	result.validate(vctx);
	STATE_CHECK( item_type );
	item_type->validate(vctx);
	length.validate(vctx);
	STATE_CHECK( length.result_type()->is_integer() );
}

void free_array_inst::validate(util::validation_context & vctx) const
{
	array.validate(vctx);
}

void array_offset_inst::validate(util::validation_context & vctx) const
{
	array.validate(vctx);
	offset.validate(vctx);
	result.validate(vctx);
	auto rt = array.result_type();
	STATE_CHECK( rt->is_array() 
		|| (rt->is_raw_pointer() && rt->sub[0]->is_array()) );
	STATE_CHECK( offset.result_type()->is_integer() );
}

void array_ptr_offset_inst::validate(util::validation_context & vctx) const
{
	array.validate(vctx);
	offset.validate(vctx);
	result.validate(vctx);
	auto rt = array.result_type();
	STATE_CHECK( rt->is_raw_pointer() );
	STATE_CHECK( offset.result_type()->is_integer() );
}

void special_field_inst::validate(util::validation_context & vctx) const
{
	object.validate(vctx);
	result.validate(vctx);
	
	switch( field )
	{
		case f_array_len:
			STATE_CHECK( object.result_type()->fundamental == type::f_array );
			return;
	}
	
	STATE_CHECK( false );
}

loop_inst::loop_inst( block & parent )
	: instruction( t_loop ) {
	cond_block = parent.create_child_block();
	body_block = parent.create_child_block();
}

void loop_inst::validate(util::validation_context & vctx) const
{
	STATE_CHECK(cond_block);
	cond_block->validate(vctx);
	STATE_CHECK(body_block);
	body_block->validate(vctx);
	cond_value.validate(vctx);
	STATE_CHECK( cond_value.result_type()->is_boolean() );
}

void promote_loss_inst::validate(util::validation_context & vctx) const
{
	source.validate(vctx);
	result.validate(vctx);
	STATE_CHECK( to_what >= w_integer && to_what <= w_binary );
	STATE_CHECK( from_what >= w_integer && from_what <= w_binary );
}

void loop_flow_inst::validate(util::validation_context &) const
{
	STATE_CHECK( flow == f_break || flow == f_next );
}

void binary_inst::validate(util::validation_context & vctx) const
{
	a.validate(vctx);
	b.validate(vctx);
	result.validate(vctx);
	STATE_CHECK( *a.result_type() == *b.result_type() );
	if( is_boolean_result() ){
		STATE_CHECK( result.result_type()->is_boolean() );
	} else {	
		STATE_CHECK( *result.result_type() == *a.result_type() );
		STATE_CHECK( a.result_type()->is_scalar() || a.result_type()->is_boolean() );
	}
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
}

void unary_inst::validate(util::validation_context & vctx) const
{
	arg.validate(vctx);
	result.validate(vctx);
	STATE_CHECK( *arg.result_type() == *result.result_type() );
	STATE_CHECK( arg.result_type()->is_scalar() ||
		(arg.result_type()->is_boolean() && op == o_not) );
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
}

void clone_inst::validate(util::validation_context & vctx) const
{
	source.validate(vctx);
	result.validate(vctx);
	if( source_value ) {
		STATE_CHECK( result.result_type()->is_raw_pointer() );
		STATE_CHECK( *result.result_type()->sub[0] == *source.result_type() );
	} else {
		STATE_CHECK( source.result_type()->is_raw_pointer() );
		STATE_CHECK( *result.result_type() == *source.result_type() );
	}
}

void cast_inst::validate(util::validation_context & vctx) const {
	source.validate(vctx);
	result.validate(vctx);
}

void alloc_uninit_inst::validate(util::validation_context & vctx) const {
	STATE_CHECK( direct_type );
	direct_type->validate(vctx);
	
	result.validate(vctx);
}

} //eon ir
