#include "ir/llvm/builder.hpp"

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
	#include "llvm/IR/Module.h"
	#include "llvm/IR/Function.h"
#pragma GCC diagnostic pop

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include <fstream>

namespace ir_llvm 
{

block_ptr block_data::get_flow( block_flow bf ) const
{
	if( bf == bf_seq && block_seq ) {
		return block_seq;
	}
	return context->get_flow(bf);
}

block_ptr block_context::get_flow( block_flow bf ) const 
{
	if( flow[bf] )
		return flow[bf];
	if( parent )
		return parent->get_flow(bf);
	return block_ptr();
}

int block_context::get_max_error_level() const
{
	if( !parent )
		return error_level;
	return std::max( error_level, parent->error_level );
}

/////////////////////////////////////////////////////////////
// function_builder
function_builder::function_builder( llvm::Module & module, llvm::Function & function, 
	type_manager & types, gen::options const & gopts ) 
	: lbuilder( function.getContext() )
	, module( module )
	, llvm_context( function.getContext() )
	, function( function )
	, types( types )
	, gopts( gopts )
{ 
	entry_block = cur_block = block_data::create( 
		make_shared<block_context>(),
		llvm::BasicBlock::Create( llvm_context, "entry", &function ) );
	all_blocks.push_back( cur_block );
	lbuilder.SetInsertPoint( cur_block->ll_block );
	
	return_path = nullptr;
	error_value = nullptr;
	error_info = nullptr;
	label_count = 0;
}

//resumes the current exception, in the current model this is no different than rethrowing
block_ptr function_builder::get_resume_block()
{
	return get_rethrow_block();
}

//rethrows a handled exception
block_ptr function_builder::get_rethrow_block()
{
	if( rethrow_block )
		return rethrow_block;
	
	auto old_block = cur_block;
	rethrow_block = create_child_block( "rethrow" );
	set_insert_point( rethrow_block );
	
	llvm::Value * error = llvm::UndefValue::get( types.error_return() );
	error = CreateInsertValue( error, CreateLoad(error_value), ef_error_code );
	error = CreateInsertValue( error, CreateLoad(error_info), ef_error_data );
	func_return( error );
	
	set_insert_point( old_block );
	return rethrow_block;
}

void function_builder::complete_blocks()
{
	propagate_flow( entry_block, block_flow_flag(bf_seq) );

#if 0
	//debugging dump
	auto flow_str = []( unsigned flow ) {
		std::string result;
		#define BFN(p) if( flow & block_flow_flag(bf_##p) ) { result += #p " "; }
		BFN(seq)
		BFN(return)
		BFN(loop_break)
		BFN(loop_next)
		BFN(error)
		#undef BFN
		return result;
	};
	
	std::ofstream pd( (std::string("/tmp/dump_") + 
		function.getName() + ".txt").str().c_str(), std::ios::binary );
		
	auto dblock = [&]( std::string const & label, block_ptr bk ) {
		pd << "\t" << label << ": "
			<< (bk ? bk->ll_block->getName().str() : "MISSING" ) 
			<< std::endl;
	};
	
	for( auto & mb : all_blocks ) {
		pd << mb->ll_block->getName().str();
		if( mb->complete )
			pd << " complete";
		if( mb.get() == entry_block.get() )
			pd << " entry";
		pd << std::endl;
		
		pd << "\tflow_in: "  << flow_str( mb->flow_in ) << std::endl;
		pd << "\tflow_out: " << flow_str( mb->flow_out ) << std::endl;
		if( mb->block_undivert ) {
			pd << "\tundivert: " << mb->block_undivert->ll_block->getName().str() << std::endl;
		}
		if( mb->undiverts ) {
			pd << "\tundiverts" << std::endl;
		};
		for( int i =0; i < mb->cvt_size; ++i ) {
			pd << "\tconvert: " << flow_str(block_flow_flag(mb->cvt_in[i]))
				<< " -> " << flow_str(block_flow_flag(mb->cvt_out[i])) << std::endl;
		}
		
		for( int8_t i = 0; i < block_flow_size; ++i ) {
			if( !(mb->flow_out & block_flow_flag(block_flow(i)) ) ) {
				continue;
			}			
			auto db = mb->get_flow(block_flow(i));
			dblock( flow_str(block_flow_flag(block_flow(i))), db );
		}
		if( mb->cond_value ) {
			dblock( "block_cond_true", mb->block_cond_true );
		}
		if( mb->block_undivert ) {
			dblock( "block_undivert", mb->block_undivert );
		}
	}
	pd.close();
#endif

	//DEBUG( "------", GetInsertBlock().getFunction()->getName().str() );

	//using indexing since "all_blocks" can be extended within complete_block (iterators would invalidate)
	for( size_t i=0; i < all_blocks.size(); ++i ) {
		complete_block( all_blocks[i] );
	}
}

void function_builder::propagate_flow( block_ptr block, unsigned flow )
{
	unsigned c = block->flow_in | flow;
	//already handled
	if( c == block->flow_in ) {
		return;
	}
	block->flow_in = c;

	//converted flow (these are applied in order since they could be chained
	//for example, a error_post block which then resumes)
	for( int i=0; i < block->cvt_size; ++i ) {
		if( c & block_flow_flag(block->cvt_in[i]) ) {
			c &= ~block_flow_flag(block->cvt_in[i]);
			c |= block_flow_flag(block->cvt_out[i]);
		}
	}

	// all diverted flow becomes seq for this block
	if( block->block_undivert ) {
		propagate_flow( block->block_undivert, c );
		c = block_flow_flag(bf_seq);
	}
	
	block->flow_out |= c;
	
	unsigned pflow = block->flow_out;
	//undiverts doesn't propagate seq flow, since previous block_undivert did proper flow
	if( block->undiverts ) {
		pflow &= ~block_flow_flag(bf_seq);
	}
	
	//conditionals don't interrupt flow
	if( block->cond_value ) {
		STATE_CHECK( block->block_cond_true );
		STATE_CHECK( block->block_seq );
		propagate_flow( block->block_cond_true, pflow );
		propagate_flow( block->block_seq, pflow );
	}
	
	//branched flow
	for( int8_t i=0; i < block_flow_size; ++i ) {
		auto bf = block_flow(i);
		if( pflow & block_flow_flag(bf) ) {
			auto f = block->get_flow(bf);
			//STATE_CHECK(f) //should make resume another block
			if( f ) {
				propagate_flow( f, block_flow_flag(bf) );
			}
		}
	}
}
		
void function_builder::complete_block( block_ptr mb )
{
	if( mb->complete ) {
		return;
	}
	
	//check for unneeded blocks
#if 0
	if( !mb->flow_in && !mb->flow_out ) {
		mb->complete = true;
		mb->ll_block->removeFromParent();
		return;
	}
#endif
	
	STATE_CHECK( mb );
	STATE_CHECK( mb->ll_block );
	set_insert_point( mb );
	
	if( mb->cond_value ) {
		//STATE_CHECK( mb->flow_out == block_flow_flag(bf_seq),
		//	std::string( "invalid-block-out-flow:" ) + mb->ll_block->getName().str() );
		STATE_CHECK( mb->block_cond_true );
		STATE_CHECK( mb->block_seq );
		lbuilder.CreateCondBr( mb->cond_value, 
			mb->block_cond_true->ll_block, mb->block_seq->ll_block );
		mb->complete = true;
		return;
	}
	
	struct case_pair {
		llvm::ConstantInt * value;
		llvm::BasicBlock * dest;
	};
	std::vector<case_pair> case_pairs;
	for( int8_t bfi=0; bfi < block_flow_size; ++bfi ) {
		auto bf = block_flow(bfi);
		if( mb->flow_out & block_flow_flag(block_flow(bf)) ) {
			auto sb = mb->get_flow(bf);
			if( !sb && bf == bf_error ) {
				sb = get_rethrow_block();
			}
			STATE_CHECK( sb );
			case_pair cp;
			cp.value = ConstInt8(bfi);
			cp.dest = sb->ll_block;
			case_pairs.push_back( cp );
		}
	};
	
	if( case_pairs.size() == 0 ) {
		//STATE_CHECK( !case_pairs.empty() );
		lbuilder.CreateUnreachable();
	} else if( case_pairs.size() == 1 ) {
		lbuilder.CreateBr( case_pairs[0].dest );
	} else {
		auto sw = lbuilder.CreateSwitch( 
			CreateLoad( return_path, "return_path" ),
			case_pairs[0].dest, 
			int(case_pairs.size()) );
		for( auto & cp : case_pairs )
			sw->addCase( cp.value, cp.dest );
	}
	
	mb->complete = true;
}


//NOTE: I'm unsure if this should actually be in builder at all and not just in gen
llvm::Value * function_builder::create_abs( llvm::Value * arg )
{
	auto in_block = cur_block;
	auto cond = create_lt( arg, const_zero( arg->getType() ) );
	
	auto neg_block = create_child_block();
	auto post_block = create_child_block();
	branch_conditional( cond, neg_block, post_block );
	
	set_insert_point( neg_block );
	auto abs = create_neg( arg );
	branch_sequence( post_block );
	
	set_insert_point( post_block );
	auto phi = create_phi( arg->getType(), 2 );
	phi->addIncoming( arg, in_block->ll_block );
	phi->addIncoming( abs, neg_block->ll_block );
	return phi;
}

llvm::Value * function_builder::call( type_manager::func_type_entry const & f_type, llvm::Value * callee,
	std::vector<llvm::Value*> const args_, llvm::Value** out_ret_ptr ) {
	
	auto args = args_;
	
	llvm::Value * ret_ptr = nullptr;
	if( !f_type.is_direct_return && f_type.ll_logical_return_type ) {
		ret_ptr = CreateAlloca( f_type.ll_logical_return_type );
		args.insert( args.begin(), ret_ptr );
	}
	if( out_ret_ptr ) {
		*out_ret_ptr = ret_ptr;
	}
	
	llvm::Value * result = lbuilder.CreateCall( callee, args );
	
	if( f_type.is_error_return ) {
			
		auto code = CreateExtractValue( result, 0 );
		auto cnd = create_neq( const_zero( types.f_error_code()), code );
			
 		auto seq_block = create_child_block( "post_call" );
 			
		auto err_block = create_child_block( "error_call" );
		branch_conditional( cnd, err_block, seq_block );			

		set_insert_point( err_block );
		CreateStore( code, error_value );
		CreateStore( CreateExtractValue( result, 1 ), error_info );
		flow_resume();
			
 		set_insert_point( seq_block );
 		
		//TODO: is it an error to not have a `ret_ptr` at this point, or does it mean there is simply no return value?
		result = llvm::UndefValue::get( types.type_void() );
	}
	
	if( ret_ptr ) {
		result = CreateLoad( ret_ptr );
	}
 		
 	return result;
}

void function_builder::create_function_entry() {
	//track whcih return path is being used, 0=normal, 1=shortened
	return_path = CreateAlloca( types.f_intx(8), 0, "_return_path" );
	CreateStore( ConstInt8(function_builder::rpv_sequential), return_path );
	
	error_value = CreateAlloca( types.f_error_code(), 0, "_error_value" );
	CreateStore( const_zero( types.f_error_code() ), error_value );
	
	error_info = CreateAlloca( types.generic_shared_ptr(), 0, "_error_info" );
	CreateStore( const_zero( types.generic_shared_ptr() ), error_info );
}

} //eon ir_llvm