#include "ir/language.hpp"

#include "ir/llvm/gen.hpp"
#include "ir/llvm/type_manager.hpp"

#include "util/validation_context.hpp"

#include "ir/llvm/builder.hpp"

#include "platform/platform.hpp"

//NOTE: I'm not entirely clear on these defines
#define __STDC_LIMIT_MACROS
#define __STDC_CONSTANT_MACROS 

// LLVM: doesn't seem to care about conversion warnings
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
	#include "llvm/IR/Verifier.h"
	#include "llvm/IR/AssemblyAnnotationWriter.h"
	//#include "llvm/Assembly/PrintModulePass.h"
	#include "llvm/IR/Constants.h"
	#include "llvm/IR/DataLayout.h"
	#include "llvm/IR/DerivedTypes.h"
	#include "llvm/ExecutionEngine/GenericValue.h"
	#include "llvm/ExecutionEngine/Interpreter.h"
	#include "llvm/LinkAllPasses.h"
	#include "llvm/Linker/Linker.h"
	#include "llvm/IR/Module.h"
	#include "llvm/Support/FormattedStream.h"
	#include "llvm/IRReader/IRReader.h"
	#include "llvm/IR/NoFolder.h"
	#include "llvm/Support/raw_ostream.h"
	#include "llvm/Support/TargetSelect.h"
	#include "llvm/Support/SourceMgr.h"
	//added for object files
	#include "llvm/MC/MCAsmBackend.h"
	#include "llvm/MC/MCAsmInfo.h"
	#include "llvm/MC/MCCodeEmitter.h"
	#include "llvm/MC/MCContext.h"
	#include "llvm/MC/MCInstPrinter.h"
	#include "llvm/MC/MCInstrInfo.h"
	#include "llvm/MC/MCObjectFileInfo.h"
	#include "llvm/MC/MCRegisterInfo.h"
	#include "llvm/MC/MCSectionMachO.h"
	#include "llvm/MC/MCStreamer.h"
	#include "llvm/MC/MCSubtargetInfo.h"
	#include "llvm/MC/SubtargetFeature.h"
	#include "llvm/Support/Host.h"
	#include "llvm/Support/TargetRegistry.h"
	#include "llvm/Support/ToolOutputFile.h"
	#include "llvm/Support/raw_os_ostream.h"
	#include "llvm/IR/LegacyPassManager.h"
#pragma GCC diagnostic pop
#undef DEBUG

#include <fstream>
#include <iostream>

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "ir/dump.hpp"

namespace
{
	template<typename T>
	void dump( T & t ) {
		std::string out;
		llvm::raw_string_ostream outs( out );
		t.print( outs, nullptr );
		std::cout << out << std::endl;
	}
	
	template<typename T>
	void dumps( T & t ) {
		std::string out;
		llvm::raw_string_ostream outs( out );
		t.print( outs );
		std::cout << out << std::endl;
	}
	
	typedef std::set<llvm::Type const*> type_set_t;
	
	void dump_type( llvm::raw_ostream & out, llvm::Type * t, type_set_t & done) {
		//stop recursion
		if( done.find(t) != done.end() ) {
			out << "%";
			return;
		}
		done.insert(t);

		if( t->isPointerTy() ) {
			dump_type( out, t->getPointerElementType(), done );
			out << "*";
		} else if( t->isStructTy() ) {
			auto sty = llvm::cast<llvm::StructType>(t);
			if( sty->hasName() ) {
				out << sty->getName() << " ";
			}
			out << "{";
			for( unsigned i=0; i< t->getStructNumElements(); ++i ) {
				if( i ) {
					out << ",";
				}
				dump_type( out, t->getStructElementType(i), done );
			}
			out << "}";
		} else {
			t->print( out );
		}

		done.erase(t);
	}

	/**
		This allows a recursive dump of llvm types. The standard "print" doesn't go into the pointed
		to types.
	*/
#if SHOW_DEBUG == 1
	std::string get_type( llvm::Type * t ) {
		std::string under;
		llvm::raw_string_ostream out(under);
		type_set_t done;
		dump_type( out, t, done );
		return out.str();
	}
#endif
}

namespace ir_llvm {

gen::gen( options gopts ) :
	gopts( gopts ) {
	llvm::InitializeNativeTarget();
	llvm_context.reset( new llvm::LLVMContext );
	module = new llvm::Module( "test", *llvm_context );
	del_module = true;
	data_layout.reset( new llvm::DataLayout( module ) );
	types.reset( new type_manager( *llvm_context, *data_layout ) );

	create_std_functions();
	
	llvm::SMDiagnostic err;	
	
	//LLVM: Not clear why I need this, especially since InitializeNativeTarget was called already
	llvm::InitializeNativeTargetAsmPrinter();
	llvm::InitializeNativeTargetAsmParser();
}

gen::~gen() {
	if( del_module ) {
		delete module;
	}
}

/**
	This wraps the Leaf-IR entry point with a final exception handling clause. Though perhaps
	we shouldn't do this and instead let Leaf terminate (since the debugger then knows where
	it is).
*/
void gen::build_wrapped_entry(ir::module const & m)
{
	auto f = m.if_get_function( "_entry" );
	if( !f ) {	
		return;
	}
	auto f_type = types->function(*f->defn_value->direct_type);
	
	//DEBUG( "FTYPE", ir::dump::get(f->defn_value->direct_type.get()));
	
	auto main = llvm::cast<llvm::Function>(module->getFunction( "_entry" ));
	STATE_CHECK( main );
	auto unhandled = module->getFunction( "_unhandled_error" );
	STATE_CHECK( unhandled );

	STATE_CHECK( !module->getFunction("main" ) );
	auto func = llvm::cast<llvm::Function>(module->getOrInsertFunction( "main",
		types->f_intx(platform::abi_int_size), //return
		types->main_argc(), types->main_argv(), //args
		(llvm::Type*)0 ) );
	func->setLinkage(llvm::GlobalValue::ExternalLinkage);
	STATE_CHECK( func );
	
	function_builder builder( *module, *func, *types, gopts );
	builder.create_function_entry();
	
	//store argc/argv in globals
	auto args_at = func->arg_begin();
	auto argc = new llvm::GlobalVariable( *module, types->main_argc(), /*constant*/false,
		llvm::GlobalValue::ExternalLinkage, /*init*/0, "__leaf_argc" );
	argc->setInitializer( llvm::UndefValue::get( types->main_argc() ) );
	builder.CreateStore( &*(args_at++), argc );
	
	auto argv = new llvm::GlobalVariable( *module, types->main_argv(), /*constant*/false,
		llvm::GlobalValue::ExternalLinkage, /*init*/0, "__leaf_argv" );
	argv->setInitializer( llvm::UndefValue::get( types->main_argv() ) );
	builder.CreateStore( &*(args_at++), argv );
	
	auto cblock = builder.create_child_block( "catch" );

	//builder.get_context()->error_land = cblock;
	builder.set_rethrow_block(cblock);

	auto result = builder.call( f_type, main, std::vector<llvm::Value*>() );
	DEBUG( "Main-TYPE", get_type( result->getType() ) );
	
	if( result->getType()->isVoidTy() ) {
		builder.func_return( builder.ConstInt( types->f_intx(platform::abi_int_size), 0 ) );
	} else {
		//allow any integral type to be used
		builder.func_return( builder.create_sext_or_trunc( result,
			types->f_intx(platform::abi_int_size) ) );
	}
	
	//only global variables can be used for the clauses
	builder.set_insert_point( cblock );
	builder.Call( unhandled );
	builder.func_return( builder.const_abi_int( 255 ) );
	builder.complete_blocks();
}

void gen::show_code( std::string const & sc )
{
	if( sc == "*" ) {
		dump( *module ); 
	} else if( !sc.empty() ) {
		for( auto p =  func_map.equal_range( sc ); p.first != p.second; ++p.first ) {
			if( p.first->first == sc ) {
				dumps( *p.first->second );
			}
		}
	}
}


int gen::execute( ir::module const & m, std::string const & sc ) //main
{
	util::validate(m);
	build_module( m );
	build_wrapped_entry( m);
	
	//print out the IR before adding imported functions (thus showing only interesting bits)
	show_code( sc );

	auto ee = create_engine();
	ee->finalizeObject();
	auto main = ee->FindFunctionNamed( "main" );
	
	std::vector<llvm::GenericValue> args(2);
	args[0].IntVal = llvm::APInt( platform::abi_int_size, 0 );
	args[1].PointerVal = nullptr;
	llvm::GenericValue gv = ee->runFunction( main, args );
	auto ret = int(gv.IntVal.getSExtValue());
	delete ee;
	return ret;
}

void gen::create_std_functions( )
{
	module->getOrInsertFunction(
		"count_malloc", types->generic_ptr(), types->f_natural_int(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"count_free", types->type_void(), types->generic_ptr(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_void_destructor", types->type_void(), types->generic_ptr(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_duplicate_release_shared", types->type_void(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_unhandled_error", types->type_void(), (llvm::Type*)0 );
	module->getOrInsertFunction(
		"_bad_fail", types->type_void(), (llvm::Type*)0 );
}

llvm::Value * gen::gen_malloc( llvm::Value * size, function_builder & bb )
{
	//TODO: called function should throw if cannot allocate
	auto func = llvm::cast<llvm::Function>( module->getFunction( "count_malloc" ) );
	STATE_CHECK( func );
	return bb.Call( func, size ); 
}

llvm::Value * gen::gen_malloc( size_t size, function_builder & bb ) 
{
	return gen_malloc( bb.const_natural_int(int64_t(size)), bb );
}

llvm::Value * gen::heap_alloc_type( llvm::Type * t, function_builder & bb )
{
	auto raw = gen_malloc( types->alloc_size( t ), bb );
	return bb.CreateBitCast( raw, types->ptr_to(t) );
}

void gen::gen_free( llvm::Value * ptr, function_builder & bb ) {
	auto func = llvm::cast<llvm::Function>( module->getFunction( "count_free" ) );
	STATE_CHECK( func );
	
	std::vector<llvm::Value*> args;
	args.push_back( bb.CreateBitCast( ptr, types->generic_ptr() ) );
	bb.Call( func, args );
}

/**
	Conditionally releases a shared object (if the pointer is no null).
	If the pointer is non-null it will be passed to `build_release_shared`, otherwise nothing happens.
*/
void gen::gen_release_optional_shared( llvm::Value * ptr, function_builder & bb ) {
	auto func = module->getFunction( "_release_opt_shared" );
	if( !func ) {
		func = llvm::cast<llvm::Function>(module->getOrInsertFunction( "_release_opt_shared",
			types->type_void(), //return
			types->generic_shared_ptr(), //args
			(llvm::Type*)0 ) );
		
		function_builder builder( *module, *func, *types, gopts );
		
		auto args_at = func->arg_begin();
		auto ptr = &*(args_at++);
		
		auto free_shared = builder.create_child_block( "free_shared" );
		auto cont = builder.create_child_block( "cont" );
		
		auto ptr_int = builder.create_ptr_to_int( ptr );
		auto cnd = builder.CreateICmpEQ( ptr_int, builder.const_natural_int(0));
		builder.branch_conditional( cnd, cont, free_shared );
		
		builder.set_insert_point( free_shared );
		build_release_shared( ptr, true, builder);
		builder.branch_sequence( cont );
		
		builder.set_insert_point( cont );
		builder.func_return_void();
		
		builder.complete_blocks();
	}
	
	//free up the error_info
	std::vector<llvm::Value*> args;
	args.push_back( ptr );
	bb.Call( func, args );
}

void gen::gen_acquire_optional_shared( llvm::Value * ptr, function_builder & bb ) {
	auto func = module->getFunction( "_acquire_opt_shared" );
	if( !func ) {
		func = llvm::cast<llvm::Function>(module->getOrInsertFunction( "_acquire_opt_shared",
			types->type_void(), //return
			types->generic_shared_ptr(), //args
			(llvm::Type*)0 ) );
		
		function_builder builder( *module, *func, *types, gopts );
		
		auto args_at = func->arg_begin();
		auto ptr = &*(args_at++);
		
		auto free_shared = builder.create_child_block( "acquire_shared" );
		auto cont = builder.create_child_block( "cont" );
		
		auto ptr_int = builder.create_ptr_to_int( ptr );
		auto cnd = builder.CreateICmpEQ( ptr_int, builder.const_natural_int(0));
		builder.branch_conditional( cnd, cont, free_shared );
		
		builder.set_insert_point( free_shared );
		build_acquire_shared( ptr, builder);
		builder.branch_sequence( cont );
		
		builder.set_insert_point( cont );
		builder.func_return_void();
		
		builder.complete_blocks();
	}
	
	//acquire up the error_info
	std::vector<llvm::Value*> args;
	args.push_back( ptr );
	bb.Call( func, args );
}

void gen::build_module( ir::module const & m ) {
	for( auto git = m.global_variable_map.begin(), end = m.global_variable_map.end(); git != end; ++git )
		build_global_variable( *git->second );
		
	for( auto fit = m.function_map.begin(), end = m.function_map.end(); fit != end; ++fit )
		setup_function( *fit->second );
	for( auto fit = m.function_map.begin(), end = m.function_map.end(); fit != end; ++fit )
		build_function( *fit->second );
}

void gen::create_symbol_cref( intptr_t & track, llvm::Value * v, symbol_cref::ref_type_t rt )
{
	STATE_CHECK( !track );
	
	symbol_deque.resize( symbol_deque.size() + 1 );
	symbol_cref & sc = symbol_deque.back();
	sc.ref_type = rt;
	sc.value = v;
	
	track = reinterpret_cast<intptr_t>(&sc);
}

void gen::build_global_variable( ir::value const & gv )
{
	auto gtype = types->value( *gv.direct_type );
	
	auto use_name = gv.name;
	if( use_name == "main" ) {
		use_name = "_main";
	}
	auto g = new llvm::GlobalVariable( *module, gtype.ll_type, /*constant*/false, 
		gv.is_export ? llvm::GlobalValue::ExternalLinkage : llvm::GlobalValue::InternalLinkage, 
		/*init*/0, use_name );
	g->setInitializer( llvm::UndefValue::get( gtype.ll_type ) );
		
	create_symbol_cref( gv.cref, g, symbol_cref::rt_global );
}

void gen::setup_function( ir::function const & f )
{
	DEBUG( "SETUP-FUNCTION", f.defn_value->name );
	
	auto func_type = types->function( *f.defn_value->direct_type );
	
	auto use_name = f.defn_value->name;
	
	auto func = llvm::cast<llvm::Function>(module->getOrInsertFunction( use_name, 
		func_type.ll_func_type ) );
	func->setLinkage( f.is_export ? llvm::GlobalValue::ExternalLinkage : llvm::GlobalValue::InternalLinkage);
	create_symbol_cref( f.defn_value->cref, func, symbol_cref::rt_value );
	
	func_map.insert( std::make_pair( f.defn_value->name, func ) );
}

void gen::build_function( ir::function const & f )
{
	DEBUG( "BUILD-FUNCTION", f.defn_value->name );
	
	auto sr = reinterpret_cast<symbol_cref*>( f.defn_value->cref );
	auto func = llvm::cast<llvm::Function>(sr->value);
	auto func_type = types->function( *f.defn_value->direct_type );
	
	function_builder builder( *module, *func, *types, gopts );
	builder.create_function_entry();
	
	//setup for a throw inside a function that doesn't have errors
	block_ptr err_block;
	if( !func_type.is_error_return ) {
		err_block = builder.create_child_block( "bad_fail" );
		builder.set_rethrow_block( err_block );
	}
	
	//connect args/return
	auto args_at = func->arg_begin();
	//either we create a local return allocation or use the provided one (first argument)
	llvm::Value * ret_value = nullptr;
	if( f.returns.size() ) {
		if( func_type.is_direct_return )
			ret_value = builder.CreateAlloca( func->getReturnType(), 0, "_return_val" );
		else
			ret_value = &*(args_at++);
	}
	
	if( func_type.is_context )	{
		auto arg = &*(args_at++);
		llvm::Value * context_ptr;
		
		if( func_type.is_context_shared ) {
			//DEBUG( "SCTX_TYPE", ir::dump::get( f.defn_value->direct_type->context.get() ) );
			llvm::Type * sctx_type = types->shared_ptr( *f.defn_value->direct_type->context ).ll_type;
			builder.context_arg = builder.CreateBitCast( arg, types->ptr_to( sctx_type ), "_context_shared" );
			context_ptr = builder.CreateStructGEP( builder.context_arg, sf_data );
		} else {
			//DEBUG( "CTX_TYPE", ir::dump::get( f.defn_value->direct_type->context.get() ) );
			llvm::Type * ctx_type = types->value( *f.defn_value->direct_type->context ).ll_type;
			context_ptr = builder.CreateBitCast( arg,  types->ptr_to( ctx_type ), "_context" );
			builder.context_arg = context_ptr;
		}
		DEBUG("CARG", get_type( builder.context_arg->getType()) );
		create_symbol_cref( f.context_arg.origin->cref, builder.context_arg, symbol_cref::rt_value );
		
		DEBUG( "CAPTURES", f.context_args.size() );
		for( size_t i=0; i < f.context_args.size(); ++i ) {
			auto & c = f.context_args[i];
			size_t parent = c.origin->capture.parent;
			size_t index = c.origin->capture.index;
			//DEBUG( "CAPTURE", i, parent, index, c.origin->name, reinterpret_cast<intptr_t>(&c.origin->cref) );
			llvm::Value * base = context_ptr;
			
			while( parent > 0 ) {
				//DEBUG( "BASETYPE", get_type(base->getType()) );
				auto sptr = builder.CreateStructLoad( base, 0, "parent" );
				//DEBUG( "SPTR", get_type(sptr->getType()) );
				base = builder.CreateStructGEP( sptr, sf_data, "parent" );
				parent--;
			}
			//DEBUG( "BASETYPE", get_type(base->getType()) );
				
			auto left = builder.CreateStructGEP( base, unsigned(index), c.origin->name.c_str() );
			create_symbol_cref( c.origin->cref, left, symbol_cref::rt_pointer );
		}
	}
	
	for( size_t i=0; i < f.args.size(); ++i )
	{
		//arguments must be copied since LLVM doesn't provide a pointer. Ultimately we should
		//optimize this to only do those which are needed (though LLVM seems to do a fine job
		//of identifying and removing those in its own)
		STATE_CHECK( args_at != func->arg_end() );
		auto local = builder.CreateAlloca( args_at->getType() );
		builder.CreateStore( &*(args_at++), local );
		create_symbol_cref( f.args[i].origin->cref, local, symbol_cref::rt_pointer );
		
		//if poiner not needed
		//create_symbol_cref( f.args[i].origin->cref, args_at++, symbol_cref::rt_value );
	}
	
	unsigned ret_index = 0;
	for( size_t i=0; i < f.returns.size(); ++i )
	{
		llvm::Value * v;
		if( f.returns.size() == 1 )
			v = ret_value;
		else
			v = builder.CreateStructGEP( ret_value, ret_index++, f.returns[i].origin->name.c_str() );
			
		create_symbol_cref( f.returns[i].origin->cref, v, symbol_cref::rt_pointer );
	}

	//build locals
	for( size_t i=0; i < f.locals.size(); ++i )
	{
		auto & local = *f.locals[i];
		auto v = builder.CreateAlloca( types->value( *local.direct_type ).ll_type, 0, local.name );
		create_symbol_cref( local.cref, v, symbol_cref::rt_pointer );
	}

	auto context = builder.get_context();
	auto end_block = builder.create_child_block( "func_ret" );
	context->flow[bf_seq] = end_block;
	context->flow[bf_return] = end_block;
	
	build_block( *f.entry_block, builder );
	
	builder.set_insert_point( end_block );
	//with no returns, or returns as a pointer, we return void
	if( func_type.is_direct_return ) {
	 	builder.func_return( builder.CreateLoad( ret_value ) );
	} else if( func_type.is_error_return ) {
		llvm::Value * non_error = llvm::UndefValue::get( types->error_return() );
		non_error = builder.CreateInsertValue( non_error, builder.const_zero( types->f_error_code() ), ef_error_code );
		non_error = builder.CreateInsertValue( non_error, builder.const_zero( types->generic_shared_ptr() ), ef_error_data );
		builder.func_return( non_error );
	} else {
		builder.func_return_void();
	}
		
	if( err_block ) {
		builder.set_insert_point( err_block );
		auto bad = module->getFunction( "_unhandled_error" );
		STATE_CHECK( bad );
		builder.Call( bad );
		builder.unreachable();
	}
	
	builder.complete_blocks();
}

void gen::build_block_part( ir::block_seq_item const & bk, function_builder & bb )
{
	if( bk.an_instruction ) {
		build_instruction( *bk.an_instruction, bb );
	}
	if( bk.a_block ) {
		build_block( *bk.a_block, bb );
	}
}

namespace {
	struct defer_blocks {
		block_ptr body, end;
	};
	
	/**
		Branches from the insertion point into a defer block structure.
		
		@param build_block will be called, with the correct insertion point, to construct the deferred handling
		@return defer_blocks
			body: block for code that follows the defer statement. All paths out of this structure will lead into
				the block set in `build_block`
			end: sequential end block that exits from the defer structure
	*/
	defer_blocks build_defer_block( function_builder & bb, 
		std::function<void(function_builder & bb)> build_block ) {
		auto end_block = bb.create_child_block( "defer_end" );
		auto cont_block = bb.create_child_block( "defer_cont" );
		auto defer_block = bb.create_child_block( "defer_block" );
		
		auto defer_body = bb.create_child_context( "defer_body" );
		defer_body->context->flow[bf_seq] = defer_block;
		defer_body->context->flow[bf_loop_break] = defer_block;
		defer_body->context->flow[bf_loop_next] = defer_block;
		defer_body->context->flow[bf_return] = defer_block;
		defer_body->context->flow[bf_error] = defer_block;
		bb.branch_sequence(defer_body);
		
		bb.set_insert_point( defer_block );
		auto old_return_path = bb.flow_divert(cont_block);
		build_block( bb );
		bb.flow_undivert( cont_block, old_return_path );
		
		//cont_block handles cases where there is a conditional branch in the defer_end block,
		//which can't be handled at the moment, so this split is done
		cont_block->block_seq = end_block;
		
		return defer_blocks{ defer_body, end_block };
	}
}

void gen::build_block( ir::block const & bk, function_builder & bb )
{
	//needed in case the context changes to go back to original context
	block_ptr seq_block;
	
	if( bk.error_block ) {
		auto error_end = bb.create_child_block( "error_end" );
		auto error_catch = bb.create_child_block( "error_catch" );
		
		auto try_body = bb.create_child_context( "try_body" );
		try_body->context->flow[bf_seq] = error_end;
		try_body->context->flow[bf_error] = error_catch;
		bb.branch_sequence( try_body );
		
		auto onfail_block = create_catch_block( error_catch, bb, bk.error_block->error_level );

		// onfail block {
		bb.set_insert_point( onfail_block );
		auto error_info = bb.CreateLoad(bb.error_info);
		create_symbol_cref( bk.error_block->error_info->origin->cref, error_info, symbol_cref::rt_value );
		
		//TODO: this should probably be a defer_error block then resume does not need to acquire the inst
		auto dblocks = build_defer_block( bb, [&]( function_builder & fbb ) {
			gen_release_optional_shared( error_info, fbb );
		});
		bb.set_insert_point( dblocks.body );
		build_block( *bk.error_block, bb );
		
		bb.set_insert_point( dblocks.end );
		bb.branch_sequence( error_end );
		// } onfail block end
		
		bb.set_insert_point( try_body );
		seq_block = error_end;
	}
	
	for( size_t i=0; i < bk.items.size(); ++i ) {
		auto & si = bk.items[i];
		
		switch( si.flow ) {
			case ir::block_flow::defer: {
				auto dblocks = build_defer_block( bb, [&]( function_builder & fbb ) {
					build_block_part( si, fbb );
				});
				
				if( !seq_block )
					seq_block = dblocks.end;
				bb.set_insert_point( dblocks.body );
				break;
			}
				
			case ir::block_flow::defer_fail: {
				
				auto end_block = bb.create_child_block( "defer_fail_end" );
				auto defer_block = bb.create_child_block( "defer_fail_block" );
				
				auto defer_body = bb.create_child_context( "defer_fail_body" );
				defer_body->context->flow[bf_error] = defer_block;
				bb.branch_sequence( defer_body );
				
				bb.set_insert_point( defer_block );
				//TODO: somewhat copied from on_error above
				if( si.a_block && si.a_block->error_info ) {
					auto error_info = bb.CreateLoad(bb.error_info);
					create_symbol_cref( si.a_block->error_info->origin->cref, error_info, symbol_cref::rt_value );
				}
				auto old_return_path = bb.flow_divert( end_block );
				build_block_part( si, bb );
				bb.flow_undivert( end_block, old_return_path );
				
				bb.set_insert_point( defer_body );
				break;
			}
			
			case ir::block_flow::sequential: {
				build_block_part( si, bb );
				break;
			}
		} //end flow-switch
	} //end items loop

	if( seq_block )
		bb.set_insert_point( seq_block );
}

/**
	Creates the block that conditionally intercepts the current error.
	
	@param catch_block where to write the catch handling
	@return the block for onfail handling (the error was caught)
*/
block_ptr gen::create_catch_block( block_ptr catch_block, function_builder & bb, int landing_type ) {
	
	STATE_CHECK( landing_type );
	
	bb.set_insert_point(catch_block);
	auto cnd = bb.CreateICmpSLT( 
		bb.const_int( types->f_error_code(), landing_type),
		bb.CreateLoad( bb.error_value )
		);
		
	auto onfail_block = bb.create_child_block( "catch" );
	auto cont_error = catch_block->context->get_flow(bf_error);
	if( !cont_error )
		cont_error = bb.get_resume_block();
	bb.branch_conditional( cnd, cont_error, onfail_block );
	
	//error is handled so revert to sequential flow (it is still handled even if part of a continuation)
	bb.set_insert_point( onfail_block );
	bb.CreateStore( bb.ConstInt8(bf_seq), bb.return_path );
	onfail_block->add_cvt( bf_error, bf_seq );
	return onfail_block;
}

void gen::build_instruction( ir::instruction const & i, function_builder & bb )
{
	//DEBUG( "INSTRUCTION: ", ir::dump::get(i) );
	switch( i.type )
	{
		#define IR_INST(type) \
			case ir::instruction::t_##type: \
				return build_##type##_inst( static_cast<ir::type##_inst const &>(i), bb );
			IR_INSTRUCTIONS
		#undef IR_INST
	}
	
	STATE_CHECK( false );
}

void gen::build_store_inst( ir::store_inst const & si, function_builder & bb )
{
	auto left = obtain_value( si.left, bb );
	auto right = obtain_value( si.right, bb );
#if 0
#if SHOW_DEBUG
	dump( *module );
	dumps( *left );
	dumps( *right );
#endif
#endif
	
	//assigning to a shared pointer assigns to its value
	if( si.left.result_type()->is_shared_pointer() ) {
		//DEBUG( "LEFT-SHARED" );
		left = bb.CreateStructGEP( left, sf_data );
	}
		
	DEBUG( "store", get_type(left->getType()), " <- ", get_type(right->getType()) );
	bb.CreateStore( right, left );
}

void gen::build_funccall_inst( ir::funccall_inst const & fi, function_builder & bb )
{
	auto ir_call_type = fi.call.result_type();
	auto func_type = types->function( *ir_call_type );
	auto call = obtain_value( fi.call, bb );
	
	std::vector<llvm::Value*> args;

	//DEBUG( "CALL", ir::dump::get(ir_call_type) );

	//deal with the context-style call (after return arg)
	if( func_type.is_context ) {
		if( fi.context ) {
			//explicit context provided
			auto val = obtain_value( *fi.context, bb ); //TODO: must be a ptr!
			args.push_back( bb.CreateBitCast( val, types->generic_ptr() ) );
		} else {
			auto fptr = bb.CreateStructGEP( call, sf_data, "fptr" );
			auto gptr = bb.CreateStructLoad( fptr, cf_ctx_ptr, "ctx_gptr" );
			args.push_back( bb.CreateBitCast( gptr, types->generic_ptr(), "ctx_ptr" ) );
			call = bb.CreateStructLoad( fptr, cf_func_ptr, "func_ptr" );
		} 
	} else {
		STATE_CHECK( !fi.context );
	}
	//then add args
	for( size_t i=0; i < fi.args.size(); ++i ) {
		auto ir_arg_type = fi.args[i].result_type();
		//DEBUG( "PUSH", ir::dump::get( ir_arg_type ) );
		auto p = obtain_value( fi.args[i], bb );
		if (ir_call_type->args[i]->fundamental == ir::type::f_abi_var_args) {
			STATE_CHECK( ir_arg_type->fundamental == ir::type::f_tuple );
			for( size_t j=0; j < ir_arg_type->sub.size(); ++j ) {
				auto q = bb.CreateExtractValue( p, j );
				args.push_back( q );
			}
		} else {
			args.push_back( p );
		}
	}

	llvm::Value * ret_ptr; //allow optimizations if using a return pointer
	auto rv = bb.call( func_type, call, args, &ret_ptr );
		
	//map return values
	for( size_t i=0; i < fi.returns.size(); ++i ) {	
		auto & r = fi.returns[i];
		STATE_CHECK( r.is_direct() ); //returns should be value_defn, not value_ref
		
		llvm::Value * v = nullptr;
		llvm::Value * vp = nullptr;
		if( fi.returns.size() == 1 ) {
			if( r.origin->need_pointer && ret_ptr ) {
				vp = ret_ptr;
			} else {
				v = rv;
			}
		} else {
			if( ret_ptr ) {
				vp = bb.CreateStructGEP( ret_ptr, unsigned(i) );
				if( !r.origin->need_pointer ) {
					v = bb.CreateLoad( vp );
				}
			} else {
				v = bb.CreateExtractValue( rv, i );
			}
		}
			
		if( r.origin->need_pointer ) {
			if( !vp ) {
				vp = bb.CreateAlloca( types->value(*r.origin->direct_type).ll_type );
				bb.CreateStore( v, vp );
			}
			create_symbol_cref( r.origin->cref, vp, symbol_cref::rt_pointer );
			
		} else {
			create_symbol_cref( r.origin->cref, v, symbol_cref::rt_value );
		}
	}
}

void gen::build_pack_tuple_inst( ir::pack_tuple_inst const & bt, function_builder & bb )
{
	if( !bt.result.origin->need_pointer ) {
		llvm::Value * cur = llvm::UndefValue::get( types->value( *bt.result.origin->direct_type ).ll_type );
	
		for( size_t i=0; i < bt.sub.size(); ++i ) {
			auto part = obtain_value( bt.sub[i], bb );
			cur = bb.CreateInsertValue( cur, part, i );
		}
	
		create_symbol_cref( bt.result.origin->cref, cur, symbol_cref::rt_value );
	} else {
		llvm::Value * cur = bb.CreateAlloca( types->value( *bt.result.origin->direct_type ).ll_type );
	
		for( size_t i=0; i < bt.sub.size(); ++i ) {
			auto part = obtain_value( bt.sub[i], bb );
			bb.CreateStructStore( cur, part, i );
		}
	
		create_symbol_cref( bt.result.origin->cref, cur, symbol_cref::rt_pointer );
	}
}

void gen::build_resolve_inst( ir::resolve_inst const & ei, function_builder & bb )
{
	auto source = obtain_value( ei.source, bb );
	create_symbol_cref( ei.result.origin->cref, source, symbol_cref::rt_value );	
}

void gen::build_struct_pointer_inst( ir::struct_pointer_inst const & sp, function_builder & bb )
{
	auto source = obtain_value( sp.struct_ptr, bb );
	auto sub = bb.CreateStructGEP( source, unsigned(sp.index) );
	create_symbol_cref( sp.result.origin->cref, sub, symbol_cref::rt_value );
}

void gen::build_malloc_inst( ir::malloc_inst const & mi, function_builder & bb ) {
	auto obj_type = types->value( *mi.direct_type );
	auto cast_ptr = heap_alloc_type( obj_type.ll_type, bb );

	create_symbol_cref( mi.result.origin->cref, cast_ptr, symbol_cref::rt_value );
}

void gen::build_alloc_uninit_inst( ir::alloc_uninit_inst const & mi, function_builder & bb ) {
	auto ret = bb.CreateAlloca( types->value( *mi.direct_type ).ll_type );
	
	ret->setMetadata( "src.inst", llvm::MDNode::get( bb.get_llvm_context(),
		llvm::MDString::get( bb.get_llvm_context(), "alloc_uninit" ) ) );
	
	create_symbol_cref( mi.result.origin->cref, ret, symbol_cref::rt_pointer );
}

/**
	There are two types of wrapper functions:
	1. The value being wrapped is a constant function value. In this case a specific wrapper is
	built which calls the target function.
	2. The value is a runtime variable. This pointer is used as the context and a generic runtime
	function is created which routes to the target function.
*/
void gen::build_bind_raw_func_inst( ir::bind_raw_func_inst const & wi, function_builder & bb ) {
	auto raw = obtain_value( wi.raw_func, bb );
	auto ctx_fp_type = types->shared_context_struct( *wi.result.origin->direct_type ).ll_type;
	auto ctx_func_type = types->function( *wi.result.origin->direct_type );
	
	auto lf = dynamic_cast<llvm::Function*>( raw );
	llvm::Function * wrap_func;
	
	static int name_count = 0;
	char buf_name[32];
	sprintf( buf_name, "_%d", name_count++ );
	std::string name = "_ctx_wrap_" + raw->getName().str() + buf_name;
	
	//OPT: generated name on type so they can be reused (optimization)
	bool need_func = true; //I was hoping for runtime we could reuse, but it is still signature specific!
	wrap_func = llvm::cast<llvm::Function>( module->getOrInsertFunction( 
		name, ctx_func_type.ll_func_type ) );
	
	if( need_func ) {
		if( !wrap_func->size() )
		{
			function_builder builder( *module, *wrap_func, *types, gopts );
	
			auto in_args = wrap_func->arg_begin();
			std::vector<llvm::Value*> args;
			
			//copy returns pointer (comes prior to context)
			if( !ctx_func_type.is_direct_return && ctx_func_type.ll_logical_return_type ) {
				args.push_back( &*(in_args++) );
			}
			auto ctx = &*(in_args++);
			while( in_args != wrap_func->arg_end() )
				args.push_back( &*(in_args++) );
			
			llvm::Value * result;
			if( lf ) {
				result = builder.direct_call( raw, args );
			} else {
				//the context ptr is actually the function ptr
				auto pptr = builder.CreateBitCast( ctx, raw->getType() );
				result = builder.direct_call( pptr, args );
			}
			if( result->getType()->isVoidTy() )
				builder.func_return_void();
			else
				builder.func_return( result );
		}
	}
	
	//build ctx func ptr
	auto fp = heap_alloc_type( ctx_fp_type, bb );
	auto fptr = init_shared( fp, nullptr, bb );

	bb.CreateStructStore( fptr, wrap_func, cf_func_ptr, "func_ptr" );
	if( !lf ) {
		//use function pointer as the context
		bb.CreateStructStore( fptr, bb.CreateBitCast( raw, types->generic_ptr() ), cf_ctx_ptr, "ctx_func" );
	}
	
	create_symbol_cref( wi.result.origin->cref, fp, symbol_cref::rt_value );
}

/**
	Build the function which takes a pointer to a shared object and releases it.
*/
llvm::Function * gen::get_release_deref_shared() {
	char const * fname = "_release_deref_shared";
	auto func = module->getFunction( fname );
	if( func ) {
		return func;
	}
	
	func = llvm::cast<llvm::Function>(module->getOrInsertFunction( fname,
		types->type_void(), types->ptr_to( types->generic_ctx_frag() ), (llvm::Type*)0));
	func->setLinkage(llvm::GlobalValue::LinkOnceODRLinkage);
	
	function_builder builder( *module, *func, *types, gopts );
	auto fptr = &*(func->arg_begin());
	
	//input is a pointer to shared ptr
	auto ctx = builder.CreateStructGEP( fptr, cf_data );
	auto bptr = builder.CreateBitCast( ctx, types->ptr_to(types->generic_shared_ptr()) );
	auto sptr = builder.CreateLoad( bptr );
	build_release_shared( sptr, true, builder );
	
	builder.func_return_void();
	builder.complete_blocks();
	return func;
}

/**
	The context destructor forwards the destruction to the contained data.  This is done so that
	the context's can be treated as normal shared objects (no special layout).
*/
llvm::Function * gen::get_context_dtor() {
	char const * fname = "_release_context";
	auto func = module->getFunction( fname );
	if( func ) {
		return func;
	}
	
	func = llvm::cast<llvm::Function>(module->getOrInsertFunction( fname,
		types->type_void(), types->ptr_to( types->generic_ctx_frag() ), (llvm::Type*)0));
	func->setLinkage(llvm::GlobalValue::LinkOnceODRLinkage);
	auto fptr = &*(func->arg_begin());
	
	function_builder builder( *module, *func, *types, gopts );
	
	auto dtor = builder.CreateStructLoad( fptr, cf_data_destructor );
	auto ctx = builder.CreateStructGEP( fptr, cf_data );
	builder.Call(dtor, builder.CreateBitCast( ctx, types->generic_ptr() ) );
	
	builder.func_return_void();
	builder.complete_blocks();
	return func;
}

void gen::build_bind_func_explicit_inst( ir::bind_func_explicit_inst const & ci, function_builder & bb ) {
	auto ctx_func = obtain_value( ci.ctx_func, bb );
	auto ctx_fp_type = types->shared_context_struct( *ci.result.origin->direct_type,
		nullptr, ci.data_shared ? types->generic_ptr() : nullptr ).ll_type;
	
	auto orig_data = obtain_value( ci.ctx_data, bb ); 
	auto ctx_data = orig_data;
	if( ci.data_shared && !ci.ctx_shared ) {
 		ctx_data = bb.CreateStructGEP( ctx_data, sf_data );
 	}
 	STATE_CHECK( ctx_data->getType()->isPointerTy() );
	auto fp = heap_alloc_type( ctx_fp_type, bb );
	auto fptr = init_shared( fp, ci.data_shared ?
		get_release_deref_shared() : nullptr, bb );

	bb.CreateStructStore( fptr, ctx_func, cf_func_ptr, "func_ptr" );
	bb.CreateStructStore( fptr, bb.CreateBitCast( ctx_data, types->generic_ptr() ), cf_ctx_ptr, "ctx_data" );
	if( ci.data_shared ) {
		bb.CreateStructStore( fptr, bb.CreateBitCast( orig_data, types->generic_ptr() ), cf_data, "ctx_offset" );
	}
	
	//expecting more generic type for this value
	auto gt = types->shared_context_struct( *ci.result.origin->direct_type ).ll_type;
	auto gp = bb.CreateBitCast( fp, types->ptr_to(gt) );
	create_symbol_cref( ci.result.origin->cref, gp, symbol_cref::rt_value );
}

/**
	The standard context builder puts the context data at the end of the same structure. There is
	still a pointer to this data inside the structure since this other style is also required (to bind
	existing contexts to functions).
*/
void gen::build_bind_func_inst( ir::bind_func_inst const & ci, function_builder & bb )
{
	auto ctx_func = obtain_value( ci.ctx_func, bb );
	auto ctx_fp_type = types->shared_context_struct( *ci.result.origin->direct_type,
		ci.ctx_data.direct_type.get() ).ll_type;

	auto fp = heap_alloc_type( ctx_fp_type, bb );

	auto fptr = init_shared( fp, 
		ci.ctx_data.static_destructor ? nullptr : get_context_dtor(), bb );

	bb.CreateStructStore( fptr, ctx_func, cf_func_ptr );
	
	if( !ci.ctx_data.static_destructor ) {
		auto dtor = obtain_value( ci.ctx_data.destructor, bb );
		dtor = bb.CreateBitCast( dtor, types->ptr_to( types->destructor() ) );
		bb.CreateStructStore( fptr, dtor, cf_data_destructor );
	}
	
	auto ctx_ptr = bb.CreateStructGEP( fptr, cf_data );
	bb.CreateStructStore( fptr, bb.CreateBitCast( ctx_ptr, types->generic_ptr() ), cf_ctx_ptr );
	
	create_symbol_cref( ci.ctx_data.result.origin->cref, ctx_ptr, symbol_cref::rt_value );
	
	//expecting more generic type for this value
	auto gt = types->shared_context_struct( *ci.result.origin->direct_type ).ll_type;
	auto gp = bb.CreateBitCast( fp, types->ptr_to(gt) );
	create_symbol_cref( ci.result.origin->cref, gp, symbol_cref::rt_value );
}

llvm::Value* gen::init_shared( llvm::Value * sptr, llvm::Value * dtor, function_builder & bb ) {
	bb.CreateStructStore( sptr, bb.const_natural_int( 1 ), sf_ref_count );
	if( !dtor ) {
		dtor = module->getFunction( "_void_destructor" );
		STATE_CHECK( dtor );
	}

	bb.CreateStructStore( sptr, 	bb.CreateBitCast( dtor, types->ptr_to( types->destructor() ) ), 
		sf_destructor, "dtor" );
	
	return bb.CreateStructGEP( sptr, sf_data, "sf_data" );
}

void gen::build_void_context_inst( ir::void_context_inst const & vc, function_builder & bb )
{
	auto ctx_fp_type = llvm::cast<llvm::StructType>(
		types->shared_context_struct( *vc.result.origin->direct_type ).ll_type );
	
	auto fp = heap_alloc_type( ctx_fp_type, bb );
	auto dptr = init_shared( fp, nullptr, bb );
	
// 	bb.CreateStructStore( dptr, 
// 		llvm::ConstantPointerNull::get( llvm::cast<llvm::PointerType>(
// 			dptr->getType()->getElementType( cf_func_ptr ) )), cf_func_ptr ); //TODO: A function which throws
	bb.CreateStructStore( dptr, module->getFunction( "_void_destructor" ), cf_data_destructor );
	
	create_symbol_cref( vc.result.origin->cref, fp, symbol_cref::rt_value );
}

void gen::build_alloc_shared_inst( ir::alloc_shared_inst const & as, function_builder & bb )
{
	auto wrap = types->shared_struct( *as.sod.result.origin->direct_type ).ll_type;
	
	auto fp = heap_alloc_type( wrap, bb );
	bb.CreateStructStore( fp, bb.const_natural_int( 1 ), sf_ref_count );
	
	if( as.sod.static_destructor )
	{
		//TODO: should actually be a pure-throwing function this should never be called
		bb.CreateStructStore( fp, module->getFunction( "_void_destructor" ), sf_destructor );
	}
	else
	{
		auto dtor = obtain_value( as.sod.destructor, bb );
		dtor = bb.CreateBitCast( dtor, types->ptr_to( types->destructor() ) );
		bb.CreateStructStore( fp, dtor, sf_destructor );
	}
	
	create_symbol_cref( as.sod.result.origin->cref, fp, symbol_cref::rt_value );
}

void gen::build_release_shared( llvm::Value * object, bool call_dtor, function_builder & bb )
{
	//dump( *module );
	auto rc = bb.CreateStructGEP( object, sf_ref_count, "ref_count" );
	auto old_count = bb.CreateAtomicRMW( llvm::AtomicRMWInst::Sub, rc, 
		bb.const_natural_int(1), llvm::Acquire );
		
	//debugging check for duplicate release of shared
	{
		auto cnd = bb.CreateICmpSLT( old_count, bb.const_natural_int(1) );
		auto fail_block = bb.create_child_block(); 
		auto post_block = bb.create_child_block();
		bb.branch_conditional( cnd, fail_block, post_block );
		
		bb.set_insert_point( fail_block );
		auto func = llvm::cast<llvm::Function>( module->getFunction( "_duplicate_release_shared" ) );
		STATE_CHECK( func );
		bb.direct_call( func );
		bb.branch_sequence( post_block );
		
		bb.set_insert_point( post_block );
	}
	auto release_block = bb.create_child_block();
	auto post_block = bb.create_child_block();
	auto cnd = bb.CreateICmpEQ( old_count, bb.const_natural_int( 1 ) );
	bb.branch_conditional( cnd, release_block, post_block );
	
	bb.set_insert_point( release_block );
	
	if( call_dtor ) {
		std::vector<llvm::Value*> args;
		args.push_back( bb.CreateBitCast( bb.CreateStructGEP( object, sf_data, "object" ), types->generic_ptr() ) );
		auto destructor = bb.CreateStructLoad( object, sf_destructor, "dtor" );
		bb.Call( destructor, args );
	}
	
	gen_free( object, bb );
	bb.branch_sequence( post_block );
	
	bb.set_insert_point( post_block );
}

void gen::build_release_shared_inst( ir::release_shared_inst const & rs, function_builder & bb ) {
	auto object = obtain_value( rs.object, bb );
	bool call_dtor = rs.destroy_how == ir::release_shared_inst::dh_dynamic;
	build_release_shared( object, call_dtor, bb );
}

void gen::build_acquire_shared( llvm::Value * object, function_builder & bb ) {
	auto rc = bb.CreateStructGEP( object, sf_ref_count, "ref_count" );
	bb.CreateAtomicRMW( llvm::AtomicRMWInst::Add, rc, bb.const_natural_int(1), llvm::Release );
}

void gen::build_acquire_shared_inst( ir::acquire_shared_inst const & as, function_builder & bb ) {
	auto object = obtain_value( as.object, bb );
	build_acquire_shared( object, bb );
}

void gen::build_switch_inst( ir::switch_inst const & si, function_builder & bb )
{
	auto start_block = bb.get_insert_block();
	
	//setup final Phi
	auto end_block = bb.create_child_block( "switch_end" );
	llvm::PHINode * phi = nullptr;
	if( si.with_result )
	{
		bb.set_insert_point(end_block);
		auto t = types->value(*si.result.origin->direct_type).ll_type;
		phi = bb.create_phi( t, unsigned(si.blocks.size()) );
		bb.set_insert_point(start_block);
	}
	
	auto cond_block = start_block;
	for( size_t i=0; i < si.blocks.size(); ++i )
	{
		bool last = (i == (si.blocks.size()-1));
		
		auto & cb = *si.blocks[i];
		STATE_CHECK( cb.cond_value || last );
		STATE_CHECK( !last || !si.with_result || !cb.cond_value );
		
		bb.set_insert_point( cond_block );
		auto then_block = bb.create_child_block( "switch_then" );
		auto else_block = last ? end_block : bb.create_child_block( "switch_else" );
		
		build_block( *cb.cond_block, bb );
		
		if( !cb.cond_value ) {
			bb.branch_sequence( then_block );
		} else {
			auto cond = obtain_value( *cb.cond_value, bb );
			bb.branch_conditional( cond, then_block, else_block );
		}
	
		bb.set_insert_point( then_block );
		build_block( *cb.expr_block, bb );
		auto phi_block = bb.get_insert_block();
		
		llvm::Value * then_val = nullptr;
		if( si.with_result )
			then_val = obtain_value( cb.expr_value, bb );
		bb.branch_sequence( end_block );
		
		if( si.with_result )
			phi->addIncoming( then_val, phi_block->ll_block );
		cond_block = else_block;
	}
	
	if( si.with_result )
		create_symbol_cref( si.result.origin->cref, phi, symbol_cref::rt_value );

	bb.set_insert_point( end_block );
}

void gen::build_loop_inst( ir::loop_inst const & lo, function_builder & bb )
{
	//in current context
	auto end_block = bb.create_child_block( "loop_end" );
	//in new context
	auto cond_block = bb.create_child_context( "loop_cond" );
	bb.branch_sequence( cond_block );
	bb.set_insert_point( cond_block );
	bb.get_context()->flow[bf_seq] = end_block;
	build_block( *lo.cond_block, bb );
	
	auto body_block = bb.create_child_block( "loop_body" );
	auto cond = obtain_value( lo.cond_value, bb );
	bb.branch_conditional( cond, body_block, end_block );
	
	auto loop_next_block = bb.create_child_block( "loop_next" );
	bb.get_context()->flow[bf_loop_next] = loop_next_block;
	loop_next_block->add_cvt( bf_loop_next, bf_seq );
	loop_next_block->block_seq = cond_block;
		
	auto loop_break_block = bb.create_child_block( "loop_break" );
	bb.get_context()->flow[bf_loop_break] = loop_break_block;
	loop_break_block->add_cvt( bf_loop_break, bf_seq );
	loop_break_block->block_seq = end_block;
	
	bb.set_insert_point( body_block );
	build_block( *lo.body_block, bb );
	bb.branch_sequence( cond_block );
		
	bb.set_insert_point( end_block );
}

void gen::build_return_inst( ir::return_inst const & /*r*/, function_builder & bb )
{
	bb.flow_return();
}

void gen::build_alloc_array_inst( ir::alloc_array_inst const & aa, function_builder & bb )
{
	auto len_type = types->f_natural_int();
	auto length_size = types->alloc_size( len_type );
	
	auto item_type = types->value( *aa.item_type ).ll_type;
	auto one_item_size = types->alloc_size( item_type );
	
	auto align = types->max_align();
	STATE_CHECK( align >= length_size );
	
	auto length = bb.create_sign_ext( obtain_value( aa.length, bb ), len_type );
	
	auto items_size = bb.create_mul( llvm::ConstantInt::get( len_type, one_item_size ),
		length );
	auto full_size = bb.create_add( llvm::ConstantInt::get( len_type, align ), items_size );
		
	auto memory = gen_malloc( full_size, bb );
	
	auto length_ptr = bb.CreateBitCast( memory, types->ptr_to( len_type ));
	bb.CreateStore( length, length_ptr );
	
	create_symbol_cref( aa.result.origin->cref, memory, symbol_cref::rt_value );
}

void gen::build_fail_inst( ir::fail_inst const & fi, function_builder & bb ) {
	//create new blocks to keep current flow and new block with fail flow
	auto post_block = bb.create_child_block( "unreachable" );
	auto fail_block = bb.create_child_block( "fail" );
	bb.branch_sequence( fail_block );
	
	bb.set_insert_point( fail_block );
	bb.CreateStore( bb.const_int(types->f_error_code(),fi.level), bb.error_value );
	if( fi.data ) {
		auto v = obtain_value( *fi.data, bb );
		bb.CreateStore( bb.CreateBitCast( v, types->generic_shared_ptr()), bb.error_info );
	} else {
		bb.CreateStore( bb.const_zero(types->generic_shared_ptr()), bb.error_info );
	}
	bb.flow_resume();
	
	bb.set_insert_point( post_block );
}

void gen::build_resume_inst( ir::resume_inst const & ri, function_builder & bb ) {
	auto error_info = obtain_value( *ri.data, bb );
	gen_acquire_optional_shared( error_info, bb );
	bb.CreateStore( bb.CreateBitCast( error_info, types->generic_shared_ptr()), bb.error_info );
	bb.flow_resume();
}


void gen::build_free_array_inst( ir::free_array_inst const & fa, function_builder & bb )
{
	gen_free( obtain_value( fa.array, bb ), bb );
}

void gen::build_special_field_inst( ir::special_field_inst const & sf, function_builder & bb )
{
	switch( sf.field )
	{	
		case ir::special_field_inst::f_array_len:
		{
			auto array = obtain_value( sf.object, bb );
			auto length_ptr = bb.CreateBitCast( array, types->ptr_to( types->f_natural_int() ));
			auto length = bb.CreateLoad( length_ptr );
			create_symbol_cref( sf.result.origin->cref, length, symbol_cref::rt_value );
			return;
		}
	}
	
	STATE_FAIL( "invalid field" );
}

void gen::build_array_offset_inst( ir::array_offset_inst const & ao, function_builder & bb )
{
	auto offset = obtain_value( ao.offset, bb );
	auto base = obtain_value( ao.array, bb );
	
	auto array_type = ao.array.result_type();
	bool is_pointer = false;
	if( array_type->is_raw_pointer() ) {
		array_type = array_type->sub[0];
		is_pointer = true;
		base = bb.CreateLoad( base );
	}
	
	//TODO: seems quite similar to the bits in build_alloc_array_inst
	auto len_type = types->f_natural_int();
	auto item_type = types->value( *array_type->sub[0] ).ll_type;
	auto one_item_size = types->alloc_size( item_type );
	auto align = types->max_align();
	
	auto item_offset = bb.create_mul( llvm::ConstantInt::get( len_type, one_item_size ), 
		bb.create_sign_ext( offset, len_type ) );
	//data starts after the size field (of align size)
	item_offset = bb.create_add( llvm::ConstantInt::get( len_type, align ), item_offset );
	
	auto raw_ptr = bb.create_inbounds_gep( base, item_offset );
	auto item_ptr = bb.CreateBitCast( raw_ptr, types->ptr_to( item_type ) );
	
	if( is_pointer ) {
		create_symbol_cref( ao.result.origin->cref, item_ptr, symbol_cref::rt_value );
	} else {
		create_symbol_cref( ao.result.origin->cref, bb.CreateLoad( item_ptr ), symbol_cref::rt_value );
	}
}

void gen::build_array_ptr_offset_inst( ir::array_ptr_offset_inst const & ao, function_builder & bb )
{
	auto offset = obtain_value( ao.offset, bb );
	auto base = obtain_value( ao.array, bb );
	
	auto array_type = ao.array.result_type();
	if( ao.lvalue ) {
		array_type = array_type->sub[0];
		base = bb.CreateLoad(base);
	}
	
	auto len_type = types->f_natural_int();
	auto item_type = types->value( *array_type->sub[0] ).ll_type;
	auto one_item_size = types->alloc_size( item_type );
	
	auto item_offset = bb.create_mul( llvm::ConstantInt::get( len_type, one_item_size ), 
		bb.create_sign_ext( offset, len_type ) );
	
	//since we calculate offset the base must be a generic pointer
	base = bb.CreateBitCast( base, types->generic_ptr() );
	auto raw_ptr = bb.create_inbounds_gep( base, item_offset );
	auto item_ptr = bb.CreateBitCast( raw_ptr, types->ptr_to( item_type ) );
	
	if( ao.lvalue ) {
		create_symbol_cref( ao.result.origin->cref, item_ptr, symbol_cref::rt_value );
	} else {
		create_symbol_cref( ao.result.origin->cref, bb.CreateLoad( item_ptr ), symbol_cref::rt_value );
	}
}

void gen::build_promote_inst( ir::promote_inst const & pi, function_builder & bb )
{
	auto source = obtain_value( pi.source, bb );
	
	#define TO_FROM( t, f ) (pi.to_what == ir::promote_inst::t && pi.from_what == ir::promote_inst::f)
	llvm::Value * promote;
	if( TO_FROM( w_integer, w_integer ) ) {
		auto target_type = types->f_intx( pi.to_bitsize );
		promote = bb.create_sign_ext( source, target_type );
	} else if( TO_FROM( w_binary, w_binary ) 
		|| TO_FROM( w_integer, w_binary ) ) {
		auto target_type = types->f_intx( pi.to_bitsize );
		promote = bb.create_zero_ext( source, target_type );
	} else if( TO_FROM( w_float, w_float ) ) {
		auto target_type = types->f_float( pi.to_bitsize );
		promote = bb.create_fp_ext( source, target_type );
	} else if( TO_FROM( w_float, w_integer ) ) {
		auto target_type = types->f_float( pi.to_bitsize );
		promote = bb.create_si_to_fp( source, target_type );
	} else if( TO_FROM( w_float, w_binary ) ) {
		auto target_type = types->f_float( pi.to_bitsize );
		promote = bb.create_ui_to_fp( source, target_type );
	} else {
		STATE_FAIL();
	}
	#undef TO_FROM	
	create_symbol_cref( pi.result.origin->cref, promote, symbol_cref::rt_value );
}

void gen::build_loss_inst( ir::loss_inst const & li, function_builder & bb )
{
	auto source = obtain_value( li.source, bb );
	
	#define TO_FROM( t, f ) (li.to_what == ir::promote_inst::t && li.from_what == ir::promote_inst::f)
	llvm::Value * loss;
	if( TO_FROM( w_integer, w_integer ) ||
		TO_FROM( w_binary, w_integer ) || 
		TO_FROM( w_integer, w_binary ) ||
		TO_FROM( w_binary, w_binary )) {
		auto target_type = types->f_intx( li.to_bitsize );
		loss = bb.create_sext_or_trunc( source, target_type );
	} else if( TO_FROM( w_integer, w_float ) ) {
		auto target_type = types->f_intx( li.to_bitsize );
		loss = bb.create_fp_to_si( source, target_type );
	} else if( TO_FROM( w_float, w_float ) ) {
		auto target_type = types->f_float( li.to_bitsize );
		loss = bb.create_fp_trunc( source, target_type );
	} else if( TO_FROM( w_float, w_integer ) ) {
		auto target_type = types->f_float( li.to_bitsize );
		loss = bb.create_si_to_fp( source, target_type );
	} else if( TO_FROM( w_float, w_binary ) ) {
		auto target_type = types->f_float( li.to_bitsize );
		loss = bb.create_ui_to_fp( source, target_type );
	} else if( TO_FROM( w_binary, w_float ) ) {
		auto target_type = types->f_intx( li.to_bitsize );
		loss = bb.create_fp_to_ui( source, target_type );
	} else {
		STATE_FAIL();
	}
	#undef TO_FROM
	create_symbol_cref( li.result.origin->cref, loss, symbol_cref::rt_value );
}

void gen::build_cast_inst( ir::cast_inst const & ci, function_builder & bb )
{
	auto source = obtain_value( ci.source, bb );
	auto src_type = source->getType();
	auto dst_type = types->value( *ci.result.result_type() ).ll_type;
	DEBUG( "cast", get_type( src_type ), get_type(dst_type) );
	llvm::Value * ret;
	
	if( src_type->isPointerTy() && dst_type->isIntegerTy() ) {
		ret = bb.create_ptr_to_int( source, dst_type );
	} else {
		ret = bb.CreateBitCast( source, dst_type );
	}
	create_symbol_cref( ci.result.origin->cref, ret, symbol_cref::rt_value );
}

void gen::build_loop_flow_inst( ir::loop_flow_inst const & lfi, function_builder & bb )
{
	switch( lfi.flow ) {
		case ir::loop_flow_inst::f_break: 
			bb.flow_loop_break();
			break;
		case ir::loop_flow_inst::f_next:
			bb.flow_loop_next();
			break;
	}
}

void gen::build_binary_inst( ir::binary_inst const & bi, function_builder & bb )
{
	auto a = obtain_value( bi.a, bb );
	auto b = obtain_value( bi.b, bb );
	
	auto rt = bi.a.result_type();
	bool sign = !rt->is_binary() && !rt->is_boolean();
	
	llvm::Value * result = nullptr;
	switch( bi.op ) {
		#define SIMPLE(t) \
			case ir::binary_inst::o_##t: \
				result = bb.create_##t( a, b, sign ); \
				break;
		SIMPLE(add)
		SIMPLE(sub)
		SIMPLE(mul)
		SIMPLE(div)
		
		SIMPLE(mod)
		
		SIMPLE(bit_shl)
		SIMPLE(bit_shr)
		
		SIMPLE(or)
		SIMPLE(and)
		SIMPLE(xor)
		SIMPLE(and_not)
		SIMPLE(nor)
		SIMPLE(nand)
		
		SIMPLE(eq)
		SIMPLE(neq)
		SIMPLE(lt)
		SIMPLE(lte)
		SIMPLE(gt)
		SIMPLE(gte)
		#undef SIMPLE
	}
	STATE_CHECK(result);
	
	create_symbol_cref( bi.result.origin->cref, result, symbol_cref::rt_value );
}

void gen::build_comment_inst( ir::comment_inst const &, function_builder & )
{
	//no effect
}

void gen::build_clone_inst( ir::clone_inst const & ci, function_builder & bb )
{
	auto source = obtain_value( ci.source, bb );
	auto temp = bb.CreateAlloca( types->value( *ci.result.result_type()->sub[0] ).ll_type );
	bb.CreateStore( ci.source_value ? source : bb.CreateLoad(source), temp );
	create_symbol_cref( ci.result.origin->cref, temp, symbol_cref::rt_value ); //not a pointer since the result is the value of the pointer
}

void gen::build_unary_inst( ir::unary_inst const & un, function_builder & bb )
{
	auto arg = obtain_value( un.arg, bb );
	
	llvm::Value * result = nullptr;
	switch( un.op ) {
		#define SIMPLE(t) \
			case ir::unary_inst::o_##t: \
				result = bb.create_##t( arg ); \
				break;
		SIMPLE(neg)
		SIMPLE(abs)
		
		SIMPLE(not)
		#undef SIMPLE
	}
	
	STATE_CHECK(result);
	create_symbol_cref( un.result.origin->cref, result, symbol_cref::rt_value );
}

llvm::Value* gen::obtain_value( ir::value_ref const & vr_, function_builder & bb )
{
	typedef boost::error_info<struct value_name,std::string> ex_value_name;
	ir::value_ref vr = vr_;
	ir::value const * v = vr.origin.get();
	
	llvm::Value * ret = nullptr;
	auto std_proc = [&]( llvm::Value * raw, symbol_cref::ref_type_t rt ) {
		ret = raw;
		auto ret_type = v->direct_type;
		
		bool is_ptr = false;
		switch( rt )
		{
			case symbol_cref::rt_value:
				break;
				
			case symbol_cref::rt_pointer:
				if( vr.how != vr.h_raw_pointer ) {
					ret = bb.CreateLoad( ret );
				} else {
					is_ptr = true;
				}
				break;
				
			case symbol_cref::rt_global:
				ret = bb.create_gep( ret, 0 );
				if( vr.how != vr.h_raw_pointer ) {
					ret = bb.CreateLoad( ret );
				} else {
					is_ptr = true;
				}
				break;
		}
		
		//resolve pre_field
		if( vr.pre_field ) 	{
			if( ret_type->is_raw_pointer() || is_ptr ) {
				ret = bb.CreateStructGEP( ret, unsigned(*vr.pre_field) );
			} else {
				ret  = bb.CreateExtractValue( ret, *vr.pre_field );
			}
		}
		
		if( vr.how == vr.h_raw_pointer ) {
			STATE_CHECK( is_ptr, "cannot get pointer to value" );
		} else if( vr.how == vr.h_deref ) {
			ret = bb.CreateLoad( ret );
		} else if( vr.how == vr.h_deref_shared ) {
			ret = bb.CreateStructLoad( ret, sf_data );
		} else if( vr.how == vr.h_raw_shared ) {
			ret = bb.CreateStructGEP( ret, sf_data );
		} else if( vr.how == vr.h_raw_array ) {
			STATE_CHECK( ret_type->is_array() );
			auto base = bb.create_inbounds_gep( ret,
				llvm::ConstantInt::get( types->f_natural_int(), types->max_align() ) );
			ret = bb.CreateBitCast( base, types->ptr_to(	
				types->value(*ret_type->item_type()).ll_type ) );
		} else /*h_value*/ {
		}
	};
	
	try {
		switch( v->form ) {
			case ir::value::f_const_boolean:
				ret = llvm::ConstantInt::get( llvm::Type::getInt1Ty( *llvm_context ), v->boolean_value );
				break;
			case ir::value::f_const_integer:
				//TODO: > int64 constant value
				ret = llvm::ConstantInt::get( types->value( *v->direct_type ).ll_type, v->int_value, true );
				break;
			case ir::value::f_const_binary:
				//TODO: > uint64 constant value
				ret = llvm::ConstantInt::get( types->value( *v->direct_type ).ll_type, v->uint_value, false );
				break;
			case ir::value::f_const_float:
				//TODO: What about 80/128 bit constants?!
				ret = llvm::ConstantFP::get( types->value( *v->direct_type ).ll_type, double(v->float_value) );
				break;
			case ir::value::f_nullptr:
				ret = llvm::ConstantPointerNull::get( llvm::cast<llvm::PointerType>(
					types->value( *v->direct_type ).ll_type) );
				break;
			case ir::value::f_undefined:
				ret = llvm::UndefValue::get( types->value( *v->direct_type ).ll_type );
				break;
				
			case ir::value::f_global:
			case ir::value::f_local:
			case ir::value::f_temp:
			case ir::value::f_return_ref:
			case ir::value::f_arg_ref:
			case ir::value::f_capture_ref: {
				DEBUG( "VALUE", v->cref, v );
				STATE_CHECK( v->cref, std::string("uninitialized-value: ") + v->name );
				auto sc = reinterpret_cast<symbol_cref*>( v->cref );
				std_proc( sc->value, sc->ref_type );
				break;
			}
				
			case ir::value::f_import_code: {
				auto type = types->function(*v->direct_type ).ll_func_type;
				auto func = llvm::cast<llvm::Function>( module->getOrInsertFunction( v->name,
					type ) );
				std_proc( func, symbol_cref::rt_value );
				break;
			}
			
			case ir::value::f_import_data: {
				auto gv = module->getGlobalVariable( v->name );
				if( !gv ) {
					auto type = types->value( *v->direct_type ).ll_type;
					gv = new llvm::GlobalVariable( *module, type, /*constant*/false,
						llvm::GlobalValue::ExternalLinkage, /*init*/0, v->name );
				}
				std_proc( gv, symbol_cref::rt_global );
				break;
			}
			
			case ir::value::f_raw_binary: {
				auto arr = llvm::ConstantDataArray::get( *llvm_context, llvm::ArrayRef<uint8_t>( v->raw_binary ) );
				auto global = new llvm::GlobalVariable( *module, arr->getType(), /*constant*/true,
					llvm::GlobalValue::InternalLinkage, /*init*/arr );
				std_proc( global, symbol_cref::rt_global );
				break;
			}
		}
		
		STATE_CHECK( ret );
		return ret;
	} catch( boost::exception & be ) {
		be << ex_value_name( v->name );
		throw;
	}
}

llvm::ExecutionEngine* gen::create_engine()
{
	llvm::TargetOptions topts;
	
	if( gopts.llvm_opt )
	{
		llvm::legacy::PassManager fpm;
		//MIGRATE: fpm.add( new llvm::DataLayout( ee->getDataLayout() ) );
		//MIGRATE: fpm.add( llvm::createBasicAliasAnalysisPass() );
		fpm.add( llvm::createPromoteMemoryToRegisterPass());
		fpm.add( llvm::createInstructionCombiningPass());
		fpm.add( llvm::createAAEvalPass() );
		
		fpm.add( llvm::createReassociatePass());
		fpm.add( llvm::createGVNPass());
		fpm.add( llvm::createConstantPropagationPass() );
		fpm.add( llvm::createFunctionInliningPass());
		fpm.add( llvm::createPartialInliningPass());
		fpm.add( llvm::createInstructionCombiningPass());
		fpm.add( llvm::createPromoteMemoryToRegisterPass());
		fpm.add( llvm::createTailCallEliminationPass());
		fpm.add( llvm::createCFGSimplificationPass());
		
		fpm.add( llvm::createDeadStoreEliminationPass() );
		fpm.add( llvm::createDeadCodeEliminationPass() );
		fpm.add( llvm::createUnreachableBlockEliminationPass() );
		
		fpm.run( *module );
		
		//print out the IR
		dump( *module );
	}
	
	std::string errStr;
	del_module = false;
	llvm::ExecutionEngine * ee = llvm::EngineBuilder( unique_ptr<llvm::Module>(module) ).
		setErrorStr( &errStr ).
		setEngineKind( llvm::EngineKind::JIT ).
		setTargetOptions( topts ).
		create();

	STATE_CHECK( ee );
	STATE_CHECK( errStr.size() == 0, errStr );
	return ee;
}

void gen::emit_object( ir::module const & m, std::string const & base_path,
	std::string const & sc )
{
	util::validate(m);
	build_module( m );
	build_wrapped_entry( m );
	show_code(sc);

	std::string err_str;
	
	//emit LL for reference (for debugging)
	{
		std::ofstream out( (base_path + ".ll").c_str(), std::ios::binary );
		llvm::raw_os_ostream rout( out );
		module->print( rout, nullptr );
	}
	
	std::string triple_name = llvm::sys::getDefaultTargetTriple();
	llvm::Triple triple(llvm::Triple::normalize(triple_name));
	llvm::Target const *  target = llvm::TargetRegistry::lookupTarget( "x86-64", triple, err_str );
	STATE_CHECK( target, err_str );
	
	llvm::TargetOptions topts;
	llvm::TargetMachine * machine = target->createTargetMachine( triple.getTriple(),
		"" /*cpu*/, "" /*featurestr*/, topts, 
		gopts.pic ? llvm::Reloc::PIC_ : llvm::Reloc::Default, 
		llvm::CodeModel::Default,
		llvm::CodeGenOpt::Default );
	
	std::error_code ec;
	llvm::tool_output_file out( (base_path + ".o").c_str(), ec, llvm::sys::fs::OpenFlags::F_RW );
	STATE_CHECK( !ec );
	
	llvm::legacy::PassManager fpm;
	bool ret = machine->addPassesToEmitFile( fpm, out.os(), llvm::TargetMachine::CGFT_ObjectFile );
	STATE_CHECK( !ret );

	module->setDataLayout( machine->createDataLayout() );
	fpm.run( *module );
	
	out.keep();
}

} //eon ir

