#include <stdint.h>
#include <stdio.h>

//NOTE: to enable you must switch to g++ in leaf.cpp (to get C++ libs)
#define SHOW_DEBUG 0
#include "util/debug.hpp"

//using namespace std;

//#include "llvm/Support/Dwarf.h"
#include "unwind.h"
#include "dwarf_eh.h"

#define C(c, b) ((((uint64_t)c)<<(64 - (8*b))))
uint64_t const leaf_exception_class = 
	C('E',1) + C('S',2) + C('\0',3) + C('\0',4) + 
	C('L',5) + C('e',6) + C('a',7) + C('f',8);
#undef C

void leaf_cleanup_exception( _Unwind_Reason_Code /*reason*/, _Unwind_Exception * exc ) 
{
	DEBUG( "CLEANING" );
	free(exc);
	DEBUG( "DONE CLEANING" );
}

extern "C" void leaf_rethrow( _Unwind_Exception * ue )
{
	DEBUG("rethrow");
	auto code = _Unwind_RaiseException( ue );
	switch( code ) {
		case _URC_END_OF_STACK:
			DEBUG( "No handler found" );
			break;
		case _URC_FATAL_PHASE1_ERROR:
			DEBUG( "Fatal phase1 error" );
			break;
		case _URC_FATAL_PHASE2_ERROR:	
			DEBUG( "Fatal phase2 error" );
			break;
		default:
			DEBUG( "Unexpected unwind error" );
			break;
	}
	abort();
}

struct leaf_exception : public _Unwind_Exception
{
	int level;
};

extern "C" void _raise_exception( int level )
{
	DEBUG( "RAISING", level );
	leaf_exception * ue = static_cast<leaf_exception*>(malloc(sizeof(leaf_exception)));
	ue->exception_cleanup = leaf_cleanup_exception;
	ue->exception_class = leaf_exception_class;
	ue->level = level;

	leaf_rethrow(ue);
}

enum handler_type {
	ht_none,
	ht_cleanup,
	ht_catch,
};

handler_type check_action_record(struct _Unwind_Context *context,
	struct dwarf_eh_lsda *lsda, dw_eh_ptr_t action_record, leaf_exception const * ue )
{
	bool any_record = false;
	
	while(action_record) {
		any_record = true;
		int64_t type_offset = read_sleb128(&action_record);
		dw_eh_ptr_t action_record_offset_base = action_record;
		int64_t displacement = read_sleb128(&action_record);
		DEBUG( "ACTION_RECORD", type_offset, displacement );
		
		if (type_offset > 0) {
			if( !ue ) {
				//foreign exception
				DEBUG( "foreign" );
			} else if( lsda->type_table ) {
				DEBUG( "TYPE", ue->level, intptr_t(lsda->type_table) );
				uint8_t * tt = get_type_table_entry(context, lsda, type_offset);
				DEBUG( "TT", intptr_t(tt) );
				auto level = tt ? *(int*)tt : 0;
				DEBUG( "level:", level, "UE", ue->level );
 				if( !level || ue->level <= level ) {
					return ht_catch;
 				}
			} else {
				DEBUG( "type-less" );
				//rcan this happen?
			}
		} else if( type_offset == 0 ) {
			//these only occur when mixed with catch clauses -- and they appear to be last
			return ht_cleanup;
		} else {
			DEBUG("filters/exception-specs not supported?");
			abort();
		}
		action_record = displacement ? action_record_offset_base + displacement : 0;
	}
	
	//if a landingpad has only a cleanup there will be no records
	return any_record ? ht_none : ht_cleanup;
}

/**
	This has been written by looking at several examples (LLVM mainly) and the ABI standard.
	There are bits I'm still not sure about yet.
*/
extern "C" _Unwind_Reason_Code leaf_exception_personality(int version,
	_Unwind_Action actions, uint64_t exception_class, struct _Unwind_Exception * ue,
	struct _Unwind_Context *context)
{
	DEBUG("HANDLE-CHECK", actions, reinterpret_cast<intptr_t>(ue), ue->exception_class );
	if( version != 1 ) {
		return _URC_FATAL_PHASE1_ERROR;
	}
	
	bool foreign = exception_class != leaf_exception_class;
	
	uint8_t const * lsda_addr = _Unwind_GetLanguageSpecificData(context);
	if( !lsda_addr )
		return _URC_CONTINUE_UNWIND; //unsure on this
		
	dwarf_eh_lsda lsda = parse_lsda(context, (unsigned char*)lsda_addr);
	dwarf_eh_action action = dwarf_eh_find_callsite(context, &lsda );
	handler_type handler = check_action_record( context, &lsda, action.action_record, 
		foreign ? nullptr : static_cast<leaf_exception const*>(ue) );

	DEBUG("FOUND", int(handler), intptr_t(lsda_addr), intptr_t(action.landing_pad) );
	//for now treat as the same. Modelling the block-flow in LLVM for cleanup becomes
	//too tricky for cleanup embedded in a caught section
	if( handler == ht_cleanup ) {
		handler = ht_catch;
	}
	
	if( actions & _UA_SEARCH_PHASE ) {
		if( handler == ht_catch && action.landing_pad )
			return _URC_HANDLER_FOUND;
		return _URC_CONTINUE_UNWIND;
	}
	
	bool exec_catch = (handler == ht_catch) && (actions & _UA_HANDLER_FRAME);
	bool exec_clean = (handler == ht_cleanup) && (actions & _UA_CLEANUP_PHASE);
	DEBUG( "CATCH", int(exec_catch), "CLEAN", int(exec_clean) );
	if( !(exec_catch || exec_clean) || !action.landing_pad ) {
		DEBUG( "CONTINUE_UNWIND" );
		return _URC_CONTINUE_UNWIND;
	}
	
	int level = foreign ? -1 : static_cast<leaf_exception const*>(ue)->level;
	DEBUG("LANDING", level );
	_Unwind_SetIP(context, (unsigned long)action.landing_pad);
	_Unwind_SetGR(context, __builtin_eh_return_data_regno(0), reinterpret_cast<uint64_t>(ue) );
	_Unwind_SetGR(context, __builtin_eh_return_data_regno(1), level );
	return _URC_INSTALL_CONTEXT;
}
