/**
	The native interface used for testing Leaf with the JIT. These are compiled directly into
	the Leaf code to limit depenedencies needed for the basic tests.
*/

#include "ir/llvm/runtime_test.hpp"
#include "util/unicode.hpp"

#include <set>
#include <cstdarg>

namespace {
	std::string trace_str, error_str;
	int num_malloc;
	std::set<void*> track_mallocs;
	jmp_buf * run_jmp_buf;
}

namespace runtime_test {

	void prepare( jmp_buf * abort_point ) {
		trace_str.clear();
		error_str.clear();
		num_malloc = 0;
		run_jmp_buf = abort_point;
	}

	std::string get_trace() {
		return trace_str;
	}

	std::string get_error() {
		return error_str;
	}

	void check_state() {
		STATE_CHECK( num_malloc == 0, "unfreed-memory-remains" );
	}
}

///////////////////////////////////////////////////////////
// Native Interface
void abort_exec( std::string const & err ) {
	error_str = err;
	longjmp( *run_jmp_buf, 1 );
}

template<typename T>
void std_trace( T v ) 
{
	std::ostringstream out;
	out << v;
	trace_str += out.str();
}

//so exe_tests/_libs/sys works without CLI args (will need a way to provide these later)
char ** __leaf_argv = new char*[0];
int __leaf_argc = 0;

extern "C" void trace_integer( int64_t v )
{ std_trace(v); }
extern "C" void trace_binary( uint64_t v )
{ std_trace(v); }
extern "C" void trace_float( double v ) 
{ std_trace(v); }
extern "C" void trace_float_high( long double v ) 
{ std_trace(v); }

extern "C" void trace_boolean( bool c ) {
	trace_str += c ? "true" : "false";
}

extern "C" void trace_char( uint32_t c ) {
	trace_str += unicode::encode8(c);
}

extern "C" void trace_abi_char( char c ) {
	trace_str += c;
}

extern "C" void trace_string_tag( intptr_t t ) {
	uint8_t const * mem = reinterpret_cast<uint8_t const*>(t);
	
	uint16_t len = *(uint16_t const*)mem;
	uint32_t const * str = (uint32_t const*)(mem+2);
	
	for( uint16_t i = 0; i < len; ++i ) {
		trace_str += (char)str[i];
	}
}

/*
	The calls to malloc are counted to ensure the test cases are properly
	freeing the memory.
*/
extern "C" void * count_malloc( int64_t sz )
{
	num_malloc++;
	void * r = malloc( sz );
	track_mallocs.insert( r );
	return r;
}

extern "C" void count_free( void * ptr )
{
	//C-free does nothing on null, and we also rely on that (to avoid an additional conditional)
	if( !ptr )
		return;
		
	auto it = track_mallocs.find( ptr );
	if( it == track_mallocs.end() ) {
		abort_exec( "free-of-unallocated-memory" );
	}
	
	track_mallocs.erase( it );
	num_malloc--;
	//DEBUG: Don't actually free now, leave around to track invalid additioanal release_shared
	//free(ptr);
}

extern "C" void _duplicate_release_shared() {
	abort_exec( "duplicate-release-shared" );
}

extern "C" void _access_unset_optional() {
	abort_exec("access-unset-optional");
}

extern "C" void _unhandled_error( ) {
	abort_exec( "unhandled-error" );
}

extern "C" void _bad_fail() {
	abort_exec( "bad-fail" );
}

extern "C" void _void_destructor( void* ) {
}

/*
	This is for the value_ptr native test.
*/
struct vpt_data {
	int a;
	uint8_t c;
	float b;
};

extern "C" void vpt_test( vpt_data * v, uint8_t val )
{
	std::ostringstream out;
	out << "VPT:";
	out << v->a;
	out << ",";
	out << v->b;
	out << ":";
	trace_str += out.str();
	
	v->c = val;
}

struct vpt_data_byte {
	uint8_t x;
	int a;
	uint8_t c;
	float b;
}  __attribute__((packed));

extern "C" void vpt_test_byte( vpt_data_byte * v, uint8_t val )
{
	std::ostringstream out;
	out << "VPT:";
	out << v->a;
	out << ",";
	out << v->b;
	out << ":";
	trace_str += out.str();
	
	v->c = val;
}

extern "C" vpt_data * vpt_test_get(int a, float b, uint8_t c) {
	static vpt_data vd;
	vd.a = a;
	vd.b = b;
	vd.c = c;
	return &vd;
}

extern "C" void array_swap( int * data, int a, int b ) {
	int c = data[a];
	data[a] = data[b];
	data[b] = c;
}

extern "C" double var_args_add( int count, ... ) {
	double s = 0;
	va_list vl;
	va_start(vl, count);
	for(int i=0; i < count; ++i) {
		s += va_arg(vl,double);
	}
	va_end(vl);
	return s;
}
