#include "ir/type.hpp"
#include "ir/function.hpp"

#include "platform/platform.hpp"

#include "ir/llvm/type_manager.hpp"

//NOTE: I'm not entirely clear on these defines
#define __STDC_LIMIT_MACROS
#define __STDC_CONSTANT_MACROS 

// LLVM: doesn't seem to care about conversion warnings
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
	#include "llvm/IR/DerivedTypes.h"
	#include "llvm/IR/Constants.h"
	#include "llvm/Support/TargetSelect.h"
	#include "llvm/IR/DataLayout.h"
#pragma GCC diagnostic pop

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "ir/dump.hpp"

namespace ir_llvm
{

type_manager::type_manager( llvm::LLVMContext & context, llvm::DataLayout & data_layout )
	: context( context )
	, data_layout( data_layout )
{
	generic_ptr_ = llvm::PointerType::get( llvm::Type::getInt8Ty(context), 0 );
	void_ = llvm::Type::getVoidTy(context);
	
	//construct destructor function type
	std::vector<llvm::Type*> args;
	args.push_back( generic_ptr_ );
	destructor_ = llvm::FunctionType::get( void_, args, false/*vararg*/ );
	
	//extype_type (exception type)
	extype_type_ = llvm::StructType::create( context, "_extype_type" );
	std::vector<llvm::Type*> fields;
	fields.push_back( f_natural_int() );
	extype_type_->setBody( fields );
	
	{ //generic shared type
		generic_shared_ = llvm::StructType::create( context, "_generic_shared" );
		std::vector<llvm::Type*> fields(sf_size);
		fields[sf_ref_count] = f_natural_int();
		fields[sf_destructor] = ptr_to( destructor() );
		fields[sf_align] = get_align_field();
		fields[sf_data] = get_align_field();
		generic_shared_->setBody( fields );
	}
	
	{ //generic context fragment
		generic_ctx_frag_ = llvm::StructType::create( context, "_generic_ctx_frag" );
		std::vector<llvm::Type*> fields(cf_size);
		fields[cf_func_ptr] = generic_ptr();
		fields[cf_ctx_ptr] = generic_ptr();
		fields[cf_data_destructor] = ptr_to( destructor() );
		fields[cf_align] = get_align_field();
		fields[cf_data] = get_align_field();
		generic_ctx_frag_->setBody( fields );
	}
	
	//error_return
	error_return_ = llvm::StructType::create( context, "_error_return" );
	fields.clear();
	fields.push_back( f_error_code() );
	fields.push_back( generic_shared_ptr() );
	error_return_->setBody( fields );
		
}

llvm::ArrayType * type_manager::get_align_field( size_t len ) {
	if( !len ) {
		len = platform::max_align;
	}
	return llvm::ArrayType::get( f_intx( unsigned(len * platform::byte_size) ), 0 );
}

llvm::PointerType * type_manager::generic_shared_ptr() {
	return ptr_to(generic_shared_);
}

bool type_manager::type_ref::operator<( type_ref const & o ) const {
	if( ctx != o.ctx )
		return ctx < o.ctx;
	if( shared != o.shared )
		return shared < o.shared;
	if( direct != o.direct )
		return direct < o.direct;

	auto cmp_ctx = [&]( ir::type const * a, ir::type const * b ) {
		if( a == b ) {
			return false;
		}
		if( !a ) {
			return true;
		}
		if( !b) {
			return false;
		}
		return *a < *b;
	};
	
	auto q = cmp_ctx(ctx_a, o.ctx_a);
	if( q ) {
		return q;
	}
		
	//assume exact types have same pointer (that is what we need to do for LLVM)
	if( ctx_b != o.ctx_b )
		return ctx_b < o.ctx_b;
	if( func_a != o.func_a ) 
		return func_a < o.func_a;
		
	return cmp_ctx(ir_type, o.ir_type);
}

llvm::IntegerType * type_manager::f_intx( unsigned bits ) {
	return llvm::IntegerType::get(context, bits);
}

llvm::IntegerType * type_manager::f_abi_char() {
	return llvm::IntegerType::get(context, platform::abi_char_size );
}

llvm::IntegerType *  type_manager::main_argc() {
	return f_intx( platform::abi_int_size );
}

llvm::PointerType * type_manager::main_argv() {
	return ptr_to( ptr_to( f_abi_char() ) );
}

llvm::IntegerType * type_manager::f_byte() {
	return f_intx( platform::byte_size );
}

llvm::Type * type_manager::f_float( unsigned bits ) {
	//TODO: float/double should be matched to platform sizes (though the same on all known platforms)
	switch( bits ) {
		case 32: return llvm::Type::getFloatTy(context);
		case 64: return llvm::Type::getDoubleTy(context);
		case 80: return llvm::Type::getX86_FP80Ty(context);
		case 128: return llvm::Type::getFP128Ty(context);
	}
	STATE_FAIL( "unsupported float bitsize" );
}

llvm::IntegerType * type_manager::f_natural_int() {
	return f_intx( platform::natural_integer_size );
}

llvm::IntegerType * type_manager::f_abi_int() {
	return f_intx( platform::abi_int_size );
}

llvm::IntegerType* type_manager::f_ptr_int() {
	return f_intx( platform::abi_ptr_size );
}

llvm::PointerType * type_manager::ptr_to( llvm::Type * t ) {
	return llvm::PointerType::get( t, 0 ); 
}

size_t type_manager::alloc_size( llvm::Type * t ) { 
	return data_layout.getTypeAllocSize( t ); 
}

size_t type_manager::max_align() {
	return platform::max_align;
}

type_manager::type_entry type_manager::value( ir::type const & t ) {
	auto tmh = type_map_helper( &t );
	if( tmh.has() )
		return tmh.get();
		
	type_entry ret;
	
	switch( t.fundamental )
	{
		case ir::type::f_integer:
		case ir::type::f_binary: //LLVM: has no distinction between signed/unsigned integers (for storage, for ops it does)
			ret.ll_type = f_intx( t.trivial_bitsize );
			ret.direct_return = true; //TODO: what about very large integers?
			break;
			
		case ir::type::f_float:
			ret.ll_type = f_float( t.trivial_bitsize );
			ret.direct_return = true;
			break;
			
		case ir::type::f_boolean:
			ret.ll_type = llvm::Type::getInt1Ty( context );
			ret.direct_return = true;
			break;
			
		case ir::type::f_tuple:
		{
			auto st = llvm::StructType::create( context, t.name.c_str() );
			ret.ll_type = st;
			//do now to allow recursion
			tmh.set(ret);
			
			std::vector<llvm::Type*> fields;
			for( unsigned i=0; i < t.sub.size(); ++i )
				fields.push_back( value( *t.sub[i] ).ll_type );
			bool packed = false;
			switch( t.pack ) {
				case ir::type::p_normal: break;
				case ir::type::p_abi: break;
				case ir::type::p_byte: packed = true; break;
			}
			st->setBody( fields, packed );
			break;
		}
		
		case ir::type::f_optional:
		{
			STATE_CHECK( t.sub.size() == 2 );
			auto st = llvm::StructType::create( context, t.name.c_str() );
			ret.ll_type = st;
			tmh.set(ret);
			
			std::vector<llvm::Type*> fields;
			fields.push_back( value( *t.sub[0] ).ll_type );
			fields.push_back( value( *t.sub[1] ).ll_type );
			st->setBody( fields );
			break;
		}

		case ir::type::f_function_ptr: {
			ret.direct_return = true;
			switch( t.access ) {
				case ir::type::a_unbound: 
				case ir::type::a_unbound_shared:
				case ir::type::a_raw: {
					auto fr = function( t );
					ret.ll_type = llvm::PointerType::get( fr.ll_func_type, 0 );
					break;
				}
				case ir::type::a_bound: {
					ret.ll_type = llvm::PointerType::get( shared_context_struct( t ).ll_type, 0 );
					break;
				}
			}
			break;
		}
		
		case ir::type::f_raw_pointer:
		{
			STATE_CHECK( t.sub.size() == 1 );
			auto pt = value( *t.sub[0] ).ll_type;
			ret.ll_type = llvm::PointerType::get( pt, 0 );
			ret.direct_return = true;
			break;
		}
		
		case ir::type::f_shared_pointer:
		{
			//LLVM: can't create an empty body pointer like for Struct, but it appears it doesn't care about
			//multiple equivalent pointers (or it manages them itself)
			auto shared = shared_struct( t ).ll_type;
			ret.ll_type = llvm::PointerType::get( shared, 0 );
			ret.direct_return = true;
			break;
		}
			
		case ir::type::f_array:
			ret.direct_return = true;
			ret.ll_type = generic_ptr();
			break;
			
		default:
			STATE_CHECK( false );
	}
	
	STATE_CHECK( ret.ll_type );
	tmh.set(ret);
	return ret;
}

type_manager::type_entry type_manager::shared_ptr( ir::type const & t ) {
	return shared_struct(t, true);
}

type_manager::type_entry type_manager::shared_struct( ir::type const & t, bool direct )
{
	ir::type const * underlying;
	if( direct ) {
		underlying = &t;
	} else {
		STATE_CHECK( t.sub.size() == 1 );
		STATE_CHECK( t.fundamental == ir::type::f_shared_pointer );
		underlying = t.sub[0].get();
	}
	
	auto tmh = type_map_helper( underlying );
	tmh.tr.shared = true;
	if( tmh.has() )
		return tmh.get();
		
	static int count = 0;
	char buf[128];
	sprintf( buf, "shared_struct_%d", count++ );
	auto st = llvm::StructType::create( context, buf );

	type_entry te;
	te.ll_type = st;
	te.direct_return = false;
	tmh.set(te);

	std::vector<llvm::Type*> fields(sf_size);
	fields[sf_ref_count] = f_natural_int();
	fields[sf_destructor] = ptr_to( destructor() );
	fields[sf_align] = get_align_field();
	fields[sf_data] = value( *underlying ).ll_type;
	st->setBody( fields );
	
	return te;
}

type_manager::func_type_entry type_manager::function( ir::type const & t )
{
	STATE_CHECK( t.fundamental == ir::type::f_function_ptr );

	auto tmh = func_type_map_helper( &t );
	if( tmh.has() )
		return tmh.get();
	//LLVM: offers no empty-body FunctionType, we have to hope it accepts duplicates
	
	func_type_entry ret;
	ret.is_error_return = t.can_fail;
	
	//collapse returns as possible
	if( t.sub.size() == 0 ) {
		ret.ll_logical_return_type = nullptr;
	} else if( t.sub.size() == 1 ) {
		auto rt = value( *t.sub[0] );
		ret.ll_logical_return_type = rt.ll_type;
		ret.is_direct_return = rt.direct_return;
	} else {
		std::vector<llvm::Type*> fields;
		for( size_t i=0; i < t.sub.size(); ++i )
			fields.push_back( value( *t.sub[i] ).ll_type );
			
		ret.ll_logical_return_type = llvm::StructType::create( context, fields );
		ret.is_direct_return = false; //LLVM:FEATURE: there are actually some structures which can be directly returned (platform/1ABI dependent)
	}

	std::vector<llvm::Type*> args;
	
	if (t.can_fail && ret.is_direct_return) {
		ret.is_direct_return = false;
	}
	
	//if we can't return the return type we put a pointer to it as the first argument
	llvm::Type * use_ret_type = ret.ll_logical_return_type;
	if( ret.ll_logical_return_type && !ret.is_direct_return ) {
		args.push_back( llvm::PointerType::get( ret.ll_logical_return_type, 0 ) );
		use_ret_type = type_void();
	}
	
	if( t.can_fail ) {
		use_ret_type = error_return();
	}
	
	if( !use_ret_type ) {
		use_ret_type = type_void();
	}
	
	//do we need a context argumentr (comes after returns value)
	if( t.access != ir::type::a_raw ) {
		args.push_back( generic_ptr() );
		ret.is_context = true;
		ret.is_context_shared = t.access == ir::type::a_unbound_shared;
	} else {
		ret.is_context = false;
	}
		
	bool varargs = false;
	for( size_t i=0; i < t.args.size(); ++i ) {
		if( t.args[i]->fundamental == ir::type::f_abi_var_args ) {
			varargs = true;
		} else {
			args.push_back( value( *t.args[i] ).ll_type );
		}
	}
		
	ret.ll_func_type = llvm::FunctionType::get( use_ret_type, args, varargs );
	tmh.set(ret);
	return ret;
}

/**
	Produces the logical type:
		shared<context<ctx_data>>
		
	@param t the function type which is being wrapped
	@param ctx the type of the cf_data field (onyl one of ctx,ll_ctx should be specified)
	@param ll_ctx the type of the cf_data field
*/
type_manager::type_entry type_manager::shared_context_struct( ir::type const & t,
	ir::type const * ctx, llvm::Type * ll_ctx )  {
	STATE_CHECK( !ctx || !ll_ctx );
	
	//the uniqueness is based on the function type
	auto tmh = type_map_helper( nullptr );
	auto func_ty = function(t);
	tmh.tr.func_a = func_ty.ll_func_type;
	tmh.tr.ctx = true;
	tmh.tr.shared = true;
	tmh.tr.ctx_a = ctx;
	tmh.tr.ctx_b = ll_ctx;
 	if( tmh.has() )
		return tmh.get();
	
	static int count = 0;
	char buf[128];
	sprintf( buf, "shared_context_%d", count++ );
	auto st = llvm::StructType::create( context, buf );
	auto ft = llvm::StructType::create( context );

	type_entry te;
	te.ll_type = st;
	te.direct_return = false;
	tmh.set( te );
	
	std::vector<llvm::Type*> sfields(sf_size);
	//for the shared object count
	sfields[sf_ref_count] = f_natural_int();
	sfields[sf_destructor] = ptr_to( destructor() );
	sfields[sf_align] = get_align_field();
	sfields[sf_data] = ft;
	
	std::vector<llvm::Type*> ffields(cf_size);
	ffields[cf_func_ptr] = ptr_to( func_ty.ll_func_type );
	ffields[cf_ctx_ptr] = generic_ptr();
	ffields[cf_data_destructor] = ptr_to( destructor() );
	ffields[cf_align] = get_align_field();
	if( ctx ) {
		ffields[cf_data] = value(*ctx).ll_type;
	} else if( ll_ctx ) {
		ffields[cf_data] = ll_ctx;
	} else {
		ffields[cf_data] = get_align_field();
	}
	ft->setBody( ffields );
	
	st->setBody( sfields );

	return te;
}

} //ir_llvm