#include "ir/std_functions.hpp"
#include "ir/builder.hpp"
#include "ir/type_manager.hpp"

namespace ir
{

namespace 
{
	/**
		Reduces repetition of some common setup work. This allows functions to be generated only
		one time for a given type and ensure they have the correct exportable name.
	*/
	class function_prep
	{
		std::string use_name;
		shared_ptr<function> cur_func;
		compiler_context & cc;
	public:
		function_prep( compiler_context & cc, std::string const & name, shared_ptr<ir::type const> item_type ) 
			: cc( cc )
		{
			use_name = name + item_type->get_export_name();
			cur_func = cc.get_cur_module().if_get_function( use_name );
		}
		
		bool has() {
			return !!cur_func;
		}
		value_defn get_cur() {
			return cur_func->defn_value;
		}
		
		function_builder create() {
			return module_builder(cc).create_function( use_name, false );
		}
	};
	
} //eon anon

value_defn get_delete_array_func( compiler_context & cc, shared_ptr<ir::type const> item_type )
{
	function_prep prep( cc, "_delete_array_", item_type );
	if( prep.has() )
		return prep.get_cur();
		
	auto fbmeta = prep.create();
	builder fb( fbmeta );
	auto array_type = cc.get_types().get_array( item_type );
	auto src = fbmeta.define_arg( array_type, "src" );
	
	//TODO: this bit is repeated in three array funcs!
	auto int_type = cc.get_types().get_natural_integer();
	auto loop_count = fb.define_local( int_type, "ndx" );
	cc.assign( value_ref::as_raw_pointer( loop_count ), fb.const_integer( int_type, 0 ), fb, fb );
	
	auto array_len = fb.add_special_field( special_field_inst::f_array_len, int_type, src );
	
	auto loop = fb.add_loop();
	
	//cond_block
	builder cond_block( *loop->cond_block );
	loop->cond_value = cond_block.add_binary( binary_inst::o_lt, 
		loop_count, array_len, cc.get_types().get_boolean() );

	//body_block
	builder body_block( *loop->body_block );
	auto slot = body_block.add_array_offset( src, loop_count );
	cc.release( slot, body_block, block_flow::defer );
	
	auto next_loop_count = body_block.add_binary( binary_inst::o_add,
		loop_count, fb.const_integer( int_type, 1 ), int_type );
	body_block.add_store_inst( value_ref::as_raw_pointer( loop_count ), next_loop_count );
	
	
	auto fa = fb.create_free_array_inst( src );
	fb.add( fa );
	
	return fbmeta.get_value_defn();
}

value_defn get_new_func( compiler_context & cc, shared_ptr<ir::type const> array_type )
{
	STATE_CHECK( array_type->fundamental == type::f_array );

	function_prep prep( cc, "_new_array_", array_type->sub[0] );
	if( prep.has() )
		return prep.get_cur();
	
	auto fbb = prep.create();
	auto int_type = cc.get_types().get_natural_integer();
	auto size_arg = fbb.define_arg( int_type, "sz" );
	auto alloc_return = fbb.define_return( array_type, "arr" );
	
	builder fbody( fbb );
	auto aa = fbody.add_alloc_array_inst( array_type, array_type->item_type(), size_arg );
	fbody.add_store_inst( value_ref::as_raw_pointer( alloc_return ), aa->result );	
	
	//initialize the array bits
	auto loop_count = fbody.define_local( int_type, "ndx" );
	cc.assign( value_ref::as_raw_pointer( loop_count ), fbody.const_integer( int_type, 0 ), fbody, fbody );
	auto loop = fbody.add_loop();
	
	builder cond_block( *loop->cond_block );
	loop->cond_value = cond_block.add_binary( binary_inst::o_lt, 
		loop_count, size_arg, cc.get_types().get_boolean() );
	
	//body_block
	builder body_block( *loop->body_block );
	auto aa_ptr = value_ref::as_raw_pointer( alloc_return );
	auto slot = body_block.add_array_offset( aa_ptr, loop_count );
	cc.initialize( slot, array_type->sub[0], 0, body_block );

	auto next_loop_count = body_block.add_binary( binary_inst::o_add,
		loop_count, fbody.const_integer( int_type, 1 ), int_type );
	body_block.add_store_inst( value_ref::as_raw_pointer( loop_count ), next_loop_count );

	return fbb.get_value_defn();
}

value_defn get_array_copy_func( compiler_context & cc, shared_ptr<ir::type const> item_type )
{
	function_prep prep( cc, "_copy_array_", item_type );
	if( prep.has() )
		return prep.get_cur();
		
	auto fbmeta = prep.create();
	auto array_type = cc.get_types().get_array( item_type );
	value_ref dst, src;
	dst = fbmeta.define_return( array_type, "dst" );
	src = fbmeta.define_arg( array_type, "src" );
	
	builder fb(fbmeta);
	auto int_type = cc.get_types().get_natural_integer();
	auto array_len = fb.add_special_field( special_field_inst::f_array_len, int_type, src );

	//create new array if requested
	auto aa = fb.add_alloc_array_inst( array_type, item_type, array_len );
	auto rptr = fb.access_raw_pointer( dst );
	fb.add_store_inst( rptr, aa->result );
	dst = rptr;
	
	//copy items
	auto loop_count = fb.define_local( int_type, "ndx" );
	cc.assign( value_ref::as_raw_pointer( loop_count ), fb.const_integer( int_type, 0 ), fb, fb );
	auto loop = fb.add_loop();
	
	//cond_block
	builder cond_block( *loop->cond_block );
	loop->cond_value = cond_block.add_binary( binary_inst::o_lt, 
		loop_count, array_len, cc.get_types().get_boolean() );

	//body_block
	builder body_block( *loop->body_block );
	auto to_slot = body_block.add_array_offset( dst, loop_count );
	auto from_slot = body_block.add_array_offset( src, loop_count );
	unsigned assign_flags =0;
	assign_flags |= compiler_context::af_no_release;
	cc.assign( to_slot, from_slot, body_block, body_block, assign_flags ); //TODO: don't all assigns need a store_block!

	auto next_loop_count = body_block.add_binary( binary_inst::o_add,
		loop_count, fb.const_integer( int_type, 1 ), int_type );
	body_block.add_store_inst( value_ref::as_raw_pointer( loop_count ), next_loop_count );

	return fbmeta.get_value_defn();
}

value_defn get_release_function( compiler_context & cc, shared_ptr<type const> rt )
{
	function_prep prep( cc, "_release_", rt );
	if( prep.has() )
		return prep.get_cur();
		
	auto fbb = prep.create();
	auto object = fbb.define_arg( cc.get_types().get_raw_pointer(rt), "obj" );
	
	builder fbody( fbb );
	cc.release_pointer( object, fbody, block_flow::sequential, compiler_context::rf_inline );
	
	auto res = fbb.get_value_defn();
	return res;
}

} //eon ir