#include "ir/type.hpp"
#include "platform/platform.hpp"
#include "util/validation_context.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "ir/dump.hpp"

#include <sstream>

namespace ir
{

shared_ptr<type> type::clone( bool with_name ) const {
	auto r = make_shared<type>( types, fundamental );
	r->trivial_bitsize = trivial_bitsize;
	r->sub = sub;
	r->args = args;
	r->pack = pack;
	r->access = access;
	r->can_fail = can_fail;
	r->copy = copy;
	if( with_name )
		r->name = name;
	return r;
}

/**
	Sub-type pairs are created and checked against the stack before calling compare.
	This allows recursive types to properly compare.
*/
int type::common_sub_compare( std::vector<shared_ptr<type const>> const & a, 
	std::vector<shared_ptr<type const>> const & o, type_pair_stack & tps ) const
{
	int diff = int(a.size()) - int(o.size());
	if( diff )
		return diff;
	for( size_t i=0; i < a.size(); ++i )
	{
		auto tp = type_pair( a[i].get(), o[i].get() );
		if( std::find( tps.begin(), tps.end(), tp ) != tps.end() )
			continue;
			
		tps.push_back( tp );
		diff = a[i]->compare( *o[i], tps );
		if( diff )
			return diff;
	}
	
	return 0;
}

int type::compare(type const & o) const
{
	type_pair_stack tps;
	return compare( o, tps );
}

int type::sub_compare( type const & o ) const
{
	type_pair_stack tps;
	return sub_compare( o, tps );
}

int type::compare( type const & o, type_pair_stack & tps ) const
{
	if( name != o.name )
		return name < o.name ? -1 : 1;
	
	int diff;
	diff = int(fundamental) - int(o.fundamental);
	if( diff )
		return diff;
		
	if( has_trivial_bitsize() ) {
		diff = trivial_bitsize - o.trivial_bitsize;
		if( diff )
			return diff;
	}
		
	diff = int(access) - int(o.access);
	if( diff )
		return diff;
		
	diff = int(pack) - int(o.pack);
	if( diff )
		return diff;
		
	diff = int(can_fail) - int(o.can_fail);
	if( diff )
		return diff;

	if( context && !o.context ) {
		return -1;
	} else if( !context && o.context ) {
		return 1;
	} else if( context && o.context ) {
		diff = context->compare( *o.context, tps );
		if( diff )
			return diff;
	}
	
	//ignore destroy -- not expecting same type with distinct destructor
	
	return sub_compare( o, tps );
}

int type::sub_compare( type const & o, type_pair_stack & tps ) const
{
	int diff = common_sub_compare( sub, o.sub, tps );
	if( diff )
		return diff;
		
	diff = common_sub_compare( args, o.args, tps );
	if( diff )
		return diff;
	
	return 0;
}

void type::validate( util::validation_context & vctx ) const
{
	if( vctx.cover(this) )
		return;
	
	STATE_CHECK( fundamental != fundamental_t(0) );
	switch( fundamental )
	{
		case f_function_ptr:
			STATE_CHECK( access >= a_bound && access <= a_unbound_shared );
			break;
			
		case f_integer:
		case f_binary:
		case f_float:
			STATE_CHECK( trivial_bitsize > 0 );
		case f_boolean:
			STATE_CHECK( sub.empty() );
			STATE_CHECK( args.empty() );
			break;
			
		case f_raw_pointer:
		case f_shared_pointer:
			STATE_CHECK( args.empty() );
			STATE_CHECK( sub.size() == 1 );
			break;
			
		case f_tuple:
			STATE_CHECK( args.empty() );
			break;
			
		case f_array:
			STATE_CHECK( sub.size() == 1 );
			break;
			
		case f_optional:
			STATE_CHECK( sub.size() == 2 );
			STATE_CHECK( sub[0]->is_boolean() );
			break;
			
		case f_abi_var_args:
			break;
	}
	
	for( auto & s : sub )
		s->validate(vctx);
	for( auto & p : args )
		p->validate(vctx);
}


bool type::is_trivial_destruction() const {
	if( destroy )
		return false;
		
	switch( fundamental )	{
		case f_integer:
		case f_float:
		case f_boolean:
		case f_raw_pointer:
		case f_binary:
			return true;
			
		case f_function_ptr:
			return access != a_bound;
		case f_array:
		case f_shared_pointer:
			return false;
			
		case f_optional:
			return sub[1]->is_trivial_destruction();
			
		case f_tuple:
			for( size_t i=0; i < sub.size(); ++i )
				if( !sub[i]->is_trivial_destruction() )
					return false;
			return true;
			
		case f_abi_var_args:
			return false;
	}
	STATE_FAIL("corrupted");
}

std::string const & type::get_export_name() const
{
	if( export_name.size() )
		return export_name;

	std::ostringstream out;
	
	if( name.size() ) {
		out << "N" << name.size() << name;
		export_name = out.str();
		return export_name;
	}
		
	auto parts = [&]( std::vector<shared_ptr<type const>> const & parts ) {
		for( size_t i=0; i < parts.size(); ++i ) {
			if( i ) {
				out << platform::export_sep;
			}
			out << parts[i]->get_export_name();
		}
	};
	
	switch( fundamental ) {
		case f_function_ptr:
			out << "C";
			out << access;
			
			if( context ) {
				out << "X";
				out << context->get_export_name();
				out << "x";
			}
			out << "P";
			parts(args);
			out << "p";
			
			out << "R";
			parts(sub);
			out << "r";
			
			out << "c";
			break;
			
		case f_raw_pointer:
			out << "R";
			out << sub[0]->get_export_name();
			out << "r";
			break;
			
		case f_shared_pointer:
			out << "S";
			out << sub[0]->get_export_name();
			out << "s";
			break;
			
		case f_tuple:
			out << "T";
			out << pack;
			parts( sub );
			out << "t";
			break;
			
		case f_array:
			out << "A";
			out << sub[0]->get_export_name();
			out << "a";
			break;
			
		case f_optional:
			out << "O";
			out << sub[1]->get_export_name();
			out << "o";
			break;
			
		case f_integer:
			out << "i" << trivial_bitsize;
			break;
			
		case f_binary:
			out << "u" << trivial_bitsize;
			break;
			
		case f_float:
			out << "f" << trivial_bitsize;
			break;
			
		case f_boolean:
			out << "b";
			break;
			
		case f_abi_var_args:
			out << "v";
			break;
	}
	
	export_name = out.str();
	return export_name;
}

} //eon ir
