#include "ir/type_manager.hpp"
#include "ir/type.hpp"
#include "ir/value.hpp"
#include "ir/module.hpp"

#include "platform/platform.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace ir {

type_manager::type_manager( ) {
}

shared_ptr<type const> type_manager::get_boolean() {
	if( !type_boolean ) {
		type_boolean = make_shared<type>( shared_from_this(), type::f_boolean);
	}
	return type_boolean;
}

shared_ptr<type const> type_manager::get_empty() {
	if( !type_empty ) {
		type_empty = make_shared<type>( shared_from_this(), type::f_tuple );
	}
	return type_empty;
}

shared_ptr<type const> type_manager::get_natural_integer() {
	return get_integer( platform::natural_integer_size );
}

shared_ptr<type const> type_manager::get_integer( int bitsize )
{
	STATE_CHECK( bitsize > 0 );
	
	auto it = integer_cache.find( bitsize );
	if( it != integer_cache.end() )
		return it->second;
		
	auto out = make_shared<type>( shared_from_this(),type::f_integer);
	out->trivial_bitsize = bitsize;
	integer_cache[bitsize] = out;
	return out;
}

shared_ptr<type const> type_manager::get_binary_for_pointer() {
	return get_binary( platform::pointer_size );
}

shared_ptr<type const> type_manager::get_binary( int bitsize ) {

	STATE_CHECK( bitsize > 0 );
		
	auto it = binary_cache.find( bitsize );
	if( it != binary_cache.end() )
		return it->second;
		
	auto out = make_shared<type>( shared_from_this(),type::f_binary);
	out->trivial_bitsize = bitsize;
	binary_cache[bitsize] = out;
	return out;
}

shared_ptr<type const> type_manager::get_raw_pointer( shared_ptr<type const> t ) {
	auto p = make_shared<type>(  shared_from_this(),type::f_raw_pointer );
	p->sub.push_back( t );
	return p;
}

shared_ptr<type const> type_manager::get_shared_pointer( shared_ptr<type const> t ) {
	auto p = make_shared<type>(  shared_from_this(),type::f_shared_pointer );
	p->sub.push_back( t );
	return p;
}

shared_ptr<type const> type_manager::get_deref( shared_ptr<type const> t ) {
	PRE_CHECK( t->fundamental == type::f_raw_pointer && t->sub.size() == 1 );
	return t->sub[0];
}

shared_ptr<type const> type_manager::get_deref_shared( shared_ptr<type const> t ) {
	PRE_CHECK( t->fundamental == type::f_shared_pointer && t->sub.size() == 1 );
	return t->sub[0];
}

shared_ptr<type const> type_manager::get_raw_shared( shared_ptr<type const> t ) {
	DEBUG( "get_raw_shared", dump::get_type(t.get()), t->sub.size() );
	PRE_CHECK( t->fundamental == type::f_shared_pointer && t->sub.size() == 1 );
	return get_raw_pointer( t->sub[0] );
}

shared_ptr<type const> type_manager::get_array( shared_ptr<type const> t ) {
	auto p = make_shared<type>(  shared_from_this(),type::f_array );
	p->sub.push_back( t );
	return p;
}

shared_ptr<type const> type_manager::remove_modifiers( shared_ptr<type const> t ) {
	switch( t->fundamental ) {
		case type::f_raw_pointer:
		case type::f_shared_pointer:
			return t->sub[0];
			
		case type::f_optional:
			return t->sub[1];
			
		default:
			return t;
	}
}

} //eon ir