#include "ir/type_manager_leaf.hpp"
#include "ir/compiler_leaf.hpp"
#include "ir/type.hpp"

#include "lang/declaration.hpp"
#include "lang/type.hpp"
#include "lang/scope.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"
#include "ir/dump.hpp"


namespace ir
{

type_manager_leaf::type_manager_leaf( compiler_leaf & comp ) : 
	comp( comp ) {
}

bool type_manager_leaf::leaf_type_key::operator<( leaf_type_key const & o ) const {
	bool r =  *leaf_type < *o.leaf_type;
	return r;
}

shared_ptr<type const> type_manager_leaf::get( leaf::type_ref const & t ) {
	return get( t.extr() );
}

shared_ptr<type const> type_manager_leaf::get_fun( shared_ptr<leaf::intr_type const> fun ) {
	DEBUG( "GET-TYPE", fun->cref, leaf::dump::get( fun ) );
	//a simple cache of types we've already processed
	auto cached = known_fun_types.if_get( fun->cref );
	if( cached ) {
		//DEBUG( "GF CACHE", dump::get_type( (*cached).get()  ) );
		return *cached;
	}

	//check for equivalent type
	leaf_type_key ltk;
	ltk.leaf_type = fun;
	auto lit = leaf_type_cache.find( ltk );
	if( lit != leaf_type_cache.end() ) {
		//DEBUG( "GF KNOWN", leaf::dump::get( *lit->leaf_type ), dump::get_type( lit->ir_type.get() ) );
		known_fun_types.set( fun->cref, lit->ir_type );
		return lit->ir_type;
	}
	
	//DEBUG( "GF NEW" );
	
	//the type object must be cached prior to recursion
	auto register_out = [&]( shared_ptr<type const> rt ) {
		known_fun_types.set( fun->cref, rt );
		//track for equivalent types
		ltk.ir_type = rt;
		leaf_type_cache.insert( ltk );
	};
	
	//does standard creation/registration
	auto make_out = [&]( type::fundamental_t f ) {
		auto out = make_shared<type>( shared_from_this(),f);
		if( fun->is_whole() ) {
			//TODO: why is this needed?
			out->name = fun->get_qualified_id();
		}
		register_out(out);
		return out;
	};
	
	switch( fun->get_fun_class() ) {
		case leaf::intr_type::f_rational:
			PRE_CHECK( false );
			
		case leaf::intr_type::f_function: {
			auto out = make_out(type::f_function_ptr);
			auto & t = fun->cast<leaf::intr_type_function>();
			switch( t.access ) {
				case leaf::intr_type_function::a_bound:
					STATE_CHECK( !t.context.is_defined() );
					out->access = type::a_bound;
					break;
				case leaf::intr_type_function::a_raw:
					STATE_CHECK( !t.context.is_defined() );
					out->access = type::a_raw;
					break;
				case leaf::intr_type_function::a_unbound:
				case leaf::intr_type_function::a_construct: //earlier typing adapts the call
					STATE_CHECK( t.context.is_defined() );
					out->access = t.context.extr().is_shared() ? type::a_unbound_shared : type::a_unbound;
					break;
			}
			auto & args = *t.args;
			auto & returns = *t.returns;
			
			for( size_t i=0; i < args.sub.size(); ++i ) {
				auto p = get( args.sub[i].type.extr() );
				out->args.push_back( p );
			}
			for( size_t i=0; i < returns.sub.size(); ++i ) {
				auto p = get( returns.sub[i].type.extr() );
				out->sub.push_back( p );
			}
			if( t.context.is_defined() ) {
				auto ctx = t.context.extr();
				out->context = get_fun(ctx.intr);
			}
			out->can_fail = t.can_fail;
			
			return out;
		}
		
		case leaf::intr_type::f_integer: {
			auto & t = fun->cast<leaf::intr_type_integer>();
			//non-distinct must always be the same "type"
			auto out = get_integer(t.data_bitsize);
			register_out(out);
			return out;
		}
		
		case leaf::intr_type::f_binary: {
			auto & t = fun->cast<leaf::intr_type_binary>();
			//must always be the same "type" instance
			auto out = get_binary(t.data_bitsize);
			register_out(out);
			return out;
		}			
		
		case leaf::intr_type::f_float: {
			auto out = make_out(type::f_float);
			auto & t = fun->cast<leaf::intr_type_float>();
			out->trivial_bitsize = t.data_bitsize;
			return out;
		}
		
		case leaf::intr_type::f_boolean: {
			auto out = get_boolean();
			register_out(out);
			return out;
		}
		
		case leaf::intr_type::f_contextual: {
			auto out = make_out(type::f_tuple);
			auto & t = fun->cast<leaf::intr_type_compound>();
			out->pack = type::p_normal;
			for( size_t i=0; i < t.sub.size(); ++i )
				out->sub.push_back( get( t.sub[i].type.extr() ) );
			return out;
		}
		
		case leaf::intr_type::f_value_list: {
			auto out = make_out(type::f_tuple);
			auto & t = fun->cast<leaf::intr_type_compound>();
			//TODO: size == 0 should probably be a special void type (it's not an object like an empty tuple is)
			if( t.sub.size() == 1 ) {
				out = get( t.sub[0].type.extr() )->clone();
				break;
			}
			
			out->pack = type::p_normal;
			for( size_t i=0; i < t.sub.size(); ++i )
				out->sub.push_back( get( t.sub[i].type.extr() ) );
		
			return out;
		}

		case leaf::intr_type::f_tuple: {
			auto out = make_out(type::f_tuple);
			auto & t = fun->cast<leaf::intr_type_tuple>();
			for( size_t i=0; i < t.sub.size(); ++i ) {
				auto p = get( t.sub[i].type.extr() );
				out->sub.push_back( p );
			}
			switch( t.pack ) {
				case leaf::intr_type_tuple::p_normal:
					out->pack = type::p_normal;
					break;
				case leaf::intr_type_tuple::p_abi:
					out->pack = type::p_abi;
					break;
				case leaf::intr_type_tuple::p_byte:
					out->pack = type::p_byte;
					break;
			}
			
			return out;
		}
		
		case leaf::intr_type::f_service:
		case leaf::intr_type::f_class: 
		case leaf::intr_type::f_closure: {
			auto out = make_out(type::f_tuple);
			if (fun->get_fun_class() == leaf::intr_type::f_service) {
				out->copy = type::c_none;
			}
			auto & t = fun->cast<leaf::intr_type_custom>();
			out->pack = type::p_normal;
			
			if( t.parent.is_defined() ) {
				auto p = get( t.parent );
				out->sub.push_back( p );
				//no sref since nothing explicitly references the parent (implicit through captures level)
			}
			
			auto local = t.get_type_decls();
			for( auto & sref : local ) {
				auto p = get( sref->type.extr() );
				out->sub.push_back( p );
				sref->xref = out->sub.size(); //+1 from actual index (so 0 is undefined)
			}
			
			if( t.destroy ) {
				out->destroy = comp.obtain_symbol_value_ref( t.destroy );
			}
			return out;
		}
			
		case leaf::intr_type::f_char: {
			auto out = get_binary(platform::char_size);
			register_out(out);
			return out;
		}
			
		case leaf::intr_type::f_array: {
			auto & t = fun->cast<leaf::intr_type_array>();
			if( t.raw ) {
				auto out = make_out(type::f_raw_pointer);
				out->sub.push_back( get( t.sub.extr() ) );
				return out;
			} else {
				auto out = make_out(type::f_array);
				out->sub.push_back( get( t.sub.extr() ) );
				return out;
			}
		}
		
		case leaf::intr_type::f_type: {
			//auto & t = fun->cast<leaf::intr_type_type>();
			//simply a tuple with no data for now (no dynamics support)
			auto out = make_out(type::f_tuple);
			return out;
		}
		
		case leaf::intr_type::f_module: {
			STATE_FAIL( "unexpected" );
		}

		case leaf::intr_type::f_compat: {
			auto & t = fun->cast<leaf::intr_type_compat>();
			if (t.type == leaf::intr_type_compat::abi_var_args) {
				auto out = make_out(type::f_abi_var_args);
				return out;
			} else if (t.type == leaf::intr_type_compat::string_tag ) {
				auto out = get_binary(platform::pointer_size);
				register_out(out);
				return out;
			} else {
				STATE_FAIL( "unexpected" );
			}
		}
	}

	STATE_FAIL();
}

shared_ptr<type const> type_manager_leaf::get( leaf::extr_type const & gt )
{
	type_key tk;
	tk.cref = gt.intr->cref;
	tk.shared = gt.is_shared();
	tk.optional = gt.is_optional();
	tk.lvalue = gt.is_lvalue();
	tk.value_ptr = gt.is_value_ptr();
	
	//check cache for this type (cref check is a shortcut)
	if( tk.cref ) {
		auto it = known_type_refs.find(tk);
		if( it != known_type_refs.end() ) {
			return it->cached;
		}
	}
	
	auto out = get_fun( gt.intr );
	tk.cref = gt.intr->cref;
	STATE_CHECK( tk.cref );
	
	if( gt.is_shared() ) {
		auto p = make_shared<type>( shared_from_this(), type::f_shared_pointer );
		p->sub.push_back( out );
		out = p;
	}
	
	if( gt.is_optional() ) {
		auto p = make_shared<type>(  shared_from_this(),type::f_optional );
		p->sub.push_back( get_boolean() );
		p->sub.push_back( out );
		out = p;
	}
	
	if( gt.is_lvalue() ) {
		auto p = make_shared<type>(  shared_from_this(),type::f_raw_pointer );
		p->sub.push_back( out );
		out = p;
	}
	
	if( gt.is_value_ptr() ) {
		auto p = make_shared<type>(  shared_from_this(),type::f_raw_pointer );
		p->sub.push_back( out );
		out = p;
	}

	tk.cached = out;
	known_type_refs.insert(tk);
	return out;
}

} //eon ir
