#include "lang/class_typer.hpp"
#include "lang/language.hpp"
#include "lang/statement_importer.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace leaf
{

/*static*/ void class_typer::process( shared_ptr<typedef_statement> ts, context ctx ) {
	STATE_CHECK( ts->class_block );
	
	shared_ptr<intr_type_classlike> class_type;
	if( ts->what == typedef_statement::w_class ) {
		class_type = make_shared<intr_type_class>();
	} else {
		class_type = make_shared<intr_type_service>();
	}
	class_type->local_complete = true;
	class_type->is_export = ts->is_export;
	class_type->set_distinct_name( ts->name );
	class_type->style = intr_type::s_distinct;
	class_type->context = ts->class_block->symbols; //TODO: MOVE into job step below, not used in parametric
	class_type->block = ts->class_block;
	class_type->block->symbols->set_base( ctx.sc, scope::bt_parent );
	ts->type = extr_type( class_type );
	ctx.sc->add_fun_type( class_type );

	//TODO: the typing of class/tuple in typer should be merged, with "type" as the common syntax
	if( ts->spec_type && ts->spec_type->params.size() ) {
		for( auto & s : ts->spec_type->params ) {
			class_type->params.push_back( { s.name } );
		}
		
	} else {
		auto ct = shared_ptr<class_typer>( new class_typer( ctx, class_type ) );
		ctx.add_pending_job( ct );
	}
	
	if(class_type->is_export) {
		ctx.add_export( class_type );
	}
}

/*static*/ void class_typer::instantiate( shared_ptr<intr_type_classlike> class_type, context proc_context ) {
	
	STATE_CHECK( class_type->block );
	STATE_CHECK( class_type->block->symbols );
	auto parent = class_type->block->symbols->get_base();
	class_type->block = class_type->block->pre_clone();
	class_type->context = class_type->block->symbols;
	
	//put parametric types into scope
	STATE_CHECK( class_type->instance_of->params.size() == class_type->applied_params.size() );
	for( size_t i=0; i < class_type->instance_of->params.size(); ++i ) {
		auto spec = type_identifier::get_spec( proc_context, class_type->applied_params[i] );
		STMT_CHECK( *class_type->block, spec.okay(), spec.explain_error() );
		auto name = class_type->instance_of->params[i].name;
		class_type->block->symbols->add_type_alias( name, spec.result );
	}
	
	DEBUG( "INSTANTIATE", dump::get( class_type ), (intptr_t)class_type->block->symbols.get() );
	
	auto nctx = context( proc_context.get_owner(), parent );
	auto ct = shared_ptr<class_typer>( new class_typer( nctx, class_type ) );
	proc_context.add_pending_job( ct );
}

class_typer::class_typer( context decl_context, shared_ptr<intr_type_classlike> class_type ) : 
	decl_context( decl_context ), 
	class_context( decl_context.get_owner(), class_type->block->symbols ),
	class_type( class_type ) {
	has_dtor = false;
	is_ctor_complete = false;
	class_block_done = false;

	auto et = extr_type(class_type);
	et.lvalue = true;
	class_type->block->symbols->mark_class_root( et );
	class_type->block->symbols->set_base( decl_context.sc, scope::bt_parent );

	auto ctor_block = make_shared<statement_block>(scope_style_t::imperative);
	ctor_block->symbols->set_base( class_type->block->symbols, scope::bt_parent );
	ctor_block->symbols->no_init_release = true;
	class_type->block->symbols->set_init_block( ctor_block );

	class_context.scan_symbols( class_type->block );
}

bool class_typer::process( typer* ) {
	if( !class_block_done ) {
		class_block_done = class_context.proc_statement_block( class_type->block, class_context );
	}

	if( !has_dtor ) {
		auto dtor_func = build_dtor();
		if( !dtor_func ) {
			has_dtor = true;
		} else if( dtor_func && dtor_func->type.is_complete() ) {
			has_dtor = true;
			auto ddecl = declaration::create();
			ddecl->form = declaration::f_single;
			ddecl->symbol = "_destroy";
			ddecl->type = dtor_func->type;
			ddecl->is_export = true;
			ddecl->lifetime = declaration::l_global;
			class_type->destroy = class_context.sc->add( ddecl );
			
			auto vs = make_shared<init_statement>();
			vs->local_complete = true;
			vs->sref = class_type->destroy;
			vs->value = dtor_func;
			vs->init_refer_value = false;
			vs->undefined = false;
			decl_context.get_static_init_block()->list.push_back( vs );
		}
	}
	
	//introduce the ctor into block
	if( ctor_funcs.size() == 0 ) {
		build_ctors();
	}
	
	if( !is_ctor_complete && ctor_funcs.size() ) {
		is_ctor_complete = complete_ctors();
	}

	if( class_block_done && is_ctor_complete && has_dtor ) {
		DEBUG( "DONE", dump::get( class_type ), class_type->is_complete() );
		return true;
	} 
	return false;
}

//TODO: this should rather find a way to inject the default and let the other code handle it normally, otherwise
//we run the risk of this not actually doing the same thing
//TODO: change is_default name
shared_ptr<expr_funcdefn> class_typer::build_implicit_ctor(bool is_default) {
	auto ctor_func = make_shared<expr_funcdefn>();
	ctor_func->block = class_context.sc->get_init_block();
	ctor_func->block->symbols->mark_function_root( ctor_func );
	
	auto cf_type = make_shared<intr_type_function>();
	cf_type->args = make_shared<intr_type_value_list>();
	cf_type->returns = make_shared<intr_type_value_list>();
	cf_type->context = extr_type(class_type);
	cf_type->access = is_default ? intr_type_function::a_construct : intr_type_function::a_unbound;
	
	auto ex = extr_type(cf_type);
	ex.literal = true;
	ctor_func->type = ex;
	ctor_func->func_type = type_identifier::get_spec( class_context, ctor_func->type ).result;
	ctor_func->mark_resolved();
	
	ctor_require_import = false;
	
	auto cdecl = declaration::create();
	cdecl->form = declaration::f_single;
	cdecl->symbol = is_default ? "default" : "$implicit";
	cdecl->type = ctor_func->type;
	cdecl->is_export = is_default ? true : false; //TODO: should be done only as required
	cdecl->lifetime = declaration::l_global;
	auto sref = class_context.sc->add( cdecl );
	
	auto vs = make_shared<init_statement>();
	vs->local_complete = true;
	vs->sref = sref;
	vs->value = ctor_func;
	vs->init_refer_value = false;
	vs->undefined = false;
	decl_context.get_static_init_block()->list.push_back( vs );
	return ctor_func;
}

void class_typer::build_ctors() {
	//check for user constructor
	auto cvars = class_type->block->symbols->get_local_vars( []( declaration const & o ) {
		if (!o.type.is_extr()) {
			return false;
		}
		auto q = o.type.extr().if_cast<intr_type_function>();
		return q && q->access == intr_type_function::a_construct;
	} );
	if( cvars.size() == 0 ) {
		ctor_funcs.push_back(build_implicit_ctor(true));
		return;
	}
	
	for( auto & cvar : cvars ) {
		STMT_CHECK( *class_type->block, cvar->source_statement->statement_form() == sf_var_statement );
		auto vs = std::static_pointer_cast<var_statement>( cvar->source_statement );
		
		STMT_CHECK( *class_type->block, vs->value->form == expression::f_funcdefn );
		auto fd = std::static_pointer_cast<expr_funcdefn>(vs->value);
		//until it's resolved we don't know what the ctor looks like
		if( !vs->value->is_resolved() ) {
			ctor_funcs.clear();
			return;
		}
		
		ctor_funcs.push_back(fd);
	}
	ctor_require_import = true;
}

bool class_typer::complete_ctors() {
	//so long as class_block is incomplete there might be more items to be added to init_block
	if( !class_block_done ) {
		return false;
	}
	
	//the init_block may not be finished yet
	auto init_block = class_context.sc->get_init_block();
	if( !class_context.proc_statement_block( init_block, class_context ) ) {	
		return false;
	}
	
	if( !ctor_require_import ) {
		return true;
	}
	
	//this optimization is failing in some cases `block/class/scenario_1.lfb`
// 	if( ctor_funcs.size() == 1 ) {
// 		auto & block = *ctor_funcs[0]->block;
// 		statement_importer importer( block.symbols );
// 		std::vector<shared_ptr<statement>> tlist;
// 		for( auto & is : class_context.sc->get_init_block()->list ) {
// 			tlist.push_back( importer.apply_statement( is, class_context.sc ) );
// 		}
// 		block.list.insert( block.list.begin(), tlist.begin(), tlist.end() );
// 		return true;
// 	}
	
	auto implicit = build_implicit_ctor(false);
	for( auto & ctor_func : ctor_funcs ) {
		auto call = make_shared<expr_funccall>();
		call->func = implicit;
		call->args = make_shared<expr_value_list>();
		
		DEBUG( "CtorType", dump::get(call->func->type));
		/*
		call->args->type = extr_type(make_shared<intr_type_value_list>());
		call->type = extr_type(make_shared<intr_type_value_list>());
		
		auto current = make_shared<expr_current>();
		call->context = current;
		current->what = expr_current::w_context;
		current->depth = 0;
		*/
		
		auto st = make_shared<expression_statement>();
		st->expr = call;
		
		auto & block = *ctor_func->block;
		block.list.insert( block.list.begin(), st );
		
		DEBUG( "Ctor", dump::get(block));
		class_context.add_pending_job( make_shared<typer::job_statement_block>(ctor_func->block ));
	}
	return true;
}

shared_ptr<expr_funcdefn> class_typer::build_dtor() {
	//check for user destructor
	auto dvars = class_type->block->symbols->get_local_vars( "destroy" );
	STMT_CHECK( *class_type->block, dvars.size() <= 1 );
	if( dvars.empty() )
		return shared_ptr<expr_funcdefn>();

	STMT_CHECK( *class_type->block, dvars[0]->source_statement->statement_form() == sf_var_statement );
	auto vs = std::static_pointer_cast<var_statement>( dvars[0]->source_statement );
	
	STMT_CHECK( *class_type->block, vs->value->form == expression::f_funcdefn );
	auto fd = std::static_pointer_cast<expr_funcdefn>(vs->value);
		
	return fd;
}

} //eon leaf