#include "lang/context.hpp"
#include "lang/typer.hpp"
#include "lang/module.hpp"
#include "lang/expression.hpp"
#include "lang/conversion_applicator.hpp"
#include "lang/dump.hpp"
#include "lang/statement_block.hpp"
#include "lang/scope.hpp"

namespace leaf {

context context::derive( shared_ptr<scope> symbols ) const {
	return context( owner, symbols, get_decl_scope() );
}

context::context( typer * owner, shared_ptr<scope> sc, shared_ptr<scope> decl_sc ) : 
	sc( sc ), 
	owner( owner ),
	decl_sc( decl_sc ) {
}

type_manager & context::types() const {
	return *owner->types;
}

bool context::no_incomplete() const {
	return owner->no_incomplete;
}

bool context::proc_statement_block( shared_ptr<statement_block> block, context ctx ) const { 
	return owner->proc_statement_block( block, ctx ); 
}

bool context::proc_statement_block( shared_ptr<statement_block> block ) const { 
	return owner->proc_statement_block( block ); 
}

void context::scan_symbols( shared_ptr<statement_block> block ) const {
	owner->scan_symbols( block );
}

void context::proc_statement( shared_ptr<statement> s, shared_ptr<scope> sc ) const {
	auto rs = owner->proc_statement( s, context(owner,sc) );
	STATE_CHECK(rs.get() == s.get());
}

void context::rewrite( shared_ptr<expression> orig, shared_ptr<expression> replace ) const {
	replace->src_loc = orig->src_loc;
	owner->progress_count++;
}

shared_ptr<statement_block> context::get_static_init_block() {
	return owner->mod->static_init_block;
}

void context::add_export( shared_ptr<declaration> vd ) {
	owner->mod->var_exports.push_back( vd );
}

void context::add_export( shared_ptr<intr_type> type ) {
	owner->mod->type_exports.push_back( type );
}

shared_ptr<expression> context::proc_expression( shared_ptr<expression> expr ) const {
	return expression_typer::process_expression( expr, *this );
}

void context::add_pending_job( shared_ptr<job> job ) const {
	owner->pending_jobs.push_back( job );
}

shared_ptr<expression> context::convert_expression( type_ref const & target, 
	shared_ptr<expression> source ) const {
	auto cost = converter().convert( target, source->type );
	return applicator().apply( cost, source );
}

shared_ptr<expression> context::if_convert_expression( type_ref const & target, 
	shared_ptr<expression> source ) const {
	auto cost = converter().convert( target, source->type );
	if( !cost.okay() ) {
		return shared_ptr<expression>();
	}
	return applicator().apply( cost, source );
}

shared_ptr<expression> context::convert_expression( type_converter::unary_convert target,
	shared_ptr<expression> source ) const
{
	auto cost = (converter().*target)( source->type );
	return applicator().apply( cost, source );
}

shared_ptr<expression> context::convert_expression( type_converter::binary_convert how,
	shared_ptr<expression> target, shared_ptr<expression> source ) const {
	auto cost = (converter().*how)( target->type, source->type );
	STMT_CHECK( *source, cost.okay(), cost.explain_error(),
		typer::tag_cerr_source_type( dump::get(source->type) ),
		typer::tag_cerr_target_type( dump::get(target->type) )
	)
	return applicator().apply( cost, source );
}

/**
	Common cloning and setup of an instanced parametric funcdefn.
*/
shared_ptr<expr_funcdefn> context::instance_funcdefn( expr_funcdefn & func, type_ref target ) const {
	auto clone = func.pre_clone_derived();
	
	clone->block->symbols->set_base( func.block->symbols->get_base(), scope::bt_parent );
	clone->type = target;
	return clone;
}

} //eon leaf