#include "lang/conversion_applicator.hpp"
#include "lang/expression.hpp"
#include "lang/declaration.hpp"
#include "lang/expression_typer/common.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace leaf {

	namespace {
	
		bool handles_incomplete( cvt_expression::form_t form ) {
			return form == cvt_expression::c_unset_optional ||
				form == cvt_expression::c_bind_context ||
				form == cvt_expression::c_param_inst ||
				form == cvt_expression::c_pack_value ||
				form == cvt_expression::c_field;
		}
	}

shared_ptr<expression> conversion_applicator::apply( type_converter::convert_t const & cost, 
	shared_ptr<expression> expr ) {
	STMT_CHECK( *expr, cost.okay(), cost.explain_error() + " " 
		+ dump::get( expr->type ) );
	if( cost.conversions ) {
		DEBUG( "CVT", typeid(*expr).name(), dump::get(expr), dump::get(cost.conversions) );
		return apply( *cost.conversions, expr );
	}
	return expr;
}

shared_ptr<expression> conversion_applicator::apply_single( cvt_expression & cvt, 
	shared_ptr<expression> expr ) {

	
	//could likely work for more than fixate, but those would be optimizations, fixation is required
	if( cvt.form == cvt_expression::c_fixate ) {
		
		size_t inner = expr->num_inner_trans();
		if( inner ) {
			for( size_t i=0; i < inner; ++i ) {
				expr->set_inner_trans( i, apply_single(
					cvt, expr->get_inner_trans( i ) ) );
			}
			expr->construct_result_type();
			return expr;
		} else {
			//DEBUG( "UNKNOWN CVT_OPTIONAL", typeid(*expr).name() );
		}
	}
	
	switch( cvt.form ) {
		#define CVT_EXPRESSION_FORM(ty) \
			case cvt_expression::c_##ty: \
				return apply_##ty( cvt, expr );
		CVT_EXPRESSION_FORMS
		#undef CVT_EXPRESSION_FORM
	}
	STATE_FAIL("corrupt conversion");
}

shared_ptr<expression> conversion_applicator::apply_parametric( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
	STMT_FAIL( *expr, "unexpected" );
}

shared_ptr<expression> conversion_applicator::apply_drop_lvalue( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {

	return drop_lvalue( expr );
}
			
shared_ptr<expression> conversion_applicator::apply_drop_literal( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
			
	auto ex = extr_type( expr->type.extr() );
	ex.literal = false;
	DEBUG( "DROP_LITERAL", dump::get( expr ), dump::get( expr->type ), dump::get( ex ) );
	expr->type = ex;
	return expr;
}
			
shared_ptr<expression> conversion_applicator::apply_addr_of( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
	auto addr = make_shared<expr_transform>();
	addr->transform = expr_transform::t_addr_of;
	addr->arg = expr;
	addr->construct_result_type();
	addr->mark_resolved();
	return addr;
}
		
shared_ptr<expression> conversion_applicator::apply_deref( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
	//STATE_CHECK( expr->type->shared );
	auto deref = make_shared<expr_transform>();
	deref->transform = expr_transform::t_deref;
	deref->arg = expr;
	deref->construct_result_type();
	deref->mark_resolved();
	return deref;
}
		
shared_ptr<expression> conversion_applicator::apply_promote( cvt_expression & cvt,
	shared_ptr<expression> expr ) {
	return apply_loss(cvt, expr);
}
		
shared_ptr<expression> conversion_applicator::apply_loss( cvt_expression & cvt,
	shared_ptr<expression> expr ) {
	
	shared_ptr<intr_type_fixed_number> dst;
	switch( cvt.to_fun ) {
		case cvt_expression::p_float:
			dst = make_shared<intr_type_float>();
			break;
		case cvt_expression::p_integer:
			dst = make_shared<intr_type_integer>();
			break;
		case cvt_expression::p_binary:
			dst = make_shared<intr_type_binary>();
			break;
	}
	dst->data_bitsize = cvt.to_bitsize;
	dst->local_complete = true;
	
	auto promote = make_shared<expr_transform>();
	promote->type = extr_type(dst);
	promote->arg = expr;
	promote->transform = cvt.form == cvt_expression::c_promote ?
		expr_transform::t_promote : expr_transform::t_loss;
	promote->mark_resolved();
	return promote;
}
		
shared_ptr<expression> conversion_applicator::apply_fixate( cvt_expression & cvt,
	shared_ptr<expression> expr ) {
		
	//fixation results in the removal of the rational constant and conversion
	//to an integer or float value
	shared_ptr<expression> use_expr = expr;
	
	if( use_expr->form == expression::f_symbol ) {
		auto & es = static_cast<expr_symbol&>(*expr);
		STATE_CHECK( es.type.extr().is_literal() );
		STATE_CHECK( es.sref && es.sref->literal_value );
		use_expr = es.sref->literal_value;
	}
	STATE_CHECK( use_expr->form == expression::f_literal_number );
	auto & cn = static_cast<expr_literal_number &>(*use_expr);

	auto num = make_shared<expr_literal_number>();
	shared_ptr<intr_type_fixed_number> ft;
	switch( cvt.to_fun ) {
		case cvt_expression::p_float:
			ft = make_shared<intr_type_float>();
			num->src_form = expr_literal_number::sf_float;
			break;
		case cvt_expression::p_integer:
			ft = make_shared<intr_type_integer>();
			num->src_form = expr_literal_number::sf_integer;
			break;
		case cvt_expression::p_binary:
			ft = make_shared<intr_type_binary>();
			num->src_form = expr_literal_number::sf_binary;
			break;
	}
	
	//create resulting expression
	ft->data_bitsize = cvt.to_bitsize;
	auto ex = extr_type(ft);
	ex.literal = true;
	num->type = ex;
	num->value = cn.value;
	num->mark_resolved();
	return num;
}
		
shared_ptr<expression> conversion_applicator::apply_bind_raw_func( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
		
	auto wrap = make_shared<expr_bind>();
	wrap->unbound = expr;
	wrap->construct_result_type();
	wrap->mark_resolved();
	return wrap;
}
		
shared_ptr<expression> conversion_applicator::apply_bind_func( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
		
	auto b = expression_typer::bind_function(expr, ctx );
	//hope it can resolve correctly now
	auto out = expression_typer::process_expression(b, ctx);
	DEBUG( "APPLY_BIND", dump::get(b), dump::get(out) );
	return out;
}
		
shared_ptr<expression> conversion_applicator::apply_collapse( cvt_expression & cvt,
	shared_ptr<expression> expr ) {

	cvt.field = 0;
	return apply_field(cvt, expr);
}

			
shared_ptr<expression> conversion_applicator::apply_field( cvt_expression & cvt,
	shared_ptr<expression> expr ) {
			
	//this is not just an optimization now, conversions of function returns force us to make value_lists
	//real things at the compiler level (they become tuples)
	if( expr->form == expression::f_value_list
		|| expr->form == expression::f_compound ) {
		auto & evl = static_cast<expr_with_sub&>(*expr);
		STMT_CHECK( *expr, cvt.field < evl.sub.size() );
		expr = evl.sub[cvt.field].expr;
	} else if( expr->is_resolved() ) {
		auto sub = make_shared<expr_field>();
		sub->base = expr;
		sub->field_index = cvt.field;
		
		intr_type_compound const * vtc = expr->type.extr().if_cast<intr_type_value_list>();
		if( !vtc ) {
			vtc = expr->type.extr().if_cast<intr_type_tuple>();
			STMT_CHECK( *expr, vtc );
		}
		sub->type = vtc->sub[cvt.field].type;
		sub->mark_resolved();
		expr = sub;
	}
	return expr;
}
		
shared_ptr<expression> conversion_applicator::apply_direct( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
	//nothing
	return expr;
}
			
shared_ptr<expression> conversion_applicator::apply_pack_value( cvt_expression & cvt,
	shared_ptr<expression> expr ) {
	return apply_pack(cvt,expr);
}
		
shared_ptr<expression> conversion_applicator::apply_pack_contextual( cvt_expression & cvt,
	shared_ptr<expression> expr ) {
	return apply_pack(cvt,expr);
}
		
shared_ptr<expression> conversion_applicator::apply_pack( cvt_expression & cvt,
	shared_ptr<expression> expr ) {

	shared_ptr<expr_with_sub> pack;
	shared_ptr<intr_type_compound> pack_type;
	bool pack_value;
	
	if( cvt.form == cvt_expression::c_pack_value ) {
		pack = make_shared<expr_value_list>();
		auto pt = make_shared<intr_type_value_list>();
		pack_type = pt;
		pack_value = true;
	} else if( cvt.form == cvt_expression::c_pack_contextual ) {
		pack = make_shared<expr_compound>();
		auto pt = make_shared<intr_type_contextual>();
		pt->sub.clear();
		pack_type = pt;
		pack_value = false;
	} else {
		pack = make_shared<expr_compound>();
		auto pt = make_shared<intr_type_tuple>();
		pt->pack = cvt.pack;
		pack_type = pt;
		pack_value = false;
	}
	//This is technically an optimization now, the type could be evaluated on a second pass (see the sub->is_complete below)
	pack->type = extr_type(pack_type);
	bool complete = true;
		
	//must avoid double evaluation if the expression may have side-effects
	//the shallow flag assumes the pack's are strict, no out-of-order items!
	bool is_literal = expr->type.extr().is_literal();
	auto shared_expr = create_shared( expr, true );
	
	for( size_t i=0; i < cvt.sub.size(); ++i ) {
		auto sub = apply( *cvt.sub[i].cvt, shared_expr );
		if( !sub->is_resolved() ) {
			complete = false;
			pack->type = type_ref();
		}
		
		if( !pack_value ) {
			sub->type.extr().literal = false;
		}
		pack_type->push_sub( cvt.sub[i].name, sub->type );
		pack->push_sub( cvt.sub[i].name, sub );
	}
	if( complete) {
		pack->mark_resolved();
	}
	//tuples are either are literal or not, value-lists are component wise literal
	if( !pack_value ) {
		pack->type.extr().literal = is_literal;
	}
	return pack;
}
		
shared_ptr<expression> conversion_applicator::apply_optional( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {

	STATE_CHECK( !expr->type.extr().is_optional() );
	auto opt = make_shared<expr_transform>();
	opt->transform = expr_transform::t_make_optional;
	opt->arg = expr;
	opt->construct_result_type();
	opt->mark_resolved();
	return opt;
}
		
shared_ptr<expression> conversion_applicator::deref_optional( shared_ptr<expression> expr ) {
	STATE_CHECK( expr->type.extr().is_optional() );
	auto opt = make_shared<expr_transform>();
	opt->transform = expr_transform::t_deref_optional;
	opt->arg = expr;
	opt->construct_result_type();
	opt->mark_resolved();
	return opt;
}
			
shared_ptr<expression> conversion_applicator::apply_deref_optional( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
		
	return deref_optional(expr);
}
		
shared_ptr<expression> conversion_applicator::apply_unset_optional( cvt_expression & cvt,
	shared_ptr<expression> /*expr*/ ) {

	auto d = make_shared<expr_temp>();
	d->init = expr_temp::i_default;
	d->type = cvt.target_type;
	d->mark_resolved();
	return d;
}
		
shared_ptr<expression> conversion_applicator::apply_set_distinct( cvt_expression & cvt,
	shared_ptr<expression> expr ) {
		
	//if we're constructing the type we can just set the distinct target during construction
	if( expr->form == expression::f_value_list
		|| expr->form == expression::f_compound ) {
		
		//TODO: see the cvt_expression todo, this should be only a fun_class
		expr->type.extr().intr = cvt.target_type.extr().intr;
		expr->mark_resolved();
		return expr;
	} else if( expr->type.extr().is<intr_type_tuple>() ) {
		//rebuild the tuple since we have no way to just magically switch types (I tried
		//a reinterpret cast, but there's no magical switching at the LLVM level possible)
		auto & tc = expr->type.extr().cast<intr_type_compound>();
		auto pack = make_shared<expr_compound>();
		pack->type = cvt.target_type;

		for( size_t i=0; i < tc.sub.size(); ++i ) {
			auto sub = make_shared<expr_field>();
			sub->base = expr;
			sub->field_index = i;
			sub->type = tc.sub[i].type;
			sub->mark_resolved();
			pack->push_sub( tc.sub[i].name, sub );
		}
		pack->mark_resolved();
		return pack;
		
	} else {
		STATE_FAIL("unsupported");
	}
}
		
shared_ptr<expression> conversion_applicator::apply_bind_context( cvt_expression & /*cvt*/,
	shared_ptr<expression> expr ) {
	
	STATE_CHECK( expr->type.extr().is<intr_type_contextual>() );
	
	auto b = make_shared<expr_bind>();
	auto & ct = expr->type.extr().cast<intr_type_contextual>();

	//needed for multi-parametrics in classes 
	//(TODO: I'm starting not to like the "multi" lookup system,  a special op is needed)
	if( expr->form == expression::f_compound ) {
		auto & comp = static_cast<expr_compound&>(*expr);
		STATE_CHECK( comp.sub.size() == 2 );
		b->context = comp.sub[0].expr;
		b->unbound = comp.sub[1].expr;
	
	} else {
		auto cn = create_shared( expr );
		
		auto sf_context = make_shared<expr_special_field>();
		sf_context->base = cn;
		sf_context->field = expr_special_field::f_contextual_context;
		sf_context->type = ct.get_context();
		if (sf_context->type.is_resolved()) {
			sf_context->mark_resolved();
		}
		b->context = sf_context;
		
		auto sf_unbound = make_shared<expr_special_field>();
		sf_unbound->base = cn;
		sf_unbound->field = expr_special_field::f_contextual_unbound;
		sf_unbound->type = ct.get_unbound();
		if (sf_unbound->type.is_resolved()) {
			sf_unbound->mark_resolved();
		}
		b->unbound = sf_unbound;
	}
	
	if( ct.get_unbound().extr().is_lvalue() ) {	
		b->unbound = drop_lvalue( b->unbound );
	}
	
	STATE_CHECK( ct.get_unbound().extr().is<intr_type_function>() );
	auto ft = ct.get_unbound().extr().cast<intr_type_function const>().clone_derived();
	ft->context = type_ref();
	ft->access = intr_type_function::a_bound;
	b->type = extr_type(ft);
	if(b->type.is_resolved()) {
		b->mark_resolved();
	}
	return b;
}
		
shared_ptr<expression> conversion_applicator::apply_make_compat( cvt_expression & cvt,
	shared_ptr<expression> expr ) {

	auto c = make_shared<expr_compat>();
	c->type = cvt.target_type;
	c->source = expr;
	c->mark_resolved();
	return c;
}
		
shared_ptr<expression> conversion_applicator::apply_param_inst( cvt_expression & cvt,
	shared_ptr<expression> expr ) {

	STATE_CHECK( expr->form == expression::f_funcdefn );
	auto & func = static_cast<expr_funcdefn&>(*expr);
	auto clone = ctx.instance_funcdefn(func, cvt.target_type);
	ctx.rewrite( expr, clone );

	return clone;
}

shared_ptr<expression> conversion_applicator::apply( cvt_expression & cvt_, shared_ptr<expression> expr )
{
	auto out = expr;
	auto cvt = &cvt_;
	do
	{
		if (!out->type.is_complete() && !handles_incomplete(cvt->form) ) {
			DEBUG( "CVT_SINGLE", typeid(*expr).name(), dump::get(out),
				dump::get(cvt), dump::get(out->type) );
			//return expr;
			return out;
		}
	
		out = apply_single( *cvt, out );
		cvt = cvt->then.get();
	} while( cvt );

	DEBUG( "APPLY", dump::get(expr->type), expr->type.is_complete(),
		dump::get(out->type), out->type.is_complete() );
		
	return out;
}

/**
	Convert the expression so it can be used as a shared expression. This shares only the
	parts that actually need to be shared.
*/
shared_ptr<expression> conversion_applicator::create_shared( shared_ptr<expression> expr,
	bool shallow_list )
{
	//TODO: why not?
	//STMT_CHECK( *expr, expr->is_resolved() );
	
	if( expr->form == expression::f_value_list
		|| expr->form == expression::f_compound ) {
		
		//in shallow cases we know for sure that we don't have to share the individual parts, this
		//is required for proper fixation in function calls
		if( shallow_list ) {
			return expr;
		}
		
		STATE_FAIL( "unexpected" ); // will this ever be needed...?
		//this greatly helps reduce wasted temporaries by allowing field and subscript to 
		//bypass a packed structure.
		auto & evl = static_cast<expr_with_sub&>(*expr);
		for( size_t i=0; i < evl.sub.size(); ++i ) {
			evl.sub[i].expr = create_shared( evl.sub[i].expr );
		}
	} else {
		//OPT: perhaps also just complex/large expressions should always be shared
		//TODO: must rethink this, the "true" is required since a shared expression cannot
		//have it's type changed (which something like drop_lvalue does). This basically
		//removes any chance to ever get rid of the shared expressiosn
		if( true || expr->may_have_side_effects() ) {
			auto shared_expr = make_shared<expr_transform>();
			shared_expr->transform = expr_transform::t_shared_expr;
			shared_expr->arg = expr;
			shared_expr->construct_result_type();
			shared_expr->mark_resolved();
			expr = shared_expr;
		}
	}
	
	return expr;
}

/**
	This function will modify the provided "expr", it does not create a clone or other
	derived expression.
*/
shared_ptr<expression> conversion_applicator::drop_lvalue( shared_ptr<expression> expr ) {
	if( expr->form == expression::f_symbol ) {
		DEBUG("DROP-LVALUE", dump::get(*expr), dump::get(expr->type));
		STMT_CHECK( *expr, expr->is_resolved() );
		STMT_CHECK( *expr, expr->type.extr().is_lvalue() );
		expr->type.extr().lvalue = false;
		expr->mark_resolved();
		return expr;
	} else if( expr->form == expression::f_value_list ) {
		auto & vl = static_cast<expr_value_list&>(*expr);
		for( auto & s : vl.sub )
			s.expr = drop_lvalue(s.expr);
		return expr;
	} /* TODO: shared bases also resulti n lvalue, thus lvalue can't be dropped there
		else if( expr->form == expression::f_field ) {
		auto & ef = static_cast<expr_field&>(*expr);
		ef.base = drop_lvalue( ef.base );
		auto cl = expr->type;
		cl.set_lvalue( false );
		expr->type = cl;
		return expr;
	} */ else if( expr->form == expression::f_transform ) {
	#if 1
		//TODO: this is perhaps an optimization. There is one test which depends on it, but it correlates
		//to a TODO in the compiler for a context call. Something suspicious
		auto & et = static_cast<expr_transform&>(*expr);
		//these types can only result in lvalue if the inner part is an lvalue
		if( et.transform == expr_transform::t_deref 
			|| et.transform == expr_transform::t_deref_optional ) {
			STMT_CHECK( *expr, et.arg->type.extr().is_lvalue() );
			et.arg = drop_lvalue( et.arg );
			
			et.type.extr().lvalue = false;
			return expr;
		}
	#endif
	}
	
	auto deref = make_shared<expr_transform>();
	deref->transform = expr_transform::t_drop_lvalue;
	deref->arg = expr;
	deref->construct_result_type();
	deref->mark_resolved();
	return deref;
}

} //eon leaf