#include "lang/dump.hpp"
#include "lang/language.hpp"
#include "lang/type_converter.hpp"
#include "parser/dump.hpp"
#include "parser/symbols.hpp"
#include "util/unicode.hpp"

#include <boost/algorithm/string.hpp>

#define SHOW_COMPLETENESS false

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf
{

dump::dump( std::ostream & out, unsigned flags )
	: leaf_util::dump_base( out )
{
	do_validate = true;
	show_full_type = false;
	show_all_types = false;
	show_spec_type = false;
	hide_var_spec_no_fundamental = false;
	keep_variant = false;
	show_spec_implicit = false;
	annotate = flags & tf_annotate;
	expand_defined = flags & tf_expand_defined;
}

dump::dump( unsigned flags )
	: leaf_util::dump_base()
{
	do_validate = true;
	show_full_type = false;
	show_all_types = false;
	show_spec_type = false;
	hide_var_spec_no_fundamental = false;
	keep_variant = false;
	show_spec_implicit = false;
	annotate = flags & tf_annotate;
	expand_defined = flags & tf_expand_defined;
}

#define CHECK_NULL( v, msg ) if( !v ) { error( "null "  msg ); return; }

#include "dump/expression.inc.cpp"
#include "dump/statement.inc.cpp"
#include "dump/type.inc.cpp"
#include "dump/type_spec.inc.cpp"

template<typename T>
void dump::validate( T const & o )
{
	if( !do_validate )
		return;
		
	try
	{
		util::validate( o );
	}
	catch( recoverable_error & re )
	{
		std::string err = boost::current_exception_diagnostic_information();
		boost::trim( err );
		boost::replace_all( err, "\n", std::string("\n") + indent_str() + "// " );
		out << "\e[31m";
		out << indent_str() << "// " << err << std::endl;
		out << "\e[0m";
	}
}

void dump::show_type( type_ref type, type_spec const * spec )
{
	if( show_spec_type ) 
		dump_type_spec( spec );
	else
		dump_type_ref( type );
}


void dump::dump_outline_types() {
	std::set<intr_type const*> ty;
	ty.swap( outline_types );
	for( auto t : ty ) {
		std::vector<std::string> markers;
		out << std::endl;
		dump_intr_type( *t, markers, true );
	}
}

void dump::write_cvt_expression( cvt_expression const * ce_ ) {
	CHECK_NULL( ce_, "cvt_expression" )
	auto & ce = *ce_;
	
	bool force_sub = false;
	switch( ce.form )
	{
		#define CODE(tag) \
			case cvt_expression::tag: \
				out << #tag; break;
		CODE(c_addr_of)
		CODE(c_drop_lvalue)
		CODE(c_deref)
		CODE(c_bind_raw_func)
		CODE(c_bind_func)
		CODE(c_bind_context)
		CODE(c_pack_value)
		CODE(c_pack_contextual)
		CODE(c_direct)
		CODE(c_collapse)
		CODE(c_optional)
		CODE(c_deref_optional)
		CODE(c_unset_optional)
		CODE(c_parametric)
		CODE(c_drop_literal)
		#undef CODE
		
		case cvt_expression::c_set_distinct: {
			out << "c_set_distinct(";
			if( ce.target_type.extr().intr->has_distinct_name() ) {
				out << ce.target_type.extr().intr->get_distinct_name();
			} else {
				out << "*anonymous*";
			}
			out << ")";
			break;
		}
		
		case cvt_expression::c_pack: {
			out << "c_pack";
			switch( ce.pack ) {
				case intr_type_tuple::p_normal: break;
				case intr_type_tuple::p_abi: out << "(abi)"; break;
				case intr_type_tuple::p_byte: out << "(byte)"; break;
			}
			force_sub = true;
			break;
		}
		case cvt_expression::c_promote:
		case cvt_expression::c_fixate: 
		case cvt_expression::c_loss: {
			if( ce.form == cvt_expression::c_promote )
				out << "c_promote(";
			else if( ce.form == cvt_expression::c_fixate )
				out << "c_fixate(";
			else
				out << "c_loss(";

			if( ce.to_fun == cvt_expression::p_float ) {
				auto nf = platform::if_get_float( ce.to_bitsize );
				STATE_CHECK(nf);
				switch( nf->precision ) {
					case -1: out << "low"; break;
					case 1: out << "high"; break;
					case 0: out << "normal";  break;
					default: STATE_FAIL();
				}
			} else { 
				if( ce.to_bitsize != platform::natural_integer_size )
					out << ce.to_bitsize << "bit";
				else
					out << "normal";
			}
			out << " ";
			
			auto fun_print = [&]( cvt_expression::fundamental_t p ) {
				switch( p ) {
					case cvt_expression::p_integer:
						out << "integer";
						break;
					case cvt_expression::p_binary:
						out << "binary";
						break;
					case cvt_expression::p_float:
						out << "float";
						break;
				}
			};
			
			if( ce.form == cvt_expression::c_fixate ) {
				fun_print( ce.to_fun );
			} else {
				fun_print( ce.from_fun );
				if( ce.from_fun != ce.to_fun ) {
					out << "_to_";
					fun_print( ce.to_fun );
				}
			}
			out << ")";
			break;
		}

		case cvt_expression::c_field:
			out << "[" << ce.field << "]";
			break;
			
		case cvt_expression::c_make_compat:
			out << "c_make_compat｢";
			dump_type_ref(ce.target_type);
			out << "｣";
			break;
			
		case cvt_expression::c_param_inst:
			out << "c_param_inst｢";
			dump_type_ref(ce.target_type);
			out << "｣";
			break;
	}
	
	if( ce.sub.size() || force_sub ) {
		out << " {";
		for( size_t i=0; i < ce.sub.size(); ++i ) 
		{
			if( i )
				out << " ";
			if( !ce.sub[i].name.empty() )
				out << ce.sub[i].name << "=";
			write_cvt_expression( ce.sub[i].cvt.get() );
		}
		out << "}";
	}
	
	if( ce.then ) {
		out << "->";
		write_cvt_expression( ce.then.get() );
	}
}

void dump::write_module( module const & mod ) {
	if( mod.name ) {
		out << "// Module Name: " << *mod.name << std::endl;
	}
	
	if( mod.libraries.size() ) {
		out << "// Libraries: ";
		for( auto & lib : mod.libraries ) {
			out << lib << " ";
		}
		out << std::endl;
	}
	
	out << "// Global Scope" << std::endl;
	for( auto & v : mod.global->get_local_decls() ) {
		dump_declaration( v.get() );
		out << std::endl << std::endl;
	}
	
	if( mod.var_exports.size() ) {
		out << "@exports {" << std::endl;
		indent();
		for( auto & ex : mod.var_exports ) {
			out << indent_str();
			write_declaration( *ex );
			out << std::endl;
		}
		unindent();
		out << "}" << std::endl;
	}
	
	out << "@static_init ";
	dump_statement_block( mod.static_init_block.get() );
	out << std::endl;
	
	for( auto & block : mod.blocks ) {
		dump_statement_block( block.get() );
		out << std::endl;
	}
	
	dump_outline_types();
	out << std::endl;
}

} //eon leaf