void dump::dump_assign_statement( assign_statement const * as ) {
	CHECK_NULL(as,"assign_statement")
	dump_expression( as->left.get() );
	switch( as->how ) {
		case assign_statement::h_value: out << " = "; break;
		case assign_statement::h_init_refer: out << " := "; break;
		case assign_statement::h_default: out << " |= "; break;
	}
	dump_expression( as->right.get() );
}

void dump::dump_declaration( declaration const * decl, unsigned flags ) {
	CHECK_NULL( decl, "declaration" );
	
	if( decl->is_export ) {
		out << "@export ";
	}
	
	switch( decl->lifetime ) {
		case declaration::l_local: break;
		case declaration::l_global: out << "@lifetime(global) "; break;
	}
	
	if( decl->import_type != declaration::it_none) {
		switch( decl->import_type ) {
			case declaration::it_none: break;
			case declaration::it_abi_code: out << "@import("; break;
			case declaration::it_abi_data: out << "@import_data("; break;
		}
		if( decl->import_module.size() ) {
			out << decl->import_module << ".";
		}
		out << decl->import_name;
		switch( decl->name_form ) {
			case declaration::nf_bare: out << ",bare"; break;
			case declaration::nf_leaf: out << ",leaf"; break;
			case declaration::nf_auto: out << ",auto"; break;
		}
		out << ") ";
	}
	
	switch( decl->form ) {
		case declaration::f_single: out << "single "; break;
		case declaration::f_multi: out << "multi "; break;
		case declaration::f_arg: out << "arg "; break; //internal (not in code directly)
		case declaration::f_return: out << "return "; break; //internal
		case declaration::f_capture: out << "capture "; break; //internal
	}
	
	out << decl->symbol;

	if( show_spec_type ) {
		if( decl->spec_type ) {
			out << " : ";
			dump_type_spec( decl->spec_type.get() );
		}
	} else {
		out << " : ";
		//the decl itself need not be valid, only the values (relevant for class parametric functions)
		dump_type_ref( decl->type, tf_no_validate );
	}

	if( decl->literal_value && !omit_value && !(flags & df_omit_value) ) {
		out << " = ";
		dump_expression( decl->literal_value.get(), ef_expand_name );
	}
}

void dump::dump_var_statement( var_statement const * vs ) {
	CHECK_NULL( vs, "var_statement" )
	
	dump_declaration( vs->decl.get() );
	
	if( vs->value ) {
		if( vs->init_refer_value )
			out << " := ";
		else
			out << " = ";
		dump_expression( vs->value.get() );
	}
}

void dump::dump_init_statement( init_statement const * is ) {
	CHECK_NULL( is, "init_statement" )
	auto decl = is->sref;

	out << "init ";
	dump_declaration( decl.get(), df_omit_value );

	if( is->undefined ) {
		out << " = undefined";
	}
		
	if( is->value ) {
		if( is->init_refer_value )
			out << " := ";
		else
			out << " = ";
		dump_expression( is->value.get(), ef_expand_name );
	}
	
	if( is->no_release ) {
		out << " no_release";
	}
		
	if( is->code ) {
		out << "{";
		out << endl();
		indent();
		dump_statement( is->code.get() );
		unindent();
		out << indent_str() << "}" << endl();
	}
}

void dump::dump_defer_statement( defer_statement const * defer ) {
	CHECK_NULL( defer, "defer_statement" )
	out << (defer->error_only ? "defer_fail" : "defer");
	dump_statement_block( defer->error.on_error.get() );
}

void dump::dump_do_statement( do_statement const * ds ) {
	CHECK_NULL( ds, "do_statement" )
	if( ds->loop ) {
		out << "loop ";
	} else {
		out << "do ";
	}

	out << "switch {" << endl();
	indent();
	for( auto & v : ds->expr_list ) {
		out << indent_str();
		if( v.cond ) {
			dump_expression( v.cond.get(), ef_isolated );
			out << " then ";
		} else {
			out << "else";
		}
		dump_statement_block( v.block.get() );
		out << endl();
	}
	unindent();
	out << indent_str() << "}";
}

void dump::dump_statement( statement const * st ) {
	CHECK_NULL(st, "statement")
	validate(*st);
	
	out << indent_str();
	dump_statement_inner( st );
	out << endl();
}

void dump::dump_statement_inner( statement const * st ) {
	CHECK_NULL(st, "statement")
	switch( st->statement_form() )
	{
		#define STATEMENT_FORM(type) \
			case sf_##type##_statement: \
				dump_##type##_statement( static_cast<type##_statement const*>(st) ); \
				break;
		STATEMENT_FORMS
		#undef STATEMENT_FORM
	}
}

void dump::dump_statement_block( statement_block const * block ) {
	CHECK_NULL(block, "statement_block")
	
	//prevent recursion
	if( shown.find(block) != shown.end() ) {
		out << "...";
		return;
	}
	shown.insert( block );
	
	if( block->reference_text.size() ) {
		out << "/* " << block->reference_text << " */ ";
	}
	out << "{" << endl();
	indent();
	
	//global symbols will already be in the static init block...?
	if( block->symbols->get_level() > 0 ) {
		out << indent_str() << "// local_decls" << endl();
		auto const & locals = block->symbols->get_local_decls(true);
		for( auto & local : locals ) {
			out << indent_str();
			dump_declaration( local.get() );
			out << endl();
		}
		out << indent_str() << "// end local_decls" << endl();
	}
	
	for( size_t i=0; i < block->list.size(); ++i ) {
		dump_statement( block->list[i].get() );		
		if( SHOW_COMPLETENESS && !block->list[i]->is_resolved() ) {
			out << " //UNRESOLVED";
		}
	}
	
	unindent();
	out << indent_str() << "} ";
	
	if( SHOW_COMPLETENESS && !block->is_complete() ) {
		out << " /* INCOMPLETE */";
	}
}

void dump::dump_typedef_statement( typedef_statement const * ts ) {
	CHECK_NULL(ts,"typedef_statement" );
	
	if( ts->is_export ) {
		out << "@export ";
	}
	
	switch( ts->what ) {
		case typedef_statement::w_prototype:
			out << "typedef " << ts->name;
			if( show_spec_type ) {
				if( ts->spec_type ) {
					out << " : ";
					dump_type_spec( ts->spec_type.get() );
				}
			} else {
				out << " : ";
				dump_type_ref( ts->type, true );
			}
			break;
			
		case typedef_statement::w_alias: {
			out << "alias " << ts->name << " : ";
			dump_type_spec( ts->spec_type.get() );
			break;
		}

		case typedef_statement::w_service:
		case typedef_statement::w_class: {
			out << (ts->what == typedef_statement::w_class ? "class " : "service ") << ts->name << " ";
			if( ts->spec_type ) {
				dump_type_spec( ts->spec_type.get() );
			}
			dump_statement_block( ts->class_block.get() );
			break;
		}
	};
	
}

void dump::dump_return_statement( return_statement const * rs ) {
	CHECK_NULL(rs,"return_statement" );
	
	out << "return";
	
	if( rs->value ) {
		out << " ";
		dump_expression( rs->value.get(), ef_isolated );
	}
}

void dump::dump_expression_statement( expression_statement const * es ) {
	CHECK_NULL( es, "expression_statement" );
	if( es->ignore ) {
		out << "ignore ";
	}
	dump_expression( es->expr.get(), ef_isolated );
}

void dump::dump_fail_statement( fail_statement const * fs ) {
	CHECK_NULL( fs, "fail_statement" );
	switch( fs->level ) {
		case fail_level::resume: out << "resume_fail"; break;
		case fail_level::pre: out << "pre_fail"; break;
		case fail_level::normal: out << "fail"; break;
		case fail_level::critical: out << "critical_fail"; break;
	}
}

void dump::dump_loop_flow_statement( loop_flow_statement const * ls ) {
	CHECK_NULL( ls, "loop_flow_statement" );
	switch( ls->flow ) {
		case loop_flow_statement::f_break: out << "break"; break;
		case loop_flow_statement::f_next: out << "next"; break;
	}
}

void dump::dump_noop_statement( noop_statement const * ns ) { 
	CHECK_NULL( ns, "noop_statement" );
	out << "noop";
}

void dump::dump_import_statement( import_statement const * is ) {
	CHECK_NULL( is, "import_statement" );
	out << "import " << is->mod_name;
}

void dump::dump_for_statement( for_statement const * fs ) {
	CHECK_NULL( fs, "for_statement" );
	out << "for " << fs->item << " in ";
	dump_expression( fs->iterable.get() );
	out << " ";
	dump_statement_block( fs->block.get() );
}

void dump::dump_error_statement( error_statement const * es ) {
	CHECK_NULL( es, "error_statement" );
	dump_statement_inner( es->inner.get() );
	
	out << " on ";
	switch( es->error.on_match ) {
		case fail_level::resume: out << "resume_fail"; break;
		case fail_level::pre: out << "pre_fail"; break;
		case fail_level::normal: out << "fail"; break;
		case fail_level::critical: out << "critical_fail"; break;
	}
	out << " ";
	dump_statement_block( es->error.on_error.get() );
}
