template<typename T>
void dump::dump_intr_type_tuple_inner( const char * o, const char * c, T const * vt, bool nameless ) {
	CHECK_NULL( vt, "intr_type(T)" )
	
	out << o;
	bool need_space = true; //to deal with empty types
	for( size_t i=0; i < vt->sub.size(); ++i )
	{
		if( i )
			out << ",";
		out << " ";

		auto & sub = vt->sub[i];
		if( !nameless ) {
			if( sub.name.size() )
				out << sub.name << " : ";
			else
				out << ": ";
		}
		auto pos = out.tellp();
		_dvtti_sub( sub.type );
		need_space = pos != out.tellp();
	}
	if( need_space )
		out << " ";
	out << c;
}

void dump::write_impl( intr_type const * vt ) {
	CHECK_NULL( vt, "intr_type" )
	std::vector<std::string> markers;
	dump_intr_type( *vt, markers );
}

void dump::dump_type_ref( type_ref const & tr, unsigned flags ) {
	if( annotate ) {
		out << "⦃";
	}
	
	if( tr.is_extr() ) {
		if( annotate ) {
			out << "tr-extr: ";
		}
		dump_extr_type( tr.extr(), flags );
	} else if( tr.is_spec() ) {
		if( annotate ) {
			out << "tr-spec: ";
		}
		dump_type_spec( &tr.spec() );
	} else {
		error( "undefined-type" );
	}
	
	if(annotate) {
		out << "⦄";
	}
}

void dump::dump_extr_type( extr_type const & tr, unsigned flags ) {
	CHECK_NULL( tr.intr.get(), "intr_type" )
	
	//show all markers/labels in alphabetical order (predictable for tests)
	std::vector<std::string> markers;
	
	//don't show rvalue, just lvalue
	if( tr.is_lvalue() )
		markers.push_back( "lvalue" );
	if( tr.is_shared() )
		markers.push_back( "shared" );
	if( tr.is_value_ptr() )
		markers.push_back( "value_ptr" );
	if( tr.is_optional() )
		markers.push_back( "optional" );
	if( tr.is_cvt_optional() )
		markers.push_back( "convert_optional" );
	if( tr.is_lossy() )
		markers.push_back( "lossy" );
	if( tr.is_literal() )
		markers.push_back( "literal" );
	
	dump_intr_type( *tr.intr, markers, flags );
}

void dump::dump_intr_type( intr_type const & vt, unsigned flags ) {
	std::vector<std::string> markers;
	dump_intr_type( vt, markers, flags );
}

void dump::dump_intr_type( intr_type const & vt, std::vector<std::string> & markers, unsigned flags ) {
	if( !(flags & tf_no_validate) ) {
		validate(vt);
	}
	
	if( vt.is_export && expand_defined ) {
		out << "@export ";
	}

	bool prefix = false;
	if( (vt.is_whole() || vt.is_instance()) && expand_defined ) {
		out << "typedef "; //could also be class... actually more likely
		prefix = true;
	}
	
	auto show_distinct = [&]( intr_type const & vt ) {
		if( vt.has_distinct_name() ) {
			if( vt.name_scope ) {
				out << vt.name_scope->qualify_name( vt.get_distinct_name(), "." );
			} else {
				out << "*namespace*." << vt.get_distinct_name();
			}
			outline_types.insert( &vt );
		} else if ( !vt.instance_of ) {
			out << "*distinct*";
		}
	};
	
	if( vt.is_whole() ) {
		show_distinct(vt);
		prefix = true;
	} 
	
	//TODO: special array case to maintain old tests only... need to clarify distinct names
	bool skip_params = false;//vt.is<intr_type_array>();
	
	if( !skip_params ) {
		if( vt.instance_of ) {
			out << "^";
			show_distinct(*vt.instance_of);
			//dump_intr_type( *vt.instance_of );
		}
			
		if( vt.applied_params.size() ) {
			out << "｢ ";
			for( size_t i=0; i < vt.applied_params.size(); ++i ) {
				if( i ) {
					out << ", ";
				}
				dump_type_ref( vt.applied_params[i] );
			}
			out << " ｣";
		}
	}
	
	if( (!vt.is_whole() && !vt.instance_of) || expand_defined ) {
		
		if( prefix ) {
			out << " = ";
		}
		
		if( vt.params.size() ) {
			out << "｢ ";
			for( size_t i=0; i < vt.params.size(); ++i ) {
				if( i ) {
					out << ", ";
				}
				out << vt.params[i].name;
			}
			out << " ｣ ";
		}
		
		switch( vt.get_fun_class() ) {
				
			case intr_type::f_char:
				out << "char";
				break;

			case intr_type::f_tuple:
			{
				auto & tt = vt.cast<intr_type_tuple>();
				switch( tt.pack ) {
					case intr_type_tuple::p_normal: break;
					case intr_type_tuple::p_abi: markers.push_back( "abi_pack"); break;
					case intr_type_tuple::p_byte: markers.push_back( "byte_pack" ); break;
				}
				if( tt.selectable ) {
					markers.push_back( "selectable" );
				}
				dump_intr_type_tuple_inner( "[", "]", &tt );
				break;
			}

			case intr_type::f_value_list:
				dump_intr_type_tuple_inner( "(", ")", &vt.cast<intr_type_value_list>() );
				break;
				
			case intr_type::f_function:
			{
				auto & tf = vt.cast<intr_type_function>();
				dump_intr_type_tuple_inner( "(", ")", tf.args.get() );
				out << " -> ";
				if( tf.returns ) {
					dump_intr_type_tuple_inner( "(", ")", tf.returns.get() );
				} else {
					out << "$implicit";
				}
				switch( tf.access ) {
					case intr_type_function::a_bound: 
						break;
					case intr_type_function::a_raw: 
						markers.push_back( "raw" ); 
						break;
					case intr_type_function::a_unbound: 
						markers.push_back( "unbound" ); 
						break;
					case intr_type_function::a_construct:
						if( tf.construct_result.is_defined() ) {
							markers.push_back( std::string( "construct(" ) +
								get( tf.construct_result )
								+ ")" );
						} else {
							markers.push_back( "construct" );
						}
						break;
				}
				if( tf.context.is_defined() ) {
					out << " context(";
					dump_type_ref( tf.context );
					out << ")";
				}
				if( !tf.can_fail ) {
					out << " no_throw";
				}
				break;
			}
				
			case intr_type::f_rational: {
				auto & fr = vt.cast<intr_type_rational>();
				out << "rational";
				if( fr.integer_signed ) {
					markers.push_back( "signed" );
				}
				if( fr.integer_bitsize ) {
					char bits[128];
					sprintf( bits, "%dbit", *fr.integer_bitsize );
					markers.push_back( bits );
				}
				break;
			}
				
			case intr_type::f_integer: {
				auto it = vt.cast<intr_type_integer>();
				out << "integer";
				if( keep_variant || it.data_bitsize != platform::natural_integer_size ) {
					char bits[128];
					sprintf( bits, "%dbit", it.data_bitsize );
					markers.push_back( bits );
				}
				break;
			}
				
			case intr_type::f_binary: {
				auto it = vt.cast<intr_type_binary>();
				out << "binary";
				if( keep_variant || it.data_bitsize != platform::natural_integer_size ) {
					char bits[128];
					sprintf( bits, "%dbit", it.data_bitsize );
					markers.push_back( bits );
				}
				break;
			}
			
			case intr_type::f_float: {
				out << "float";
				auto ft = vt.cast<intr_type_float>();
				auto nf = platform::if_get_float( ft.data_bitsize );
				if( !nf ) {
					error( "invalid-size-variant" );
				} else if( nf->precision != 0 ) {
					if( nf->precision == -1 )
						markers.push_back("low");
					else if( nf->precision == 1 )
						markers.push_back("high" );
					else {
						STATE_FAIL("unexpected now");
					}
				} else if( keep_variant ) {
					markers.push_back( "normal" );
				}
					
				break;
			}
				
			case intr_type::f_boolean:
				out << "boolean";
				break;
		
			case intr_type::f_array: {
				auto & at = vt.cast<intr_type_array>();
				out << "array｢ ";
				dump_type_ref( at.sub );
				out << " ｣";
				if( at.raw ) {
					out << " raw";
				}
				break;
			}
			
			case intr_type::f_type: {
				auto & tt = vt.cast<intr_type_type>();
				out << "type｢ ";
				dump_type_ref( tt.sub );
				out << " ｣";
				break;
			}
			
			case intr_type::f_module: {
				out  << "module";
				break;
			}
			
			case intr_type::f_service:
			case intr_type::f_class: 
			case intr_type::f_closure: {
				auto & tc = vt.cast<intr_type_custom>(); 
				out << "{";
				indent();
				for( auto & sr : tc.get_type_decls() ) {
					out << std::endl;
					out << indent_str();
					dump_declaration( sr.get() );
				}
				unindent();
				out << std::endl << indent_str() << "}";
				break;
			}
			
			case intr_type::f_contextual:
				out << "contextual",
				dump_intr_type_tuple_inner( "｢", "｣", &vt.cast<intr_type_contextual>(), true );
				break;
				
			case intr_type::f_compat: {
				auto & tc = vt.cast<intr_type_compat>();
				switch( tc.type ) {
					case intr_type_compat::abi_var_args: out << "abi_var_args"; break;
					case intr_type_compat::string_tag: out << "string_tag"; break;
				}
				break;
			}
		}
	}
		
	std::sort( markers.begin(), markers.end() );
	for( size_t i=0; i < markers.size(); ++i ) {
		out << " " << markers[i];
	}
}
