#include "lang/language.hpp"

#include <typeinfo>

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

//common pattern for recursive validation
#define VALIDATE_RECURSE() \
	if( vctx.cover(this) ) \
		return; \

namespace leaf {

///////////////////////////////////////////////////////////
// expression
/*static*/ char const * expression::name_of( form_t f ) {
	switch( f ) {
		#define EXPRESSION_FORM(tag) case f_##tag: return #tag;
		EXPRESSION_FORMS
		#undef EXPRESSION_FORM
	}
	return "invalid";
}

void expression::validate(util::validation_context & vctx) const {
	type.validate(vctx);
	STATE_CHECK( is_resolved() );
}

void expression::pre_copy( expression const & o ) {
	src_loc = o.src_loc;
}

void expression::mark_resolved() {
	STATE_CHECK( type.get_extant() >= extant_level::resolved );
	local_resolved = true;
}

bool expression::is_resolved() const {
	return get_extant() >= extant_level::resolved;
}

extant_level expression::get_extant() const {
	if( !local_resolved ) {
		return extant_level::incomplete;
	}
	return type.get_extant();
}

///////////////////////////////////////////////////////////
// expr_literal_number
void expr_literal_number::validate(util::validation_context & vctx) const {
	expression::validate(vctx);
	
	STATE_CHECK( type.is_extr() );
	switch( src_form )
	{
		case sf_rational:
			STATE_CHECK( type.extr().is<intr_type_rational>() );
			break;
		case sf_integer:
			STATE_CHECK( type.extr().is<intr_type_integer>() );
			break;
		case sf_binary:
			STATE_CHECK( type.extr().is<intr_type_binary>() );
			break;
		case sf_float:
			STATE_CHECK( type.extr().is<intr_type_float>() );
			break;
		case sf_boolean:
			STATE_CHECK( type.extr().is<intr_type_boolean>() );
			break;
	}
}

void expr_literal_number::pre_copy( expr_literal_number const & o ) {
	super::pre_copy(o);
	src_form = o.src_form;
	value = o.value;
	val_boolean = o.val_boolean;
}

///////////////////////////////////////////////////////////
// expr_symbol
void expr_symbol::validate(util::validation_context & vctx) const {
	expression::validate(vctx);
	
	STATE_CHECK( symbol.size() );
	STATE_CHECK( sref );
	sref->validate(vctx);
	
	//TODO: check if the type of the expression is actually compatible with the declaration (not exactly
	//the same since lvalue could be dropped)
}

void expr_symbol::pre_copy( expr_symbol const & o ) {
	super::pre_copy(o);
	symbol = o.symbol;
	param = o.param;
}

///////////////////////////////////////////////////////////
// expr_with_block
void expr_with_block::validate(util::validation_context & vctx) const {
	expression::validate(vctx);
	STATE_CHECK( block );
	block->validate(vctx);
}

void expr_with_block::pre_copy( expr_with_block const & o ) {
	super::pre_copy(o);
	block = o.block->pre_clone();
}

///////////////////////////////////////////////////////////
// expr_field_like
void expr_field_like::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	
	STATE_CHECK( base );
	base->validate(vctx);
	STATE_CHECK( base->type.is_extr() );
}

void expr_field_like::pre_copy( expr_field_like const & o ) {
	super::pre_copy(o);
	base = o.base->pre_clone();
	optional = o.optional;
}


///////////////////////////////////////////////////////////
// expr_subscript
void expr_subscript::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	
	STATE_CHECK( base->type.extr().is<intr_type_array>() );
	
	STATE_CHECK( offset );
	offset->validate(vctx);
	STATE_CHECK( offset->type.is_extr() );
	STATE_CHECK( offset->type.extr().is<intr_type_integer>() );
}

void expr_subscript::pre_copy( expr_subscript const & o ) {
	super::pre_copy(o);
	offset = o.offset->pre_clone();
}

///////////////////////////////////////////////////////////
// expr_field
void expr_field::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	STATE_CHECK( field.size() || field_index );
	STATE_CHECK( field_index || sref );
	
	//TODO: unsure about this
	//STATE_CHECK( base->type.is_addressable() == type.is_lvalue() );
}

void expr_field::pre_copy( expr_field const & o ) {
	super::pre_copy(o);
	field = o.field;
	field_index = o.field_index;
	STATE_CHECK( !o.sref );
}


///////////////////////////////////////////////////////////
// expr_switch
void expr_switch::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	STATE_CHECK( expr_list.size() );
	for( auto & eli : expr_list ) {
		eli.validate(vctx);
		//TODO: verify all value's are the exact same type
	}
}

void expr_switch::expr_list_item::validate(util::validation_context & vctx) const {
	if( cond ) { //should only be allowed in last item
		cond->validate(vctx);
		STATE_CHECK( cond->type.extr().is<intr_type_boolean>() );
	}
	STATE_CHECK( value );
	value->validate(vctx);
}

bool expr_switch::may_have_side_effects() const {
	for( auto & eli : expr_list ) {
		if( (eli.cond && eli.cond->may_have_side_effects()) || 
			eli.value->may_have_side_effects() )
			return true;
	}
	return false;
}

void expr_switch::construct_result_type() {
	STATE_CHECK( expr_list.size() );
	type = expr_list[0].value->type;
}

void expr_switch::pre_copy( expr_switch const & o ) {
	super::pre_copy(o);
	expr_list.resize( o.expr_list.size() );
	for( size_t i=0; i < expr_list.size(); ++i ) {
		if( o.expr_list[i].cond ) {
			expr_list[i].cond = o.expr_list[i].cond->pre_clone();
		}
		expr_list[i].value = o.expr_list[i].value->pre_clone();
	}
}

///////////////////////////////////////////////////////////
// expr_with_sub
bool expr_with_sub::may_have_side_effects() const {
	for( auto & s : sub ) {
		if( s.expr->may_have_side_effects() ) {
			return true;
		}
	}
	return false;
}

void expr_with_sub::validate(util::validation_context & vctx) const {
	expression::validate(vctx);
	for( auto & s : sub ) {
		STATE_CHECK( s.expr );
		s.expr->validate(vctx);
	}
}

void expr_with_sub::pre_copy( expr_with_sub const & o ) {
	super::pre_copy(o);
	for( auto & s : o.sub ) {
		push_sub( s.name, s.expr->pre_clone() );
	}
}

///////////////////////////////////////////////////////////
// expr_funccall
void expr_funccall::validate(util::validation_context & vctx) const {
	VALIDATE_RECURSE() //could be a recursive function call chain
	expression::validate(vctx);
	STATE_CHECK( func );
	func->validate(vctx);
	STATE_CHECK( args );
	args->validate(vctx);
	STATE_CHECK( args->type.is_extr() && args->type.extr().is<intr_type_value_list>() );
	if( context ) {
		//TODO: context->validate(vctx);
	}
	
	auto & ftype = func->type.extr().cast<intr_type_function>();
	bool need_context = ftype.is_unbound();
	STATE_CHECK( !need_context || context );
}

void expr_funccall::pre_copy( expr_funccall const & o ) {
	super::pre_copy(o);
	func = o.func->pre_clone();
	args = o.args->pre_clone();
	if(o.context ) {
		context = o.context->pre_clone();
	}
}

///////////////////////////////////////////////////////////
// expr_transform
void expr_transform::validate(util::validation_context & vctx) const
{
	expression::validate(vctx);
	STATE_CHECK( arg );
}

void expr_transform::construct_result_type()
{
	switch( transform ) {
		case t_deref: {
			type = arg->type.extr();
			type.extr().set_ref_value();
			break;
		}
			
		case t_drop_lvalue: {
			type = arg->type.extr();
			type.extr().lvalue = false;
			break;
		}
			
		case t_shared_expr: {
			type = arg->type;
			break;
		}
			
		case t_promote:
		case t_loss:
			STMT_FAIL(*this);
			
		case t_make_optional: {
			type = arg->type.extr();
			type.extr().optional = true;
			type.extr().cvt_optional = true;
			break;
		}
			
		case t_deref_optional: {
			//this intentionally retains lvalue
			type = arg->type.extr();
			type.extr().optional = false;
			type.extr().cvt_optional = false;
			break;
		}
			
		case t_addr_of:  {
			auto arr = arg->type.extr().if_cast<intr_type_array>();
			if( arr ) {
				auto ra = make_shared<intr_type_array>();
				ra->sub = arr->sub;
				ra->raw = true;
				type = extr_type(ra);
			} else {
				type = arg->type.extr();
				type.extr().lvalue = false;
				type.extr().set_ref_value_ptr();
			}
			break;
		}
			
		case t_code_val: {
			static extr_type charint( make_shared<intr_type_integer>(platform::char_size) );
			type = charint;
			break;
		}
			
		case t_char_val: {
			static extr_type char_type( make_shared<intr_type_char>() );
			type = char_type;
			break;
		}
	}
}

///////////////////////////////////////////////////////////
// expr_new_func
void expr_new_func::validate(util::validation_context & vctx) const {
	expression::validate(vctx);
	new_type.validate(vctx);
}

void expr_new_func::pre_copy( expr_new_func const & o ) {
	super::pre_copy( o );
	if( o.new_spec_type ) {
		new_spec_type = make_shared<type_spec>(*o.new_spec_type);
	}
}

////////////////////////////////////////////////////////
// expr_temp
void expr_temp::pre_copy( expr_temp const & /*o*/ ) {
	STATE_FAIL( "not-expected" );
}

void expr_temp::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	STATE_CHECK( init >= init_min_ && init <= init_max_ );
	
	if( init == i_intrinsic ) {
		STATE_CHECK( intrinsic_value );
		intrinsic_value->validate(vctx);
	}
}

////////////////////////////////////////////////////////
// expr_block
void expr_block::pre_copy( expr_block const & o ) {
	super::pre_copy( o );
	if( o.result ) {
		result = o.result->pre_clone();
	}
}

////////////////////////////////////////////////////////
// expr_special_field
void expr_special_field::validate(util::validation_context & vctx) const {
	expression::validate(vctx);
	STATE_CHECK( base );
	base->validate(vctx);
	//TODO: type validation...
}

void expr_binary::validate(util::validation_context & vctx) const
{
	expression::validate(vctx);
	STATE_CHECK(a);
	STATE_CHECK(b);
	a->validate(vctx);
	b->validate(vctx);
	STATE_CHECK( a->type.same_as( b->type ) );
	STATE_CHECK( a->type.is_extr() );
	STATE_CHECK( b->type.is_extr() );
	if( is_boolean_result() ) {
		STATE_CHECK( type.extr().is<intr_type_boolean>() );
	} else {
		STATE_CHECK( type.same_as( a->type ) );
	}
		
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
}

void expr_unary::validate(util::validation_context & vctx) const
{
	expression::validate(vctx);
	STATE_CHECK(arg);
	arg->validate(vctx);
	STATE_CHECK( type.same_as( arg->type ) );
};

///////////////////////////////////////////////////////////
// expr_conditional
void expr_conditional::validate(util::validation_context & vctx) const {
	expression::validate(vctx);
	STATE_CHECK(cond);
	STATE_CHECK(a);
	cond->validate(vctx);
	a->validate(vctx);
	
	STATE_CHECK( op >= min_op_ && op <= max_op_ );
	switch( op ) {
		case o_optional_or:
			STATE_CHECK(b);
			b->validate(vctx);
			
			STATE_CHECK( cond->type.extr().is_optional() );
			STATE_CHECK( a->type.same_as( b->type ) );
			STATE_CHECK( type.same_as( a->type ) );
			break;
			
		case o_optional_cond:
			STATE_CHECK( cond->type.extr().is<intr_type_boolean>() );
			STATE_CHECK( type.extr().is_optional() );
			STATE_CHECK( type.extr().intr->same_as( *a->type.extr().intr ) );
			break;
	}
}

size_t expr_conditional::num_inner_trans() const { 
	return 1; 
}

shared_ptr<expression> expr_conditional::get_inner_trans( size_t ndx ) {
	STMT_CHECK( *this, ndx == 0 );
	return a; //TODO: this seems wrong, test cases must be incomplete for conversion_applicator::apply_single
}

void expr_conditional::set_inner_trans( size_t ndx, shared_ptr<expression> expr ) {
	STMT_CHECK( *this, ndx == 0 );
	a = expr;
}

void expr_conditional::construct_result_type() {
	switch( op ) {
		case o_optional_or:
			type = a->type;
			break;
		
		case o_optional_cond: {
			type = a->type.extr();
			type.extr().optional = true;
			//optional is being constructed, thus it is convertable
			type.extr().cvt_optional = true;
			break;
		}
	}
}

void expr_conditional::pre_copy( expr_conditional const & o ) {
	super::pre_copy(o);
	op = o.op;
	cond = o.cond->pre_clone();
	if( o.a ) {
		a = o.a->pre_clone();
	}
	if( o.b ) {
		b = o.b->pre_clone();
	}
}


///////////////////////////////////////////////////////////
// expr_text
void expr_text::pre_copy( expr_text const & o ) {
	super::pre_copy(o);
	what = o.what;
	text = o.text;
}

///////////////////////////////////////////////////////////
// expr_funcdefn
void expr_funcdefn::pre_copy( expr_funcdefn const & o ) {
	super::pre_copy(o);
	clone = o.clone;
	if( o.func_type ) {
		func_type = make_shared<type_spec>(*o.func_type);
	}
	for( auto & p : o.arg_defaults ) {
		arg_defaults[p.first] = p.second->pre_clone();
	}
}

shared_ptr<expr_funcdefn> expr_funcdefn::if_find_instance( type_ref result ) const {
	for( auto & inst : instances ) {
		if( result == inst->applied_type ) {
			return inst;
		}
	}
	
	return shared_ptr<expr_funcdefn>();
}

///////////////////////////////////////////////////////////
// expr_bind
void expr_bind::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	STATE_CHECK( unbound );
	unbound->validate(vctx);
	
	if( context ) {
		context->validate(vctx);
	} else if(!clone) {
		STATE_CHECK(unbound->type.extr().cast<intr_type_function>().access ==
			intr_type_function::a_raw);
	}
}

void expr_bind::construct_result_type() {
	auto ft = unbound->type.extr().cast<intr_type_function const>().clone_derived();
	ft->context = type_ref();
	ft->access = intr_type_function::a_bound;
	type = extr_type(ft);
}

///////////////////////////////////////////////////////////
// expr_current
void expr_current::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	STATE_CHECK( what >= min_what_ && what <= max_what_ );
}

void expr_current::pre_copy( expr_current const & o ) {
	super::pre_copy(o);
	what = o.what;
}

///////////////////////////////////////////////////////////
// expr_cast
void expr_cast::validate(util::validation_context & vctx ) const {
	super::validate(vctx);
	STATE_CHECK(source);
	source->validate( vctx );
	STATE_CHECK(spec_type);
}

void expr_cast::pre_copy( expr_cast const & o ) {
	super::pre_copy(o);
	if( o.spec_type ) {
		spec_type = make_shared<type_spec>(*o.spec_type);
	}
	if( o.source ) {
		source = o.source->pre_clone();
	}
}

/////////////////////////////////////////////////////////////
// expr_compat
void expr_compat::validate(util::validation_context & vctx) const {
	super::validate(vctx);
	STATE_CHECK(source);
	source->validate(vctx);
}

void expr_compat::pre_copy( expr_compat const & o ) {
	super::pre_copy(o);
	if (o.source) {
		source = o.source->pre_clone();
	}
}

/////////////////////////////////////////////////////////////
// expr_type
void expr_type::validate( util::validation_context & vctx ) const {
	super::validate(vctx);
	inner_type.validate( vctx );
}

void expr_type::pre_copy( expr_type const & o ) {
	super::pre_copy(o);
	if (o.inner_type_spec) {
		inner_type_spec = make_shared<type_spec>(*o.inner_type_spec);
	}
}

//////////////////////////////////////////////////////////////
// expr_tag
void expr_tag::validate( util::validation_context & vctx ) const {
	super::validate(vctx);
	STATE_CHECK(value);
	value->validate(vctx);
	//TODO: validate the type of source `literal array｢char｣`
}

void expr_tag::pre_copy( expr_tag const & o ) {
	super::pre_copy(o);
	if( o.value ) {
		value = o.value->pre_clone();
	}
}

} //eon leaf