//evaluation of builtins with function syntax
#include "lang/expression_typer/common.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {
namespace expression_typer {
namespace {

/**
	Extended context for the evaluation of a builtin -- since they all have a similar form.
*/
struct builtin_context : public context {
	shared_ptr<expr_symbol> expr;
	shared_ptr<expr_value_list> args;
	intr_type_value_list const * args_type;
	
	builtin_context( context ctx, intr_type_value_list const * args_type ) :
		context( ctx ),
		args_type( args_type ) {
	}
};

struct fm_entry {
	bool require_args;
	
	fm_entry() :
		require_args(true) {
	}
};
/**
	Check status of optional: True if set, False if unset.
*/
result builtin_has( builtin_context & bc ) {
	
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );
	auto cost = bc.converter().convert_nonlvalue_optional( bc.args_type->sub[0].type );
	bc.args->sub[0].expr = bc.applicator().apply( cost, bc.args->sub[0].expr );
		
	auto have = make_shared<expr_special_field>();
	have->src_loc = bc.expr->src_loc;
	have->field = expr_special_field::f_optional_set;
	have->base = bc.args->sub[0].expr;

	have->type = extr_type( make_shared<intr_type_boolean>() );
	have->mark_resolved();
	return have;
}

/**
	Allow a lossy conversion on the expression.
*/
result builtin_lossy( builtin_context & bc ) {
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );
	auto lt = bc.args->sub[0].expr;
	auto t = extr_type( lt->type.extr() );
	t.lossy = true;
	lt->type = t;
	return lt;
}

/**
	Obtain code-point integer value for a character.
*/
result builtin_code_val( builtin_context & bc ) {
	static type_ref char_type( extr_type(make_shared<intr_type_char>()) );
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );

	auto er = make_shared<expr_transform>();
	er->transform = expr_transform::t_code_val;
	er->arg = bc.convert_expression( char_type, bc.args->sub[0].expr );
	er->construct_result_type();
	er->mark_resolved();
	return er;
}

/**
	Obtain a character value for the given code-point integer.
*/
result builtin_char_val( builtin_context & bc ) {
	static type_ref charint( extr_type(make_shared<intr_type_integer>(platform::char_size)) );
	
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );
	auto er = make_shared<expr_transform>();
	er->transform = expr_transform::t_char_val;
	er->arg = bc.convert_expression( charint, bc.args->sub[0].expr );
	er->construct_result_type();
	er->mark_resolved();
	return er;
}

/**
	Pack/unpack convert to/from a tuple and value-list.
*/
result builtin_pack( builtin_context & bc ) {
	auto pack = bc.expr->symbol == "pack";
	
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );
	auto vl = bc.args->sub[0].expr;
	
	STMT_CHECK( *vl, pack ? vl->type.extr().is<intr_type_value_list>() : vl->type.extr().is<intr_type_tuple>() );
	size_t len = vl->type.extr().cast<intr_type_compound>().sub.size();
	//build a unpacking/packing conversion
	auto result = make_shared<cvt_expression>( 
		pack ? cvt_expression::c_pack : cvt_expression::c_pack_value );
	for( size_t i=0; i < len; ++i ) {
		auto unpack = make_shared<cvt_expression>( cvt_expression::c_field );
		unpack->field = i;
		result->sub.push_back({ unpack, "" }); //TODO: transfer name
	}
	return bc.applicator().apply( *result, bc.applicator().create_shared( bc.args->sub[0].expr ) ); //TODO: is the shallow shared okay?
}

/**
	Construction of the range type tuple. The range type is a tuple, but has a distinct name.
*/
result builtin_create_range( builtin_context & bc ) {
	static type_ref index_integer( extr_type(make_shared<intr_type_integer>()) ); //TODO: use bc.types()
	STMT_CHECK( *bc.args, bc.args->sub.size() == 2 );
	
	auto raw = make_shared<expr_compound>();
	raw->src_loc = bc.expr->src_loc;
	bool literal = true;
	for( int i=0; i < 2; ++i ) {
		auto cost = bc.converter().convert_retain_literal( index_integer, bc.args->sub[i].expr->type );
		auto sub = bc.applicator().apply( cost, bc.args->sub[i].expr );
		raw->push_sub( sub );
		literal = literal && sub->type.extr().is_literal();
	}

	//TODO: type can be directly set on expr_compound
	auto cast = make_shared<expr_cast>();
	cast->src_loc = bc.expr->src_loc;
	cast->source = raw;
	cast->spec_type = make_shared<type_spec>();
	cast->spec_type->set<pt_literal>( literal );
	cast->spec_type->set<pt_fun_prototype>( bc.types().get_integer_range() );
	
	bc.rewrite( bc.expr, cast );
	return cast;
}

result builtin_unopt( builtin_context & bc ) {
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );

	auto expr = bc.args->sub[0].expr;
	auto cost = bc.converter().remove_optional( expr->type );
	STMT_CHECK( *bc.args, cost.okay(), cost.explain_error() );
	return bc.applicator().apply( cost, expr );
}

result builtin_map( builtin_context & bc ) {
	STMT_CHECK( *bc.args, bc.args->sub.size() == 2 );
	
	auto tuple = process_expression(bc.args->sub[0].expr,bc);
	auto func = process_expression(bc.args->sub[1].expr,bc);

	//defer to non-builtin if it doesn't match our expected type
	if( !tuple->type.is_extr() || !tuple->type.extr().is<intr_type_tuple>()
		|| !func->type.is_extr() || !func->type.extr().is<intr_type_function>() ) {
		return result();
	}
	
	auto func_shared = func->may_have_side_effects() ? bc.applicator().create_shared(func) : func;
	auto tuple_shared = tuple->may_have_side_effects() ? bc.applicator().create_shared(tuple) : tuple;
	
	auto & tuple_type = tuple->type.extr().cast<intr_type_tuple>();
	size_t len = tuple_type.sub.size();
	
	auto result = make_shared<expr_compound>();
	for( size_t i=0; i < len; ++i) {
		auto part = make_shared<expr_field>();
		part->base = tuple_shared;
		part->field_index = i;
		
		auto f = make_shared<expr_funccall>();
		f->func = func_shared;
		auto args = make_shared<expr_value_list>();
		args->push_sub(part);
		f->args = args;
		
		result->push_sub( tuple_type.sub[i].name, f);
	}
	
	bc.rewrite(bc.expr, result);
	return result; 
}

result builtin_foreach( builtin_context & bc ) {
	STMT_CHECK( *bc.args, bc.args->sub.size() == 2 );
	
	auto tuple = process_expression(bc.args->sub[0].expr,bc);
	auto func = process_expression(bc.args->sub[1].expr,bc);

	//defer to non-builtin if it doesn't match our expected type
	if( !tuple->type.is_extr() || !tuple->type.extr().is<intr_type_tuple>()
		|| !func->type.is_extr() /*|| !func->type.extr().is<intr_type_function>()*/ ) { //selectable is also okay
		return result();
	}
	
	auto func_shared = func->may_have_side_effects() ? bc.applicator().create_shared(func) : func;
	auto tuple_shared = tuple->may_have_side_effects() ? bc.applicator().create_shared(tuple) : tuple;
	
	auto & tuple_type = tuple->type.extr().cast<intr_type_tuple>();
	size_t len = tuple_type.sub.size();
	
	auto res = make_shared<expr_block>();
	res->block = make_shared<statement_block>(scope_style_t::imperative);
	for( size_t i=0; i < len; ++i) {
		auto part = make_shared<expr_field>();
		part->base = tuple_shared;
		part->field_index = i;
		
		auto f = make_shared<expr_funccall>();
		f->func = func_shared;
		auto args = make_shared<expr_value_list>();
		args->push_sub(part);
		f->args = args;
		auto es = make_shared<expression_statement>();
		es->expr = f;
		
		res->block->list.push_back(es);
	}
	
	bc.rewrite( bc.expr, res );
	return res; 
}

/* TODO: This would be a lot easier if we could import macros directly from Leaf syntax itself. The structure
	is basically that in trap_func_pseudo.lfb -- except it uses a block and not a function. Or perhaps
	a simplified creator syntax as we introduce more microcode. */
result builtin_trap( builtin_context & bc ) {
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );
	
	COND_PROC_EXPRESSION( bc.args->sub[0].expr, bc, bc.expr );
	auto value = bc.args->sub[0].expr;
	//unforunately trapping requires fixation of the value
	//TODO: this should somehow guarantee that `a` in `var a = expr` and `unopt(a.value) in `var a = trap expr` 
	//have the same type (minus optional bit)
	auto value_type_r = bc.converter().fixate(value->type);
	STMT_CHECK( *value, value_type_r.okay() );
	auto value_type = value_type_r.result.extr();
	value_type.set_lvalue(false); //TODO: hmm, shouldn't fixate take care of this?
	
	auto cost = bc.converter().convert( value_type, value->type );
	value = bc.applicator().apply( cost, value );
	
	auto block = make_shared<expr_block>();
	block->block = make_shared<statement_block>(scope_style_t::imperative);
	
	auto decl = declaration::create();
	block->result = decl->create_expr_symbol();
	{ //var ⁑ei_symbol : ⁑trapped_error｢value_type｣
		decl->symbol = "⁑ei_symbol";
		auto ts = make_shared<type_spec>();
		auto & pts = ts->push_part<pt_symbol>( type_spec_symbol::create("⁑trapped_error") ); //use protoype instead
		pts.sub.push_back( type_identifier::get_spec(bc,value_type).result );
		decl->spec_type = ts;
	
		auto vs = make_shared<var_statement>();
		vs->decl = decl;
		block->block->list.push_back(vs);
	}
	
	/* do { 
			⁑ei_symbol.value = value
		} on fail {
			⁑ei_symbol.error = current_fail
		}
	*/
	auto ds = make_shared<error_statement>();
	ds->safe_skip = true;
	block->block->list.push_back(ds);
	{
		auto as = make_shared<assign_statement>();
		as->right = value;
	
		auto left = make_shared<expr_field>();
		left->field = "value";
		left->base = decl->create_expr_symbol();
		as->left = left;
		
		ds->inner = as;
	}
	{
		auto as = make_shared<assign_statement>();
		
		auto err = make_shared<expr_symbol>();
		err->symbol = "current_fail";
		as->right = err;
		
		auto left = make_shared<expr_field>();
		left->field = "error";
		left->base = decl->create_expr_symbol();
		as->left = left;
		
		auto block = make_shared<statement_block>(scope_style_t::imperative);
		block->list.push_back(as);
		ds->error.on_match = fail_level::normal;
		ds->error.on_error = block;
	}
	
	
	bc.rewrite( bc.expr, block );
	DEBUG( dump::get(block), dump::get(decl->spec_type) );
	return block;
}

typedef bool (*check_op_type)( extr_type const & );
bool check_op_ibf( extr_type const & t ) {
	return t.is<intr_type_integer>() || t.is<intr_type_binary>() || t.is<intr_type_float>();
}
bool check_op_ibfxct( extr_type const & t ) {
	if( t.is<intr_type_integer>() || t.is<intr_type_binary>() || t.is<intr_type_float>() || t.is<intr_type_boolean>() || t.is<intr_type_char>()) {
		return true;
	}
	
	auto compat = t.if_cast<intr_type_compat>();
	if( compat && compat->type == intr_type_compat::string_tag) {	
		return true;
	}
	
	return false;
}
bool check_op_if( extr_type const & t ) {
	return t.is<intr_type_integer>() || t.is<intr_type_float>();
}
bool check_op_b( extr_type const & t ) {
	return t.is<intr_type_binary>();
}
bool check_op_i( extr_type const & t ) {
	return t.is<intr_type_integer>();
}
// bool check_op_x( extr_type const & t ) {
// 	return t.is<intr_type_boolean>();
// }
bool check_op_xb( extr_type const & t ) {
	return t.is<intr_type_boolean>() || t.is<intr_type_binary>();
}

enum class binary_op_type {
	standard,
	binary,
	binary_or_boolean,
};

struct binary_entry : public fm_entry { 
	expr_binary::op_t op; 
	check_op_type check;  
	binary_op_type type;
	
	binary_entry( expr_binary::op_t op_, check_op_type check_, binary_op_type type_ ) :
		op( op_ ),
		check( check_ ),
		type( type_ ) {
	}
	binary_entry() {
	}
};
//construct a static map of symbols to binary builtin ops
typedef std::map<std::string,binary_entry> bin_op_map_t;
bin_op_map_t bin_op_map;
struct bom_init {
	bom_init() {
		#define OP(name,op,type) bin_op_map[#name] = binary_entry(expr_binary::o_##op, check_op_##type, binary_op_type::standard );
		OP(add,add,ibf)
		OP(sub,sub,ibf)
		OP(mul,mul,ibf)
		OP(div,div,ibf)
		
		OP(mod,mod,i)
		
		#define BITOP(name,op,type) bin_op_map[#name] = binary_entry(expr_binary::o_##op, check_op_##type, binary_op_type::binary);
		BITOP(bit_shl,bit_shl,b)
		BITOP(bit_shr,bit_shr,b)
		
		OP(eq,eq,ibfxct)
		OP(neq,neq,ibfxct)
		OP(lt,lt,ibfxct)
		OP(lte,lte,ibfxct)
		OP(gt,gt,ibfxct)
		OP(gte,gte,ibfxct)
		
		#define BBOP(name,op,type) bin_op_map[#name] = binary_entry(expr_binary::o_##op, check_op_##type, binary_op_type::binary_or_boolean);
		BBOP(or,or,xb)
		BBOP(and,and,xb)
		BBOP(xor,xor,xb)
		BBOP(and_not,and_not,xb)
		BBOP(nor,nor,xb)
		BBOP(nand,nand,xb)
		#undef OP
		#undef BITOP
		#undef BBOP
	}
} _bom_init;

struct unary_entry : public fm_entry { 
	expr_unary::op_t op; 
	check_op_type check;  
	
	unary_entry( expr_unary::op_t op_, check_op_type check_ ) :
		op(op_),
		check(check_) {
	}
	unary_entry() {
	}
};
typedef std::map<std::string,unary_entry> un_op_map_t;
un_op_map_t un_op_map;
struct uom_init {
	uom_init() {
		#define OP(name,op,type) un_op_map[#name] = unary_entry( expr_unary::o_##op, check_op_##type );
		OP(neg,neg,if)
		OP(abs,abs,if)
		OP(not,not,xb)
		#undef OP
	}
} _uom_init;

typedef result (*builtin_func)( builtin_context & bc );
struct func_entry : public fm_entry {
	builtin_func func;
	
	func_entry( builtin_func func_ ) :
		func(func_) {
	}
	func_entry() {
	}
};

typedef std::map<std::string, func_entry> func_map_t;
func_map_t func_map;

struct fm_init {
	fm_init() {
		func_map["has"] = func_entry( builtin_has );
		func_map["lossy"] = func_entry( builtin_lossy );
		func_map["code_val"] = func_entry( builtin_code_val );
		func_map["char_val"] = func_entry( builtin_char_val );
		func_map["pack"] = func_entry( builtin_pack );
		func_map["unpack"] = func_entry( builtin_pack );
		func_map["create_range"] = func_entry( builtin_create_range );
		func_map["unopt"] = func_entry( builtin_unopt );
		func_map["trap"] = func_entry( builtin_trap );
		
		func_map["map"] = func_entry( builtin_map );
		func_map["map"].require_args = false;
		func_map["foreach"] = func_entry( builtin_foreach );
		func_map["foreach"].require_args = false;
	}
} _fm_init;

/**
	Standard binary operations.
*/
result process_binary( builtin_context & bc, binary_entry const & bop ) {

	STMT_CHECK( *bc.args, bc.args->sub.size() == 2 );
	type_converter tc = bc.converter();
	DEBUG( "IN-A", dump::get(bc.args_type->sub[0].type) );
	DEBUG( "IN-B", dump::get(bc.args_type->sub[1].type) );
	for( int i=0; i < 2; ++i ) {
		auto cost = tc.convert_any_scalar( bc.args_type->sub[i].type );
		if( !cost.okay() ) {
			//string_tag's are also okay for some of the operations, but they aren't really scalar so not part of the above
			cost = tc.convert( extr_type(bc.types().get_string_tag()), bc.args_type->sub[i].type );
		}
		if( !cost.okay() ) {
			return result();
		}
		bc.args->sub[i].expr = bc.applicator().apply( cost, bc.args->sub[i].expr );
	}

	type_ref result_type;
	if( bc.expr->param ) {
		STMT_CHECK( *bc.expr, bc.expr->param->sub.size() == 1 );
		auto pt = type_identifier::determine( bc, bc.expr->param->sub[0].type );
		STMT_CHECK( *bc.expr, pt.okay(), pt.explain_error() ); //TODO: unknown identifier
		result_type= pt.result;
	} else {
		auto unify = tc.unify( bc.args->sub[0].expr->type, bc.args->sub[1].expr->type );
		STMT_CHECK( *bc.args, unify.okay(), unify.explain_error() );
		result_type = unify.result;
	}

	DEBUG( "CVT-A", dump::get(bc.args_type->sub[0].type) );
	DEBUG( "CVT-B", dump::get(bc.args_type->sub[1].type) );
	DEBUG( "RESULT", dump::get(result_type), result_type.extr().is<intr_type_boolean>() );
	DEBUG( "OP", bc.expr->symbol );
	//constant folding
	if( result_type.extr().is_literal() && result_type.extr().is<intr_type_number>() ) {
		auto a = expression_typer::extract_literal_number( *bc.args->sub[0].expr );
		auto b = expression_typer::extract_literal_number( *bc.args->sub[1].expr );
		boost::optional<util::number> number;
		
		if( bc.expr->symbol == "add" ) {
			number = a + b;
		} else if( bc.expr->symbol == "sub" ) {
			number = a - b;
		} else if( bc.expr->symbol == "mul" ) {
			number = a * b;
		} else if( bc.expr->symbol == "div" ) {
			number = a / b;
		}

		if( number ) {
			//TODO: it's very uncertain what type should result here. If both are rational it's
			//clearly rational, but if it is mixed...
			auto num = make_shared<expr_literal_number>();
			num->src_form = expr_literal_number::sf_rational;
			num->value = *number;
			//num->local_complete = true;
			return bc.proc_expression( num ); //gets a new type for the resulting rational
		}
	}

	type_ref arg_type;
	DEBUG( "TYPE", (int)bop.type, result_type.extr().is<intr_type_boolean>() );
	if( bop.type == binary_op_type::binary || 
		(bop.type == binary_op_type::binary_or_boolean && !result_type.extr().is<intr_type_boolean>()) ) {
		type_spec ts;
		ts.set<pt_fun_class>(intr_type::f_binary);
		auto resolve = type_identifier::constrain( bc, ts, result_type );
		STMT_CHECK( *bc.args, resolve.okay() );
		arg_type = resolve.result;
	} else {
		auto fix = tc.fixate( result_type );
		STMT_CHECK( *bc.args, fix.okay() );
		arg_type = fix.result;
	}
	STMT_CHECK( *bc.args, bop.check( arg_type.extr() ), "invalid-binary-op-type" );
			
	for( int i=0; i <2; ++i ) {
		auto cost = tc.convert( arg_type, bc.args->sub[i].expr->type );
		bc.args->sub[i].expr = bc.applicator().apply( cost, bc.args->sub[i].expr );
	}
	
	auto bin = make_shared<expr_binary>();
	bin->src_loc = bc.expr->src_loc;
	bin->op = bop.op;
	bin->a = bc.args->sub[0].expr;
	bin->b = bc.args->sub[1].expr;
	if( bin->is_boolean_result() ) {
		bin->type = extr_type(make_shared<intr_type_boolean>());
	} else {
		bin->type = arg_type;
	}
	//DEBUG( "BOP-TYPE", dump::get(bin->type) );
	bin->mark_resolved();
	return bin;
}

/**
	Standard unary operations.
*/
result process_unary( builtin_context & bc, unary_entry const & uop ) {
	STMT_CHECK( *bc.args, bc.args->sub.size() == 1 );

	DEBUG( "IN", dump::get( bc.args_type->sub[0].type ));
	auto tc = bc.converter();
	auto cost = tc.convert_any_scalar( bc.args_type->sub[0].type );
	if( !cost.okay() ) {
		return result();
	}
	auto arg = bc.applicator().apply( cost, bc.args->sub[0].expr );

	//limited literal folding
	if( arg->type.extr().is<intr_type_number>() && arg->type.extr().is_literal() ) {
		auto a = expression_typer::extract_literal_number( *arg );
		boost::optional<util::number> number;
		
		if( bc.expr->symbol == "neg" ) {
			number = a.neg();
		} else if( bc.expr->symbol == "abs" ) {
			number = a.abs();
		}
			
		if( number ) {
			auto num = make_shared<expr_literal_number>();
			num->src_form = expr_literal_number::sf_rational;
			num->value = *number;
			//num->local_complete = true;
			return bc.proc_expression( num );
		}
	}
	
	auto fix = tc.fixate( arg->type );
	STMT_CHECK( *bc.args, fix.okay() );
	DEBUG( "OP", dump::get(fix.result) );
	STMT_CHECK( *bc.args, uop.check( fix.result.extr() ), "invalid-unary-op-type" );
	cost = tc.convert( fix.result, arg->type );
	arg = bc.applicator().apply( cost, arg );
	
	auto un = make_shared<expr_unary>();
	un->src_loc = bc.expr->src_loc;
	un->op = uop.op;
	un->arg = arg;
	un->type = fix.result;
	un->mark_resolved();
	return un;
}

} //eon anon

/**
	The binary/unary operators cannot be processed as normal functions at this time since they
	have an unlimited number of size variations. Ultimately one would hope that a proper parametric
	function definition would be able to capture the logic being applied here.
	
	@return if null it means it isn't a builtin and normal funccall processing should continue
*/
result process_funccall_builtin( shared_ptr<expr_funccall> expr, context ctx ) {
	if( expr->func->form != expression::f_symbol ) {
		return result();
	}
	auto es = std::static_pointer_cast<expr_symbol>(expr->func);

	auto it = func_map.find( es->symbol );
	auto bop = bin_op_map.find( es->symbol );
	auto uop = un_op_map.find( es->symbol );
	fm_entry * fm_entry;
	if( it != func_map.end() ) {
		fm_entry = &it->second;
	} else if( bop != bin_op_map.end() ) {
		fm_entry = &bop->second;
	} else if( uop != un_op_map.end() ) {
		fm_entry = &uop->second;
	} else {
		return result();
	}

	intr_type_value_list const * args_type = nullptr;
	auto args = std::static_pointer_cast<expr_value_list>(expr->args);
	
	if( fm_entry->require_args ) {
		COND_PROC_EXPRESSION( expr->args, ctx, expr );
		STMT_CHECK( *expr, expr->args->form == expression::f_value_list );

		args_type = &args->type.extr().cast<intr_type_value_list>();
	}
	
	builtin_context bc( ctx, args_type );
	bc.expr = es;
	bc.args = args;
	
	if( it != func_map.end() ) {
		return (it->second.func)( bc );
	}
	
	if( bop != bin_op_map.end() ) {
		return process_binary( bc, bop->second );
	}

	if( uop != un_op_map.end() ) {
		return process_unary( bc, uop->second );
	}
	
	STATE_FAIL( "ummm" );
}

/**
	Obtain the literal number associated with this expression.
*/
util::number extract_literal_number( expression const & expr ) {
	STMT_CHECK( expr, expr.type.extr().is_literal() )
	STMT_CHECK( expr, expr.type.extr().is<intr_type_number>() );
	
	expression const * use_expr = &expr;
	
	if( expr.form == expression::f_symbol ) {
		auto & es = static_cast<expr_symbol const&>(expr);
		STMT_CHECK( expr, es.sref && es.sref->literal_value );
		use_expr = es.sref->literal_value.get();
	}
	
	STMT_CHECK( expr, use_expr->form == expression::f_literal_number );
	auto & cn = static_cast<expr_literal_number const&>( *use_expr );
	return cn.value;
}

shared_ptr<expr_compound> access_literal_tuple( shared_ptr<expression> expr ) {
	STMT_CHECK( *expr, expr->type.extr().is_literal() );
	STMT_CHECK( *expr, expr->type.extr().is<intr_type_tuple>() );
	
	if( expr->form == expression::f_symbol ) {
		auto & es = static_cast<expr_symbol const &>(*expr);
		STMT_CHECK( *expr, es.sref && es.sref->literal_value );
		expr = es.sref->literal_value;	
	}
	
	STMT_CHECK( *expr, expr->form == expression::f_compound );
	return std::static_pointer_cast<expr_compound>(expr);
}

} //eon expression_typer

} //eon leaf
