#include "lang/expression_typer/common.hpp"

#define SHOW_DEBUG 0
#include "lang/dump.hpp"
#include "util/debug.hpp"

namespace leaf {
namespace expression_typer {

shared_ptr<expression> resolve_scope_symbol_type( scope_symbol const & msym,
	shared_ptr<type_spec const> param, std::string symbol ) {
	if( msym.is_fun_type() ) {
		auto ex = make_shared<expr_type>();
		auto ts = make_shared<type_spec>();
		auto & part = ts->set<pt_fun_prototype>(msym.fun_type);
		if( param ) {
			for( auto & p : param->sub ) {
				part.sub.push_back( p.type );
			}
		}
		ex->inner_type_spec = ts;
		return ex;
	}
	
	if( msym.is_type_alias() ) {
		auto ex = make_shared<expr_type>();
		if( param ) {
			auto ts = make_shared<type_spec>();
			//TODO: Giving the symbol here seems wrong, it's already been resolved and there should
			//be a way to provide the resolved type_spec
			type_spec_symbol tss;
			tss.parts.push_back( symbol );
			auto & part = ts->push_part<pt_symbol>( tss );
			for( auto & p : param->sub ) {
				part.sub.push_back( p.type );
			}
			ex->inner_type_spec = ts;
		} else {
			//NOTE: this seems like an optimization
			ex->inner_type_spec = msym.type_alias;
		}
		return ex;
	}
	
	return shared_ptr<expression>();
}

shared_ptr<expression> process_symbol( shared_ptr<expr_symbol> expr, context ctx ) {
	if( expr->is_resolved() ) {
		return expr;
	}
	
	auto complete = [&]( shared_ptr<expr_symbol> es ) {
		STATE_CHECK( es->sref );
		if( !es->sref->type.is_extr() ) {
			return;
		}
		
		auto rtype = es->sref->type.extr();
		rtype.lvalue = true;
		es->type = rtype;
		if (es->type.is_complete() ) {
			es->mark_resolved();
		}
	};

	//sref already decided (either here, or other creation)
	if( expr->sref ) {
		complete(expr);
		return expr;
	}

	auto const & msym = ctx.sc->if_get_symbol( expr->symbol );
	STMT_CHECK( *expr, !!msym, "unknown-symbol", error::tag_error_symbol( expr->symbol ) );

	auto tex = resolve_scope_symbol_type( *msym, expr->param, expr->symbol );
	if( tex ) {
		ctx.rewrite(expr,tex);
		return tex;
	}
	
	STMT_CHECK( *expr, !expr->param, "unsupported-param" );
	STMT_CHECK( *expr, msym->is_var(), "invalid-symbol-type", error::tag_error_symbol( expr->symbol ) );

	auto const & match = msym->vars;
	STMT_CHECK( *expr, match.size() > 0, "invalid-symbol-count", error::tag_error_symbol( expr->symbol ) );
	for( auto & m : match ) {
		//this is mainly an optimization for type identification in value-list references (to avoid determing the type there)
		if( !m->type.is_defined() && m->auto_resolve ) {
			auto r = type_identifier::determine( ctx,m->spec_type );
			if( r.okay() ) {
				m->type = r.result;
			}
		}
		
		STMT_CHECK( *expr, ctx.sc->can_capture_var(m), "illegal-capture",
			error::tag_error_symbol( expr->symbol ) );
	}
	
	if( match.size() == 1 ) {
		//unknown types can't be evaluated yet
		if( match[0]->type.get_extant() == extant_level::incomplete || !match[0]->type.is_extr() ) {
			return expr;
		}

		//TODO: this should be only an optimization, yet a few tests still require it!
		if( match[0]->type.extr().is_literal() ) {
			if( !match[0]->literal_value ) {
				return expr;
			}
			return match[0]->literal_value;
		}
			
		//must become available in our scope
		expr->sref = ctx.sc->capture_var(match[0]);
		complete(expr);
		return expr;
	}
	
	auto select = make_shared<intr_type_tuple>();
	select->selectable = true;
	auto multi = make_shared<expr_compound>();
	multi->src_loc = expr->src_loc;

	for( auto & sr : match ) {
		auto sym = make_shared<expr_symbol>();
		sym->src_loc = expr->src_loc;
		sym->symbol = sr->symbol;
		sym->sref = ctx.sc->capture_var( sr );
		complete(sym);
		select->push_sub( sym->type );
		multi->push_sub( sym );
	}
	multi->type = extr_type(select);
	//multi->type.set_literal( true );
	if( multi->type.is_complete() ) {
		multi->mark_resolved();
	}
	ctx.rewrite(expr, multi);
	return multi;
}

shared_ptr<expression> process_literal_number( shared_ptr<expr_literal_number> expr,
	context /*ctx*/ ) {
	
	shared_ptr<intr_type> st;
	switch( expr->src_form ) 
	{
		case expr_literal_number::sf_rational: {
			auto tr = make_shared<intr_type_rational>();
			if( expr->value.is_integral() ) {
				if( expr->value.is_negative() ) {
					tr->integer_bitsize = expr->value.signed_bits();
					tr->integer_signed = true;
				} else {
					tr->integer_bitsize = expr->value.unsigned_bits();
					tr->integer_signed = false;
				}
			}
			st = tr;
			break;
		}
		case expr_literal_number::sf_integer:
		case expr_literal_number::sf_float:
		case expr_literal_number::sf_binary:
			STMT_FAIL( *expr, "not expecting integer/float type const" );
			break;
		case expr_literal_number::sf_boolean:
			st = make_shared<intr_type_boolean>();
			break;
	}
	auto ex = extr_type(st);
	ex.literal = true;
	expr->type = ex;
	expr->mark_resolved();
	return expr;
}

shared_ptr<expression> process_text( shared_ptr<expr_text> expr, context ctx ) {
	switch( expr->what ) {
		case expr_text::w_string: {
			//TODO: make a common isntantiation function
			auto at = make_shared<intr_type_array>();
			at->sub = extr_type( make_shared<intr_type_char>() );
			at->instance_of = ctx.types().get_primordial_array();
			at->applied_params.push_back( at->sub.extr() );
			
			auto t = extr_type( at );
			t.literal = true;
			expr->type = t;
			break;
		}
		case expr_text::w_char: {
			auto t = extr_type( make_shared<intr_type_char>() );
			t.literal = true;
			expr->type = t;
			break;
		}
	}
	expr->mark_resolved();
	return expr;
}

shared_ptr<expression> process_compound( shared_ptr<expr_compound> expr, 
	context ctx ) {
	auto tt = make_shared<intr_type_tuple>();
	
	//we can retain the literal flag is all components are literal
	bool all_literal = true;
	for( size_t i=0; i < expr->sub.size(); ++i ) 	{
		auto sub = process_expression( expr->sub[i].expr, ctx ); 
		if( sub->get_extant() < extant_level::resolved )
			return expr;
			
		//lvalues cannot be contained inside tuples
		auto cost = ctx.converter().convert_any_nonlvalue( sub->type );
		expr->sub[i].expr = ctx.applicator().apply( cost, sub );
		all_literal = all_literal && expr->sub[i].expr->type.extr().is_literal();
		
		//lose literal on components, only whole tuple or none (don't touch source expressions though)
		auto net = expr->sub[i].expr->type.extr();
		net.literal = false; 
		tt->push_sub( expr->sub[i].name, net );
	}
	
	auto t = extr_type(tt);
	t.literal = all_literal;
	expr->type = t;
	expr->mark_resolved();
	return expr;
}

shared_ptr<expression> process_value_list( shared_ptr<expr_value_list> expr, context ctx ) {
	auto tt = make_shared<intr_type_value_list>();
	expr->type = extr_type(tt);
	bool resolved = true;
	for( size_t i=0; i < expr->sub.size(); ++i ) 	{
		expr->sub[i].expr = process_expression( expr->sub[i].expr, ctx );
		tt->push_sub( "", expr->sub[i].expr->type );
		if( !expr->sub[i].expr->is_resolved() ) {
			resolved = false;
		}
		//unlike tuples the parts of a value_list retain literal status and the whole does not.
	}
	if( resolved ) {
		expr->mark_resolved();
	}
	return expr;
}

shared_ptr<expression> process_block( shared_ptr<expr_block> expr, context ctx ) {
	expr->block->symbols->set_base( ctx.sc, scope::bt_local );
	ctx.proc_statement_block( expr->block );
	
	if( expr->result ) {
		COND_PROC_EXPRESSION( expr->result, ctx, expr );
		expr->type = expr->result->type;
	} else {
		expr->type = extr_type(make_shared<intr_type_value_list>()); //no parts though
	}
	expr->mark_resolved();
	return expr;
}

shared_ptr<expression> process_transform( shared_ptr<expr_transform> expr,
	context ) {
	//only internal so far, thus should be complete
	STATE_CHECK( expr->is_resolved() );
	return expr;
}

shared_ptr<expression> process_special_field( shared_ptr<expr_special_field> expr, context ctx) {
	COND_PROC_EXPRESSION( expr->base, ctx, expr );
	switch( expr->field ) {
		case expr_special_field::f_array_size:
			expr->type = extr_type(make_shared<intr_type_integer>());
			break;
		case expr_special_field::f_optional_set:
			expr->type = extr_type( make_shared<intr_type_boolean>() );
			break;
		case expr_special_field::f_contextual_context:
		case expr_special_field::f_contextual_unbound:
			//use in conversion_applicator knows the types
			STATE_CHECK( expr->is_resolved() );
			return expr;
	}

	expr->mark_resolved();
	return expr;
}

shared_ptr<expression> process_binary( shared_ptr<expr_binary> expr, context ) {
	//only internal so far, thus should be complete
	STATE_CHECK( expr->is_resolved() );
	return expr;
}

shared_ptr<expression> process_unary( shared_ptr<expr_unary> expr, context ) {
	//only internal so far, thus should be complete
	STATE_CHECK( expr->is_resolved() );
	return expr;
}

shared_ptr<expression> process_compat( shared_ptr<expr_compat> expr, context ) {
	//only internal so far, thus should be complete
	STATE_CHECK( expr->is_resolved() );
	return expr;
}

shared_ptr<expression> process_cast( shared_ptr<expr_cast> expr, context ctx ) {
	COND_PROC_EXPRESSION( expr->source, ctx, expr );
	
	//DEBUG( "CAST_CONSTRAIN", dump::get( *expr->spec_type ),	dump::get( expr->source->type ) );
	auto it = type_identifier::constrain( ctx, expr->spec_type, expr->source->type );
	STMT_CHECK( *expr, it.okay(), it.explain_error() + " invalid-cast" );
	return ctx.convert_expression( it.result, expr->source );
}

shared_ptr<expression> process_expression( shared_ptr<expression> expr, 
	context ctx ) {
	//only type things once
	if( expr->get_extant() >= extant_level::resolved )
		return expr;
		
	switch( expr->form ) {
		#define EXPRESSION_FORM(form) \
			case expression::f_##form: \
				expr = process_##form( std::static_pointer_cast<expr_##form>(expr), ctx ); \
				break;
		EXPRESSION_FORMS
		#undef EXPRESSION_FORM
	}
	return expr;
}

shared_ptr<expression> process_switch( shared_ptr<expr_switch> expr, 
	context ctx )
{	
	static type_ref cond_type( extr_type(make_shared<intr_type_boolean>()) );
	bool have_else = false;
	
	std::vector<type_ref> types;
	for( auto & item : expr->expr_list ) {
		if( item.cond ) {
			STATE_CHECK( !have_else, "conditions post-else statement" );
			COND_PROC_EXPRESSION( item.cond, ctx, expr );
			item.cond = ctx.convert_expression( cond_type, item.cond );
		} else {
			have_else = true;
		}

		COND_PROC_EXPRESSION( item.value, ctx, expr );
		types.push_back( item.value->type );
	}

	STMT_CHECK( *expr, have_else, "missing an else statement" );

	//types are unified so the switch has a single type result
	auto icost = ctx.converter().unify( &types[0], types.size() );
	STMT_CHECK( *expr, icost.okay(), "no unifying type" );
	for( auto & item : expr->expr_list ) {
		item.value = ctx.convert_expression( icost.result, item.value );
	}
	
	expr->construct_result_type();
	expr->mark_resolved();
	return expr;
}


shared_ptr<expression> process_conditional( shared_ptr<expr_conditional> expr,
	context ctx ) {
	static type_ref cond_type( extr_type(make_shared<intr_type_boolean>()) );
	
	COND_PROC_EXPRESSION( expr->cond, ctx, expr );
	
	switch( expr->op ) {
		case expr_conditional::o_optional_or: {
			COND_PROC_EXPRESSION( expr->b, ctx, expr );
			type_converter tc = ctx.converter();
			
			//only makes sense if first part is optional
			auto cost = tc.convert_nonlvalue_optional( expr->cond->type );
			expr->cond = ctx.applicator().apply( cost, expr->cond );
			STMT_CHECK( *expr->cond, expr->cond->type.extr().is_optional() );

			auto rtype = expr->cond->type.extr();
			rtype.optional = false;
			
			auto unify = tc.unify( rtype, expr->b->type );
			STMT_CHECK( *expr, unify.okay(), std::string("unable-to-unify-optional-or: ") +
				dump::get(rtype) + " with " + dump::get(expr->b->type));
			auto fix = tc.fixate( unify.result );
			STMT_CHECK( *expr, fix.okay(), "unable-to-fixate-optional-or" );
			
			//the `cond` and `a` are supposed to be the same, thus we convert the condition
			//to an optional form of the unified type instead of converting the parts (this
			//allows proper fixation)
			auto ctype = fix.result.extr();
			ctype.optional = true;
			cost = tc.convert( ctype, expr->cond->type );
			expr->cond = ctx.applicator().apply( cost, expr->cond );

			//will be used twice, thus must be marked as shared
			expr->cond = ctx.applicator().create_shared( expr->cond ); 
			
			cost = tc.convert( ctype, expr->cond->type );
			expr->a = ctx.applicator().apply( cost, expr->cond );
			expr->a = ctx.applicator().deref_optional( expr->a ); //TODO: initial convert should drop optional!
			cost = tc.convert( fix.result, expr->b->type );
			expr->b = ctx.applicator().apply( cost, expr->b );
			
			expr->construct_result_type();
			expr->mark_resolved();
			return expr;
		}
		
		case expr_conditional::o_optional_cond: {
			COND_PROC_EXPRESSION( expr->a, ctx, expr );
			auto cost = ctx.converter().convert_any_nonlvalue( expr->a->type );
			expr->a = ctx.applicator().apply( cost, expr->a );
			
			expr->cond = ctx.convert_expression( cond_type, expr->cond );
			expr->construct_result_type();			
			expr->mark_resolved();
			return expr;
		}
	}
	
	STMT_FAIL(*expr);
}

shared_ptr<expression> process_bind( shared_ptr<expr_bind> expr,
	context ctx ) {
	
	//only need input level for context now
	if( expr->context ) {
		expr->context = process_expression( expr->context, ctx );
		if( expr->context->type.get_extant() < extant_level::input ) {
			return expr;
		}
	}
	
	COND_PROC_EXPRESSION( expr->unbound, ctx, expr );
	//discovered in class/function_import_name.lfb test, by being a selectable the lvalue is retained
	//longer (not dropped during resolution if currently undefined)
	//TODO: it feels like this should happen somewhere else...
	if( expr->unbound->type.extr().is_lvalue() ) {
		expr->unbound = ctx.applicator().drop_lvalue( expr->unbound );
	}

	expr->construct_result_type();
	expr->mark_resolved();
	return expr;
}

shared_ptr<expression> process_current( shared_ptr<expr_current> expr,
	context ctx ) {
	
	switch(expr->what) {
		case expr_current::w_this: {
			int depth;
			auto sc = ctx.sc->find_class_root(depth);
			STMT_CHECK(*expr,sc);
			
			expr->depth = depth;
			expr->what = expr_current::w_context;
		
			auto cl = sc->get_context_type(true);
			expr->type = cl;
			expr->mark_resolved();
			return expr;
		}
		
		case expr_current::w_context: {
			auto sc = ctx.sc->resolve_depth(expr->depth);
			STMT_CHECK(*expr,sc);
			
			auto cl = sc->get_context_type(true);
			expr->type = cl;
			expr->mark_resolved();
			return expr;
		}
		
	}
	
	STMT_FAIL(*expr,"invalid-what");
		
}

shared_ptr<expression> process_type( shared_ptr<expr_type> expr, context ctx ) {
	if( !expr->inner_type.is_defined() ) {
		STMT_CHECK( *expr, expr->inner_type_spec );
		auto r = type_identifier::determine( ctx, expr->inner_type_spec );
		if( !r.okay() ) {
			//TODO: what?
			STMT_FAIL( *expr, r.explain_error() );
			return expr;
		}
		
		DEBUG( "PType", dump::get(expr->inner_type_spec), dump::get(r.result));
		expr->inner_type = r.result;
	}
	
	STMT_CHECK( *expr, expr->inner_type.is_extr());
	
	auto t = make_shared<intr_type_type>();
	t->sub = expr->inner_type.extr();
	expr->type = extr_type(t);
	expr->mark_resolved();
	return expr;
}

shared_ptr<expression> process_tag( shared_ptr<expr_tag> expr, context ctx ) {
	expr->value = process_expression(expr->value, ctx);
	
	//TODO: simplify to checking against a type string for clarity
	STMT_CHECK( *expr->value, expr->value->type.is_extr() );
	auto estring = expr->value->type.extr();
	auto earray = estring.if_cast<intr_type_array>();
	STMT_CHECK( *expr->value, estring.is_literal() && earray && earray->sub.extr().is<intr_type_char>() );
	
	expr->type = extr_type(ctx.types().get_string_tag());
	expr->mark_resolved();
	
	return expr;
}

} //eon expression_typer
} //eon leaf