#include "lang/expression_typer/common.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {
namespace expression_typer {

/**
	Common code for subscript and field on tuples.
*/
shared_ptr<expression> access_field_index( shared_ptr<expression> ex,
	shared_ptr<expression> base, size_t index,  context /*ctx*/ ) {
	bool is_lvalue = base->type.extr().is_addressable();

	//DEBUG("FIELD_IN", dump::get(*ex), dump::get(*base), index );
	//to support rational fixation this folding of constant lookups to known tuples is required
	if( base->type.extr().is_literal() && base->type.extr().is<intr_type_tuple>() ) {
		auto tup = access_literal_tuple( base );
		STMT_CHECK( *ex, index < tup->sub.size() );
		STMT_CHECK( *ex, tup->sub[index].expr->type.extr().is_literal() );
		return tup->sub[index].expr;
	}
	
	//convert to field access
	auto & bt = base->type.extr().cast<intr_type_tuple>();
	STMT_CHECK( *ex, index < bt.sub.size() );
	
	shared_ptr<expr_field> fex;
	//process_field caller will already have a field
	if( ex->form == expression::f_field ) {
		fex = std::static_pointer_cast<expr_field>(ex);
	} else {
		fex = make_shared<expr_field>();
		fex->src_loc = ex->src_loc;
		fex->base = base;
	}
	
	fex->field_index = index;
	auto t = bt.sub[index].type.extr();
	if( is_lvalue ) {
		t.lvalue = true;
	}
	fex->type = t;
	fex->mark_resolved();
	return fex;
}

enum field_flags {
	//TODO: no_bind is a hack
	ff_no_bind = 1 << 0,
	ff_global_only = 1 << 1,
};

//TODO: try to combine more of this with process_symbol. In theory it's simply a different orign scope lookup
shared_ptr<expression> access_scope_field( shared_ptr<expr_field> field, shared_ptr<scope const> sc, 
	context ctx, unsigned flags = 0 ) {
	auto & t = *field->base->type.extr().intr;
	DEBUG( "SCOPE-IN", dump::get(field->base->type), field->field );
	
	//TODO: Local symbol only (using if_get_local_symbol causes classes to break somehow)
	auto const & msym = sc->if_get_symbol( field->field );
	if( !msym ) {
		STMT_FAIL( *field, std::string( "unknown-field: " ) + field->field );
	}
	
	auto bind = [&](shared_ptr<expr_field>  ef )->shared_ptr<expression> {
		shared_ptr<expression> res;
		
		DEBUG( "SCOPE-BIND", dump::get(ef->type) );
		bool lit = false;
		if( ef->sref->type.extr().is_literal() ) {
			STMT_CHECK( *field, ef->sref->literal_value );
			res = ef->sref->literal_value;
			lit = true;
		} else if( t.is<intr_type_module>() || ef->sref->lifetime == declaration::l_global ) {
			auto sym = ef->sref->create_expr_symbol();
			sym->src_loc = field->src_loc;
			sym->mark_resolved();
			res = sym;
		} else {
			res = ef;
		}
		
		if( !lit && field->base->type.extr().is_lvalue() ) {
			auto t = res->type.extr();
			t.lvalue = true;
			res->type = t;
		}
		
		//bind member functions (TODO: only "member" functions, that info is lost now, hmmm)
		auto * vtf = res->type.extr().if_cast<intr_type_function>();
		if( vtf && vtf->is_unbound() && !(flags & ff_no_bind)) {
			auto cx = make_shared<expr_compound>();
			cx->push_sub( ef->base );
			cx->push_sub( res );
			cx->src_loc = field->src_loc;
				
			auto ct = make_shared<intr_type_contextual>();
			ct->get_context() = ef->base->type;
			
			auto t = res->type.extr();
			if( !lit && ef->base->type.extr().is_lvalue() ) {
				t.lvalue = true;
			}
			ct->get_unbound() = t;
			cx->type = extr_type( ct );
			if( cx->type.get_extant() >= extant_level::resolved ) {
				cx->mark_resolved();
			}
			return cx;
		}
			
		if(res->type.get_extant() >= extant_level::resolved ) {
			res->mark_resolved();
		}
		DEBUG( "SCOPE-OUT", dump::get(res->type) );
		return res;
	};

	auto tex = resolve_scope_symbol_type( *msym, field->param, field->field );
	if( tex ) {
		ctx.rewrite( field, tex );
		return tex;
	}
	
	STMT_CHECK( *field, msym->is_var(), "invalid-symbol-type", error::tag_error_symbol( field->field ) );
	
	for( auto & sr : msym->vars ) {
		if( sr->type.get_extant() < extant_level::input ) {
			return field;
		}
		//FEATURE: possibly just filter them out instead
		if( (flags & ff_global_only) && !sr->is_lifetime_global() ) {
			STMT_FAIL( *field, std::string( "expecting-global-lifetime: " + field->field ) );
		}
	}
		
	if( msym->vars.size() == 1 ) {
		field->sref = msym->vars[0];
		field->type = field->sref->type;
			
		auto q = bind( field );
		DEBUG( "RESCOPE-IN", dump::get(field->base->type), field->field );
		return q;
	} else {
		auto select = make_shared<intr_type_tuple>();
		select->selectable = true;
		auto multi = make_shared<expr_compound>();
		multi->src_loc = field->src_loc;

		auto shared_base = ctx.applicator().create_shared( field->base, false );
		for( auto & sr : msym->vars ) {
			auto sym = make_shared<expr_field>();
			sym->src_loc = field->src_loc;
			sym->field = sr->symbol;
			sym->sref = sr;
			sym->type = sr->type;
			sym->base = shared_base;
			
			auto r = bind(sym);
			
			select->push_sub( r->type );
			multi->push_sub( r );
		}
		multi->type = extr_type(select);
		if( multi->type.is_complete() ) {
			multi->mark_resolved();
		}
		return multi;
	}
}

namespace {
	shared_ptr<expression> rewrite_field_optional( shared_ptr<expr_field_like> field, context ctx ) {
		STMT_CHECK( *field, field->base->type.extr().is_optional(), "require-optional-base" );
		auto shared_base = ctx.applicator().create_shared( field->base );
		auto cond = make_shared<expr_conditional>();
		cond->op = expr_conditional::o_optional_cond;
		
		auto has = make_shared<expr_special_field>();
		has->field = expr_special_field::f_optional_set;
		has->base = shared_base;
		cond->cond = has;
		
		auto cost = ctx.converter().remove_optional(shared_base->type);
		field->base = ctx.applicator().apply(cost, shared_base);
		field->optional = false;
		cond->a = field;
		
		ctx.rewrite(field, cond);
		return cond;
	}
}

shared_ptr<expression> process_subscript( shared_ptr<expr_subscript> ex, 
	context ctx ) {
	COND_PROC_EXPRESSION( ex->base, ctx, ex );
	COND_PROC_EXPRESSION( ex->offset, ctx, ex );
	
	//check if we have a range subscript (ex: `q#1..3`)
	bool range = false;
	auto cost = ctx.converter().convert_retain_literal( extr_type(ctx.types().get_integer_range()), 
		ex->offset->type );
	if( cost.okay() ) {
		ex->offset = ctx.applicator().apply( cost, ex->offset );
		range = true;
	} else {
		static auto vti = extr_type(make_shared<intr_type_integer>()); //TODO: use types()
		cost = ctx.converter().convert_retain_literal( type_ref(vti), ex->offset->type );
		ex->offset = ctx.applicator().apply( cost, ex->offset );
	}
	
	if( ex->optional ) {
		return rewrite_field_optional(ex, ctx);
	}
	
	cost = ctx.converter().convert_require_value( ex->base->type );
	ex->base = ctx.applicator().apply( cost, ex->base );

	if( ex->base->type.extr().is<intr_type_tuple>() ) {
		
		auto & tup_type = ex->base->type.extr().cast<intr_type_tuple>();
		//FEATURE: support constexpr propagation and allow it here
		//TODO: use literal type evaluation here
		if( range ) {
			
			STMT_CHECK( *ex, ex->offset->form == expression::f_compound );
			auto et = static_cast<expr_compound*>(ex->offset.get());
			STMT_CHECK( *ex, et->sub.size() == 2 );
			
			auto integer_of = [&](size_t index ) {
				STMT_CHECK( *ex, et->sub[index].expr->form == expression::f_literal_number );
				auto ecn  = static_cast<expr_literal_number*>(et->sub[index].expr.get());
				STMT_CHECK( *ex, ecn->src_form == expr_literal_number::sf_integer );
				return ecn->integer_value();
			};
			auto begin = integer_of(0);
			auto end = integer_of(1);
			STMT_CHECK( *ex, begin <= end && begin >= 0, "invalid-range-index" );
			STMT_CHECK( *ex, unsigned(end) <= tup_type.sub.size(), "index-out-of-range" );
			
			auto result = make_shared<cvt_expression>( cvt_expression::c_pack );
			result->pack = tup_type.pack;
			for( size_t i=0; i < size_t(end-begin); ++i ) {
				auto unpack = make_shared<cvt_expression>( cvt_expression::c_field );
				unpack->field = size_t(begin) + i;
				result->sub.push_back({ unpack, "" }); //TODO: preserve name
			}
			auto ca = ctx.applicator();
			if( ex->base->type.extr().is_lvalue() ) {
				ex->base = ca.drop_lvalue( ex->base );
			}
			auto rex = ca.apply( *result, ca.create_shared( ex->base ) );
			rex->mark_resolved();
			return rex;
			
		} else {
			auto num = extract_literal_number(*ex->offset);
			STMT_CHECK( *ex, num.is_integral(), "invalid-tuple-index-type" );
			STMT_CHECK( *ex, !num.is_negative() && size_t(num.as_int()) < tup_type.sub.size(), 
				"index-out-of-range" );
			return access_field_index( ex, ex->base, unsigned(num.as_int()), ctx );
		}
		
	} else if( ex->base->type.extr().is<intr_type_array>() ) {

		STMT_CHECK(*ex,!range);
		auto & at = ex->base->type.extr().cast<intr_type_array>();
		STMT_CHECK(*ex, at.sub.is_extr() );
		
		auto t = at.sub.extr();
		if( ex->base->type.extr().is_lvalue() ) {
			t.lvalue = true;;
		}
		ex->type = t;
		ex->mark_resolved();
		return ex;
		
	} else {
		STMT_FAIL( *ex, std::string("index-not-supported: ") + dump::get(ex->base->type) );
	}
}


shared_ptr<expression> process_field( shared_ptr<expr_field> field, context ctx ) {
	STMT_CHECK( *field, field->base );
	field->base = process_expression(field->base,ctx);
	DEBUG("FIELD", dump::get(*field->base), dump::get(field->base->type),
		field->field );
	auto extant = field->base->get_extant();
	if( extant < extant_level::input ) {
		return field;
	}
	
	if( field->optional ) {
		return rewrite_field_optional(field, ctx);
	}

	auto cost = ctx.converter().convert_require_value( field->base->type );
	field->base = ctx.applicator().apply( cost, field->base );
	auto & t = *field->base->type.extr().intr;
	auto tuple_t = t.if_cast<intr_type_compound>();
	
	if( t.is<intr_type_tuple>() ) {
		if( extant < extant_level::resolved) {
			return field;
		}
		if( field->field == "size" ) {
			auto ecn = make_shared<expr_literal_number>();
			ecn->src_form = expr_literal_number::sf_rational;
			ecn->set_integer_value( int64_t(tuple_t->sub.size()) );
			return process_expression( ecn, ctx );
		}
		
		size_t index;
		if ( field->field_index ) {
			index = *field->field_index;
		} else {
			auto i = tuple_t->find_name( field->field );
			STMT_CHECK( *field, i, std::string("no field named: " ) + field->field );
			index = *i;
		}
		
		return access_field_index( field, field->base, index, ctx );
		
	} else if( t.is<intr_type_array>() ) {
		if( extant < extant_level::resolved) {
			return field;
		}
		auto & a = t.cast<intr_type_array>();
		STMT_CHECK( *field, field->field == "size" );
		STMT_CHECK( *field, !a.raw, "raw-array-has-no-size" );
		
		auto sf = make_shared<expr_special_field>();
		sf->field = expr_special_field::f_array_size;
		sf->base = field->base; //TODO: actually need conversion to array?
		auto cl = sf->base->type.extr();
		cl.lvalue = false;
		sf->base->type = cl; //TODO: see above
		ctx.rewrite( field, sf );
		return sf;
		
	} else if( t.is<intr_type_module>() || t.is<intr_type_classlike>() ) {
		auto & c = t.cast<intr_type_scope>();
		return access_scope_field( field, c.context, ctx );

	} else if( t.is<intr_type_type>() ) {
		auto tt = t.cast<intr_type_type>();
		auto sub = tt.sub.extr();
		
		if ( field->field == "name" ) {
			auto tex = make_shared<expr_text>();
			tex->what = expr_text::w_string;
			tex->text = unicode::decode8(dump::get(sub));
			tex->src_loc = field->src_loc;
			return process_text(tex, ctx);
			
		} else if( sub.is<intr_type_scope>() ) {
			STMT_CHECK( *field, !sub.is_decorated() );
			
			auto & st = sub.cast<intr_type_scope>();
			auto q = access_scope_field( field, st.context, ctx, ff_no_bind | ff_global_only );
			//TODO: actually should be done by access_scope_field somehow
			if( !q->is_resolved() ) {
				return field;
			}
			
			//TODO: only for constructors?
			if( sub.is_shared() ) {
				//TODO: somewhat copied from funccall_type.cpp
				auto ntype = q->type.extr();
				auto nftype = ntype.intr->cast<intr_type_function>().clone_derived();
				nftype->construct_result = sub;
				ntype.intr = nftype;
				q->type = ntype;
			}
			
			return q;
		}  else {
			STMT_FAIL( *field, "unknown-field: " + field->field );
		}
	} else {
		STMT_FAIL( *field, "type-lacks-fields: " + dump::get(field->base->type) );
	}
	
	STMT_FAIL( *field, "unreachable" );
}

}} //eon leaf::expression_typer