//typing of function calls
#include "lang/expression_typer/common.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {
namespace expression_typer {

shared_ptr<expr_bind> bind_function( shared_ptr<expression> expr, context ctx ) {

	auto & ftype = expr->type.extr().cast<intr_type_function>();
	
	auto b = make_shared<expr_bind>();
	b->unbound = expr;

	int depth = 0;
	auto sc = ctx.sc->find_context( ftype.context.extr().intr, depth );
	STMT_CHECK( *expr, sc, std::string("no-context-found: ") + dump::get(ftype) );
	
	auto t = make_shared<expr_current>();
	t->what = expr_current::w_context;
	t->depth = depth;
	t->which_scope = sc;
	b->context = t;
	
	ctx.rewrite( expr, b );
	return b;
}

namespace {
	/**
		Determine what function is actually being called.
	*/
	result resolve_function( shared_ptr<expr_funccall> expr, context ctx ) {
		
		shared_ptr<expression> func_sym = expr->func;
		auto cost = ctx.converter().convert_any_callable( func_sym->type );
		expr->func = ctx.applicator().apply( cost, func_sym );
		func_sym = expr->func;
		
		//TODO: see note in type_converter::convert_any_callable. Something isn't right here. Perhaps
		//the selectable should be handled prior to covnert_any_callable and match should take a
		//conversion function to check on each element.
		auto * st = func_sym->type.extr().if_cast<intr_type_tuple>();
		if(!st || !st->selectable ) {
			return result();
		}
		
		//don't need completeness on args as apply_call_signature can resolve further
		expr->args = process_expression( expr->args, ctx );
		if( expr->args->get_extant() < extant_level::input ) {
			return expr;
		}

		auto matches = ctx.converter().match_function( *st, *expr->args->type.extr().intr );
		if( matches.size() != 1 ) {
			
			std::ostringstream out;
			out << (matches.size() ? "multiple-matching-symbols" : "no-matching-symbol") << std::endl;
			out<< "Args: ";
			leaf::dump d( out );
			d.write( expr->args->type );
			out << std::endl;
			out << "Candidates: " << std::endl;
			for( size_t i=0; i < st->sub.size(); ++i ) {
				d.write( st->sub[i].type );
				out << std::endl;
			}
			STMT_FAIL( *expr, out.str(), 
				error::tag_error_symbol( dump::get(func_sym)  ) );
		}
		auto func_index = matches[0];

		if( func_sym->form == expression::f_compound ) {
			//avoid creating the tuple and just get the indexed item
			//NOTE: This is not an optimization but required for certain types (like parametrics)
			//TODO: captured srefs will still be captured resulting in unnecessarily large closures
			auto & fst = static_cast<expr_compound&>(*func_sym);
			STMT_CHECK( *func_sym, func_index < fst.sub.size() );
			func_sym = fst.sub[func_index].expr;
		} else {
			auto sub = make_shared<expr_field>();
			sub->src_loc = func_sym->src_loc;
			sub->base = func_sym;
			sub->field_index = func_index;
			sub->type = st->sub[func_index].type;
			STMT_CHECK( *func_sym, !sub->type.contains_parametric() );
			sub->mark_resolved();
			func_sym = sub;
		}
		
		//reconvert (see TODO:)
		auto fcost = ctx.converter().convert_any_callable( func_sym->type );
		func_sym = ctx.applicator().apply( fcost, func_sym );
		expr->func = func_sym;

		return result();
	}

	/**
		Determine an unbound function is being called and look for the appropriate context.
	*/
	result resolve_call_binding( shared_ptr<expr_funccall> expr, context ctx ) {
		if( expr->context ) {
			return result();
		}

		auto & ftype = expr->func->type.extr().cast<intr_type_function>();
		if( !ftype.is_unbound() ) {
			return result();
		}

		expr->func = bind_function( expr->func, ctx );
		return expr;
	}

	result instance_call_parametric( shared_ptr<expr_funccall> expr, context ctx ) {
		if( !expr->func->type.contains_parametric() ) {
			return result();
		}
		
		//apply fixation to get real types
		auto fix = ctx.converter().fixate( expr->args->type );
		STMT_CHECK( *expr, fix.okay(), fix.explain_error() );
		DEBUG("call_param", dump::get( fix.result ) );

		
		shared_ptr<expression> in_func;
		if( expr->func->form == expression::f_bind ) {
			auto & bind = static_cast<expr_bind&>(*expr->func);
			in_func = bind.unbound;
		} else {
			in_func = expr->func;
		}
		
		//resolve literal (TODO: abstract a bit)
		if( in_func->form == expression::f_symbol ) {
			auto & sym = static_cast<expr_symbol const&>(*in_func);
			STMT_CHECK( *expr, sym.sref && sym.sref->literal_value );
			in_func = sym.sref->literal_value;
		}
		
		STATE_CHECK( in_func->form == expression::f_funcdefn );
		auto & func = static_cast<expr_funcdefn&>(*in_func);
		//TODO: we don't really want to allow_param, we just don't want this to fail on inferred return values.
		// thus we check for contains_parametric on the result
		//see share/unit_tests/block/type/ctor_class_alt_func_2.lfb
		auto rt = type_identifier::parameterize_type( ctx, func.type, fix.result, type_identifier::df_allow_param );
		STMT_CHECK( *expr, rt.okay(), rt.explain_error() );
		// the input type must be fully known since we don't have a chance later to update it
		if( rt.result.get_extant() < extant_level::input || rt.result.contains_parametric() ) {
			return expr;
		}
		DEBUG( "Result", dump::get(rt.result), (int)rt.result.get_extant() );
		
		auto use = func.if_find_instance( rt.result );
		if( !use ) {
			use = ctx.instance_funcdefn(func, rt.result);
			use->instance_of = std::static_pointer_cast<expr_funcdefn>(in_func);
			//TODO: this is actually wrong. It assumes every future instance will use it in the exact
			//same way. At least it must be a shared_expression, but probably we need some kind
			//or intermediary
			use->applied_type = rt.result;
			func.instances.push_back(use);
			STATE_CHECK( func.instances.size() < 10 );
		}
		ctx.rewrite( in_func, use );
		
		if( expr->func->form == expression::f_bind ) {
			auto & bind = static_cast<expr_bind&>(*expr->func);
			bind.unbound = use;
			bind.type = type_ref(); //force type revaluation
		} else {
			expr->func = use;
		}

		DEBUG( "Done it?" );
		return expr;
	}
	
	result apply_call_signature( shared_ptr<expr_funccall> expr, context ctx ) {
		shared_ptr<intr_type_value_list const> func_args_type, func_returns_type;
		auto ft = expr->func->type.extr().if_cast<intr_type_function>();
		STMT_CHECK( *expr->func, ft );
		func_args_type = ft->args;
		func_returns_type = ft->returns;
		
		//now we need the return type
		if( !func_returns_type || !func_returns_type->is_complete() ) {
			return expr;
		}
		
		DEBUG("CallSig-A", dump::get(func_args_type), dump::get(expr->args->type));
		//convert arguments (even if incomplete since the conversion for parametric instances can force some types)
		auto pcost = ctx.converter().convert_function_call( *func_args_type,
			expr->args->type.extr().cast<intr_type_value_list>() );
		//there's at least one call path where the args aren't completely processed yet
		//found by unit_tests/block/type/class_shared.lfb
		if (!pcost.okay() && !expr->args->is_resolved()) {
			return expr;
		}
		
		DEBUG( "Convert" );
		STMT_CHECK( *expr, pcost.okay(), pcost.explain_error() + ": Converting from " +
			dump::get(expr->args->type) + " to " + dump::get(func_args_type) );
		expr->args = ctx.applicator().apply( pcost, expr->args );
		if( !expr->args->is_resolved() ) {
			return expr;
		}
		
		auto & args = expr->args->type.extr().cast<intr_type_value_list>();
		for ( auto & sub : args.sub ) {
			STMT_CHECK( *expr, sub.type.extr().allow_copy(), "no-copy-allowed" );
		}
		
		//1-value-list's are always collapsed
		if( func_returns_type->sub.size() == 1 )
			expr->type = func_returns_type->sub[0].type;
		else
			expr->type = extr_type(func_returns_type);

		return result();
	}
	
} //eon anon

/**
	Functions work with value-lists, but are internall represented by a tuple. So the
	function call operation takes a tuple, but the actual function works with a list.
	The result of the function-call is also a list. This is done to get the desired
	syntax dealing with function calls and multiple-returns.
*/
shared_ptr<expression> process_funccall( shared_ptr<expr_funccall> expr, 
	context ctx ) {

	//builtin's aren't really function calls, handle them first
	auto builtin = process_funccall_builtin( expr, ctx );
	if( builtin ) {
		return *builtin;
	}
	
	DEBUG( "FUNCCALL", dump::get(*expr));
	//the first part of a function call can be processed without needing the returns
	expr->func = process_expression( expr->func, ctx );
	if( expr->func->type.get_extant() < extant_level::input ) {
		return expr;
	}
	
	auto typecall = process_funccall_type( expr, ctx );	
	if( typecall ) {
		return *typecall;
	}
	
	auto ctorcall = process_funccall_ctor( expr, ctx );
	if( ctorcall ) {
		return *ctorcall;
	}

	DEBUG( "Resolve" );
	auto which = resolve_function( expr, ctx );
	if( which ) {
		return *which;
	}

	DEBUG( "Bind" );
	auto bind = resolve_call_binding( expr, ctx );
	if( bind ) {
		return *bind;
	}
	
	DEBUG( "Args" );
	expr->args = process_expression( expr->args, ctx );
	
	auto param = instance_call_parametric( expr, ctx );
	if( param ) {
		return *param;
	}
	
	DEBUG( "callSig" );
	auto sig = apply_call_signature( expr, ctx );
	if( sig ) {
		return *sig;
	}
	
	expr->mark_resolved();
	return expr;
}

} //eon expression_typer
} //eon leaf