//evaluation of builtins with function syntax
#include "lang/expression_typer/common.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {
namespace expression_typer {

namespace {
	extr_type resolve_new_func_type( expression const & where,
		type_ref type ) {
		STMT_CHECK( where, type.extr().is<intr_type_array>() );
		
		auto vt = make_shared<intr_type_function>();
		vt->access = intr_type_function::a_raw;
		
		//size of the array
		auto p = make_shared<intr_type_value_list>();
		auto i = make_shared<intr_type_integer>();
		p->push_sub( "size", extr_type(i) );
		vt->args = p;
		
		//created array
		auto r = make_shared<intr_type_value_list>();
		r->push_sub( type );
		vt->returns = r;
		
		return extr_type(vt);
	}
	
	shared_ptr<expression> process_type_scalar( shared_ptr<expr_funccall> expr, context ctx, type_ref t ) {
		auto args = static_cast<expr_value_list const*>(expr->args.get());
		
		STMT_CHECK( *expr, args->sub.size() < 2, "invalid-argument-count" );
		
		//this is an optimization
// 		if( t.is_extr() && !t.extr().is_decorated && args->sub.size() == 1) {
// 			auto result = ctx.converter().convert( t, args->sub[0].expr->type );
// 			STMT_CHECK( *expr, result.okay(), result.explain_error() );
// 			return ctx.applicator().apply( result, args->sub[0].expr );
// 		}
		
		auto nexpr = make_shared<expr_temp>();
		nexpr->init = expr_temp::i_default;
		nexpr->src_loc = expr->src_loc;
		nexpr->type = t;
		nexpr->mark_resolved();
		
		if( args->sub.size() == 1) {
			nexpr->init = expr_temp::i_intrinsic;
			auto strip = extr_type(t.extr().intr);
			auto result = ctx.converter().convert( strip, args->sub[0].expr->type );
			STMT_CHECK( *expr, result.okay(), result.explain_error() );
			nexpr->intrinsic_value = ctx.applicator().apply( result, args->sub[0].expr );
		}
		
		return nexpr;
	}
	
	shared_ptr<expression> process_type_array( shared_ptr<expr_funccall> expr, context ctx, type_ref t ) {
		auto create = make_shared<expr_funccall>();
		
		auto func = make_shared<expr_new_func>();
		func->new_type = t;
		func->type = resolve_new_func_type( *expr, t );
		func->mark_resolved();
		
		create->func = func;
		create->args = expr->args;
		ctx.rewrite( expr, create );
		return create;
	}

	shared_ptr<expression> process_type_class( shared_ptr<expr_funccall> expr, context ctx, type_ref t ) {
		auto call = create_default_ctor_call( *expr, ctx, t );
		if( !call ) {
			return expr;
		}
		call->args = expr->args;
		ctx.rewrite(expr, call);
		return call;
	}
	
}

result process_funccall_type( shared_ptr<expr_funccall> expr, context ctx ) {
	auto func_sym = expr->func;	
	if( !func_sym->type.extr().is<intr_type_type>() ) {
		return result();
	}
	auto tt = func_sym->type.extr().cast<intr_type_type>();
	
	COND_PROC_EXPRESSION( expr->args, ctx, expr );
	STMT_CHECK( *expr, expr->args->form == expression::f_value_list );
	STMT_CHECK( *expr, tt.sub.is_extr() );

	auto ex = tt.sub.extr();
	if( ex.is<intr_type_scalar>() ) {
		return process_type_scalar( expr, ctx, ex );
	} else if( ex.is<intr_type_array>() ) {
		return process_type_array(expr, ctx, ex );
	} else if( ex.is<intr_type_classlike>() ) {
		return process_type_class(expr, ctx, ex ); 
	} 

	STMT_FAIL( *expr, "unsupported-type-call" );
}

shared_ptr<expression> process_new_func( shared_ptr<expr_new_func> expr, context /*ctx*/ ) {
	STATE_CHECK( expr->is_resolved() );
	return expr;
}

shared_ptr<expression> process_temp( shared_ptr<expr_temp> expr, context /*ctx*/ ) {
	STATE_CHECK( expr->is_resolved() );
	return expr;
}

/**
	Constructs a funccall to the default constructor of a type. This does not set the `args` of the resulting
	call. The call is also not processed, allowing the caller to add a context.
*/
shared_ptr<expr_funccall> create_default_ctor_call( util::source_provider const & source, context ctx, type_ref t ) {
	STMT_CHECK( source, t.is_extr() && !t.extr().is_decorated() );
	
	auto & custom = t.extr().cast<intr_type_classlike>();
	auto ctors = custom.context->get_local_vars( "default" );
	if( ctors.size() != 1 ) {
		STMT_CHECK( source, !ctx.no_incomplete(), std::string( "missing-ctor: " ) + 
			dump::get( custom ) );
		return shared_ptr<expr_funccall>();
	}
	
	if( !ctors[0]->type.is_complete() ) {
		STMT_CHECK( source, !ctx.no_incomplete(), std::string( "incomplete-ctor: " ) +
			dump::get( custom ) );
		return shared_ptr<expr_funccall>();
	}
	
	//introduce symbol for constructor
	auto ctor = ctors[0]->create_expr_symbol();

	//force a new construction type to allow different extrinsic modifiers
	auto ntype = ctor->type.extr();
	auto nftype = ntype.intr->cast<intr_type_function>().clone_derived();
	nftype->construct_result = t;
	ntype.intr = nftype;
	ctor->type = ntype;
	ctor->mark_resolved();
	
	DEBUG( "TYPE_CLASS", dump::get(ctor->type), dump::get(t) );
	//rewrite as constructor call
	auto fc = make_shared<expr_funccall>();
	fc->func = ctor;
	return fc;
}

} //eon expressiong_typer
} //eon leaf
