#include "lang/expression_typer/common.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {
namespace expression_typer {

namespace {
	
	void set_defaults_optional( shared_ptr<expr_funcdefn> func, context ) {
		STMT_CHECK( *func, !!func->func_type );
		//force items with default to be optional before creating the type
		STMT_CHECK( *func, func->func_type->sub.size() == 2 );
		auto spec_args = make_shared<type_spec>( *func->func_type->sub[0].type );
		for( size_t i=0; i < spec_args->sub.size(); ++i ) {
			auto dit = func->arg_defaults.find(i);
			if( dit != func->arg_defaults.end() ) {
				
				STMT_CHECK( *func, spec_args->sub[i].type->get<pt_optional>(true) );
					
				auto dup = make_shared<type_spec>(*spec_args->sub[i].type);
				dup->set<pt_optional>( true );
				spec_args->sub[i].type = dup;
			}
		}
		func->func_type->sub[0].type = spec_args;
	}

	result determine_type( shared_ptr<expr_funcdefn> func, context ctx,
		shared_ptr<intr_type_function const> & func_type ) {
		//establish the type the first time here
		if( !func->type.is_defined() ) {
			DEBUG( "DETERMINE_TYPE", dump::get(*func->func_type, dump::tf_annotate),
				dump::get(func) );
			auto rt = type_identifier::determine( ctx, func->func_type, 
				type_identifier::df_imply_context | type_identifier::df_allow_param );
			if( rt.unknown_identifier() ) {
				STMT_CHECK( *func, !ctx.no_incomplete(), rt.explain_error() );
				return func;
			}
		
			func_type = std::dynamic_pointer_cast<intr_type_function const>( rt.result.extr().intr );
		} else {
			func_type = std::dynamic_pointer_cast<intr_type_function const>( func->type.extr().intr );
		}
		STMT_CHECK( *func, func_type );
		return result();
	}
	
	void set_base( shared_ptr<expr_funcdefn> func, context ctx,
		shared_ptr<intr_type_function const> & func_type ) {
		statement_block & func_block = *func->block;
		
		if( !func_block.symbols->has_base() ) {
			auto bt = func->clone ? scope::bt_clone :
				func_type->is_raw() ? scope::bt_free :
				scope::bt_parent;
			func_block.symbols->set_base( ctx.get_decl_scope(), bt );
		}
		func_block.symbols->mark_function_root(func);
	}
	
	result check_parametric( shared_ptr<expr_funcdefn> func, context,
		shared_ptr<intr_type_function const> func_type ) {
		//nothing to be done for parametrics yet (only check args, returns handled later)
		if( func_type->args->is_level_parametric() ) {
			auto t = extr_type( func_type );
			t.literal = true;
			func->type = t;
			return func;
		}
		
		return result();
	}
	
	void add_args( shared_ptr<expr_funcdefn> func, context,
		shared_ptr<intr_type_function const> func_type ) {
		if( func->added_args ) {
			return;
		}
		
		std::vector<shared_ptr<statement>> extra_block;
		func->added_args = true;
		//introduce a var_statement into the function block for each argument
		for( size_t i=0; i < func_type->args->sub.size(); ++i ) {
			auto & arg = func_type->args->sub[i];
			STATE_CHECK( arg.name.size() );

			auto vs = declaration::create();
			vs->form = declaration::f_arg;
			vs->symbol = arg.name;
			vs->type = arg.type;
			
			auto dit = func->arg_defaults.find(i);
			if( dit != func->arg_defaults.end() ) {
				vs->symbol = "$arg_" + arg.name;
				
				//add a new var that is not optional
				auto nvs = make_shared<var_statement>();
				
				auto decl = declaration::create();
				decl->symbol = arg.name;
				
				auto no_opt_type = arg.type.extr();
				no_opt_type.set_optional(false);
				decl->type = no_opt_type;
				
				nvs->decl = decl;
				//func->block->symbols->add( decl, scope::add_in_order );
				DEBUG("PosteR");
				
				auto def_expr = make_shared<expr_conditional>();
				def_expr->op = expr_conditional::o_optional_or;
 				auto sym = make_shared<expr_symbol>(); //TODO: a direct reference to the declaration would be better/safer
 				sym->symbol = vs->symbol;
				def_expr->cond = sym;
				def_expr->b = dit->second;
				nvs->value = def_expr;
				
				extra_block.push_back( nvs );
// 				auto as = make_shared<assign_statement>();
// 				as->how = assign_statement::h_default;
// 				auto sym = make_shared<expr_symbol>(); //TODO: a direct reference to the declaration would be better/safer
// 				sym->symbol = arg.name;
// 				as->left = sym;
// 				as->right = dit->second;
// 				extra_block.push_back( as );
			}
			
			func->block->symbols->add( vs, scope::add_in_order );
		}
		
		//put extra statements in order at beginning of block
		func->block->list.insert( func->block->list.begin(), extra_block.begin(), extra_block.end() );
		func->arg_defaults.clear();
	}
	
	void maybe_add_returns( shared_ptr<expr_funcdefn> func, context,
		shared_ptr<intr_type_function const> func_type ) {
		
		//auto & func_type = func->type.fun_cast<intr_type_function>();
		if( !func->added_returns && func_type->returns ) {
			std::vector<shared_ptr<statement>> extra_block;
			//and for each return (if specified, otherwise they don't need to be initialized...?)
			//OPT: no unnamed returns need to be initialized actually..
			auto & returns = *func_type->returns;
			STATE_CHECK( returns.is<intr_type_value_list>() );
			
			for( size_t i=0; i < returns.sub.size(); ++i ) {
				auto & ret = returns.sub[i];

				auto vs = declaration::create();
				vs->form = declaration::f_return;
				vs->symbol = ret.name; //returns may be anonymous
				vs->type = ret.type;

				auto init = make_shared<init_statement>();
				init->sref = func->block->symbols->add( vs, scope::add_in_order );
				init->no_release = true;
				extra_block.push_back( init );
			}
			
			func->added_returns = true;
			//put extra statements in order at beginning of block
			func->block->list.insert( func->block->list.begin(), extra_block.begin(), extra_block.end() );
		}
	}
		
	/**
		Here a decision is made about what context type the function will have. The current
		approach is a fairly rigid selection based on scope and presence of captures. The ideal
		solution would simply be to mark everything as a shared context, and then optimize
		later. But for "clone" functions they must be evaluated now to ensure they only use
		available captures.
		
		Note: the ideal solution has very few use cases that require it. The fixed structure below
		is way beyond what most languages offer and is very sufficient  -- an optimier can
		still remove any redudnant things left in from here.
	*/
	result process_block( shared_ptr<expr_funcdefn> func, context ctx,
		shared_ptr<intr_type_function> func_type ) {
		
		if( func->block->symbols->get_base()->is_declarative() || 
			func->block->symbols->get_base()->is_dynamic() ) {
			STMT_CHECK( *func, func->block );
			//place onto pending blocks (so this block can see later values)
			if( !func->pend_block ) {
				//pushing an empty block would mess with progress tracking (completed_statements)
				if( !func->block->list.empty() ) {
					ctx.add_pending_job( make_shared<typer::job_statement_block>(func->block) );
				}
				func->pend_block = true;
			}
			
			if( !func->block->symbols->get_base()->is_class_root() || !func_type->context.is_defined() ) {
				//STMT_CHECK( *func, captures.empty() ); //can't really check since block isn't processed yet!
				func_type->access = intr_type_function::a_raw;
			}
		} else {
			STMT_CHECK( *func, !func_type->context.is_defined() ); //hmm, should have been a class...
			
			//note this ctx is not used by the block, symbols has its own
			if( !ctx.proc_statement_block( func->block ) ) {
				return func;
			}
			auto captures = func->block->symbols->get_captures();
			STMT_CHECK( *func, func->block->is_complete() );
			
			if( captures.empty() ) {
				func_type->access = intr_type_function::a_raw;
			} else if( func->clone ) {
				//create a custom type for the captures
				func_type->access = intr_type_function::a_unbound;
				auto func_ctx = make_shared<intr_type_closure>();
				func_ctx->style = intr_type::s_distinct;
				func_ctx->what = intr_type_closure::w_captures;
				func_ctx->context = func->block->symbols;
				func_type->context = extr_type(func_ctx);
			} else {
				//attach to parent shared refers context
				func_type->access = intr_type_function::a_unbound;
				func_type->context = func->block->symbols->get_base()->get_context_type(true);
			}
		}
		
		return result();
	}
	
	result complete_returns( shared_ptr<expr_funcdefn> func, context ctx,
		shared_ptr<intr_type_function> func_type ) {
		//if block is complete but had no value returned
		if( !func_type->returns || func_type->returns->is_level_parametric() 
			|| func->pend_returns.size() ) { //TODO: I don't like the pend_returns check here, the type is already known!!!
			if( !func->block->is_complete() ) {
				return func;
			}
		
			if( func->pend_returns.empty() ) {
				auto evl = make_shared<intr_type_value_list>();
				func_type->returns = evl;
			} else {
				//unify all returned types
				std::vector<type_ref> types;
				for( auto & t : func->pend_returns ) {
					STATE_CHECK(t.rs->value);
					types.push_back( t.rs->value->type );
				}
				auto unif = ctx.converter().unify( &types[0], types.size() );
				STMT_CHECK(*func, unif.okay());
				
				//infer the resulting type
				auto inf_type = type_identifier::infer_value_list( ctx, unif.result, false );
				STMT_CHECK(*func, inf_type.okay());
				func_type->returns = inf_type.result;
				
				//TODO: should I perhaps just scan the function again, marking return as half-complete?
				for( auto & pr : func->pend_returns ) {
					pr.rs->local_complete = false; //TODO: yeah,ugly
					ctx.proc_statement( pr.rs, pr.sc );
				}
				func->pend_returns.clear();
			}
			maybe_add_returns( func, ctx, func_type );
		}
		
		return result();
	}
	
} //eon anon

shared_ptr<expression> process_funcdefn( shared_ptr<expr_funcdefn> func, 
	context ctx ) {

	set_defaults_optional( func, ctx );
	
	shared_ptr<intr_type_function const> func_type;
	auto res = determine_type( func, ctx, func_type );
	if( res ) {
		return *res;
	}
	
	set_base( func, ctx, func_type );
	
	res = check_parametric( func, ctx, func_type );
	if( res ) {
		return *res;
	}
	
	add_args( func, ctx, func_type );
	
	maybe_add_returns( func, ctx, func_type);

	auto nfunc_type = func_type->clone_derived();
	//a preliminary incomplete type
	auto t =extr_type(nfunc_type);
	t.literal = true;
	func->type = t;
	DEBUG( "FUNCDEFN", dump::get(func->type), (int)func->type.get_extant() );
	
	res = process_block( func, ctx, nfunc_type );
	if( res ) {
		return *res;
	}
	
	res = complete_returns( func, ctx, nfunc_type );
	if( res ) {
		return *res;
	}
	
	//type is now complete
	nfunc_type->local_complete = true;
	
	ctx.get_owner()->get_module()->funcdefns.insert(func);

	func->mark_resolved();
	
	//if not unbound the "clone" changes nothing
	if( func->clone && nfunc_type->is_unbound() ) {
		STMT_CHECK(*func, func->block->symbols->get_base()->is_imperative() );
			
		auto wrap = make_shared<expr_bind>();
		wrap->clone = true;
		wrap->unbound = func;
		wrap->construct_result_type();	
		wrap->mark_resolved();
		return wrap;
	}

	return func;
}

} //eon expression_typer
} //eon leaf