#include "lang/type.hpp"
#include "lang/type_spec.hpp"
#include "lang/scope.hpp"
#include "lang/dump.hpp"

#include "util/validation_context.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

//common pattern for recursive validation
#define VALIDATE_RECURSE() \
	if( vctx.cover(this) ) \
		return; \

#define COMPARE_RECURSE(ret_expr) \
	if( vctx.in(this) ) \
		return ret_expr; \
	util::validation_scope vscope( vctx, this );
	
namespace leaf {

//CPP: The neverending story of whether this should be needed or not!
/*static*/ intr_type::fun_class_t const intr_type::f_invalid;

/*static*/ char const* intr_type::name_of( fun_class_t f ) {
	switch( f )
	{
		#define FUN_CLASS_TYPE(s) \
			case f_##s: return #s;
		INTR_TYPE_FUN_CLASSES
		#undef FUN_CLASS_TYPE
	}
	return "corrupt";
}

extant_level min_extant( extant_level into, extant_level from ) {
	//normal combination erases input level, only valid in am inferred function return
	if( from == extant_level::input) {
		from = extant_level::incomplete;
	}
	return std::min( into, from );
}

extant_level min_extant_return( extant_level into, extant_level from ) {
	return std::min( into, from );
}

extr_type::extr_type( serial_in & si ) {
	read( si );
}

void extr_type::validate(util::validation_context & vctx) const
{
	STATE_CHECK( intr );
	intr->validate(vctx);
	
	if( is<intr_type_value_list>() ) {
		//value-lists are not real values
		STATE_CHECK( !is_shared() );
		STATE_CHECK( !is_lvalue() );
		STATE_CHECK( !is_optional() );
	}
}

void intr_type::clear_param( ) {
	params.clear();
	instance_of.reset();
	instances.clear();
	name_scope.reset();
	clear_distinct_name();
}

intr_type::intr_type( fun_class_t fun_class ) {
	this->fun_class = fun_class;
	cref = 0;
	style = s_plain;
	//start true since only a few special paths produce incomplete types
	local_complete = true;
	is_export = false;
}

intr_type::~intr_type() {
}

std::string intr_type::get_qualified_id() const {
	std::string use;
	if( has_distinct_name() ) {
		use = *distinct_name;
	} else {
		use = get_qualified_id_part();
	}
	
	if( !name_scope ) {
		return use;
	}
	return name_scope->qualify_name( use, ":" );
}

void intr_type::validate(util::validation_context &) const {
	STATE_CHECK( fun_class >= f_array && fun_class <= f_value_list, dump::get(this) );
	STATE_CHECK( get_extant() >= extant_level::resolved, dump::get(this) );
	STATE_CHECK( !has_distinct_name() || style != s_plain, dump::get(this) );
}

int intr_type::compare( intr_type const & o, bool logical ) const
{
	util::validation_context vctx;
	return compare_impl( o, logical, vctx );
}

#define CINT( x, y ) { intptr_t q = intptr_t(x) - intptr_t(y); if( q ) return int(q); }
int intr_type::compare_impl( intr_type const & o, bool /*logical*/, util::validation_context & ) const
{
	CINT( fun_class, o.fun_class )
	
	CINT( applied_params.size(), o.applied_params.size() );
	for( size_t i=0; i < applied_params.size(); ++i ) {
		int c = applied_params[i].compare( o.applied_params[i], false );
		if( c ) {
			return c;
		}
	}

	if( is_whole() ) {
		if( !o.is_whole() ) {
			return -1;
		}
		
		auto a = get_qualified_id();
		auto b = o.get_qualified_id();
		if( a < b ) {
			return -1;
		} 
		if(  b < a ) {
			return 1;
		}
		return 0;
	}
	
	if( instance_of ) {
		if( !o.instance_of ) {
			return 1;
		}
		int ic = instance_of->compare( *o.instance_of, false );
		if( ic != 0 ) {
			return ic;
		}
	} else if( o.instance_of ) {
		return -1;
	}
	
	return 0;
}

void intr_type::complete_with( intr_type const & o ) {
	STATE_CHECK( !local_complete );
	params = o.params;
	local_complete = true;
}

shared_ptr<intr_type const> intr_type::if_find_instance( std::vector<extr_type> const & params ) const {
	for( auto & inst : instances ) {
		if( params.size() != inst->applied_params.size() ) {
			continue;
		}
		bool okay = true;
		for( size_t i=0; i < params.size(); ++i ) {
			if( params[i] != inst->applied_params[i] ) {
				okay = false;
				break;
			}
		}
		if( okay ) {
			return inst;
		}
	}
	
	return shared_ptr<intr_type const>();
}

///////////////////////////////////////////////////////////
// type_ref
int extr_type::compare( extr_type const & o, bool logical ) const {
	util::validation_context vctx;
	return compare_impl( o, logical, vctx );
}

int extr_type::compare_impl( extr_type const& o, bool logical, util::validation_context & vctx ) const {
	STATE_CHECK( intr )
	CINT( lvalue, o.lvalue )
	CINT( reference, o.reference )
	CINT( optional, o.optional )
	if( !logical ) {
		CINT( literal, o.literal )
		CINT( cvt_optional, o.cvt_optional )
	}
	return intr->compare_impl( *o.intr, logical, vctx );
}

std::string extr_type::get_qualified_id() const {
	STATE_CHECK( intr );
	auto t = intr->get_qualified_id();
	t += "$";
	if( lvalue ) {
		t += "L";
	}
	if( optional ) {
		t += "O";
	}
	if( lossy ) {
		t += "Y";
	}
	if( cvt_optional ) {
		t += "C";
	}
	if( literal ) {
		t += "T";
	}
	switch( reference ) {
		case r_value: t += "V"; break;
		case r_shared: t += "S"; break;
		case r_value_ptr: t += "P"; break;
	}
	return t;
}

#define COMPARE_SUPER(t_,o_,logical,vctx) \
	int q = t_::compare_impl( o_, logical, vctx ); \
	if( q ) \
		return q;

///////////////////////////////////////////////////////////
// type_spec
bool type_spec::extrinsic_free() const
{
	//TODO: this seems incomplete
	return get<pt_reference>(extr_type::r_value) == extr_type::r_value
		&& !has<pt_optional>()
		&& !has<pt_lvalue>();
}

///////////////////////////////////////////////////////////
// intr_type_rational
int intr_type_rational::compare_impl( intr_type const& o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER(intr_type_number,o,logical,vctx);
	auto & ov = static_cast<intr_type_rational const&>(o);

	if( integer_signed != ov.integer_signed )
		return integer_signed < ov.integer_signed ? -1 : 1;
		
	bool a = !!integer_bitsize;
	bool b = !!ov.integer_bitsize;
	if( a != b )
		return a < b ? -1 : 1;
	if( a )
		return int(*integer_bitsize) - int(*ov.integer_bitsize);
	return 0;
}

std::string intr_type_rational::get_qualified_id_part() const {
	std::string t = "$R";
	if( integer_signed ) {
		t += "+";
	}
	if( integer_bitsize ) {
		char buf[32];
		sprintf( buf, "%d", *integer_bitsize );
		t += buf;
	}
	return t;
}

///////////////////////////////////////////////////////////
// intr_type_fixed_number
int intr_type_fixed_number::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER(intr_type_number,o,logical,vctx)
	auto & ov = static_cast<intr_type_fixed_number const &>(o);
		
	if( data_bitsize != ov.data_bitsize )
		return data_bitsize < ov.data_bitsize ? -1 : 1;
		
	return 0;
}

void intr_type_fixed_number::validate(util::validation_context & vctx) const
{
	intr_type::validate(vctx);
	STATE_CHECK( data_bitsize > 0 );
}

void intr_type_float::validate(util::validation_context & vctx) const
{
	intr_type_fixed_number::validate(vctx);
	STATE_CHECK( platform::if_get_float( data_bitsize ) );
}

std::string intr_type_fixed_number::get_qualified_id_end() const {
	char buf[32];
	sprintf( buf, "%d", data_bitsize );
	return buf;
}

///////////////////////////////////////////////////////////
// intr_type_compound
int intr_type_compound::compare_impl( intr_type const & o, bool logical, 
	util::validation_context & vctx ) const {
	
	COMPARE_RECURSE(0)
	COMPARE_SUPER(intr_type,o,logical,vctx)
	auto & co = static_cast<intr_type_compound const&>(o);
	
	if( sub.size() != co.sub.size() )
		return sub.size() < co.sub.size() ? -1 : 1;
	
	for( size_t i=0; i < sub.size(); ++ i )
	{
		if( sub[i].name != co.sub[i].name )
			return sub[i].name < co.sub[i].name ? -1 : 1;
			
		int q = sub[i].type.compare_impl( co.sub[i].type, logical, vctx );
		if( q )
			return q;
	}

	return 0;
}

bool intr_type_compound::is_fixated_impl(util::validation_context & vctx) const {
	COMPARE_RECURSE(true)
	
	for( auto & s : sub ) {
		if( !s.type.is_fixated_impl(vctx) )
			return false;
	}
	return true;
}

extant_level intr_type_compound::get_extant_impl(util::validation_context & vctx) const {
	COMPARE_RECURSE(extant_level::complete)
	
	auto level = super::get_extant_impl(vctx);
		
	for( auto & s : sub ) {
		auto slevel = s.type.get_extant_impl(vctx);
		level = min_extant( level, slevel );
	}
	
	return level;
}

extant_level intr_type_contextual::get_extant_impl( util::validation_context & vctx ) const {
	COMPARE_RECURSE(extant_level::complete);
	
	//TODO: we need to skip this level to be correct, but it's very questionable, see note in header file
	auto level = super::super::get_extant_impl(vctx);
	level = min_extant(level, get_context().get_extant_impl(vctx) );
	//contextual can preserve input level
	level = min_extant_return(level, get_unbound().get_extant_impl(vctx) );
	
	return level;
}

//TODO: common with above, this isn't right...
extant_level intr_type_tuple::get_extant_impl( util::validation_context & vctx ) const {
	
	if (!selectable)
		return super::get_extant_impl(vctx);
		
	COMPARE_RECURSE(extant_level::complete);
	auto level = super::super::get_extant_impl(vctx);
		
	for( auto & s : sub ) {
		auto slevel = s.type.get_extant_impl(vctx);
		level = min_extant_return( level, slevel );
	}
	
	return level;
}

bool intr_type_compound::contains_parametric_impl( util::validation_context & vctx ) const {
	COMPARE_RECURSE(false);
	
	for( auto & s: sub ) {
		if( s.type.contains_parametric_impl(vctx) ) {
			return true;
		}
	}
	return false;
}

void intr_type_compound::validate(util::validation_context & vctx) const {
	VALIDATE_RECURSE()
	
	intr_type::validate(vctx);
	for( auto & s : sub )
		s.type.validate(vctx);
}

boost::optional<size_t> intr_type_compound::find_name( std::string const & name ) const {
	for( size_t i=0; i < sub.size(); ++i ) {
		if( name == sub[i].name )
			return i;
	}
	return boost::optional<size_t>();
}

void intr_type_compound::complete_with( intr_type_compound const & o ) {
	super::complete_with( o );
	sub = o.sub;
}

std::string intr_type_compound::get_qualified_id_end() const {
	std::string t = "<";
	for( size_t i=0; i < sub.size(); ++i ) {
		if( i ) {
			t += ",";
		}
		t += sub[i].name;
		t += "=";
		t += sub[i].type.get_qualified_id();
	}
	t += ">";
	return t;
}

///////////////////////////////////////////////////////////
// intr_type_value_list
//CPP: not really needed, but GCC requires some function to appear somewhere!
void intr_type_value_list::validate(util::validation_context & vctx) const
{
	intr_type_compound::validate(vctx);
}

///////////////////////////////////////////////////////////
// intr_type_tuple
void intr_type_tuple::validate(util::validation_context & vctx) const
{
	intr_type_compound::validate(vctx);
	//DEBUG( "PACK", pack, p_normal, p_byte );
	STATE_CHECK( pack >= p_normal && pack <= p_byte );
}

int intr_type_tuple::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER( intr_type_compound, o, logical, vctx );
			
	auto & tt = static_cast<intr_type_tuple const&>(o);
	if( pack != tt.pack )
		return pack < tt.pack ? -1 : 1;
	return 0;
}

void intr_type_tuple::complete_with( intr_type_tuple const & o ) {
	super::complete_with( o );
	selectable = o.selectable;
	pack = o.pack;
}

std::string intr_type_tuple::get_qualified_id_part() const {
	std::string t = "$T";
	switch( pack ) {
		case p_normal: t += "n"; break;
		case p_abi: t += "a"; break;
		case p_byte: t += "b"; break;
	}
	if( selectable ) {
		t += "s";
	}
	//distinct?
	return t += super::get_qualified_id_end();
}

///////////////////////////////////////////////////////////
// intr_type_wrap
int intr_type_wrap::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER( intr_type, o, logical, vctx )
	
	auto & wo = static_cast<intr_type_wrap const&>(o);
	return sub.compare_impl( wo.sub, logical, vctx );
}

void intr_type_wrap::validate(util::validation_context & vctx) const
{
	VALIDATE_RECURSE()
	intr_type::validate(vctx);
	sub.validate(vctx);
}

std::string intr_type_wrap::get_qualified_id_end() const {
	return sub.is_extr() ? "<" + sub.extr().get_qualified_id() + ">" : "";
}

///////////////////////////////////////////////////////////
// intr_type_array
int intr_type_array::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const {
	COMPARE_SUPER( super, o, logical, vctx )
	
	auto & ao = static_cast<intr_type_array const&>(o);
	if( ao.raw != raw ) {
		return raw < ao.raw ? -1 : 1;
	}
	return 0;
}

std::string intr_type_array::get_qualified_id_part() const {
	std::string t = "$A";
	if( raw ) {
		t += 'r';
	}
	return t + super::get_qualified_id_end();
}


///////////////////////////////////////////////////////////
// intr_type_type
int intr_type_type::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const {
	COMPARE_SUPER( super, o, logical, vctx );
	return 0;
}

std::string intr_type_type::get_qualified_id_part() const {
	std::string t = "$Y";
	return t + super::get_qualified_id_end();
}


///////////////////////////////////////////////////////////
// intr_type_function
int intr_type_function::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_RECURSE(0)
	COMPARE_SUPER( intr_type, o, logical, vctx )
	auto & fo = o.cast<intr_type_function>();
	
	if( access != fo.access ) {
		return access < fo.access ? -1 : 1;
	}
		
	if( can_fail != fo.can_fail) {
		return can_fail < fo.can_fail ? -1 : 1;
	}
		
	q = args->compare_impl( *fo.args, logical, vctx );
	if( q ) {
		return q;
	}
	
	if( returns && fo.returns ) {
		q = returns->compare_impl( *fo.returns, logical, vctx );
		if( q )
			return q;
	} else {
		return !returns ? -1 : 1;
	}
		
	auto cr = construct_result.compare_impl( fo.construct_result, logical, vctx );
	if( cr ) {
		return cr;
	}
	return context.compare_impl( fo.context, logical, vctx );
}

bool intr_type_function::is_fixated_impl(util::validation_context & vctx) const
{
	COMPARE_RECURSE(true)
	
	if( !args || !args->is_fixated_impl(vctx) || !returns || !returns->is_fixated_impl(vctx) )
		return false;
	
	if( construct_result.is_defined() && !construct_result.is_fixated_impl(vctx) ) {
		return false;
	}
	
	return !context.is_defined() || context.is_fixated_impl(vctx);
}
	
void intr_type_function::validate(util::validation_context & vctx) const
{
	VALIDATE_RECURSE()
	
	intr_type::validate(vctx);
	STATE_CHECK(args);
	args->validate(vctx);
	STATE_CHECK(returns);
	returns->validate(vctx);
	
	if( construct_result.is_defined() ) {
		construct_result.validate(vctx);
	}
	
	if( is_unbound() ) {
		STATE_CHECK(context.is_defined() && context.is_extr());
		context.validate(vctx);
		//the ir layer expects these to be only custom types
		STATE_CHECK( context.extr().is<intr_type_custom>() );
	} else {
		STATE_CHECK(!context.is_defined());
	}
}

extant_level intr_type_function::get_extant_impl( util::validation_context & vctx ) const {
	COMPARE_RECURSE(extant_level::complete);
	
	auto level = super::get_extant_impl(vctx);
	if( args ) {
		level = min_extant( level, args->get_extant_impl(vctx) );
	} else {
		return extant_level::incomplete;
	}
	
	if( returns ) {
		level = min_extant( level, returns->get_extant_impl(vctx) ); //TODO: hmm, shouldn't minimum be "input"?
	} else {
		level = min_extant_return( level, extant_level::input );
	}
	
	return level;
}

bool intr_type_function::contains_parametric_impl( util::validation_context & vctx ) const {
	COMPARE_RECURSE(false);
	return (args && args->contains_parametric_impl(vctx) ) ||	
		(returns && returns->contains_parametric_impl(vctx));
}

std::string intr_type_function::get_qualified_id_part() const {
	std::string t = "$C";
	switch( access ) {
		case a_bound: t += "b"; break;
		case a_raw: t += "r"; break;
		case a_unbound: t += "u"; break;
		case a_construct: t += "c"; break;
	}
	t += "<";
	t += args->get_qualified_id();
	t += ",";
	t += returns->get_qualified_id();
	t += ",";
	if( context.is_defined() ) {
		t += context.get_qualified_id();
	}
	if( !can_fail ) {
		t += "N";
	}
	t += ">";
	return t;
}

///////////////////////////////////////////////////////////
// intr_type_scope
int intr_type_scope::compare_impl( intr_type const & o, bool logical, util::validation_context & vctx ) const
{
	COMPARE_SUPER(intr_type,o,logical,vctx);
	intptr_t idf = reinterpret_cast<intptr_t>(this) - 
		reinterpret_cast<intptr_t>(&o);
	if( idf != 0 )
		return idf < 0 ? -1 : 1;
	return 0;
}

void intr_type_scope::validate(util::validation_context & vctx) const
{
	VALIDATE_RECURSE()
	intr_type::validate(vctx);
	STATE_CHECK( context );
	//context->validate(vctx); //TODO: hmmm...
}

std::string intr_type_scope::get_qualified_id_part() const {
	char buf[64];
	sprintf( buf, "$S<%ld>", intptr_t(this) );
	return buf;
}

///////////////////////////////////////////////////////////
// intr_type_contextual
void intr_type_contextual::validate( util::validation_context & vctx ) const {
	super::validate( vctx );
	STATE_CHECK( sub.size() == 2 );
}

///////////////////////////////////////////////////////////
// intr_type_custom
void intr_type_custom::validate( util::validation_context & vctx ) const {
	super::validate( vctx );
	STATE_CHECK( instance_of || style == s_distinct );
}

///////////////////////////////////////////////////////////
// intr_type_classlike
var_vector_t intr_type_classlike::get_type_decls() const {
	return context->get_local_decls( true );
}

void intr_type_classlike::validate( util::validation_context & vctx ) const {
	if( params.size() ) {
		return;
	}
	super::validate( vctx );
}

///////////////////////////////////////////////////////////
// intr_type_closure
var_vector_t intr_type_closure::get_type_decls() const {
	auto sc = dynamic_cast<scope const*>(context.get() );
	STATE_CHECK( sc );
	
	switch( what ) {
		case w_share_context:
			return sc->get_context_vars();
		case w_captures:
			return sc->get_captures();
	}
	STATE_FAIL();
}

///////////////////////////////////////////////////////////
// intr_type_compat
int intr_type_compat::compare_impl( intr_type const & o, bool logical,util::validation_context & vctx ) const { 
	COMPARE_SUPER(super,o,logical,vctx);
	auto & ov = static_cast<intr_type_compat const&>(o);
	auto type_diff = type - ov.type;
	if( type_diff != 0 ) {
		return type_diff;
	}
	return 0;
}

} //eon leaf