	#include "lang/scope.hpp"
#include "lang/declaration.hpp"
#include "lang/statement_block.hpp"
#include "lang/expression.hpp"
#include "lang/serial.hpp"
#include "lang/type_spec.hpp"

#define SHOW_DEBUG 0
#include "lang/dump.hpp"
#include "util/debug.hpp"

namespace leaf
{

scope::scope( scope_style_t ss) {
	scope_style = ss;
	class_root = false;
	base_type = bt_none;
	fref = 0;
	no_init_release = false;
	pseudo = false;
}

/*static*/ shared_ptr<scope> scope::create( scope_style_t style  ) {
	auto sc = shared_ptr<scope>( new scope(style) );
	return sc;
}

/*static*/ shared_ptr<scope> scope::create_pseudo( scope_style_t style  ) {
	auto sc = shared_ptr<scope>( new scope(style) );
	sc->pseudo = true;
	return sc;
}

void scope::detach_base() {
	PRE_CHECK( base_type && this->base );
	base_type = bt_none;
	base.reset();
}

void scope::set_base( shared_ptr<scope> base, base_type_t type ) {
	//allow calling set_base with same base again (simplifies typer logic)
	PRE_CHECK( !base_type || (base_type == type && this->base.get() == base.get()) );
	this->base = base;
	base_type = type;
}

bool scope::is_base_reachable( shared_ptr<scope const> base ) const {
	if( this == base.get() ) {
		return true;
	}
	
	if (this->base) {
		return this->base->is_base_reachable(base);
	}
	return false;
}

void scope::set_name( std::string name ) {
	scope_name = name;
}

void scope::mark_function_root(shared_ptr<expr_funcdefn> expr) {
	PRE_CHECK( !class_root );
	function_root = expr;
}

shared_ptr<expr_funcdefn> scope::get_function_root_expr() const {
	PRE_CHECK(function_root);
	return function_root;
}

void scope::mark_class_root( type_ref context ) {
	PRE_CHECK( context.is_extr() && context.extr().lvalue );
	PRE_CHECK( !function_root )
	class_root = true;
	this->context = context;
}

shared_ptr<declaration> scope::add( shared_ptr<declaration> decl, unsigned flags )
{
	DEBUG("ADD", decl->symbol, dump::get(decl->type), dump::get(decl->spec_type) );
	PRE_CHECK( !decl->var_scope );
	bool is_anon = decl->form == declaration::f_return && decl->symbol.empty();
	
	if( !is_anon) {
		auto it = symbols.find(decl->symbol);
		if( decl->form == declaration::f_single || decl->form == declaration::f_arg
			|| decl->form == declaration::f_return ) {
			PRE_CHECK( it == symbols.end(), "duplicate-symbol", error_symbol( decl->symbol ) );
			
			auto & sym = symbols[decl->symbol];
			sym.what = scope_symbol::w_single_var;
			sym.vars.push_back( decl );
			
		} else if( decl->form == declaration::f_multi ) {
			STATE_CHECK( it == symbols.end() || it->second.is_multi_var(), "mismatched-multi-symbol", 
				error_symbol( decl->symbol ) );
				
			auto & sym = symbols[decl->symbol];
			sym.what = scope_symbol::w_multi_var;
			sym.vars.push_back( decl );
			
		} else {
			STATE_FAIL("invalid-declaration-form");
		}
		
	}

	decl->var_scope = shared_from_this();
	if( flags & add_in_order )
		in_order.push_back( decl );
	
	return decl;
}

bool scope::is_local( declaration const & vd ) const {
	if( vd.var_scope.get() == this ) {
		return true;
	}
	if( !base || base_type != bt_local ) {
		return false;
	}
	return base->is_local(vd);
}

bool scope::is_strict_local( declaration const & vd ) const {
	return vd.var_scope.get() == this;
}

scope & scope::get_local_base() {
	auto sc = this;
	while( sc->base_type == bt_local ) {
		sc = sc->base.get();
	}
	return * sc;
}

void scope::capture_var_add_to_context(shared_ptr<declaration> sr) {
	STATE_CHECK( base_type != bt_local, "no-context-in-local-scopes" );
	STATE_CHECK( sr->form != declaration::f_return, "return-not-supported-in-capture" );
	
	bool has = false;
	for( auto & c : context_vars ) {
		if( c.get() == sr.get() ) {
			has = true;
			break;
		}
	}
	if( !has ) {	
		context_vars.push_back(sr);
	}
}

shared_ptr<declaration> scope::capture_var( shared_ptr<declaration> sr, bool shared ) {
	if( pseudo ) {
		return sr;
	}
	
	if( sr->form == declaration::f_capture ) {
		//shortcut
		if( sr->var_scope.get() == this ) {
			return sr;
		}
		
		sr = sr->ultimate_capture;
	}
	
	//globals are always by reference, thus no capture
	if( sr->get_root_depth() == 0 || sr->lifetime == declaration::l_global ) {
		return sr;
	}
	
	return capture_var_impl(sr, shared);
}
	
shared_ptr<declaration> scope::capture_var_impl(shared_ptr<declaration> sr, bool shared) {
	
	//locals may need to be marked as shared. This must be done prior to the bt_local check before
	//since the base can't scan into it's child scopes looking for locals
	if( is_local(*sr) ) {
		
		if( shared ) {
			//capturing done in the true base
			get_local_base().capture_var_add_to_context(sr);
		}
		return sr;
	}
	
	//capturing is always done in the base
 	if( base_type == bt_local ) {
 		return base->capture_var_impl( sr, shared );
 	}

	//if already captured
	auto sym_it = symbols.find( sr->symbol );
	if( sym_it != symbols.end() ) {
		STATE_CHECK( sym_it->second.what == scope_symbol::w_capture_var );
		for( auto & decl : sym_it->second.vars ) {
			if( decl->ultimate_capture.get() == sr.get() ) {
				return decl;
			}
		}
	}
	
	auto local = declaration::create(*sr);
	local->form = declaration::f_capture;
	local->var_scope = shared_from_this();
	PRE_CHECK( base_type == bt_parent || base_type == bt_clone );
	local->proximal_capture = base->capture_var_impl( sr, shared | (base_type == bt_parent) );
	local->ultimate_capture = sr;
	
	auto & syms = symbols[sr->symbol];
	syms.what = scope_symbol::w_capture_var;
	syms.vars.push_back( local );

	return local;
}

bool scope::can_capture_var( shared_ptr<declaration> vs ) const {
	if( vs->is_lifetime_global() || vs->get_root_depth() == 0 ) {
		return true;
	}
	
	if( vs->var_scope.get() == this ) {
		return true;
	}
	
	switch( base_type ) {
		case bt_none:
		case bt_free:
			return false;
			
		case bt_clone:
		case bt_parent:
		case bt_local:
		case bt_namespace:
			return base->can_capture_var( vs );
	}

	STATE_FAIL( "unreachable" );
}

var_vector_t scope::get_captures() const {
	var_vector_t ret;
	for( auto & sym : symbols ) {
		if( sym.second.what == scope_symbol::w_capture_var ) {
			ret.insert( ret.end(), sym.second.vars.begin(), sym.second.vars.end() );
		}
	}
	
	return ret;
}

bool scope::requires_shared_parent() const {
	if( base_type == bt_clone ) {
		return false;
	}
	for( auto & sym : symbols ) {
		if( sym.second.what == scope_symbol::w_capture_var ) {
			return true;
		}
	}
	return false;
}

shared_ptr<declaration> scope::import( shared_ptr<declaration> sr )
{
	if( sr->ultimate_capture ) {
		sr = sr->ultimate_capture;
	}
	return capture_var( sr );
}

var_vector_t scope::get_local_vars( std::string const & name ) const {
	var_vector_t vars;
	
	auto it = symbols.find(name);
	if( it != symbols.end() ) {
		for( auto & var : it->second.vars ) {
			//captures aren't local (adding a capture shouldn't change the result of this function)
			if( var->form != declaration::f_capture ) {
				vars.push_back( var );
			}
		}
	}
	return vars;
}

var_vector_t scope::get_local_vars( decl_match_func criteria ) const {
	var_vector_t ret;
	
	for( auto & sym : symbols ) {
		for( auto & var : sym.second.vars ) {
			if( criteria(*var) ) {
				ret.push_back( var );
			}
		}
	}
	return ret;
}
	
var_vector_t scope::get_local_decls( bool local_life_only ) const {
	var_vector_t ret;
	
	//local_life_only = false;
	for( auto & sym : symbols ) {	
		for( auto & var : sym.second.vars ) {
			if( var->is_first_class()
				&& (!local_life_only || var->is_lifetime_local()) ) {
				ret.push_back( var );
			}
		}
	}
	return ret;
}

bool scope::has_local( declaration const * sr ) const {
	//TODO: should this also check lifetime?
	for( auto & sym : symbols ) {	
		for( auto & var : sym.second.vars ) {
			if( var.get() == sr ) {
				return true;
			}
		}
	}
	return false;
}

boost::optional<scope_symbol> scope::if_get_local_symbol( std::string const & name ) const {
	auto it = symbols.find(name);
	if( it != symbols.end() && !it->second.is_multi_var()) {
		return it->second;
	}
	
	return boost::optional<scope_symbol>();
}

/**
	Collect all the multi_var across the scope levels, or return the single non-multi_var symbol.
*/
boost::optional<scope_symbol> scope::if_get_symbol( std::string const & name ) const {
	auto it = symbols.find(name);
	if( it != symbols.end() && !it->second.is_multi_var()) {
		return it->second;
	}
	
	if( !base_type ) {
		if( it == symbols.end() ) {
			return boost::optional<scope_symbol>();
		}
		return it->second;
	}
	
	auto const & p = base->if_get_symbol( name );
	if( it == symbols.end() ) {
		return p;
	}
	if( !p ) {
		return it->second;
	}
	
	if( !p->is_multi_var() ) {
		return it->second;
	}
	
	scope_symbol build = it->second;
	for( auto & v : p->vars ) {
		build.vars.push_back(v);
	}
	return build;
}

unsigned scope::get_level() const {
	if( base_type == bt_namespace )
		return base->get_level();
	else if( base_type )
		return base->get_level() + 1;
	return 0;
}

void scope::add_fun_type( shared_ptr<intr_type const> type ) {
	PRE_CHECK( type->has_distinct_name() );
	add_fun_type( type->get_distinct_name(), type );
	PRE_CHECK( type->is_whole(), error_symbol( type->get_distinct_name() ) );
}
	
void scope::add_fun_type( std::string const & name, shared_ptr<intr_type const> type ) {
	auto it = symbols.find( name );
	PRE_CHECK( it == symbols.end(), "duplicate-symbol", error_symbol( name ) );
	
	auto & sym = symbols[name];
	sym.what = scope_symbol::w_fun_type;
	sym.fun_type = type;
	type->name_scope = shared_from_this();
}

void scope::add_type_alias( std::string const & name, shared_ptr<type_spec const> spec ) {
	auto it = symbols.find(name);
	PRE_CHECK( it == symbols.end(), "duplicate-symbol", error_symbol(name) );
	
	auto & sym = symbols[name];
	sym.what = scope_symbol::w_type_alias;
	sym.type_alias = spec;
}

cerr_impl<scope_symbol> scope::resolve_symbol( std::string const & name ) const {
	auto dot = name.find_first_of( '.' );
	if( dot != std::string::npos ) {
		auto ns_name = name.substr(0,dot);
		auto nss = if_get_symbol( ns_name ); 
		if( !nss || !nss->is_var() ) {
			return cerr::invalid_symbol_dot;
		}
		auto ns = nss->vars;

		if( ns.size() != 1 || !ns[0] || !ns[0]->type.is_extr() ) {	
			return cerr::invalid_symbol_dot;
		}
		
		auto mtype = ns[0]->type.extr().if_cast<intr_type_module>();
		if( !mtype ) {
			return cerr::invalid_symbol_dot;
		}
		
		return mtype->context->resolve_symbol( name.substr( dot+1 ) );
	} 
	
	auto q = if_get_symbol( name );
	if( !!q ) {
		return *q;
	}
	return cerr::unknown_identifier;
}

scope const * scope::if_get_function_root() const {
	if( function_root )
		return this;
	return base->if_get_function_root();
}

scope const & scope::get_function_root() const {
	if( function_root )
		return *this;
		
	PRE_CHECK( base_type );
	return base->get_function_root();
}

std::string scope::qualify_name( std::string const & local, std::string const & join ) const {
	auto use = local;
	
	//function_root's are anonymous namespaces. These must exist to prevent global vars in
	//the function from conflicting with other names. The acutal name need only be unique.
	if( function_root ) {
		//refer to block/lifetime/name_conflict_multi.lfb for failing case
		//TODO: move qualify_name to the declaration, it must be able to produce unique names
		//for all declarations
		/*if( function_root->bound_decl ) {
			//friendlier name, not strictly necessary
			return function_root->bound_decl->var_scope->qualify_name( 
				function_root->bound_decl->symbol + join + local, join );
		} else */ {
			char buf[128];
			sprintf( buf, "f%ld", intptr_t(function_root.get()) );
			return buf + join + use;
		}
	}

	if( class_root ) {
		STATE_CHECK( context.is_defined() );
		return context.extr().intr->get_qualified_id() + join + use;
	}
	
	if( base_type == bt_namespace ) {
		use = base->qualify_name( scope_name + join + use, join );
	} else if( base ) {
		use = base->qualify_name( use, join );
	}
	return use;
}

type_ref scope::get_context_type( bool force ) const {
	if( is_class_root() ) {
		return context;
	}
	
	if( base_type == bt_local ) {
		return base->get_context_type(force);
	}
	
	if( context_type.is_defined() || !force ) {
		return context_type;
	}
	
	auto crt = make_shared<intr_type_closure>();
	crt->style = intr_type::s_distinct;
	crt->context = shared_from_this();
	crt->what = intr_type_closure::w_share_context;

	if( requires_shared_parent() ) {
		STATE_CHECK(base);
		crt->parent = base->get_context_type(true);
	}
	
	auto et = extr_type(crt);
	et.set_ref_shared();
	context_type = et;
	return context_type;
}

void scope::read_minimal( serial_in & in ) {
	in.read( class_root );
	in.read( base_type );
	in.read( scope_name );
	
	base = in.read_scope_ref();
	context.read(in);
}

void scope::write_minimal( serial_out & out ) const {
	//NOTE: scope_style is handled in serial_out/in as it is required by the ctor
	out.write( class_root );
	out.write( base_type );
	out.write( scope_name );
	
	out.write_ref( base );
	context.write(out);
}

void scope::validate(util::validation_context & vctx) const {
	if( vctx.cover(this ) ) {
		return;
	}
	for( auto & s : symbols ) {
		if( s.second.fun_type ) {
			s.second.fun_type->validate(vctx);
		}
		for( auto & v : s.second.vars ) {
			v->validate(vctx);
		}
	}
}

std::string scope::desc_structure() const {
	std::string desc;
	switch( base_type ) {
		case bt_none: break;
		case bt_parent: desc = std::string("parent ") + base->desc_structure(); break;
		case bt_clone: desc = std::string("clone ") + base->desc_structure(); break;
		case bt_free: desc = std::string("free ") + base->desc_structure(); break;
		case bt_local: desc = std::string("local ") + base->desc_structure(); break;
		case bt_namespace: desc = std::string("ns:") + scope_name + " " + base->desc_structure(); break;
	}
	
	if( class_root ) {
		desc = "class " + desc;
	}
	if( function_root ) {
		desc = "function " + desc;
	}
	
	switch(scope_style) {
		case scope_style_t::declarative: desc = "declarative " + desc; break;
		case scope_style_t::imperative: desc = "imperative " + desc; break;
		case scope_style_t::dynamic: desc = "dynamic " + desc; break;
	}
	
	return ": " + desc;
}

shared_ptr<scope> scope::find_context( shared_ptr<intr_type const> context_type, int & depth ) {
	depth = 0;
	auto sc = shared_from_this();
	while( true ) {
		auto sc_ctx = sc->get_context_type(false);
		if( sc_ctx.is_extr() && sc_ctx.extr().intr == context_type ) {
			return sc;
		}
		if( !sc->has_base()) {
			break;
		}
		if( sc->base_type == bt_parent || sc->base_type == bt_clone ) {
			depth++;
		}
		sc = sc->get_base();
	}
	return shared_ptr<scope>();
}

shared_ptr<scope> scope::find_class_root( int & depth ) {
	depth = 0;
	auto sc = shared_from_this();
	while( !sc->is_class_root() ) {
		if( !sc->has_base() ) {
			return shared_ptr<scope>();
		}
		if( sc->base_type == bt_parent || sc->base_type == bt_clone ) {
			depth++;
		}
		sc = sc->get_base();
	}
	return sc;
}

shared_ptr<scope> scope::resolve_depth( int depth ) {
	auto sc = shared_from_this();
	DEBUG( "Resolve", depth, (int)sc->base_type );
	while( depth > 0 || !(sc->base_type == bt_parent || sc->base_type == bt_clone) ) {
		if( sc->base_type == bt_none ) {
			return shared_ptr<scope>();
		}
		if( sc->base_type == bt_parent || sc->base_type == bt_clone) {
			depth--;
		}
		DEBUG( "__up");
		sc = sc->base;
	}
	return sc;
}

} //eon leaf