#include "lang/serial.hpp"
#include "lang/language.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace leaf {

///////////////////////////////////////////////////////////
// serial_out
serial_out::serial_out( std::ostream & out, serial_resolver & resolver ) :
	out(out),
	resolver(resolver) {
}

void serial_out::add_type_ref( type_ref t ) {
	write( serial_item::t_type_ref );
	t.write(*this);
}

void serial_out::add_type_spec( type_spec const & ts ) {
	write( serial_item::t_type_spec );
	ts.write(*this);
}

void serial_out::write_pending( shared_ptr<declaration const> vd ) {
	auto id = queue_declaration.id_of(vd);
	write( serial_item::t_declaration );
	write( id );
	write( vd->form );
	write( vd->import_name );
	write( vd->import_type );
	write( vd->import_module );
	write( vd->name_form );
	write( vd->symbol );
	write( vd->lifetime );
	write( vd->is_export );
	
	bool st = !!vd->spec_type;
	write( st );
	if( st ) {
		vd->spec_type->write( *this );
	}
	
	bool literal = bool(vd->literal_value) && (!is_exporting() || vd->literal_value->export_literal());
	if( !literal && vd->type.is_extr() && vd->type.extr().is_literal() ) {
		//must strip off literal from type
		auto cp = vd->type.extr();
		cp.literal = false;
		type_ref(cp).write( *this );
	} else {
		vd->type.write( *this );
	}
	
	write( literal );
	if( literal ) {
		write_ref( vd->literal_value );
	}
	
	write_ref( vd->var_scope );
}

void serial_out::write_pending( shared_ptr<intr_type const> t ) {
	auto id = queue_intr_type.id_of(t);
	write( serial_item::t_intr_type );
	write( id );
	write( t->get_fun_class() );
	write_ref( t->name_scope );
	t->write(*this);
}

void serial_out::write_pending( shared_ptr<expression const> t ) {
	auto id = queue_expression.id_of(t);
	DEBUG( "WriteExpr", id );
	write( serial_item::t_expression );
	write( id );
	write( t->form );
	t->write( *this );
}

void serial_out::write_pending( shared_ptr<scope const> t ) {
	auto id = queue_scope.id_of(t);
	write( serial_item::t_scope );
	write( id );
	write( t->get_scope_style() );
	t->write_minimal( *this );
}

void serial_out::write_pending( shared_ptr<statement const> t ) {
	auto id = queue_statement.id_of(t);
	write( serial_item::t_statement );
	write( id );
	//write( t->statement_form() ); //done in ref
	t->write( *this );
}

template<typename T>
void serial_out::process_pending( T & q ) {
	typename T::pending_t sw;
	sw.swap( q.pending );
	for( auto tptr : sw ) {
		write_pending( tptr );
	}
}

void serial_out::done() {
	while( queue_intr_type.pending.size() || 
		queue_declaration.pending.size() ||
		queue_expression.pending.size() ||
		queue_scope.pending.size() ||
		queue_statement.pending.size() ) {
		process_pending( queue_intr_type );
		process_pending( queue_declaration );
		process_pending( queue_expression );
		process_pending( queue_scope );
		process_pending( queue_statement );
	}
	write( serial_item::t_end_marker );
}

void serial_out::write_ref( shared_ptr<intr_type const> t )
{
	if( t ) {
		auto id = queue_intr_type.id_of( t );
		write( id );
		write( t->get_fun_class() );
		
		auto en = resolver.get_ref_name( *t );
		bool has_ref_name = !!en;
		write( has_ref_name );
		if( en ) {
			write( *en );
		} else {
			queue_intr_type.add( t );
		}
	} else {
		intptr_t id = 0;
		write( id );
	}
}

void serial_out::write_ref( shared_ptr<declaration const> vd ) {
	if( vd ) {
		auto id = queue_declaration.id_of(vd);
		write(id);
		queue_declaration.add(vd);
	} else {
		intptr_t id = 0;
		write(id);
	}
}

void serial_out::write_ref( shared_ptr<expression const> expr ) {
	if( expr ) {
		auto id = queue_expression.id_of(expr);
		write( id );
		write( expr->form );
		queue_expression.add( expr );
	} else {
		intptr_t id = 0;
		write(id);
	}
}

void serial_out::write_ref( shared_ptr<scope const> sc ) {
	if( sc ) {
		auto id = queue_scope.id_of( sc );
		write( id );
		write( sc->get_scope_style() );
		queue_scope.add( sc );
	} else {
		intptr_t id = 0;
		write( id );
	}
}

void serial_out::write_ref( shared_ptr<statement const> st ) {
	if( st ) {
		auto id = queue_statement.id_of(st);
		write( id );
		auto form = st->statement_form();
		write( form );
		queue_statement.add( st );
	} else {
		intptr_t id = 0;
		write( id );
	}
}

void serial_out::write( std::string const & t ) {	
	write( t.size() );
	out.write( t.data(), t.size() );
}


///////////////////////////////////////////////////////////
// serian_in
serial_in::serial_in( std::istream & in, serial_resolver & resolver ) :
	in( in ),
	resolver( resolver ) {
	load_all_items();
}
	
type_ref serial_in::read_primary_type_ref() {
	STATE_CHECK( loaded_items.size() > 0 );
	STATE_CHECK( loaded_items[0].type == serial_item::t_type_ref );
	return *loaded_items[0].v_type_ref;
}

type_spec serial_in::read_primary_type_spec() {
	STATE_CHECK( loaded_items.size() > 0 );
	STATE_CHECK( loaded_items[0].type == serial_item::t_type_spec );
	return *loaded_items[0].v_type_spec;
}

void serial_in::load_all_items()
{	
	bool done = false;
	while( !done ) {
		serial_item sitem;
		read( sitem.type );
		
		if( sitem.type == serial_item::t_key_marker ) {
			read( sitem.key_ref );
			read( sitem.type );
				
			switch( sitem.type ) {
				case serial_item::t_intr_type:
					sitem.v_intr_type = read_intr_type_ref();
					break;
				case serial_item::t_declaration:
					sitem.v_declaration = read_declaration_ref();
					break;
				case serial_item::t_expression:
					sitem.v_expression = read_expression_ref();
					break;
				case serial_item::t_statement:
					sitem.v_statement = read_statement_ref();
					break;
				case serial_item::t_scope:
					sitem.v_scope = read_scope_ref();
					break;
				default:
					STATE_FAIL( "corrupt-serial" );
			}
			
		} else {

			switch( sitem.type ) {
				case serial_item::t_end_marker:
					done = true;
					break;
					
				case serial_item::t_type_ref:
					sitem.v_type_ref = type_ref( *this );
					break;
					
				case serial_item::t_extr_type:
					sitem.v_extr_type = extr_type( *this );
					break;
					
				case serial_item::t_intr_type:
					sitem.v_intr_type = read_intr_type();
					break;
					
				case serial_item::t_declaration:
					sitem.v_declaration = read_declaration();
					break;
					
				case serial_item::t_expression:
					sitem.v_expression = read_expression();
					break;
					
				case serial_item::t_statement:
					sitem.v_statement = read_statement();
					break;
					
				case serial_item::t_scope:
					sitem.v_scope = read_scope();
					break;
					
				case serial_item::t_type_spec: {
					type_spec ts;
					ts.read(*this);
					sitem.v_type_spec = ts;
					break;
				}

				default:
					STATE_FAIL( "corrupt-serial" );
			}
		}
		
		if( !done ) {
			loaded_items.push_back( sitem );
			if( sitem.key_ref != serial_item::t_none ) {
				key_items.push_back( sitem );
			}
		}
	}
	
	complete();
}

shared_ptr<scope> serial_in::get_key_root_scope() {
	for( auto & v : key_items ) {
		if( v.key_ref == serial_item::t_root &&
			v.type == serial_item::t_scope ) {
			return v.v_scope;
		}
	}
	return shared_ptr<scope>();
}

void serial_in::complete() {
	STATE_CHECK( queue_intr_type.pending.empty() );
	STATE_CHECK( queue_declaration.pending.empty() );
	STATE_CHECK( queue_expression.pending.empty() );
	STATE_CHECK( queue_scope.pending.empty() );
	
	for( auto & pl : post_loaders ) {
		(*pl)();
	}
	post_loaders.clear();
}

shared_ptr<intr_type> serial_in::read_intr_type() {
	intptr_t id;
	read(id);
	intr_type::fun_class_t cl;
	read(cl);

	auto sc = read_scope_ref();
		
	auto res = queue_intr_type.pop_pending(id);
	if( !res ) {
		res = create_intr_type(cl);
	}
	queue_intr_type.known[id] = res;
	
	STATE_CHECK( res->get_fun_class() == cl );
	res->read(*this);
	
	if( sc ) {
		//must be done after fully loaded
		struct post_load : post_loader {
			shared_ptr<scope> sc;
			shared_ptr<intr_type> t;
			void operator()() {
				sc->add_fun_type( t );
			}
		};
		auto pl = make_shared<post_load>();
		pl->sc = sc;
		pl->t = res;
		post_loaders.push_back( pl );
	}
	
	return res;
}

shared_ptr<expression> serial_in::read_expression() {
	intptr_t id;
	read(id);
	DEBUG( "ReadExpr", id );
	expression::form_t form;
	read(form);
	
	auto expr = queue_expression.pop_pending(id);
	STATE_CHECK( !!expr );
	
	STATE_CHECK( expr->form == form );
	expr->read(*this);
	return expr;
}

shared_ptr<statement> serial_in::read_statement() {
	intptr_t id;
	read( id );
	
	auto st = queue_statement.pop_pending(id);
	STATE_CHECK( !!st );
	
	st->read( *this );
	return st;
}

shared_ptr<intr_type> serial_in::read_intr_type_ref()
{
	intptr_t id;
	read(id);
	if( !id )
		return shared_ptr<intr_type>();
		
	intr_type::fun_class_t cl;
	read(cl);
	
	bool has_ref_name;
	read(has_ref_name);
	if( has_ref_name ) {
		std::string ref_name;
		read(ref_name);
		//TODO: we should return const
		return std::const_pointer_cast<intr_type>(resolver.get_intr_type( ref_name ));
	}
	
	auto known = queue_intr_type.find(id);
	if( known ) {
		return known;
	}
	
	auto load = create_intr_type(cl);
	queue_intr_type.add_pending( id, load );
	return load;
}

shared_ptr<declaration> serial_in::read_declaration_ref() {
	intptr_t id;
	read(id);

	if( id == 0 ) {
		return shared_ptr<declaration>();
	}
	
	auto known = queue_declaration.find(id);
	if( known ) {
		return known;
	}
		
	auto load = declaration::create();
	queue_declaration.add_pending( id, load );
	return load;
}

shared_ptr<expression> serial_in::read_expression_ref() {
	intptr_t id;
	read(id);
	if( id == 0 ) {
		return shared_ptr<expression>();
	}
	
	expression::form_t form;
	read(form);
	
	auto known = queue_expression.find(id);
	if( known ) {
		return known;
	}
	
	auto s = create_expression(form);
	queue_expression.add_pending(id, s );
	return s;
}

shared_ptr<statement> serial_in::read_statement_ref() {
	intptr_t id;
	read(id);
	if( id == 0 ) {
		return shared_ptr<statement>();
	}
	
	statement_form_t form;
	read(form);
	
	auto known = queue_statement.find(id);
	if( known ) {
		return known;
	}
	
	auto s = create_statement(form);
	queue_statement.add_pending( id, s );
	return s;
}

shared_ptr<scope> serial_in::read_scope_ref() {
	intptr_t id;
	read(id);
	if( id == 0 ) {
		return shared_ptr<scope>();
	}
	
	scope_style_t style;
	read(style);
	
	auto known = queue_scope.find(id);
	if( known ) {
		return known;
	}
	
	auto sc = scope::create( style );
	queue_scope.add_pending( id, sc );
	return sc;
}

shared_ptr<expression> serial_in::create_expression(expression::form_t form) {
	shared_ptr<expression> load;
	switch( form ) {
		#define EXPRESSION_FORM(form) \
			case expression::f_##form: load = make_shared<expr_##form>(); break;
		EXPRESSION_FORMS
		#undef EXPRESSION_FORM
	}
	return load;
}

shared_ptr<statement> serial_in::create_statement( statement_form_t form ) {
	shared_ptr<statement> s;
	switch(form) {
		#define STATEMENT_FORM(form) \
			case sf_##form##_statement: \
				s = make_shared<form##_statement>(); \
				break;
		STATEMENT_FORMS
		#undef STATEMENT_FORM
	}
	return s;
}


shared_ptr<intr_type> serial_in::create_intr_type(intr_type::fun_class_t cl) {
	shared_ptr<intr_type> load;
	switch( cl ) {
		#define FUN_CLASS_TYPE(s) case intr_type::f_##s: \
			load = make_shared<intr_type_##s>(); \
			break;
		INTR_TYPE_FUN_CLASSES
		#undef FUN_CLASS_TYPE
	}
	return load;
}

void serial_in::read( std::string & t ) {
	size_t len;
	read( len );
	t.resize( len );
	in.read( (char*)t.data(), len );
}

shared_ptr<declaration> serial_in::read_declaration() {
	intptr_t id;
	read(id);
	
	//TODO: add_declaration requires this, but it'd be nicer not to need a speical form for declaration
	auto vd = queue_declaration.pop_pending(id);
	if ( !vd ) {
		vd = declaration::create();
		queue_declaration.known[id] = vd;
	}
	
	read( vd->form );
	read( vd->import_name );
	read( vd->import_type );
	read( vd->import_module );
	read( vd->name_form );
	read( vd->symbol );
	read( vd->lifetime );
	read( vd->is_export );
	
	bool st;
	read( st );
	if( st ) {
		vd->spec_type = make_shared<type_spec>( *this );
	}
	
	vd->type.read(*this);
	
	bool literal;
	read( literal );
	if( literal ) {
		vd->literal_value = read_expression_ref();
	}
	
	auto sc = read_scope_ref();
	if( sc ) { //may not be set if inside parametric
		sc->add( vd );
	}
	
	return vd;
}

shared_ptr<scope> serial_in::read_scope() {
	intptr_t id;
	read(id);
	
	scope_style_t style;
	read(style);
	
	//STATE_CHECK( queue_scope.known.find(id) == queue_scope.known.end() );
	auto sc = queue_scope.pop_pending(id);
	STATE_CHECK( !!sc );
	//	if( !sc ) {
	//	sc = scope::create( style );
	//}
	//queue_scope.known[id] = sc;

	sc->read_minimal( *this );
	return sc;
}

} //eon leaf
