#include "lang/expression.hpp"
#include "lang/serial.hpp"
#include "lang/statement_block.hpp"

namespace leaf {

void expression::write_impl( serial_out & so ) const {
	type.write( so );
	so.write_ref( bound_decl );
}

void expression::read_impl( serial_in & si ) {
	type.read( si );
	bound_decl = si.read_declaration_ref();
}

/////////////////////////////////////////////////
// expr_literal_number
void expr_literal_number::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write( src_form );
	so.write( val_boolean );
	value.write( so.get_stream() );
}

void expr_literal_number::read_impl( serial_in & si ) {
	super::read_impl(si);
	si.read( src_form );
	si.read( val_boolean );
	value.read( si.get_stream() );
}

/////////////////////////////////////////////////
// expr_text
void expr_text::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write(what);
	unicode::write( text, so.get_stream() );
}

void expr_text::read_impl( serial_in & si ) {
	super::read_impl(si);
	si.read(what);
	unicode::read( text, si.get_stream() );
}

/////////////////////////////////////////////////
// expr_with_sub
void expr_with_sub::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write( sub.size() );
	for( auto & s : sub ) {
		so.write( s.name );
		so.write_ref( s.expr );
	}
}

void expr_with_sub::read_impl( serial_in & si ) {
	super::read_impl(si);
	size_t sz;
	si.read(sz);
	sub.resize(sz);
	for( size_t i=0; i < sz; ++i ) {
		si.read( sub[i].name );
		sub[i].expr = si.read_expression_ref();
	}
}

//////////////////////////////////////////////////
// expr_with_block
void expr_with_block::write_impl( serial_out & so ) const {
	super::write_impl(so);
	block->write(so);
}

void expr_with_block::read_impl( serial_in & si ) {
	super::read_impl(si);
	block = make_shared<statement_block>(si);
}


//////////////////////////////////////////////////
// expr_funccall
void expr_funccall::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write_ref(func);
	so.write_ref(args);
	so.write_ref(context);
}

void expr_funccall::read_impl( serial_in & si ) {
	super::read_impl(si);
	func = si.read_expression_ref();
	args = si.read_expression_ref();
	context = si.read_expression_ref();
}

//////////////////////////////////////////////////
// expr_funcdefn

void expr_funcdefn::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write(clone);
	bool e=  !!func_type;
	so.write(e);
	if(e) {
		func_type->write(so);
	}
}

void expr_funcdefn::read_impl( serial_in & si ) {
	super::read_impl(si);
	si.read(clone);
	
	bool e;
	si.read(e);
	if( e ) {
		func_type = make_shared<type_spec>(si);
	}
}

bool expr_funcdefn::export_literal() const {
	// parametrics need to be exported to be compiled in including code
	return type.contains_parametric();
}

//////////////////////////////////////////////////
// expr_funcdefn
void expr_symbol::write_impl( serial_out & so )  const {
	super::write_impl(so);
	so.write(symbol);
	so.write_ref(sref);
}

void expr_symbol::read_impl( serial_in & si ) {
	super::read_impl(si);
	si.read(symbol);
	sref = si.read_declaration_ref();
}

//////////////////////////////////////////////////
// expr_field_like
void expr_field_like::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write_ref( base );
	so.write( optional );
}

void expr_field_like::read_impl( serial_in & si ) {
	super::read_impl(si);
	base = si.read_expression_ref();
	si.read( optional );
}

//////////////////////////////////////////////////
// expr_field
void expr_field::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write(field);
	so.write(field_index);
	so.write_ref(sref); //TODO: is this part of the serialization?
}

void expr_field::read_impl( serial_in & si ) {
	super::read_impl(si);
	si.read(field);
	si.read(field_index);
	sref = si.read_declaration_ref();
}

/////////////////////////////////////////////////
// expr_subscript
void expr_subscript::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write_ref(offset);
}

void expr_subscript::read_impl( serial_in & si ) {
	super::read_impl(si);
	offset = si.read_expression_ref();
}

/////////////////////////////////////////////////
// expr_type
void expr_type::write_impl( serial_out & so ) const {
	super::write_impl(so);
	inner_type_spec->write(so);
}

void expr_type::read_impl( serial_in & si ) {
	super::read_impl(si);
	inner_type_spec = make_shared<type_spec>(si);
}

/////////////////////////////////////////////////
// expr_tag
void expr_tag::write_impl( serial_out & so ) const {
	super::write_impl(so);
	so.write_ref(value);
}

void expr_tag::read_impl( serial_in & si ) {
	super::read_impl(si);
	value = si.read_expression_ref();
}

} //eon leaf