#include "lang/serial.hpp"
#include "lang/type.hpp"
#include "lang/declaration.hpp"
#include "lang/scope.hpp"
#include "lang/statement_block.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {

///////////////////////////////////////////////////////////
// intr_type
namespace {
	void xwrite( serial_out & so, intr_type::param_spec const & o ) {
		so.write( o.name );
	}
	
	void xread( serial_in & si, intr_type::param_spec & i ) {
		si.read( i.name );
	}
}

void intr_type::write( serial_out & so ) const {
	PRE_CHECK( !is_whole() || is_export, "needed-type-not-exported", 
		error::tag_error_symbol( get_qualified_id() ) );
	
	//the header info, class/pointer is handled by serial_out
	so.write(distinct_name);
	so.write(local_complete);
	so.write(style);
	
	bool has_inst_of = !!instance_of;
	so.write(has_inst_of);
	if( has_inst_of ) {
		so.write_ref( instance_of );
	}
	
	size_t plen = params.size();
	so.write( plen );
	for( auto & ps : params ) {
		xwrite( so, ps );
	}
	
	size_t aplen = applied_params.size();
	so.write( aplen );
	for( auto & aps : applied_params ) {
		aps.write( so );
	}
}

void intr_type::read( serial_in & si ) {
	si.read(distinct_name);
	si.read(local_complete);
	si.read(style);
	
	bool has_inst_of;
	si.read( has_inst_of );
	if( has_inst_of ) {
		instance_of = si.read_intr_type_ref();
	}
	
	size_t plen;
	si.read(plen);
	for( size_t i=0; i < plen; ++i ) {
		param_spec ps;
		xread( si, ps );
		params.push_back(ps);
	}

	size_t aplen;
	si.read( aplen );
	for( size_t i=0; i < aplen; ++i ) {
		extr_type et( si );
		applied_params.push_back(et);
	}
}

///////////////////////////////////////////////////////////
// intr_type_fixed
void intr_type_fixed_number::write( serial_out & so ) const {
	super::write(so);
	so.write(data_bitsize);
}

void intr_type_fixed_number::read( serial_in & si ) {
	super::read(si);
	si.read(data_bitsize);
}

///////////////////////////////////////////////////////////
// extr_type
void extr_type::write( serial_out & so ) const {
	so.write(lvalue);
	so.write(optional);
	so.write(lossy);
	so.write(cvt_optional);
	so.write(reference);
	so.write(literal);
	so.write_ref( intr );
}

void extr_type::read( serial_in & si ) {
	si.read(lvalue);
	si.read(optional);
	si.read(lossy);
	si.read(cvt_optional);
	si.read(reference);
	si.read(literal);
	intr = si.read_intr_type_ref();
}

///////////////////////////////////////////////////////////
// intr_type_funtion
void intr_type_function::write( serial_out & so ) const {
	super::write(so);
	
	so.write_ref( args );
	so.write_ref( returns );
	context.write(so);
	so.write(access);
	so.write(can_fail);
}

void intr_type_function::read( serial_in & si ) {
	super::read(si);
	
	args = si.read_intr_type_ref_known<intr_type_value_list>();
	returns = si.read_intr_type_ref_known<intr_type_value_list>();
	context.read(si);
	si.read(access);
	si.read(can_fail);
}

///////////////////////////////////////////////////////////
// intr_type_compount
void intr_type_compound::write( serial_out & so ) const {
	super::write(so);
	
	so.write( sub.size() );
	for( auto & sr : sub ) {
		so.write( sr.name );
		sr.type.write( so );
	}
}

void intr_type_compound::read( serial_in & si ) {
	super::read(si);
	
	size_t len;
	si.read( len );
	sub.resize( len );
	for( size_t i=0; i < len; ++i ) {
		si.read( sub[i].name );
		sub[i].type.read(si);
	}
}

///////////////////////////////////////////////////////////
// intr_type_tuple
void intr_type_tuple::write( serial_out & so ) const {
	super::write(so);
	
	so.write(pack);
	so.write(selectable);
}

void intr_type_tuple::read( serial_in & si ) {
	super::read(si);
	
	si.read(pack);
	si.read(selectable);
}

///////////////////////////////////////////////////////////
// intr_type_wrap
void intr_type_wrap::write( serial_out & so ) const {
	super::write(so);
	sub.write(so);
}

void intr_type_wrap::read( serial_in & si ) {
	super::read(si);
	sub.read(si);
}

///////////////////////////////////////////////////////////
// intr_type_array
void intr_type_array::write( serial_out & so ) const {
	super::write(so);
	so.write(raw);
}

void intr_type_array::read( serial_in & si ) {
	super::read(si);
	si.read(raw);
}


///////////////////////////////////////////////////////////
// intr_type_type
void intr_type_type::write( serial_out & so ) const {
	super::write(so);
}

void intr_type_type::read( serial_in & si ) {
	super::read(si);
}


///////////////////////////////////////////////////////////
// intr_type_rational
void intr_type_rational::write( serial_out & so ) const {
	super::write(so);
	so.write(integer_bitsize);
	so.write(integer_signed);
}

void intr_type_rational::read( serial_in & si ) {
	super::read(si);
	si.read(integer_bitsize);
	si.read(integer_signed);
}

///////////////////////////////////////////////////////////
// intr_type_scope
void intr_type_scope::write( serial_out & so ) const {
	super::write(so);
	
}

void intr_type_scope::read( serial_in & si ) {
	super::read(si);
	
}

///////////////////////////////////////////////////////////
// intr_type_custom
void intr_type_custom::write( serial_out & so ) const {
	super::write(so);

	auto sc = std::dynamic_pointer_cast<scope const>( context );
	STATE_CHECK( sc );
	so.write_ref( sc );
	
	auto srs = context->get_local_decls();
	so.write( srs.size() );
	for( auto & sr : srs ) {
		so.write_ref( sr );
	}
	
	so.write_ref( destroy );
}

void intr_type_custom::read( serial_in & si ) {
	/**
		Adding to the scope must be done after the declaration has been fully loaded.
	*/
	struct post_loader : serial_in::post_loader {
		shared_ptr<scope> sc;
		var_vector_t decls;
		
		void operator()() {
			for( auto & decl : decls ) {
				sc->add( decl );
			}
		}
	};
	super::read(si);

	auto pl = make_shared<post_loader>();
	si.add_post_loader( pl );
	
	pl->sc = si.read_scope_ref();
	context = pl->sc;
	
	size_t len;
	si.read(len);
	for( size_t i=0; i < len; ++i ) {
		auto sc = si.read_declaration_ref();
		//pl->decls.push_back( si.read_declaration_ref() );
	}
	//si.read( mctx->name_prefix );
	
	destroy = si.read_declaration_ref();
}

///////////////////////////////////////////////////////////
// intr_type_classlike
void intr_type_classlike::write( serial_out & so ) const {
	super::write(so);
	block->write(so);
}

void intr_type_classlike::read( serial_in & si ) {
	super::read(si);
	block = make_shared<statement_block>(si);
}

//////////////////////////////////////////////////////////
// intr_type_compat
void intr_type_compat::write( serial_out & so ) const {
	super::write(so);
	so.write( type );
}

void intr_type_compat::read( serial_in & si ) {
	super::read(si);
	si.read( type );
}


} //eon leaf
