#include "lang/type_converter.hpp"
#include "lang/type_identifier.hpp"
#include "lang/type.hpp"
#include "lang/type_spec.hpp"
#include "lang/scope.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#if SHOW_DEBUG
#include "lang/dump.hpp"
#endif

namespace leaf
{

namespace 
{
	cvt_expression const * skip_zero_cost( cvt_expression const * c )
	{
		while( c && (
			c->form == cvt_expression::c_set_distinct ||
			c->form == cvt_expression::c_addr_of ||
			c->form == cvt_expression::c_drop_lvalue ||
			c->form == cvt_expression::c_bind_raw_func ||
			c->form == cvt_expression::c_bind_func ||
			c->form == cvt_expression::c_bind_context ||
			c->form == cvt_expression::c_collapse ||
			c->form == cvt_expression::c_field ||
			c->form == cvt_expression::c_deref_optional ||
			c->form == cvt_expression::c_direct ||
			c->form == cvt_expression::c_optional ||
			c->form == cvt_expression::c_unset_optional ||
			c->form == cvt_expression::c_drop_literal ||
			c->form == cvt_expression::c_param_inst) ) {
			c = c->then.get();
		}
		return c;
	}
	
	//must follow the order of form_t
	int compare_cvt_expression( cvt_expression const * a, cvt_expression const * b ) 
	{
		#define COST(v) (v ? v->form : cvt_expression::c_direct)
		auto next = [&]( bool consume, cvt_expression const * c ) {
			if( consume && c )
				c = c->then.get();
			return skip_zero_cost(c);
		};

		//parametric always costs more than non-parametric, but all parametric conversions cost the same
		bool param_a = COST(a) == cvt_expression::c_parametric;
		bool param_b = COST(b) == cvt_expression::c_parametric;
		if( param_a != param_b ) {
			return param_a ? 1 : -1;
		} else if( param_a && param_b ) {
			return 0;
		}
		
		a = skip_zero_cost( a );
		b = skip_zero_cost( b );

		bool deref_a = COST(a) == cvt_expression::c_deref;
		bool deref_b = COST(b) == cvt_expression::c_deref;
		if( deref_a != deref_b )
			return deref_a ? 1 : -1;
		a = next( deref_a, a );
		b = next( deref_b, b );

		//TODO: add lossy conversions (uncertain of cost, they must be explicit, but still distinguishable)
		
		bool fix_a = COST(a) == cvt_expression::c_fixate;
		bool fix_b = COST(b) == cvt_expression::c_fixate;
		if( !fix_a && !fix_b ) {
			//no cost
		} else if( !fix_a || !fix_b ) {
			//fixation costs more than not fixating
			return fix_a ? 1 : -1;
		} else if( a->to_fun != b->to_fun ) {
			return a->to_fun < b->to_fun ? -1 : 1;
		} else if( a->to_bitsize != b->to_bitsize ) {
			//natural size preferred
			int natural_size = a->to_fun == cvt_expression::p_float 
				? platform::natural_float_size : platform::natural_integer_size;
			if( a->to_bitsize == natural_size )
				return -1;
			if( b->to_bitsize == natural_size )
				return 1;
				
			//TODO:
			//otherwise the smallest (actually I've decided now there should be no preference beyond natural)
			//return a->to_bitsize < b->to_bitsize ? -1 : 1;
		} 
		a = next( fix_a, a );
		b = next( fix_b, b );
		
		bool promote_a = COST(a) == cvt_expression::c_promote;
		bool promote_b = COST(b) == cvt_expression::c_promote;
		if( !promote_a && !promote_b ) {
			//no cost
		} else if( !promote_a || !promote_b ) {
			//promotion costs more than not promotion
			return promote_a ? 1 : -1;
		} else if( a->to_fun != b->to_fun || a->from_fun != b->from_fun ) {
			bool nox_a = a->no_cross_promote();
			bool nox_b = b->no_cross_promote();
			if( nox_a && nox_b ) {
				//no cost
			} else if( nox_a != nox_b ) {
				//cross promotion costlier than not
				return nox_a ? -1 : 1;
			} else {
				return a->to_fun < b->to_fun ? -1 : 1;
			}
		} else if( a->to_bitsize != b->to_bitsize ) {
			//same promotion prefer lower bitsize
			return a->to_bitsize < b->to_bitsize ? -1 : 1;
		}
		a = next( promote_a, a );
		b = next( promote_b, b );
		
		bool pack_a = COST(a) == cvt_expression::c_pack || COST(a) == cvt_expression::c_pack_value ||
			COST(a) == cvt_expression::c_pack_contextual;
		bool pack_b = COST(b) == cvt_expression::c_pack || COST(b) == cvt_expression::c_pack_value ||
			COST(b) == cvt_expression::c_pack_contextual;
		if( pack_a || pack_b ) {
			size_t xsub = std::max( a ? a->sub.size() : 0, b ? b->sub.size() : 0 );
			int cmp = 0;
			for( size_t i=0; i < xsub; ++i ) {
				//in sub-matching the exact cost differences are not relevant, only whether each part
				//was less/more costly than the other
				int sub_cmp = compare_cvt_expression(
					(a && i < a->sub.size()) ? a->sub[i].cvt.get() : 0,
					(b && i < b->sub.size()) ? b->sub[i].cvt.get() : 0 );
				if( sub_cmp < 0 )
					cmp--;
				else if( sub_cmp > 0 )
					cmp++;
			}
			if( cmp )
				return cmp;
			//NOTE: packing type conversion doesn't cost anything
		}
		a = next( pack_a, a );
		b = next( pack_b, b );
		return 0;
	}
	
	shared_ptr<cvt_expression> make_cvt( cvt_expression::form_t form ) {
		return make_shared<cvt_expression>(form);
	}
	
	/**
		A simple wrapper to a conversion chain that allows tracking the start and end.
	*/
	struct converter {
		shared_ptr<cvt_expression> start;
		shared_ptr<cvt_expression> current;
	
		/**	
			@param all_none if non-null and a conversion happens then set to false,
				otherwise unchanged
		*/
		void extend( shared_ptr<cvt_expression> then, bool * all_none = nullptr ) {
			if( !then || then->is_none() )
				return;
				
			if( all_none )
				*all_none = false;
				
			if( !start )
				start = then;
			if( current )
				current->then = then;
			//must skip to end of chain
			current = then;
			while( current->then )
				current = current->then;
		}
		
		bool extend( type_converter::convert_t then, bool * all_none = nullptr ) {
			if( !then.okay() )
				return false;
			extend( then.conversions, all_none );
			return true;
		}
		
		#define CVT(type) \
			shared_ptr<cvt_expression> type() { \
				auto t = make_cvt( cvt_expression::c_##type ); \
				extend( t ); \
				return t; \
			}
		CVT(collapse)
		CVT(deref_optional)
		CVT(deref)
		CVT(drop_lvalue)
		CVT(optional)
		CVT(bind_raw_func)
		CVT(bind_func)
		CVT(bind_context)
		CVT(fixate)
		CVT(promote)
		CVT(loss)
		CVT(field)
		CVT(set_distinct)
		CVT(addr_of)
		CVT(drop_literal)
		CVT(pack_contextual)
		CVT(make_compat)
		CVT(param_inst)
		#undef CVT
		
		shared_ptr<cvt_expression> promote_loss( bool loss ) {
			auto t = make_cvt( loss ? cvt_expression::c_loss : cvt_expression::c_promote );
			extend(t);
			return t;
		}
	};
	
} //eon anon
int type_converter::convert_t::compare_cost( convert_t const & o ) const
{ 
	return compare_cvt_expression( conversions.get(), o.conversions.get() );
}

type_converter::type_converter( context const & ctx )
	: ctx( ctx )
{
}

type_converter::~type_converter()
{
}

type_converter::convert_t type_converter::convert_list(
	intr_type_compound const& to, intr_type_compound const & from_, 
	unsigned sub_flags, cvt_expression::form_t form )
{
	converter cvt;
	
	intr_type_compound const * from = &from_;
	//1-value-lists within lists can be collapsed. We do only the outermost item here since
	//internal collapsing is done in convert_impl
	if( form == cvt_expression::c_pack_value ) {
		while( true ) {
			if( to.sub.size() == from->sub.size() )
				break;
				
			if( from->sub.size() != 1 || !from->sub[0].type.extr().is<intr_type_value_list>() )
				return cerr::mismatch_size;
				
			cvt.collapse();
			from = &from->sub[0].type.extr().cast<intr_type_value_list>();
		}
	} else {
		if( to.sub.size() != from->sub.size() )
			return cerr::mismatch_size;
	}
		
	auto pack = make_cvt( form );
	bool all_none = true;
	//check for pack type mismatch
	if( form == cvt_expression::c_pack ) {
		auto & left = to.cast<intr_type_tuple>();
		auto & right = from->cast<intr_type_tuple>();
		pack->pack = left.pack;
		//force unpack/pack if the packing doesn't match
		if( left.pack != right.pack )
			all_none = false;
	}
	
	for( size_t i=0; i < to.sub.size(); ++i )	{
		converter scvt;
		
		if( !i && from->sub.size() == 1 && form == cvt_expression::c_pack_value ) {
			scvt.collapse();
		} else {
			auto f = scvt.field();
			f->field = i;
		}
		
		auto n = convert_impl( to.sub[i].type, from->sub[i].type, sub_flags );
		if( !scvt.extend( n, &all_none ) )
			return n;
			
		std::string use_name = to.sub[i].name;
		//tuple conversions must match names if not anonymous (do at this level since in non-tuple
		//conversion the names don't matter)
		if( form == cvt_expression::c_pack && to.sub[i].name != from->sub[i].name ) {
			if( from->sub[i].name.empty() ) {
				all_none = false;
			} else {
				return cerr::mismatch_field_name;
			}
		}
		
		pack->sub.push_back( { scvt.start, to.sub[i].name } );
	}
	if( all_none ) {
		return cvt.start;
	}
		
#if 0
	//special reduction to reduce the form `c_pack_value {cvt->c_collapse}` to simply `cvt`
	if( pack->sub.size() == 1 ) {
		shared_ptr<cvt_expression> prev;
		auto last = pack->sub[0];
		while( last->then ) {
			prev = last;
			last = last->then;
		}
			
		if( last->form == cvt_expression::c_collapse ) {
			if( prev ) {
				prev->then.reset();
				pack = pack->sub[0];
			}
		}
	}
#endif
	cvt.extend( pack );
	return cvt.start;
}

type_converter::convert_t type_converter::convert_function_call( intr_type_value_list const & to,
	intr_type_value_list const & from, bool allow_parametric )
{
	convert_t cost;

	if( from.sub.size() > to.sub.size() )
		return cerr::mismatch_size;
			
	auto pack = make_cvt( cvt_expression::c_pack_value );
	
	for( size_t i=0; i < to.sub.size(); ++i )	{
		if( allow_parametric && to.sub[i].type.contains_parametric() ) {
			//parametric types can't be converted, we just put placeholders here
			auto f = make_cvt(cvt_expression::c_parametric);
			pack->sub.push_back( { f, to.sub[i].name } );
		} else if( i >= from.sub.size() ) {
			//potential optional parameter matching
			if( !to.sub[i].type.extr().is_optional() )
				return cerr::mismatch_size;
				
			auto f = make_cvt(cvt_expression::c_unset_optional);
			f->target_type = to.sub[i].type;
			pack->sub.push_back( { f, to.sub[i].name } ) ;
		} else {
			auto f = make_cvt(cvt_expression::c_field);
			f->field = i;
			auto n = convert_impl( to.sub[i].type, from.sub[i].type, cf_allow_cvt );
			if( !n.okay() )
				return n;
			f->chain(n.conversions);
			pack->sub.push_back( { f, to.sub[i].name } );
		}
	}
	cost.conversions = pack;
	return cost;
}

/**
	TODO: there is something wrong here, possibly with the name. For selectable it doesn't
	actually return a callable type. Moreso, for contextual within selectable it doesn't even
	return a list of callable types. There are various workarounds to this now:
		- expression_typer::process_funccall
		- type_converter::match_function
*/
type_converter::convert_t type_converter::convert_any_callable( type_ref const & from_ ) {
	type_ref const * from = &from_;
	converter cvt;
	
	while( true ) {
		auto vl = from->extr().if_cast<intr_type_value_list>();
		if( vl && vl->sub.size() == 1 ) {
			cvt.collapse();
			from = &vl->sub[0].type;
			continue;
		}
		break;
	}
	
	if( from->extr().is_lvalue() ) {
		cvt.drop_lvalue();
	}
 	if( from->extr().is_optional() ) {
 		return cerr::mismatch_optional;
 	}
	if( from->extr().is_refer() ) {
		cvt.deref();
	}
	
	//TODO: should check that this is valid
	if( from->extr().is<intr_type_contextual>() ) {
		auto & ct = from->extr().cast<intr_type_contextual>();
		from = &ct.get_unbound();
		
		auto c = convert_unqualified( *from );
		if( !c.okay() )
			return c;
			
		if( c.conversions ) {
			auto pack = make_cvt( cvt_expression::c_pack_contextual );
			
			auto b = make_cvt( cvt_expression::c_field );
			b->field = 0;
			pack->sub.push_back( { b, "" } );
			
			converter u;
			auto f = u.field();
			f->field = 1;
			u.extend( c );
			pack->sub.push_back( { f, "" } );
			
			cvt.extend(pack);
		}
		
		cvt.bind_context();
	}
		
	auto tuple = from->extr().if_cast<intr_type_tuple>();
	if( !from->extr().is<intr_type_function>() && !(tuple && tuple->selectable) )	
		return cerr::requires_function;
		
	return cvt.start;
}

type_converter::convert_t type_converter::convert_any_common( type_ref const & from_, 
	cerr list_fail, unsigned flags ) {
	if( !from_.is_extr() ) {
		return cerr::invalid_type_ref;
	}
	
	return convert_any_common( from_.extr(), list_fail, flags );
}

type_converter::convert_t type_converter::convert_any_common( extr_type const & from_, 
	cerr list_fail, unsigned flags ) {
	extr_type const * from = &from_;
	converter cvt;
	
	while( true ) {
		auto fvl = from->if_cast<intr_type_value_list>();
		if( !fvl )
			break;
		if( !fvl->sub.size() )
			return list_fail;
		
		//collapse	
		if( fvl->sub.size() == 1 ) {
			cvt.collapse( );
			from = &fvl->sub[0].type.extr();
			continue;
		}
		
		//or unpack/convert/pack parts
		bool all_none = true;
		auto pack = make_cvt( cvt_expression::c_pack_value );
		
		for( size_t i=0; i < fvl->sub.size(); ++ i ) {
			auto f = make_cvt(cvt_expression::c_field);
			f->field = i;
			auto n = convert_any_common( fvl->sub[i].type, list_fail, flags );
			if( !n.okay() )
				return n;
			if( f->chain(n.conversions) ) {
				all_none = false;
			}
			pack->sub.push_back( { f, "" } );
		}
		
		if( !all_none )
			cvt.extend( pack );
		return cvt.start;
	}
	
	//unless converting to an explicit type the contextual is lost
	auto * c = from->if_cast<intr_type_contextual>();
	if( c ) {
		if( flags & acf_retain_contextual ) {
			auto pack = make_cvt( cvt_expression::c_pack_contextual );
			
			auto b = make_cvt( cvt_expression::c_field );
			b->field = 0;
			pack->sub.push_back( { b, "" } );
			
			auto f = make_cvt( cvt_expression::c_field );
			f->field = 1;
			auto unbound = convert_any_common( *from, list_fail, flags & ~acf_retain_contextual);
			if( !unbound.okay() ) {
				return unbound;
			}
			pack->sub.push_back( { unbound.conversions, "" } );
			
			cvt.extend(pack);
		} else {
			from = &c->get_unbound().extr();
			auto cf = cvt.field(); //TODO: hmmm, but do we want addiitonal converters just for this special field (consider using expr_special_field enum somehow)
			cf->field = 1;
		}
	}
	
	bool is_lvalue = from->is_lvalue();
	bool is_optional = from->is_optional();
	bool is_shared = from->is_shared();
	bool is_value_ptr = from->is_value_ptr();
	
	bool deref = (is_shared && !(flags & acf_retain_shared)) ||
		(is_value_ptr && !(flags & acf_retain_value_ptr));
 	bool deref_optional = is_optional &&
 		(flags & acf_deref_optional);
	bool drop_lvalue = is_lvalue &&
		(deref || !(flags & acf_retain_lvalue)); //deref_optional retains lvalue status
	
	if( drop_lvalue ) {
		cvt.drop_lvalue();
		is_lvalue = false;
	}
 	if( deref_optional ) {
 		cvt.deref_optional();
 		is_optional = false;
 	}
	if( deref ) {
		cvt.deref();
		is_shared = false;
		is_value_ptr = false;
	}
		
	if( flags & acf_require_value && is_optional ) {
		return cerr::requires_value;
	}
	if( flags & acf_require_lvalue && !is_lvalue ) {
		return cerr::requires_lvalue;
	}
	if( flags & acf_require_optional && !is_optional ) {
		return cerr::requires_optional;
	}
	if( flags & acf_require_shared && !is_shared ) {
		return cerr::requires_shared;
	}
	if( flags & acf_require_refer && !(is_shared || is_value_ptr) ) {
		return cerr::requires_refer;
	}
	if( !(flags & acf_require_optional) && !(flags & acf_retain_optional) && is_optional ) {
		return cerr::mismatch_optional;
	}
		
	if( from->is<intr_type_scalar>() ) {
		//nop
	} else if( flags & acf_require_scalar ) {
		return cerr::requires_scalar;
	}
	
	return cvt.start;
}

type_converter::convert_t type_converter::convert_impl( type_ref const & to, 
	type_ref const & from_, unsigned flags ) {
	if (!to.is_extr() || !from_.is_extr())  {
		return cerr::expect_extr_type;
	}
	return convert_impl( to.extr(), from_.extr(), flags );
}
	
type_converter::convert_t type_converter::convert_impl( extr_type const & to_, 
	extr_type const & from_, unsigned flags_ ) {
	
	auto flags = flags_;
	
	//DEBUG( "CVT", dump::get( to_ ), dump::get( from_ ) );
	extr_type const * from = &from_;
	type_ref holder;
	
	if( to_.is<intr_type_value_list>() ) {
		if( !from->is<intr_type_value_list>() )
			return cerr::requires_value_list;
		return convert_list( to_.cast<intr_type_value_list>(), 
			from->cast<intr_type_value_list>(), flags, cvt_expression::c_pack_value );
	}
 	
	converter cvt;
	bool allow_cvt = flags & cf_allow_cvt;

	//a source value-list of 1 however can just be removed
	while( true ) {
		auto from_vl = from->if_cast<intr_type_value_list>();
		if( from_vl && from_vl->sub.size() == 1 ) {
			cvt.collapse();
			from = &from_vl->sub[0].type.extr();
			continue;
		}
		break;
	}
	
	extr_type to = to_;
	if( flags & cf_to_lvalue ) {
		if( !to.is_lvalue() )
			return cerr::requires_lvalue;
		flags = flags & (~cf_to_lvalue);
		to.set_lvalue(false);
	}
	if( flags & cf_to_refer ) {
		if( !to.is_refer() )
			return cerr::requires_refer;
		flags = flags & (~cf_to_refer);
		to.set_ref_value();
	}
	
	auto from_tuple = from->if_cast<intr_type_tuple>();
	auto to_tuple = to.if_cast<intr_type_tuple>();
	if( from_tuple && from_tuple->selectable
		&& (!to_tuple || !to_tuple->selectable) ) {
		return select_convert( to, *from, flags );
	}
	
	bool lvalue_cvt = false;
	bool post_addr_of = false;
	
	auto if_addr_of = [&]() {
		if( from->is_lvalue() ) {
			if( to.is_value_ptr() && !from->is_value_ptr() ) {
				post_addr_of = true;
			} else {
				cvt.drop_lvalue();
			}
		}
	};
	if( from->is_lvalue() != to.is_lvalue() ) {
		if( !allow_cvt ) {
			return cerr::mismatch_lvalue;
		} else if( !to.is_lvalue() ) {
			if_addr_of();
		} else {
			return cerr::requires_lvalue;
		}
	} else if( to.is_lvalue() ) {
		//converting to an lvalue always requires an exact match
		lvalue_cvt = true;
	}
	
	bool make_optional = false;
	if( to.is_optional() != from->is_optional() ) {
		if( !allow_cvt ) {
			return cerr::mismatch_optional;
		} else if( lvalue_cvt && to.is_optional() ) {
			return cerr::requires_optional;
 		} else  if( !to.is_optional() ) {
			return cerr::mismatch_optional;
		} else {
			//optional has to happen after whatever else changed
			//TODO: this should happen at end of conversion, it's distinct from the deref_optional
			make_optional = true;
		}
	} else if( to.is_optional() ) {
		//need a strict comparison now as optional from type cannot be converted
		//convertable optionals are however exempt form this restriction
		allow_cvt = from->is_cvt_optional();
	}
	
	if( to.is_shared() != from->is_shared() ) {
		if( !allow_cvt ) {
			return cerr::mismatch_shared;
		} else if( !to.is_shared() ) {
			cvt.deref();
		} else {
			return cerr::requires_shared;
		}
	} else if( !post_addr_of && to.is_value_ptr() != from->is_value_ptr() ) {
		if( !allow_cvt ) {
			return cerr::mismatch_value_ptr;
		} else if( !to.is_value_ptr() ) {
			cvt.deref();
		} else {
			return cerr::requires_value_ptr;
		}
	} else if( to.is_refer() ) {
		allow_cvt = false; //requires strict conversion now
	}
	
	if( post_addr_of )
		cvt.addr_of();
		
	//no more lvalue conversion steps
	if( lvalue_cvt )
		allow_cvt = false;
	
	bool need_distinct = false;
	if( to.intr->is_whole() ) {
		//non-distinct types may convert to compatible distinct types
		if( from->intr->is_whole() ) {
			if( *from->intr != *to.intr )
				return cerr::mismatch_defined;
				
			//skip rest since no conversion is necessary (required for recursion)
			goto skip_check;
		} else {
			//must do after all part conversions/checks
			need_distinct = true;
		}
	} else if( from->intr->is_whole() ) {
		return cerr::mismatch_defined;
	} else {
		//ensure parametric instances match (special case for raw array)
		auto from_array = from->if_cast<intr_type_array>(); 
		auto to_array = to.if_cast<intr_type_array>();
		if( to_array && to_array->raw ) {
			if( from_array && from_array->raw ) {
				allow_cvt = false;
			} else if( allow_cvt && from_array && !from_array->raw ) {
				if( !from_array->sub.extr().is_plain() )
					return cerr::invalid_array_cvt;
				if( from_array->sub.extr() != to_array->sub.extr() )
					return cerr::invalid_array_cvt;
				cvt.addr_of();
				allow_cvt = false;
				//from = &from_array->sub;
			}
		} else if( to.intr->instance_of && from->intr->instance_of ) {
			//TODO: a strict equality check would make more sense, but we need to resolve exporting types first
			if(to.intr->instance_of->compare( *from->intr->instance_of, false ) != 0 ||
				to.intr->applied_params.size() != from->intr->applied_params.size() ) {
				return cerr::mismatch_params;
			}
			
			for( size_t i=0; i < to.intr->applied_params.size(); ++i ) {
				if( to.intr->applied_params[i].compare( from->intr->applied_params[i], true ) != 0 ) {
					return cerr::mismatch_params;
				}
			}
		} else if( from->intr->instance_of ) {
			return cerr::mismatch_params;
		}
	}
	
	{	//scope for skip_check goto
		auto to_function = to.if_cast<intr_type_function>();
		auto from_function = from->if_cast<intr_type_function>();
		if( to_function && from_function ) {
			if( to_function->can_fail != from_function->can_fail ) {
				return cerr::mismatch_function_throw;
			}
			
			if( from_function->contains_parametric() ) {
				auto pt = type_identifier::parameterize_type( ctx, *from, to );
				if( !pt.okay() ) {
					return convert_t(pt);
				}

				//DEBUG( "PT", dump::get(pt.result));
				holder = pt.result;
				auto q = cvt.param_inst();
				q->target_type = pt.result;
				from = &holder.extr();
				from_function = &from->cast<intr_type_function>();
			}
			
			if( to_function->access != from_function->access ) {
				if( !allow_cvt )
					return cerr::mismatch_function_access;
					
				if( to_function->access == intr_type_function::a_bound
					&& from_function->access == intr_type_function::a_raw ) {
					cvt.bind_raw_func();
				} else if( to_function->access == intr_type_function::a_bound
					&& from_function->is_unbound() ) {
					cvt.bind_func();
				} else {
					return cerr::mismatch_function_access;
				}
			}
			
			//here the function itself is being converted, not the parameters for a function
			//call, thus only a very strict comparison can be allowed.
			if( !to_function->returns != !from_function->returns) {
				return cerr::incomplete;
			}
			if( to_function->returns ) {
				auto ci = convert_impl( extr_type(to_function->returns), 
					extr_type(from_function->returns), 0 );
				if( !cvt.extend( ci ) )
					return ci;
			}
			
			auto ci = convert_impl( extr_type(to_function->args), 
				extr_type(from_function->args), 0 );
			if( !cvt.extend( ci ) )
				return ci;
			
			goto skip_check;
		}	
	}

	//beyond here no incomplete types can be converted (some might need to move before this
	//to support incomplete literal variables, like functions)
	if( !to.is_complete() || !from->is_complete() ) {
		return cerr::incomplete;
	}

	if( allow_cvt && to.is<intr_type_integral>() && from->is<intr_type_rational>() ) {
		auto & from_r = from->cast<intr_type_rational>();
		if( !from_r.integer_bitsize )
			return cerr::range_loss;
		if( to.is<intr_type_binary>() && from_r.integer_signed )
			return cerr::mismatch_signed;
			
		auto & to_fnvt = to.cast<intr_type_fixed_number>();
		if( *from_r.integer_bitsize > to_fnvt.data_bitsize )
			return cerr::range_loss;
			
		auto n = cvt.fixate();
		n->to_fun = to.is<intr_type_integer>() ? cvt_expression::p_integer : cvt_expression::p_binary;
		n->to_bitsize = to_fnvt.data_bitsize;
		
	} else if( allow_cvt && to.is<intr_type_integral>() && from->is<intr_type_float>() ) {
		if( !from->is_lossy() )
			return cerr::mismatch_fundamental;
			
		auto & to_fnvt = to.cast<intr_type_fixed_number>();
		
		auto n = cvt.loss();
		n->from_fun = cvt_expression::p_float;
		n->to_fun = to.is<intr_type_integer>() ? cvt_expression::p_integer : cvt_expression::p_binary;
		n->to_bitsize = to_fnvt.data_bitsize;
		
	} else if( allow_cvt && to.is<intr_type_binary>() && from->is<intr_type_integer>() ) {
		if( !from->is_lossy() )
			return cerr::mismatch_fundamental;
			
		auto & to_fnvt = to.cast<intr_type_fixed_number>();
		
		auto n = cvt.loss();
		n->from_fun = cvt_expression::p_integer;
		n->to_fun = cvt_expression::p_binary;
		n->to_bitsize = to_fnvt.data_bitsize;
			
	} else if( allow_cvt &&  to.is<intr_type_float>() && from->is<intr_type_rational>() ) {
		auto & to_fnvt = to.cast<intr_type_fixed_number>();
		//TODO: check that range fits in the target type, we don't care about precision here

		auto n = cvt.fixate();
		n->to_fun = cvt_expression::p_float;
		n->to_bitsize = to_fnvt.data_bitsize;
		
	} else if( allow_cvt &&  to.is<intr_type_float>() && from->is<intr_type_integral>() ) {
		auto & from_nt= from->cast<intr_type_fixed_number>();
		auto & to_nt = to.cast<intr_type_float>();
		
		//-1 since one of those is the sign-bit whtih the float also has in addition to its significand
		auto nf = platform::if_get_float(to_nt.data_bitsize);
		if( !nf )
			return cerr::invalid_size;
		bool loss = (from_nt.data_bitsize-(from->is<intr_type_integer>() ? 1 : 0)) > nf->significand_size;
		if( loss && !from->is_lossy() )
			return cerr::precision_loss;
			
		auto n = cvt.promote_loss(loss);
		n->to_bitsize = to_nt.data_bitsize;
		n->from_fun = from->is<intr_type_integer>() ? cvt_expression::p_integer : cvt_expression::p_binary;
		n->to_fun = cvt_expression::p_float;

	} else if( allow_cvt &&  to.is<intr_type_integer>() && from->is<intr_type_binary>() ) {
		auto & to_fnvt = to.cast<intr_type_fixed_number>();
		auto & from_fnvt = from->cast<intr_type_fixed_number>();
		//requires 1-extra bit to store sign
		bool loss = from_fnvt.data_bitsize >= to_fnvt.data_bitsize;
		if( loss && !from->is_lossy() )
			return cerr::range_loss;
			
		auto n = cvt.promote_loss(loss);
		n->from_fun =cvt_expression::p_binary; 
		n->to_fun = cvt_expression::p_integer;
		n->to_bitsize = to_fnvt.data_bitsize;

	} else if( allow_cvt && !to.is<intr_type_contextual>() && from->is<intr_type_contextual>() ) {
		auto * to_func = to.if_cast<intr_type_function>();
		if( !to_func ) {
			return cerr::unsupported;
		}
		
		auto & from_ctx = from->cast<intr_type_contextual>();
		auto from_func = from_ctx.get_unbound().extr().if_cast<intr_type_function>();
		if( !from_func ) {
			return cerr::unsupported;
		}
		if( from_func->access != intr_type_function::a_unbound ) {
			return cerr::unsupported;
		}
		//DEBUG( "CTX", dump::get(from_func->context), dump::get(from_ctx.get_context()));
		//TODO: is conversion actually needed here?
		if( from_func->context.extr().intr != from_ctx.get_context().extr().intr ) {
			return cerr::mismatch_fundamental;
		}

		if( to_func->access == intr_type_function::a_bound ) {
			cvt.bind_context();
		} else if( to_func->access == intr_type_function::a_unbound ) {
			if( to_func->context != from_func->context ) {
				return cerr::mismatch_fundamental;
			}
			
			auto cf = cvt.field(); //TODO: hmmm, but do we want addiitonal converters just for this special field (consider using expr_special_field enum somehow)
			cf->field = 1;
			
			auto sub = convert_impl( to_, from_ctx.get_unbound().extr(), flags_ );
			if(!cvt.extend(sub)) {
				return sub;
			}
			return cvt.start;
		} else {
			return cerr::unsupported;
		}

	} else if( to.is<intr_type_compat>() ) {
		auto & to_compat = to.cast<intr_type_compat>();
		if( to_compat.type == intr_type_compat::abi_var_args ) {
			if (!from->is<intr_type_tuple>()) {
				return cerr::mismatch_fundamental;
			}
			
			auto c = cvt.make_compat();
			c->target_type = extr_type(to.intr);
		} else if( to_compat.type == intr_type_compat::string_tag ) {
			if( !from->is<intr_type_compat>()) {
				return cerr::mismatch_fundamental;
			}
			
			auto & from_compat = from->cast<intr_type_compat>();
			if( from_compat.type != to_compat.type ) {
				return cerr::mismatch_fundamental; //TODO: not a fundamental mismatch
			}
		} else {
			return cerr::mismatch_fundamental;
		}
		
	} else 	if( to.intr->get_fun_class() != from->intr->get_fun_class() ) {
		return cerr::mismatch_fundamental;
		
		//past here the fundamental is the same
	} else if( to.is<intr_type_integral>() || to.is<intr_type_float>() ) {
		auto & to_fnvt = to.cast<intr_type_fixed_number>();
		auto & from_fnvt = from->cast<intr_type_fixed_number>();
		if( to_fnvt.data_bitsize != from_fnvt.data_bitsize ) {
		
			if( !allow_cvt )
				return cerr::mismatch_data_size;

			bool loss = to_fnvt.data_bitsize < from_fnvt.data_bitsize;
			if( loss && !from->is_lossy() ) {
				return to.is<intr_type_float>() ? cerr::precision_loss : cerr::range_loss;
			}
			auto n = cvt.promote_loss(loss);
			n->to_bitsize = to_fnvt.data_bitsize;
			if( to.is<intr_type_float>() ) {
				n->from_fun = n->to_fun = cvt_expression::p_float;
			} else if( to.is<intr_type_binary>() ) {
				n->from_fun = n->to_fun = cvt_expression::p_binary;
			} else {
				n->from_fun = n->to_fun = cvt_expression::p_integer;
			}
		}
		
	} else if( to.is<intr_type_tuple>() || to.is<intr_type_contextual>() ) {
		auto ci = convert_list( to.cast<intr_type_compound>(), 
			from->cast<intr_type_compound>(), allow_cvt ? cf_allow_cvt : 0, 
			to.is<intr_type_tuple>() ? cvt_expression::c_pack : cvt_expression::c_pack_contextual );
		if( !cvt.extend( ci ) )
			return ci;
			
	}

skip_check:
	if( to.is_literal() != from->is_literal() ) {
		if( !from->is_literal() ) {
			return cerr::requires_literal;
		}
		if( !(flags & cf_retain_literal) ) {
			cvt.drop_literal();
		}
	}
	
	if( need_distinct ) {
		auto sd = cvt.set_distinct();
		sd->target_type = to;
	}
	
	if( make_optional )
		cvt.optional();

	return cvt.start;
}

type_converter::resolve_t type_converter::fixate( type_ref const & from ) {
	resolve_t it;
	try {
		if( from.is_extr() ) {
			it.result = fixate_impl( from.extr() );
		} else {
			it.error = cerr::invalid_type_ref;
		}
	} catch( cerr e ) {
		it.error = e;
	}
	return it;
}

/**
	Fixation always picks the type which would be inferred from the same type. This ensures
	consistency. It doesn't always guarantee the best conversion, but here the consistency
	is more important. Such generic fixation is only used when no other type is identified in
	the expression.
*/
type_ref type_converter::fixate_impl( extr_type const & from ) {
	if( from.is<intr_type_rational>() ) {
		auto & trat = from.cast<intr_type_rational>();	
		if( trat.integer_bitsize ) {
			if( *trat.integer_bitsize > platform::natural_integer_size )
				throw cerr::large_rational_integer;
				
			auto n = make_shared<intr_type_integer>();
			n->data_bitsize = platform::natural_integer_size;
			return extr_type(n);
		}
		
		auto n = make_shared<intr_type_float>();
		n->data_bitsize = platform::natural_float_size;
		return extr_type(n);
	}

	if( from.is<intr_type_value_list>() || from.is<intr_type_tuple>() ) {
		shared_ptr<intr_type_compound> comp
			= from.cast<intr_type_compound>().clone_derived();
		for( size_t i=0; i < comp->sub.size(); ++ i ) {
			if( comp->sub[i].type.is_extr()) {
				comp->sub[i].type = fixate_impl( comp->sub[i].type.extr() );
			}
		}
		return extr_type(comp);
	}
	
	return from;
}

type_converter::resolve_t type_converter::unify( type_ref const & a, type_ref const & b ) {
	resolve_t ut;
	try {
		if( a.is_extr() && b.is_extr() ) {
			ut.result = unify_impl( a.extr(), b.extr() );
		} else {
			ut.error = cerr::mismatch_type_ref;
		}
	} catch( cerr e ) {
		ut.error = e;
	}
	return ut;
}

type_converter::resolve_t type_converter::unify( type_ref const * begin, size_t len ) {
	PRE_CHECK( len );
	if( len == 1 ) {
		resolve_t ut;
		ut.result = begin[0];
		return ut;
	}
	if( len == 2 ) {
		return unify( begin[0], begin[1] );
	}
	
	auto a = unify( begin, len / 2 );
	auto b = unify( begin + len/2, len - len/2 );
	if( a.error != cerr::okay )
		return a;
	if( b.error != cerr::okay )
		return b;
	return unify( a.result, b.result );
}

type_ref type_converter::unify_value_list( extr_type const & a, extr_type const & b, unsigned flags ) {
	
	auto alist = a.if_cast<intr_type_value_list>();
	auto blist = b.if_cast<intr_type_value_list>();
	
	//1-lists can collapse
	if( flags & uf_allow_collapse) {
		if( !alist ) {
			if( blist->sub.size() != 1 )
				throw cerr::mismatch_size;
			return unify_impl( a, blist->sub[0].type.extr() );
		}
		if( !blist ) {
			if( alist->sub.size() != 1 )
				throw cerr::mismatch_size;
			return unify_impl( alist->sub[0].type.extr(), b );
		}
	} else if( !alist || !blist ) {
		throw cerr::mismatch_size;
	}
	
	//if same size unify the components
	if( alist->sub.size() == blist->sub.size() ) {
		auto ulist = make_shared<intr_type_value_list>();
		for( size_t i=0; i < alist->sub.size(); ++ i ) {
			std::string name = alist->sub[i].name;
			if( blist->sub[i].name.size() ) {
				if( name == blist->sub[i].name ) {
					//keep
				} else if( name.size() ) {
					throw cerr::mismatch_field_name;
				} else {
					name = blist->sub[i].name;
				}
			}
			ulist->push_sub( name, unify_impl( alist->sub[i].type.extr(), blist->sub[i].type.extr() ) );
		}
		return extr_type(ulist);
	}
	
	if( flags & uf_allow_collapse ) {
		//again, 1-list can collapse
		if( alist->sub.size() == 1 )
			return unify_impl( alist->sub[0].type.extr(), b );
		if( blist->sub.size() == 1 )
			return unify_impl( a, blist->sub[0].type.extr() );
	}
	
	throw cerr::mismatch_size;
}

/**
	unify_impl is supposed to unify according to the rules of type conversion. I see no reasonable
	way to share the code from above to ensure that. Thus it'll just have to be covered with
	clear rules, good tests, and ultimately a very thorough specification.
*/
type_ref type_converter::unify_impl( extr_type const & a, extr_type const & b ) {
	//DEBUG( "UNIFY", dump::get(a), "::", dump::get(b));
	//equivalent types are always fine
	if( a == b )
		return a;
	
	if( a.is<intr_type_value_list>() || b.is<intr_type_value_list>() ) {
		return unify_value_list(a, b, uf_allow_collapse);
	}
	
	//literals are lost if not the same
	if( a.is_literal() != b.is_literal() ) {
		auto ac = a;
		ac.set_literal(false);
		auto bc = b;
		bc.set_literal(false);
		return unify_impl( ac, bc );
	}
	
	//lvalue's must be exact (but failed previous exact test), but lvalue can be dropped
	if( a.is_lvalue() || b.is_lvalue() ) {
		auto ac = a;
		ac.set_lvalue( false );
		auto bc = b;
		bc.set_lvalue( false );
		return unify_impl( ac, bc );
	}
	
	if( a.is_optional() != b.is_optional() ) {
		if( a.is_optional() ) {
			auto bc = b;
			bc.set_optional( true );
			bc.set_cvt_optional( true );
			return unify_impl( a, bc );
		} else {
			auto ac = a;
			ac.set_optional( true );
			ac.set_cvt_optional( true );
			return unify_impl( ac, b );
		}
	}

	//if both sides are cvt_optional it may be retained (since a conversion can flow through to both)
	if( a.is_cvt_optional() != b.is_cvt_optional() ) {
		auto ac = a;
		ac.set_cvt_optional( false );
		auto bc = b;
		bc.set_cvt_optional( false );
		return unify_impl( ac, bc );
	}
	
	if( a.is_shared() != b.is_shared() ) {
		if( a.is_shared() ) {
			auto ac = a;
			ac.set_ref_value();
			return unify_impl( ac, b );
		} else {
			auto bc = b;
			bc.set_ref_value();
			return unify_impl( a, bc );
		}
	}
	
	auto arat = a.if_cast<intr_type_rational>();
	auto brat = b.if_cast<intr_type_rational>();
	if( a.intr->get_fun_class() != b.intr->get_fun_class() ) {
		
		auto xrat = arat ? arat : brat;
		auto xalt = arat ? &b : &a;
		
		if( xrat ) {
			//special case for intr_type_integral, we can unify the bitsize
			auto xint = xalt->if_cast<intr_type_integral>();
			auto xbits = xrat->integer_bitsize;
			if( xbits && xrat && xint ) {
				if( *xbits <= xint->data_bitsize )
					return *xalt;
				auto dup = xint->clone_derived();
				dup->data_bitsize = *xbits;
				return extr_type(dup);
			}
			
			//otherwise rationals always unify to the other type, even if they could not be converted to that
			//type. This provides a consistency in how types of unified/fixated.
			if( xrat && xalt->is<intr_type_scalar>() ) {
				return *xalt;
			}
		}
		
		//a lossy value unifies to the other type
		if( b.is_lossy() ) {
			if( a.is_lossy() ) {
				throw cerr::unify_ambiguous;
			}
			return a;
		} else if( a.is_lossy() ) {
			return b;
		}
		
		throw cerr::mismatch_fundamental;
	}
	
	if( a.is<intr_type_integral>() || a.is<intr_type_float>() ) {
		auto & anum = a.cast<intr_type_fixed_number>();
		auto & bnum = b.cast<intr_type_fixed_number>();
		if( anum.data_bitsize < bnum.data_bitsize )
			return b;
		return a;
	}

	auto atuple = a.if_cast<intr_type_tuple>();
	auto btuple = b.if_cast<intr_type_tuple>();
	if( atuple ) {
		if( atuple->sub.size() != btuple->sub.size() )
			throw cerr::mismatch_size;
		
		auto ntuple = make_shared<intr_type_tuple>();
		for( size_t i=0; i < atuple->sub.size(); ++i ) {
			//names can be added but not removed
			std::string use_name = atuple->sub[i].name;
			if( use_name.empty() ) {
				use_name = btuple->sub[i].name;
			} else if( !btuple->sub[i].name.empty() && use_name != btuple->sub[i].name ) {
				throw cerr::mismatch_field_name;
			}
			ntuple->push_sub( use_name, unify_impl( atuple->sub[i].type.extr(), 
				btuple->sub[i].type.extr() ) );
		}
		return extr_type(ntuple);
	}
	
	if( arat && brat ) {
		//mixing with non-integer becomes non-integer
		if( !arat->integer_bitsize )
			return a;
		if( !brat->integer_bitsize )
			return b;
			
		//otherwise an integer of the larger bitsize
		if( *arat->integer_bitsize > *brat->integer_bitsize )
			return a;
		else
			return b;
	}
	
	auto afunc = a.if_cast<intr_type_function>();
	auto bfunc = b.if_cast<intr_type_function>();
	if( afunc && bfunc ) {
		auto c = make_shared<intr_type_function>();
		c->can_fail = afunc->can_fail || bfunc->can_fail;
		if( afunc->construct_result.is_defined() || bfunc->construct_result.is_defined() ) {
			if( !afunc->construct_result.is_extr() || !bfunc->construct_result.is_extr() ) {
				throw cerr::unify_impossible;
			}
			c->construct_result = unify_impl( afunc->construct_result.extr(), bfunc->construct_result.extr() );
		}
		
		c->args = std::dynamic_pointer_cast<intr_type_value_list const>(unify_value_list(
			extr_type(afunc->args), extr_type(bfunc->args)).extr().intr);
		c->returns = std::dynamic_pointer_cast<intr_type_value_list const>(
			unify_value_list(extr_type(afunc->returns), extr_type(bfunc->returns)).extr().intr);
		STATE_CHECK(c->args && c->returns );
		return extr_type(c);
	}
	throw cerr::unify_impossible;
}

/**
	Does a conversion by selecting the sub element which has the best conversion cost.
*/
type_converter::convert_t type_converter::select_convert( extr_type const & to, 
	extr_type const & from, unsigned flags )
{
	auto & from_tuple = from.cast<intr_type_tuple>();

	std::vector<convert_t> match;
	converter cvt;
	
	//TODO: this pattern is repeated, make generic
	if( from.is_lvalue() ) {
		cvt.drop_lvalue();
	}
	if( from.is_refer() ) {
		cvt.deref();
	}
	
	for( size_t i=0; i < from_tuple.sub.size(); ++i ) {
		auto & tref = from_tuple.sub[i];
		auto cost = convert_impl( to, tref.type, flags );
		if( !cost.okay() )
			continue;

		if( !match.empty() ) {
			int cmp = match[0].compare_cost( cost );
			if( cmp > 0 ) { //better match
				match.clear();
			} else if( cmp < 0 ) { //same cost
				continue;
			}
		}
		
		//add field access
		auto field = make_shared<cvt_expression>(cvt_expression::c_field);
		field->field = i;
		field->then = cost.conversions;
		cost.conversions = field;
		match.push_back( cost );
	}
	
	if( match.empty() )
		return cerr::select_none;
	if( match.size() > 1 )
		return cerr::select_multiple;
	
	cvt.extend( match[0] );
	return cvt.start;
}

/**
	Find a function which has the lowest call conversion cost from the given arguments.
*/
std::vector<size_t> type_converter::match_function( intr_type_tuple const & select,
	intr_type const & args_ ) {
	
	std::vector<size_t> matches;
	
	struct match_t {
		size_t what;
		type_converter::convert_t cost;
	};
	std::vector<match_t> match;
	
	auto args = args_.cast<intr_type_value_list>();
	for( size_t i=0; i < select.sub.size(); ++i )
	{
		intr_type_function const * func;
		if( select.sub[i].type.extr().is<intr_type_contextual>() ) {
			//ignore all extrinisics and decorators since match works on args signature only
			auto & vt = select.sub[i].type.extr().cast<intr_type_contextual>();
			func = vt.get_unbound().extr().if_cast<intr_type_function>();
		} else {
			func = select.sub[i].type.extr().if_cast<intr_type_function>();
		}
		if( !func ) {
			continue;
		}
		
		STATE_CHECK( func->args );
		auto & func_args = *func->args;
		
		//OPT: should return matching conversions, otherwise caller does again
		if( func_args.sub.size() < args.sub.size() )
			continue;

		match_t check;
		check.what = i;
		check.cost = convert_function_call( func_args, args, true );
		if( check.cost.okay() ) {
			if( match.empty() ) {
				match.push_back( check );
			} else {
				int cmp = match[0].cost.compare_cost( check.cost );
				if( cmp == 0 ) {
					match.push_back( check );
				} else if( cmp > 0 ) {
					match.clear();
					match.push_back( check );
				} //else most costly conversion
			}
		}
	}
	
	for( auto & m : match ) {
		matches.push_back(m.what);
	}
	return matches;
}


} //eon leaf