#include "lang/class_typer.hpp"
#include "lang/context.hpp"
#include "lang/declaration.hpp"
#include "lang/expression_typer/common.hpp"
#include "lang/scope.hpp"
#include "lang/type.hpp"
#include "lang/type_identifier.hpp"
#include "lang/type_spec.hpp"
#include "platform/platform.hpp"
#include "util/defer.hpp"

#include <map>

#define SHOW_DEBUG 0
#include "util/debug.hpp"
#include "lang/dump.hpp"

namespace leaf {
namespace type_identifier {

typedef std::map<std::string, type_ref> param_map_t;
typedef std::vector<extr_type> param_vector_t;

resolve_t morph( context const & ctx, type_spec const * constraint, type_ref source, unsigned flags );
resolve_spec_t morph_spec( context const & ctx, type_spec const * constraint, type_ref source, unsigned flags );
resolve_value_list_t morph_value_list( context const & ctx, type_ref source, unsigned flags );

resolve_spec_t expand( context const & ctx, type_spec const & cvt, param_map_t const & params, unsigned flags = 0 );
cerr_base expand_into( context const & ctx, type_spec & to, type_spec const & from,
	param_map_t const & params, unsigned flags );
	

void complete_defaults( context const & ctx, type_spec & ts );
void apply_infer( context const & ctx, type_spec & ts );

context create_function_scope(context const & ctx) {
	auto fs = scope::create_pseudo( scope_style_t::declarative );
	fs->set_base( ctx.sc, scope::bt_parent );

	return ctx.derive( fs );
}

#include "type_identifier/get_spec.inc.cpp"
#include "type_identifier/instantiate.inc.cpp"
#include "type_identifier/defaults.inc.cpp"
#include "type_identifier/determine.inc.cpp"
#include "type_identifier/parameterize.inc.cpp"
#include "type_identifier/expand.inc.cpp"

resolve_t infer( context const & ctx, type_spec const & constraint, type_ref const & source, bool init_refer ) {
	return morph( ctx, &constraint, source, df_infer | (init_refer ? df_refer : 0) );
}
resolve_t infer( context const & ctx, shared_ptr<type_spec const> constraint, 
	type_ref const & source, bool init_refer ) {
	return morph( ctx, constraint.get(), source, df_infer | (init_refer ? df_refer : 0) ); 
}
resolve_spec_t infer_spec( context const & ctx, type_spec const & spec, type_ref assign, bool init_refer ) {
	return morph_spec( ctx, &spec, assign, df_infer | (init_refer ? df_refer : 0) );
}
resolve_t infer( context const & ctx, type_ref const & source, bool init_refer ) {
	return morph( ctx, nullptr, source, df_infer | (init_refer ? df_refer : 0) );
}

/**
	Constrains the given type to the indicated spec. This is a lesser reduction
	than "infer" since it may retain literals and virtual types like rationals.
*/
resolve_t constrain( context const & ctx, type_spec const & constraint, type_ref const & source ) {
	return morph( ctx, &constraint, source, df_allow_param );
}
resolve_t constrain( context const & ctx, shared_ptr<type_spec const> constraint, type_ref const & source ) {
	return morph( ctx, constraint.get(), source, df_allow_param ); 
}

/**
	Infers a value-list (such as for inferred return types).
*/
resolve_value_list_t infer_value_list( context const & ctx, type_ref source, bool init_refer ) {
	return morph_value_list( ctx, source, df_infer | (init_refer ? df_refer : 0) );
}

resolve_t morph( context const & ctx, type_spec const * constraint_, type_ref source, unsigned flags ) {
	auto constrained = morph_spec( ctx, constraint_, source, flags );
	if( !constrained.okay() )
		return resolve_t(constrained);
		
	//OPT: this needlessly expands the source again
	return determine( ctx, *constrained.result, flags );
}

resolve_spec_t morph_spec( context const & ctx, type_spec const * constraint_, type_ref source, unsigned flags ) {
	
	static type_spec const empty_spec;
	auto & constraint = constraint_ ? *constraint_ : empty_spec;
	
	//decompose value-list
	while( source.is_extr() ) {
		auto vl = source.extr().if_cast<intr_type_value_list>();
		if( !vl )
			break;
			
		if( vl->sub.size() != 1 ) {
			return cerr::disallow_value_list;
		}
		source = vl->sub[0].type;
	}
	
	//we need to work with an expanded spec
	auto ex_constraint = expand( ctx, constraint );
	if( !ex_constraint.okay() ) {
		return ex_constraint;
	}
	complete_defaults( ctx, *ex_constraint.result );
	
	auto source_spec = get_spec( ctx, source );
	STATE_CHECK( source_spec.okay(), source_spec.explain_error() );
	if( flags & df_infer ) {
		apply_infer( ctx, *source_spec.result );
	}

	//init_refer requires a shared type
	if( flags & df_refer ) {
		source_spec.result->set<pt_reference>( extr_type::r_shared );
	}
	
	param_map_t pt;
	auto r = expand_into( ctx, *ex_constraint.result, *source_spec.result, pt, flags );
	if( !r.okay() ) {
		return resolve_spec_t(r);
	}
	return ex_constraint.result;
}

resolve_value_list_t morph_value_list( context const & ctx, type_ref source, unsigned flags ) {
	//decompose value-list
	while( true ) {
		auto vl = source.extr().if_cast<intr_type_value_list>();
		if( !vl ) {
			auto part = morph( ctx, nullptr, source, flags );
			if( !part.okay() ) {
				return resolve_value_list_t(part);
			}
			auto list = make_shared<intr_type_value_list>();
			list->push_sub(part.result);
			
			return list;
		}
			
		if( vl->sub.size() == 1 ) {
			source = type_ref(vl->sub[0].type);
			continue;
		}
		
		auto list = make_shared<intr_type_value_list>();
		for( auto & sub : vl->sub ) {
			auto part = morph( ctx, nullptr, sub.type, flags );
			if( !part.okay() ) {
				return resolve_value_list_t(part);
			}
			list->push_sub(part.result);
		}
		return list;
	}
	
}

} //eon type_identifier
} //eon leaf