//included in ../type_identifier.cpp

/*
	Determination produces a resolved extr_type from a type_spec.
	
	If df_allow_param is specified the result may be a parametric, otherwise it is a extr_type.
*/

resolve_t determine( context const & ctx, shared_ptr<type_spec const> vt, unsigned flags ) {
	if( !vt )
		return cerr::null_spec;
	return determine( ctx, *vt, flags );
}

namespace {
	/* Convert the extrinsic parts of the type_spec into an extr_type */
	extr_type determine_extr_type( shared_ptr<intr_type const> const & intr, type_spec const & vt ) {
		STATE_CHECK(intr);
		
		extr_type tr(intr);
		
		if( vt.has<pt_lvalue>() ) {
			tr.set_lvalue( vt.get<pt_lvalue>() );
		}
		if( vt.has<pt_reference>() ) {
			tr.set_reference( vt.get<pt_reference>() );
		} else if( intr->is<intr_type_service>() ) {
			tr.set_ref_shared();
		}
		if( vt.has<pt_optional>() ) {
			tr.set_optional( vt.get<pt_optional>() );
		}
		if( vt.has<pt_cvt_optional>() ) {
			tr.set_cvt_optional( vt.get<pt_cvt_optional>() );
		}
		if( vt.has<pt_lossy>() ) {
			tr.set_lossy( vt.get<pt_lossy>() );
		}
		if( vt.has<pt_literal>() ) {
			tr.set_literal( vt.get<pt_literal>() );
		}
		
		return tr;
	}
	
	/* Will instantiate a parametric intrinsic type if the parameters have been specified. If any are unresolved
		it will return unresolved_param. */
	resolve_intr_type_t determine_prototype_param_part( context const & ctx, type_spec::part const & part, 
		shared_ptr<intr_type const> prototype, unsigned flags  ) {
		param_vector_t pt;
		for( size_t i = 0; i < part.sub.size(); ++i ) {
			//this forces instantances to use fully defined extrinsics
			auto c = determine( ctx, *part.sub[i], flags );
			if( !c.okay() ) { 
				return resolve_intr_type_t(c.error);
			}
			if (!c.result.is_extr()) {
				return cerr::unresolved_param;
			}
			pt.push_back( c.result.extr() );
		}
		
		return instantiate( ctx, prototype, pt );
	}
	
	#pragma GCC diagnostic push
	#pragma GCC diagnostic ignored "-Wunused-parameter"
	
	typedef cerr_impl<shared_ptr<intr_type>> resolve_intr_type_tm;
	
	#define DETFUNC(s) resolve_intr_type_tm determine_intr_type_##s( \
		context const & ctx, type_spec const & vt, unsigned flags )
	DETFUNC(boolean) {
		return resolve_intr_type_tm(make_shared<intr_type_boolean>());
	}
		
	DETFUNC(char) {
		return resolve_intr_type_tm(make_shared<intr_type_char>());
	}

	DETFUNC(tuple) {
		auto tt = make_shared<intr_type_tuple>();
		if( vt.has<pt_pack>() )
			tt->pack = vt.get<pt_pack>();
		if( vt.has<pt_selectable>() ) {
			tt->selectable = vt.get<pt_selectable>();
		}
			
		for( size_t i=0; i < vt.sub.size(); ++i ) {
			auto pt = determine( ctx, vt.sub[i].type, flags | df_sub_type );
			if( !pt.okay() )
				return resolve_intr_type_tm(pt);
			tt->push_sub( vt.sub[i].name, pt.result );
		}
		return resolve_intr_type_tm(tt);
	}
	
	DETFUNC(value_list) {
		//create a new scope to allow subsequence params/returns to reference previous ones
		auto func_ctx = create_function_scope(ctx);
	
		auto tt = make_shared<intr_type_value_list>();
		for( size_t i=0; i < vt.sub.size(); ++i ) {
			auto pt = determine( func_ctx, vt.sub[i].type, flags ); //do not add df_sub_type here (it may already be applied however)
			if( !pt.okay() )
				return resolve_intr_type_tm(pt);
			tt->push_sub( vt.sub[i].name, pt.result );
			
			if( vt.sub[i].name.size() ) {
				auto sd = declaration::create();
				sd->symbol = vt.sub[i].name;
				sd->type = pt.result;
				//DEBUG( "ADD-B", sd->symbol );
				func_ctx.sc->add(sd);
			}
			
		}
		return resolve_intr_type_tm(tt);
	}
	
	DETFUNC(function) {
		auto ft = make_shared<intr_type_function>();
		
		bool immediate_member = (flags && df_imply_context) &&
			ctx.get_decl_scope()->is_class_root() && 
			vt.get<pt_member>(true) && !(flags & df_sub_type);
			
		if( vt.has<pt_context>() ) {
			ft->context = extr_type(vt.get<pt_context>());
		} else if( vt.has<pt_context_name>() ) {
			ft->access = intr_type_function::a_unbound;
			
			auto fctx = ctx.sc->resolve_symbol( 
				vt.get<pt_context_name>() );
			if( !fctx.okay() ) {
				return resolve_intr_type_tm( fctx.error, vt.get<pt_context_name>() );
			}
			if( fctx.result.what != scope_symbol::w_fun_type
				|| !fctx.result.fun_type->is<intr_type_custom>() ) {
				return cerr::invalid_context;
			}
			ft->context = extr_type(std::static_pointer_cast<intr_type_custom const>(fctx.result.fun_type));
		} else if( immediate_member || vt.get<pt_member>(false) ) {
			ft->access = intr_type_function::a_unbound;
			ft->context = ctx.get_decl_scope()->get_class_scope_context();
		}
		
		if( vt.has<pt_function_access>() ) {
			ft->access = vt.get<pt_function_access>();
		}
		
		if( vt.has<pt_function_throw>() ) {
			ft->can_fail = vt.get<pt_function_throw>();
		}
		
		if( vt.sub.size() != 1 && vt.sub.size() != 2 )
			return cerr::mismatch_size;

		auto args = determine( ctx, vt.sub[0].type, flags | df_sub_type );
		if( !args.okay() )
			return resolve_intr_type_tm(args);
		if( !args.result.extr().is<intr_type_value_list>() )
			return cerr::requires_value_list;
		ft->args = std::static_pointer_cast<intr_type_value_list const>(args.result.extr().intr);
		
		if( vt.sub.size() > 1 ) {
			auto returns = determine( ctx, vt.sub[1].type, flags | df_sub_type );
			DEBUG( "RETURNS", returns.okay(), dump::get(returns.result) );
			if( !returns.okay() )
				return resolve_intr_type_tm(returns);
			if( returns.result.is_spec() ) {
				//TODO: this isn't quite right, we really only care about an intrinsic return, not if somehow
				//it ended up parametric another way (which would be an error)
				//leave ft->returns as null
			} else {
				if( !returns.result.extr().is<intr_type_value_list>() )
					return cerr::requires_value_list;
				ft->returns = std::static_pointer_cast<intr_type_value_list const>(returns.result.extr().intr);
			}
		}
		
		return resolve_intr_type_tm(ft);
	}
	
	DETFUNC(array) {
		auto & p = vt.get_part(pt_fun_class);
		if( p.sub.size() != 1 )
			return cerr::mismatch_size;
		auto at = make_shared<intr_type_array>();
		if( vt.has<pt_function_access>() && vt.get<pt_function_access>() == intr_type_function::a_raw ) {
			at->raw = true;
		}
		auto rt = determine( ctx, p.sub[0], flags | df_sub_type );
		if( !rt.okay() ) {
			return resolve_intr_type_tm(rt);
		}
		if( !rt.result.is_extr() ) {
			return cerr::unresolved_param;
		}
		
		at->sub = rt.result.extr();
		return resolve_intr_type_tm(at);
	}
	
	//TODO: shares a lot with f_array (both intr_type_wrap)
	DETFUNC(type) {
		auto & p = vt.get_part(pt_fun_class);
		if( p.sub.size() != 1 )
			return cerr::mismatch_size;
		auto at = make_shared<intr_type_type>();
		auto rt = determine( ctx, p.sub[0], flags | df_sub_type );
		if( !rt.okay() ) {
			return resolve_intr_type_tm(rt);
		}
		if( !rt.result.is_extr() ) {
			return cerr::unresolved_param;
		}
		
		at->sub = rt.result.extr();
		return resolve_intr_type_tm(at);
	}
	
	DETFUNC(contextual) {
		auto & p = vt.get_part(pt_fun_class);
		if( p.sub.size() != 2 ) {
			return cerr::mismatch_size;
		}
		
		auto ct = make_shared<intr_type_contextual>();
		auto rt = determine( ctx, p.sub[0], flags | df_sub_type );
		if( !rt.okay() ) {
			return resolve_intr_type_tm(rt);
		}
		ct->get_context() = rt.result;
		
		rt = determine( ctx, p.sub[1], flags | df_sub_type );
		if( !rt.okay() ) {
			return resolve_intr_type_tm(rt);
		}
		ct->get_unbound() = rt.result;
		
		return resolve_intr_type_tm(ct);
	}

	template<typename T>
	resolve_intr_type_tm determine_intr_type_fixed_number( type_spec const & vt ) {
		auto it = make_shared<T>();
		if( vt.has<pt_data_bitsize>() ) {
			it->data_bitsize = vt.get<pt_data_bitsize>();
		} else {
			it->data_bitsize = platform::natural_integer_size;
		}
		return resolve_intr_type_tm(it);
	}
	
	DETFUNC(integer) {
		return determine_intr_type_fixed_number<intr_type_integer>(vt);
	}
	DETFUNC(binary) {
		return determine_intr_type_fixed_number<intr_type_binary>(vt);
	}
	
	DETFUNC(float) {
		auto ft = make_shared<intr_type_float>();
		if( vt.has<pt_data_bitsize>() ) {
			if( vt.has<pt_precision_variant>() ) {
				return cerr::redundant_bitsize;
			}
			ft->data_bitsize = vt.get<pt_data_bitsize>();
			if( !platform::if_get_float( ft->data_bitsize ) )
				return cerr::unknown_variant;
		} else if( vt.has<pt_precision_variant>() ) {
			auto nf = platform::if_get_float_precision( vt.get<pt_precision_variant>() );
			if( !nf )
				return cerr::unknown_variant;
			ft->data_bitsize = nf->size;
		} else {
			ft->data_bitsize = platform::natural_float_size;
		}
		return resolve_intr_type_tm(ft);
	}
	
	DETFUNC(rational) {
		auto rt = make_shared<intr_type_rational>();
		if( vt.has<pt_data_bitsize>() )
			rt->integer_bitsize = vt.get<pt_data_bitsize>();
			
		auto signed_ = vt.if_get<pt_signed>();
		if( signed_ ) {
			rt->integer_signed = *signed_;
		}
		return resolve_intr_type_tm(rt);
	}

	DETFUNC(compat) {
		auto rt = make_shared<intr_type_compat>();
		rt->type = vt.get<pt_compat_type>();
		return resolve_intr_type_tm(rt);
	}
	
	DETFUNC(service) {
		STATE_FAIL( "not-expected-here" );
	}
	DETFUNC(module) {
		STATE_FAIL( "not-expected-here" );
	}
	DETFUNC(class) {
		STATE_FAIL( "not-expected-here" );
	}
	DETFUNC(closure) {
		STATE_FAIL( "not-expected-here" );
	}
	#pragma GCC diagnostic pop
	
	
	/* Determine the intrinsic part of a type_spec. If the top-level type contains parameters this
		will return unresolved_param. If the sub-types contain parameters and the flags
		allow_params then the returns type will include type_spec components. */
	resolve_intr_type_t determine_intr_type( context const & ctx, type_spec const & vt, unsigned flags ) {
		shared_ptr<intr_type const> prototype = 
			vt.get<pt_fun_prototype>( shared_ptr<intr_type const>() ); //may be null
		
		auto use_intr_type = prototype;
		
		if( prototype && prototype->contains_parametric() ) {
			auto pres = determine_prototype_param_part( ctx, vt.get_part( pt_fun_prototype ), prototype, flags );
			if( !pres.okay() ) {
				return pres;
			}
			use_intr_type = pres.result; 
			
		} else if( !prototype && vt.has<pt_param>()) {
			return cerr::unresolved_param;
			
		} else if( !prototype ) {
			resolve_intr_type_tm ret;
			
			auto fund = vt.get<pt_fun_class>( intr_type::f_invalid );
			switch( fund ) {
				#define FUN_CLASS_TYPE(s) \
					case intr_type::f_##s: ret = determine_intr_type_##s( ctx, vt, flags ); break;
				INTR_TYPE_FUN_CLASSES
				#undef FUN_CLASS_TYPE
				
				default:
					//a missing intrinsic is a param spec
					DEBUG("Missing-instrinsic");
					return cerr::unresolved_param;
			}
			
			if( !ret.okay() ) {
				return resolve_intr_type_t(ret);
			}
			
			for( auto & p : vt.params ) {
				intr_type::param_spec ps;
				ps.name = p.name;
				ret.result->params.push_back( ps );
			}
			
			use_intr_type = ret.result;
		}
		
		return use_intr_type;
	}
}

resolve_t determine( context const & ctx, type_spec const & vt_, unsigned flags ) {
	auto ex = expand( ctx, vt_ );
	if( !ex.okay() ) {
		return resolve_t(ex);
	}
	auto & vt = *ex.result;

	auto use_intr_type = determine_intr_type( ctx, vt, flags );
	if( use_intr_type.error == cerr::unresolved_param && (flags & df_allow_param)) {
		return type_ref::create_spec( vt );
	} else if( !use_intr_type.okay() ) {
		return resolve_t(use_intr_type);
	}
	
	auto tr = determine_extr_type( use_intr_type.result, vt );
		
	//apply constraints
	if( vt.get<pt_tuple>(false) && !tr.is<intr_type_tuple>() ) {
		return cerr::requires_tuple;
	}
	
	auto q = type_ref(tr);
	return q;
}

