//included in ../type_identifier.cpp

/**
	Expand using the current params of the type
*/
resolve_spec_t expand( context const & ctx, type_spec const & vt ) {

	//replace these with their appropriate placeholders (just retains the notation)
	param_map_t retain;
	for( auto & p : vt.params ) {
		type_spec ts;
		ts.push_part<pt_param>( p.name );
		retain[p.name] = type_ref::create_spec(ts); //TODO: p.type constraints?
	}
	
	auto r = expand( ctx, vt, retain );
	if( r.okay() ) {
		r.result->params = vt.params;
	}
	return r;
}

/** 
	Expands into an empty type_spec
*/
resolve_spec_t expand( context const & ctx, type_spec const & from, 
	param_map_t const & params, unsigned flags ) {
	
	auto to = make_shared<type_spec>();
	auto r = expand_into( ctx, *to, from, params, flags | df_single );
	if( !r.okay() ) {
		return resolve_spec_t(r);
	}

	return to;
}

/**
	Resolves all known elements of the `from` spec and merges them into the target `to` spec.
	
	This resolves known parameters and symbols from the scope.
	
	This does not add any defaults.
*/
cerr_base expand_into( context const & ctx, type_spec & to, type_spec const & from,
	param_map_t const & params, unsigned flags ) {
	
	//special allowance to merge value-list into signature parameters
	if( from.get_any_fun_class() == intr_type::f_value_list &&
		to.get_any_fun_class() == intr_type::f_function ) {
		auto fp = make_shared<type_spec>();
		fp->set<pt_fun_class>(intr_type::f_function);
		fp->push_sub( make_shared<type_spec>(from) );
		//DEBUG("param->signature");
		return expand_into( ctx, to, *fp, params, flags );
	}
		
#if SHOW_DEBUG
	DEBUG( "EXPAND_INTO", &to, dump::get(to),  dump::get( from ) );
	//for( auto & mt : params ) {
		//DEBUG( "MT", mt.first, dump::get( mt.second ) );
	//}
#endif
	
	for( auto & from_part : from.get_parts() ) {
		auto copy_params = [&]( type_spec::part & to_part ) -> cerr {
			
			for( size_t i=0; i < from_part.sub.size(); ++i ) {
				//the parameter symbols must be resolved in our namespace
				auto pex = expand( ctx, *from_part.sub[i], params );
				if( !pex.okay() ) {
					return pex.error;
				}
				to_part.sub.push_back( pex.result );
			}
			return cerr::okay;
		};
		
		auto merge_params = [&]( type_spec::part & to_part ) -> cerr {
			
			for( size_t i=0; i < from_part.sub.size(); ++i ) {
				auto pex = expand( ctx, *from_part.sub[i], params );
				if( !pex.okay() ) {
					return pex.error;
				}
				if( i >= to_part.sub.size() ) {
					to_part.sub.push_back( pex.result );
				} else {
					auto q = make_shared<type_spec>(*to_part.sub[i]);
					auto rex = expand_into( ctx, *q, *pex.result, params, flags );
					if( !rex.okay() ) {
						return rex.error;
					}
					to_part.sub[i] = q;
				}
			}
			
			return cerr::okay;
		};
		
		auto merge = [&]( context const & ctx, type_spec const & other ) -> cerr_base {
			if( from_part.sub.size() != other.params.size() ) {
				return cerr::mismatch_params;
			}
			
			param_map_t pm;
			for( size_t i=0; i < other.params.size(); ++i ) {
				//the parameter symbols must be resolved in our namespace
				auto pex = expand( ctx, *from_part.sub[i], params );
				if( !pex.okay() ) {
					return pex.error;
				}
				//DEBUG( "MAP", other.params[i].name, dump::get( pex.result ) );
				pm[ other.params[i].name ] = type_ref::create_spec( *pex.result );
			}
			return expand_into( ctx, to, other, pm, flags );
		};
		
		auto merge_type_ref = [&]( type_ref tr ) -> cerr_base {
			if( tr.is_spec() ) {
				return merge( ctx, tr.spec() );
			} else if( tr.is_extr() ) {
				auto ts = get_spec( ctx, tr );
				STATE_CHECK( ts.okay(), ts.explain_error() );
				return merge( ctx, *ts.result );
			}
			STATE_FAIL( "unsupported-type-ref" );
		};
		
		if( is_singular_part_type( from_part.type ) ) {
			auto hp = to.has_alternate_part( from_part.type );
			if( hp ) {
				if( !(from_part.value == to.get_part(*hp).value) )  {
					if( flags & (df_instantiate | df_single) ) { //TODO: shouldn't need df_instantiate, callers set df_single
						return cerr::redundant_spec;
					}
				}

				merge_params( to.get_part(*hp) );
				continue;
			}
			auto & s = to.set_part( from_part.type, from_part.value );
			auto c = copy_params( s );
			if( c != cerr::okay ) {
				return c;
			}
			continue;
		}
		
		if( from_part.type == pt_param ) {
			auto name_ = boost::get<std::string>(&from_part.value);
			STATE_CHECK( name_ );
			auto pit = params.find( *name_ );
			if( pit == params.end() ) {
				//not mapped at this level so just copy across
				auto s = to.push_part<pt_param>(*name_);
				auto c = copy_params( s );
				if( c != cerr::okay ) {
					return c;
				}
			} else {
				auto c = merge_type_ref( pit->second );
				if( !c.okay() ) {
					return c;
				}
			}
			continue;
		}
		
		STATE_CHECK( from_part.type == pt_symbol );
		
		auto name_parts = boost::get<type_spec_symbol>(&from_part.value);
		STATE_CHECK( name_parts);
		auto &name = name_parts->primary();
		
		
		//locally defined param names take precedence
		auto pt = params.find( name );
		if( pt != params.end() ) {
			auto r = merge_type_ref( pt->second );
			if( !r.okay() ) {
				return r;
			}
			continue;
		}
		
		if( name == "type_infer" ) {
			if( !from_part.expr ) {
				return cerr::missing_expression;
			}
			//cannot modify source since it might be shared, such as in parametric functions
			auto dup = from_part.expr->pre_clone(); 
			//and a new scope to prevent captures from causing clashes (actually, the capturing to higher
			//scopes must be blocked entirely somehow, otherwise this can have side-effects)
			auto func_ctx = create_function_scope(ctx);
			
			auto pex = expression_typer::process_expression(dup, 	
				/*TODO: kludged here*/ context( func_ctx.get_owner(), func_ctx.sc ) );
			if( pex->get_extant() < extant_level::resolved) {
				//TODO: only if param type allowed do the merge
				to.get_parts().push_back(from_part);
				continue;
				//return cerr::unknown_identifier;
			}
			auto ptype = pex->type.extr().if_cast<intr_type_value_list>();
			if( !ptype || ptype->sub.size() != 1 ) {
				return cerr::invalid_expression;
			}
			auto as = get_spec( func_ctx, ptype->sub[0].type );
			STATE_CHECK( as.okay(), as.explain_error() );
			apply_infer( func_ctx, *as.result );
			auto c = merge( func_ctx, *as.result);
			if( !c.okay() ) {
				return c;
			}
			continue;
		}
		
		if( name == "strip_optional" ) {
			if( from_part.sub.size() != 1) {
				return cerr::invalid_type_ref;
			}
			auto sub_spec = expand( ctx, *from_part.sub[0], params, flags );
			if( !sub_spec.okay() ) {
				return sub_spec;
			}
			if( sub_spec.result->has_pt_param() ) {
				DEBUG( "Has Param still" );
				type_spec_symbol tss;
				tss.parts.push_back( "strip_optional" );
				auto & part = to.push_part<pt_symbol>( tss );
				part.sub.push_back( sub_spec.result );
			} else {
			
				type_spec spec_copy = *sub_spec.result;
				spec_copy.clear<pt_optional>();
				spec_copy.clear<pt_cvt_optional>();
			
				DEBUG( "Strip", dump::get(from_part.sub[0]), dump::get(spec_copy) );
				auto ex_sub = expand_into( ctx, to, spec_copy, params, flags );
				if( !ex_sub.okay() ) {
					return ex_sub;
				}
			}
			continue;
		}
		
		auto symbol = ctx.sc->resolve_symbol( name );
		if( !symbol.okay() ) {
			return cerr_base( symbol.error, name );
		}
		
		if( symbol.result.what == scope_symbol::w_type_alias ) {
			auto r = merge( ctx, *symbol.result.type_alias );
			if( !r.okay() ) {
				return r;
			}
			continue;
		}
		
		if( symbol.result.what == scope_symbol::w_fun_type ) {
			if( to.has<pt_fun_class>() ) {
				return cerr::redundant_fundamental;
			}
			
			auto & s = to.set<pt_fun_prototype>( symbol.result.fun_type );
			copy_params( s );
			continue;
		}

		DEBUG( "Type", (int)symbol.result.what );
		return cerr_base( cerr::invalid_symbol_type, name );
	}
	
	if( from.sub.size() && from.get_any_fun_class() == to.get_any_fun_class() ) { 
		//create a new scope to allow subsequence params/returns to reference previous ones
		auto func_ctx = create_function_scope(ctx);
	
		for( size_t i=0; i < from.sub.size(); ++i ) {
			auto r = expand( func_ctx, *from.sub[i].type, params );
			if( !r.okay() ) {
				return r.error;
			}
			
			if( i < to.sub.size() ) {
				if( to.sub[i].type ) {
					auto tc = make_shared<type_spec>( *to.sub[i].type );
					auto mr = expand_into( func_ctx, *tc, *r.result, params, flags );
					if( !mr.okay() ){
						return mr;
					}
					to.sub[i].type = tc;
				} else {
					to.sub[i].type = r.result;
				}

				//it turns out that we can't take the `from` name since it may clash with existing `to` names
				//see: unit_tests/block/parametric/function/arg_replace.lfb
				//if( to.sub[i].name.empty() ) {
				//	to.sub[i].name = from.sub[i].name;
				//}
			} else {
				to.push_sub( from.sub[i].name, r.result );
			}
			
			if( to.sub[i].name.size() ) {
				auto sd = declaration::create();
				sd->symbol = to.sub[i].name;
				sd->spec_type = r.result;
				sd->auto_resolve = true;
				func_ctx.sc->add(sd);
			}
		}
	}
	
	DEBUG( "=>", &to, dump::get(to) );
	return cerr::okay;
}
