//included in ../type_identifier.cpp

resolve_spec_t get_spec( context const & ctx, type_ref assign_ ) {
	
	//DEBUG( "GET_SPEC", dump::get( assign_ ) );
	
	if( assign_.is_spec() ) {
		return make_shared<type_spec>(assign_.spec());
	}
	
	if( !assign_.is_defined() ) {
		return make_shared<type_spec>();
	}
	
	STATE_CHECK( assign_.is_extr() );
	return get_spec(ctx, assign_.extr());
}
	
resolve_spec_t get_spec( context const & ctx, extr_type assign ) {
	auto rt = get_spec( ctx, *assign.intr );
	if( !rt.okay() ) {
		return rt;
	}
	
	auto spec = rt.result;
	spec->set<pt_optional>( assign.is_optional() );
	spec->set<pt_reference>( assign.get_reference() );
	spec->set<pt_lvalue>( assign.is_lvalue() );
	spec->set<pt_lossy>( assign.is_lossy() );
	spec->set<pt_literal>( assign.is_literal() );

	return spec;
}
	
resolve_spec_t get_spec( context const & ctx, intr_type const & intr ) {
	auto spec = make_shared<type_spec>();

	if( intr.instance_of ) {
		auto & pt = spec->set<pt_fun_prototype>( intr.instance_of );
		for( auto & p : intr.applied_params ) {
			auto res = get_spec( ctx, p );
			if( !res.okay() ) {
				return res;
			}
			pt.sub.push_back( res.result );
		}
	} else if( intr.is_whole() ){
		spec->set<pt_fun_prototype>( intr.shared_from_this() );
	} else switch( intr.get_fun_class() ) {
	
		case intr_type::f_rational: {
			auto & rt = intr.cast<intr_type_rational>();
			spec->set<pt_fun_class>( intr_type::f_rational );
			if( rt.integer_bitsize ) {
				spec->set<pt_data_bitsize>( *rt.integer_bitsize );
			}
			spec->set<pt_signed>( rt.integer_signed );
			break;
		}
		case intr_type::f_value_list:
		case intr_type::f_tuple: {
			if( spec->sub.empty() ) {
				auto & comp = intr.cast<intr_type_compound>();
				spec->set<pt_fun_class>( intr.get_fun_class() );
				//each part of the tuple must be inferred (this similar to the above when the spec matches)
				spec->sub.resize( comp.sub.size() );
				for( size_t i=0; i < comp.sub.size(); ++i ) {
					auto lit = get_spec( ctx, comp.sub[i].type );
					if( !lit.okay() ) {
						return lit;
					}
					spec->sub[i].type = lit.result;
					spec->sub[i].name = comp.sub[i].name;
				}
			}
			
			if( intr.is<intr_type_tuple>() ) {
				auto & tt = intr.cast<intr_type_tuple>();
				if( !spec->has<pt_pack>() ) {
					spec->set<pt_pack>( tt.pack );
				}
				if( !spec->has<pt_selectable>() ) {
					spec->set<pt_selectable>( tt.selectable );
				}
			}
			break;
		}
		case intr_type::f_function: {
			auto & func = intr.cast<intr_type_function>();

			spec->set<pt_fun_class>( intr_type::f_function );
			spec->set<pt_function_access>( func.access );
			if( func.context.is_extr() ) {
				spec->set<pt_context>( func.context.extr() );
			}
			size_t sub_size = func.returns ? 2 : 1;
			spec->sub.resize( sub_size );
			//TODO: Perhaps this should not be inferred and they should just be wrapped as throwing?
			if( !func.can_fail )
				spec->set<pt_function_throw>(false);
			
			for( size_t i=0; i < sub_size; ++i ) {
				auto store = get_spec( ctx, extr_type(i ? func.returns : func.args) );
				if( !store.okay() )
					return store;
				spec->sub[i].type = store.result;
			}
			break;
		}
		case intr_type::f_contextual: {
			auto & ct = intr.cast<intr_type_contextual>();
			
			auto & sr = spec->set<pt_fun_class>( intr_type::f_contextual );
			
			auto context = get_spec( ctx, ct.get_context() );
			if( !context.okay() ) {
				return context;
			}
			
			auto unbound = get_spec( ctx, ct.get_unbound() );
			if( !unbound.okay() ) {
				return unbound;
			}
			
			sr.sub.push_back( context.result );
			sr.sub.push_back( unbound.result );
			break;
		}
		case intr_type::f_array: {
			auto & arr = intr.cast<intr_type_array>();

			//return to symbol space here to generate the array spec
			auto store = get_spec( ctx, arr.sub.extr() );
			if( !store.okay() ) {
				return store;
			}
			
			auto & sr = spec->set<pt_fun_class>( intr_type::f_array );
			sr.sub.push_back( store.result );
			
			if( arr.raw ) {
				spec->set<pt_function_access>( intr_type_function::a_raw );
			}
			break;
		}
		
		//TODO: shares a lot with f_array
		case intr_type::f_type: {
			auto & arr = intr.cast<intr_type_type>();

			//return to symbol space here to generate the array spec
			auto store = get_spec( ctx, arr.sub.extr() );
			if( !store.okay() ) {
				return store;
			}
			
			auto & sr = spec->set<pt_fun_class>( intr_type::f_type );
			sr.sub.push_back( store.result );
			break;
		}
			
		case intr_type::f_integer:
		case intr_type::f_binary:
		case intr_type::f_float: {
			auto & num = intr.cast<intr_type_fixed_number>();
			spec->set<pt_fun_class>( intr.get_fun_class() );
			if( !spec->has<pt_data_bitsize>() ) {
				spec->set<pt_data_bitsize>( num.data_bitsize );
			}
			break;
		}
		
		case intr_type::f_boolean:
		case intr_type::f_char: {
			spec->set<pt_fun_class>( intr.get_fun_class() );
			break;
		}
			
		case intr_type::f_service:
		case intr_type::f_module:
		case intr_type::f_class:
		case intr_type::f_closure: {
			STATE_FAIL( "not-expected-here" );
		}
		
		case intr_type::f_compat: {
			auto & ct = intr.cast<intr_type_compat>();
			spec->set<pt_fun_class>( intr_type::f_compat );
			spec->set<pt_compat_type>( ct.type );
			break;
		}
		
	}

	if( intr.params.size() ) {
		for( auto & s : intr.params ) {
			spec->push_param( s.name, shared_ptr<type_spec const>() );
		}
	}
	
	return spec;
}
