#include "lang/type_manager.hpp"
#include "lang/type.hpp"
#include "lang/scope.hpp"
#include "lang/type_spec.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {

namespace {
	/*
		The type parser doesn't understand the type symbols that it is parsing. This is a mapping of predefined symbols to their actual type.
		
		These are `type_spec` instead of `extr_type` as they are not meant to be complete types. An `integer` for example should leave open the option to specify the bitsize, and of course all should be allowed to specify their own extrinsic qualifiers.
		
		`type_identifier` will locate these via the scope. Previously a lookup into this table was hard-coded, which would treat them as globally recognized predefined types. That may be better for efficiency of lookup and could be added back later. It depends on whether these predefined types should be allowed to be overridden in the scope.
	*/
	typedef std::map<std::string, type_spec const> predefined_types_t;
	predefined_types_t predefined_types;
	void predefined_types_add( std::string const & name, type_spec const & vts ) {
		predefined_types.insert( predefined_types_t::value_type{ name, vts } );
	}
		
	struct predefined_types_init
	{
		predefined_types_init() {
			{
				type_spec vts;
				vts.set<pt_fun_class>( intr_type::f_binary );
				vts.set<pt_data_bitsize>(8);
				predefined_types_add("octet", vts);
			}
			{
				type_spec vts;
				vts.set<pt_fun_class>( intr_type::f_binary );
				vts.set<pt_data_bitsize>( platform::abi_ptr_size );
				predefined_types_add("abi_ptr", vts);
			}
			{
				type_spec vts;
				vts.set<pt_fun_class>( intr_type::f_binary );
				vts.set<pt_data_bitsize>( 8 );
				predefined_types_add("abi_char", vts);
			}
			{
				type_spec vts;
				vts.set<pt_pack>( intr_type_tuple::p_abi );
				predefined_types_add("abi_pack", vts);
			}
			{
				type_spec vts;
				vts.set<pt_pack>( intr_type_tuple::p_byte );
				predefined_types_add("byte_pack", vts);
			}
			{
				type_spec vts;
				auto & p = vts.set<pt_fun_class>( intr_type::f_type );
				
				auto st = make_shared<type_spec>();
				st->push_part<pt_param>( "Item" );
				p.sub.push_back( st );
				
				//need to indicate it is a parametric
				type_spec::sub_ref sr;
				sr.name = "Item";
				vts.params.push_back( sr );
				
				predefined_types_add("type", vts);
			}
			{
				type_spec vts;
				vts.set<pt_signed>( true );
				predefined_types_add("signed", vts );
			}
			{
				type_spec vts;
				auto & p = vts.set<pt_fun_class>( intr_type::f_contextual );
				
				auto st = make_shared<type_spec>();
				st->push_part<pt_param>( "Context" );
				p.sub.push_back( st );
				
				st = make_shared<type_spec>();
				st->push_part<pt_param>( "Unbound" );
				p.sub.push_back( st );
				
				type_spec::sub_ref sr;
				sr.name = "Context";
				vts.params.push_back(sr);
				sr.name = "Unbound";
				vts.params.push_back(sr);
				
				predefined_types_add( "contextual", vts );
			}
			{
				type_spec vts;
				vts.set<pt_fun_class>( intr_type::f_compat );
				vts.set<pt_compat_type>( intr_type_compat::abi_var_args );
				predefined_types_add("abi_var_args", vts);
			}
			{
				type_spec vts;
				vts.set<pt_fun_class>( intr_type::f_compat );
				vts.set<pt_compat_type>( intr_type_compat::string_tag );
				predefined_types_add("string_tag", vts);
			}
			
			simple( "rational", intr_type::f_rational );
			simple( "float", intr_type::f_float );
			simple( "integer", intr_type::f_integer );
			simple( "binary", intr_type::f_binary );
			simple( "char", intr_type::f_char );
			simple( "boolean", intr_type::f_boolean );
			
			native_int( "short", platform::abi_short_size );
			native_int( "int", platform::abi_int_size );
			native_int( "long", platform::abi_long_size );
			native_int( "long_long", platform::abi_long_long_size );
			
			float_( "abi_float", platform::abi_float_size );
			float_( "abi_double", platform::abi_double_size );
			float_( "abi_long_double", platform::abi_long_double_size );
		}
		
		void simple( char const * name, intr_type::fun_class_t f ) {
			type_spec vts;
			vts.set<pt_fun_class>( f );
			predefined_types_add(name, vts);
		}
		
		void float_( char const * name, int bit_size ) {
			auto nf = platform::if_get_float( bit_size );
			STATE_CHECK(nf);
			type_spec vts;
			vts.set<pt_fun_class>( intr_type::f_float );
			vts.set<pt_precision_variant>(nf->precision);
			predefined_types_add(name, vts);
		}
		
		void native_int( char const * name_bit, int bit_size ) {
			type_spec vts;
			vts.set<pt_data_bitsize>(bit_size);
			vts.set<pt_fun_class>( intr_type::f_integer );
			char name[64];
			sprintf( name, "abi_%s", name_bit );
			predefined_types_add( name,  vts);
			
			//also the unsigned variant
			sprintf( name, "abi_u%s", name_bit );
			predefined_types_add( name, vts );
		}
		
	} predefined_types_init_;
	
} //eon anon

type_manager::type_manager( shared_ptr<scope> root ) :
	root( root ) {

	for( auto & pt : predefined_types ) {
		auto cur = root->resolve_symbol( pt.first );
		if( cur.okay() ) {
			continue;
		}
		
		root->add_type_alias( pt.first, make_shared<type_spec>(pt.second) );
	}
	
	auto cpa = root->resolve_symbol( "array" );
	if (!cpa.okay()) {
		auto pa = make_shared<intr_type_array>();
		pa->set_distinct_name(  "array" );
		pa->style = intr_type::s_inherent;
		intr_type::param_spec ps;
		ps.name = "Item";
		pa->params.push_back( ps );
		type_spec ts;
		ts.set<pt_param>("Item");
		pa->sub = type_ref::create_spec(ts);
		root->add_fun_type( pa );
		primordial_array = pa;
	} else {
		STATE_CHECK( cpa.result.what == scope_symbol::w_fun_type &&
			cpa.result.fun_type->is<intr_type_array>() );
		primordial_array = std::dynamic_pointer_cast<intr_type_array const>(cpa.result.fun_type);
	}
	
	auto cpar = root->resolve_symbol( "raw_array" );
	if (!cpar.okay()) {
		auto pa = make_shared<intr_type_array>();
		pa->set_distinct_name(  "raw_array" );
		pa->style = intr_type::s_inherent;
		pa->raw = true;
		intr_type::param_spec ps;
		ps.name = "Item";
		pa->params.push_back( ps );
		type_spec ts;
		ts.set<pt_param>("Item");
		pa->sub = type_ref::create_spec(ts);
		root->add_fun_type( pa );
		primordial_raw_array = pa;
	} else {
		STATE_CHECK( cpa.result.what == scope_symbol::w_fun_type &&
			cpar.result.fun_type->is<intr_type_array>() );
		primordial_raw_array = std::dynamic_pointer_cast<intr_type_array const>(cpar.result.fun_type);
	}

	//TODO: part of predefined, grab from there
	auto tt = make_shared<intr_type_compat>();
	tt->type = intr_type_compat::string_tag;
	string_tag = tt;
	
// 	auto cpt = root->resolve_symbol( "⁑primordial_type" );
// 	if (!cpt.okay()) {
// 		auto pa = make_shared<intr_type_type>();
// 		pa->set_distinct_name(  "⁑primordial_type" );
// 		intr_type::param_spec ps;
// 		ps.name = "Item";
// 		pa->params.push_back( ps );
// 		root->add_fun_type( pa );
// 		primordial_type = pa;
// 	} else {
// 		STATE_CHECK( cpt.result.what == scope_symbol::w_fun_type &&
// 			cpt.result.fun_type->is<intr_type_type>() );
// 		primordial_type = std::dynamic_pointer_cast<intr_type_type const>(cpt.result.fun_type);
// 	}
}
	
shared_ptr<intr_type_tuple const> type_manager::get_integer_range() {
	if (!integer_range) {
		auto cur = root->resolve_symbol( "⁑integer_range" );
		STATE_CHECK( cur.okay(), cur.explain_error(), error::tag_error_symbol( "⁑integer_range" ) );
		STATE_CHECK( cur.result.what == scope_symbol::w_fun_type &&
			cur.result.fun_type->is<intr_type_tuple>() );
		integer_range = std::dynamic_pointer_cast<intr_type_tuple const>(cur.result.fun_type);
	}
	return integer_range;
}

shared_ptr<intr_type_array const> type_manager::get_primordial_array() {
	return primordial_array;
}

shared_ptr<intr_type_type const> type_manager::get_primordial_type() {
	return primordial_type;
}

shared_ptr<intr_type_compat const> type_manager::get_string_tag() {
	return string_tag;
}

shared_ptr<intr_type_class const> type_manager::get_error_info() {
	if( !error_info ) {
		auto eir = root->resolve_symbol( "⁑error_info" );
		STATE_CHECK( eir.okay(), eir.explain_error(), error::tag_error_symbol( "⁑error_info" ) );
		STATE_CHECK( eir.result.what == scope_symbol::w_fun_type &&
			eir.result.fun_type->is<intr_type_class>() );
		error_info = std::dynamic_pointer_cast<intr_type_class const>(eir.result.fun_type);
	}
	return error_info;
}

} //eon leaf

