#include "lang/type.hpp"
#include "lang/serial.hpp"
#include "lang/type_spec.hpp"

namespace leaf {

type_ref::type_ref( serial_in & si ) {
	read( si );
}

/*static*/ type_ref type_ref::create_spec( type_spec const & ts ) {
	type_ref tr;
	tr.type_ = t_spec;
	tr.spec_ = make_shared<type_spec>( ts );
	return tr;
}

bool type_ref::is_fixated_impl( util::validation_context & vctx ) const {
	switch( type_ ) {
		case t_undefined:
			return false;
			
		case t_extr:
			return extr_ && extr_->is_fixated_impl( vctx );
		
		case t_spec:
			return false;
	}
	
	STATE_FAIL( "corrupt-type-ref" );
	return false;
}

extant_level type_ref::get_extant_impl( util::validation_context & vctx ) const {
	switch( type_ ) {
		case t_undefined:
			return extant_level::incomplete;
			
		case t_extr:
			return extr_ ? extr_->get_extant_impl( vctx ) : extant_level::incomplete;
			
		case t_spec:
			return spec_ ? extant_level::parametric : extant_level::incomplete;
	}
	
	STATE_FAIL( "corrupt-type-ref" );
	return extant_level::incomplete;
}

bool type_ref::contains_parametric_impl( util::validation_context & ctx ) const {
	switch( type_ ) {
		case t_undefined: 
			return false;
			
		case t_extr:
			return extr_ && extr_->contains_parametric_impl(ctx);
			
		case t_spec:
			return true;
	}
	
	STATE_FAIL( "corrupt-type-ref" );
	return false;
}

void type_ref::validate(util::validation_context & vctx) const {
	switch( type_ ) {
		case t_undefined:
			STATE_FAIL( "undefined-type-ref" );
			
		case t_extr:
			STATE_CHECK( extr_ );
			extr_->validate( vctx );
			break;
			
		case t_spec:
			STATE_CHECK( spec_ );
			break;
	}
}

int type_ref::compare( type_ref const& o, bool logical ) const {
	util::validation_context vctx;
	return compare_impl( o, logical, vctx );
}

int type_ref::compare_impl( type_ref const & o, bool logical, util::validation_context & vctx ) const {
	if( type_ != o.type_ ) {
		return type_ < o.type_ ? -1 : 1;
	}
	switch( type_ ) {
		case t_undefined:
			return 0;
			
		case t_extr:
			if( !!extr_ != !!o.extr_ ) {
				return !!extr_ < !!o.extr_ ? -1 : 1;
			}
			if( extr_ ) {
				return extr_->compare_impl( o.extr(), logical, vctx );
			}
			return 0;
			
		case t_spec:
			return 0; //TODO: param_spec_, or somehow fail?
	}
	
	STATE_FAIL( "corrupt-type-def" );
	return 0;
}

void type_ref::write( serial_out & so ) const {
	so.write( type_ );
	bool e = !!extr_;
	so.write( e );
	if( extr_ ) {
		extr_->write( so );
	}
	
	e = !!spec_;
	so.write(e);
	if( spec_ ) {
		spec_->write( so );
	}
}

void type_ref::read( serial_in & si ) {
	extr_.reset();
	
	si.read( type_ );
	bool e;
	si.read( e );
	if( e ) {
		extr_ = extr_type( si );
	}
	
	si.read( e );
	if( e ) { 
		spec_ = make_shared<type_spec>( si );
	}
}

std::string type_ref::get_qualified_id() const {
	switch( type_ ) {
		case t_undefined:
			STATE_FAIL( "undefined-type-ref" );
			
		case t_spec:
			return "$U";
			//TODO: must include type_spec...?
			
		case t_extr:
			PRE_CHECK( extr_ );
			return extr_->get_qualified_id();
	}
	
	STATE_FAIL( "corrupt-type-ref" );
}

} //eon leaf