#include "lang/type_spec.hpp"
#include "lang/serial.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace leaf {

std::string name_of( part_type_t pt ) {
	switch(pt) {
		#define TYPE_SPEC_PART(s) case pt_##s: return #s ;
		TYPE_SPEC_PARTS
		#undef TYPE_SPEC_PART
	}
	return "corrupted";
}

bool type_spec_symbol::operator==(type_spec_symbol const &o) const {
	if( parts.size() != o.parts.size() ) {
		return false;
	}
	
	for( size_t i=0; i < parts.size(); ++i ) {
		if( parts[i] != o.parts[i]) {
			return false;
		}
	}
	
	return true;
}

boost::optional<part_type_t> type_spec::has_alternate_part( part_type_t p ) const {
	//bitsize
	if( p == pt_precision_variant ) {
		if( has<pt_data_bitsize>() ) {
			return pt_data_bitsize;
		}
	}
	if( p == pt_data_bitsize ) {
		if( has<pt_precision_variant>() ) {
			return pt_precision_variant;
		}
	}
	
	//intrinsic
	if( p == pt_fun_class ) {
		if( has<pt_fun_prototype>() ) {
			return pt_fun_prototype;
		}
	}
	if( p == pt_fun_prototype ) {
		if( has<pt_fun_class>() ) {
			return pt_fun_class;
		}
	}
	
	//context
	if( p == pt_context_name ) {
		if( has<pt_context>() ) {
			return pt_context;
		}
	}
	if( p == pt_context ) {
		if( has<pt_context_name>() ) {
			return pt_context_name;
		}
	}
	
	if( has_part(p) ) {
		return p;
	}
	return boost::optional<part_type_t>();
}

namespace {
	template<typename T>
	void write_part( serial_out & so, type_spec::part const & p ) {
		so.write( boost::get<T>(p.value ) );
	}
	
	template<>
	void write_part<shared_ptr<intr_type const>>( serial_out & so, type_spec::part const & p ) {
		so.write_ref( boost::get<shared_ptr<intr_type const>>(p.value) );
	}
	
	template<>
	void write_part<extr_type>( serial_out & so, type_spec::part const & p ) {
		boost::get<extr_type>(p.value).write(so);
	}
	
	template<>
	void write_part<type_spec_symbol>( serial_out & so, type_spec::part const & p ) {
		so.write( boost::get<type_spec_symbol>(p.value).parts );
	}
	
	template<typename T>
	void read_part( serial_in & si, type_spec::part & p ) {
		T val;
		si.read(val);
		p.value = val;
	}
	
	template<>
	void read_part<shared_ptr<intr_type const>>( serial_in & si, type_spec::part & p ) {
		p.value = si.read_intr_type_ref();
	}
	
	template<>
	void read_part<extr_type>( serial_in & si, type_spec::part & p ) {
		extr_type val(si);
		p.value = val;
	}
	
	template<>
	void read_part<type_spec_symbol>( serial_in & si, type_spec::part & p ) {
		type_spec_symbol tss;
		si.read(tss.parts);
		p.value = tss;
	}
	
	void write_sub_ref( serial_out & so, type_spec::sub_ref const& sr ) {
		so.write(sr.name);
		sr.type->write(so);
	}
	
	type_spec::sub_ref read_sub_ref( serial_in & si ) {
		type_spec::sub_ref sr;
		si.read(sr.name);
		auto ts = make_shared<type_spec>();
		ts->read(si);
		sr.type = ts;
		return sr;
	}
}

void type_spec::write( serial_out & so ) const {
	size_t len = parts.size();
	so.write(len);
	
	for( auto & p : parts ) {
		so.write( p.type );
		switch( p.type ) {
			#define TYPE_SPEC_PART(s) \
				case pt_##s: \
					write_part<typename part_type_descriptor<pt_##s>::type>( so, p ); \
					break;
			TYPE_SPEC_PARTS
			#undef TYPE_SPEC_PART
		}
		
		size_t slen = p.sub.size();
		so.write(slen);
		for( auto & ts : p.sub ) {
			ts->write(so);
		}
		
		// "node_expr" excluded on purpose (it's just a link for debugging...?)
		
		so.write_ref( p.expr );
		DEBUG( "TSWrite", reinterpret_cast<intptr_t>(p.expr.get() ));
	}
	
	len = sub.size();
	so.write(len);
	for( auto & s : sub ) {
		write_sub_ref( so, s );
	}
	
	len = params.size();
	so.write(len);
	for( auto & s : params ) {
		write_sub_ref( so, s );
	}
}

void type_spec::read( serial_in & si ) {
	size_t len;
	si.read( len );

	for( size_t i=0; i < len; ++i ) {
		part p;
		si.read( p.type );
		switch( p.type ) {
			#define TYPE_SPEC_PART(s) \
				case pt_##s: \
					read_part<typename part_type_descriptor<pt_##s>::type>( si, p ); \
					break;
			TYPE_SPEC_PARTS
			#undef TYPE_SPEC_PART
		}
		
		size_t slen;
		si.read(slen);
		for( size_t j=0; j < slen; ++j) {
			auto ts = make_shared<type_spec>();
			ts->read(si);
			p.sub.push_back(ts);
		}
		p.expr = si.read_expression_ref();
		
		parts.push_back(p);
	}
	
	si.read(len);
	for( size_t i=0; i < len; ++i ) {
		sub.push_back( read_sub_ref( si ) );
	}
	
	si.read(len);
	for( size_t i=0; i < len; ++i ) {
		params.push_back( read_sub_ref( si ) );
	}
}

bool type_spec::has_pt_param() const {
	if( has<pt_param>() ) { 
		return true;
	}
	
	for( auto & s : sub ) {
		if( s.type->has_pt_param() ) {
			return true;
		}
	}
	
	return false;
}

} //eon leaf