#include "lang/class_typer.hpp"
#include "lang/conversion_applicator.hpp"
#include "lang/expression_visitor.hpp"
#include "lang/language.hpp"
#include "lang/module_loader.hpp"
#include "lang/serial.hpp"
#include "lang/statement_visitor.hpp"
#include "lang/type_converter.hpp"
#include "lang/typer.hpp"
#include "lang/scope.hpp"
#include "lang/module.hpp"
#include "lang/type_manager.hpp"
#include "lang/expression_typer/common.hpp"

#include "lang/dump.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

#include <fstream>
#include <iostream>

#define RETURN_IF_UNRESOLVED( obj, stmt ) \
	if( !(obj).is_resolved() ) { \
		STMT_CHECK( *stmt, !no_incomplete, "unresolved-typing" ); \
		return stmt; \
	}

#define RETURN_IF_INCOMPLETE( obj, stmt ) \
	if( !(obj).is_complete() ) { \
		STMT_CHECK( *stmt, !no_incomplete, "incomplete-typing" ); \
		return stmt; \
	}
	
namespace leaf
{

///////////////////////////////////////////////////////////
// job_statement_block
bool typer::job_statement_block::process( typer * ty ) {
	auto ctx = context( ty, sb->symbols );
	auto ret = ty->proc_statement_block( sb, ctx );
	return ret;
}

///////////////////////////////////////////////////////////
// typer
typer::typer( boost::optional<std::string> module_name, shared_ptr<module_loader> mloader ) :
	mod( new module() ),
	mloader( mloader ) {
	
	mod->name = module_name;
	mod->root = scope::create( scope_style_t::declarative );
	if( module_name ) {
		mod->global = scope::create( scope_style_t::declarative );
		mod->global->set_base( mod->root, scope::bt_namespace );
		mod->global->set_name( *module_name );
	} else {
		mod->global = mod->root;
	}
	
	init();
}

typer::typer( shared_ptr<scope> root_scope ) :
	mod( new module() ) {
	
	mod->root = root_scope;
	mod->global = root_scope;
	
	init();
}

void typer::init() {
	types = make_shared<type_manager>(mod->root);
	
	no_incomplete = false;
	
	mod->static_init_block = make_shared<statement_block>(scope_style_t::imperative);
	//it is a proper scope
	mod->static_init_block->symbols->set_base( mod->global, scope::bt_parent );
}

shared_ptr<statement_block> typer::add_root_block( statement_vector_t block ) {
	auto b = make_shared<statement_block>( scope_style_t::declarative );
	b->symbols = mod->root;
	//TODO:  this should be a static_init block, but not relevant yet if only imports are declared here
	b->list = block;
	
	mod->blocks.push_back( b );
	scan_symbols( b );
	pending_jobs.push_back( make_shared<job_statement_block>(b) );
	return b;
}

shared_ptr<statement_block> typer::add_global_block( statement_vector_t block ) {
	auto b = make_shared<statement_block>( scope_style_t::declarative );
	b->symbols = mod->global;
	b->symbols->set_init_block( mod->static_init_block );
	b->list = block;
	
	mod->blocks.push_back( b );
	scan_symbols( b );
	pending_jobs.push_back( make_shared<job_statement_block>(b) );
	return b;
}

shared_ptr<statement_block> typer::add_dynamic_block( statement_vector_t block ) {
	auto b = make_shared<statement_block>( scope_style_t::dynamic );
	b->list = block;
	b->symbols->set_base( mod->global, scope::bt_namespace );
	
	mod->blocks.push_back( b );
	pending_jobs.push_back( make_shared<job_statement_block>(b) );
	return b;
}

void typer::full_complete_type() {
	try {
		complete_type();
	} catch( recoverable_error & re ) {
		re << error_typer(true);
		throw;
	}
}

shared_ptr<expression> typer::full_type( shared_ptr<expression> expr, unsigned flags ) {
	try {
		auto ctx = context(this,mod->global);
		auto result = ctx.proc_expression( expr );
		//rough loop to resolve new expression results
		while( !result->is_resolved() && result != expr ) {
			result = ctx.proc_expression( result );
		}
		STMT_CHECK( *expr, result->is_resolved() );
		complete_type();
		
		if( flags & tef_convert_inferred ) {
			auto it = type_identifier::infer( ctx, shared_ptr<type_spec>(), result->type, false );
			STMT_CHECK( *expr, it.okay(), it.explain_error() );
			auto cost = type_converter(ctx).convert( it.result, result->type );
			result = ctx.applicator().apply( cost, result );
		}
		
		return result;
	} catch( recoverable_error & re ) {
		re << error_typer(true);
		throw;
	}
}

/**
	In a declarative scope the statements and blocks have no particular order. This allows
	forward referencing of both variables and types. Rather than do a fixed 2-pass approach
	we just keep looping so long as progress is being made.
*/
void typer::complete_type() {
	auto proc_jobs = [&,this]() {
		job_vector_t jobs;
		jobs.swap( pending_jobs );

		for( auto & j : jobs ) {
			auto r = j->process( this );
			//DEBUG( "JOB", typeid( *j ).name(), r );
			if(!r) {
				pending_jobs.push_back( j );
			} else {
				progress_count++;
			}
		}
	};

	auto loop = [&,this]() {
		while( pending_jobs.size() ) {
			progress_count = 0;
			completed_statements = 0;
			auto old_pj = pending_jobs.size();
			proc_jobs();
			
			//if nothing happened let it run once more to report errors
			if( !completed_statements && !progress_count && old_pj == pending_jobs.size() ) {
				//TODO: need an error reporter to always report this
				std::cout << "\nNo progress being made" << std::endl;
				no_incomplete = true;
				proc_jobs();
				STATE_FAIL( "no progress and no error?!" );
				return false; //not reached now...
			}
		}
		return true;
	};
	
	if( loop() ) {
		//once everything is done finish up by processing the static_init_block
		auto job = make_shared<job_statement_block>(mod->static_init_block);
		pending_jobs.push_back( job );
		loop();
	}
}

shared_ptr<statement> typer::proc_assign_statement( shared_ptr<assign_statement> as, 
	context ctx )
{
	STMT_CHECK( *as, !ctx.sc->is_declarative() );
	as->left = ctx.proc_expression( as->left );
	as->right = ctx.proc_expression( as->right );
	//DEBUG( "LT", dump::get( as->left->type ) );
   	//DEBUG( "RT", dump::get( as->right->type ) );
	RETURN_IF_UNRESOLVED( *as->left, as )
	RETURN_IF_UNRESOLVED( *as->right, as )
		
	auto tc = ctx.converter();
	
	//a reference init means the left must resolve to a shared lvalue first (which allows optionals to work)
	type_converter::unary_convert uhow;
	switch( as->how ) {
		case assign_statement::h_value:
		case assign_statement::h_default:
			uhow = &type_converter::convert_any_lvalue;
			break;
		case assign_statement::h_init_refer:
			uhow = &type_converter::convert_any_refer;
			break;
	}
	as->left = ctx.convert_expression( uhow, as->left );
	
	type_converter::binary_convert bhow;
	switch( as->how ) {
		case assign_statement::h_value:
			bhow = &type_converter::convert_to_lvalue;
			break;
		case assign_statement::h_init_refer:
			bhow = &type_converter::convert_to_refer;
			break;
		case assign_statement::h_default:
			STMT_CHECK( *as, as->left->type.extr().is_optional() );
			bhow = &type_converter::convert_to_lvalue;
			break;
	}
	STMT_CHECK( *as, as->right->type.extr().allow_copy(), "no-copy-allowed" );
	STMT_CHECK( *as, as->left->type.extr().allow_copy(), "no-copy-allowed" );
	//DEBUG( "ASSIGN", dump::get(as->left->type), dump::get(as->right->type) );
	as->right = ctx.convert_expression( bhow, as->left, as->right );
	as->local_complete = true;
	return as;
}

shared_ptr<statement> typer::proc_var_statement( shared_ptr<var_statement> vs, 
	context ctx ) {
	STMT_CHECK( *vs, vs->decl );
	vs->decl->source_statement = vs;

	//parametric variables don't really exist...
	bool parametric = false;
	bool literal = vs->decl->spec_type
		&& vs->decl->spec_type->get<pt_literal>( false );

	DEBUG( "VAR", vs->decl->symbol, ctx.sc.get(), literal ? "LITERAL"  : "" );
	//where to initialize (null is here)
	shared_ptr<statement_block> init_block;
	if( vs->decl->lifetime == declaration::l_global ) {
		init_block = mod->static_init_block;
	} else if( ctx.sc->has_init_block() ) {
		init_block = ctx.sc->get_init_block();
	}
	
	//an init_block needs to be evaluated in a different context
	auto value_ctx = ctx.sc->has_init_block() ? ctx.derive( ctx.sc->get_init_block()->symbols ) : ctx;
	
	if( vs->value ) {
		vs->value = value_ctx.proc_expression( vs->value );
	
		DEBUG( "value-type", dump::get(vs->value->type));
		//the type may not be complete yet, but we can work with it nonetheless
		if( vs->value->type.get_extant() < extant_level::input ) {
			STMT_CHECK( *vs, !no_incomplete, "incomplete-typing" );
			//DEBUG( "VAR-INCOMPLETE", vs->decl->symbol, dump::get(vs->value) );
			return vs;
		}
		
		if( literal ) {
			STMT_CHECK( *vs, !vs->init_refer_value );
		
			//literals constrain their type rather than infer it
			auto it = type_identifier::constrain( ctx, vs->decl->spec_type, vs->value->type );
			STMT_CHECK( *vs, it.okay(), it.explain_error() );
			vs->decl->type = type_ref(it.result);
			
			parametric = vs->decl->type.contains_parametric();
		} else {
			auto it = type_identifier::infer( ctx, vs->decl->spec_type, vs->value->type, vs->init_refer_value );
			if( it.unknown_identifier() ) {
				STMT_CHECK( *vs, !no_incomplete, it.explain_error() );
				return vs;
			}
			STMT_CHECK( *vs, it.okay(), it.explain_error() );
			//the value need not yet be complete, but the inferred type must be
			RETURN_IF_INCOMPLETE( it.result, vs);
			
			vs->decl->type = type_ref(it.result);
			//literal can be inferred, but only a "literal" can retain it
			vs->decl->type.extr().set_literal( false );
		}
		
		DEBUG( "OUT", vs->decl->symbol, dump::get( vs->decl->type ), dump::get( vs->value->type ),
			vs->value->type.is_complete(), dump::get(*vs->value) );
		vs->value->bound_decl = vs->decl;
	} else {
		auto rt = type_identifier::determine( ctx, vs->decl->spec_type );
		if( rt.unknown_identifier() ) {
			STMT_CHECK( *vs, !no_incomplete, rt.explain_error() );
			return vs;
		}
		STMT_CHECK( *vs, rt.okay(), rt.explain_error() );
		vs->decl->type = type_ref(rt.result);
	}
	
	STMT_CHECK( *vs, literal || vs->decl->type.is_fixated() );
	
	if( vs->value ) {
		if( !parametric ) {
			auto tc = value_ctx.converter();
			type_converter::convert_t cost = 
				vs->init_refer_value 
					? tc.convert_to_refer( vs->decl->type, vs->value->type )
					: tc.convert( vs->decl->type, vs->value->type );
			if (cost.error == cerr::incomplete) {
				STMT_CHECK( *vs, !no_incomplete, cost.explain_error() );
				return vs;
			}
			DEBUG("var conversion", dump::get( cost.conversions ),
				dump::get( vs->decl->type ), dump::get( vs->value->type ) );
			vs->value = ctx.applicator().apply( cost, vs->value );
		}
		
		if( literal ) {
			if( vs->value->type.is_complete() ) {
				STMT_CHECK( *vs, vs->value->type.extr().is_literal() );
			}
			vs->decl->literal_value = vs->value;
		}
		
		if( !parametric ) {
			STMT_CHECK( *vs, vs->value->is_resolved() || !no_incomplete, "incomplete" );
			RETURN_IF_UNRESOLVED(*vs->value, vs);
		}
		
	}

	
	if( vs->init_refer_value )
		STMT_CHECK( *vs, vs->decl->type.extr().is_shared() );

	//the symbol is not available until after the var_statement
	auto init = make_shared<init_statement>();
	init->src_loc = vs->src_loc;
	if( vs->sref ) {
		init->sref = vs->sref;
	} else {
		//TODO: same code as in symbol_scanner
		switch( vs->decl->form )
		{
			case declaration::f_single:
				init->sref = ctx.sc->add( vs->decl );
				break;
			
			case declaration::f_multi:
				init->sref = ctx.sc->add( vs->decl );
				break;
				
			case declaration::f_arg:
			case declaration::f_return:
			case declaration::f_capture:
				STMT_FAIL(*vs);
		}
	}

	//fixated values get initialized now, non-fixated are only in scope
	shared_ptr<statement> ret;
	if( vs->decl->type.is_fixated() ) {
		init->value = vs->value;
		init->init_refer_value = vs->init_refer_value;
		
		if( init_block ) {
			if( init_block->symbols->no_init_release ) {
				init->no_release = true;
			}
			init->sref = init_block->symbols->capture_var( init->sref );
			init_block->list.push_back( init );
			
			ret = make_shared<noop_statement>();
			ret->local_complete = true;
		} else {
			ret = proc_init_statement( init, ctx );
		}
	} else { 
		ret = make_shared<noop_statement>();
		ret->local_complete = true;
	}
	
	//export handling
	if( vs->decl->is_lifetime_global() && 
		ctx.sc->is_class_root() && ctx.sc->get_class_scope_context().extr().intr->is_export ) {
		vs->decl->is_export = true;
	}
	if( vs->decl->is_export ) {
		mod->var_exports.push_back( vs->decl );
	}

	//track import library
	if( vs->decl->import_module.size() ) {
		mod->libraries.insert( vs->decl->import_module );
	}
	//DEBUG( "END_VAR", vs->decl->symbol, dump::get( vs->decl->type ), 	(int)vs->decl->type.get_extant(), ret->is_complete() );
	return ret;
}

/**
	init is generated by the typer itself in var_statement. This function assumes that is how it
	was created. This separation was only done to deal with incomplete recursive types.
*/
shared_ptr<statement> typer::proc_init_statement( shared_ptr<init_statement> is, context ctx )
{
	STMT_CHECK( *is, is->sref );
	STMT_CHECK( *is, is->sref->type.is_extr() );
	auto type = is->sref->type;
	
	//this avoids class construction since optionals are empty by default
	if( type.extr().is_optional() ) { 
		//nop
	} else if( type.extr().is<intr_type_classlike>() && !is->value && !is->code ) {
		auto fc = expression_typer::create_default_ctor_call( *is, ctx, is->sref->type );
		if( !fc ) {
			return is;
		}
		
		//add call to the constructor
		auto es = make_shared<expression_statement>();
		es->expr = fc;
		
		is->code = es;
		is->undefined = true;

		//initialize directly on the variable's context
		auto sym = is->sref->create_expr_symbol();
		fc->context = sym;
		
		fc->args = make_shared<expr_value_list>();
	}
	
	if( is->value ) {
		STMT_CHECK( *is, type.extr().allow_copy(), "no-copy-allowed" );
		is->value = ctx.proc_expression( is->value );
		RETURN_IF_UNRESOLVED( *is->value, is );
	} 
	
	if( is->code ) {
		is->code = proc_statement( is->code, ctx );
		RETURN_IF_UNRESOLVED( *is->code, is );
	}

	is->local_complete = true;
	return is;
}

shared_ptr<statement> typer::proc_statement( shared_ptr<statement> s, context ctx )
{
	if( s->is_resolved() ) {
		return s;
	}
	
	auto was_level = s->get_extant();
	
	//DEBUG( "STATEMENT: ", typeid( *s.get() ).name(), dump::get(s) );
	shared_ptr<statement> result;
	switch( s->statement_form() )
	{
		#define STATEMENT_FORM(type) \
			case sf_##type##_statement: \
				result = proc_##type##_statement( std::static_pointer_cast<type##_statement>(s), ctx ); \
				break;
		STATEMENT_FORMS
		#undef STM
	}
	STMT_CHECK( *s, result );
	
	if( was_level != result->get_extant() ) {
		completed_statements++;
	}
	if( s.get() != result.get() ) {
		//DEBUG( "STATEMENT: ", typeid( *s.get() ).name(), dump::get(s) );
		//DEBUG( "Altered", typeid( *result.get() ).name(), dump::get(result) );
	}
	return result;
}

shared_ptr<statement> typer::proc_do_statement( shared_ptr<do_statement> ds, context ctx )
{
	STMT_CHECK( *ds, !ctx.sc->is_declarative() );
	static extr_type cond_type( make_shared<intr_type_boolean>() );
	
	for( size_t i=0; i < ds->expr_list.size(); ++i )
	{
		auto & ex = ds->expr_list[i];
		if( ex.cond ) {
			ex.cond = ctx.proc_expression( ex.cond );
			RETURN_IF_UNRESOLVED( *ex.cond, ds );
			ex.cond = ctx.convert_expression( cond_type, ex.cond );
		}
		
		ex.block->symbols->set_base( ctx.sc, scope::bt_local );
		if( !proc_statement_block( ex.block ) )
			return ds; //incomplete
	}
	
	ds->local_complete = true;
	return ds;
}

bool typer::proc_error_block( shared_ptr<statement> st, error_block & block, context ctx ) {
	if( !block.error_info ) {
		STMT_CHECK( *st, block.on_error );
		block.on_error->symbols->set_base( ctx.sc, scope::bt_local );
			
		auto decl = declaration::create();
		decl->form = declaration::f_single;
		decl->symbol = "current_fail";
		
		auto shared_error = extr_type(ctx.types().get_error_info());
		shared_error.set_ref_shared();
		decl->type = shared_error;
		
		block.on_error->symbols->add( decl );
		block.error_info = decl;
	}
	
	return proc_statement_block( block.on_error );
}

shared_ptr<statement> typer::proc_error_statement( shared_ptr<error_statement> es, context ctx ) {
	es->inner = proc_statement( es->inner, ctx );
	
	if( !proc_error_block( es, es->error, ctx ) ) {
		return es;
	}
	
	if( es->inner->is_resolved() ) {
		STMT_CHECK( *es, !(es->inner->get_flow() & (flow_return | flow_loop_break | flow_loop_next)), 
			"terminator-handled" );
		if( !es->inner->is_skippable() && !es->safe_skip ) {
			STMT_CHECK(*es, es->error.on_error->is_terminator(), "on-fail-must-terminate" );
		}
	
		es->local_complete = true;
	}
	return es;
}

shared_ptr<statement> typer::proc_defer_statement( shared_ptr<defer_statement> ds, context ctx ) {
	STMT_CHECK( *ds, !ctx.sc->is_declarative() );
	STATE_CHECK(ds->error.on_error);
	ds->error.on_error->symbols->set_base( ctx.sc, scope::bt_local );
	
	if( ds->error_only ) {
		if( proc_error_block( ds, ds->error, ctx ) ) {
			ds->local_complete = true;
		}
	} else if( proc_statement_block( ds->error.on_error ) ) {
			ds->local_complete = true;
	}
	return ds;
}

shared_ptr<statement> typer::proc_fail_statement( shared_ptr<fail_statement> fs, context ctx ) {
	STMT_CHECK( *fs, !ctx.sc->is_declarative() );
	
	if( fs->tag ) {
		fs->tag = ctx.proc_expression( fs->tag );
		RETURN_IF_UNRESOLVED(*fs->tag, fs);
		
		fs->tag = ctx.convert_expression( extr_type(types->get_string_tag()), fs->tag );
	}
	
	fs->local_complete = true;
	return fs;
}

bool typer::proc_statement_block( shared_ptr<statement_block> block ) {
	auto ctx = context( this, block->symbols );
	return proc_statement_block( block, ctx );
}

/**
	Declarative statement blocks  can be processed piece-wise. Any incomplete statement
	is simply moved to the end and tried again later. After enough passes this should result
	in an appropriate ordering of complete statements.
	
	Sequence blocks can only be processed in order. Any incomplete statement stops the
	processing and pushes the whole block on the pending list.
*/
bool typer::proc_statement_block( shared_ptr<statement_block> block, context ctx )
{
	if( ctx.sc->is_declarative() ) {
		std::vector<shared_ptr<statement>> done, after;
		
		for( size_t i=0; i < block->list.size(); ++i ) {	
			auto st = proc_statement( block->list[i], ctx );
			if( st->is_resolved() ) {
				done.push_back(st);
			} else {
				after.push_back(st);
			}
		}
		
		if( after.size() ) {
			done.insert( done.end(), after.begin(), after.end() );
		}
		block->list.swap( done );
		if( after.size() ) {
			return false;
		}
		
	} else {
		//TODO: catch "resume" being used in non-handler
		for( size_t i=0; i < block->list.size(); ++i ) {	
			if( i && !(block->list[i-1]->get_flow() & flow_sequence) ) 
				STMT_FAIL( *block->list[i], "code-post-terminator" );
			block->list[i] = proc_statement( block->list[i], ctx );
			if( !block->list[i]->is_resolved() ) {
				return false;
			}
		}
		block->update_flow();
	}
	return true;
}

shared_ptr<statement> typer::proc_typedef_statement( shared_ptr<typedef_statement> ts, context ctx )
{
	if( ts->local_complete )
		return ts;
		
	if( ts->what == typedef_statement::w_prototype ) {
	
		if( !ts->proc_tuple ) {
			//introduce incomplete type first to allow recursion
			auto inc_tuple = make_shared<intr_type_tuple>();
			inc_tuple->local_complete = false;
			inc_tuple->set_distinct_name( ts->name );
			inc_tuple->style = ts->is_inherent ? intr_type::s_inherent : intr_type::s_distinct;
			inc_tuple->is_export = ts->is_export;
			ctx.sc->add_fun_type( inc_tuple );
			ts->proc_tuple = inc_tuple;
			DEBUG( "AddTypedef", ts->name );
		}

		if( ts->spec_type ) {
			STMT_CHECK( *ts, !ts->proc_tuple->local_complete );
			
			//'type' statements create new classes from plain tuples
			STMT_CHECK( *ts, ts->spec_type->extrinsic_free() );
			STMT_CHECK( *ts, ts->spec_type->get<pt_fun_class>( intr_type::f_invalid ) == intr_type::f_tuple );

			auto nt = type_identifier::determine( ctx, ts->spec_type, type_identifier::df_allow_param );
			if( nt.okay() ) {
				STMT_CHECK( *ts, !nt.result.extr().intr->is_whole() );
				auto nt_tuple = nt.result.extr().if_cast<intr_type_tuple>();
				STMT_CHECK( *ts, nt_tuple );
				ts->proc_tuple->complete_with( *nt_tuple );
			} else if( nt.unknown_identifier() ) {
				STMT_CHECK( *ts, !no_incomplete, nt.explain_error() );
				return ts;
			} else {
				STMT_FAIL( *ts, nt.explain_error() );
			}
		} else {
			//not necessary, but has cleaner dump
			auto noop = make_shared<noop_statement>();
			noop->local_complete = true;
			return noop;
		}
		
		//expected in dumps
		ts->type = extr_type(ts->proc_tuple);
	
	} else if( ts->what == typedef_statement::w_alias ) {
		auto err = type_identifier::expand( ctx, *ts->spec_type );
		if( err.unknown_identifier() ) {
			STMT_CHECK( *ts, !no_incomplete, err.explain_error() );
			//aliases cannot be partially constructure, they only exist in the scope once fully definable
			return ts;
		}
		ctx.sc->add_type_alias( ts->name, ts->spec_type );
	} else if( ts->what == typedef_statement::w_class || ts->what == typedef_statement::w_service ) {
		class_typer::process( ts, ctx );
	}
	
	ts->local_complete = true;
	return ts;
}

shared_ptr<statement> typer::proc_return_statement( shared_ptr<return_statement> rs, context ctx ) {
	STMT_CHECK( *rs, !ctx.sc->is_declarative() );
	auto & func_scope = ctx.sc->get_function_root();
	auto func_expr = func_scope.get_function_root_expr();
	shared_ptr<intr_type_value_list const> return_type;
	if( func_expr->type.is_defined() && func_expr->type.extr().is<intr_type_function>() ) {
		return_type = func_expr->type.extr().cast<intr_type_function>().returns;
	}
	
	if( !rs->value || rs->local_complete ) {
		rs->local_complete = true;
		return rs;
	}
		
	rs->value = ctx.proc_expression( rs->value );
	//DEBUG( "RETURN", dump::get(rs->value->type), rs->value->is_complete() );
	RETURN_IF_UNRESOLVED(*rs->value, rs);

	//convert 1-value-list return to value for type conversion
	if( return_type ) {
		type_ref use_return_type;
		if( return_type->sub.size() == 1 )
			use_return_type = return_type->sub[0].type;
		else
			use_return_type = extr_type(return_type);
		
		rs->value = ctx.convert_expression( use_return_type, rs->value );
		rs->local_complete = true;
		
		STMT_CHECK( *rs, use_return_type.extr().allow_copy(), "no-copy-allowed" );
	} else {
		expr_funcdefn::pend_return pr;
		pr.rs = rs;
		pr.sc = ctx.sc;
		func_expr->pend_returns.push_back(pr);
		rs->local_complete = true;
	}
	return rs;
}

shared_ptr<statement> typer::proc_for_statement( shared_ptr<for_statement> fs, context /*ctx*/ ) {
	static extr_type cond_type( make_shared<intr_type_boolean>() );
	
	//all declarations are local to a new block
	auto outer = make_shared<do_statement>();

	auto iterable = make_shared<var_statement>();
	iterable->decl = declaration::create();
	iterable->decl->form = declaration::f_single;
	iterable->decl->symbol = std::string("_") + fs->item + "_iterator";
	iterable->value = fs->iterable;
	
	do_statement::expr_list_item eli;
	eli.block = make_shared<statement_block>( scope_style_t::imperative );
	eli.block->list.push_back( iterable );
	outer->expr_list.push_back( eli );
	
	//the inner block is the looping part
	auto cond_sym = make_shared<expr_symbol>();
	cond_sym->symbol = iterable->decl->symbol;
	
	auto cond_func = make_shared<expr_field>();
	cond_func->base = cond_sym;
	cond_func->field = "is_more";
	
	auto cond_call = make_shared<expr_funccall>();
	cond_call->func = cond_func;
	cond_call->args = make_shared<expr_value_list>();
	
	auto inner = make_shared<do_statement>();
	inner->loop = true;
	do_statement::expr_list_item inner_eli;
	inner_eli.block = make_shared<statement_block>( scope_style_t::imperative );
	inner_eli.cond = cond_call;
	inner->expr_list.push_back( inner_eli );
	
	eli.block->list.push_back( inner );
	
	//create the item access
	auto item_func = make_shared<expr_field>();
	item_func->base = cond_sym;
	item_func->field = "get";
	
	auto item_call = make_shared<expr_funccall>();
	item_call->func = item_func;
	item_call->args = make_shared<expr_value_list>();
	
	auto item = make_shared<var_statement>();
	item->decl = declaration::create();
	item->decl->form = declaration::f_single;
	item->decl->symbol = fs->item;
	item->value = item_call;
	
	inner_eli.block->list.push_back( item );
	
	//step to next value
	auto step_func = make_shared<expr_field>();
	step_func->base = cond_sym;
	step_func->field = "step";
	
	auto step_call = make_shared<expr_funccall>();
	step_call->func = step_func;
	step_call->args = make_shared<expr_value_list>();
	
	auto stat = make_shared<expression_statement>();
	stat->expr = step_call;
	inner_eli.block->list.push_back( stat );
	
	//add the block
	auto body = make_shared<do_statement>();
	do_statement::expr_list_item body_eli;
	body_eli.block = fs->block;
	body->expr_list.push_back( body_eli );
	inner_eli.block->list.push_back( body );

	return outer;
}

shared_ptr<statement> typer::proc_loop_flow_statement( shared_ptr<loop_flow_statement> ls, 
	context ctx )
{
	STMT_CHECK( *ls, !ctx.sc->is_declarative() );
	ls->local_complete = true;
	return ls;
}

shared_ptr<statement> typer::proc_expression_statement( shared_ptr<expression_statement> es, 
	context ctx )
{
	//STMT_CHECK( *es, !ctx.sc->is_declarative() );
	es->expr = ctx.proc_expression( es->expr );
	RETURN_IF_UNRESOLVED( *es->expr, es )
	//values cannot be implicitly ignored
	auto * vl = es->expr->type.extr().if_cast<intr_type_value_list>();
	if( !es->ignore ) {
		STMT_CHECK( *es, vl && vl->sub.size() == 0, "cannot-ignore-value" );
	}
	
	es->local_complete = true;
	return es;
}

shared_ptr<statement> typer::proc_noop_statement( shared_ptr<noop_statement> ns, context /*ctx*/ ) {
	ns->local_complete = true;
	return ns;
}

struct typer_serial_resolver : public serial_resolver {
	typer & owner;
	typer_serial_resolver( typer & owner ) :
		owner(owner) {
	}
	
	boost::optional<std::string> get_ref_name( intr_type const & t ) {
		if( t.style == intr_type::s_inherent ) {
			return t.get_distinct_name();
		}
		return boost::optional<std::string>();
	}
	std::shared_ptr<intr_type const> get_intr_type( std::string const & ref_name ) {
		auto q = owner.mod->root->if_get_symbol( ref_name );
		STATE_CHECK( !!q && q->what == scope_symbol::w_fun_type, 
			std::string("invalid-import-name: ") + ref_name );
		return q->fun_type;
	}
};

serial_resolver & typer::get_serial_resolver() {
	if( !tsr ) {
		tsr.reset( new typer_serial_resolver( *this ) );
	}
	return *tsr;
}

shared_ptr<statement> typer::proc_import_statement( shared_ptr<import_statement> is, context ctx )
{
	if( is->local_complete )
		return is;
		
	STATE_CHECK( mloader );
	STMT_CHECK( *is, mod->libraries.find(is->mod_name) == mod->libraries.end(), 
		"duplicate-import", error::tag_error_symbol(is->mod_name) );
	mod->libraries.insert( is->mod_name );

	auto in_file_name = mloader->resolve_import( is->mod_name );
	std::ifstream in_file( in_file_name, std::ios::binary );

	serial_in sin( in_file, get_serial_resolver() );
	
	//the imported module becomes a variable with a distinct type
	auto mdecl = declaration::create();
	mdecl->form = declaration::f_single;
	mdecl->symbol = is->mod_name;
	mdecl->source_statement = is;
	ctx.sc->add( mdecl );
	
	auto mscope = sin.get_key_root_scope();
	//the scope is always in root (though the name may not be). This ensures the qualified
	//names are not incorrectly embedded.
	mscope->set_base( mod->root, scope::bt_namespace );
	mscope->set_name( is->mod_name );
	
	auto mtype = make_shared<intr_type_module>();
	mtype->set_distinct_name( std::string("module_") + is->mod_name ); //name is arbitrary
	mtype->style = intr_type::s_distinct;
	mtype->local_complete = true;
	mtype->context = mscope;
	mdecl->type = extr_type(mtype);
	
	for( auto & sitem : sin.get_all_items() ) {
		switch( sitem.type ) {
			case leaf::serial_item::t_invalid:
			case leaf::serial_item::t_end_marker:
			case leaf::serial_item::t_extr_type:
			case leaf::serial_item::t_expression:
			case leaf::serial_item::t_scope:
			case leaf::serial_item::t_key_marker:
			case leaf::serial_item::t_intr_type:
			case leaf::serial_item::t_type_ref:
			case leaf::serial_item::t_type_spec:
			case leaf::serial_item::t_statement:
				break;
				
			case leaf::serial_item::t_declaration:
				//TODO: should export itself set the fields appropriately for import?
				auto decl = sitem.v_declaration;
				//DEBUG( "IMPORT", sitem.v_declaration->symbol, decl->is_lifetime_global(), decl->is_export );
				//if( decl->var_scope && !decl->is_lifetime_global() ) {
				if( !decl->is_export ) {
					break;
				}
				decl->import_name = decl->symbol;
				decl->import_module = is->mod_name;
				decl->name_form = declaration::nf_auto;
				decl->import_type = declaration::it_abi_data;

				/*if( decl->var_scope ) {
					DEBUG("UH OH");
				} else {
					mscope->add( decl );
				}*/
				break;
		}
	}
	
	is->local_complete = true;
	return is;
}


namespace {
	struct symbol_scanner : public stat_expr_transformer, public expression_transformer {
		typer & owner;
		
		symbol_scanner( typer & owner )
			: stat_expr_transformer( static_cast<expression_transformer&>(*this) )
			, expression_transformer( false )
			, owner( owner ) {
		}
		
		shared_ptr<statement> istmt_var( shared_ptr<var_statement> vs, shared_ptr<scope> sc ) {
			switch( vs->decl->form ) {
				case declaration::f_single:
				case declaration::f_multi:
					vs->sref = sc->add( vs->decl );
					break;
					
				case declaration::f_arg:
				case declaration::f_return:
				case declaration::f_capture:
					break;
			}
			return vs;
		}
	};
} //eon anon

/**
	Supporting overloads requires all possible symbols to be known the moment a symbol is
	encountered. This requires pre-scanning all the code for symbols. If this didn't happen
	an overload may resolve to a type from previous declarations, but may not include the
	correct one. This is then just extended to work with all symbols in a declaration scope.
*/
void typer::scan_symbols( shared_ptr<statement_block> block )
{
	symbol_scanner ss( *this );
	ss.apply_statement_block( block );
}

} //eon leaf