#include "parser/intr_type_parser.hpp"
#include "lang/type_spec.hpp"
#include "parser/node_parser.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

#include "regex_patterns.hpp"

namespace leaf
{

	namespace
	{
		bool is_single_end_char( src_symbol_t c ) {
			return c == ',' || c == ')' || c == '}' || c == sym_close_param || c == ';' || c == ']' || c == '\n';
		}
		
		bool is_inline_break_char( src_symbol_t c ) {
			return c == '(' || c == '{' || c == '[';
		}
	}

shared_ptr<type_spec> intr_type_parser::parse_tuple( source & src, src_symbol_t end, unsigned flags ) {
	auto vt = make_shared<type_spec>();
	extend_with_tuple( vt, src, end, flags );
	return vt;
}
	
void intr_type_parser::extend_with_tuple( shared_ptr<type_spec> vt, source & src, src_symbol_t end,
	unsigned flags ) {
	
	STATE_CHECK( !vt->has<pt_fun_class>() );
	STATE_CHECK( vt->sub.empty() );
	vt->set<pt_fun_class>( intr_type::f_tuple );

	static boost::regex re_symbol( SYMBOL_REGEX );

	bool at_name = true, linger_sep = false;
	std::string cur_name;
	
	while( true )
	{
		STATE_CHECK( src.has_next(source::skip_line_ends), "unexpected end of type tuple" );

		//check for completion
		auto nc = src.peek_next();
		if( is_single_end_char(nc) ) {
			if( cur_name.size() ) {
				//names without types are also okay
				vt->push_sub( cur_name, make_shared<type_spec>() );
			} else if( linger_sep ) {
				//simple separators also denote an empty entry
				vt->push_sub( make_shared<type_spec>() );
			}
			
			if( nc == end )	{
				src.next();
				break;
			} else 	if( nc == ',' ) {
				src.next();
				at_name = true;
				linger_sep = false;
				cur_name.clear();
				continue;
			}
			src.throw_unexpected( "type-end-char" );
		}
		
		static boost::regex re_op( OP_REGEX_MULTI );
		if( src.match( re_op ) ) {
			//is a separator required (this comes before the re_symbol check since a symbol is optional)
			if( src.cur_match() == ":" ) {
				linger_sep = true;
				if( !at_name )
					src.throw_unexpected( "type-separator" );
					
				at_name = false;
				continue;
			} else {
				src.throw_unexpected( "incomplete type tuple" );
			}
		}

		if( at_name && !(flags & ptf_no_names) )
		{
			if( cur_name.size() )
				src.throw_expecting( "type separator" );
			if( !src.match( re_symbol ) )
				src.throw_expecting( "symbol" );
			
			cur_name = src.cur_match();
		}
		else
		{
			auto s = parse_single(src, pf_line_cont);
			vt->push_sub( cur_name, s );
			cur_name = "";
			linger_sep = false;
		}
	}
}

void intr_type_parser::parse_signature( source & src, shared_ptr<type_spec> const &sig )
{
	auto t_params = parse_tuple( src, ')' );
	t_params->set<pt_fun_class>( intr_type::f_value_list );

	src.skip_all_space();
	static boost::regex re_arrow( "->" );
	if( !src.match( re_arrow ) ) {
		//nope, just a value list
		sig->set<pt_fun_class>( intr_type::f_value_list );
		sig->sub = t_params->sub;
		return;
	}

	static boost::regex re_extra( "\\$implicit" );
	src.skip_space();
	shared_ptr<type_spec> t_return;
	auto nc = src.peek_next();
	if( nc == '(' ) {
		src.next();
		t_return = parse_tuple( src, ')' );
		t_return->set<pt_fun_class>( intr_type::f_value_list );
	} else if( src.match( re_extra ) ) {
		//leave t_return null
	} else {
		src.throw_expecting( "return-specifier" );
	}

	if( sig->has<pt_fun_class>() || sig->sub.size() != 0 ) {
		src.throw_invalid( "duplicate-signature" );
	}
	
	sig->set<pt_fun_class>( intr_type::f_function );
	sig->push_sub( t_params );
	if( t_return ) {
		sig->push_sub( t_return );
	}
}

shared_ptr<type_spec> intr_type_parser::parse_single( source & src, unsigned flags, bool * empty ) {
	auto vt = make_shared<type_spec>();
	bool line_cont = flags & pf_line_cont;
	bool first_token = true;
	bool end_close_paren = false;
	bool is_symbol = false;
	
	auto is_inline = [&]() {
		return !end_close_paren && (flags & pf_inline);
	};
	
	while( true ) {
		bool was_first = first_token;
		bool was_symbol = is_symbol;
		is_symbol = false;
		
		if( !src.has_next( (end_close_paren || line_cont) ? source::skip_line_ends : 0) ) {
			break;
		}

		auto nc = src.peek_next();
		if( end_close_paren ) {
			//in this mode only a close-paren ends the type (though other end characters will
			//still be flagged as invalid usually
			if( nc == ')' )
				break;
		} else {
			static boost::regex re_op( "=|:=" );
			if( src.peek_match(re_op) || is_single_end_char(nc) )
				break;
			if( (flags & pf_inline) && is_inline_break_char(nc)
				&& !(first_token && nc == '(') /*see below*/) {
				break;
			}
		}

		if( nc == sym_open_param ) {
			src.next();
			auto sub = parse_tuple( src, sym_close_param, was_first ? 0 : ptf_no_names );
			if( was_first ) {
				vt->params = sub->sub;
			} else if( was_symbol ) {
				for( auto & s : sub->sub ) {
					vt->get_parts().back().sub.push_back( s.type );
				}
			} else {
				src.throw_unexpected( "params-type" );
			}
			continue;
		}
		
		if( first_token ) {
			first_token = false;
			if( nc == '(' ) {
				src.next();
				if( flags & pf_inline ) {
					//inline starting with open-paren is terminated by matching close-paren
					end_close_paren = true;
					continue;
				} else {
					parse_signature( src, vt );
				}
				continue;
			}
		}
		
		if( nc == '[' ) {
			src.next();
			extend_with_tuple( vt, src, ']' );
			continue;
		}

		static boost::regex re_marker( "(lvalue|rvalue|shared|raw|unbound|"
			"high|low|normal|optional|selectable|member|!member|convert_optional|value_ptr|value|"
			"lossy|literal|no_throw|tuple|any|construct)" EXCLUDE_SYMBOL_REGEX_CHAR );
		static boost::regex re_symbol( TYPE_SYMBOL_REGEX );
		static boost::regex re_bitsize( "([0-9]+)(bit)" EXCLUDE_SYMBOL_REGEX_CHAR );
		static boost::regex re_context( "context\\(([^)]*)\\)" EXCLUDE_SYMBOL_REGEX_CHAR );

		if( src.match( re_marker ) ) {
			std::string f( src.cur_match() );
			if( f == "any" ) {
				//does not thing but serve as a placeholder to force a tuple
			} else if( f == "lvalue" ) {
				STATE_CHECK( !vt->has<pt_lvalue>() );
				vt->set<pt_lvalue>(true);
			} else if( f == "rvalue" ) {
				STATE_CHECK( !vt->has<pt_lvalue>() );
				vt->set<pt_lvalue>(false);
			} else if( f == "shared" ) {
				STATE_CHECK( !vt->has<pt_reference>() );
				vt->set<pt_reference>(extr_type::r_shared);
			} else if( f == "value_ptr" ) {
				STATE_CHECK( !vt->has<pt_reference>() );
				vt->set<pt_reference>( extr_type::r_value_ptr );
			} else if( f == "value" ) {
				STATE_CHECK( !vt->has<pt_reference>() );
				vt->set<pt_reference>( extr_type::r_value );
			} else if( f == "raw" ) {
				STATE_CHECK( !vt->has<pt_function_access>() ); 
				vt->set<pt_function_access>( intr_type_function::a_raw );
			} else if( f == "unbound" ) {
				STATE_CHECK( !vt->has<pt_function_access>() );
				vt->set<pt_function_access>( intr_type_function::a_unbound );
			} else if( f == "construct" ) {
				STATE_CHECK( !vt->has<pt_function_access>() );
				vt->set<pt_function_access>( intr_type_function::a_construct );
			} else if( f == "no_throw" ) {
				STATE_CHECK( !vt->has<pt_function_throw>() );
				vt->set<pt_function_throw>( false );
			} else if( f == "optional" ) {
				STATE_CHECK( !vt->has<pt_optional>() );
				vt->set<pt_optional>( true );
			} else if( f == "convert_optional" ) {
				STATE_CHECK( !vt->has<pt_cvt_optional>() );
				vt->set<pt_cvt_optional>( true );
			} else if( f == "high" ) {
				STATE_CHECK( !vt->has<pt_precision_variant>() );
				vt->set<pt_precision_variant>(1);
			} else if( f == "low" ) {
				STATE_CHECK( !vt->has<pt_precision_variant>() );
				vt->set<pt_precision_variant>(-1);
			} else if( f == "normal" ) {
				STATE_CHECK( !vt->has<pt_precision_variant>() );
				vt->set<pt_precision_variant>(0);
			} else if( f == "selectable" ) {
				STATE_CHECK( !vt->has<pt_selectable>() );
				vt->set<pt_selectable>(true);
			} else if( f == "member" ) {
				STATE_CHECK( !vt->has<pt_member>() );
				vt->set<pt_member>( true );
			} else if( f == "!member" ) {
				STATE_CHECK( !vt->has<pt_member>() );
				vt->set<pt_member>( false );
			} else if( f == "lossy" ) {
				STATE_CHECK( !vt->has<pt_lossy>() );
				vt->set<pt_lossy>( true );
			} else if( f == "literal" ) {
				STATE_CHECK( !vt->has<pt_literal>() );
				vt->set<pt_literal>( true );
			} else if( f == "tuple" ) {
				STATE_CHECK( !vt->has<pt_tuple>() );
				vt->set<pt_tuple>( true );
			} else {
				STATE_CHECK( false );
			}
		} else if( src.match( re_bitsize ) ) {
			STATE_CHECK( !vt->has<pt_data_bitsize>() );
			vt->set<pt_data_bitsize>( atoi( src.cur_match(1).c_str() ) );
		} else if( src.match( re_context ) ) {
			STATE_CHECK( !vt->has<pt_context_name>() );
			vt->set<pt_context_name>( src.cur_match(1).c_str() );
		} else if( src.match( re_symbol ) ) {
			type_spec_symbol tss;
			tss.parts.push_back( src.cur_match() );
			auto &p = vt->push_part<pt_symbol>( tss );
			is_symbol = true;

			if( !is_inline() ) {
				auto nc = src.peek_next();
				if(nc == '(') {
					p.node_expr = node_parser::parse_value_list(src);
				}
			}
		} else {
			src.throw_invalid( "type syntax" );
		}
	}

	//an empty type is valid -- simply unknown
	if( empty )
		*empty = first_token;

	return vt;
}

intr_type_parser::intr_type_parser() {
}

/*static*/ node_type_spec intr_type_parser::parse( source & src ) {
	intr_type_parser p;
	return p.parse_single( src );
}

/*static*/ node_type_spec intr_type_parser::parse_inline( source & src) {
	intr_type_parser p;
	return p.parse_single( src, pf_inline );
}

/*static*/ node_type_spec intr_type_parser::if_parse_inline( source & src ) {
	intr_type_parser p;
	bool empty;
	auto r = p.parse_single( src, pf_inline, &empty );
	if( empty )
		return shared_ptr<type_spec>();
	return r;
}

/*static*/ node_type_spec intr_type_parser::parse_inline_param( source & src ) {
	intr_type_parser p;
	
	auto nc = src.next();
	if( nc != sym_open_param ) {
		src.throw_expecting( "open-param" );
	}
	
	auto t = p.parse_tuple( src, sym_close_param,  ptf_no_names );
	t->set<pt_fun_class>( intr_type::f_value_list );
	return t;
}

} //eon leaf
