#include "parser/node_converter.hpp"
#include "parser/node_parser.hpp"
#include "parser/node.hpp"

#include "lang/language.hpp"
#include "lang/dump.hpp"
#include "util/unicode.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

#define NODE_CHECK( node, cond, ... )	\
	if( !(cond) ) { auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
		x << error::tag_error_condition( #cond ); \
		x << util::source_location::error( (node).src_loc ); \
		BOOST_THROW_EXCEPTION( x ); }
#define NODE_FAIL( node, ... )	{ \
	auto x = error::create_exception<recoverable_error>( __VA_ARGS__ ); \
	x << util::source_location::error( (node).src_loc ); \
	BOOST_THROW_EXCEPTION( x ); }

#define EXPECT_FUNCTION( src ) \
	NODE_CHECK( (src), (src).form == node::f_funccall ); \
	NODE_CHECK( (src), (src).sub.size() == 2 ); \
	auto & symbol = *(src).sub[0]; \
	NODE_CHECK( symbol, !symbol.text.empty() ); \
	auto & args = *(src).sub[1]; \
	NODE_CHECK( args, args.form == node::f_value_list );

#define STD_FUNCTION( src, sym ) \
	EXPECT_FUNCTION( src ); \
	NODE_CHECK( symbol, symbol.text == sym );

#define STD_FUNCTION2( src, syma, symb ) \
	EXPECT_FUNCTION( src ); \
	NODE_CHECK( symbol, symbol.text == syma || symbol.text == symb);
	
namespace leaf
{

/*static*/ shared_ptr<expression> node_converter::expression_convert( node const & src )
{
	try {
		node_converter nc;
		return nc.convert_expression( src );
	} catch( recoverable_error & re ) {
		re << error_converter(true);
		throw;
	}
}

/*static*/ shared_ptr<statement> node_converter::convert( node const & src )
{
	try {
		node_converter nc;
		return nc.convert_statement( src );
	} catch( recoverable_error & re ) {
		re << error_converter(true);
		throw;
	}
}

/*static*/ node_converter::statement_vector_t node_converter::convert_block( node const & src,
	scope_style_t style ) {
	try {
		node_converter nc;
		
		NODE_CHECK( src, src.form == node::f_block );
		statement_vector_t block;
		nc.convert_block( block, src, style );
		return block;
	} catch( recoverable_error & re ) {
		re << error_converter(true);
		throw;
	}
}


shared_ptr<statement> node_converter::convert_statement( node const & src )
{
	auto * primary = &src;
	if( src.form == node::f_on ) {
		return convert_xon_statement( src );
	}
	
	if( src.form == node::f_expr_list ) {
		auto sz = src.sub.size();
		if( sz == 0 ) {
			return make_shared<noop_statement>();
		} else if( (sz > 1) && (src.sub[1]->text == "then" || src.sub[1]->text == "else" )) {
			return convert_conditional_statement(src);
		}
		primary = src.sub[0].get();
	}
	
	//special declarations/flow
	if( primary->form == node::f_symbol ) {
		auto & decl = *primary;
		if( decl.text == "class" || decl.text == "service" ) {
			return convert_class_statement(src);
		} else if( decl.text == "typedef" ) {
			return convert_typedef_statement( src, false );
		} else if( decl.text == "alias" ) {
			return convert_typedef_statement( src, true );
		} else if( decl.text == "var" || decl.text == "defn" || decl.text == "literal" || decl.text == "multi" ) {
			return convert_var_statement( src );
		} else if( decl.text == "do" || decl.text == "loop" ) {
			return convert_do_statement( src );
		} else if( decl.text == "defer" ) {
			return convert_defer_statement( src, false );
		} else if( decl.text == "defer_fail" ) {
			return convert_defer_statement( src, true );
		} else if( decl.text == "import" ) {
			return convert_import_statement( src );
		} else if( decl.text == "⁑assign" ) {
			return convert_assign_statement( src, ah_value );
		} else if( decl.text == "⁑assign_value" ) {
			return convert_assign_statement( src, ah_init_refer );
		} else if( decl.text == "⁑assign_default" ) {
			return convert_assign_statement( src, ah_default );
		} else if( decl.text == "return" ) {
			return convert_return_statement( src );
		} else if( decl.text == "resume_fail" ) {
			return convert_resume_statement( src );
		} else if( decl.text == "break" ) {
			return convert_break_statement( src, false );
		} else if( decl.text == "next" ) {
			return convert_break_statement( src, true );
		} else if( decl.text == "pre_fail" ) {
			return convert_fail_statement( src, fail_level::pre );
		} else if( decl.text == "fail" ) {
			return convert_fail_statement( src, fail_level::normal );
		} else if( decl.text == "critical_fail" ) {
			return convert_fail_statement( src, fail_level::critical );
		} else if( decl.text == "for" ) {
			return convert_for_statement( src );
		} else if( decl.text == "ignore" ) {
			return convert_ignore_expression_statement( src );
		}
	}
	
	if( src.form == node::f_funccall ) {
		//some pseudo-functions become statements
		NODE_CHECK( src, src.sub.size() == 2 );
		if( src.sub[0]->form == node::f_symbol ) {
			auto & symbol = src.sub[0]->text;
			
			if( symbol == "⁑assign" )
				return convert_assign_statement( src, ah_value );
			if( symbol == "⁑assign_value" )
				return convert_assign_statement( src, ah_init_refer );
			if( symbol == "⁑assign_default" )
				return convert_assign_statement( src, ah_default );
				
			//some decls may also look like functions with a trailing value-list
			if( symbol == "return" )
				return convert_return_statement( src );
		}
	} 
	return convert_expression_statement( src );	
}

//TODO: temporary naming, remove above once it works
shared_ptr<statement> node_converter::convert_xon_statement( node const & src ) {
	NODE_CHECK( src, src.sub.size() == 3 );
	
	auto xon = make_shared<error_statement>();
	xon->src_loc = src.src_loc;
	xon->inner = convert_statement( *src.sub[0] );
	
	auto & match = *src.sub[1];
	NODE_CHECK( match, match.form == node::f_symbol );
	
	auto & ons = match.text;
	if( ons == "pre_fail" ) {
		xon->error.on_match = fail_level::pre;
	} else if( ons == "fail" ) {
		xon->error.on_match = fail_level::normal;
	} else if( ons == "critical_fail" ) {
		xon->error.on_match = fail_level::critical;
	} else {
		NODE_FAIL( match, "invalid-on-match" );
	}
	xon->error.on_error = create_statement_block( *src.sub[2], scope_style_t::imperative );
	
	return xon;
}

shared_ptr<expression_statement> node_converter::convert_expression_statement(
	node const & src ) {
	auto es = make_shared<expression_statement>();
	es->expr = convert_expression( src );
	es->src_loc = es->expr->src_loc;
	return es;
}

shared_ptr<expression_statement> node_converter::convert_ignore_expression_statement(
	node const & src ) {
	NODE_CHECK( src, src.sub.size() == 2 );
	auto es = make_shared<expression_statement>();
	es->expr = convert_expression( *src.sub[1] );
	es->src_loc = src.sub[0]->src_loc;
	es->ignore = true;
	return es;
}

shared_ptr<expression> node_converter::convert_expression( node const & src ) {
	shared_ptr<expression> ex;
	
	bool omit_type = false;
	switch( src.form )
	{
		case node::f_symbol: {
			if( src.text == "this" ) {
				auto ec = make_shared<expr_current>();
				ec->what = expr_current::w_this;
				ex = ec;
			} else if( src.text == "typedef" ) {
				NODE_FAIL( src, "type-not-allowed-here" );
			} else if( src.text == "type" ) {
				ex = convert_type_expression( src );
			} else {
				auto es = make_shared<expr_symbol>();
				es->symbol = src.text;
				es->param = convert_type(src.param);
				ex = es;
			}
			break;
		}
			
		case node::f_literal_number: {
			auto & ncn = static_cast<node_literal_number const&>(src);
			auto ecn = make_shared<expr_literal_number>();
			ecn->src_form = expr_literal_number::sf_rational;
			//TODO: need an explicit high-prec float still
			ecn->value = ncn.value;
			ex = ecn;
			break;
		}
			
		case node::f_literal_boolean: {
			auto ecn = make_shared<expr_literal_number>();
			if( src.text == "true" )
				ecn->val_boolean = true;
			else if( src.text == "false" )
				ecn->val_boolean = false;
			else	
				NODE_FAIL( src, "invalid boolean constant" );
			ecn->src_form = expr_literal_number::sf_boolean;
			ex = ecn;
			break;
		}

		case node::f_value_list: {
			auto evl = make_shared<expr_value_list>();
			convert_sub( *evl, src );
			ex = evl;
			break;
		}

		case node::f_tuple: {
			auto et = make_shared<expr_compound>();
			convert_sub( *et, src );
			ex = et;
			break;
		}

		case node::f_funccall: {
			//special functions
			NODE_CHECK( src, src.sub.size() == 2 );
			if( src.sub[0]->form == node::f_symbol ) {
				auto & symbol = src.sub[0]->text;
				
				if( symbol == "⁑field" || symbol == "⁑optional_field" )
					return convert_field_expression( src );
				if( symbol == "select" )
					return convert_select_expression( src );
				if( symbol == "⁑ifelse" )
					return convert_ifelse_expression( src );
				if( symbol == "⁑optional_or" )
					return convert_optional_or_expression( src );
				if( symbol == "⁑optional_cond" )
					return convert_optional_cond_expression( src );
				if( symbol == "⁑subscript" || symbol == "⁑optional_subscript" )
					return convert_subscript_expression( src );
				if( symbol == "string_tag" )
					return convert_tag_expression( src );
			}
				
			auto ef = make_shared<expr_funccall>();
			ef->func = convert_expression( *src.sub[0] );
			ef->args = convert_expression( *src.sub[1] );
			ex = ef;
			break;
		}

		case node::f_funcdefn: {
			ex = convert_funcdefn_expression( src );
			omit_type = true;
			break;
		}
			
		case node::f_block: {
			auto eb = make_shared<expr_block>();
			eb->block = create_statement_block( src, scope_style_t::imperative );
			ex = eb;
			break;
		}

		case node::f_literal_string: {
			auto es = make_shared<expr_text>();
			es->what = expr_text::w_string;
			es->text = unicode::decode8(src.text);
			ex = es;
			break;
		}
		
		case node::f_literal_char: {
			auto es = make_shared<expr_text>();
			es->what = expr_text::w_char;
			es->text = unicode::decode8(src.text);
			NODE_CHECK( src, es->text.size() == 1 );
			ex = es;
			break;
		}
			
		case node::f_decl_name:
		case node::f_type:
		case node::f_named:
		case node::f_on:
		case node::f_expr_list:
		case node::f_implied:
		case node::f_param_list:
			NODE_FAIL( src, "node type should not appear here" );
	}

	NODE_CHECK( src, ex);
	if( src.type && !omit_type ) {
		auto cast = make_shared<expr_cast>();
		cast->source = ex;
		cast->spec_type = convert_type(src.type);
		ex = cast;
	}
	ex->src_loc = src.src_loc;
	return ex;
}

shared_ptr<expression> node_converter::convert_field_expression( node const & src ) {
	STD_FUNCTION2( src, "⁑field", "⁑optional_field" );
	NODE_CHECK( src, args.sub.size() == 2 );
	
	auto fe = make_shared<expr_field>();
	fe->src_loc = src.src_loc;
	fe->base = convert_expression( *args.sub[0] );
	
	NODE_CHECK( src, args.sub[1]->form == node::f_symbol );
	fe->field = args.sub[1]->text;

	fe->param = convert_type(args.sub[1]->param);
	
	fe->optional = symbol.text == "⁑optional_field";
	return fe;
}

shared_ptr<expression> node_converter::convert_subscript_expression( node const & src ) {
	STD_FUNCTION2( src, "⁑subscript", "⁑optional_subscript" );
	NODE_CHECK( src, args.sub.size() == 2 );
	
	auto es = make_shared<expr_subscript>();
	es->base = convert_expression( *args.sub[0] );
	es->offset = convert_expression( *args.sub[1] );
	es->optional = symbol.text == "⁑optional_subscript";
	return es;
}


void node_converter::convert_sub( expr_with_sub & out, node const & src )
{
	for( size_t i=0; i < src.sub.size(); ++i ) {
		auto & item = src.sub[i];
		if( item->form == node::f_funccall && item->sub.size() == 2
			&& item->sub[0]->text == "⁑assign" ) {
			auto & args = *item->sub[1];
			NODE_CHECK( src, args.sub.size() == 2 );
			auto ex = convert_expression( *args.sub[1] );
			auto name = args.sub[0]->text;
			out.push_sub( name, ex );
		} else {
			out.push_sub( convert_expression( *item ) );
		}
	}
}


shared_ptr<expression> node_converter::convert_ifelse_expression( node const & src )
{
	STD_FUNCTION( src, "⁑ifelse" )
	NODE_CHECK( src, args.sub.size() == 3 );

	auto ex = make_shared<expr_switch>();
	
	expr_switch::expr_list_item eli;
	eli.cond = convert_expression( *args.sub[0] );
	eli.value = convert_expression( *args.sub[1] );
	ex->expr_list.push_back( eli );
	
	//literal_bool true
	expr_switch::expr_list_item eli_else;
	eli_else.value = convert_expression(*args.sub[2]);
	ex->expr_list.push_back( eli_else );
	
	return ex;
}

shared_ptr<expression> node_converter::convert_optional_or_expression( node const & src ) {
	STD_FUNCTION( src, "⁑optional_or" );
	NODE_CHECK( src, args.sub.size() == 2 );
	
	auto ex = make_shared<expr_conditional>();
	ex->op = expr_conditional::o_optional_or;
	ex->cond = convert_expression( *args.sub[0] );
	//a is not set here, implied from cond
	ex->b = convert_expression( *args.sub[1] );
	return ex;
}

shared_ptr<expression> node_converter::convert_optional_cond_expression( node const & src ) {
	STD_FUNCTION( src, "⁑optional_cond" );
	NODE_CHECK( src, args.sub.size() == 2 );
	
	auto ex = make_shared<expr_conditional>();
	ex->op = expr_conditional::o_optional_cond;
	ex->cond = convert_expression( *args.sub[0] );
	ex->a = convert_expression( *args.sub[1] );
	//b is not used
	return ex;
}

shared_ptr<expression> node_converter::convert_type_expression( node const & src ) {
	NODE_CHECK( src, src.sub.size() == 0 );
	auto t = convert_type(src.param); 
	NODE_CHECK( src, t->sub.size() == 1 );
	
	auto ex = make_shared<expr_type>();
	ex->inner_type_spec = t->sub[0].type;
	return ex;
}

shared_ptr<expression> node_converter::convert_select_expression( node const & src ) {
	STD_FUNCTION( src, "select" );
	
	auto ex = make_shared<expr_switch>();
	ex->src_loc = src.src_loc;

	for( auto & ni : args.sub ) {
		expr_switch::expr_list_item eli;
		
		//in thsi form the conditional has special meaning
		if( ni->form == node::f_funccall 
			&& ni->sub[0]->form == node::f_symbol
			&& ni->sub[0]->text == "⁑optional_cond" ) {
			STATE_CHECK( ni->sub[1]->sub.size() == 2 );
			eli.cond = convert_expression( *ni->sub[1]->sub[0] );
			eli.value = convert_expression( *ni->sub[1]->sub[1] );
		} else {
			eli.value = convert_expression( *ni );
		}
		ex->expr_list.push_back( eli );
	}
	
	return ex;
}

shared_ptr<expression> node_converter::convert_tag_expression( node const & src ) {
	STD_FUNCTION( src, "string_tag" );
	NODE_CHECK( src, args.sub.size() == 1 );
	
	auto ex = make_shared<expr_tag>();
	ex->src_loc = src.src_loc;
	ex->value = convert_expression( *args.sub[0] );
	return ex;
}


shared_ptr<do_statement> node_converter::convert_do_statement( node const & list )
{
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];
	
	auto value = convert_expression( src );
	
	shared_ptr<do_statement> ds( new do_statement );
	ds->src_loc = src.src_loc;
	//TODO: eventually this must be cleaned up to accept anything and let the typer
	//deal with it. (It would require block's as closures first though...)
	if( value->form == expression::f_switch ) {
		auto & es = static_cast<expr_switch&>(*value);
		for( auto & item : es.expr_list ) {
			
			do_statement::expr_list_item eli;
			eli.cond = item.cond;
			NODE_CHECK( src, item.value->form == expression::f_block );
			eli.block = static_cast<expr_block&>(*item.value).block;
			ds->expr_list.push_back( eli );
		}
		
	} else if( value->form == expression::f_conditional ) {
		auto & ec = static_cast<expr_conditional&>(*value);
		do_statement::expr_list_item eli;
		if( ec.op == expr_conditional::o_optional_cond ) {
			eli.cond = ec.cond;
			NODE_CHECK( src, ec.a->form == expression::f_block );
			eli.block = static_cast<expr_block&>(*ec.a).block;
		} else {
			NODE_FAIL( src, "expecting optional_cond" );
		}
		ds->expr_list.push_back( eli );
	} else if( value->form == expression::f_block ) {
		do_statement::expr_list_item eli;
		eli.block = static_cast<expr_block*>(value.get())->block;
		ds->expr_list.push_back( eli );
	} else {
		NODE_FAIL( src, "expecting switch or block" );
	}
	
	if( list.sub[0]->text == "loop" ) {
		ds->loop = true;
	}
	
	return ds;
}

shared_ptr<statement_block> node_converter::create_statement_block( node const & src,
	scope_style_t scope_style ) {
	shared_ptr<statement_block> block( new statement_block( scope_style ) );
	block->src_loc = src.src_loc;
	convert_block( block->list, src, scope_style );
	return block;
}

void node_converter::convert_block( statement_vector_t & block, node const & src,
	scope_style_t scope_style ) {
	for( size_t i=0; i < src.sub.size(); ++i ) {
		shared_ptr<statement> sub = convert( *src.sub[i] );
		if( scope_style == scope_style_t::declarative ) {
			NODE_CHECK( *src.sub[i], sub->is_declarative(), "expect-declarative-block" );
		}

		block.push_back( sub );
	}
}

/**
	An "assign funcall" or "symbol" form can convert to a var_statement.
*/
shared_ptr<var_statement> node_converter::convert_var_statement( node const & list )
{
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];

	node const* symbol_node = &src;
	node const* value_node = nullptr;
	
	bool init_refer_value = false;
	if( src.form == node::f_funccall )
	{
		NODE_CHECK( src, src.sub.size() == 2 );
		
		if( src.sub[0]->text == "⁑assign" )
			init_refer_value = false;
		else if( src.sub[0]->text == "⁑assign_value" )
			init_refer_value = true;
		else
			NODE_FAIL( src, "invalid input node" );
		
		NODE_CHECK( src, src.sub[0]->form == node::f_symbol );
		NODE_CHECK( src, src.sub[1]->sub.size() == 2 );
		symbol_node = src.sub[1]->sub[0].get();
		value_node = src.sub[1]->sub[1].get();
	}
	
	NODE_CHECK( src, symbol_node->form == node::f_symbol );

	shared_ptr<var_statement> vs( new var_statement );
	vs->src_loc = src.src_loc;
	vs->init_refer_value = init_refer_value;
	vs->decl = declaration::create();
	vs->decl->symbol = symbol_node->text;
	vs->decl->spec_type = convert_type(symbol_node->type);
	
	auto ensure_spec_type = [&]() {
		if( !vs->decl->spec_type ) {
			vs->decl->spec_type = make_shared<type_spec>();
		}
	};
	
	
	auto dstr = list.sub[0]->text;
	if( dstr == "var" ) {
		vs->decl->form = declaration::f_single;
	} else if( dstr == "defn" ) {
		vs->decl->form = declaration::f_multi;
		vs->decl->lifetime = declaration::l_global;
		ensure_spec_type();
		vs->decl->spec_type->set<pt_literal>( true );
	} else if( dstr == "literal" ) {
		ensure_spec_type();
		vs->decl->spec_type->set<pt_literal>( true );
		vs->decl->form = declaration::f_single;
	} else if( dstr == "multi" ) {
		vs->decl->form = declaration::f_multi;
	} else {
		STATE_FAIL(dstr);
	}
	
	for( auto it = list.annotations.begin(); it != list.annotations.end(); ++it ) {
		if( it->first == "import" || it->first == "import_data" ) {
			vs->decl->spec_type->set<pt_literal>( false ); //in case defn
			auto & sub = it->second->sub;
			NODE_CHECK( *it->second, sub.size() >= 1 && sub.size() <= 2 );
			vs->decl->import_type = it->first == "import" ? declaration::it_abi_code : declaration::it_abi_data;
			
			//split name into module.name if needed
			auto & sym = *sub[0];
			NODE_CHECK( sym, sym.form == node::f_literal_string );
			size_t dot = sym.text.find_first_of( '.' );
			if( dot != std::string::npos ) {
				vs->decl->import_module = sym.text.substr(0,dot);
				vs->decl->import_name = sym.text.substr(dot+1);
			} else {
				vs->decl->import_name = sym.text;
			}
			
			//extract optional form
			if( sub.size() > 1 ) {
				auto & form = *sub[1];
				NODE_CHECK( form, form.form == node::f_symbol );
				if( form.text == "leaf" ) {
					vs->decl->name_form = declaration::nf_leaf;
				} else if( form.text == "bare" ) {
					vs->decl->name_form = declaration::nf_bare;
				} else {
					NODE_FAIL( form, "import-unknown-name-form" );
				}
			}
		} else if( it->first == "export" ) {
			vs->decl->is_export = true;
		} else if( it->first == "lifetime" ) {
			auto & sub = it->second->sub;
			NODE_CHECK( *it->second, sub.size() == 1 );
			auto & life = *sub[0];
			NODE_CHECK( life, life.form == node::f_symbol );
			if( life.text == "local" ) {
				vs->decl->lifetime = declaration::l_local;
			} else if( life.text == "global" ) {
				vs->decl->lifetime = declaration::l_global;
			} else {
				NODE_FAIL( life, "unknown-lifetime" );
			}
		} else {
			NODE_FAIL( *it->second, "unknown-annotation" );
		}
	}

	if( value_node )
		vs->value = convert_expression( *value_node );
	
	return vs;
}

shared_ptr<assign_statement> node_converter::convert_assign_statement( node const & src, 
	assign_how_t how )
{
	NODE_CHECK( src, src.form == node::f_funccall );
	NODE_CHECK( src, src.sub.size() == 2 );
	
	node const & func = *src.sub[0];
	NODE_CHECK( func, func.form == node::f_symbol );

	node const & args = *src.sub[1];
	NODE_CHECK( args, args.form == node::f_value_list );
	NODE_CHECK( args, args.sub.size() == 2 );

	shared_ptr<assign_statement> as( new assign_statement );
	as->src_loc = src.src_loc;
	as->left = convert_expression(*args.sub[0]);
	switch( how ) {
		case ah_value: as->how = assign_statement::h_value; break;
		case ah_init_refer: as->how = assign_statement::h_init_refer; break;
		case ah_default: as->how = assign_statement::h_default; break;
	}
	
	as->right = convert_expression(*args.sub[1]);
	return as;
}

shared_ptr<defer_statement> node_converter::convert_defer_statement( node const & list, bool error_only )
{
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];
	
	shared_ptr<expression> value = convert_expression( src );
	
	NODE_CHECK( src, value->form == expression::f_block );
	
	shared_ptr<defer_statement> defer( new defer_statement );
	defer->src_loc = src.src_loc;
	defer->error_only = error_only;
	defer->error.on_error = static_cast<expr_block*>(value.get())->block;
	defer->error.on_match = fail_level::critical; //should always run
	
	return defer;
}

namespace {
	void add_annotations( node const & src, shared_ptr<typedef_statement> ts ) {
		for( auto & anno : src.annotations ) {
			if( anno.first == "inherent" ) {
				ts->is_inherent = true;
			} else if( anno.first == "export" ) {
				ts->is_export = true;
			} else {
				NODE_FAIL( src, "unknown-annotation", error::tag_error_symbol( anno.first ) );
			}
		}
	}
}

shared_ptr<typedef_statement> node_converter::convert_typedef_statement( node const & src, bool alias )
{
	NODE_CHECK( src, src.sub.size() == 2 );
	NODE_CHECK( src, src.sub[1]->form == node::f_symbol );
	
	auto ts = make_shared<typedef_statement>();
	ts->src_loc = src.src_loc;
	ts->name = src.sub[1]->text;
	ts->what = alias ? typedef_statement::w_alias : typedef_statement::w_prototype;
	ts->spec_type = convert_type(src.sub[1]->type); //may be null
	
	add_annotations( src, ts );
	return ts;
}

shared_ptr<typedef_statement> node_converter::convert_class_statement( node const & src )
{
	NODE_CHECK( src, src.sub.size() >= 3 && src.sub.size() <= 4 );
	
	NODE_CHECK( src, src.sub[1]->form == node::f_symbol );
	auto ts = make_shared<typedef_statement>();
	ts->src_loc = src.src_loc;
	ts->name = src.sub[1]->text;
	
	int subAt = 2;
	
	if( src.sub[subAt]->form == node::f_param_list ) {
		auto spec = make_shared<type_spec>();
		for( auto & s : src.sub[subAt]->sub ) {
			NODE_CHECK( *s, s->form == node::f_symbol );
			spec->push_param( s->text, convert_type(s->type, ct_create_empty) );
		}
		
		ts->spec_type = spec;
		subAt++;
	}
	
	NODE_CHECK( src, src.sub[subAt]->form == node::f_block  );
	ts->what = src.sub[0]->text == "class" ? typedef_statement::w_class : typedef_statement::w_service;
	ts->class_block = create_statement_block( *src.sub[subAt], scope_style_t::declarative );
	
	add_annotations( src, ts );
	return ts;
}


shared_ptr<return_statement> node_converter::convert_return_statement( node const & src )
{
	NODE_CHECK( src, src.sub.size() <= 2, "too-many-return-values" );
	auto rs = make_shared<return_statement>();
	rs->src_loc = src.src_loc;	
	if( src.sub.size() == 2 ) {
		rs->value = convert_expression( *src.sub[1] );
	}
	return rs;
}

shared_ptr<statement> node_converter::convert_resume_statement( node const & src ) {
	NODE_CHECK( src, src.sub.size() == 0 );
	
	auto rs = make_shared<fail_statement>( fail_level::resume );
	rs->src_loc = src.src_loc;
	return rs;
}

shared_ptr<statement> node_converter::convert_fail_statement( node const & src, fail_level level ) {
	NODE_CHECK( src, src.sub.size() <= 2 );
	
	auto fs = make_shared<fail_statement>( level );
	fs->src_loc = src.src_loc;
	
	if( src.sub.size() == 2 ) {
		fs->tag = convert_expression( *src.sub[1] );
	}
	
	return fs;
}


shared_ptr<statement> node_converter::convert_break_statement( node const & src, bool cont ) {
	NODE_CHECK( src, src.sub.size() == 0 );
	
	auto bs = make_shared<loop_flow_statement>();
	bs->src_loc = src.src_loc;
	bs->flow = cont ? loop_flow_statement::f_next : loop_flow_statement::f_break;
	return bs;
}

/**
	Conver the node into a type. This works since the node contains
	as much information as the type -- though it must validate that only type-correct
	specs have been used.
*/
shared_ptr<type_spec> node_converter::convert_to_type( shared_ptr<node> args,
	sig_assigns_t * assigns )
{
	auto type = make_shared<type_spec>();
	if( args->form != node::f_implied ) {
		type->set<pt_fun_class>( intr_type::f_value_list );
	}

	for( size_t i=0; i < args->sub.size(); ++i )
	{
		auto ex = args->sub[i];
		
		if( ex->form == node::f_symbol ) {
			//
		} else if( ex->form == node::f_funccall ) {
			STD_FUNCTION( *ex, "⁑assign" );
			NODE_CHECK( args, assigns, "default value not allowed here" );
			NODE_CHECK( args, args.sub.size() == 2 );
			ex = args.sub[0];
		
			sig_assign sa;
			sa.index = i;
			sa.value = convert_expression( *args.sub[1] );
			assigns->push_back(sa);
		} else {
			NODE_FAIL( *args, "invalid type syntax" );
		}
		
		type->push_sub( ex->text, convert_type(ex->type, ct_create_empty) );
	}
	
	return type;
}

/**
	Convert what we though was an node into a type.
*/
shared_ptr<type_spec> node_converter::convert_to_type( 
	shared_ptr<type_spec> type,
	shared_ptr<node> args,
	shared_ptr<node> returns, sig_assigns_t & arg_assigns )
{
	NODE_CHECK( *args, !type->has<pt_fun_class>() );
	NODE_CHECK( *args, type->sub.empty() );
	type->set<pt_fun_class>( intr_type::f_function );
	type->push_sub( convert_to_type( args, &arg_assigns ) );
	type->push_sub( convert_to_type( returns, nullptr ) );
	
	return type;
}

shared_ptr<expression> node_converter::convert_funcdefn_expression( node const & src ) {
	NODE_CHECK( src, src.sub.size() == 3 );
	
	auto ef = make_shared<expr_funcdefn>();
	sig_assigns_t assigns;
	if( src.type ) {
		ef->func_type = convert_type(src.type, ct_create_empty);
		//look for special symbols
		for( auto iter = ef->func_type->get_parts().begin(); iter != ef->func_type->get_parts().end(); ) {
			if( iter->type == pt_symbol && boost::get<type_spec_symbol>(iter->value).primary() == "clone" ) {
				ef->clone = true;
				iter = ef->func_type->get_parts().erase(iter);
			} else {
				++iter;
			}
		}
	} else {
		ef->func_type = make_shared<type_spec>();
	}
	convert_to_type( ef->func_type, src.sub[0], src.sub[1], assigns );
	ef->block = create_statement_block( *src.sub[2], scope_style_t::imperative );
	
	for( auto & sa : assigns ) {
		NODE_CHECK( src, sa.index < ef->func_type->sub[0].type->sub.size() );
		ef->arg_defaults[sa.index] = sa.value;
	}
	
	return ef;
}

shared_ptr<statement> node_converter::convert_import_statement( node const & list ) {
	NODE_CHECK( list, list.sub.size() == 2 );
	auto & src = *list.sub[1];
	
	NODE_CHECK( src, src.sub.size() == 0 );
	NODE_CHECK( src, src.form == node::f_symbol );
	
	auto is = make_shared<import_statement>();
	is->mod_name = src.text;
	
	return is;
}

shared_ptr<for_statement> node_converter::convert_for_statement( node const  & list ) {
	NODE_CHECK( list, list.sub.size() == 5 );
	NODE_CHECK( list, list.sub[2]->text == "in" );
	auto & item = *list.sub[1];
	auto & iterable = *list.sub[3];
	auto & block = *list.sub[4];
	
	auto fs = make_shared<for_statement>();
	NODE_CHECK( item, item.form == node::f_symbol );
	fs->item = item.text;
	fs->iterable = convert_expression( iterable );
	fs->block = create_statement_block( block, scope_style_t::imperative );
	
	return fs;
}

shared_ptr<statement> node_converter::convert_conditional_statement( node const & list ) {
	
	auto cond = convert_expression( *list.sub[0] );
	auto ds = make_shared<do_statement>();
	ds->src_loc = list.src_loc;

	auto convert_subblock = [&]( size_t from, size_t to ) -> shared_ptr<statement_block> {
		
		shared_ptr<statement> st;
		if( to - from == 1) {
			if( list.sub[from]->form == node::f_block ) {
				return create_statement_block( *list.sub[from], scope_style_t::imperative );
			}
			st = convert_statement( *list.sub[from] );
		} else {
			auto copy = list;
			copy.sub.erase( copy.sub.begin() + to, copy.sub.end() );
			copy.sub.erase( copy.sub.begin(), copy.sub.begin() + from );
			st = convert_statement( copy );
		}
		
		auto block = make_shared<statement_block>(scope_style_t::imperative);
		block->src_loc = st->src_loc;
		block->list.push_back( st );
		return block;
	};
	
	int then_at = -1, else_at = -1;
	for (size_t i=1; i < list.sub.size(); ++i ) {
		auto t = list.sub[i]->text;
		if (t == "then") {
			NODE_CHECK( *list.sub[i], then_at == -1, "no-conditional-nesting" );
			then_at = (int)i;
		} else if (t == "else") {
			NODE_CHECK( *list.sub[i], else_at == -1, "no-conditional-nesting" );
			else_at = (int)i;
		}
	}
	
	if( then_at != -1) {
		do_statement::expr_list_item eli;
		eli.cond = cond;
		eli.block = convert_subblock( then_at + 1, else_at == -1 ? list.sub.size() : else_at );
		
		ds->expr_list.push_back( eli );
	}
	
	if( else_at != -1 ) {
		if( then_at == -1 ) {
			//create empty then
			do_statement::expr_list_item eli;
			eli.cond = cond;
			eli.block = make_shared<statement_block>(scope_style_t::imperative);
			ds->expr_list.push_back(eli);
		}
		
		do_statement::expr_list_item eli;
		//no cond for else
		eli.block = convert_subblock( else_at+1, list.sub.size() );
		ds->expr_list.push_back( eli );
	}
	
	return ds;
}

shared_ptr<type_spec> node_converter::convert_type( node_type_spec const & nts, unsigned flags ) {
	if( !nts ) {
		return (flags & ct_create_empty) ? make_shared<type_spec>() : shared_ptr<type_spec>();
	}

	auto r = make_shared<type_spec>();
	*r = *nts.parsed_type_spec();
	convert_type_spec(*r);
	return r;
}

void node_converter::convert_type_spec( type_spec & ts) {
	for( auto & p : ts.get_parts() ) {
		if( p.node_expr ) {	
			p.expr = convert_expression(*p.node_expr);
			p.node_expr = shared_ptr<node>();
		}
	}
	
	for( auto & s : ts.sub ) {
		s.type = convert_type( s.type );
	}
}

} //eon leaf