#include "parser/node_parser.hpp"
#include "parser/intr_type_parser.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

#include "regex_patterns.hpp"
#include "util/unicode.hpp"

namespace
{
	enum precedence_level
	{
		pl_on = 10,
		pl_assign = 50,
		pl_then = 75,
		pl_compare = 100,
		pl_binary = 200,
		pl_unary = 300,
		//refer to unit_tests/block/tuple/call_prec.lfb to see equal precidence justification
		pl_function = 400,
		pl_field = 400,
	};
} //eon anon

namespace leaf
{

typedef boost::basic_regex<src_char_t> regex;

/*static*/ shared_ptr<node> node_parser::parse_as_block( source & src )
{
	try {
		return parse_compound( src, 0, 0, node::f_block );
	} catch( recoverable_error & re ) {
		re << error_parser(true);
		throw;
	}
}

/*static*/ shared_ptr<node> node_parser::parse_single( source & src )
{
	try {
		node_parser ep;
		return ep.parse( src, line_ends_terminate );
	} catch( recoverable_error & re ) {
		re << error_parser(true);
		throw;
	}
}

void node_parser::push_op( op const & o )
{
	DEBUG( "push_op: ", o.text );

	is_value = false;

	while( true )
	{
		if( op_stack.empty() )
		{
			op_stack.push_back( o );
			return;
		}

		//if higher precendence then just push and keep processing
		if( o.precedence > op_stack.back().precedence )
		{
			op_stack.push_back( o );
			return;
		}
		//if equal precedence then check left/right associativity
		if( o.precedence == op_stack.back().precedence
			&& o.right_assoc )
		{
			op_stack.push_back( o );
			return;
		}

		pop_op();
		is_value = false;
	}
}

shared_ptr<node> node_parser::pop_out()
{
	STATE_CHECK( out_stack.size() > 0 );
	shared_ptr<node> ex = out_stack.back();
	out_stack.pop_back();
	return ex;
}

void node_parser::pop_op()
{
	op o = op_stack.back();
	op_stack.pop_back();

	DEBUG( "pop_op: ", o.text );

	auto fex = make_shared<node>( node::f_funccall, o.src_loc );
	fex->sub.resize( 2 );

	fex->sub[0] = make_shared<node>( node::f_symbol, o.src_loc );
	fex->sub[0]->text = o.text;
	
	switch( o.form )
	{
		case op::f_ternary:
		{
			STATE_CHECK( out_stack.size() >= 3, "ternary operator requires three arguments" );

			auto p3 = pop_out();
			auto p2 = pop_out();
			auto p1 = pop_out();
			
			auto part = make_shared<node>( node::f_value_list, p1->src_loc );
			part->sub.resize(3);
			part->sub[0] = p1;
			part->sub[1] = p2;
			part->sub[2] = p3;
			fex->sub[1] = part;

			push_out(fex);
			break;
		}
		
		case op::f_binary:
		{
			//normalize to a function
			STATE_CHECK( out_stack.size() >= 2, "binary operator requires two arguments",
				util::source_location::error( o.src_loc ) );

			auto p2 = pop_out();
			auto p1 = pop_out();

			//convert special functions into switch
			//the ?| becomes ternary when used together
			if( o.text == "⁑optional_or" && p1->form == node::f_funccall
				&& p1->sub.size() > 0 && p1->sub[0]->text == "⁑optional_cond" ) {
				DEBUG( "TCELSE", p1->form, p1->sub[0]->text, p1->sub.size() );
				p1->sub[0]->text = "⁑ifelse";
				p1->sub[1]->sub.push_back( p2 );
				fex = p1;
			} else if( o.text == "named" ) {
				fex->form = node::f_named;
				fex->text = "named";
				fex->sub[0] = p1;
				fex->sub[1] = p2;
			} else {
				auto part = make_shared<node>( node::f_value_list, p1->src_loc );
				part->sub.resize( 2 );
				part->sub[1] = p2;
				part->sub[0] = p1;
				fex->sub[1] = part;
			}
			push_out( fex );
			break;
		}

		case op::f_unary:
		{
			fex->sub[1] = make_shared<node>( node::f_value_list, o.src_loc );
			auto p = pop_out();
			fex->sub[1]->sub.resize( 1 );
			fex->sub[1]->sub[0] = p;

			push_out( fex );
			break;
		}

		case op::f_funccall:
		{
			STATE_CHECK( out_stack.size() >= 2, "function operator requires function and tuple" );
				
			fex->sub[1] = pop_out();
			fex->sub[0] = pop_out();

			push_out( fex );
			break;
		}
	}
}

void node_parser::push_out( shared_ptr<node> const & ex )
{
	DEBUG( "push_out", (int)ex->form, ex->text, out_stack.size() );
	out_stack.push_back( ex );
	is_value = true;
}

void node_parser::push_value( shared_ptr<node> const & value ) {
	//pending stack must be collapsed to allow multiple expressions
	if( is_value ) {
		while( op_stack.size() )
			pop_op();
	}
	push_out( value );
}
	
namespace
{
	/**
		A static table of binary operators. Note how awkward it is to do this type of
		thing in C++.
	*/
	struct operator_map
	{
		struct op
		{
			char const * symbol;
			char const * alt_symbol;
			char const * name;
			unsigned precedence;
			bool right_assoc;
		};

		struct bstrcmp
		{
			bool operator()( char const * a, char const * b ) const
			{ return strcmp( a, b ) < 0; }
		};

		typedef std::map<char const*, op,bstrcmp> ops_map_t;
		ops_map_t ops_map;

		operator_map( std::initializer_list<op> base );

		op const & get( char const * text, util::source_location loc ) const {
			ops_map_t::const_iterator it = ops_map.find( text );
			STATE_CHECK( it != ops_map.end(), "unknown operator",
					node_parser::error_symbol( text ), util::source_location::error(loc) );
			return it->second;
		}
		
		bool has( char const * text ) const {
			return ops_map.find( text ) != ops_map.end();
		}

	};

	operator_map::operator_map( std::initializer_list<op> base )
	{
		for( auto it = base.begin(); it != base.end(); ++it )
		{
			ops_map[it->symbol] = *it;
			if (it->alt_symbol) {
				ops_map[it->alt_symbol] = *it;
			}
		}
	}
	
	operator_map binary_operators({
		{ "=", nullptr, "⁑assign", pl_assign, false }, //though in C it is right associative I don't want to support chaining of = (such as a=b=c;)
		{ ":=", nullptr, "⁑assign_value", pl_assign, false },
		{ "|=", nullptr, "⁑assign_default", pl_assign, false },

		{ ".", nullptr, "⁑field", pl_field, false },
		{ "?.", nullptr, "⁑optional_field", pl_field, false },
		{ "#", nullptr, "⁑subscript", pl_field, false },
		{ "?#", nullptr, "⁑optional_subscript", pl_field, false },
		//above subscript to allow common syntax: item#x..y
		{ "..", nullptr, "create_range", pl_field+1, false }, 
		
		{ "?", nullptr, "⁑optional_cond", pl_then, true },
		{ "|", nullptr, "⁑optional_or", pl_then-1, true },

		{ "or", "∨", "or", pl_compare-1, false },
		{ "and", "∧", "and", pl_compare-1, false },
		{ "xor", "↮", "xor", pl_compare-1, false },
		{ "and_not", "↛", "and_not", pl_compare-1, false }, //material nonimplication
		{ "nor", "↓", "nor", pl_compare-1, false },
		{ "nand", "↑", "nand", pl_compare-1, false },
		{ "==", nullptr, "eq", pl_compare, false },
		{ "!=", nullptr, "neq", pl_compare, false },
		{ "<", nullptr, "lt", pl_compare, false },
		{ ">", nullptr, "gt", pl_compare, false },
		{ "<=", nullptr, "lte", pl_compare, false },
		{ ">=", nullptr, "gte", pl_compare, false },
		
		{ "+", nullptr, "add", pl_binary, false },
		{ "-", nullptr, "sub", pl_binary, false },
		{ "++", nullptr, "concat", pl_binary, false },
		{ "*", nullptr, "mul", pl_binary + 10, false },
		{ "/", nullptr, "div", pl_binary + 10, false },
		{ "^", nullptr, "pow", pl_binary + 50, true },
		
	});

	operator_map unary_operators({
		{ "not", "¬", "not", pl_compare-1, true },
		{ "-", nullptr, "neg", pl_unary, true },
		{ "else", nullptr, "else", pl_then, true },
		{ "trap", nullptr, "trap", pl_assign+1, true },
	});
} //eon anon

node_parser::op node_parser::binary_op( std::string const & text, util::source_location loc )
{
	op o( loc );
	o.form = op::f_binary;

	if( text == "<-" ) {
		o.precedence = pl_function;
		o.form = op::f_funccall;
		o.text = text;
	} else {
		auto  & sop = binary_operators.get( text.c_str(), loc);
		o.precedence = sop.precedence;
		o.right_assoc = sop.right_assoc;
		o.text = sop.name;
	}

	return o;
}

node_parser::op node_parser::unary_op( std::string const & text, util::source_location loc )
{
	op o( loc );
	o.form = op::f_unary;

	auto & sop = unary_operators.get( text.c_str(), loc );
	o.precedence = sop.precedence;
	o.right_assoc = sop.right_assoc;
	o.text = sop.name;

	return o;
}

/*static*/ shared_ptr<node> node_parser::parse_compound( source & src,
	src_symbol_t start, src_symbol_t end, node::form_t ct )
{
	DEBUG( "parse_compound", end );
	auto tuple = make_shared<node>( ct, src.cur_location() );

	if( start && (!src.has_next(source::skip_line_ends) || src.next() != start) )
		src.throw_expecting( "compound" );
	
	unsigned flags = 0;
	if( ct == node::f_block ) {
		flags |= line_ends_terminate;
	}
	if( ct == node::f_param_list ) {
		flags |= param_ends_terminate;
	}
		
	while( true )
	{
		if( !src.has_next() ) {
			if( !end )
				break;
			src.throw_unexpected( "end-of-compound" );
		}

		//check for completion
		auto peeked_at = src.cur_loc();
		auto nc = src.peek_next();
		if( nc == end ) {
			src.next();
			break;
		}
		
		//was it an item, and what is the block type
		if( nc == ';' || (ct == node::f_block && nc == '\n') ) {
			if( ct != node::f_block )
				src.throw_invalid( "statement-out-of-block" );
			tuple->form = node::f_block;
			
			src.next();
			src.skip_space(source::skip_line_ends);
			continue;
		}
		if( nc == ',' ) {
			if( ct == node::f_block )
				src.throw_invalid( "expression-in-block" );
			
			src.next();
			src.skip_space(source::skip_line_ends);
			continue;
		}
		
		if( nc == '\n' ) {
			src.next();
			continue;
		}
		
		node_parser np;
		auto part = np.parse( src, flags );
		tuple->sub.push_back( part );
		
		//this appeared with an unmatched-end tag, like ) inside a { block.
		//TODO: fix that particular case to report a proper error, leave this generic check in place though
		if( src.cur_loc() == peeked_at ) {
			src.throw_unexpected( "parser-stuck-parsing-nothing" );
		}
	}

	return tuple;
}

bool node_parser::parse_single_string( src_symbol_t nc, source & src  ) {
	auto ex = make_shared<node>( node::f_literal_string, src.cur_location() );
	ex->text = parse_string( src, nc );
	if( nc == '\'' ) {
		ex->form = node::f_literal_char;
		if( unicode::len8( ex->text ) != 1 )
			src.throw_expecting( "char constant" );
	}
	push_value( ex );
	return true;
}

bool node_parser::parse_single_operator( std::string opstr, source & src ) {
	//indicate type
	if( opstr == ":" ) {
		if( out_stack.size() == 0 ) {
			//introduce anonymous typed node
			push_value( make_shared<node>(node::f_symbol, src.cur_location() ) );
		}
		out_stack.back()->type = intr_type_parser::parse( src );
	}
	
	//this could be a function call, or a value
	else if( opstr == "(" ) {
	
		if( is_value ) {
			//is it a function call (this is treated as a binary operator of the function and its arguments as a tuple)
			op o( src.cur_location() );
			o.form = op::f_funccall;
			o.precedence = pl_function;
			o.text = "<-";
			push_op( o );
		}
		
		//parses like a tuple but is a value-list
		shared_ptr<node> ex = parse_compound( src, 0, ')', node::f_value_list );
		push_value( ex );
	}
	
	//otherwise we could expect a prefix unary operator. Use the same regex however to
	//also avoid operator ambiguity
	else if( opstr == "->" ) {
		auto func = make_shared<node>( node::f_funcdefn, src.cur_location() );
		
		//specified or default empty args
		if( out_stack.size() && out_stack.back()->form == node::f_value_list ) {
			func->sub.push_back( out_stack.back() ); 
			out_stack.pop_back();
		} else {
			auto args = make_shared<node>( node::f_value_list, src.cur_location() );
			func->sub.push_back( args );
		}
		
		src.skip_space();
		if( src.peek_next() == '(' ) {
			auto returns = parse_value_list( src );
			func->sub.push_back( returns );
		} else {
			auto returns = make_shared<node>( node::f_implied, src.cur_location() );
			func->sub.push_back( returns );
		}
		
		func->type = intr_type_parser::if_parse_inline( src );
		
		auto block = parse_block( src );
		func->sub.push_back( block );
		push_out( func ); //not push_value
		
	} else if( is_value && binary_operators.has( opstr.c_str() ) ) {
		//is_value also checked to allow shared prefix operators (like -)
		push_op( binary_op( opstr, src.cur_location() ) );
		
	} else {
		push_op( unary_op( opstr, src.cur_location() ) );
	}
	
	return true;
}

bool node_parser::parse_single_special( std::string special, source & src ) {
	if( special =="on" ) {
		
		auto xon = make_shared<node>( node::f_on, src.cur_location() );
		xon->sub.push_back( collect_out(src.cur_location()) );
		if( !parse_single_token(src,line_ends_terminate) || !parse_single_token(src,line_ends_terminate) ) {
			src.throw_expecting( "on-fail-syntax" );
		}
		if( out_stack.size() != 2 ) {
			src.throw_unexpected( "on-fail-syntax-tokens" );
		}
		
		xon->sub.push_back( out_stack[0] );
		xon->sub.push_back( out_stack[1] );
		out_stack.clear();
		
		push_value( xon );
		
	} else if( special == "new" ) {
		auto nt = make_shared<node>( node::f_type, src.cur_location() );
		nt->type = intr_type_parser::parse_inline( src );
		
		auto fex = make_shared<node>( node::f_symbol, src.cur_location() );
		fex->text = "new";
		fex->sub.push_back( nt );
		
		auto ex = parse_compound( src, '(', ')', node::f_value_list );
		fex->sub.push_back( ex );
	
		push_value(fex);
		
	} else {
		STATE_FAIL( "failed regex" );
	}
	return true;
}

bool node_parser::parse_single_token( source & src, unsigned flags ) {
	if( !src.has_next() ) {
		return false;
	}

	auto nc = src.peek_next();
	
	//a line-end terminates the current expression
	if( nc == '\n' ) {
		src.next();
		if( (flags & line_ends_terminate) )
			return false;
		return true;
	}
	
	if( nc == sym_close_param && (flags & param_ends_terminate) ) {
		return false;
	}
	
	if( nc == ',' || nc == ';' || nc == '@' || nc == ']' || nc == ')' || nc == '}' ) {
		return false;
	}
	
	if( nc == '{' ) {
		push_value( parse_block( src ) );
		return true;
	} 
	
	if( nc == '[' ) {
		push_value( parse_tuple( src ) );
		return true;
	}
	
	if( nc == '\"' || nc == '\'' ) {
		return parse_single_string( nc, src );
	}
	
	if( nc == sym_open_param ) {
		if( !is_value) {
			src.throw_unexpected( "param-type" );
		}
		auto q = intr_type_parser::parse_inline_param( src );
		if( out_stack.back()->param ) {
			src.throw_unexpected( "duplicate-param-type" );
		}
		out_stack.back()->param = q;
		return q;
	}
	
	//paired symbols are however consume one-at-a-time
	bool is_operator = src.match( leaf_parse::re_operator );
	if( is_operator ) {
		return parse_single_operator( src.cur_match(), src );
	}
	
	if( src.match( leaf_parse::re_special ) ) {
		return parse_single_special( src.cur_match(), src );
	}

	//requires special identification to avoid a trailing () to be identified as a function call
	static boost::regex re_decl( "(return|var|defn|multi|ignore|new|class|service)" EXCLUDE_SYMBOL_REGEX_CHAR );
	
	if( src.match( leaf_parse::re_bool ) ) { //must be before symbol
		auto ex = make_shared<node>( node::f_literal_boolean, src.cur_location());
		ex->text = src.cur_match();
		push_value( ex );
		
	} else if( src.match( re_decl ) ) {
		auto symbol = src.cur_match();
		
		//special parsing
		if( flags & first_token ) {
			if( symbol == "class" || symbol == "service" ) {
				parse_class( symbol, src );
				return false;
			}
		}
		auto ex = make_shared<node>( node::f_symbol, src.cur_location());
		ex->text = symbol;
		push_value( ex );
		is_value = false;
		
	} else if( src.match( leaf_parse::re_symbol ) ) {
		auto symbol = src.cur_match();
		auto ex = make_shared<node>( node::f_symbol, src.cur_location());
		ex->text = symbol;
		push_value( ex );
		
	} else if( src.match( leaf_parse::re_base_number ) ) {
		auto ex = make_shared<node_literal_number>(src.cur_location());
		unsigned f = 0;
		char c = src.cur_match(1)[0];
		if( c == 'x' )
			ex->value = util::number::parse_hexadecimal( src.cur_match(2), f );
		else if( c == 'o'  )
			ex->value = util::number::parse_octal( src.cur_match(2), f );
		else if( c == 'b'  )
			ex->value = util::number::parse_binary( src.cur_match(2), f );
		ex->was_float = f & util::number::pdf_had_decimal;
		push_value( ex );
		
	} else if( src.match( leaf_parse::re_number ) ) {
		auto ex = make_shared<node_literal_number>(src.cur_location());
		unsigned f;
		ex->value = util::number::parse_decimal( src.cur_match(), f );
		ex->was_float = f & util::number::pdf_had_decimal;
		push_value( ex );
		
	} else {
		src.throw_expecting( "value" );
	}
	
	return true;
}


/**
	Parsing of a single statement. Though some parts are recursive this outer
	wrapper has a specific form distinct to that (allowing additional markers
	like declarators and annotations).
*/
shared_ptr<node> node_parser::parse( source & src, unsigned flags ) {
	reset();
	
	src.skip_space(source::skip_line_ends);
	
	//pre-annotaitons
	node::annotate_map_t anno;
	parse_annotations( src, anno );
	src.skip_space( source::skip_line_ends );
	auto decl_loc = src.cur_location();
	
	bool is_first_token = true;
	while( parse_single_token( src, flags | (is_first_token ? first_token :0 ) ) ) {
		is_first_token = false;
	}

	auto result = collect_out( decl_loc );
	
	result->annotations.swap(anno);
	
	//from start
	result->src_loc = decl_loc;
	return result;
}

shared_ptr<node> node_parser::collect_out(util::source_location loc) {
	while( op_stack.size() )
		pop_op();
		
	shared_ptr<node> result;
	
	if( out_stack.size() == 1 ) {
		//simple single node result
		result = out_stack[0];
	} else {
		//expression list result
		result = make_shared<node>( node::f_expr_list, loc );
		result->sub.swap(out_stack);
	}
	
	out_stack.clear();
	return result;
}

void node_parser::parse_annotations( source & src, node::annotate_map_t & anno ) {
	//check annotations
	while( src.has_next() && src.peek_next() == '@' ) {
		static regex re_anno( "@(" SYMBOL_REGEX ")" );
		if( !src.match( re_anno ) )
			src.throw_expecting( "annotation" );
		std::string name = src.cur_match( 1 );

		//value-list used for annotation value
		if( src.has_next() && src.peek_next() == '(' ) {
			anno[name] = parse_value_list( src );
		} else {
			anno[name] = make_shared<node>(node::f_value_list, src.cur_location() );
		}
	}
}

/**
	The delimiters and escape sequences are handled as individual characters even
	though we may are parsing utf-8 files. This is okay since all these ASCII chars
	are represented exactly the same in utf-8, with no ambiguity.
	
	The supported escapes are a common set coming from C,C++,Java,Python.
*/
/*static*/ std::string node_parser::parse_string( source & src, src_symbol_t end )
{
	if( !src.has_next() )
		src.throw_expecting( "string open-quotes" );
	src.next();
	
	//parse fixed length hex sequences
	auto parse_hex = [&](int len) {
		unsigned val = 0;
		for( int i=0; i < len; ++i ) {
			val <<= 4;
			auto v = src.next();
			if( v >= '0' && v <= '9' ) {
				val += (v-'0');
			} else if( v >= 'a' && v <= 'f' ) {
				val += (v-'a') + 10;
			} else if( v >= 'A' && v <= 'F' ) {
				val += (v-'A') + 10;
			} else {
				src.throw_invalid( "escape-sequence" );
			}
		}
		return val;
	};
	
	std::string ret;
	while( src.has_raw_next() )
	{
		auto n = src.raw_next();
		if( n == end )
			return ret;

		if( n == '\\' ) {
			auto sn = src.next();
			if( sn == '0' ) {
				ret += '\0';
			} else if( sn == '\'' ) {
				ret += char(0x27);
			} else if( sn == '\"' ) {
				ret += char(0x22);
			} else if( sn == '\\' ) {
				ret += char(0x5c);
			} else if( sn == 'a' ) {
				ret += char(0x07);
			} else if( sn == 'b' ) {
				ret += char(0x08);
			} else if( sn == 'f' ) {
				ret += char(0x0c);
			} else if( sn == 'n' ) {
				ret += char(0x0a);
			} else if( sn == 'r' ) {
				ret += char(0x0d);
			} else if( sn == 't' ) {
				ret += char(0x09);
			} else if( sn == 'v' ) {
				ret += char(0x0b);
			} else if( sn == 'x' ) {
				ret += unicode::encode8(parse_hex(2));
			} else if( sn == 'u' ) {
				ret += unicode::encode8(parse_hex(4));
			} else if( sn == 'U' ) {
				ret += unicode::encode8(parse_hex(8));
			} else {
				src.throw_unexpected( "escape-sequence" );
			}
		} else {
			ret += n;
		}
	}

	src.throw_expecting( "string close-quotes" );
	return "not reached";
}

void node_parser::parse_class( std::string style, source & src ) {
	auto csym = make_shared<node>( node::f_symbol, src.cur_location() );
	csym->text = style;
	push_out(csym);
	
	if( !src.has_next() || !src.match( leaf_parse::re_symbol ) ) {
		src.throw_expecting( "class-name" );
	}
	auto sym = make_shared<node>( node::f_symbol, src.cur_location() );
	sym->text = src.cur_match();
	push_out(sym);
	
	src.skip_space();
	auto nc = src.peek_next();
	if( nc == sym_open_param ) {
		auto params = parse_param_list( src );
		push_out( params );
	}
	
	auto block = parse_block( src );
	push_out( block );
}

} //eon leaf