#include "runner/runner.hpp"
#include "runner/expecter.hpp"

#include <boost/algorithm/string/predicate.hpp>
#include <boost/filesystem.hpp>
namespace fs = boost::filesystem;

#include <iostream>
#include <fstream>

#include <dlfcn.h>
#include <unistd.h>

// Root Leaf data files (these are generated files, in the build directory)
#include "runner/base.leaf.hpp"

#define SHOW_DEBUG 0
#include "util/debug.hpp"

shared_ptr<ir::module> runner::build_module() 
{
	shared_ptr<ir::module> m;
	try {
		STATE_CHECK( input_files.size() );
		auto & first_file = input_files[0];
		
		if( boost::ends_with( first_file, ".lfx" ) ) {
			STATE_CHECK( input_files.size() == 1, "only one lfx file supported as input" );
			m = get_expr();
		} else if( boost::ends_with( first_file, ".lfb" ) ) {
			STATE_CHECK( input_files.size() == 1, "only one lfb file supported as input" );
			m = get_block();
		} else if( boost::ends_with( first_file, ".leaf" ) ) {
			m = get_full();
		} else {
			STATE_FAIL( "unknown input-file type" );
		}
	} catch( recoverable_error & re ) {
		if( expect ) {
			expect->catch_compile_exception(re);
			return shared_ptr<ir::module>();
		}
		throw;
	}
	
	if( dump_ir ) {
		ir::dump d( std::cout );
		d.set_show_all_types( dump_ir_types );
		d.write( m.get() );
	}
	return m;
}

void runner::x_dump_parse( leaf::node const & n )
{
	if( dump_parse_flat )
		std::cout << leaf::node_dump::dump_flat(n) << std::endl;
		
	if( !dump_parse )
		return;
	
	leaf::node_dump nd( std::cout );
	nd.header( "// -- Parse Tree --" );
	nd.dump( n );
	std::cout << std::endl;
}

void runner::x_dump_post_type( leaf::module const & mod ) {
	if( !dump_post_type )
		return;
		
	std::cout << "// -- Post-Typing Leaf Tree --" << std::endl;
	leaf::dump d( std::cout );
	d.set_show_all_types( dump_all_types );
	d.write_module( mod );
}

void runner::x_dump_pre_type( std::vector<shared_ptr<leaf::statement>> const & sb )
{
	if( !dump_pre_type )
		return;
		
	std::cout << "// -- Pre-Typing Leaf Tree --" << std::endl;
	leaf::dump d( std::cout );
	d.set_show_spec_type( true );
	d.set_validate( false );
	for( auto & s : sb ) {
		d.write_statement( *s );
		d.line_break();
	}
	std::cout << std::endl;
}

shared_ptr<leaf::expression> runner::parse_expr( std::string const & expr )
{
	auto src = leaf::source::create( expr );

	shared_ptr<leaf::node> n = leaf::node_parser::parse_single( *src );
	x_dump_parse( *n );
	
	auto ex = leaf::node_converter::expression_convert( *n );
	if( dump_pre_type )
	{
		leaf::dump d( std::cout );
		d.set_show_spec_type( true );
		d.set_validate( false );
		d.write( ex );
		std::cout << std::endl;
	}
	
	//the expression must be converted to its inferred type to ensure it is representable as
	//an actual compiled type (such as removing rationals)
	ex = module_typer->full_type( ex, leaf::typer::tef_convert_inferred );
	
	if( dump_post_type ) {
		leaf::dump d( std::cout );
		d.set_show_all_types( dump_all_types );
		d.write( ex );
		std::cout << std::endl;
	}

	return ex;
}

runner::statement_vector_t runner::load_parse_block( std::string const & filename, 
	leaf::scope_style_t scope_style, bool dump ) {
	DEBUG( "Block-Load", filename );
	auto s = leaf::source::create( util::load_file( filename ) );
	s->set_name( filename );
	
	return load_parse_block( s, scope_style, dump);
}

runner::statement_vector_t runner::load_parse_block( shared_ptr<leaf::source> src,
	leaf::scope_style_t scope_style, bool dump ) {
	
	DEBUG( "Block-Parse" );
	auto n = leaf::node_parser::parse_as_block( *src );
	DEBUG( "Block-Parse-Dump" );
	if( dump ) {
		x_dump_parse( *n );
	}
	DEBUG( "Block-Post-Parse" );
	
	auto block = leaf::node_converter::convert_block( *n, scope_style );
	if( dump ) {
		x_dump_pre_type( block );
	}
	DEBUG( "Block-Post-Convert" );
	
	return block;
}

void runner::do_typing( bool base )
{
	module_typer->full_complete_type();
	
	if( !base || dump_base ) {
		x_dump_post_type( *module );
	}
	
	if( !base && list_functions ) {
		std::cout <<  "Function List" << std::endl;
		for( auto & fd : module->funcdefns ) {
			std::string name = fd->bound_decl ? fd->bound_decl->symbol : "anonymous";
			std::cout << name << " : " << leaf::dump::get( fd->type ) << std::endl;
		}
	}
}

void runner::init_module() {
	module_typer = make_shared<leaf::typer>( module_name, mloader );
	module = module_typer->get_module();

	add_base_block( create_string_source("base.leaf",dataBaseLeaf) );
}

void runner::add_base_block( shared_ptr<leaf::source> src ) {
	auto list = load_parse_block( src, leaf::scope_style_t::declarative, false );
	auto block = module_typer->add_root_block( list );
	block->reference_text = src->get_name();
	do_typing(true);
}

shared_ptr<ir::module> runner::get_expr() {
	shared_ptr<leaf::expression> ex = parse_expr( util::load_file( input_files.at(0) ) );
	ir::compiler_leaf c( module->global );
	
	shared_ptr<ir::module> m = c.compile_expr( *ex );
	auto f = m->if_get_function( "eval_expr" );
	PRE_CHECK(f);
	main_trace_function( c, f );

	return m;
}

shared_ptr<ir::module> runner::get_block()
{
	auto list = load_parse_block( input_files.at(0), 	leaf::scope_style_t::dynamic );
	auto block = module_typer->add_dynamic_block( list );
	do_typing();
	
	ir::compiler_leaf c( module->global );
	return c.compile_block( *block, *module->static_init_block );
}

shared_ptr<ir::module> runner::get_full() {

	for( auto & file : input_files ) {
		auto list = load_parse_block( file, leaf::scope_style_t::declarative );
		auto block = module_typer->add_global_block( list );
		block->reference_text = file;
	}
	do_typing();

	ir::compiler_leaf c( module->global );
	return c.compile_module( *module );
}

int runner::execute( shared_ptr<ir::module> m ) {
	//load explicit libraries (mainly for testing)
	std::vector<void*> lib_handles;
	auto load_lib = [&]( std::string const & path ) {
		auto handle = dlopen( path.c_str(), RTLD_NOW | RTLD_GLOBAL );
		STATE_CHECK( handle, std::string( "failed loading library: " )
			+ path + ": " + dlerror() );
		lib_handles.push_back( handle );
	};
	
	for( auto & lib : libraries ) {
		load_lib( lib );
	}
	//load imported libraries
	for( auto & lib : module->libraries ) {
		load_lib( mloader->resolve_dynload(lib) );
	}

	//setup capturing of stdout
	volatile int old_stdout = 0; //GCC: complains it will get clobbered by longjmp without volatile
	char out_filename[128] = "/tmp/output_XXXXXX";
	if( capture_stdout ) {
		old_stdout = dup(STDOUT_FILENO);
 		int out_file = mkstemp(out_filename);
 		fflush( stdout );
 		
 		dup2( out_file, STDOUT_FILENO );
 	}
 	
 	auto end_redirect = [&]() {
		if( !old_stdout ) {
			return;
		}
		
 		fflush( stdout );
 		dup2(old_stdout, STDOUT_FILENO);
 	};
	auto clean_redirect = [&]() {
 		unlink(out_filename);
	};
	
	//volatile to not be clobberd by setjmp/longjmp
	volatile int return_value = 0;
	
	try {
		jmp_buf abort_point;
		int i = setjmp(abort_point);
		if( i ) {
			end_redirect();
			if( !expect || !expect->catch_runtime_abort( runtime_test::get_error() ) ) {
				throw recoverable_error( std::string( "abort: " ) + runtime_test::get_error() );
			}
		} else {
			runtime_test::prepare( &abort_point );
			ir_llvm::gen lg( gen_opts );
			int rv = lg.execute( *m, dump_llvm );
			end_redirect();
			
			runtime_test::check_state();
			if( expect ) {
				std::string output = runtime_test::get_trace();
				if( old_stdout ) {
					output = util::load_file( out_filename );
				}
				//DEBUG( "OUTPUT:", output );
				expect->check_result( output, rv );
			} else {
				return_value = rv;
			}
		}
		
		clean_redirect();
	} catch( ... ) {
		end_redirect();
		clean_redirect();
		throw;
	}
	
	return return_value;
}

void runner::emit_object( shared_ptr<ir::module> m, std::string const & path_base, gen_what_t gw )
{
	ir_llvm::gen lg( gen_opts );
	lg.emit_object( *m, path_base, dump_llvm );
	
	if( gw == gw_shared ) {
	#if defined(IS_DARWIN)
		std::string cmd( "gcc -dynamiclib -Wl,-undefined,dynamic_lookup,-o," );
		cmd += mloader->dyn_name(path_base) + ",-init,"
			+ "_" + *module_name + platform::export_sep + "_init_module"
			//there just is no "fini" on OSX!
			//+ " " + mloader->resolve_dyn_core( "runtime" )
			+ " " + mloader->obj_name(path_base);
	#else
		std::string cmd( "gcc -shared -Wl,--eh-frame-hdr,--build-id,--export-dynamic,-o," );
		cmd += path_base + ".so,-init,"
			+ *module_name + platform::export_sep + "_init_module,-fini,"
			+ *module_name + platform::export_sep + "_final_module"
			//+ " " + mloader->resolve_dyn_core( "runtime" )
			+ "," + path_base + ".o";
	#endif
	
	#if 0
		//These work, but don't produce proper shared libraries (ldd says statically linked) (they also don't
		//contain the shared library dependencies)
		std::string cmd( "ld --eh-frame-hdr --build-id --export-dynamic -shared -o " );
		cmd += path_base + ".so -init "
			+ *module_name + platform::export_sep + "_init_module -fini "
			+ *module_name + platform::export_sep + "_final_module"
			//+ " " + mloader->resolve_dyn_core( "runtime" )
			+ " " + path_base + ".o";
	#endif
	
 		for( auto & lib : module->libraries ) {
 			cmd = cmd + " " + mloader->resolve_dynlink( lib );
 		}
		int ret = std::system( cmd.c_str() );
		STATE_CHECK( ret == 0, cmd );
	} else if( gw == gw_exe ) {
	#if defined(IS_DARWIN)
		std::string cmd( "gcc -Wl -o " );
		cmd += path_base + " " + mloader->obj_name(path_base)
			//+ " " + mloader->resolve_static_core( "runtime_support" )
			+ " " + mloader->resolve_dyn_core( "runtime" )
			;
	#else
		//NOTE: switch to g++ for rare debugging needs with iostream
		std::string cmd( "gcc -Wl,--eh-frame-hdr,--build-id,--export-dynamic -o " );
		cmd += path_base + " " + path_base + ".o"
			+ " " + mloader->resolve_static_core( "runtime_support" )
			//+ " " + mloader->resolve_dyn_core( "runtime" )
			;
	#endif
			
		for( auto & lib : module->libraries ) {
			cmd = cmd + " " + mloader->resolve_dynlink( lib );
		}
		
		int ret = std::system( cmd.c_str() );
		STATE_CHECK( ret == 0, cmd );
	}

	if( gw == gw_shared ) {
		//detach scope and create leaf object file (this breaks the module!)
		module->global->detach_base();
		std::ofstream lo( mloader->leaflib_name(path_base), std::ios::binary );
		leaf::serial_out sout( lo, module_typer->get_serial_resolver() );
		sout.set_is_exporting();
		sout.write_key_ref( leaf::serial_item::t_root, module->global );
		for( auto & ex : module->var_exports ) {
			sout.add_declaration( ex );
		}
		for( auto & ex : module->type_exports ) {
			sout.add_intr_type( ex );
		}
		sout.done();
	}
}

//TODO: more of a util function
shared_ptr<leaf::source> runner::create_string_source( std::string const & name, std::string const & content ) {
	auto s = leaf::source::create( content );
	s->set_name( name );
	return s;
}
