#define BOOST_TEST_MODULE LeafTest
#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>
#include <boost/test/unit_test_monitor.hpp>
#include <boost/exception/all.hpp>
#include <boost/test/unit_test_log_formatter.hpp>

#include <iostream>

using namespace boost::unit_test;

class my_formatter : public unit_test_log_formatter {
	std::vector<std::string> names;
public:
	virtual void log_start(std::ostream & out, counter_t amount) {
		out << "Running " << amount << " tests..." << std::endl;
	}
	
	virtual void log_finish(std::ostream & out) {
		out.flush();
	}
	
	virtual void log_build_info(std::ostream &) {
	}
	
	virtual void test_unit_start(std::ostream &, test_unit const & tu ) {
		names.push_back( tu.p_name );
	}
	
	void write_name( std::ostream & out ) {
		auto first = true;
		for (auto & str : names ) {
			if( !first ) {
				out << "/";
			}
			first = false;
			out << str;
		}
	}
	
	virtual void test_unit_finish(std::ostream & out, test_unit const & /*tu*/, unsigned long) {
		if (entries.size()) {
			out << std::endl;
			write_name( out );
			out << std::endl;
			dump_entries(out);
		}
		
		names.pop_back();
	}
	
	virtual void log_exception_start(std::ostream & out, log_checkpoint_data const & cdata, 
		boost::execution_exception const & ex) {
	
		write_name( out );
		auto & loc = ex.where();
		out << loc.m_file_name << ":" << loc.m_line_num << " somewhere in ";
		
		for (auto & str : names ) {
			out << str << ".";
		}
		out << " exception " << loc.m_function << ": " << ex.what();

		if( !cdata.m_file_name.is_empty() ) {
			out << '\n';
			out << "checkpoint @ " << cdata.m_file_name << ":" << cdata.m_line_num << ": " <<
				cdata.m_message;
		}

		out << std::endl;
	}
// 	virtual void log_exception_finish(std::ostream &) {
// 	}
	
	std::vector<std::string> entries;
	std::stringstream entry;

	void dump_entries(std::ostream & out) {
		for (auto & str : entries) {
			out << str << std::endl;
		}
		entries.clear();
	}
	
	virtual void test_unit_skipped(std::ostream & out, test_unit const & tu) {
		write_name(out);
		out << "/" << tu.p_name << " skipped" << std::endl;
	}

	virtual void log_entry_start(std::ostream &, log_entry_data const & entry_data, log_entry_types type) {
		if (type != BOOST_UTL_ET_MESSAGE) {
			switch (type ) {
				case BOOST_UTL_ET_INFO: entry << "info: "; break;
				case BOOST_UTL_ET_WARNING: entry << "warning: "; break;
				case BOOST_UTL_ET_ERROR: entry << "error: "; break;
				case BOOST_UTL_ET_FATAL_ERROR: entry << "fatal: "; break;
				case BOOST_UTL_ET_MESSAGE: break;
			}
			entry << entry_data.m_file_name << ":" << entry_data.m_line_num << ": ";
		}
	}
	
	void log_entry_value( std::ostream&, lazy_ostream const& value ) {
		entry << value;
	}
	
	
	virtual void log_entry_value(std::ostream &, const_string value) {
		entry << value;
	}
	
	virtual void log_entry_finish(std::ostream &) {
		entries.push_back(entry.str());
		entry.str("");
		entry.clear();
	}
	
	virtual void entry_context_start(std::ostream &, log_level) {
	}
	
	virtual void log_entry_context(std::ostream &, const_string) {
	}
	
	virtual void entry_context_finish(std::ostream &) {
	}
};

//otherwise boost just swallows the exceptions... :(
void boost_exception_translator( const boost::exception & /*ex*/ )
{
	//still in exception, so this works
	BOOST_MESSAGE(boost::current_exception_diagnostic_information());
	throw;
}

void terminator()
{
	std::cerr << "Unexpected exception" << std::endl;
	std::cerr << boost::current_exception_diagnostic_information();
	exit(255);
}

bool init_all() {	
	if (!init_unit_test()) {	
		return false;
	}
	
	unit_test_log.set_formatter( new my_formatter() );
	return true;
}

int main( int argc, char** argv )
{
	std::set_terminate( &terminator );

	//show exceptions
	unit_test_monitor.register_exception_translator<boost::exception>( 
		&boost_exception_translator );
	
	//force boost to show entry/exit (otherwise error output is hard to track)
	setenv( "BOOST_TEST_LOG_LEVEL", "test_suite", 0 );
	return unit_test_main( &init_all, argc, argv );
}

