/**
	Tests of expression typing. This stays are the expression level and directly modifies
	the scope to add required variables. (Contrast to full_type_test which has no special
	access).
*/
#define SHOW_DEBUG 0
#include "expr_common.hpp"


BOOST_AUTO_TEST_SUITE( expr_type_test )

struct tester : public expr_tester {
	
	void operator()( char const * code, char const * expect, unsigned flags = 0 ) {
		expect_test et;
		et.add( "Source", code );
		et.add( "Expect", expect );
		
		try {
			shared_ptr<expression> ex = parse_expr( code );
			auto clone = ex->pre_clone();
			
			auto check = [&](shared_ptr<expression> ex, char const* msg=nullptr) {
				#if SHOW_DEBUG
				{
					std::cout << typeid(*ex).name() << std::endl;
					leaf::dump d( std::cout );
					d.set_validate(false);
					d.set_show_spec_type( true );
					d.write( ex );
					std::cout << std::endl;
				}
				#endif
				
				ex = the_typer->full_type( ex, flags );
				et.check_actual( dump::get(ex->type), expect, msg );
			};
			
			check( ex );
			check( clone, "Cloned" );
			
		} catch( ... ) {
			et.except();
		}
	}
};

BOOST_AUTO_TEST_CASE( basic )
{
	tester t;

	t( "1", "rational 1bit literal" );
	//t( "-1", "rational 2bit signed" ); //TODO: need constant folding first

	t.add_var( "a", "integer" );
	t( "a", "integer lvalue" );

	t.add_var( "b", "float" );
	t( "b", "float lvalue" );
	
	//lvalue is always lost in tuples
	t( "[a,b]", "[ : integer, : float ]" );
	t( "[[a],1]","[ : [ : integer ], : rational 1bit ]" );

	//but preserved in value-lists
	t( "(a,b)", "( : integer lvalue, : float lvalue )" );
	
	//typing does not collapse 1-value-lists, only type-conversion does
	t( "(a)", "( : integer lvalue )" );
}

BOOST_AUTO_TEST_CASE( literal ) {
	tester t;
	
	t( "[1,2]", "[ : rational 1bit, : rational 2bit ] literal" );
	
	t.add_var( "a", "integer" );
	t( "[ 1, a ]", "[ : rational 1bit, : integer ]" );
	
	t( " 'a' ", "char literal" );
	t( " false ", "boolean literal" );
	t( "\"hello\"", "^array｢ char ｣ literal" );
	t( "[ 'a', false ]", "[ : char, : boolean ] literal" );
}

BOOST_AUTO_TEST_CASE( boolean )
{
	tester t;
	
	t( "true", "boolean literal" );
	t( "false", "boolean literal" );
	
	t.add_var( "a", "boolean" );
	t( "a", "boolean lvalue" );
}

BOOST_AUTO_TEST_CASE( binary_function )
{
	tester t;

	t.add_var( "madd", "(:integer,:integer)->(:integer)" );
	t( "madd(1,2)", "integer" );

	t.add_var( "a", "integer" );
	t.add_var( "b", "integer" );
	t( "madd(madd(a,3),b)", "integer" );

	t.add_var( "foo", "(:integer,:integer)->(:float,:integer)" );
	t("foo(madd(1,2),b)","( : float, : integer )" );
}

BOOST_AUTO_TEST_CASE( core_ops )
{
	tester t;
	t( "1+2", "rational 2bit literal" ); //TODO: hmmm, perhaps just a rational literal, no size again...
}

/**
 	For rationals a normal "float" is the preferred conversion to "low/high". For fixed floats the
 	closest higher matching float. When mixed in signature the preferrence aspect is the lowest
 	priority.
*/
BOOST_AUTO_TEST_CASE( float_preference )
{
	tester t;

	t.add_var( "fs", "float low" );
	t.add_var( "fn", "float normal" );
	t.add_var( "fl", "float high" );
	
	t.add_multi( "fo", "(:float)->(:float)" );
	t.add_multi( "fo", "(:float low)->(:float low)" );
	t.add_multi( "fo", "(:float high)->(:float high)" );
	t( "fo( 1.5 )", "float" );
	t( "fo(fs)", "float low" );
	t( "fo(fl)", "float high" );
	t( "fo(fn)", "float" );
	
	t.add_multi( "foo", "(:float,:float)->(:float)" );
	t.add_multi( "foo", "(:float low,:float low)->(:float low)" );
	t.add_multi( "foo", "(:float high,:float high)->(:float high)" );
	
	t( "foo( 3.5, 4.5 )", "float" );
	//t( "foo( fs, 4.5 )", "float low" ); //undecided now, this is ambiguous in the current rules
	t( "foo( 3.5, fl )", "float high" );
	t( "foo( fs, fn )", "float" );
	t( "foo( fs, fl )", "float high" );
}

BOOST_AUTO_TEST_CASE( integer_preference )
{
	tester t;

	//use size variations large enough to hopefully be platform independent
	t.add_var( "is", "integer 16bit" );
	t.add_var( "in", "integer" );
	t.add_var( "il", "integer 128bit" );
	
	t.add_multi( "fo", "(:integer)->(:integer)" );
	t.add_multi( "fo", "(:integer 16bit)->(:integer 16bit)" );
	t.add_multi( "fo", "(:integer 128bit)->(:integer 128bit)" );
	t( "fo( 1 )", "integer" );
	t( "fo(is)", "integer 16bit" );
	t( "fo(il)", "integer 128bit" );
	t( "fo(in)", "integer" );
	
	t.add_multi( "foo", "(:integer,:integer)->(:integer)" );
	t.add_multi( "foo", "(:integer 16bit,:integer 16bit)->(:integer 16bit)" );
	t.add_multi( "foo", "(:integer 128bit,:integer 128bit)->(:integer 128bit)" );
	
	t( "foo( 3, 4 )", "integer" );
	//t( "foo( is, 4 )", "integer 16bit" ); //undecided now, currently ambiguous
	t( "foo( 3, il )", "integer 128bit" );
	t( "foo( is, in )", "integer" );
	t( "foo( is, il )", "integer 128bit" );
}

BOOST_AUTO_TEST_CASE( unsigned_preference )
{
	tester t;
	
	t.add_var( "un", "binary" );
	t.add_var( "in", "integer" );
	
	t.add_multi( "num", "(:integer)->(:integer)" );
	t.add_multi( "num", "(:binary)->(:binary)" );

	t( "num(13)", "integer" ); //integer conversion is preferred
	t( "num(un)", "binary" );
	t( "num(in)", "integer" );
	
	t.add_multi( "under", "(:binary)->(:binary)" );
	t( "under(13)", "binary" );
	
	t.add_var( "unbit32", "binary 32bit" );
	t( "num(unbit32)", "binary" );
}

BOOST_AUTO_TEST_CASE( rational_match )
{
	tester t;
	t.add_var( "madd", "(:float,:float)->(:float)" );
	t( "madd(1,2)", "float" ); //rationals kept until conversion to float
}

BOOST_AUTO_TEST_CASE( multi_match )
{
	tester t;
	t.add_multi( "madd", "( : integer, : integer)->( : integer)" );
	t.add_multi( "madd", "( : float, : float)->( : float)" );
	t( "madd(1,2)", "integer" );
	t( "madd(1.0,2.0)", "integer" ); //rationals are best match for integer still .0 does not change the value
	t( "madd(1.1,2)", "float" ); //one matching float will be enough to get float version
	t( "madd(1,2.1)", "float" );
	
 	t.add_multi( "promote", "( : integer)->( : float)" );
 	t( "madd(promote(1),2.5)", "float" );
 	
 	t.add_multi( "trunc", "(:float)->(:integer)" );
 	t( "madd(1,trunc(2.5))", "integer" );
}

BOOST_AUTO_TEST_CASE( funcdefn )
{
	tester t;
	//testing the anonymous name, not the raw bit
	t( "()->(:integer){}", "( ) -> ( : integer ) literal raw" );
}

BOOST_AUTO_TEST_CASE( tuple_names ) {
	tester t;
	t( "[ a =12 ]", "[ a : rational 4bit ] literal" );
	t( "[ 3, b = 4.5, c= [ 0 ] ]", "[ : rational 2bit, b : rational, c : [ : rational 1bit ] ] literal" );
}

BOOST_AUTO_TEST_CASE( tuple_field ) {
	tester t;
	t.add_var( "q", "[ a : integer ]" );
	t( "q.a", "integer lvalue" );
	t( "q#0", "integer lvalue" );
	
	t.add_var( "s", "[ a : integer ] shared" );
	t( "s.a", "integer lvalue" );
	t( "s#0", "integer lvalue" );
	t( "(s : [ a : integer ] shared).a", "integer lvalue" );
	
	//remove support for auto-optional derefencing
	/*t.add_var( "t", "[ a : integer ] optional" );
	t( "t.a", "integer lvalue" );
	t( "t#0", "integer lvalue" );
	
	t.add_var( "u", "[ a : integer ] shared optional" );
	t( "u.a", "integer lvalue" );
	t( "u#0", "integer lvalue" );*/
	
	t.add_var( "v", "[ a : integer ] value_ptr" );
	t( "v.a", "integer lvalue" );
	t( "v#0", "integer lvalue" );
	//TODO: an explicit "drop_lvalue" command since we're still not 100% this cast  should remove the lvalue
	t( "(v : [ a : integer ] value_ptr).a", "integer lvalue" );
	
}

BOOST_AUTO_TEST_CASE( strings ) {
	tester t;
	t( " \"hello\" ", "^array｢ char ｣ literal" );
	
	t( " 'b' ", "char literal" );
}

BOOST_AUTO_TEST_CASE( subscript ) {
	tester t;
	t( "[1,2]#1", "rational 2bit literal" );
	
	t.add_var( "arr", "array<:float:>" );
	t( "arr#5", "float lvalue" );
	t.add_var( "index", "integer" );
	t( "arr#index", "float lvalue" );
}

BOOST_AUTO_TEST_CASE( unary ) {
	tester t;
	t.add_var( "a", "integer 15bit" );
	t( "-a", "integer 15bit" );
	
	t.add_var( "b", "float high" );
	t( "abs(b)", "float high" );
	
	t.add_var( "c", "binary 39bit" );
	t( "not c", "binary 39bit" );
}

BOOST_AUTO_TEST_CASE( conditional ) {
	tester t;
	t.add_var( "a", "integer" );
	t( "true ? a", "integer convert_optional optional" );
	
	t.add_var( "b", "optional integer" );
	t( "b | 5", "integer" );
	
	//TODO: this has convert_optional, it's unclear if it should
	//t( "true ? b", "optional integer" );
	
	t.add_var( "c", "boolean" );
	t( "c ? a | 3", "integer" );
}

BOOST_AUTO_TEST_CASE( special_field ) {
	tester t;
	t.add_var( "arr", "array<:float:>" );
	t( "arr.size", "integer" );
	
	t.add_var( "opt", "optional float" );
	t( "has(opt)", "boolean" );
}

BOOST_AUTO_TEST_CASE( class_field ) {
	tester t;
	auto ctx = t.add_class( "pine" ).context;
	
	auto sd = declaration::create();
	sd->form = declaration::f_single;
	sd->symbol = "needle";
	sd->type = t.determine_type( "integer" );
	ctx->add(sd );
	
	sd = declaration::create();
	sd->form = declaration::f_single;
	sd->symbol = "cone";
	sd->type = t.determine_type( "()->() context(pine)" );
	ctx->add(sd );
	
	sd = declaration::create();
	sd->form = declaration::f_single;
	sd->symbol = "twig";
	sd->type = t.determine_type( "(:integer)->(:float) context(pine)" );
	ctx->add(sd );
	
	t.add_var( "p", "pine" );
	t( "p", "pine lvalue" );
	t( "p.needle", "integer lvalue" );
	t( "p.cone", "contextual｢ pine lvalue, ( ) -> ( ) context(pine) lvalue unbound ｣" );
	t( "p.cone()", "( )" );
	t( "p.twig(12)", "float" );
	
	sd = declaration::create();
	sd->form = declaration::f_multi;
	sd->symbol = "br";
	sd->type = t.determine_type( "(:integer)->() context(pine)" );
	ctx->add(sd );
	
	sd = declaration::create();
	sd->form = declaration::f_multi;
	sd->symbol = "br";
	sd->type = t.determine_type( "()->() context(pine)" );
	ctx->add(sd );
	
	t( "p.br", 
		"[ : contextual｢ pine lvalue, ( : integer ) -> ( ) context(pine) lvalue unbound ｣, "
		": contextual｢ pine lvalue, ( ) -> ( ) context(pine) lvalue unbound ｣ ] selectable" );
}

BOOST_AUTO_TEST_CASE( type ) {
	tester t;
	
 	t( "type｢integer｣", "type｢ integer ｣" );
 	t( "integer", "type｢ integer ｣" );
 	t( "type｢array｢integer｣｣", "type｢ ^array｢ integer ｣ ｣");
	t( "array｢integer｣", "type｢ ^array｢ integer ｣ ｣");
}

BOOST_AUTO_TEST_SUITE_END()
