#include <boost/test/unit_test.hpp>

#include <iostream>
#include <cstring>

#include "parser/intr_type_parser.hpp"
#include "lang/dump.hpp"
#include "dummy_resolver.hpp"

using namespace leaf;

BOOST_AUTO_TEST_SUITE( intr_type_parse_test )

struct tester {
	bool is_inline;
	bool is_test_serial;
	
	tester() {
		is_inline = false;
		is_test_serial = true;
	}
	
	node_type_spec parse( char const * code ) {
		auto s = source::create( code, std::strlen( code ) );
		if( is_inline )
			return intr_type_parser::parse_inline( *s );
		return intr_type_parser::parse( *s );
	}

	void operator()( char const * code, char const * expect ) {
		auto vt = parse( code ).parsed_type_spec();

		leaf::dump d;
		d.write_type_spec( vt );
		std::string result = d.str();
		if( result != expect )
		{
			std::cerr <<
				"Failed: " << code << std::endl << 
				"Expect: " << expect << std::endl << 
				"Actual: " << result << std::endl;
			BOOST_FAIL( "Failed" );
		}
		
		if( is_test_serial ) {
			test_serial(code);
		}

		BOOST_CHECK(true);
	}
	
	void test_serial( char const * code ) {
		auto ts = parse(code).parsed_type_spec();
		
		dummy_resolver dr;
		std::ostringstream outs;
		leaf::serial_out out( outs, dr );
		out.add_type_spec( *ts );
		out.done();
		
		std::istringstream ins( outs.str() );
		leaf::serial_in in( ins, dr );
		auto load_ts = in.read_primary_type_spec();
		
		auto orig = leaf::dump::get( ts);
		auto load = leaf::dump::get( load_ts );
		if( orig != load ) {
			std::cerr <<
				"Serial Test" << std::endl << 
				"Failed: " << code << std::endl <<
				"Expect: " << orig << std::endl <<
				"Actual: " << load << std::endl;
			BOOST_FAIL( "Failed" );
		}
	}

	void operator()( char const * code ) {
		auto s = source::create( code, std::strlen( code ) );
		try
		{
			intr_type_parser::parse( *s );
			BOOST_FAIL( "Should not have parsed: " << code );
		}
		catch( recoverable_error & re )
		{
			BOOST_CHECK(true);
		}
	}
};

BOOST_AUTO_TEST_CASE( basic ) {
	tester t;
	t( "", "" );
	t( "char", "char" );
	t( "integer 17bit", "17bit integer" );
	t( "integer value_ptr", "integer value_ptr" );
}

BOOST_AUTO_TEST_CASE( tuple ) {
	tester t;
	t( "[\n]", "[ ]" );
	t( "[:]", "[ : ]" );
	t( "[:,a,\nb:,]", "[ : , a : , b : ]" );
	t( "[:[:integer]\n,:[],:[:float,:integer]]",
		"[ : [ : integer ], : [ ], : [ : float, : integer ] ]" );
}

BOOST_AUTO_TEST_CASE( tuple_deco ) {
	tester t;
	t( "lvalue []", "[ ] lvalue" );
}

BOOST_AUTO_TEST_CASE( value_list ) {
	tester t;
	t( "()" , "( )" );
	t( "(:integer)", "( : integer )" );
	t( "(a:integer)", "( a : integer )" );
	t( "(a)", "( a : )" );
	t( "(a,b)", "( a : , b : )" );
}

BOOST_AUTO_TEST_CASE( signature ) {
	tester t;
	t( "(a:integer)->()", "( a : integer ) -> ( )" );
	t( "(a:float,)->()", "( a : float ) -> ( )" );
	t( "(:integer,:integer)->()", "( : integer, : integer ) -> ( )" );
	t( "()->() raw", "( ) -> ( ) raw" );
	t( "()->() construct", "( ) -> ( ) construct" );
	
	//incomplete
	t( "()->$implicit", "( ) -> $implicit" );
}

BOOST_AUTO_TEST_CASE( signature_param ) {
	tester t;
	t( "(a)->()", "( a : ) -> ( )" );
}

//ensure that these fail
BOOST_AUTO_TEST_CASE( broken ) {
	tester t;
	t( "{a;}" );
	t( "(a}" );
	t( "(a,b,c;d)" );
	t( "({a=b})" );
	t( "integer 128bit 64bit" ); //only one size_variant allowed
	t( "float high low" ); //only one variant allowed
	
	//these are disallowed due to reserved words, though they are in a different namespace
	//TODO: test_noparse( "{integer}" )
	//TODO: test_noparse( "{integer,float}" )
}

BOOST_AUTO_TEST_CASE( token_match ) {
	tester t;
	t( "integerblah", "integerblah" );
	t( "shorty", "shorty" );
	t( "32bitsy" );
	t( "under_score", "under_score" );
}

BOOST_AUTO_TEST_CASE( user_types ) {
	tester t;
	t( "point", "point" );
	t( "opt integer modifier", "integer modifier opt" );
	t( "[:ragged,:integer opt]", "[ : ragged, : integer opt ]" );
}

BOOST_AUTO_TEST_CASE( array ) {
	tester t;
	t( "array<:integer:>", "array｢ integer ｣" );
	t( "array<:shared:>", "array｢ shared ｣" );
	t( "shared array<:array<:float shared:>:>", "array｢ array｢ float shared ｣ ｣ shared" );
	t( "[:array<:[:integer,:float]:>]", "[ : array｢ [ : integer, : float ] ｣ ]" );
	
	t( "array｢integer｣", "array｢ integer ｣" );
}

BOOST_AUTO_TEST_CASE( inline_type ) {
	tester t;
	t.is_inline = true;
	t( "integer(", "integer" );
	t( "(\n [ : ()->() ] )", "[ : ( ) -> ( ) ]" );
	t( "(shared integer) garbage", "integer shared" );
}

BOOST_AUTO_TEST_CASE( trailing ) {
	tester t;
	//t( "integer @item(\"value\")", "integer" );
	t( "float;", "float" );
	//line-ends only okay within tuple decl
	t( "integer\n8bit", "integer" );
}

BOOST_AUTO_TEST_CASE( selectable ) {
	tester t;
	t( "selectable [ ]", "[ ] selectable" );
}

BOOST_AUTO_TEST_CASE( lossy ) {
	tester t;
	t( "lossy integer", "integer lossy" );
}

BOOST_AUTO_TEST_CASE( parametrics ) {
	tester t;
	t( "<:T:> T", "｢ T : ｣ T" );
	t( "<:T,Q:shared:> array<:Q:>", "｢ T : , Q : shared ｣ array｢ Q ｣" );
	t( "<:T:> ( x : T, y : shared T ) -> ()",
		"｢ T : ｣ ( x : T, y : T shared ) -> ( )" );
		
	t( "pine<:shared:>", "pine｢ shared ｣" );
	t( "pine<:any:>", "pine｢ any ｣" );
	
	t( "multi｢a,b｣", "multi｢ a, b ｣" );
	t( "embed｢multi｢a,b｣,c｣", "embed｢ multi｢ a, b ｣, c ｣" );
}

BOOST_AUTO_TEST_CASE( context ) {
	tester t;
	t( "()->() context(integer)", "( ) -> ( ) context(integer)" );
	t( "contextual<:integer,float:>", "contextual｢ integer, float ｣" );
}

BOOST_AUTO_TEST_CASE( tuple_constraint ) {
	tester t;
	t( "tuple", "tuple" );
}

BOOST_AUTO_TEST_CASE( expression ) {
	tester t;
	t.is_test_serial = false; //since we don't serialize the `node_expr`, only `expr`
	t( "type_infer(pine)" , "type_infer([value-list [pine]])" );
	t( "type_infer(1+2)", "type_infer([value-list [funccall [add] [value-list [1] [2]]]])" );
}

BOOST_AUTO_TEST_CASE( scope_name ) {
	tester t;
	t( "pine.cone", "pine.cone" );
}

BOOST_AUTO_TEST_SUITE_END()