#include "util/number.hpp"

#include <boost/test/unit_test.hpp>

#include <limits>
#include <sstream>

BOOST_AUTO_TEST_SUITE( number_test )

using util::number;

#define CHECK_DECIMAL( ex, in ) \
	BOOST_CHECK_EQUAL( ex, number::parse_decimal( in ).to_string() );
#define CHECK_BINARY( ex, in ) \
	BOOST_CHECK_EQUAL( ex, number::parse_binary( in ).to_string() );
#define CHECK_OCTAL( ex, in ) \
	BOOST_CHECK_EQUAL( ex, number::parse_octal( in ).to_string() );
#define CHECK_HEX( ex, in ) \
	BOOST_CHECK_EQUAL( ex, number::parse_hexadecimal( in ).to_string() );

#define NPD( str ) number::parse_decimal( str )
unsigned npf( std::string const & str ) {
	unsigned f;
	number::parse_base( str, f, 10 );
	return f;
}

BOOST_AUTO_TEST_CASE( basic ) {
	number n;
	CHECK_DECIMAL( "15", "15" );
 	CHECK_DECIMAL( "19", "+19" );
 	CHECK_DECIMAL( "-3", "-3" );
	CHECK_DECIMAL( "123456789", "123_456_789" );
 	
 	CHECK_DECIMAL( "15", "15.0" );
	CHECK_DECIMAL( "1/2", "0.5" );
	CHECK_DECIMAL( "197/100", "1.97" );
	CHECK_DECIMAL( "13/10000", "0.0013" );
	CHECK_DECIMAL( "7/10", "0.7000" );
	CHECK_DECIMAL( "13/1000", "0.0130" );
	
	CHECK_DECIMAL( "1500", "1.5e3" );
	CHECK_DECIMAL( "34567/100", "3.45_67e+2" );
	CHECK_DECIMAL( "17/10000", "17e-4" );
	
	//lenient parsing, expect higher level parser to catch nonsense
	CHECK_DECIMAL( "-1/2", "-.5" );
	CHECK_DECIMAL( "0", "." );
	CHECK_DECIMAL( "0", "e100" );
	CHECK_DECIMAL( "7", "7." );
}

BOOST_AUTO_TEST_CASE( binary ) {
	number n;
	CHECK_BINARY( "0", "0" );
	CHECK_BINARY( "1", "1" );
	CHECK_BINARY( "2", "0010" );
	CHECK_BINARY( "15", "1111" );
	CHECK_BINARY( "1/2", "0.1" );
	CHECK_BINARY( "1/4", ".01" );
	CHECK_BINARY( "3/4", "0.11" );
}

BOOST_AUTO_TEST_CASE( octal ) {
	number n;
	CHECK_OCTAL( "74", "112" );
	CHECK_OCTAL( "0","0");
	CHECK_OCTAL( "1/8", "0.1" );
	CHECK_OCTAL( "1/32", "0.02" );
}

BOOST_AUTO_TEST_CASE( hexadecimal ) {
	number n;
	CHECK_HEX( "0", "0" );
	CHECK_HEX( "15", "0F" );
	CHECK_HEX( "2989", "bad" );
	CHECK_HEX( "1/16", "0.1" );
	CHECK_HEX( "5/256", "0.05" );
}

BOOST_AUTO_TEST_CASE( flags ) {
	BOOST_CHECK_EQUAL( number::pdf_is_exact, npf( "123" ) );
	BOOST_CHECK_EQUAL( number::pdf_is_exact | number::pdf_had_decimal, npf( "12.5" ) );
	BOOST_CHECK_EQUAL( number::pdf_is_exact | number::pdf_had_decimal, npf( "12." ) );
	BOOST_CHECK_EQUAL( number::pdf_is_exact | number::pdf_had_exponent, npf( "1e2" ) );
	BOOST_CHECK_EQUAL( number::pdf_is_exact | number::pdf_had_exponent | number::pdf_had_decimal, npf( ".6E2" ) );
}

BOOST_AUTO_TEST_CASE( info ) {
	BOOST_CHECK( !NPD( "15" ).is_negative() );
	BOOST_CHECK( NPD( "-18" ).is_negative() );
	
	BOOST_CHECK( NPD( "398" ).is_integral() );
	BOOST_CHECK( !NPD( "0.3" ).is_integral() );
	BOOST_CHECK( NPD( "5.0" ).is_integral() );
	BOOST_CHECK( NPD( "0.33e4" ).is_integral() );
	BOOST_CHECK( !NPD( "123e-1" ).is_integral() );
	BOOST_CHECK( !NPD("0.025e1").is_integral() );
}

BOOST_AUTO_TEST_CASE( convert ) {
	BOOST_CHECK_EQUAL( 123, NPD("12.3e1").as_int() );
	BOOST_CHECK_EQUAL( 123, NPD("12.3e1").as_uint() );
	BOOST_CHECK_EQUAL( -1231, NPD("-1231").as_int() );
	
	//these should be exact, not just close
	BOOST_CHECK_EQUAL( 0.5, NPD("5e-1").as_double() );
	BOOST_CHECK_EQUAL( 0.25, NPD("0.025e1").as_double() );

	//check boundaries on specific types
	BOOST_CHECK_EQUAL( std::numeric_limits<int8_t>::min(), int8_t(NPD("-128").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<int8_t>::max(), int8_t(NPD("127").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<uint8_t>::max(), uint8_t(NPD("255").as_uint()) );
	
	BOOST_CHECK_EQUAL( std::numeric_limits<int16_t>::min(), int16_t(NPD("-32768").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<int16_t>::max(), int16_t(NPD("32767").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<uint16_t>::max(), uint16_t(NPD("65535").as_uint()) );
	
	BOOST_CHECK_EQUAL( std::numeric_limits<int32_t>::min(), int32_t(NPD("-2147483648").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<int32_t>::max(), int32_t(NPD("2147483647").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<uint32_t>::max(), uint32_t(NPD("4294967295").as_uint()) );
	
	BOOST_CHECK_EQUAL( std::numeric_limits<int64_t>::min(), int64_t(NPD("-9223372036854775808").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<int64_t>::max(), int64_t(NPD("9223372036854775807").as_int()) );
	BOOST_CHECK_EQUAL( std::numeric_limits<uint64_t>::max(), uint64_t(NPD("18446744073709551615").as_uint()) );
}

BOOST_AUTO_TEST_CASE( fits ) {
	#define FITS( val, size ) BOOST_CHECK( NPD(val).fits_signed(size) )
	#define NFITS( val, size ) BOOST_CHECK( NPD(val).fits_signed(size) )
	FITS("123",8);
	FITS("127",8);
	FITS("-128",8);
	NFITS("128",8);
	NFITS("-129",8);
	
	FITS("32767",16);
	NFITS("32768",16);
	FITS("-32768",16);
	NFITS("-32769",16);
	
	#define FITU( val, size ) BOOST_CHECK( NPD(val).fits_unsigned(size) )
	#define NFITU( val, size ) BOOST_CHECK( NPD(val).fits_unsigned(size) )
	FITU("255",8);
	NFITU("256",8);
	FITU("0",8);
	
	FITU("65535",16);
	NFITU("65536",16);
	
	FITU("0",1);
	FITU("1",1);
	NFITU("2",1);
}

BOOST_AUTO_TEST_CASE( integer_bits ) {
	#define SIZE( val, size ) BOOST_CHECK_EQUAL( size, NPD(val).signed_bits() );
	SIZE( "1", 2 );
	SIZE( "-1", 2 );
	SIZE( "100", 8 ); //note we still need a sign bit even if positive
	SIZE( "-1000", 11 );
	SIZE( "1e32", 108 );
}

BOOST_AUTO_TEST_CASE( operations ) {
	BOOST_CHECK_EQUAL( "-25/2", NPD( "12.5" ).neg().to_string() );
}

BOOST_AUTO_TEST_CASE( serial ) {
	auto f = []( std::string const & s ) {
		std::ostringstream out;
		auto n = number::parse_decimal(s);
		n.write( out );
		
		std::istringstream in( out.str() );
		number m;
		m.read( in );
		
		BOOST_CHECK_MESSAGE( n == m, std::string(s) + ": " + n.to_string() + " == " + m.to_string() );
	};
	
	f( "2" );
	f( "0" );
	f( "0.5" );
	f( "348.34834534582" );
	f( "-345893458934534.3453895348275" );
	f( "1e1000" );
}

BOOST_AUTO_TEST_SUITE_END()
