#include "parser/source.hpp"
#include "parser/dump.hpp"
#include "parser/node_parser.hpp"
#include "expect_test.hpp"

using namespace leaf;

BOOST_AUTO_TEST_SUITE( parse_test )

struct tester {
	bool as_block;
	
	tester() :
		as_block(false) { 
	}
	
	void operator()( char const * code, char const * expect ) {
		expect_test et;
		et.add( "Code", code );
		auto src = source::create( code, std::strlen( code ) );
		
		shared_ptr<node> lst = as_block 
			? node_parser::parse_as_block( *src )
			: node_parser::parse_single( *src );
		std::string result = node_dump::dump_flat( *lst );
		et.check_actual( result, expect );
	}
	
	void operator()( char const * code ) {
		auto src = source::create( code, std::strlen( code ) );
		try {	
			shared_ptr<node> lst = node_parser::parse_single( *src );
			BOOST_FAIL( "Should not have parsed: " << code );
		} catch( recoverable_error & re ) {
			BOOST_CHECK( true );
		}
	}
};

BOOST_AUTO_TEST_CASE( token_match ) {
	tester t;
	//TODO: unsure of this now -- what level they are rejected...
	//t( "123abc" );
	t( "+-" );
	
	//no accidental keyword detection
	t( "newly", "[newly]" );
	t( "returning", "[returning]" );
	t( "falsed", "[falsed]" );
}

BOOST_AUTO_TEST_CASE( basic ) {
	tester t;
	t( "a", "[a]" );
	t( "(a,b)", "[value-list [a] [b]]" );
	t( "sum(1,2)", 
		"[funccall "
			"[sum] "
			"[value-list [1] [2]]"
		"]" );
		
	t( "{ a; b; }", "[block [a] [b]]" );
	t( "a := b", "[funccall [⁑assign_value] [value-list [a] [b]]]" );
	
	t( "- d",
		"[funccall [neg] [value-list [d]]]" );
		
	t( "a = lossy(b)",
		"[funccall [⁑assign] [value-list [a] [funccall [lossy] [value-list [b]]]]]" );
	t( "() () ()",
		"[funccall [funccall [value-list] [value-list]] [value-list]]" );
}

BOOST_AUTO_TEST_CASE( expr_list ) {
	tester t;
 	t( "a b c", 
 		"[expr-list [a] [b] [c]]" );
	t( "a(b+c) d",
		"[expr-list [funccall [a] [value-list [funccall [add] [value-list [b] [c]]]]] [d]]" );
 	t( "- d c",
 		"[expr-list [funccall [neg] [value-list [d]]] [c]]" );
	//function call has higher precedence than unary minus
	t( "- d ( a - b )  c", 
		"[expr-list [funccall [neg] [value-list [funccall [d] [value-list [funccall [sub] [value-list [a] [b]]]]]]] [c]]" );
	t( "decl name {}", 
		"[expr-list [decl] [name] [block]]" );
	t( "() a ()",
		"[expr-list [value-list] [funccall [a] [value-list]]]" );
}

BOOST_AUTO_TEST_CASE( var_assign ) {
	tester t;
	t( "var abc = def",
		"[expr-list [var] [funccall [⁑assign] [value-list [abc] [def]]]]" );
	t( "var abc = ()->() {}",
		"[expr-list [var] [funccall [⁑assign] [value-list [abc] [funcdefn [value-list] [value-list] [block]]]]]" );
}

BOOST_AUTO_TEST_CASE( base_consts ) {
	tester t;
	t( "0x1F", "[31]" );
	t( "0x0.1", "[1/16f]" );
	t( "0b1011", "[11]" );
	t( "0b1.1", "[3/2f]" );
	t( "0o12", "[10]" );
	t( "0o0.4", "[1/2f]" );
}

BOOST_AUTO_TEST_CASE( thenelse ) {
	tester t;
	t( "a ? b", 
		"[funccall [⁑optional_cond] [value-list [a] [b]]]" );
	t( "a ? b | c",
		"[funccall [⁑ifelse] [value-list [a] [b] [c]]]" );
	t( "a ? b | c ? d | e",
		"[funccall [⁑ifelse] [value-list [a] [b] [funccall [⁑ifelse] [value-list [c] [d] [e]]]]]" );
	t( "a ? (b) | (c)",
		"[funccall [⁑ifelse] [value-list "
			"[a] "
			"[value-list [b]] "
			"[value-list [c]]"
		"]]" );
	t( "a < 5 ? (c,d) | (x,y)",
		"[funccall [⁑ifelse] [value-list "
			"[funccall [lt] [value-list [a] [5]]] "
			"[value-list [c] [d]] "
			"[value-list [x] [y]]"
		"]]" );
	t( "a ? (b ? c) | d",
		"[funccall [⁑ifelse] [value-list [a] [value-list [funccall [⁑optional_cond] [value-list [b] [c]]]] [d]]]" );
}

BOOST_AUTO_TEST_CASE( select ) {
	tester t;
	//not much needs to be tested since this is normal block parsing
	t( "select( a ? b )",
		"[funccall [select] [value-list [funccall [⁑optional_cond] [value-list [a] [b]]]]]" );
	t( "select( q )",
		"[funccall [select] [value-list [q]]]" );
	t( "select( a ? b, c ? d, e )",
		"[funccall [select] [value-list [funccall [⁑optional_cond] [value-list [a] [b]]] [funccall [⁑optional_cond] [value-list [c] [d]]] [e]]]" );
}

BOOST_AUTO_TEST_CASE( do_then ) {
	tester t;
	t( "do a ? b | c",
		"[expr-list [do] [funccall [⁑ifelse] [value-list [a] [b] [c]]]]" );
	t( "do cond_var", 
		"[expr-list [do] [cond_var]]" );
	//naked tuple?
	//t( "do a ? b, c ? d",	"[do tuple [cond [a] [b]] [cond [c] [d]]" );
}

BOOST_AUTO_TEST_CASE( subscript ) {
	tester t;
	t( "a#3", "[funccall [⁑subscript] [value-list [a] [3]]]" );
	t( "a+b#3", "[funccall [add] [value-list [a] [funccall [⁑subscript] [value-list [b] [3]]]]]" );
	t( "a#1#2", "[funccall [⁑subscript] [value-list [funccall [⁑subscript] [value-list [a] [1]]] [2]]]" );
	t( "(a+b)#1", "[funccall [⁑subscript] [value-list [value-list [funccall [add] [value-list [a] [b]]]] [1]]]" );
	t( "(a+b)#1#0", "[funccall [⁑subscript] [value-list [funccall [⁑subscript] [value-list [value-list [funccall [add] [value-list [a] [b]]]] [1]]] [0]]]" );
	t( "a#(3-b)", "[funccall [⁑subscript] [value-list [a] [value-list [funccall [sub] [value-list [3] [b]]]]]]" );
}

BOOST_AUTO_TEST_CASE( field ) {
	tester t;
	t( "a.b", "[funccall [⁑field] [value-list [a] [b]]]" );
	t( "a.b.c", "[funccall [⁑field] [value-list [funccall [⁑field] [value-list [a] [b]]] [c]]]" );
}

BOOST_AUTO_TEST_CASE( return_ ) {
	tester t;
	t( "return", "[return]" );
	t( "return a", "[expr-list [return] [a]]" );
	t( "return(a)", "[expr-list [return] [value-list [a]]]" );
	t( "return (x)-v", "[expr-list [return] [funccall [sub] [value-list [value-list [x]] [v]]]]" );
	//TODO: naked value-lists in return -- probably not allowed, it would be ambiguous parsing
	//t( "return a,b", "[expr-list [return] [value-list [a] [b]]]" );
}

BOOST_AUTO_TEST_CASE( tuple_name ) {
	tester t;
	t( "[ q = val ]", "[tuple [funccall [⁑assign] [value-list [q] [val]]]]" );
}

BOOST_AUTO_TEST_CASE( funcdefn ) {
	//this isn't too deeply tested since it goes a lot into type parsing, plus any actual
	//problems will be caught by higher level tests (basic checked here as it overlaps
	//the named syntax)
	tester t;
	t( "()->() { }", 
		"[funcdefn [value-list] [value-list] [block]]" );
	t( "->{}", 
		"[funcdefn [value-list] [implied] [block]]" );
	t( "->(){}", 
		"[funcdefn [value-list] [value-list] [block]]" );
	t( "()->{}", 
		"[funcdefn [value-list] [implied] [block]]" );
		
	t( "(a,b:integer)->(:float) { }", 
		"[funcdefn [value-list [a] [b] : integer] [value-list [] : float] [block]]" );
	t( "(:float,:integer)->(:float) { }", 
		"[funcdefn [value-list [] : float [] : integer] [value-list [] : float] [block]]" );
	t( "func add : (:integer,:integer)->(:integer)", 
		"[expr-list [func] [add] : ( : integer, : integer ) -> ( : integer )]" );
		
	t( "()->() member { }", "[funcdefn [value-list] [value-list] [block]] : member" );
	
	t( "a = -> () { }",
		"[funccall [⁑assign] [value-list [a] [funcdefn [value-list] [value-list] [block]]]]" );
		
	t( "-> raw { }",
		"[funcdefn [value-list] [implied] [block]] : raw" );
	t( "-> clone { }",
		"[funcdefn [value-list] [implied] [block]] : clone" );
}

BOOST_AUTO_TEST_CASE( annotations ) {
	tester t;
	t( "@anno() a", "[a] @anno [value-list]" );
	t( "@anno(q) a : integer", "[a] : integer @anno [value-list [q]]" );
}

BOOST_AUTO_TEST_CASE( on ) {
	tester t;
	t( "a on any {}","[on [a] [any] [block]]" );
	t( "func() on any {}","[on [funccall [func] [value-list]] [any] [block]]" );
}

BOOST_AUTO_TEST_CASE( resume ) {
	tester t;
	t( "resume_fail", "[resume_fail]" );
}

BOOST_AUTO_TEST_CASE( as_block ) {
	tester t;
	t.as_block = true;
	t( "a", "[block [a]]" );
	t( "a; b;", "[block [a] [b]]" );
	//t( "a,b", "[tuple [a] [b]]" ); //nothing needs to parse directly as tuples now
	//t( "a:integer,b:float", "[tuple [a] : integer [b] : float]" );
	t( "a:integer;b:float", "[block [a] : integer [b] : float]" );
	t( "@anno() var a:integer = b;b:float", 
		"[block [expr-list [var] [funccall [⁑assign] [value-list [a] : integer [b]]]] @anno [value-list] [b] : float]" );
}

BOOST_AUTO_TEST_CASE( classes ) {
	tester t;
	t( "class name { var a; }", "[expr-list [class] [name] [block [expr-list [var] [a]]]]" );
	t( "class pine<:T:> { }", "[expr-list [class] [pine] [param-list [T]] [block]]" );
	t( "class pine｢T｣ { }", "[expr-list [class] [pine] [param-list [T]] [block]]" );
	t( "class cone<:x : integer:> { }", "[expr-list [class] [cone] [param-list [x] : integer] [block]]" );
	t( "class needle <:Q : integral,y:float,z:Other<:Q:>:> {}", 
		"[expr-list [class] [needle] [param-list [Q] : integral [y] : float [z] : Other｢ Q ｣] [block]]" );
}

BOOST_AUTO_TEST_CASE( line_ends ) {
	tester t;
	t.as_block = true;
	t( "one\ntwo", "[block [one] [two]]" );
	t( "\npost;\nthree\n\n", "[block [post] [three]]" );
}

BOOST_AUTO_TEST_CASE( strings ) {
	tester t;
	t( " \"hello\" ", "[\"hello\"]" );
	t( " 'a' ", "['a']" );
}

BOOST_AUTO_TEST_CASE( unicode ) {
	tester t;
	t ( "new a｢b｣()", "[new [] : a｢ b ｣ [value-list]]" );
}

BOOST_AUTO_TEST_CASE( for_loop ) {
	tester t;
	//this doesn't actually need any special parsing, but it's good to test
	t( "for i in items { }",
		"[expr-list [for] [i] [in] [items] [block]]" );
	t( "for i in get(items) { }",
		"[expr-list [for] [i] [in] [funccall [get] [value-list [items]]] [block]]" );
}

BOOST_AUTO_TEST_CASE( range )  {
	tester t;
	//TODO: should this be perhaps "__create_range" as it isn't truly a function...
	t( "4..5", "[funccall [create_range] [value-list [4] [5]]]" );
}

BOOST_AUTO_TEST_CASE( inline_param ) {
	tester t;
	
	t( "bit_shl｢binary 32bit｣( 1, 10 )", "[funccall [bit_shl [param ( : 32bit binary )]] [value-list [1] [10]]]" );
	t( "bit_shl｢binary,integer｣", "[bit_shl [param ( : binary, : integer )]]" );
}

BOOST_AUTO_TEST_CASE( optional_field ) {
	tester t;
	
	t( "a?.b", "[funccall [⁑optional_field] [value-list [a] [b]]]" );
	t( "a?#b", "[funccall [⁑optional_subscript] [value-list [a] [b]]]" );
	//t( "a#b?.c?#d" , "");
}

BOOST_AUTO_TEST_CASE( trap ) {	
	tester t;
	
	t( "trap a", "[funccall [trap] [value-list [a]]]" );
	t( "trap a + b", "[funccall [trap] [value-list [funccall [add] [value-list [a] [b]]]]]" );
	t( "var a = trap a + b", "[expr-list [var] [funccall [⁑assign] [value-list [a] [funccall [trap] [value-list [funccall [add] [value-list [a] [b]]]]]]]]");
	//t( "trap(a,b)", "[funccall [trap] [value-list [a] [b]]]" ); //TODO: collapse value-list in operators
	t( "pine( trap a, b )", "[funccall [pine] [value-list [funccall [trap] [value-list [a]]] [b]]]" );
	t( "var a = trap 15", "[expr-list [var] [funccall [⁑assign] [value-list [a] [funccall [trap] [value-list [15]]]]]]" );
	
	t.as_block = true;
	t( "var a = trap 15\nq", "[block [expr-list [var] [funccall [⁑assign] [value-list [a] [funccall [trap] [value-list [15]]]]]] [q]]" );
}

BOOST_AUTO_TEST_CASE( on_fail ) {
	tester t;

	t( "pine() on fail {}", "[on [funccall [pine] [value-list]] [fail] [block]]" );
	t( "a + b on fail {}", "[on [funccall [add] [value-list [a] [b]]] [fail] [block]]" );
	
	t( "var a = b on fail {}", "[on [expr-list [var] [funccall [⁑assign] [value-list [a] [b]]]] [fail] [block]]" );
	t( "a() on fail {}", "[on [funccall [a] [value-list]] [fail] [block]]" );
}

BOOST_AUTO_TEST_SUITE_END()