#include <boost/test/unit_test.hpp>

#include "lang/scope.hpp"
#include "lang/declaration.hpp"

using namespace leaf;

#define SHOW_DEBUG 0
#include "util/debug.hpp"

BOOST_AUTO_TEST_SUITE( scope_test )

shared_ptr<declaration> vs_single( std::string const & name )
{
	auto vs = declaration::create();
	vs->symbol = name;
	vs->form = declaration::f_single;
	return vs;
}

shared_ptr<declaration> vs_multi( std::string const & name )
{
	auto vs = declaration::create();
	vs->symbol = name;
	vs->form = declaration::f_multi;
	return vs;
}

BOOST_AUTO_TEST_CASE( basic )
{
	auto s = scope::create( scope_style_t::imperative /*doens't matter here*/ );
	
	s->add( vs_single( "a" ) );
	auto sr = s->if_get_symbol( "a" )->vars[0];
	BOOST_CHECK_EQUAL( sr->symbol, "a" );
}

BOOST_AUTO_TEST_CASE( parent )
{
	auto sa=scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative ), 
		sc = scope::create( scope_style_t::imperative );
	sb->set_base( sa, scope::bt_parent );
	sc->set_base( sb, scope::bt_parent );
	
	BOOST_CHECK_EQUAL( 0, sa->get_level() );
	BOOST_CHECK_EQUAL( 2, sc->get_level() );
	
	sa->add( vs_single( "a" ) );
	
	auto sr = sc->if_get_symbol( "a" )->vars[0];
	BOOST_CHECK( !sc->is_local(*sr) );
	BOOST_CHECK_EQUAL( 0, sr->get_root_depth() );

	//decl overrides one from parent
	sb->add( vs_single( "a" ) );
	sr = sc->if_get_symbol( "a" )->vars[0];
	BOOST_CHECK_EQUAL( 1, sr->get_root_depth() );
	
	//multi's can exist at multiple levels (might be crazy though) //TODO: perhaps only nearest should resolve?
	sa->add( vs_multi( "b" ) );
	sb->add( vs_multi( "b" ) );
	sc->add( vs_multi( "b" ) );
	auto all = sc->if_get_symbol( "b" );
	BOOST_CHECK( !!all );
	BOOST_REQUIRE_EQUAL( 3, all->vars.size() );
	//the order shouldn't really be guaranteed
	BOOST_CHECK( sc->is_local( *all->vars[0] ) );
	BOOST_CHECK( !sc->is_local( *all->vars[1] ) );
	BOOST_CHECK( !sc->is_local( *all->vars[2] ) );
	
	auto vall = sc->get_local_vars( "b" );
	BOOST_CHECK_EQUAL( 1, vall.size() );
	vall = sc->get_local_vars( "a" );
	BOOST_CHECK_EQUAL( 0, vall.size() );
}

BOOST_AUTO_TEST_CASE( local )
{
	auto sa = scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative ), 
		sc = scope::create( scope_style_t::imperative ), 
		sd = scope::create( scope_style_t::imperative );
	sb->set_base( sa, scope::bt_parent );
	sc->set_base( sb, scope::bt_parent );
	sd->set_base( sc, scope::bt_local );
	
	BOOST_CHECK_EQUAL( 2, sc->get_level() );
	BOOST_CHECK_EQUAL( 3, sd->get_level() );
	
	sa->add( vs_single( "a" ) );
	auto sr = sd->if_get_symbol( "a" )->vars[0];
	BOOST_CHECK_EQUAL( 0, sr->get_root_depth() );
	BOOST_CHECK( !sd->is_local(*sr) );
	
	sc->add( vs_single( "b" ) );
	sr = sd->if_get_symbol( "b" )->vars[0];
	BOOST_CHECK_EQUAL( 2, sr->get_root_depth() );
	BOOST_CHECK( sd->is_local(*sr) );
	
	sd->add( vs_single( "c" ) );
	BOOST_CHECK( !sc->if_get_symbol( "c" ) );
	
	sr = sa->if_get_symbol( "a" )->vars[0];
	auto cr = sd->capture_var( sr );
	BOOST_CHECK_EQUAL( sr.get(), cr.get() );
	sb->add( vs_single("d") );
	sr = sd->if_get_symbol( "d" )->vars[0];
	cr = sd->capture_var( sr );
	BOOST_CHECK_EQUAL( declaration::f_capture, cr->form );
	BOOST_CHECK_EQUAL( sc.get(), cr->var_scope.get() );
	sr = sc->capture_var( sr );
	BOOST_CHECK_EQUAL( cr.get(), sr.get() );
}

BOOST_AUTO_TEST_CASE( capture )
{
	auto sa = scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative ), 
		sc = scope::create( scope_style_t::imperative ), 
		sd = scope::create( scope_style_t::imperative );
	sb->set_base( sa, scope::bt_parent );
	sc->set_base( sb, scope::bt_parent );
	sd->set_base( sc, scope::bt_parent );
	
	sb->add( vs_single( "a" ) );
	auto sr = sd->if_get_symbol( "a" )->vars[0];
	auto local = sd->capture_var( sr );
	BOOST_CHECK_EQUAL( local->ultimate_capture.get(), sr.get() );
	
	{
		auto const & cm = sd->get_captures();
		BOOST_REQUIRE_EQUAL( 1,  cm.size() );
	}
	
	{
		auto const & cm = sc->get_captures();
		BOOST_REQUIRE_EQUAL( 1,  cm.size() );
	}

	{
		auto const & cm = sb->get_captures();
		BOOST_REQUIRE_EQUAL( 0,  cm.size() );
	}

	//capture from a parent scope_ref
	sb->add( vs_single("b") );
	sr = sb->if_get_symbol( "b" )->vars[0];
	local = sc->capture_var( sr );
	BOOST_CHECK_EQUAL( 2, local->get_root_depth() );
	BOOST_CHECK_EQUAL( sc.get(), local->var_scope.get() );
	
}

BOOST_AUTO_TEST_CASE( namespace_basic )
{
	auto sa = scope::create( scope_style_t::imperative ), 
		sb = scope::create( scope_style_t::imperative );
	sb->set_base( sa, scope::bt_namespace );
	
	sa->add( vs_single("only_a") );
	sb->add( vs_single("only_b") );
	sa->add( vs_single("both") );
	sb->add( vs_single("both") );
	
	auto sr = sb->if_get_symbol( "only_a" )->vars[0];
	BOOST_CHECK_EQUAL( sa.get(), sr->var_scope.get() );
	BOOST_CHECK( !sb->is_local(*sr) );
	
	sr = sb->if_get_symbol( "both" )->vars[0];
	BOOST_CHECK_EQUAL( sb.get(), sr->var_scope.get() );
	sr = sa->if_get_symbol( "both" )->vars[0];
	BOOST_CHECK_EQUAL( sa.get(), sr->var_scope.get() );
	
	BOOST_CHECK( !sa->if_get_symbol( "only_b" ) );
	
	auto vars = sb->if_get_symbol( "both" )->vars;
	BOOST_CHECK_EQUAL( 1, vars.size() );
}

BOOST_AUTO_TEST_SUITE_END()
