/**
	Covers the parsing and basic operations on statements. Does not include any typing operations.
*/
#include "type_converter_common.hpp"

#include "lang/dump.hpp"
#include "lang/expression.hpp"

#include "parser/source.hpp"
#include "parser/node_parser.hpp"
#include "parser/node_converter.hpp"

#include <boost/test/unit_test.hpp>
#include<boost/algorithm/string.hpp>

#include <iostream>
#include <cstring>

#define SHOW_DEBUG 0
#include "util/debug.hpp"

using namespace leaf;

BOOST_AUTO_TEST_SUITE( statement_test )

std::string dump_statement( statement const & st ) {
	leaf::dump d;
	d.set_flat();
	d.set_validate(false);
	d.set_show_spec_type(true); //no typing yet, thus still in spec_type
	d.set_hide_var_spec_no_fundamental(true);
	d.write_statement( st );
	
	auto  r= d.str();
	boost::trim(r);
	return r;
}

void test_parse( char const * code, char const * expect )
{
	expect_test et;
	et.add( "Source", code );

	try {
		auto s = source::create( code, std::strlen( code ) );

		//DEBUG( "-->", code );

		auto n = node_parser::parse_single( *s );
		auto st = node_converter::convert( *n );
		et.check_actual( dump_statement( *st ), expect );
		
		//check that the clone produces the same output
		auto clone = st->pre_clone();
		et.check_actual( dump_statement( *clone ), expect, "Cloned" );
	} catch(...) {
		et.except();
	}
}

void test_noparse( char const * code )
{
	//DEBUG( "-->", code );

	auto s = source::create( code, std::strlen( code ) );
	try
	{
		shared_ptr<node> node( node_parser::parse_single( *s ) );
		shared_ptr<statement> st = node_converter::convert( *node );
		shared_ptr<expression> ex = std::dynamic_pointer_cast<expression>( st );
		
		BOOST_FAIL( "Should not have parsed: " << code );
	}
	catch( recoverable_error & re )
	{
		BOOST_CHECK( true );
	}
}

BOOST_AUTO_TEST_CASE( basic )
{
	test_parse( "a",
		"a" );
	test_parse( "a +b ",
		"add <- ( a, b )" );
	//unary negative
	test_parse( "a + -3*c",
		"add <- ( a, mul <- ( neg <- ( 3/1 ), c ) )" );
	//chain of exponents
	test_parse( "a*3^ -c^d - 2",
		"sub <- ( mul <- ( a, pow <- ( 3/1, pow <- ( neg <- ( c ), d ) ) ), 2/1 )");
	//basic tuple
	test_parse( " [ a, 1, c + d ]",
		"[ a, 1/1, add <- ( c, d ) ]" );
	//empty tuple
	test_parse( "[]",
		"[ ]" );
	//tuple concatenation
	test_parse( "[1,2,c] ++ [[],[1],[[2]]]",
		"concat <- ( [ 1/1, 2/1, c ], [ [ ], [ 1/1 ], [ [ 2/1 ] ] ] )" );
#if 0
	//NOTE: not a valid syntax now, may be reintroduced later
	//expanded function call using tuple
	test_parse( "a<-[1,2]",
		"a <- [ 1/1, 2/1 ]" );
	//expanded function call using value-list
	test_parse( "a<-(1,2)",
		"a <- ( 1/1, 2/1 )" );
	//call function with tuple var
	test_parse( "a<-q",
		"a <- q" );
	//chain function call expanded
	test_parse( "a<-q<-c",
		"a <- q <- c" );
	//mixed syntax for previous
	test_parse( "get_func<-[](a)",
		"get_func <- [ ] <- ( a )" );
#endif
	//complex tuple
	test_parse( "[ cos(5/c), log(rand()), uniq([1,2,1]) ]",
		"[ cos <- ( div <- ( 5/1, c ) ), log <- ( rand <- ( ) ), uniq <- ( [ 1/1, 2/1, 1/1 ] ) ]" );
	//override precedence
	test_parse( "(a+b)*c",
		"mul <- ( ( add <- ( a, b ) ), c )" );
	//excess parens (kept now, dealt with during typing)
	test_parse( "((x/(c-d)))",
		"( ( div <- ( x, ( sub <- ( c, d ) ) ) ) )" );
	//unary operator post other op
	test_parse( "a- -b",
		"sub <- ( a, neg <- ( b ) )" );
	//call function on return of function
	test_parse( "get_func()(a)",
		"get_func <- ( ) <- ( a )" );
	//nonsensical call of tuple, but syntactically parsable
	test_parse( "[1](x)",
		"[ 1/1 ] <- ( x )" );
	//in fact any value is callable as a function
	test_parse( "(a+b)(q)",
		"( add <- ( a, b ) ) <- ( q )" );
	//a series of unary operators
	test_parse( "- -(4)",
		"neg <- ( neg <- ( ( 4/1 ) ) )" );
	//a unary tuple
	test_parse( "-[1]",
		"neg <- ( [ 1/1 ] )" );
}

BOOST_AUTO_TEST_CASE( declarations )
{
	test_parse( "var a", "single a" );
	test_parse( "var a = b", "single a = b" );
	test_parse( "defn a := c", "@lifetime(global) multi a : literal := c" );
	test_parse( "@lifetime(local) defn a := c", "multi a : literal := c" );
	test_parse( "var a : integer = b", "single a : integer = b" );
	test_parse( "var a : shared := b", "single a : shared := b" );
}

BOOST_AUTO_TEST_CASE( boolean )
{
	test_parse( "true", "true" );
	test_parse( "false", "false" );
}

BOOST_AUTO_TEST_CASE( strings )
{
	test_parse( " \"hello\"", "\"hello\"" );
}

BOOST_AUTO_TEST_CASE( not_parsable )
{
	//not parsable

	test_noparse( "- -" ); //requires a value
	test_noparse( "a--b" ); //not a valid operator
	test_noparse( "a+" ); //missing binary operand
}

/**
	Stuff which has syntactic meaning but is invalid semantically.
	
	Note that only internals can take value-lists (these are private types).
	
	TODO: fix these
*/
// BOOST_AUTO_TEST_CASE( nonsense )
// {
// 	//test_noparse( "()" ); //now parses, is it nonsense (a void expression)
// 	//test_noparse( "-()" ); //also parses, but nonsense
// 	//test_noparse( "()+{}" ); //excess parens still don't define a value
// 	//test_noparse( "(a,b)+(c,d)" ); //value-lists can only be final expression or in a function calls
// }

BOOST_AUTO_TEST_CASE( value_list )
{
	test_parse( "(a,b)",
		"( a, b )" );
	//previously this would reduce to just "a" but that is typing dependent, thus not done during parsing
	test_parse( "(a)",
		"( a )" );
	//simple function
	test_parse( "sin(1)",
		"sin <- ( 1/1 )" );
	//PREVIOUSLY: 1-value-list should always collapse, but 2's are still value-lists
	//test_parse( "defn(((1.5));(a;b))",
	//	"defn <- ( 1.5; ( a; b ) )" );
		
	//from multi_match test (had a problem at one point)
	test_parse( "1 + trunc(2.5)",
		"add <- ( 1/1, trunc <- ( 5/2 ) )" );

	//this is syntactically valid nonsene (calling a function of a non-1-value-list
	test_parse( "(a,b)(c,d)",
		"( a, b ) <- ( c, d )" );
	//it works since 1-value-list's work
	test_parse( "(a)(c,d)",
		"( a ) <- ( c, d )" );
}

/**
	Supporting assignment as just another operation simplifies a lot of things.
*/
BOOST_AUTO_TEST_CASE( assignment )
{
	test_parse( "a= b",
		"a = b" );
	
	//value-lists must be assignable then
	test_parse( "(a,b)=(c,d)",
		"( a, b ) = ( c, d )" );
}

/**
	Any part of an expression can be typed. This forces the type at this particular
	point. For clarity, and lack of ambiguity, the type can only appear in a proper
	value-list with a value-only entry.
	NOTE: I've lifted this ambiguity check to ease the parsing. It must just be know that
	':' has a very high precedence.
*/
BOOST_AUTO_TEST_CASE( inline_type )
{
	test_parse( "(a:integer) + b", 
		"add <- ( ( a : integer ), b )" ); //during typing this will be reduced however to ( a : integer; b )
		
	//the ambiguity is below
	test_noparse( "a : integer + b");
	//test_parse( "a : integer + b", "add <- ( a : integer, b )" );
	//test_noparse( "a + b : integer");
	test_parse( "a + b : integer", "add <- ( a, b : integer )" );
	
	//this is okay though
	test_parse( "a : integer",
		"a : integer" );
	//and to type an expression do this
	test_parse( "(a+b):float",
		"( add <- ( a, b ) ) : float" );
	test_parse( "((a:float) + (b:float)) : float",
		"( add <- ( ( a : float ), ( b : float ) ) ) : float" );
		
	test_parse( "((a+b):integer)+((b+c):integer)",
		"add <- ( ( ( add <- ( a, b ) ) : integer ), ( ( add <- ( b, c ) ) : integer ) )" );
	test_parse( "( (b()) : (:integer)->(:integer) )(a:integer)",
		"( ( b <- ( ) ) : ( : integer ) -> ( : integer ) ) <- ( a : integer )" );
		
	//wasn't allow to parse before (first one)
	test_parse( " a = b : integer", "a = b : integer" );
	test_parse( "a = (b : integer)", "a = ( b : integer )" );
}

BOOST_AUTO_TEST_CASE( statement_block ) {
	test_parse( "a = ()->() { }",
		"a = ( ) -> ( ) { }" );
		
	test_parse( "b := (a:integer,c)->(d:float) { d=c; }",
		"b := ( a : integer, c : ) -> ( d : float ) { d = c }" );
}

BOOST_AUTO_TEST_CASE( parametric_type ) {
	test_parse( "typedef simple : other", "typedef simple : other" );
	test_parse( "typedef param : <:T:> optional T", "typedef param : ｢ T : ｣ T optional" );
}

BOOST_AUTO_TEST_CASE( class_params ) {
	test_parse( "class pine<: T :> { }", "class pine ｢ T : ｣ { }" );
}

BOOST_AUTO_TEST_SUITE_END()
