#include "lang/declaration.hpp"
#include "lang/dump.hpp"
#include "lang/scope.hpp"
#include "lang/statement_block.hpp"
#include "lang/type_converter.hpp"
#include "lang/type.hpp"
#include "lang/type_identifier.hpp"
#include "lang/typer.hpp"
#include "parser/intr_type_parser.hpp"
#include "parser/node_converter.hpp"
#include "runner/runner.hpp"

#include <sstream>

#include "util/debug.hpp"
#include "expect_test.hpp"

using namespace leaf;

//use anon since debug output may differ per file
namespace {

struct type_converter_tester {
	shared_ptr<runner> run; //may be null
	shared_ptr<scope> sc; //never null
	shared_ptr<typer> the_typer; //never null
	
	context the_context() {
		return context( the_typer.get(), sc );
	}
	
	type_converter_tester() : 
		sc( scope::create(scope_style_t::declarative) ),
		the_typer( std::make_shared<typer>(sc) ) {
	}
	type_converter_tester( shared_ptr<scope> sc ) :
		sc( sc ),
		the_typer( std::make_shared<typer>(sc) ) {
	}
	
	enum style_runner_t { style_runner };
	type_converter_tester( style_runner_t ) :
		run( std::make_shared<runner>() ) {
		run->init_module();
		sc = run->module->root;
		the_typer = std::make_shared<typer>(sc);
	}
	
	shared_ptr<type_spec> parse_spec( char const * code) {
		auto s = source::create( code, std::strlen(code) );
		auto vts = intr_type_parser::parse( *s );
		return node_converter::convert(vts);
	}

	type_ref determine_type( char const * code, unsigned flags = 0 ) {
		auto vts = parse_spec( code );
		auto vt = type_identifier::determine( the_context(), vts, flags );
		STATE_CHECK(vt.okay(),vt.explain_error());
		DEBUG( code, dump::get(vts), dump::get(vt.result) );
		return vt.result;
	}
	
	shared_ptr<intr_type_value_list const> parse_value_list( char const * code ) {
		auto vts = parse_spec( code );
		auto vt = type_identifier::determine( the_context(), vts );
		STATE_CHECK(vt.okay(),vt.explain_error());
		auto vl = std::dynamic_pointer_cast<intr_type_value_list const>(vt.result.extr().intr);
		STATE_CHECK(vl);
		return vl;
	}
	
	/**
		Adds a new fundamental class type. This is expected to be a tuple and mimics what is
		done in typer::proc_typedef_statement to support recursion.
	*/
	void add_defined_type( std::string name, char const * code,
		shared_ptr<scope> where = shared_ptr<scope>() ) {
		
		auto inc_tuple = make_shared<intr_type_tuple>();
		inc_tuple->is_export = true;
		inc_tuple->set_distinct_name( name );
		inc_tuple->style = intr_type::s_distinct;
		inc_tuple->local_complete = false;
		add_defined_type( inc_tuple, where );
		
		auto raw = determine_type( code, type_identifier::df_allow_param );
		PRE_CHECK( !raw.extr().intr->is_whole() );
		auto nt_tuple = raw.extr().if_cast<intr_type_tuple>();
		PRE_CHECK( nt_tuple );
		inc_tuple->complete_with( *nt_tuple );
	}
	
	void add_defined_type( shared_ptr<intr_type> type, 
		shared_ptr<scope> where = shared_ptr<scope>() ) {
		
		if( !where ) {
			where = sc;
		}
		
		where->add_fun_type( type );
	}
	
	struct add_class_return {
		shared_ptr<scope> context;
		shared_ptr<intr_type_classlike> type;
	};
	add_class_return add_class( std::string const & name ) {
		auto ctx = make_shared<intr_type_class>();
		ctx->is_export = true; //to allow serialization tests
		ctx->style = intr_type::s_distinct;
		ctx->set_distinct_name( name );
		ctx->block = make_shared<statement_block>( scope_style_t::declarative );
		ctx->block->symbols->set_base( sc, scope::bt_parent );

		auto et = extr_type(ctx);
		et.lvalue = true;
		auto csc = scope::create(scope_style_t::declarative);
		csc->mark_class_root( et );
		ctx->context = csc;
		add_defined_type( ctx );
		
		add_class_return acr;
		acr.context = csc;
		acr.type = ctx;
		return acr;
	}
	
	void add_type_alias( std::string name, char const * code ) {
		auto s = source::create( code, std::strlen(code) );
		auto spec = node_converter::convert(intr_type_parser::parse( *s ));
		auto err = type_identifier::expand( the_context(), *spec );
		PRE_CHECK( err.okay(), err.explain_error() );
		sc->add_type_alias( name, spec );
	}
	
	void add_var( std::string const & name, char const * type_spec ) {
		auto sd = declaration::create();
		sd->form = declaration::f_single;
		sd->symbol = name;
		sd->type = determine_type(type_spec);
		sc->add( sd );
	}
};

} //eon anon
