#define SHOW_DEBUG 0
#include "expr_common.hpp"

BOOST_AUTO_TEST_SUITE( type_converter_parameterize_type )

//needs expr_tester due to use of `type_infer`
struct tester : public expr_tester {
	
	void operator()( char const * param_type, char const * apply, char const * expect ) {
		expect_test ex;
		ex.add( "Parametric", param_type );
		ex.add( "Apply", apply );
	
		type_identifier::resolve_t resolve;
		try {
			auto t_param = type_ref::create_spec( *parse_spec(param_type) );
			auto t_apply = type_ref::create_spec( *parse_spec(apply) );
		
			resolve = type_identifier::parameterize_type( the_context(), t_param, t_apply,
				type_identifier::df_allow_param);
		} catch(...) {
			ex.except();
		}
		
		ex.check( resolve.okay(), resolve.explain_error() );
		ex.check_actual( dump::get( resolve.result ), expect );
	}
	
	void operator()( char const * param_type, char const * apply, cerr expect ) {
		expect_test ex;
		ex.add( "Parametric", param_type );
		ex.add( "Apply", apply );
	
		type_identifier::resolve_t resolve;
		try {
			auto t_param = type_ref::create_spec( *parse_spec(param_type) );
			auto t_apply = type_ref::create_spec( *parse_spec(apply) );
		
			resolve = type_identifier::parameterize_type( the_context(), t_param, t_apply );
		} catch(...) {
			ex.except();
		}
		
		ex.check( resolve.error == expect, resolve.explain_error(), "wrong error" );
	}
};

BOOST_AUTO_TEST_CASE( non_parametric ) {
	tester t;
	t( "integer", "float", "integer" );
	t( "shared integer", "float", "integer shared" );
	t( "( : integer, : float )", "float", "( : integer, : float )" );
}

BOOST_AUTO_TEST_CASE( singles ) {
	tester t;
	t( "", "integer", "integer" );
	t( "", "float high", "float high" );
	t( "", "( : boolean, : float )", "( : boolean, : float )" );
	t( "", "integer lvalue", "integer" );
	t( "", "integer literal", "integer" );
}

BOOST_AUTO_TEST_CASE( references ) {
	tester t;
	t( "value_ptr", "integer", "integer value_ptr" );
}

BOOST_AUTO_TEST_CASE( value_list ) {
	tester t;
	t( "( : )", "( : integer )", "( : integer )" );
	t( "( : boolean, : )", "( : integer, : float low )", "( : boolean, : float low )" );
	t( "( : ( x : ) )", "( : ( : integer ) )", "( : ( x : integer ) )" );
	t( "( x : )", "( : integer )", "( x : integer )" );
	t( "(  : )", "( x : integer)", "( : integer )" );
	t( "( : ( : )->())", "( x : ( y : float )->() )", "( : ( : float ) -> ( ) )");
}

BOOST_AUTO_TEST_CASE( unknown ) {
	tester t;
	t( "", "", "" );
	t( "( : , : )", "( : integer )", "( : integer, : )" );
}

BOOST_AUTO_TEST_CASE( signature ) {
	tester t;
	t( "(:)->()", "(:integer)->()", "( : integer ) -> ( )" );
	t( "(:)->()", "(x:integer)->()", "( : integer ) -> ( )" );
	//t( "(:unknown)->unknown", "(:integer)->()", "( : integer ) -> unknown" ); //unknown what should be done

	t( "(:)->(:)", "(:)->(:boolean)", "( : ) -> ( : boolean )" );
	t( "()->$implicit", "()->()", "( ) -> ( )" );
	t( "( : ()->$implicit )->$implicit", "(:()->())->()", "( : ( ) -> ( ) ) -> ( )" );
	t( "( : ( x : )->$implicit ) -> $implicit",
		"( : ( : integer) -> ( ) ) -> $implicit",
		"( : ( x : integer ) -> ( ) ) -> $implicit" );
	//The change that arg names are not applied makes this fail... It could in theory work, but applying
	//a param to a param type doesn't make sense, does it?
 	//t( "( : array｢integer｣ lvalue, : ( x : ) -> $implicit ) -> $implicit literal )",
 	//	"( x : array｢ any ｣, f : ( : type_infer( x#0 ) ) -> ( ) ) -> $implicit literal",
 	//	"( x : ^array｢ integer ｣, f : ( x : integer ) -> ( ) ) -> $implicit" );
	
	//only params
	t( "(:)->(:integer)", "(:float)", "( : float ) -> ( : integer )" );
	t( "(:)->$implicit", "(:float)", "( : float ) -> $implicit" );
	t( "(:)->$implicit", "(:integer literal)", "( : integer ) -> $implicit" );
	
	//with context
	t.add_class( "pine" );
	t( "()->() context(pine)", "()", "( ) -> ( ) context(pine) unbound" );
	t( "(:)->() context(pine)", "(:integer)", "( : integer ) -> ( ) context(pine) unbound" );
}

BOOST_AUTO_TEST_CASE( class_param ) {
	tester t;
	
	auto cl = t.add_class( "cone" );
	cl.type->params.push_back( { "T" } );
	
	t( "( : cone<:any:> )->()", "( : cone<:integer:> )", "( : ^cone｢ integer ｣ ) -> ( )" );
}

BOOST_AUTO_TEST_CASE( same_symbol_name ) {
	tester t;
	
	t( "( x : (:integer)->())", "( : (x)->())", "( x : ( : integer ) -> ( ) )" );
	t( "( x, : ( : type_infer(x) )->() )", "( :integer, : (x)->())", "( x : integer, : ( : integer ) -> ( ) )" );
	
	//TODO: These should not be allowed as constraints, but perhaps as conversions?!
	//there's no way to capture this failure at the moment (duplicate-symbol is an exception string)
	//t( "( x, : integer)", "(:integer, x)", cerr::duplicate_symbol);
	//t( "( x, : type_infer(x))", "(:integer, x)", cerr::duplicate_symbol);
}

BOOST_AUTO_TEST_CASE( array_infer ) {
	tester t;
	
	t( "( x : array<:any:>)", "( : array<:integer:> )", "( x : ^array｢ integer ｣ )" );
	t( "( x : array<:any:>, : type_infer(x#0) )", "( : array<:integer:> )", 
		"( x : ^array｢ integer ｣, : integer )" );
	t( "( x : array<:any:>, : ( : type_infer(x#0) ) ->() )", "( : array<:integer:> )", 
		"( x : ^array｢ integer ｣, : ( : integer ) -> ( ) )" );
	t( "( x : array<:any:>, : ( : type_infer(x#0) ) ->() )->()", "( : array<:integer:> )", 
		"( x : ^array｢ integer ｣, : ( : integer ) -> ( ) ) -> ( )" );
	t( "( x : array<:any:>, : ( : type_infer(x#0) ) ->() )", "( : array<:integer:>, : ( : )->() )", 
		"( x : ^array｢ integer ｣, : ( : integer ) -> ( ) )" );
}

BOOST_AUTO_TEST_CASE( symbol_scope ) {
	tester t;
	
	t( "( x, :, z : type_infer(x#0))", "( : [ : integer], x : integer, z : integer )", 
		"( x : [ : integer ], : integer, z : integer )" );
 	t( "( x, : ( : type_infer(x#0), : type_infer(x#0))->() )",
 		"( : [ : integer ], : ( x : integer, y : integer)->() )", 
 		"( x : [ : integer ], : ( : integer, : integer ) -> ( ) )" );
}


BOOST_AUTO_TEST_SUITE_END()