#define SHOW_DEBUG 0
#include "type_converter_common.hpp"
#include "dummy_resolver.hpp"

#include <sstream>

BOOST_AUTO_TEST_SUITE( type_converter_test )

struct tester : public type_converter_tester
{
	typedef type_converter::convert_t (type_converter::*from_func)( type_ref const & );
	typedef type_converter::convert_t (type_converter::*to_from_func)( type_ref const &,
		type_ref const & );
		
	from_func use_from_func;
	to_from_func use_to_from_func;
	bool check_serial;
	
	tester( from_func use_from_func ) : 
		use_from_func( use_from_func ), 
		use_to_from_func( nullptr ),
		check_serial( true ) { 
	}
	tester( to_from_func use_to_from_func ) : 
		use_from_func( nullptr ), 
		use_to_from_func( use_to_from_func ),
		check_serial( true ) {
	}
	tester() : 
		use_from_func( nullptr ), 
		use_to_from_func( &type_converter::convert ),
		check_serial( true ) {
	}
	
	unsigned determine_flags = 0;
	
	void operator()( char const * sto, char const * sfrom, 
		char const * expect )
	{
		expect_test ex;
		if( sto )
			ex.add( "To", sto );
		ex.add( "From", sfrom );
		
		type_converter::convert_t cost;

		type_ref to, from;
		if( sto )
			to = determine_type( ex.platform_fix(sto).c_str(), determine_flags );
		from = determine_type( ex.platform_fix(sfrom).c_str(), determine_flags );

		DEBUG( "FROM", dump::get( from ) );
		if( check_serial ) {
			//do serialization tests on all tested types, good way to get coverage
			if( sto )
				test_serial( to );
			test_serial( from );
		}

		//DEBUG( leaf::dump::get(to), "==", leaf::dump::get(from));
		try {
			if( use_to_from_func )
				cost = (the_context().converter().*use_to_from_func)( to, from );
			else
				cost = (the_context().converter().*use_from_func)( from );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( cost.okay(), cost.explain_error() );
			
		std::ostringstream out;
		if( cost.conversions )
			out << dump::get( cost.conversions );
		ex.check_actual( out.str(), expect );
	}
	void operator()( char const * sfrom, char const * expect ) {
		(*this)( nullptr, sfrom, expect ); 
	}
	
	void operator()( char const * sto, char const * sfrom, cerr expect )
	{
		expect_test ex;
		if( sto )
			ex.add( "To", sto );
		ex.add( "From", sfrom );
		
		type_ref to, from;
		if( sto )
			to = determine_type( sto );
		from = determine_type( sfrom );

		if( check_serial ) {
			if( sto )
				test_serial( to );
			test_serial( from );
		}

		type_converter::convert_t cost;
		try {
			if( use_to_from_func )
				cost = (the_context().converter().*use_to_from_func)( to, from );
			else
				cost = (the_context().converter().*use_from_func)( from );
		} catch( ... ) {
			ex.except();
		}
		ex.check( cost.error == expect, cost.explain_error(), "wrong error" );
	}
	void operator()( char const * sfrom, cerr expect ) {
		(*this)( nullptr, sfrom, expect );
	}
	
	void test_serial( type_ref t ) {
		dummy_resolver dr;
		
		std::ostringstream outs;
		leaf::serial_out out( outs, dr );
		out.add_type_ref( t );
		out.done();
		
		std::istringstream ins( outs.str()  );
		leaf::serial_in in( ins, dr );
		auto load = in.read_primary_type_ref();
		
		if( t != load ) {
			std::cout << "Serialization failed" << std::endl;
			std::cout << "Orig: " << leaf::dump::get( t ) << std::endl;
			std::cout << "Load: " << leaf::dump::get( load ) << std::endl;
			BOOST_CHECK(false);
		}
	}
};

BOOST_AUTO_TEST_CASE( basic ) {
	tester t;
	t( "integer", "integer", "" );
	t( "integer shared", "integer shared", "" );
	t( "integer lvalue", "integer lvalue", "" );
	t( "integer lvalue shared", "integer lvalue shared", "" );
	t( "integer shared", "integer lvalue shared", "c_drop_lvalue" );
	t( "integer", "integer shared lvalue", "c_drop_lvalue->c_deref" );
	
 	t( "integer", "integer lvalue",	 "c_drop_lvalue" );
 	t( "integer", "integer shared", "c_deref" );
	t( "[]", "[] lvalue", "c_drop_lvalue" );
	
}

BOOST_AUTO_TEST_CASE( functions ) {
	tester t;
	t( "()->()", "()->()", "" );
	t( "()->()", "()->() raw lvalue", "c_drop_lvalue->c_bind_raw_func" );
	t( "()->()", "()->() raw shared lvalue", "c_drop_lvalue->c_deref->c_bind_raw_func" );
	t( "(:integer)->()", "(:integer)->()", "" );
	t( "()->()", "()->() raw", "c_bind_raw_func" );
	
	t( "()->()", "()->() unbound", "c_bind_func" );
	
	t( "()->() optional", "()->()", "c_optional" );
	t( "()->() optional", "()->() unbound", "c_bind_func->c_optional" );
	
	t.add_class("pine");
	t.check_serial = false;
	t( "()->() context(pine) optional", "()->() context(pine)", "c_optional" );
	t( "()->() context(pine) optional","contextual<:pine, ()->() context(pine):>", "[1]->c_optional" );
	
	//incomplete functions can convert if they have the same incompleteness (used for literal defn)
	t( "()->$implicit", "()->$implicit", "" );
}

BOOST_AUTO_TEST_CASE( param_func ) {
	tester t;
	t.determine_flags = type_identifier::df_allow_param;
	t( "( : integer )->()", "( : )->()", "c_param_inst｢( : integer ) -> ( )｣" );

}

BOOST_AUTO_TEST_CASE( floats ) {
	tester t;
	t( "float", "float", "" );
	t( "float high", "float", "c_promote(high float)" );
	t( "float", "rational", "c_fixate(normal float)" );
	t( "float", "integer 10bit", "c_promote(normal integer_to_float)" );
	t( "float low", "integer", cerr::precision_loss );
	t( "float", "binary 10bit", "c_promote(normal binary_to_float)" );
	t( "float low", "binary", cerr::precision_loss );
	
	//unfortunately these are tied directly to the standard float sizes(though 64-bit is almost universal)
	t( "float", "binary 53bit", "c_promote(normal binary_to_float)" );
	t( "float", "binary 54bit", cerr::precision_loss );
	t( "float", "integer 54bit", "c_promote(normal integer_to_float)" );
	t( "float", "integer 55bit", cerr::precision_loss );
}

BOOST_AUTO_TEST_CASE( lvalue_strict )
{
	tester t;
	t( "float lvalue", "integer lvalue", cerr::mismatch_fundamental );
	t( "lvalue [ : integer ]", "lvalue [ : integer 2bit ]", cerr::mismatch_data_size );
	t( "lvalue [ : float shared ]", "lvalue [ : float ]", cerr::mismatch_shared );
	t( "lvalue shared []", "lvalue shared []", "" );
	
	//contrast to retained, optional wraps shared so you can't drop the shared here
	t( "integer optional lvalue", "integer optional shared lvalue", cerr::mismatch_shared );
}

BOOST_AUTO_TEST_CASE( lvalue_retain )
{
	tester t;
	t( "float lvalue", "lvalue optional float", cerr::mismatch_optional );
	t( "float lvalue", "lvalue shared float", "c_deref" );
	t( "integer lvalue", "lvalue integer optional shared", cerr::mismatch_optional );
	t( "integer shared lvalue", "integer shared optional lvalue", cerr::mismatch_optional );
}

BOOST_AUTO_TEST_CASE( tuples )
{
	tester t;
	t( "[]", "[]", "" );
	t( "[:[]]", "[:[]]", "" );
	t( "[:integer,:float]", "[:integer,:float]", "" );
	t( "[:integer]", "[:integer lvalue]", "c_pack {[0]->c_drop_lvalue}" );
	t( "[:integer,:float]", "[:integer lvalue,:float shared]", 
		"c_pack {[0]->c_drop_lvalue [1]->c_deref}" );
	t( "[:integer,:float]", "[:integer,:float shared]",
		"c_pack {[0] [1]->c_deref}" );
	t( "[:[:integer,:float]]", "[:[:integer,:float shared]]",
		"c_pack {[0]->c_pack {[0] [1]->c_deref}}" );
		
	//packing styles
	t( "[] abi_pack", "[]", "c_pack(abi) {}" );
	t( "[:integer]", "[:integer] byte_pack", "c_pack {[0]}" );
}

BOOST_AUTO_TEST_CASE( tuple_names )
{
	tester t;
	t( "[ a : integer ]", "[ :integer ]", "c_pack {a=[0]}" );
	t( "[ : integer]", "[ a : integer ]", cerr::mismatch_field_name );
}

BOOST_AUTO_TEST_CASE( value_lists )
{
	tester t;
 	t( "(:integer)", "(:integer)", "" );
 	
	t( "integer", "(:(:integer))", "c_collapse->c_collapse" );
	t( "integer", "(:integer)", "c_collapse" );
	t( "integer", "(:integer lvalue shared)", "c_collapse->c_drop_lvalue->c_deref" );

 	t( "(:integer,:float)", "(:(:(:integer),:float))", "c_collapse->c_pack_value {[0]->c_collapse [1]}" );
 	t( "(:(:integer),:float)", "(:(:integer),:float)", "" );
 	t( "(:(:integer),:float)", "(:(:(:integer),:float))", "c_collapse" );

 	//NOTE: The optimization that makes the shorted form is removed from type_converter. I'd to
 	//avoid that the "typer" and further actually require that optimization.
	//t( "(:integer)", "(:(:integer))", "c_collapse" );
	t( "(:integer)", "(:(:integer))", "c_pack_value {c_collapse->c_collapse}" );
 	//t( "(:(:integer),:float)", "(:(:(:integer)),:float)", "c_pack_value {[0]->c_collapse [1]}" );
 	t( "(:(:integer),:float)", "(:(:(:integer)),:float)", "c_pack_value {[0]->c_pack_value {c_collapse->c_collapse} [1]}" );
 	
 	//1-value-list collapsing is one-direction only
 	t( "(:integer)", "integer", cerr::requires_value_list );
 	
 	//lvalue should be preserved
 	t( "integer lvalue", "(:integer lvalue)", "c_collapse" );
 	t( "integer lvalue", "(:(:integer lvalue))", "c_collapse->c_collapse" );
}

BOOST_AUTO_TEST_CASE( invalid )
{
	tester t;
 	t( "integer", "float", cerr::mismatch_fundamental );
 	t( "boolean", "[]", cerr::mismatch_fundamental );
 	t( "integer shared", "integer", cerr::requires_shared );
 	t( "integer lvalue", "integer", cerr::requires_lvalue );
 	t( "[:integer]", "[:integer,:float]", cerr::mismatch_size );
 	t( "()->() raw", "()->()", cerr::mismatch_function_access );
 	t( "integer lvalue", "integer shared", cerr::requires_lvalue );
 	t( "integer lvalue", "float lvalue", cerr::mismatch_fundamental );
 	t( "[:integer] shared", "[:integer shared] shared", cerr::mismatch_shared );
 	
 	
	t( "()->()", "(:integer)->()", cerr::mismatch_size );
	t( "()->()", "()->(:integer)", cerr::mismatch_size );
	t( "(:integer)->()", "()->()", cerr::mismatch_size );
	t( "()->(:integer)", "()->()", cerr::mismatch_size );
	t( "(:integer)->()", "(:integer shared)->()", cerr::mismatch_shared );
}

BOOST_AUTO_TEST_CASE( assign )
{
	tester t( &type_converter::convert_to_lvalue );
	t( "integer lvalue", "integer", "" );
	t( "integer lvalue", "integer lvalue", "c_drop_lvalue" );
	t( "integer lvalue", "integer shared", "c_deref" );
	t( "integer", "integer", cerr::requires_lvalue );
	t( "(:integer lvalue)", "(:integer)", "" );
	t( "(:integer)", "(:integer)", cerr::requires_lvalue );
	t( "(:integer)->(:integer) lvalue", "()->() raw", cerr::mismatch_size );
	
	t( "integer optional lvalue", "integer", "c_optional" );                                                                                                                                                                                                                            
 	t.add_defined_type( "point", "[ x : float, y : float ]" );
 	t( "point lvalue", "point", "" );
	t( "point lvalue optional", "point", "c_optional" );
	
	t( "integer value_ptr lvalue", "integer value_ptr", "" );
	t( "integer value_ptr lvalue", "integer", cerr::requires_value_ptr );
	t( "integer lvalue", "integer value_ptr", "c_deref" );
	t( "integer lvalue", "integer value_ptr lvalue", "c_drop_lvalue->c_deref" );
	
	//with lossy
	t( "integer 8bit lvalue", "integer lossy", "c_loss(8bit integer)" );
	t( "integer 8bit lvalue", "integer lossy lvalue", "c_drop_lvalue->c_loss(8bit integer)" );
}                                                                         

BOOST_AUTO_TEST_CASE( to_refer )
{
	tester t( &type_converter::convert_to_refer );
	t( "integer shared", "integer shared", "c_deref" );
	t( "float shared", "float", "" );
	t( "boolean", "boolean", cerr::requires_refer );
	t( "integer shared", "integer shared lvalue", "c_drop_lvalue->c_deref" );
	
	t( "integer value_ptr", "integer value_ptr", "c_deref" );
	t( "float value_ptr", "float", "" );
	t( "boolean", "boolean", cerr::requires_refer );
	t( "integer value_ptr", "integer value_ptr lvalue", "c_drop_lvalue->c_deref" );
	t( "integer value_ptr", "integer lvalue", "c_drop_lvalue" );
}

BOOST_AUTO_TEST_CASE( any_callable )
{
	tester t( &type_converter::convert_any_callable );
	t( "()->()", "" );
	t( "()->() lvalue", "c_drop_lvalue" );
	t( "()->() shared", "c_deref" );
	t( "( :(:integer)->(:boolean) )", "c_collapse" );
	t( "()->() optional", cerr::mismatch_optional );
	
	t( "( : [ ] selectable )", "c_collapse" );
	
	t( "integer", cerr::requires_function );
	t( "( :()->(), :()->() )", cerr::requires_function );

	t.add_class("pine");
	t.check_serial = false;
	t( "contextual<:pine, ()->() context(pine):>", "c_bind_context" );
	t( "contextual<:pine lvalue, ()->() context(pine):>", "c_bind_context" );
}

BOOST_AUTO_TEST_CASE( any_refer )
{
	tester t( &type_converter::convert_any_refer );
	t( "shared integer", "" );
	t( "(:shared integer)", "c_collapse" );
	t( "shared integer lvalue", "c_drop_lvalue" );
	t( "( : shared integer, : shared float )", "" );//no pack expected if all sub-parts are empty

	t( "value_ptr integer", "" );
	t( "(:value_ptr integer)", "c_collapse" );
	t( "value_ptr integer lvalue", "c_drop_lvalue" );
	t( "( : value_ptr integer, : value_ptr float )", "" );//no pack expected if all sub-parts are empty
	
	t( "value_ptr integer lvalue", "c_drop_lvalue" );
	
	t( "optional integer", cerr::requires_refer );
	t( "integer", cerr::requires_refer );
	t( "()", cerr::requires_shared ); //TODO: error is suspicious (expect refer instead)
}

BOOST_AUTO_TEST_CASE( any_lvalue )
{
	tester t( &type_converter::convert_any_lvalue );
	t( "integer lvalue", "" );
	t( "( : integer lvalue, : lvalue float)", "" );
	t( "( : integer lvalue )", "c_collapse" );
	
	t( "integer lvalue value_ptr", "" );
	
	t( "integer", cerr::requires_lvalue );
	t( "()", cerr::requires_lvalue );
}

BOOST_AUTO_TEST_CASE( remove_optional ) {
	tester t( &type_converter::remove_optional );
	t( "integer lvalue", "" );
	t( "integer lvalue optional", "c_deref_optional" ); //lvalue status is retained
	t( "integer optional shared", "c_deref_optional" );
	t( "integer optional lvalue", "c_deref_optional" );
	t( "rational 16bit optional", "c_deref_optional" ); //though an optional rational is not actually allowed
}

BOOST_AUTO_TEST_CASE( nonlvalue_optional )
{
	tester t( &type_converter::convert_nonlvalue_optional );
	t( "optional integer", "" );
	t( "shared optional integer", "" );
	t( "optional lvalue float", "c_drop_lvalue" );
	t( "(:optional boolean)", "c_collapse" );
	
	t( "integer", cerr::requires_optional );
}

BOOST_AUTO_TEST_CASE( any_scalar )
{
	tester t( &type_converter::convert_any_scalar );
	t( "integer", "" );
	t( "binary", "" );
	t( "optional integer", cerr::mismatch_optional );
	t( "shared float", "c_deref" );
	t( "(: shared integer)", "c_collapse->c_deref" );
	t( "integer value_ptr", "c_deref" );
	
	//rationals are also scalar
	t( "rational lvalue", "c_drop_lvalue" );
	
	//booleans are also scalar
	t( "boolean lvalue", "c_drop_lvalue" );
	
	//chars are also scalar
	t( "char literal", "" );
	
	//live cases causing problems
	t( "float lvalue", "c_drop_lvalue" );
	t( "integer lossy lvalue", "c_drop_lvalue" ); //retain lossy
}

BOOST_AUTO_TEST_CASE( unqualified )
{
	tester t( &type_converter::convert_unqualified );
	t( "integer", "" );
	t( "shared integer", "c_deref" );
	t( "( : float lvalue shared )",  "c_collapse->c_drop_lvalue->c_deref" );
}

BOOST_AUTO_TEST_CASE( defined )
{
	tester t;
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t.add_defined_type( "complex", "[ r: float, i : float ]" );
	
	//OPT: ideally this could just be "c_set_distinct(point)" since the parts are already compatible
	t( "point", "[ : float, : float ]", "c_pack {x=[0] y=[1]}->c_set_distinct(point)" );
	t( "point", "[ x : float, y : float ]", "c_set_distinct(point)" );
	t( "point", "[ a : float, b : float ]", cerr::mismatch_field_name );
	t( "point", "[ : integer 8bit, y:integer 7bit ]", 
		"c_pack {"
			"x=[0]->c_promote(normal integer_to_float) "
			"y=[1]->c_promote(normal integer_to_float)"
		"}->c_set_distinct(point)" ); //integer promotion
	
	t( "point", "complex", cerr::mismatch_defined );
	
 	t( "point", "point shared", "c_deref" );
 	t( "point", "point lvalue", "c_drop_lvalue" );
 	t( "point", "point shared lvalue", "c_drop_lvalue->c_deref" );
 	t( "point", "( :point lvalue )", "c_collapse->c_drop_lvalue" );
 	
 	t( "point", "[ : float, : float ] lvalue", "c_drop_lvalue->"
 		"c_pack {x=[0] y=[1]}->c_set_distinct(point)" );

	t( "[ a : point ]", "[ : [ : float, : float ] ]",
		"c_pack {a=[0]->c_pack {x=[0] y=[1]}->c_set_distinct(point)}" );
	
	t.add_defined_type( "rect", "[ ul: point, br: point ]" );
	t( "rect", "[ :point, :point ]", "c_pack {ul=[0] br=[1]}->c_set_distinct(rect)" );
	//OPT: also ideally this would be just be "c_set_distinct(rect)"
	t( "rect", "[:[x:float,:float],br:[:float,:float]]", 
		"c_pack {ul=[0]->c_pack {x=[0] y=[1]}->c_set_distinct(point) br=[1]->c_pack {x=[0] y=[1]}->c_set_distinct(point)}->c_set_distinct(rect)" );
}

BOOST_AUTO_TEST_CASE( integer_range ) {
	//exposed internally as ⁑integer_range
	tester t;
	t.add_defined_type( "range", "[ begin : integer, end : integer ]"  );
	
	t( "range", "[:integer,:integer]", "c_pack {begin=[0] end=[1]}->c_set_distinct(range)" );
	t( "range", "[:integer PBIT,:integer PBIT]", "c_pack {begin=[0] end=[1]}->c_set_distinct(range)" );
	
}


BOOST_AUTO_TEST_CASE( array )
{
	tester t;
	t.check_serial = false;
	t( "array｢integer｣", "array｢integer｣", "" );
	t( "array｢integer｣", "array｢integer｣ lvalue", "c_drop_lvalue" );
	t( "array｢integer｣", "array｢integer｣ literal", "c_drop_literal" );
	t( "raw_array｢char｣", "array｢char｣", "c_addr_of" );
	//TODO: t( "array｢integer｣", "array｢float｣", cerr::mismatch_fundamental );
}

BOOST_AUTO_TEST_CASE( promote )
{
	tester t;
	t( "integer", "integer 3bit", "c_promote(normal integer)" );
	t( "integer 60bit", "integer 11bit", "c_promote(60bit integer)" );
	t( "integer 128bit", "integer", "c_promote(128bit integer)" );
	t( "integer 200bit", "integer 200bit", "" );
	//deref must come first
	t( "integer 60bit", "shared integer 8bit", "c_deref->c_promote(60bit integer)" );
	
	//references to types cannot be promoted
	t( "shared integer", "shared integer 8bit", cerr::mismatch_data_size );
	t( "lvalue integer", "lvalue integer 8bit", cerr::mismatch_data_size );
	
	t( "binary", "binary 16bit", "c_promote(normal binary)" );
	t( "binary 100bit", "binary 73bit", "c_promote(100bit binary)" );
	
	//binary to signed (note that integer->binary is not a promotion, would be a truncation)
	t( "integer", "binary", cerr::range_loss );
	t( "integer 41bit", "binary 40bit", "c_promote(41bit binary_to_integer)" );
}

BOOST_AUTO_TEST_CASE( optional )
{
	tester t;
	t( "integer", "optional integer", cerr::mismatch_optional );
	t( "integer optional", "integer", "c_optional" );
	t( "integer optional", "integer optional", "" );
	
	t( "integer optional lvalue", "integer lvalue", cerr::requires_optional );
	t( "integer shared optional", "integer shared", "c_optional" );
	t( "integer optional", "integer shared", "c_deref->c_optional" );
	t( "integer optional", "integer shared optional", cerr::mismatch_shared );
	t( "integer optional", "integer shared", "c_deref->c_optional" );
	t( "integer optional", "integer shared lvalue", "c_drop_lvalue->c_deref->c_optional" );
	t( "integer shared optional", "integer shared lvalue", "c_drop_lvalue->c_optional" );
}

BOOST_AUTO_TEST_CASE( fixation )
{
	tester t;
	//NOTE: these tests are somewhat unfortunately tied to the platform size... may need to fix on limited platforms
	t( "integer", "rational 29bit", "c_fixate(normal integer)" );
	t( "optional integer 6bit", "rational 2bit", "c_fixate(6bit integer)->c_optional" );
	t( "binary", "rational 32bit", "c_fixate(normal binary)");
	t( "binary 32bit", "rational 32bit", "c_fixate(32bit binary)");
	t( "binary 32bit", "rational 32bit", "c_fixate(32bit binary)");
	
	t( "float", "rational", "c_fixate(normal float)" );
	
	t( "binary", "rational signed 15bit", cerr::mismatch_signed );
	
}

BOOST_AUTO_TEST_CASE( convert_optional )
{
	tester t;
	t( "optional integer", "rational 3bit optional convert_optional", "c_fixate(normal integer)" );
}

BOOST_AUTO_TEST_CASE( shared_or_unqualified )
{
	tester t( &type_converter::convert_shared_or_unqualified );
	t( "float", "" );
	t( "lvalue integer", "c_drop_lvalue" );
	t( "shared optional [ : integer ]", cerr::mismatch_optional );
}

BOOST_AUTO_TEST_CASE( selectable ) {
	tester t;
	t( "float", "[ : integer, : float ] selectable", "[1]" );
	t( "integer 20bit", "[ :integer 21bit, : integer, : integer 10bit ] selectable", "[2]->c_promote(20bit integer)" );
	t( "float optional", "[ : float, : rational ] selectable", "[0]->c_optional" );
	t( "()->(:integer)", "[ :()->(x:integer) raw, :()->(x:float) ] selectable", "[0]->c_bind_raw_func" );
	
	t( "float", "[ : integer, : float ] selectable lvalue", "c_drop_lvalue->[1]" );
		
	//if target is selectable source doesn't select
	t( "[ : integer, : float ] selectable", "[ : integer, : float ] selectable", "" );
	
	t ("(:integer)->() no_throw", "[ : ()->(), : (:integer)->() no_throw literal lvalue raw ] selectable", 
		"[1]->c_drop_lvalue->c_bind_raw_func->c_drop_literal" );
	
}

BOOST_AUTO_TEST_CASE( value_ptr ) {
	tester t;
	t( "float value_ptr", "float lvalue", "c_addr_of" );
	t( "integer value_ptr", "integer shared lvalue", "c_deref->c_addr_of" );
	
	t( "integer value_ptr", "integer value_ptr", "" );
	t( "integer value_ptr", "integer value_ptr lvalue", "c_drop_lvalue" );
	
	t( "integer value_ptr", "integer", cerr::requires_value_ptr );
}

BOOST_AUTO_TEST_CASE( array_ptr )
{
	tester t;
	t.check_serial = false;
	t( "raw_array<:float:>", "array<:float:>", "c_addr_of" );
	t( "raw_array<:float:>", "array<:float optional:>", cerr::invalid_array_cvt );
	t( "raw_array<:float:>", "array<:rational:>", cerr::invalid_array_cvt );
	t( "raw_array<:float:>", "shared lvalue array<:float:>", "c_drop_lvalue->c_deref->c_addr_of" ) ;
	
	t( "raw_array<:float:>", "float", cerr::mismatch_fundamental );
	
	t( "raw_array<:float:>", "raw_array<:float:>", "" );
}

/**
	Be aware in this test the recursive type is not actually valid in Leaf. The type
	system supports it, but it creates an infinitely sized structure! This isn't the
	concern of the type system, thus it should always work here.
*/
BOOST_AUTO_TEST_CASE( recursive )
{
	tester t;
	t.add_defined_type( "point", "[ x : integer, pt : point ]" );
	
	t( "point", "point", "" );
	t( "point optional", "(:point)", 
		"c_collapse->c_optional" );
	t( "point", "[ : integer, : point ] shared", 
		"c_deref->c_pack {x=[0] pt=[1]}->c_set_distinct(point)" );
	t( "point", "(:[ : integer, : point lvalue ])", 
		"c_collapse->c_pack {x=[0] pt=[1]->c_drop_lvalue}->c_set_distinct(point)" );
	t( "point", "[ : integer, : [ : integer, : point ] ]", 
		"c_pack {x=[0] pt=[1]->c_pack {x=[0] pt=[1]}->c_set_distinct(point)}->c_set_distinct(point)" );
}

BOOST_AUTO_TEST_CASE( lossy ) 
{
	tester t;
	t( "integer 8bit", "integer lossy", "c_loss(8bit integer)" );
	t( "float low", "float lossy", "c_loss(low float)" );
	t( "integer", "float lossy", "c_loss(normal float_to_integer)");
	t( "float low", "integer 128bit lossy", "c_loss(low integer_to_float)" );
	t( "integer", "binary lossy", "c_loss(normal binary_to_integer)" );
	t( "binary", "integer lossy", "c_loss(normal integer_to_binary)" );
	t( "integer 10bit", "integer 20bit lossy", "c_loss(10bit integer)" );

	t( "integer 8bit", "integer lossy", "c_loss(8bit integer)" );
	
	//no lossy conversion is needed here, normal fixation
	t( "float", "rational 91bit lossy", "c_fixate(normal float)" );
	//it doesn't seem to make sense to allow lossy rational conversion, either the rational is a valid
	//size or its guaranteed to be a loss (not likely wanted).
	t( "integer 30bit", "rational 40bit lossy", cerr::range_loss);
}

BOOST_AUTO_TEST_CASE( literal ) {
	tester t;
	t( "integer", "integer literal", "c_drop_literal" );
	t( "boolean literal", "boolean", cerr::requires_literal );
	t( "literal [ : integer ]", "literal [ : rational 8bit ]", "c_pack {[0]->c_fixate(normal integer)}" );
	t( "[ : integer, : boolean ]", "literal [ : integer, : boolean ]", "c_drop_literal" );
	
	t( "integer literal", "rational 2bit literal", "c_fixate(normal integer)" );
	//TODO: test for invalid types (top-level can only be literal if all parts are literal)
	//  literal [ : integer, : boolean literal ]
	
	//TODO: unclear still
 	//t.add_defined_type( "point", "[ x : float, y : float ]" );
 	//t( "point", "[ : float literal, : float literal ] : literal", "" );
	
}

BOOST_AUTO_TEST_CASE( retain_literal ) {
	tester t( &type_converter::convert_retain_literal );
	t( "integer", "integer literal", "" );
	t( "[ : integer, : boolean ]", "literal [ : integer, : boolean ]", "" );
	t( "[ : integer ]", "literal [ : rational 8bit ]", "c_pack {[0]->c_fixate(normal integer)}" );
}

BOOST_AUTO_TEST_CASE( contextual ) {
	tester t;
	t.check_serial = false; //loaded contexts will not be the same... hmmm
	t.add_class( "pine" );
	t( "()->()", "contextual<:pine, ()->() context(pine):>", "c_bind_context" );
	t( "()->()", "contextual<:pine lvalue, ()->() context(pine) lvalue:>", "c_bind_context" );
	
	t( "contextual<:pine lvalue, ()->() context(pine):>", 
		"contextual<:pine lvalue, ()->() context(pine) lvalue:>", "c_pack_contextual {[0] [1]->c_drop_lvalue}" );
		
	t( "()->() context(pine)", "contextual<:pine,()->() context(pine):>", "[1]" );
	t( "()->() context(pine)", "contextual<:pine,()->() context(pine) lvalue:>", "[1]->c_drop_lvalue" );
	
	t.add_class( "cone" );
	t( "()->() context(cone)", "contextual<:pine,()->() context(pine):>", cerr::mismatch_fundamental );
}

BOOST_AUTO_TEST_CASE( parametric ) {
	tester t;
	t.check_serial = false; //TODO: hmm...
	t.add_defined_type( "pine",  "<:T:> [:T]" );
	
	t( "pine<:integer:>", "pine<:integer:> lvalue", "c_drop_lvalue" );
	t( "pine<:integer:> shared", "pine<:integer:> lvalue shared", "c_drop_lvalue" );
	t( "pine<:integer:>", "pine<:integer 8bit:>", cerr::mismatch_params );
	
	auto cl = t.add_class( "cone" );
	cl.type->params.push_back( { "T" } );
	
	t( "cone<:integer:>", "cone<:integer:> lvalue", "c_drop_lvalue" );
	t( "cone<:integer:> shared", "cone<:integer:> lvalue shared", "c_drop_lvalue" );
	t( "cone<:integer:>", "cone<:integer 8bit:>", cerr::mismatch_params );
	
}

BOOST_AUTO_TEST_CASE( abi_var_args ) {
	tester t;
	
	t( "abi_var_args", "[ : integer ]", "c_make_compat｢abi_var_args｣" );
	t( "abi_var_args", "[ : char, : float ]", "c_make_compat｢abi_var_args｣" );
	t( "abi_var_args", "[ : boolean ] lvalue", "c_drop_lvalue->c_make_compat｢abi_var_args｣" );
	t( "abi_var_args optional", "[ : boolean ]", "c_make_compat｢abi_var_args｣->c_optional" );
	
	t( "abi_var_args", "integer", cerr::mismatch_fundamental );
}

BOOST_AUTO_TEST_CASE( string_tag ) {
	tester t;
	t( "string_tag", "string_tag", "" );
}

BOOST_AUTO_TEST_SUITE_END()
