/**
	Tests of get_qualified_id are just frozen versions of what was output to ensure it doesn't change.
*/

#define SHOW_DEBUG 0
#include "type_converter_common.hpp"

#include "platform/platform.hpp"
#include <cstdio>
#include <stdarg.h>

BOOST_AUTO_TEST_SUITE( type_identifier_determine )

struct tester : public type_converter_tester
{
	bool keep_variant;
	bool expand_defined;
	bool annotate;
	bool allow_param;
	
	tester() {
		keep_variant = false;
		expand_defined = false;
		annotate = false;
		allow_param = false;
	}
	tester( shared_ptr<scope> sc ) :
		type_converter_tester( sc ) {
		keep_variant = false;
		expand_defined = false;
		annotate = false;
		allow_param = false;
	}
	
	type_ref ex( char const * spec, char const * expect_fmt, ... ) {
		char expect[1024];
		va_list ap;
		va_start(ap, expect_fmt);
		std::vsprintf( expect, expect_fmt, ap );
		va_end(ap);
		
		return operator()( spec, expect );
	}

	unsigned get_flags() {
		return allow_param ? type_identifier::df_allow_param : 0;
	}
	
	type_ref operator()( char const* spec, char const * expect,
		char const * expect_id = nullptr ) {
		expect_test ex;
		ex.add( "Spec", spec );

		type_identifier::resolve_t result;
		try {
			auto s = source::create( spec, std::strlen(spec) );
			auto vts = node_converter::convert( intr_type_parser::parse( *s ) );
			result = type_identifier::determine( the_context(), vts, get_flags() );
			STATE_CHECK(result.okay(),result.explain_error());
		} catch(...) {
			ex.except();
		}
	
		auto get_str = [&]( type_ref ex ) {
			leaf::dump d( expand_defined ? dump::tf_expand_defined : 0 );
			d.set_keep_variant( keep_variant );
			d.set_validate(false);
			d.set_annotate(annotate);
			d.write( ex );
			return d.str();
		};
		ex.check_actual( get_str( result.result ), expect );
			
		if( expect_id ) {
			ex.check_actual( result.result.get_qualified_id(), ex.platform_fix(expect_id) );
		}
		
		//round-trip test
		try {
			auto nspec = type_identifier::get_spec( the_context(), result.result );
			result = type_identifier::determine( the_context(), nspec.result, get_flags() );
			STATE_CHECK( result.okay(), result.explain_error() );
		} catch( ... ) {
			ex.except();
		}
		ex.check_actual( get_str( result.result ), expect, "get_spec" );
		
		return result.result;
	}
	
	type_ref operator()( char const * spec, cerr expect ) {
		expect_test ex;
		ex.add( "Spec", spec );
		
		type_identifier::resolve_t result;
		try {
			auto vt_spec = parse_spec(spec);
			result = type_identifier::determine( the_context(), vt_spec, get_flags() );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( result.error == expect, result.explain_error(), "wrong error" );
		return result.result;
	}
};

BOOST_AUTO_TEST_CASE( basic ) {
	tester t;
	t( "integer ", "integer", "$I%PBIT$V" );
	t( "float lvalue", "float lvalue", "$F%PBIT$LV" );
	t( "boolean", "boolean", "$B$V" );
	t( "shared integer", "integer shared", "$I%PBIT$S" );
	t( "float normal", "float" );
	t( "integer 17bit", "integer 17bit", "$I17$V" );
	t( "rational optional", "rational optional", "$R$OV" );
	t( "binary", "binary" );
	t( "binary 90bit", "binary 90bit", "$X90$V" );
	
	t( "no_such_symbol", cerr::unknown_identifier );
	//t( "", cerr::missing_intrinsic ); //TODO: uncertain, is this a param type?
	//t( "lvalue", cerr::missing_intrinsic ); //TODO: ^
}

BOOST_AUTO_TEST_CASE( literal ) {
	tester t;
	t( "literal rational", "rational literal" );
	t( "literal optional value_ptr boolean", "boolean literal optional value_ptr" );
	
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t( "point literal", "point literal", "point$TV" );
}

BOOST_AUTO_TEST_CASE( alt_names ) {
	tester t;
	t( "octet", "binary 8bit" );
}

BOOST_AUTO_TEST_CASE( native_types ) {
	tester t;
	t.keep_variant = true;
	t.ex( "abi_int", "integer %dbit", platform::abi_int_size );
	t.ex( "abi_short", "integer %dbit", platform::abi_short_size );
	t.ex( "abi_long", "integer %dbit", platform::abi_long_size );
	t.ex( "abi_long_long", "integer %dbit", platform::abi_long_long_size );
	//NOTE: this is not likely truly platform independent
	t( "abi_float", "float low" );
	t( "abi_double", "float normal" );
	t( "abi_long_double", "float high" );
	
	//TODO: these are temporary until appropriate types appear
	t.ex( "abi_char", "binary %dbit", platform::abi_char_size );
	t.ex( "abi_ptr", "binary %dbit", platform::abi_ptr_size );
}

BOOST_AUTO_TEST_CASE( tuple ) {
	tester t;
	t( "[:integer,:float]",
		"[ : integer, : float ]" );
	t( "[]", "[ ]" );
	t( "[:]", cerr::unresolved_param );
	t( "[:[:integer],:[],:[:float,:integer]]",
		"[ : [ : integer ], : [ ], : [ : float, : integer ] ]" );
		
	t( "abi_pack [ : integer, : float]", "[ : integer, : float ] abi_pack", "$Ta<=$I64$V,=$F64$V>$V" );
	t( "[ ] abi_pack", "[ ] abi_pack" );
	
	t.allow_param = true;
	t( "[:]", "[ : ]" );
}

BOOST_AUTO_TEST_CASE( tuple_deco ) {
	tester t;
	t( "lvalue []", "[ ] lvalue" );
	
	t( "selectable [ :integer, : float ]", "[ : integer, : float ] selectable" );
}

BOOST_AUTO_TEST_CASE( value_list ) {
	tester t;
	t( "()" , "( )" );
	t( "(:integer)", "( : integer )" );
	t( "(a:integer)", "( a : integer )" );
	//t( "(a)", cerr::okay ); //TODO: undecided if this can be parsed as such...
}

BOOST_AUTO_TEST_CASE( signature ) {
	tester t;
	/*t( "(a:integer)->()", "( a : integer ) -> ( )" );
	t( "(a:float,)->()", "( a : float ) -> ( )" );
	t( "(:integer,:integer)->()", "( : integer, : integer ) -> ( )" );
	t( "()->() raw", "( ) -> ( ) raw" );
	
	t( "()->$implicit", "( ) -> $implicit" );*/
	
	//back reference
	t( "(x:integer,y:type_infer(x))->()", "( x : integer, y : integer ) -> ( )" );
}

BOOST_AUTO_TEST_CASE( signature_param ) {
	tester t;
	t.allow_param = true;
	
	t.add_defined_type( "p1", "<:T:> [:T]" );
	auto q = t( "( q )->()", "( q : ) -> ( )" ); //TODO: this should fail if allow_param == false
	BOOST_CHECK( extant_level::parametric == q.get_extant() );

	q = t( "( q : p1<:integer:> )->()", "( q : ^p1｢ integer ｣ ) -> ( )" );
	BOOST_CHECK( extant_level::complete == q.get_extant() );

	q = t( "( q : p1<:shared:> )->()", "( q : prototype(p1)｢ shared ｣ ) -> ( )" );
	BOOST_CHECK( extant_level::parametric == q.get_extant() );
	
	q = t( "( q : p1<:any:> )->()", "( q : prototype(p1)｢ any ｣ ) -> ( )" );
	BOOST_CHECK( extant_level::parametric == q.get_extant() );
	
	q = t( "( q : array<:any:> )->()", "( q : prototype(array)｢ any ｣ ) -> ( )" );
	BOOST_CHECK( extant_level::parametric == q.get_extant() );
}

BOOST_AUTO_TEST_CASE( array_value_list ) {
	tester t;
	t.allow_param = true;
	
	t( "( x : array<:any:>, : type_infer(x#0) )", 
		"( x : prototype(array)｢ any ｣, : type_infer( x#(0/1) ) )" );
}

BOOST_AUTO_TEST_CASE( lossy ) {
	tester t;
	t( "integer lossy", "integer lossy" );
}

//ensure that these fail
BOOST_AUTO_TEST_CASE( broken ) {
	tester t;
	//only one fundamental allowed
	t( "integer float", cerr::redundant_spec ); //TODO: specified: cerr::redundant_fundamental ); 
	
	//these are disallowed due to reserved words, though they are in a different namespace
	//TODO: test_noparse( "{integer}" )
	//TODO: test_noparse( "{integer,float}" )
}

BOOST_AUTO_TEST_CASE( array ) {
	tester t;
	t( "array<:integer:>", "^array｢ integer ｣" );
 	t( "array<:shared char:>", "^array｢ char shared ｣", "$A<$C$S>$V" );
 	t( "shared array<:array<:float shared:>:>", "^array｢ ^array｢ float shared ｣ ｣ shared" );
 	t( "[:array<:[:integer,:float]:>]", "[ : ^array｢ [ : integer, : float ] ｣ ]" );
 	
 	t( "raw_array<:integer:>", "^raw_array｢ integer ｣" );
 	t( "raw_array<:raw_array<:integer:> :>", "^raw_array｢ ^raw_array｢ integer ｣ ｣" );
}

BOOST_AUTO_TEST_CASE( type ) {
	tester t;
	t( "type<:integer:>", "type｢ integer ｣" );
	t( "type<:type<:array<:float:>:>:>", "type｢ type｢ ^array｢ float ｣ ｣ ｣" );
}

BOOST_AUTO_TEST_CASE( classes ) {
	tester b;
	auto ctx = b.add_class( "happy" ).context;
	
	tester t( ctx );
	t( "()->() member", "( ) -> ( ) context(happy lvalue) unbound" );
	
	//only the outermost function should be implicitly a member function
	t( "()->( : ()->() ) member", "( ) -> ( : ( ) -> ( ) ) context(happy lvalue) unbound",
		"$Cu<$V<>,$V<=$Cb<$V<>,$V<>,>$V>,happy$LV>$V" );
}

BOOST_AUTO_TEST_CASE( classes_direct ) {
	tester t;
	t.add_class( "cone" );
	t( "()->() context(cone)", "( ) -> ( ) context(cone) unbound" );
}

BOOST_AUTO_TEST_CASE( alias ) {
	tester t;
	t.add_type_alias( "sinter", "integer 8bit" );
	t( "sinter", "integer 8bit" );
	t( "shared sinter", "integer 8bit shared" );
	//TODO: t( "sinter float", cerr::redundant_spec ); //cerr::redundant_fundamental );
	t( "[ : sinter ]", "[ : integer 8bit ]" );
	
	t.add_type_alias( "ptr", "optional shared" );
	t( "float ptr", "float optional shared" );
	
	t.add_type_alias( "point", "[ x : integer, y : integer ]" );
	t( "point", "[ x : integer, y : integer ]" );
	
	t( "[ : sinter, pt : point ] ptr", "[ : integer 8bit, pt : [ x : integer, y : integer ] ] optional shared" );
}


BOOST_AUTO_TEST_CASE( alias_distinct ) {
	tester t;
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t.add_type_alias( "point_shared", "point shared" );
	t( "point_shared", "point shared" );
}

#if 0
//TODO: should rework modules first, a bit sloppy
BOOST_AUTO_TEST_CASE( scoped )
{
	tester t;
	auto ns = scope::create(scope_style_t::declarative);
	ns->set_base( t.sc, scope::bt_namespace );
	ns->set_name( "branch" );
	t.add_defined_type( "cone", "[ a : integer ]" );
	t( "branch.cone", "branch.cone" );
}
#endif

/*
	Completeness indicates that a type has all its parts identified, even if they are unknown. That is,
	unknown indicates a parametric type, not something which is simply not yet available (like a missing
	symbol).
*/
BOOST_AUTO_TEST_CASE( complete ) {	
	tester t;
	t.allow_param = true;
	auto u = t("","");
	BOOST_CHECK( extant_level::parametric == u.get_extant() );
	
	u = t("(x)->()", "( x : ) -> ( )");
	BOOST_CHECK( extant_level::parametric == u.get_extant() );
	
	u = t("integer","integer");
	BOOST_CHECK( extant_level::complete == u.get_extant() );
}

BOOST_AUTO_TEST_CASE( parametric ) {
	tester t;
	t.allow_param = true;
	t.add_defined_type( "p1", "<:T:> [:T]" );
	t.add_defined_type( "p2", "<:T:> [:T shared]" );
	t.add_type_alias( "tup", "<:T, Q:> [ : T, : Q ]" );
	
	//t( "p1<lvalue>", cerr::unresolved_param ); //hmmm
	
	t.allow_param = true;

	//t( "<T> T", "< T > %T" ); //Not a valid parametric type, would require a special "distinct" type
	t( "<:T:> [:T]", "｢ T ｣ [ : %T ]" );
	t( "<:T,Q:> [:T,a:[:Q] shared]", "｢ T, Q ｣ [ : %T, a : [ : %Q ] shared ]" ); 
	//t( "<T,Q> [:T,a:[:Q shared]]", "< T, Q > [ : %T, a : [ : %Q shared ] ]" ); //TODO: I'm not sure this should be allowed
	
	t.expand_defined = true;
	//t( "p1", "p1" ); //TODO: should be an error?
	//TODO: there's a flip-flopping on the @export labels below, it's not clear if its'a dump issue or typing issue
	t( "p1<:integer:>", "@export typedef ^p1｢ integer ｣ = [ : integer ]" );
	t( "p1<:float optional:>", "@export typedef ^p1｢ float optional ｣ = [ : float optional ]" );
	
	t( "p2<:integer:>", cerr::redundant_spec );
	//t( "p2<float lossy>", "@export type ^p2< float lossy > = [ : float lossy shared ]" );
	//t( "p2<binary shared>", "@export type ^p2< binary shared > = [ : binary shared ]" ); //TODO: uncertain, dobule spec "shared"
	
	t.expand_defined = false;
	t( "tup<:char,boolean:>", "[ : char, : boolean ]" );
	t( "tup<:p1<:float:>,integer:>", "[ : ^p1｢ float ｣, : integer ]" );
	
	t.annotate = true;
	t( "p1<:shared:>", "⦃tr-spec: prototype(p1)｢ shared ｣⦄" );
	t( "array<:shared:>", "⦃tr-spec: prototype(array)｢ shared ｣⦄" );
}

BOOST_AUTO_TEST_CASE( contextual ) {
	tester t;
	t( "contextual<:integer,integer:>", "contextual｢ integer, integer ｣" );
	
	t.add_class( "pine" );
	t( "()->() context(pine)", "( ) -> ( ) context(pine) unbound" );
	t( "contextual<:pine,()->() context(pine):>", "contextual｢ pine, ( ) -> ( ) context(pine) unbound ｣" );
	//TODO: t( "contextual<pine,()->() context(pine) unbound>", "contextual< pine, ( ) -> ( ) context(pine) unbound >" );
}

BOOST_AUTO_TEST_CASE( class_parametric ) {
	tester t;
	auto cl = t.add_class( "cone" );
	cl.type->params.push_back( { "T" } );
	
	auto pcl = t.add_class( "pine" );
	pcl.type->params.push_back( { "R" } );
	pcl.type->params.push_back( { "S" } );
	
	//t( "cone", "" ); //will need to be done eventually to refer to parametric type as a whole
	t( "cone<:integer:>", "^cone｢ integer ｣" );
	t( "pine<:cone<:float:>,pine<:integer,float:>:>", "^pine｢ ^cone｢ float ｣, ^pine｢ integer, float ｣ ｣" );
	
	t( "cone<:any:>", cerr::unresolved_param );
	
	t.allow_param = true;
	auto q = t("(x : cone<:any:> )->()", "( x : prototype(cone)｢ any ｣ ) -> ( )");
	BOOST_CHECK( extant_level::parametric == q.get_extant() );	
}

BOOST_AUTO_TEST_CASE( abi_var_args ) {
	tester t;
	t( "abi_var_args", "abi_var_args" );
}

BOOST_AUTO_TEST_CASE( expression ) {
	tester t;
	//t("type_infer(false) optional", "boolean"); //TODO: requires an intrinsic filter, or type_infer makes all extrinsics optional
 	t("type_infer(1)", "integer" );
 	t("type_infer(1+2.5)", "float" );
	t("type_infer(false)", "boolean");
	
	t.add_var( "pine", "integer" );
	t( "[ : type_infer(pine) ]", "[ : integer ]" );
	
	t.add_var( "floaty", "float" );
	t( "type_infer(floaty)", "float" );
}

BOOST_AUTO_TEST_CASE( typedef_optional ) {
	tester t;
	t.expand_defined = true;
	
	t.add_defined_type( "add_opt", "｢T｣ [value : optional strip_optional｢T｣]" );
	t( "add_opt｢integer｣", "@export typedef ^add_opt｢ integer ｣ = [ value : integer optional ]" );
	t( "add_opt｢optional integer｣", "@export typedef ^add_opt｢ integer optional ｣ = [ value : integer optional ]" );
}

BOOST_AUTO_TEST_CASE( strip_optional ) {
	tester t;
	
	t( "strip_optional｢integer｣", "integer" );
	t( "strip_optional｢optional integer｣", "integer" );
}

BOOST_AUTO_TEST_SUITE_END()