#define SHOW_DEBUG 0
#include "type_converter_common.hpp"

BOOST_AUTO_TEST_SUITE( type_identifier_infer )

struct tester_infer : public type_converter_tester {
	bool init_refer;
	bool allow_param;
	
	tester_infer() {
		init_refer = false;
		allow_param = false;
	}
	tester_infer( shared_ptr<scope> sc ) :
		type_converter_tester( sc ) {
		init_refer = false;
		allow_param = false;
	}
	
	type_ref operator()( char const * spec, char const * assign, 
		char const * expect ) {
		expect_test ex;
		ex.add( "Spec", spec );
		ex.add( "Assign", assign );
		
		type_identifier::resolve_t result;

		unsigned flags = allow_param ? type_identifier::df_allow_param : 0;
		auto vt_spec = parse_spec( spec );
		auto vt_assign = determine_type( assign, flags );

		try {
			result = type_identifier::infer( the_context(), *vt_spec, vt_assign, init_refer );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( result.okay(), result.explain_error() );
		ex.check_actual( dump::get( result.result, 0 ), ex.platform_fix(expect) );
		return result.result;
	}
	
	void operator()( char const * spec, char const * assign, cerr expect ) {
		expect_test ex;
		ex.add( "Spec", spec );
		ex.add( "Assign", assign );
		
		auto vt_spec = parse_spec( spec );
		auto vt_assign = determine_type( assign );

		type_identifier::resolve_t result;
		try {
			result = type_identifier::infer( the_context(), *vt_spec, vt_assign, init_refer );
		} catch( ... ) {
			ex.except();
		}

		ex.check( result.error == expect, result.explain_error(), "wrong error" );
	}
	
	void operator()( char const * spec, type_ref vt_assign, 
		char const * expect ) {
		expect_test ex;
		ex.add( "Spec", spec );
		ex.add( "Assign", dump::get(vt_assign) );
		
		type_identifier::resolve_t result;
		
		auto vt_spec = parse_spec( spec );

		try {
			result = type_identifier::infer( the_context(), *vt_spec, vt_assign, init_refer );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( result.okay(), result.explain_error() );
		ex.check_actual( dump::get( result.result ), ex.platform_fix(expect) );
	}
	
};

struct tester_infer_spec : public type_converter_tester {
	
	void operator()( char const * spec, char const * assign, 
		char const * expect ) {
		expect_test ex;
		ex.add( "Spec", spec );
		ex.add( "Assign", assign );
		
		type_identifier::resolve_spec_t result;
		
		auto vt_spec = parse_spec( spec );
		auto vt_assign = determine_type( assign );

		try {
			result = type_identifier::infer_spec( the_context(), *vt_spec, vt_assign, false );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( result.okay(), result.explain_error() );
		ex.check_actual( dump::get( result.result ), ex.platform_fix(expect) );
	}
	
};

struct tester_infer_list : public type_converter_tester {
	bool init_refer;
	
	tester_infer_list() {
		init_refer = false;
	}
	
	void operator()( char const * assign, char const * expect ) {
		expect_test ex;
		ex.add( "Assign", assign );
		
		type_identifier::resolve_value_list_t result;
		
		auto vt_assign = determine_type( assign );

		try {
			result = type_identifier::infer_value_list( the_context(), vt_assign, init_refer );
		} catch( ... ) {
			ex.except();
		}
		
		ex.check( result.okay(), result.explain_error() );
		ex.check_actual( dump::get( result.result ), ex.platform_fix(expect) );
	}
};

BOOST_AUTO_TEST_CASE( basic ) {
	tester_infer t;
	
	t( "", "integer", "integer" );
	t( "shared", "float", "float shared" );
	t( "value_ptr", "float", "float value_ptr" );
	t( "float", "float shared", "float" );
	t( "integer", "integer 16bit", "integer" );
	t( "float", "float low", "float" );
	t( "", "integer lvalue 16bit", "integer 16bit" );
	t( "", "(:(:[:integer,:float]))", "[ : integer, : float ]" );
	t( "optional", "boolean", "boolean optional" );
	t( "boolean", "optional boolean", "boolean" );
	t( "binary lvalue", "binary", "binary lvalue" );
	t( "binary", "binary lvalue", "binary" );
	t( "binary 8bit", "binary 1bit lvalue", "binary 8bit" );
	
	t( "", "(:integer,:integer)", cerr::disallow_value_list );
	
	//since a fundamental is provider the modifiers are not taken on the outermost type
	t( "[:]", "[ : integer ] lvalue", "[ : integer ]" ); //!
	t( "[:]", "[ : integer ] shared", "[ : integer ]" );
}

BOOST_AUTO_TEST_CASE( rational ) {
	tester_infer t;
	
	t( "", "rational 19bit", "integer" );
	t( "", "rational", "float" );
	t( "integer 16bit", "rational 88bit", "integer 16bit" ); //note that conversion is not checked in inference (the assingment may not be possible)
}

BOOST_AUTO_TEST_CASE( functions ) {
	tester_infer t;
	t( "()->() raw", "()->()", "( ) -> ( ) raw" );
	//TODO: t( "()->()", "()->() raw", "( ) -> ( )" ); //hmmm, something must add non-raw explicitly
	t( "", "()->() unbound", "( ) -> ( ) unbound" ); //NOTE: this isn't really a valid type though, it needs a context!
	t( "()->()", "()->()", "( ) -> ( )" );
	
	//TODO: I suspect the default should still be "throw" and all no_throw can be converted
	t( "", "()->() no_throw", "( ) -> ( ) no_throw" );
}

//came from examples which weren't working
BOOST_AUTO_TEST_CASE( practical ) {
	tester_infer t;
	t( "float low", "float low", "float low" );
	t( "", "( a: optional boolean )->()", "( a : boolean optional ) -> ( )" );
	t( "", "( a: optional boolean value_ptr )->()", "( a : boolean optional value_ptr ) -> ( )" );
	t( "", "shared float", "float shared" );
	t( "", "value_ptr boolean", "boolean value_ptr" );
	
	t( "binary 5bit", "rational 3bit", "binary 5bit" );
	t( "integer 11bit", "binary 10bit", "integer 11bit" );
	//TODO: t( "float", "integer 15bit", "float" ); //unclear what should happen (error, or float)
}

BOOST_AUTO_TEST_CASE( infer_defined ) {
	tester_infer t;
	t.add_defined_type( "point", "[ x : float, y : float ]" );

	//TODO: really need a better way to handle user types
	t( "", "point", "point" );//"prototype(point)" );
	t( "optional", "point", "point optional" );// "prototype(point) optional" );
	t( "integer", "point", "integer" );
	t( "point", "integer", "point" );
	t( "point", "[:float,:float]", "point" ); //"prototype(point)" );
	
	t( "", "array<:point:>", "^array｢ point ｣" ); //"array< prototype(point) >" );
}

BOOST_AUTO_TEST_CASE( array ) {
	tester_infer t;
	type_ref q = t( "", "array<:integer:>", "^array｢ integer ｣" );
	BOOST_CHECK_EQUAL( intr_type::s_plain, q.extr().intr->style );
	t( "", "raw_array<:integer:>", "^raw_array｢ integer ｣" );
}

BOOST_AUTO_TEST_CASE( infer_init_refer ) {
	tester_infer t;
	t.init_refer = true;
	t( "", "integer", "integer shared" );
	
	//TODO: need that spec_type has shared set to "false" by default, not undefined
	//t( "integer", "integer", cerr::mismatch_shared ); //since non-specific shared is off by default
}

BOOST_AUTO_TEST_CASE( infer_tuple ) {
	tester_infer t;
	t( "[:]", "[ : integer ]", "[ : integer ]" );
	t( "[a,:]", "[:float,b:boolean]", "[ a : float, : boolean ]" );
	t( "[:optional,:128bit]", "[:integer,:integer]", "[ : integer optional, : integer 128bit ]" );
	
	t( "", "[ : rational 2bit ]", "[ : integer ]" );
	t( "optional", "[ : rational 2bit, : [ ] ]", "[ : integer, : [ ] ] optional" );
	
	t( "[ : integer, : float ] abi_pack", "[ : rational 4bit, : rational 8bit ]",
		"[ : integer, : float ] abi_pack" );
}

BOOST_AUTO_TEST_CASE( selectable ) {
	tester_infer t;
	//nothing can be inferred from selectables, so it becomes a 'determine' call
	t( "integer", "selectable [:float,:integer]", "integer" );
	//t( "", "selectable [:float,:integer]", "!lvalue" ); //TODO: lvalue nonsense!
	t( "selectable", "[ : float, : integer ]", "[ : float, : integer ] selectable" );
	
	//selectable can however be retained if marked in the target
	t( "selectable", "[ : float, : integer ] selectable", "[ : float, : integer ] selectable" );
	
	//not a useful (nor valid now) type, but infer doesn't care
	//t( "selectable", "integer", "integer selectable" ); //apparently it does care now
	t( "selectable", "[]", "[ ] selectable" );
}

BOOST_AUTO_TEST_CASE( contextual ) {
	tester_infer t;
	t.add_class( "cone" );
	t( "", "contextual<:cone,()->() context(cone):>", 
		"contextual｢ cone, ( ) -> ( ) context(cone) unbound ｣" );
	//TODO: unsafe! This was hidden before by the automatic binding, but we really need to
	//address "reference" types to clarify what happens
	//t( "", "contextual<cone lvalue,()->() context(cone)>", 
	//	"contextual< cone lvalue, ( ) -> ( ) context(cone) unbound >" );
	
	//TODO: t( "", "contextual<cone lvalue,()->() context(cone) lvalue>", 
	//	"contextual< cone lvalue, ( ) -> ( ) context(cone) unbound >" );
}

BOOST_AUTO_TEST_CASE( class_members ) {
	tester_infer b;
	auto ctx = b.add_class("happy").context;
	
	tester_infer t( ctx );
	t( "", "()->() member", "( ) -> ( ) context(happy lvalue) unbound" );
}

BOOST_AUTO_TEST_CASE( value_list ) {
	tester_infer_list t;
	
	t( "integer", "( : integer )" );
	t( "(:rational)", "( : float )" );
	t( "(:(:rational 12bit))", "( : integer )" );
	t( "( : rational, : boolean )", "( : float, : boolean )" );
	t( "(:( : rational, : boolean ))", "( : float, : boolean )" );
}

BOOST_AUTO_TEST_CASE( literal ) {
	tester_infer t;
	t( "", "integer literal", "integer" );
	t( "literal", "boolean", "boolean literal" );
	t( "literal", "[ : float, : integer ]", "[ : float, : integer ] literal" );
	t( "", "[ : rational 1bit, : rational 2bit ] literal", "[ : integer, : integer ]" );
	t( "8bit", "rational 4bit literal", "integer 8bit" );
	t( "octet", "rational 4bit", "binary 8bit" );
	
	t.add_defined_type( "point", "[ x : float, y : float ]" );
	t( "literal point", "[ : rational 1bit, : rational 2bit ] literal", "point literal" );
	t( "point", "[ : rational 1bit, : rational 2bit ] literal", "point" );
}

BOOST_AUTO_TEST_CASE( anon_context ) {
	tester_infer t;
	
	auto ct = make_shared<intr_type_function>();
	ct->args = t.parse_value_list("()");
	ct->returns = t.parse_value_list("()");
	ct->access = intr_type_function::a_bound;
	t( "", extr_type(ct), "( ) -> ( )" );
	
	auto closure = make_shared<intr_type_closure>();
	closure->style = intr_type::s_distinct;
	closure->what = intr_type_closure::w_share_context;
	closure->context = t.sc;
	ct->context = extr_type(closure);
	ct->access = intr_type_function::a_unbound;
	t( "", extr_type(ct), "( ) -> ( ) context(*distinct*) unbound" );
	
	auto child_scope = scope::create( scope_style_t::imperative );
	child_scope->set_base( t.sc, scope::bt_parent );
	auto child_closure = make_shared<intr_type_closure>();
	child_closure->style = intr_type::s_distinct;
	child_closure->what = intr_type_closure::w_share_context;
	child_closure->context = child_scope;
	ct->context = extr_type(child_closure);
	t( "", extr_type(ct), "( ) -> ( )" );
}

BOOST_AUTO_TEST_CASE( param_func ) {
	tester_infer t;
	t.allow_param = true;
	
	t( "(:integer)->()", "(:)->()","( : integer ) -> ( )" );
	t( "(x:integer)->()", "(x)->$implicit","( x : integer ) -> ( )" );
	t( "(:array<:integer:>)->()", "(:array<:any:>)->()", "( : ^array｢ integer ｣ ) -> ( )");
	t( "(:array<:integer:>)->()", "(:array<:any:>)-> $implicit", "( : ^array｢ integer ｣ ) -> ( )");
}

BOOST_AUTO_TEST_SUITE_END()