#include "util/number.hpp"
#include "core_types.hpp"

#include <boost/regex.hpp>

#define SHOW_DEBUG 0
#include "util/debug.hpp"

namespace util
{

number::number()
{
}

number::number( number const & o ) {
	value = o.value;
}

number::number(mpq_class ovalue) {
	value = ovalue;
}

number & number::operator=(number const & o) {
	value = o.value;
	return *this;
}

bool  number::operator==(number const & o ) const {
	return value == o.value;
}

namespace 
{
	template<typename T>
	mpz_class parse_bitz( T begin, T end, int base, int & digits )
	{
		digits = 0;
		if( begin == end )
			return mpz_class( 0 );
		//strip separators (just used for visual clarification)
		char buf[end - begin + 1];
		char * c = buf;
		while( begin != end ) {
			if( *begin != '_' ) {
				*c++ = *begin;
				digits++;
			}
			begin++;
		}
		*c = 0;
		return mpz_class( buf, base );
	}
} //eon anon

/*static*/ number number::parse_base( std::string const & str, unsigned & meta, int base )
{
	meta = pdf_is_exact;
	
	//this is intentionally far more lenient than the one in node_parser so this class is flexible
	//note that due to conflict on 'e' hex can't have exponent, for simplicity/clarity is thus only
	//accepted on exact decimal representation (which I think is a common expectation)
	static boost::regex re_decimal( "([+-]?)([0-9_]*)(\\.([0-9_]*))?([eE]([+-]?)([0-9]+))?" );
	static boost::regex re_other( "([+-]?)([0-9a-fA-F_]*)(\\.([0-9a-fA-F_]*))?" );
	bool is_decimal = base == 10;
	boost::match_results<std::string::const_iterator> what;
	if( !boost::regex_match( str, what, is_decimal ? re_decimal : re_other ) )
		PRE_FAIL( std::string( "invalid number: " + str ) );
		
	mpz_class part;
	mpq_class combine;
	bool negative = false;
	
	if( what[1].matched && *what[1].first == '-' )
		negative = true;
		
	int digits;
	if( what[2].matched )
		combine = parse_bitz( what[2].first, what[2].second, base, digits );
	
	if( what[3].matched ) {
		meta |= pdf_had_decimal;
		part = parse_bitz( what[4].first, what[4].second, base, digits );
		if( digits ) {
			mpz_class div(base);
			mpz_pow_ui( div.get_mpz_t(), div.get_mpz_t(), digits );
			combine += mpq_class( part, div );
		}
	}
	
	if( is_decimal && what[5].matched ) {
		meta |= pdf_had_exponent;
		mpz_class e = parse_bitz( what[7].first, what[7].second, base, digits );
		mpz_class ebase(base), ebase_exp;
		
		//TODO: check that "e" is constrained to unsigned long, check some limit as well
		mpz_pow_ui( ebase_exp.get_mpz_t(), ebase.get_mpz_t(), e.get_ui() );
		
		if( *what[6].first == '-' )
			combine /= ebase_exp;
		else
			combine *= ebase_exp;
	}

	if( negative )
		combine *= -1;
	combine.canonicalize();
	return number( combine );
}

std::string number::to_string(bool force_denom) const
{
	std::string ret = value.get_str(10);
	if( is_integral() && force_denom )
		ret += "/1";
	return ret;
}

bool number::is_negative() const
{
	return sgn(value) < 0;
}

bool number::is_integral() const
{
	//assume canonical form already
	return value.get_den() == 1;
}

unsigned number::signed_bits() const
{
	mpz_class ival( value );
	return unsigned( mpz_sizeinbase( ival.get_mpz_t(), 2 ) + 1 );
}

unsigned number::unsigned_bits() const
{
	mpz_class ival( value );
	return unsigned( mpz_sizeinbase( ival.get_mpz_t(), 2 ) );
}

bool number::fits_signed( unsigned bits ) const
{
	if( bits < 2 )
		return 0;
	bits--; //drop sign
	
	mpz_class base(2);
	mpz_class max;
	mpz_pow_ui( max.get_mpz_t(), base.get_mpz_t(), bits );
	
	mpz_class ival( value );
	if( sgn(ival) < 0 ) {
		return ival >= (-max-1);
	} else {
		return ival <= max;
	}
}

bool number::fits_unsigned( unsigned bits ) const
{
	mpz_class base(2);
	mpz_class max;
	mpz_pow_ui( max.get_mpz_t(), base.get_mpz_t(), bits );
	
	mpz_class ival( value );
	return ival <= max;
}

//TODO: All the below conversions do not support the large integer types, 64-bit, 128-bit (where possible)
//or long double!!! (well get_si does work on platforms where long is 64bit)
intmax_t number::as_int() const
{
	return mpz_class(value.get_num() / value.get_den()).get_si();
}

uintmax_t number::as_uint() const
{
	return mpz_class(value.get_num() / value.get_den()).get_ui();
}

double number::as_double() const
{
	double d = value.get_d();
	return d;
}

number number::neg() const {
	number r;
	mpq_neg( r.value.get_mpq_t(), value.get_mpq_t() );
	return r;
}

number number::abs() const {
	number r;
	mpq_abs( r.value.get_mpq_t(), value.get_mpq_t() );
	return r;
}

number number::operator+( number const & o ) const {
	number r;
	mpq_add( r.value.get_mpq_t(), value.get_mpq_t(), o.value.get_mpq_t() );
	return r;
}

number number::operator-( number const & o ) const {
	number r;
	mpq_sub( r.value.get_mpq_t(), value.get_mpq_t(), o.value.get_mpq_t() );
	return r;
}

number number::operator*( number const & o ) const {
	number r;
	mpq_mul( r.value.get_mpq_t(), value.get_mpq_t(), o.value.get_mpq_t() );
	return r;
}

number number::operator/( number const & o ) const {
	number r;
	mpq_div( r.value.get_mpq_t(), value.get_mpq_t(), o.value.get_mpq_t() );
	return r;
}

void number::write( std::ostream & out ) const {
#if 1
	int8_t neg = value.get_num() < 0;
	out.write( (char*)&neg, sizeof(neg) );
	
	size_t c;
	void * raw = mpz_export( nullptr, &c, 1, 1, 0, 0, value.get_num().get_mpz_t() );
	out.write( (char*)&c, sizeof(c) );
	out.write( (char*)raw, c );
	free(raw);
	
	raw = mpz_export( nullptr, &c, 1, 1, 0, 0, value.get_den().get_mpz_t() );
	out.write( (char*)&c, sizeof(c) );
	out.write( (char*)raw, c );
	free(raw);
#else
	//GMP: lacks serialization functions
	auto n = value.get_num().get_str(62);
	auto d = value.get_den().get_str(62);
	
	auto s = n.size();
	out.write( (char*)&s, sizeof(s) );
	out.write( (char*)&n[0], s );
	
	s = d.size();
	out.write( (char*)&s, sizeof(s));
	out.write( (char*)&d[0], s );
#endif
}

void number::read( std::istream & in ) {
#if 1
	mpz_class num, den;
	size_t s;
	std::vector<uint8_t> v;
	
	int8_t neg;
	in.read( (char*)&neg, sizeof(neg) );
	
	in.read( (char*)&s, sizeof(s) );
	v.resize(s);
	in.read( (char*)&v[0], s );
	mpz_import( num.get_mpz_t(), s, 1, 1, 0, 0, &v[0] );
	
	in.read( (char*)&s, sizeof(s) );
	v.resize(s);
	in.read( (char*)&v[0], s );
	mpz_import( den.get_mpz_t(), s, 1, 1, 0, 0, &v[0] );
	
	DEBUG( "NUM-READ", int(neg), num, den );
	value = mpq_class(num) / mpq_class(den);
	if( neg ) {
		value = -value;
	}
#else
	std::string n, d;
	size_t s;
	
	in.read( (char*)&s, sizeof(s) );
	n.resize(s);
	in.read( (char*)&n[0], s );
	
	in.read( (char*)&s, sizeof(s) );
	d.resize(s);
	in.read( (char*)&d[0], s );

	DEBUG( "NUM-READ", n, d );
	value = mpq_class(n,62) / mpq_class(d,62);
#endif
}


} //eon util
