#include "util/unicode.hpp"

/* 
	Reference: Unicode Section 3.9, Unicode Encoding Forms
	http://www.unicode.org/versions/Unicode6.2.0/ch03.pdf#G7404 
*/

namespace unicode {

#define BIN(b7,b6,b5,b4,b3,b2,b1,b0) ((b7<<7) | (b6<<6) | (b5<<5) | (b4<<4) | (b3<<3) | (b2<<2) | (b1<<1) | b0)
uint8_t const
	c_mask = BIN(1,1,0,0,0,0,0,0),
	c_val = BIN(1,0,0,0,0,0,0,0),
	
	b_mask_1 = BIN(1,0,0,0,0,0,0,0),
	b_val_1 = BIN(0,0,0,0,0,0,0,0),
	b_mask_2 = BIN(1,1,1,0,0,0,0,0),
	b_val_2 = BIN(1,1,0,0,0,0,0,0),
	b_mask_3 = BIN(1,1,1,1,0,0,0,0),
	b_val_3 = BIN(1,1,1,0,0,0,0,0),
	b_mask_4 = BIN(1,1,1,1,1,0,0,0),
	b_val_4 = BIN(1,1,1,1,0,0,0,0);
	
uint32_t const
	range_1 = BIN(0,1,1,1,1,1,1,1),
	range_2 = (BIN(0,0,0,0,0,1,1,1) << 8) | 0xFF,
	range_3 = (BIN(1,1,1,1,1,1,1,1) << 8) | 0xFF,
	range_4 = (BIN(0,0,0,1,1,1,1,1) << 16) | 0xFFFF;

size_t len8( uint8_t const * str, size_t size ) {
	size_t len = 0;
	size_t at = 0;
	while( at < size ) {
		uint8_t c = str[at];
		if( (c & b_mask_1) == b_val_1 ) {
			at+=1;
		} else if( (c & b_mask_2) == b_val_2 ) {
			at+=2;
		} else if( (c & b_mask_3) == b_val_3 ) {
			at+=3;
		} else if( (c & b_mask_4) == b_val_4 ) {
			at+=4;
		} else {
			STATE_FAIL( "invalid-utf-8" );
		}
		len++;
	}
	return len;
}

code_bit next8( uint8_t const * str, size_t max ) {
	uint32_t c = *str;
	
	auto clen = [&]( size_t len ) {
		STATE_CHECK( len <= max, "truncated-utf-8" )
		for( size_t i=1; i < len; i++ ) {
			STATE_CHECK( (str[i] & c_mask) == c_val, "invalid-utf-8-cont" )
		}
	};
	
	if( (c & b_mask_1) == b_val_1 ) {
		clen(1);
		return code_bit{ c & ~b_mask_1, 1 };
	} else if( (c & b_mask_2) == b_val_2 ) {
		clen(2);
		return code_bit{ ((c & ~b_mask_2) << 6) 
			| (str[1] & ~c_mask), 2 };
	} else if( (c & b_mask_3) == b_val_3 ) {
		clen(3);
		return code_bit{ ((c & ~b_mask_3) << 12) 
			| ((str[1] & ~c_mask) << 6)
			| (str[2] & ~c_mask), 3 };
	} else if( (c & b_mask_4) == b_val_4 ) {
		clen(4);
		return code_bit{ ((c & ~b_mask_4) << 18) 
			| ((str[1] & ~c_mask) << 12)
			| ((str[2] & ~c_mask) << 6)
			| ((str[3] & ~c_mask)), 4 };
	} else {
		STATE_FAIL( "invalid-utf-8" );
	}
}

codestr decode8( uint8_t const * str, size_t size ) {
	codestr out;
	
	size_t at = 0;
	
	while( at < size ) {
		auto bit = next8( str + at, size - at);
		out.push_back( bit.code );
		at += bit.count;
	}
	
	return out;
}

std::string encode8( uint32_t const * str, size_t size ) {
	std::string ret;
	auto end = str + size;
	while( str < end ) {
		uint32_t cp = *str++;
		if( cp <= range_1 ) {
			ret += char(cp);
		} else if( cp <= range_2 ) {
			ret += char( (cp>>6) | b_val_2 );
			ret += char( (cp & ~c_mask) | c_val );
		} else if( cp <= range_3 ) {
			ret += char( (cp>>12)  | b_val_3 );
			ret += char( ((cp>>6) & ~c_mask) | c_val );
			ret += char( (cp & ~c_mask) | c_val );
		} else if( cp <= range_4 ) {
			ret += char( (cp>>18)  | b_val_4 );
			ret += char( ((cp>>12) & ~c_mask) | c_val );
			ret += char( ((cp>>6) & ~c_mask) | c_val );
			ret += char( (cp & ~c_mask) | c_val );
		} else {
			STATE_FAIL( "invalid-unicode-value" );
		}
	}
	return ret;
}

void write( codestr const & str, std::ostream & out ) {
	size_t sz = str.size();
	out.write( (char*)&sz, sizeof(sz) );
	out.write( (char*)&str[0], sizeof(str[0]) * sz );
}

void read( codestr & str, std::istream & in ) {
	size_t sz;
	in.read( (char*)&sz, sizeof(sz) );
	str.resize(sz);
	in.read( (char*)&str[0], sizeof(str[0]) * sz );
}

} //eon unicode